/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "flitter.language.vm",
        "sources": [
            "src/flitter/language/vm.pyx"
        ]
    },
    "module_name": "flitter.language.vm"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flitter__language__vm
#define __PYX_HAVE_API__flitter__language__vm
/* Early includes */
#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include "pythread.h"
#include <stdint.h>

#include "Python.h"

#if PY_MINOR_VERSION >= 13
    static inline double perf_counter() {
        PyTime_t result;
        PyTime_PerfCounter(&result);
        return PyTime_AsSecondsDouble(result);
    }
#else
    static inline double perf_counter() {
        return _PyTime_AsSecondsDouble(_PyTime_GetPerfCounter());
    }
#endif
    
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/flitter/language/vm.pyx",
  "<stringsource>",
  "contextvars.pxd",
  "src/flitter/language/vm.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "src/flitter/model.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7flitter_5model_Vector;
struct __pyx_obj_7flitter_5model_Matrix33;
struct __pyx_obj_7flitter_5model_Matrix44;
struct __pyx_obj_7flitter_5model_Quaternion;
struct __pyx_obj_7flitter_5model_Node;
struct __pyx_obj_7flitter_5model_StateDict;
struct __pyx_obj_7flitter_5model_Context;
struct __pyx_obj_7flitter_8language_2vm_VectorStack;
struct __pyx_obj_7flitter_8language_2vm_Instruction;
struct __pyx_obj_7flitter_8language_2vm_InstructionVector;
struct __pyx_obj_7flitter_8language_2vm_InstructionNode;
struct __pyx_obj_7flitter_8language_2vm_InstructionStr;
struct __pyx_obj_7flitter_8language_2vm_InstructionTuple;
struct __pyx_obj_7flitter_8language_2vm_InstructionInt;
struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple;
struct __pyx_obj_7flitter_8language_2vm_InstructionLabel;
struct __pyx_obj_7flitter_8language_2vm_InstructionJump;
struct __pyx_obj_7flitter_8language_2vm_InstructionFunc;
struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt;
struct __pyx_obj_7flitter_8language_2vm_Program;
struct __pyx_obj_7flitter_8language_2vm_Function;
struct __pyx_obj_7flitter_8language_2vm_LoopSource;
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr;
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute;
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_opt_args_7flitter_5model_6Vector_match;
struct __pyx_opt_args_7flitter_5model_4Node_get;

/* "model.pxd":63
 *     cdef str as_string(self)
 *     cpdef uint64_t hash(self, bint floor_floats)
 *     cpdef object match(self, int64_t n=?, type t=?, default=?)             # <<<<<<<<<<<<<<
 *     cdef str repr(self)
 *     cdef Vector neg(self)
 */
struct __pyx_opt_args_7flitter_5model_6Vector_match {
  int __pyx_n;
  int64_t n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};

/* "model.pxd":212
 *     cpdef void append_vector(self, Vector nodes)
 *     cdef bint _equals(self, Node other)
 *     cpdef object get(self, str name, int n=?, type t=?, object default=?)             # <<<<<<<<<<<<<<
 *     cdef Vector get_vec(self, str name, Vector default)
 *     cdef Vector get_fvec(self, str name, int n, Vector default)
 */
struct __pyx_opt_args_7flitter_5model_4Node_get {
  int __pyx_n;
  int n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_call;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_append;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_func;

/* "flitter/language/vm.pxd":34
 * 
 * 
 * cdef enum OpCode:             # <<<<<<<<<<<<<<
 *     Add
 *     Append
 */
enum __pyx_t_7flitter_8language_2vm_OpCode {
  __pyx_e_7flitter_8language_2vm_Add,
  __pyx_e_7flitter_8language_2vm_Append,
  __pyx_e_7flitter_8language_2vm_Attributes,
  __pyx_e_7flitter_8language_2vm_BeginFor,
  __pyx_e_7flitter_8language_2vm_BranchFalse,
  __pyx_e_7flitter_8language_2vm_BranchTrue,
  __pyx_e_7flitter_8language_2vm_Call,
  __pyx_e_7flitter_8language_2vm_CallFast,
  __pyx_e_7flitter_8language_2vm_Ceil,
  __pyx_e_7flitter_8language_2vm_Compose,
  __pyx_e_7flitter_8language_2vm_Drop,
  __pyx_e_7flitter_8language_2vm_Dup,
  __pyx_e_7flitter_8language_2vm_EndFor,
  __pyx_e_7flitter_8language_2vm_Eq,
  __pyx_e_7flitter_8language_2vm_Floor,
  __pyx_e_7flitter_8language_2vm_FloorDiv,
  __pyx_e_7flitter_8language_2vm_Fract,
  __pyx_e_7flitter_8language_2vm_Func,
  __pyx_e_7flitter_8language_2vm_Exit,
  __pyx_e_7flitter_8language_2vm_Ge,
  __pyx_e_7flitter_8language_2vm_Gt,
  __pyx_e_7flitter_8language_2vm_Import,
  __pyx_e_7flitter_8language_2vm_IndexLiteral,
  __pyx_e_7flitter_8language_2vm_Jump,
  __pyx_e_7flitter_8language_2vm_Label,
  __pyx_e_7flitter_8language_2vm_Le,
  __pyx_e_7flitter_8language_2vm_Literal,
  __pyx_e_7flitter_8language_2vm_LiteralNode,
  __pyx_e_7flitter_8language_2vm_LiteralNodes,
  __pyx_e_7flitter_8language_2vm_LocalDrop,
  __pyx_e_7flitter_8language_2vm_LocalLoad,
  __pyx_e_7flitter_8language_2vm_LocalPush,
  __pyx_e_7flitter_8language_2vm_Lookup,
  __pyx_e_7flitter_8language_2vm_LookupLiteral,
  __pyx_e_7flitter_8language_2vm_Lt,
  __pyx_e_7flitter_8language_2vm_Mod,
  __pyx_e_7flitter_8language_2vm_Mul,
  __pyx_e_7flitter_8language_2vm_MulAdd,
  __pyx_e_7flitter_8language_2vm_Ne,
  __pyx_e_7flitter_8language_2vm_Neg,
  __pyx_e_7flitter_8language_2vm_Next,
  __pyx_e_7flitter_8language_2vm_Not,
  __pyx_e_7flitter_8language_2vm_Pos,
  __pyx_e_7flitter_8language_2vm_Pow,
  __pyx_e_7flitter_8language_2vm_Range,
  __pyx_e_7flitter_8language_2vm_Slice,
  __pyx_e_7flitter_8language_2vm_SliceLiteral,
  __pyx_e_7flitter_8language_2vm_Export,
  __pyx_e_7flitter_8language_2vm_Sub,
  __pyx_e_7flitter_8language_2vm_Tag,
  __pyx_e_7flitter_8language_2vm_TrueDiv,
  __pyx_e_7flitter_8language_2vm_Xor,
  __pyx_e_7flitter_8language_2vm_MAX
};

/* "flitter/language/vm.pxd":21
 * 
 *     cpdef VectorStack copy(self)
 *     cpdef void drop(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef void push(self, Vector vector)
 *     cpdef Vector pop(self)
 */
struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop {
  int __pyx_n;
  int64_t count;
};

/* "flitter/language/vm.pxd":159
 *     cpdef int64_t new_label(self)
 *     cpdef Program dup(self)
 *     cpdef Program drop(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef Program label(self, int64_t label)
 *     cpdef Program jump(self, int64_t label)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop {
  int __pyx_n;
  int64_t count;
};

/* "flitter/language/vm.pxd":195
 *     cpdef Program slice(self)
 *     cpdef Program slice_literal(self, Vector value)
 *     cpdef Program call(self, int64_t count, tuple names=?)             # <<<<<<<<<<<<<<
 *     cpdef Program call_fast(self, function, int64_t count)
 *     cpdef Program tag(self, str name)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_call {
  int __pyx_n;
  PyObject *names;
};

/* "flitter/language/vm.pxd":199
 *     cpdef Program tag(self, str name)
 *     cpdef Program attributes(self, tuple names)
 *     cpdef Program append(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef Program compose(self, int64_t count)
 *     cpdef Program begin_for(self, int64_t count)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_append {
  int __pyx_n;
  int64_t count;
};

/* "flitter/language/vm.pxd":205
 *     cpdef Program end_for(self)
 *     cpdef Program export(self, str name)
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=?)             # <<<<<<<<<<<<<<
 *     cpdef Program exit(self)
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_func {
  int __pyx_n;
  int64_t ncaptures;
};

/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */
struct __pyx_obj_7flitter_5model_Vector {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtab;
  int64_t length;
  PyObject *objects;
  double *numbers;
  double _numbers[16];
  uint64_t _hash;
};


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix33 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":128
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix44 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":178
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */
struct __pyx_obj_7flitter_5model_Quaternion {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":198
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */
struct __pyx_obj_7flitter_5model_Node {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtab;
  PyObject *kind;
  PyObject *_tags;
  PyObject *_attributes;
  int _attributes_shared;
  PyObject *_children;
};


/* "model.pxd":222
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */
struct __pyx_obj_7flitter_5model_StateDict {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtab;
  PyObject *_changed_keys;
  PyObject *_state;
};


/* "model.pxd":231
 * 
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 *     cdef readonly dict names
 *     cdef readonly set captures
 */
struct __pyx_obj_7flitter_5model_Context {
  PyObject_HEAD
  PyObject *names;
  PyObject *captures;
  struct __pyx_obj_7flitter_5model_StateDict *state;
  struct __pyx_obj_7flitter_5model_Node *root;
  PyObject *path;
  struct __pyx_obj_7flitter_5model_Context *parent;
  PyObject *exports;
  PyObject *errors;
  PyObject *logs;
  PyObject *references;
  PyObject *stack;
  PyObject *lnames;
  PyObject *dependencies;
};


/* "flitter/language/vm.pxd":14
 * 
 * 
 * cdef class VectorStack:             # <<<<<<<<<<<<<<
 *     cdef PyObject** vectors
 *     cdef int64_t top
 */
struct __pyx_obj_7flitter_8language_2vm_VectorStack {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack *__pyx_vtab;
  PyObject **vectors;
  int64_t top;
  int64_t size;
  int64_t max_size;
};


/* "flitter/language/vm.pxd":90
 * 
 * 
 * cdef class Instruction:             # <<<<<<<<<<<<<<
 *     cdef readonly OpCode code
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_Instruction {
  PyObject_HEAD
  enum __pyx_t_7flitter_8language_2vm_OpCode code;
};


/* "flitter/language/vm.pxd":94
 * 
 * 
 * cdef class InstructionVector(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly Vector value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionVector {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *value;
};


/* "flitter/language/vm.pxd":98
 * 
 * 
 * cdef class InstructionNode(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly Node value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionNode {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  struct __pyx_obj_7flitter_5model_Node *value;
};


/* "flitter/language/vm.pxd":102
 * 
 * 
 * cdef class InstructionStr(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly str value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionStr {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *value;
};


/* "flitter/language/vm.pxd":106
 * 
 * 
 * cdef class InstructionTuple(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionTuple {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *value;
};


/* "flitter/language/vm.pxd":110
 * 
 * 
 * cdef class InstructionInt(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionInt {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t value;
};


/* "flitter/language/vm.pxd":114
 * 
 * 
 * cdef class InstructionIntTuple(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t ivalue
 *     cdef readonly tuple tvalue
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t ivalue;
  PyObject *tvalue;
};


/* "flitter/language/vm.pxd":119
 * 
 * 
 * cdef class InstructionLabel(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t label
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionLabel {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t label;
};


/* "flitter/language/vm.pxd":123
 * 
 * 
 * cdef class InstructionJump(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t label
 *     cdef readonly int64_t offset
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionJump {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t label;
  int64_t offset;
};


/* "flitter/language/vm.pxd":128
 * 
 * 
 * cdef class InstructionFunc(InstructionJump):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionFunc {
  struct __pyx_obj_7flitter_8language_2vm_InstructionJump __pyx_base;
  PyObject *name;
  PyObject *parameters;
  int64_t ncaptures;
};


/* "flitter/language/vm.pxd":134
 * 
 * 
 * cdef class InstructionObjectInt(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly object obj
 *     cdef readonly int64_t value
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *obj;
  int64_t value;
};


/* "flitter/language/vm.pxd":139
 * 
 * 
 * cdef class Program:             # <<<<<<<<<<<<<<
 *     cdef readonly dict pragmas
 *     cdef readonly list instructions
 */
struct __pyx_obj_7flitter_8language_2vm_Program {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_8language_2vm_Program *__pyx_vtab;
  PyObject *pragmas;
  PyObject *instructions;
  int linked;
  PyObject *path;
  PyObject *top;
  PyObject *initial_lnames;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *stack;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *lnames;
  PyObject *compiler_errors;
  int simplify;
  int64_t next_label;
};


/* "flitter/language/vm.pyx":480
 * 
 * 
 * cdef class Function:             # <<<<<<<<<<<<<<
 *     cdef readonly str __name__
 *     cdef readonly Vector vself
 */
struct __pyx_obj_7flitter_8language_2vm_Function {
  PyObject_HEAD
  PyObject *__name__;
  struct __pyx_obj_7flitter_5model_Vector *vself;
  PyObject *parameters;
  PyObject *defaults;
  struct __pyx_obj_7flitter_8language_2vm_Program *program;
  int64_t address;
  int record_stats;
  PyObject *captures;
  int64_t call_depth;
  int64_t hash;
};


/* "flitter/language/vm.pyx":535
 * 
 * 
 * cdef class LoopSource:             # <<<<<<<<<<<<<<
 *     cdef Vector source
 *     cdef int64_t stride
 */
struct __pyx_obj_7flitter_8language_2vm_LoopSource {
  PyObject_HEAD
  struct __pyx_obj_7flitter_5model_Vector *source;
  int64_t stride;
  int64_t position;
  int64_t iterations;
};


/* "flitter/language/vm.pyx":779
 * 
 *     def __str__(self):
 *         return '\n'.join(str(instruction) for instruction in self.instructions)             # <<<<<<<<<<<<<<
 * 
 *     cdef Instruction last_instruction(self):
 */
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_instruction;
};


/* "flitter/language/vm.pyx":1092
 *         return self
 * 
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats):             # <<<<<<<<<<<<<<
 *         global CallOutCount, CallOutDuration
 *         cdef VectorStack stack=context.stack, lnames=context.lnames
 */
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute {
  PyObject_HEAD
  int64_t __pyx_v_pc;
};


/* "flitter/language/vm.pyx":1454
 *             if instruction is not None:
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_instruction;
};



/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */

struct __pyx_vtabstruct_7flitter_5model_Vector {
  struct __pyx_obj_7flitter_5model_Vector *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_compose)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_symbol)(PyObject *);
  int64_t (*allocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  void (*deallocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*copy)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*intern)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*fill_range)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*isinstance)(struct __pyx_obj_7flitter_5model_Vector *, PyObject *, int __pyx_skip_dispatch);
  int (*is_finite)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*as_bool)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*as_double)(struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*as_integer)(struct __pyx_obj_7flitter_5model_Vector *);
  PyObject *(*as_string)(struct __pyx_obj_7flitter_5model_Vector *);
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Vector *, int, int __pyx_skip_dispatch);
  PyObject *(*match)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_6Vector_match *__pyx_optional_args);
  PyObject *(*repr)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*neg)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pos)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*abs)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ceil)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floor)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*fract)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*round)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  struct __pyx_obj_7flitter_5model_Vector *(*add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul_add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*sub)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*truediv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floordiv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mod)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pow)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*eq)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ne)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*gt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ge)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*lt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*le)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*compare)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*slice)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*item)(struct __pyx_obj_7flitter_5model_Vector *, int64_t, int __pyx_skip_dispatch);
  double (*squared_sum)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*normalize)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*dot)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*cross)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*clamp)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  double (*minimum)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*maximum)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*concat)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtabptr_7flitter_5model_Vector;


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix33 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_rotate)(double);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Matrix33 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix33 *__pyx_vtabptr_7flitter_5model_Matrix33;


/* "model.pxd":128
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix44 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_project)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_ortho)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_look)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_x)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_y)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_z)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_x)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_y)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_z)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*immul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse_transpose_matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix44 *__pyx_vtabptr_7flitter_5model_Matrix44;


/* "model.pxd":178
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */

struct __pyx_vtabstruct_7flitter_5model_Quaternion {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Quaternion *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_euler)(struct __pyx_obj_7flitter_5model_Vector *, double);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_between)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*copy)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*mmul)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *);
  struct __pyx_obj_7flitter_5model_Vector *(*conjugate)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*normalize)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*inverse)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*exponent)(struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*slerp)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Quaternion *__pyx_vtabptr_7flitter_5model_Quaternion;


/* "model.pxd":198
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */

struct __pyx_vtabstruct_7flitter_5model_Node {
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Node *);
  struct __pyx_obj_7flitter_5model_Node *(*copy)(struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*add_tag)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch);
  void (*set_attribute)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*append)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*append_vector)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*_equals)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *);
  PyObject *(*get)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_4Node_get *__pyx_optional_args);
  struct __pyx_obj_7flitter_5model_Vector *(*get_vec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*get_fvec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int, struct __pyx_obj_7flitter_5model_Vector *);
  double (*get_float)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, double);
  int64_t (*get_int)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int64_t);
  int (*get_bool)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
  PyObject *(*get_str)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, PyObject *);
  void (*repr)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
};
static struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtabptr_7flitter_5model_Node;


/* "model.pxd":222
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */

struct __pyx_vtabstruct_7flitter_5model_StateDict {
  struct __pyx_obj_7flitter_5model_Vector *(*get_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
  void (*set_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*contains)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
};
static struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtabptr_7flitter_5model_StateDict;


/* "flitter/language/vm.pyx":246
 * 
 * 
 * cdef class VectorStack:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 */

struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack {
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *(*copy)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  void (*drop)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop *__pyx_optional_args);
  void (*push)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*pop)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  PyObject *(*pop_tuple)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*pop_list)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*pop_dict)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*pop_composed)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*peek)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*peek_at)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  void (*poke)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*poke_at)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack *__pyx_vtabptr_7flitter_8language_2vm_VectorStack;


/* "flitter/language/vm.pyx":765
 * 
 * 
 * cdef class Program:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, lnames=()):
 *         self.pragmas = {}
 */

struct __pyx_vtabstruct_7flitter_8language_2vm_Program {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *(*last_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *);
  struct __pyx_obj_7flitter_8language_2vm_Instruction *(*pop_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*push_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_8language_2vm_Instruction *);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*link)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  void (*use_simplifier)(struct __pyx_obj_7flitter_8language_2vm_Program *, int, int __pyx_skip_dispatch);
  int64_t (*new_label)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*dup)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*drop)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*label)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*jump)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*branch_true)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*branch_false)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*import_)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_push)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_load)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_drop)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lookup)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lookup_literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*range)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*neg)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*pos)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ceil)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*floor)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*fract)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*not_)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*add)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*sub)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mul)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mul_add)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*truediv)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*floordiv)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mod)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*pow)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*eq)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ne)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*gt)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lt)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ge)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*le)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*xor)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*slice)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*slice_literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*call)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_call *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*call_fast)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*tag)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*attributes)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*append)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_append *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*compose)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*begin_for)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*next)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*end_for)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*export)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*func)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_func *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*exit)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  void (*_execute)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Context *, int64_t, int);
};
static struct __pyx_vtabstruct_7flitter_8language_2vm_Program *__pyx_vtabptr_7flitter_8language_2vm_Program;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(enum __pyx_t_7flitter_8language_2vm_OpCode value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7flitter_8language_2vm_OpCode __Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_f_7flitter_8language_2vm_11VectorStack_copy(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop *__pyx_optional_args); /* proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, PyObject *__pyx_v_keys, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_f_7flitter_8language_2vm_7Program_last_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_f_7flitter_8language_2vm_7Program_pop_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_push_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instruction); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_link(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7flitter_8language_2vm_7Program_use_simplifier(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_v_simplify, int __pyx_skip_dispatch); /* proto*/
static int64_t __pyx_f_7flitter_8language_2vm_7Program_new_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_dup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop *__pyx_optional_args); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_jump(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_branch_true(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_branch_false(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_import_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_push(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_load(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_offset, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lookup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lookup_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_range(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_neg(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_pos(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ceil(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_floor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_fract(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_not_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_sub(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mul(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mul_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_truediv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_floordiv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mod(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_pow(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_eq(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ne(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_gt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ge(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_le(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_xor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_slice(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_slice_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_call(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_call *__pyx_optional_args); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_call_fast(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_function, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_tag(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_attributes(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_append(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_append *__pyx_optional_args); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_compose(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_begin_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_next(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_end_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_export(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_func(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_func *__pyx_optional_args); /* proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_exit(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7flitter_8language_2vm_7Program__execute(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, int64_t __pyx_v_pc, int __pyx_v_record_stats); /* proto*/

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "cython" */

/* Module declarations from "flitter.model" */
static uint64_t *__pyx_vp_7flitter_5model_HASH_START = 0;
#define __pyx_v_7flitter_5model_HASH_START (*__pyx_vp_7flitter_5model_HASH_START)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_null_ = 0;
#define __pyx_v_7flitter_5model_null_ (*__pyx_vp_7flitter_5model_null_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_true_ = 0;
#define __pyx_v_7flitter_5model_true_ (*__pyx_vp_7flitter_5model_true_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_false_ = 0;
#define __pyx_v_7flitter_5model_false_ (*__pyx_vp_7flitter_5model_false_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_minusone_ = 0;
#define __pyx_v_7flitter_5model_minusone_ (*__pyx_vp_7flitter_5model_minusone_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_inf_ = 0;
#define __pyx_v_7flitter_5model_inf_ (*__pyx_vp_7flitter_5model_inf_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_nan_ = 0;
#define __pyx_v_7flitter_5model_nan_ (*__pyx_vp_7flitter_5model_nan_)

/* Module declarations from "flitter.timer" */

/* Module declarations from "libc.math" */

/* Module declarations from "flitter.language.vm" */
static PyObject *__pyx_v_7flitter_8language_2vm_static_builtins = 0;
static PyObject *__pyx_v_7flitter_8language_2vm_dynamic_builtins = 0;
static PyObject *__pyx_v_7flitter_8language_2vm_builtins = 0;
static char const *__pyx_v_7flitter_8language_2vm_ContextFunc;
static int64_t __pyx_v_7flitter_8language_2vm_DEFAULT_STACK_SIZE;
static int64_t __pyx_v_7flitter_8language_2vm_MAX_STACK_SIZE;
static int64_t __pyx_v_7flitter_8language_2vm_MAX_CALL_DEPTH;
static PyObject *__pyx_v_7flitter_8language_2vm_all_builtins = 0;
static int64_t *__pyx_v_7flitter_8language_2vm_StatsCount;
static double *__pyx_v_7flitter_8language_2vm_StatsDuration;
static double __pyx_v_7flitter_8language_2vm_CallOutDuration;
static int64_t __pyx_v_7flitter_8language_2vm_CallOutCount;
static PyObject *__pyx_v_7flitter_8language_2vm_OpCodeNames = 0;
static PyObject *__pyx_f_7flitter_8language_2vm_initialize_stats(void); /*proto*/
static int64_t __pyx_f_7flitter_8language_2vm_increase(struct __pyx_obj_7flitter_8language_2vm_VectorStack *); /*proto*/
static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE int64_t __pyx_f_7flitter_8language_2vm_push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *); /*proto*/
static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, struct __pyx_obj_7flitter_5model_Vector *); /*proto*/
static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_call_helper(struct __pyx_obj_7flitter_5model_Context *, struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *, PyObject *, PyObject *, int, double *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_import_module(struct __pyx_obj_7flitter_5model_Context *, PyObject *, int, double *, int); /*proto*/
static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_execute_append(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_execute_attributes(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_execute_tag(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_Instruction__set_state(struct __pyx_obj_7flitter_8language_2vm_Instruction *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionVector__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionNode__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionStr__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionTuple__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionInt__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionIntTuple__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionLabel__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionJump__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionFunc__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionObjectInt__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_Function__set_state(struct __pyx_obj_7flitter_8language_2vm_Function *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_LoopSource__set_state(struct __pyx_obj_7flitter_8language_2vm_LoopSource *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "flitter.language.vm"
extern int __pyx_module_is_main_flitter__language__vm;
int __pyx_module_is_main_flitter__language__vm = 0;

/* Implementation of "flitter.language.vm" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_enumerate;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = " ";
static const char __pyx_k_L[] = ".L";
static const char __pyx_k_d[] = "+d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_Eq[] = "Eq";
static const char __pyx_k_Ge[] = "Ge";
static const char __pyx_k_Gt[] = "Gt";
static const char __pyx_k_Le[] = "Le";
static const char __pyx_k_Lt[] = "Lt";
static const char __pyx_k_Ne[] = "Ne";
static const char __pyx_k__2[] = " (";
static const char __pyx_k__3[] = ")";
static const char __pyx_k__6[] = "#";
static const char __pyx_k__7[] = ": ";
static const char __pyx_k__8[] = "\n";
static const char __pyx_k__9[] = "'";
static const char __pyx_k_eq[] = "eq";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_ge[] = "ge";
static const char __pyx_k_gt[] = "gt";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_le[] = "le";
static const char __pyx_k_lt[] = "lt";
static const char __pyx_k_ne[] = "ne";
static const char __pyx_k_Add[] = "Add";
static const char __pyx_k_Dup[] = "Dup";
static const char __pyx_k_L_2[] = " .L";
static const char __pyx_k_L_3[] = " L";
static const char __pyx_k_Mod[] = "Mod";
static const char __pyx_k_Mul[] = "Mul";
static const char __pyx_k_Neg[] = "Neg";
static const char __pyx_k_Not[] = "Not";
static const char __pyx_k_Pos[] = "Pos";
static const char __pyx_k_Pow[] = "Pow";
static const char __pyx_k_Sub[] = "Sub";
static const char __pyx_k_Tag[] = "Tag";
static const char __pyx_k_Xor[] = "Xor";
static const char __pyx_k__12[] = ".";
static const char __pyx_k__21[] = "";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_dup[] = "dup";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_mod[] = "mod";
static const char __pyx_k_mul[] = "mul";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_neg[] = "neg";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_not[] = "not_";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_pow[] = "pow";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_sub[] = "sub";
static const char __pyx_k_tag[] = "tag";
static const char __pyx_k_top[] = "top";
static const char __pyx_k_xor[] = "xor";
static const char __pyx_k_Call[] = "Call";
static const char __pyx_k_Ceil[] = "Ceil";
static const char __pyx_k_Drop[] = "Drop";
static const char __pyx_k_Exit[] = "Exit";
static const char __pyx_k_Func[] = "Func";
static const char __pyx_k_Jump[] = "Jump";
static const char __pyx_k_Next[] = "Next";
static const char __pyx_k_None[] = "None";
static const char __pyx_k__163[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_call[] = "call";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_code[] = "code";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_drop[] = "drop";
static const char __pyx_k_exit[] = "exit";
static const char __pyx_k_from[] = "' from '";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_item[] = "item";
static const char __pyx_k_jump[] = "jump";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_link[] = "link";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_next[] = "next";
static const char __pyx_k_null[] = "null";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_peek[] = "peek";
static const char __pyx_k_poke[] = "poke";
static const char __pyx_k_push[] = "push";
static const char __pyx_k_root[] = "root";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_true[] = "true";
static const char __pyx_k_Floor[] = "Floor";
static const char __pyx_k_Fract[] = "Fract";
static const char __pyx_k_Label[] = "Label";
static const char __pyx_k_Range[] = "Range";
static const char __pyx_k_Slice[] = "Slice";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_false[] = "false";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_fract[] = "fract";
static const char __pyx_k_items[] = " items";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_noise[] = "noise";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_slice[] = "slice";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_stats[] = "stats";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_total[] = "total";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_Append[] = "Append";
static const char __pyx_k_EndFor[] = "EndFor";
static const char __pyx_k_Export[] = "Export";
static const char __pyx_k_Import[] = "Import";
static const char __pyx_k_Lookup[] = "Lookup";
static const char __pyx_k_MulAdd[] = "MulAdd";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_export[] = "export";
static const char __pyx_k_import[] = "import_";
static const char __pyx_k_ivalue[] = "ivalue";
static const char __pyx_k_lnames[] = "lnames";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_loguru[] = "loguru";
static const char __pyx_k_lookup[] = "lookup";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_objptr[] = "objptr";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_tvalue[] = "tvalue";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_vector[] = "vector";
static const char __pyx_k_Compose[] = "Compose";
static const char __pyx_k_Literal[] = "Literal";
static const char __pyx_k_Program[] = "Program";
static const char __pyx_k_TrueDiv[] = "TrueDiv";
static const char __pyx_k_compose[] = "compose";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_end_for[] = "end_for";
static const char __pyx_k_execute[] = "execute";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_items_2[] = "items";
static const char __pyx_k_literal[] = "literal";
static const char __pyx_k_mul_add[] = "mul_add";
static const char __pyx_k_peek_at[] = "peek_at";
static const char __pyx_k_poke_at[] = "poke_at";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_set_top[] = "set_top";
static const char __pyx_k_truediv[] = "truediv";
static const char __pyx_k_BeginFor[] = "BeginFor";
static const char __pyx_k_CallFast[] = "CallFast";
static const char __pyx_k_FloorDiv[] = "FloorDiv";
static const char __pyx_k_Function[] = "Function";
static const char __pyx_k_builtins[] = "builtins";
static const char __pyx_k_duration[] = "duration";
static const char __pyx_k_floordiv[] = "floordiv";
static const char __pyx_k_function[] = "function";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_import_2[] = "__import__";
static const char __pyx_k_max_size[] = "max_size";
static const char __pyx_k_overhead[] = "overhead";
static const char __pyx_k_pop_dict[] = "pop_dict";
static const char __pyx_k_pop_list[] = "pop_list";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_set_path[] = "set_path";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_simplify[] = "simplify";
static const char __pyx_k_Bad_stack[] = "Bad stack";
static const char __pyx_k_LocalDrop[] = "LocalDrop";
static const char __pyx_k_LocalLoad[] = "LocalLoad";
static const char __pyx_k_LocalPush[] = "LocalPush";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_begin_for[] = "begin_for";
static const char __pyx_k_call_fast[] = "call_fast";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_functions[] = "functions";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_ncaptures[] = "ncaptures";
static const char __pyx_k_new_label[] = "new_label";
static const char __pyx_k_pop_tuple[] = "pop_tuple";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_Attributes[] = "Attributes";
static const char __pyx_k_Bad_lnames[] = "Bad lnames";
static const char __pyx_k_BranchTrue[] = "BranchTrue";
static const char __pyx_k_Cannot_pop[] = "Cannot pop ";
static const char __pyx_k_LoopSource[] = "LoopSource";
static const char __pyx_k_Program_eq[] = "Program.eq";
static const char __pyx_k_Program_ge[] = "Program.ge";
static const char __pyx_k_Program_gt[] = "Program.gt";
static const char __pyx_k_Program_le[] = "Program.le";
static const char __pyx_k_Program_lt[] = "Program.lt";
static const char __pyx_k_Program_ne[] = "Program.ne";
static const char __pyx_k_StackError[] = "StackError";
static const char __pyx_k_attributes[] = "attributes";
static const char __pyx_k_item_stack[] = " item stack";
static const char __pyx_k_items_from[] = " items from ";
static const char __pyx_k_local_drop[] = "local_drop";
static const char __pyx_k_local_load[] = "local_load";
static const char __pyx_k_local_push[] = "local_push";
static const char __pyx_k_name_patch[] = "name_patch";
static const char __pyx_k_parameters[] = "parameters";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_set_pragma[] = "set_pragma";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_BranchFalse[] = "BranchFalse";
static const char __pyx_k_Cannot_drop[] = "Cannot drop ";
static const char __pyx_k_Cannot_peek[] = "Cannot peek ";
static const char __pyx_k_Cannot_poke[] = "Cannot poke ";
static const char __pyx_k_Instruction[] = "Instruction";
static const char __pyx_k_LiteralNode[] = "LiteralNode";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_Program_add[] = "Program.add";
static const char __pyx_k_Program_dup[] = "Program.dup";
static const char __pyx_k_Program_mod[] = "Program.mod";
static const char __pyx_k_Program_mul[] = "Program.mul";
static const char __pyx_k_Program_neg[] = "Program.neg";
static const char __pyx_k_Program_not[] = "Program.not_";
static const char __pyx_k_Program_pos[] = "Program.pos";
static const char __pyx_k_Program_pow[] = "Program.pow";
static const char __pyx_k_Program_run[] = "Program.run";
static const char __pyx_k_Program_sub[] = "Program.sub";
static const char __pyx_k_Program_tag[] = "Program.tag";
static const char __pyx_k_Program_xor[] = "Program.xor";
static const char __pyx_k_SharedCache[] = "SharedCache";
static const char __pyx_k_Stack_empty[] = "Stack empty";
static const char __pyx_k_VectorStack[] = "VectorStack";
static const char __pyx_k_branch_true[] = "branch_true";
static const char __pyx_k_depth_stack[] = " depth stack";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_IndexLiteral[] = "IndexLiteral";
static const char __pyx_k_LiteralNodes[] = "LiteralNodes";
static const char __pyx_k_Program_call[] = "Program.call";
static const char __pyx_k_Program_ceil[] = "Program.ceil";
static const char __pyx_k_Program_drop[] = "Program.drop";
static const char __pyx_k_Program_exit[] = "Program.exit";
static const char __pyx_k_Program_func[] = "Program.func";
static const char __pyx_k_Program_jump[] = "Program.jump";
static const char __pyx_k_Program_link[] = "Program.link";
static const char __pyx_k_Program_next[] = "Program.next";
static const char __pyx_k_SliceLiteral[] = "SliceLiteral";
static const char __pyx_k_branch_false[] = "branch_false";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_log_vm_stats[] = "log_vm_stats";
static const char __pyx_k_native_funcs[] = "(native funcs)";
static const char __pyx_k_pop_composed[] = "pop_composed";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_record_stats[] = "record_stats";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_Bad_root_node[] = "Bad root node";
static const char __pyx_k_Error_calling[] = "Error calling ";
static const char __pyx_k_LookupLiteral[] = "LookupLiteral";
static const char __pyx_k_Missing_lname[] = "Missing lname";
static const char __pyx_k_Program_floor[] = "Program.floor";
static const char __pyx_k_Program_fract[] = "Program.fract";
static const char __pyx_k_Program_label[] = "Program.label";
static const char __pyx_k_Program_range[] = "Program.range";
static const char __pyx_k_Program_slice[] = "Program.slice";
static const char __pyx_k_StackOverflow[] = "StackOverflow";
static const char __pyx_k_get_with_root[] = "get_with_root";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_per_execution[] = "per_execution";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_slice_literal[] = "slice_literal";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_InstructionInt[] = "InstructionInt";
static const char __pyx_k_InstructionStr[] = "InstructionStr";
static const char __pyx_k_Program_append[] = "Program.append";
static const char __pyx_k_Program_export[] = "Program.export";
static const char __pyx_k_Program_import[] = "Program.import_";
static const char __pyx_k_Program_lookup[] = "Program.lookup";
static const char __pyx_k_RecursionError[] = "RecursionError";
static const char __pyx_k_StackUnderflow[] = "StackUnderflow";
static const char __pyx_k_lookup_literal[] = "lookup_literal";
static const char __pyx_k_use_simplifier[] = "use_simplifier";
static const char __pyx_k_InstructionFunc[] = "InstructionFunc";
static const char __pyx_k_InstructionJump[] = "InstructionJump";
static const char __pyx_k_InstructionNode[] = "InstructionNode";
static const char __pyx_k_NOISE_FUNCTIONS[] = "NOISE_FUNCTIONS";
static const char __pyx_k_Program_compose[] = "Program.compose";
static const char __pyx_k_Program_end_for[] = "Program.end_for";
static const char __pyx_k_Program_execute[] = "Program.execute";
static const char __pyx_k_Program_literal[] = "Program.literal";
static const char __pyx_k_Program_mul_add[] = "Program.mul_add";
static const char __pyx_k_Program_set_top[] = "Program.set_top";
static const char __pyx_k_Program_truediv[] = "Program.truediv";
static const char __pyx_k_VectorStack_pop[] = "VectorStack.pop";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_static_builtins[] = "static_builtins";
static const char __pyx_k_InstructionLabel[] = "InstructionLabel";
static const char __pyx_k_InstructionTuple[] = "InstructionTuple";
static const char __pyx_k_Program_floordiv[] = "Program.floordiv";
static const char __pyx_k_Program_set_path[] = "Program.set_path";
static const char __pyx_k_STATIC_FUNCTIONS[] = "STATIC_FUNCTIONS";
static const char __pyx_k_Unable_to_import[] = "Unable to import '";
static const char __pyx_k_VectorStack_copy[] = "VectorStack.copy";
static const char __pyx_k_VectorStack_drop[] = "VectorStack.drop";
static const char __pyx_k_VectorStack_peek[] = "VectorStack.peek";
static const char __pyx_k_VectorStack_poke[] = "VectorStack.poke";
static const char __pyx_k_VectorStack_push[] = "VectorStack.push";
static const char __pyx_k_dynamic_builtins[] = "dynamic_builtins";
static const char __pyx_k_DYNAMIC_FUNCTIONS[] = "DYNAMIC_FUNCTIONS";
static const char __pyx_k_InstructionVector[] = "InstructionVector";
static const char __pyx_k_Program_begin_for[] = "Program.begin_for";
static const char __pyx_k_Program_call_fast[] = "Program.call_fast";
static const char __pyx_k_Program_new_label[] = "Program.new_label";
static const char __pyx_k_Circular_import_of[] = "Circular import of '";
static const char __pyx_k_Program_attributes[] = "Program.attributes";
static const char __pyx_k_Program_local_drop[] = "Program.local_drop";
static const char __pyx_k_Program_local_load[] = "Program.local_load";
static const char __pyx_k_Program_local_push[] = "Program.local_push";
static const char __pyx_k_Program_set_pragma[] = "Program.set_pragma";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_items_below_top_of[] = " items below top of ";
static const char __pyx_k_InstructionIntTuple[] = "InstructionIntTuple";
static const char __pyx_k_Program_branch_true[] = "Program.branch_true";
static const char __pyx_k_VectorStack_peek_at[] = "VectorStack.peek_at";
static const char __pyx_k_VectorStack_poke_at[] = "VectorStack.poke_at";
static const char __pyx_k_flitter_language_vm[] = "flitter.language.vm";
static const char __pyx_k_InstructionObjectInt[] = "InstructionObjectInt";
static const char __pyx_k_Program_branch_false[] = "Program.branch_false";
static const char __pyx_k_VectorStack_pop_dict[] = "VectorStack.pop_dict";
static const char __pyx_k_VectorStack_pop_list[] = "VectorStack.pop_list";
static const char __pyx_k_read_flitter_program[] = "read_flitter_program";
static const char __pyx_k_str___locals_genexpr[] = "__str__.<locals>.genexpr";
static const char __pyx_k_Program_slice_literal[] = "Program.slice_literal";
static const char __pyx_k_Unable_to_import_from[] = "Unable to import from '";
static const char __pyx_k_VectorStack_pop_tuple[] = "VectorStack.pop_tuple";
static const char __pyx_k_pyx_unpickle_Function[] = "__pyx_unpickle_Function";
static const char __pyx_k_Program_lookup_literal[] = "Program.lookup_literal";
static const char __pyx_k_Program_use_simplifier[] = "Program.use_simplifier";
static const char __pyx_k_Jump_outside_of_program[] = "Jump outside of program";
static const char __pyx_k_Program___reduce_cython[] = "Program.__reduce_cython__";
static const char __pyx_k_VM_exception_processing[] = "VM exception processing:\n{} <--";
static const char __pyx_k_VM_execution_statistics[] = "VM execution statistics:";
static const char __pyx_k_pyx_unpickle_LoopSource[] = "__pyx_unpickle_LoopSource";
static const char __pyx_k_Function___reduce_cython[] = "Function.__reduce_cython__";
static const char __pyx_k_Stack_cannot_grow_beyond[] = "Stack cannot grow beyond ";
static const char __pyx_k_Unrecognised_instruction[] = "Unrecognised instruction: ";
static const char __pyx_k_VectorStack_pop_composed[] = "VectorStack.pop_composed";
static const char __pyx_k_pyx_unpickle_Instruction[] = "__pyx_unpickle_Instruction";
static const char __pyx_k_Bad_function_return_stack[] = "Bad function return stack";
static const char __pyx_k_Program___setstate_cython[] = "Program.__setstate_cython__";
static const char __pyx_k_Stack_overflow_in_program[] = "Stack overflow in program";
static const char __pyx_k_Bad_function_return_lnames[] = "Bad function return lnames";
static const char __pyx_k_Function___setstate_cython[] = "Function.__setstate_cython__";
static const char __pyx_k_LoopSource___reduce_cython[] = "LoopSource.__reduce_cython__";
static const char __pyx_k_15s_9d_x_8_3f_s_7_3f_s_4_1f[] = "- {:15s} {:9d} x {:8.3f}\302\265s = {:7.3f}s ({:4.1f}%)";
static const char __pyx_k_Instruction___reduce_cython[] = "Instruction.__reduce_cython__";
static const char __pyx_k_Program_has_not_been_linked[] = "Program has not been linked";
static const char __pyx_k_VectorStack___reduce_cython[] = "VectorStack.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_InstructionInt[] = "__pyx_unpickle_InstructionInt";
static const char __pyx_k_pyx_unpickle_InstructionStr[] = "__pyx_unpickle_InstructionStr";
static const char __pyx_k_src_flitter_language_vm_pyx[] = "src/flitter/language/vm.pyx";
static const char __pyx_k_15s_9d_x_8_0f_ns_7_3f_s_4_1f[] = "- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)";
static const char __pyx_k_LoopSource___setstate_cython[] = "LoopSource.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_InstructionFunc[] = "__pyx_unpickle_InstructionFunc";
static const char __pyx_k_pyx_unpickle_InstructionJump[] = "__pyx_unpickle_InstructionJump";
static const char __pyx_k_pyx_unpickle_InstructionNode[] = "__pyx_unpickle_InstructionNode";
static const char __pyx_k_Instruction___setstate_cython[] = "Instruction.__setstate_cython__";
static const char __pyx_k_VectorStack___setstate_cython[] = "VectorStack.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_InstructionLabel[] = "__pyx_unpickle_InstructionLabel";
static const char __pyx_k_pyx_unpickle_InstructionTuple[] = "__pyx_unpickle_InstructionTuple";
static const char __pyx_k_InstructionInt___reduce_cython[] = "InstructionInt.__reduce_cython__";
static const char __pyx_k_InstructionStr___reduce_cython[] = "InstructionStr.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_InstructionIntTup[] = "__pyx_unpickle_InstructionIntTuple";
static const char __pyx_k_pyx_unpickle_InstructionObject[] = "__pyx_unpickle_InstructionObjectInt";
static const char __pyx_k_pyx_unpickle_InstructionVector[] = "__pyx_unpickle_InstructionVector";
static const char __pyx_k_Flitter_language_stack_based_vi[] = "\nFlitter language stack-based virtual machine\n";
static const char __pyx_k_InstructionFunc___reduce_cython[] = "InstructionFunc.__reduce_cython__";
static const char __pyx_k_InstructionJump___reduce_cython[] = "InstructionJump.__reduce_cython__";
static const char __pyx_k_InstructionNode___reduce_cython[] = "InstructionNode.__reduce_cython__";
static const char __pyx_k_InstructionObjectInt___setstate[] = "InstructionObjectInt.__setstate_cython__";
static const char __pyx_k_Program__execute_locals_genexpr[] = "Program._execute.<locals>.genexpr";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))";
static const char __pyx_k_InstructionFunc___setstate_cytho[] = "InstructionFunc.__setstate_cython__";
static const char __pyx_k_InstructionIntTuple___reduce_cyt[] = "InstructionIntTuple.__reduce_cython__";
static const char __pyx_k_InstructionIntTuple___setstate_c[] = "InstructionIntTuple.__setstate_cython__";
static const char __pyx_k_InstructionInt___setstate_cython[] = "InstructionInt.__setstate_cython__";
static const char __pyx_k_InstructionJump___setstate_cytho[] = "InstructionJump.__setstate_cython__";
static const char __pyx_k_InstructionLabel___reduce_cython[] = "InstructionLabel.__reduce_cython__";
static const char __pyx_k_InstructionLabel___setstate_cyth[] = "InstructionLabel.__setstate_cython__";
static const char __pyx_k_InstructionNode___setstate_cytho[] = "InstructionNode.__setstate_cython__";
static const char __pyx_k_InstructionObjectInt___reduce_cy[] = "InstructionObjectInt.__reduce_cython__";
static const char __pyx_k_InstructionStr___setstate_cython[] = "InstructionStr.__setstate_cython__";
static const char __pyx_k_InstructionTuple___reduce_cython[] = "InstructionTuple.__reduce_cython__";
static const char __pyx_k_InstructionTuple___setstate_cyth[] = "InstructionTuple.__setstate_cython__";
static const char __pyx_k_InstructionVector___reduce_cytho[] = "InstructionVector.__reduce_cython__";
static const char __pyx_k_InstructionVector___setstate_cyt[] = "InstructionVector.__setstate_cython__";
static const char __pyx_k_Recursion_depth_exceeded_for_fun[] = "Recursion depth exceeded for func '";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_8[] = "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_9[] = "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))";
/* #### Code section: decls ### */
static int __pyx_pf_7flitter_8language_2vm_11Instruction___init__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_2__str__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_4code___get__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_17InstructionVector___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_15InstructionNode___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, struct __pyx_obj_7flitter_5model_Node *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_14InstructionStr___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_16InstructionTuple___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_14InstructionInt___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_ivalue, PyObject *__pyx_v_tvalue); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6ivalue___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6tvalue___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_16InstructionLabel___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_5label___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_15InstructionJump___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_5label___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_6offset___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_15InstructionFunc___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int64_t __pyx_v_ncaptures); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_4name___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_10parameters___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_9ncaptures___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_obj, int64_t __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_3obj___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_2vm_11VectorStack___cinit__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_size, int64_t __pyx_v_max_size); /* proto */
static void __pyx_pf_7flitter_8language_2vm_11VectorStack_2__dealloc__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_7flitter_8language_2vm_11VectorStack_4__len__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_6__str__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_8copy(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_10drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_12push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_14pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_16pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_18pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_20pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, PyObject *__pyx_v_keys); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_22pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_24peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_26peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_28poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_30poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_4size___get__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_8max_size___get__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_32__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_34__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static Py_hash_t __pyx_pf_7flitter_8language_2vm_8Function___hash__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_2__call__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8__name_____get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_5vself___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_10parameters___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8defaults___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_7program___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_7address___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_12record_stats___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8captures___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_10call_depth___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_4hash___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_10LoopSource___reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_10LoopSource_2__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_log_vm_stats(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static int __pyx_pf_7flitter_8language_2vm_7Program___cinit__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_lnames); /* proto */
static Py_ssize_t __pyx_pf_7flitter_8language_2vm_7Program_2__len__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_4__str__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_6set_path(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8set_top(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_top); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_10set_pragma(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_12execute(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, PyObject *__pyx_v_lnames, int __pyx_v_record_stats); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_14run(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, int __pyx_v_record_stats); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_16link(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_18use_simplifier(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_v_simplify); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_20new_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_22dup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_24drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_26label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_28jump(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_30branch_true(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_32branch_false(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_34import_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_36literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_38local_push(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_40local_load(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_offset); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_42local_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_44lookup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_46lookup_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_48range(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_50neg(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_52pos(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_54ceil(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_56floor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_58fract(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_60not_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_62add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_64sub(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_66mul(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_68mul_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_70truediv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_72floordiv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_74mod(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_76pow(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_78eq(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_80ne(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_82gt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_84lt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_86ge(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_88le(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_90xor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_92slice(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_94slice_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_96call(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, PyObject *__pyx_v_names); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_98call_fast(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_function, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_100tag(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_102attributes(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_104append(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_106compose(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_108begin_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_110next(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_112end_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_114export(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_116func(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int64_t __pyx_v_ncaptures); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_118exit(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8_execute_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_7pragmas___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_12instructions___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_4path___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_3top___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_14initial_lnames___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_5stack___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_6lnames___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_15compiler_errors___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8simplify___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_120__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_122__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_2__pyx_unpickle_Instruction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_4__pyx_unpickle_InstructionVector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_6__pyx_unpickle_InstructionNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_8__pyx_unpickle_InstructionStr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_10__pyx_unpickle_InstructionTuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_12__pyx_unpickle_InstructionInt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_14__pyx_unpickle_InstructionIntTuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_16__pyx_unpickle_InstructionLabel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_18__pyx_unpickle_InstructionJump(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_20__pyx_unpickle_InstructionFunc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_22__pyx_unpickle_InstructionObjectInt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_24__pyx_unpickle_Function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_2vm_26__pyx_unpickle_LoopSource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7flitter_8language_2vm_VectorStack(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_Instruction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionVector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionStr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionTuple(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionIntTuple(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionFunc(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionObjectInt(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_Program(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_Function(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm_LoopSource(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_1__execute(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_setdefault = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_update = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_clear = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7flitter_5model_Vector;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix33;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix44;
  PyTypeObject *__pyx_ptype_7flitter_5model_Quaternion;
  PyTypeObject *__pyx_ptype_7flitter_5model_Node;
  PyTypeObject *__pyx_ptype_7flitter_5model_StateDict;
  PyTypeObject *__pyx_ptype_7flitter_5model_Context;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_7flitter_8language_2vm_VectorStack;
  PyObject *__pyx_type_7flitter_8language_2vm_Instruction;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionVector;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionNode;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionStr;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionTuple;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionInt;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionIntTuple;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionLabel;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionJump;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionFunc;
  PyObject *__pyx_type_7flitter_8language_2vm_InstructionObjectInt;
  PyObject *__pyx_type_7flitter_8language_2vm_Program;
  PyObject *__pyx_type_7flitter_8language_2vm_Function;
  PyObject *__pyx_type_7flitter_8language_2vm_LoopSource;
  PyObject *__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr;
  PyObject *__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute;
  PyObject *__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr;
  #endif
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_VectorStack;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_Instruction;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionVector;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionNode;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionStr;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionTuple;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionInt;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionLabel;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionJump;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionFunc;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_Program;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_Function;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_LoopSource;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f;
  PyObject *__pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f;
  PyObject *__pyx_n_u_Add;
  PyObject *__pyx_n_u_Append;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_u_Attributes;
  PyObject *__pyx_kp_u_Bad_function_return_lnames;
  PyObject *__pyx_kp_u_Bad_function_return_stack;
  PyObject *__pyx_kp_u_Bad_lnames;
  PyObject *__pyx_kp_u_Bad_root_node;
  PyObject *__pyx_kp_u_Bad_stack;
  PyObject *__pyx_n_u_BeginFor;
  PyObject *__pyx_n_u_BranchFalse;
  PyObject *__pyx_n_u_BranchTrue;
  PyObject *__pyx_n_u_Call;
  PyObject *__pyx_n_u_CallFast;
  PyObject *__pyx_kp_u_Cannot_drop;
  PyObject *__pyx_kp_u_Cannot_peek;
  PyObject *__pyx_kp_u_Cannot_poke;
  PyObject *__pyx_kp_u_Cannot_pop;
  PyObject *__pyx_n_u_Ceil;
  PyObject *__pyx_kp_u_Circular_import_of;
  PyObject *__pyx_n_u_Compose;
  PyObject *__pyx_n_s_DYNAMIC_FUNCTIONS;
  PyObject *__pyx_n_u_Drop;
  PyObject *__pyx_n_u_Dup;
  PyObject *__pyx_n_u_EndFor;
  PyObject *__pyx_n_u_Eq;
  PyObject *__pyx_kp_u_Error_calling;
  PyObject *__pyx_n_u_Exit;
  PyObject *__pyx_n_u_Export;
  PyObject *__pyx_n_u_Floor;
  PyObject *__pyx_n_u_FloorDiv;
  PyObject *__pyx_n_u_Fract;
  PyObject *__pyx_n_u_Func;
  PyObject *__pyx_n_s_Function;
  PyObject *__pyx_n_s_Function___reduce_cython;
  PyObject *__pyx_n_s_Function___setstate_cython;
  PyObject *__pyx_n_u_Ge;
  PyObject *__pyx_n_u_Gt;
  PyObject *__pyx_n_u_Import;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9;
  PyObject *__pyx_n_u_IndexLiteral;
  PyObject *__pyx_n_s_Instruction;
  PyObject *__pyx_n_s_InstructionFunc;
  PyObject *__pyx_n_s_InstructionFunc___reduce_cython;
  PyObject *__pyx_n_s_InstructionFunc___setstate_cytho;
  PyObject *__pyx_n_s_InstructionInt;
  PyObject *__pyx_n_s_InstructionIntTuple;
  PyObject *__pyx_n_s_InstructionIntTuple___reduce_cyt;
  PyObject *__pyx_n_s_InstructionIntTuple___setstate_c;
  PyObject *__pyx_n_s_InstructionInt___reduce_cython;
  PyObject *__pyx_n_s_InstructionInt___setstate_cython;
  PyObject *__pyx_n_s_InstructionJump;
  PyObject *__pyx_n_s_InstructionJump___reduce_cython;
  PyObject *__pyx_n_s_InstructionJump___setstate_cytho;
  PyObject *__pyx_n_s_InstructionLabel;
  PyObject *__pyx_n_s_InstructionLabel___reduce_cython;
  PyObject *__pyx_n_s_InstructionLabel___setstate_cyth;
  PyObject *__pyx_n_s_InstructionNode;
  PyObject *__pyx_n_s_InstructionNode___reduce_cython;
  PyObject *__pyx_n_s_InstructionNode___setstate_cytho;
  PyObject *__pyx_n_s_InstructionObjectInt;
  PyObject *__pyx_n_s_InstructionObjectInt___reduce_cy;
  PyObject *__pyx_n_s_InstructionObjectInt___setstate;
  PyObject *__pyx_n_s_InstructionStr;
  PyObject *__pyx_n_s_InstructionStr___reduce_cython;
  PyObject *__pyx_n_s_InstructionStr___setstate_cython;
  PyObject *__pyx_n_s_InstructionTuple;
  PyObject *__pyx_n_s_InstructionTuple___reduce_cython;
  PyObject *__pyx_n_s_InstructionTuple___setstate_cyth;
  PyObject *__pyx_n_s_InstructionVector;
  PyObject *__pyx_n_s_InstructionVector___reduce_cytho;
  PyObject *__pyx_n_s_InstructionVector___setstate_cyt;
  PyObject *__pyx_n_s_Instruction___reduce_cython;
  PyObject *__pyx_n_s_Instruction___setstate_cython;
  PyObject *__pyx_n_u_Jump;
  PyObject *__pyx_kp_u_Jump_outside_of_program;
  PyObject *__pyx_kp_u_L;
  PyObject *__pyx_kp_u_L_2;
  PyObject *__pyx_kp_u_L_3;
  PyObject *__pyx_n_u_Label;
  PyObject *__pyx_n_u_Le;
  PyObject *__pyx_n_u_Literal;
  PyObject *__pyx_n_u_LiteralNode;
  PyObject *__pyx_n_u_LiteralNodes;
  PyObject *__pyx_n_u_LocalDrop;
  PyObject *__pyx_n_u_LocalLoad;
  PyObject *__pyx_n_u_LocalPush;
  PyObject *__pyx_n_u_Lookup;
  PyObject *__pyx_n_u_LookupLiteral;
  PyObject *__pyx_n_s_LoopSource;
  PyObject *__pyx_n_s_LoopSource___reduce_cython;
  PyObject *__pyx_n_s_LoopSource___setstate_cython;
  PyObject *__pyx_n_u_Lt;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_kp_u_Missing_lname;
  PyObject *__pyx_n_u_Mod;
  PyObject *__pyx_n_u_Mul;
  PyObject *__pyx_n_u_MulAdd;
  PyObject *__pyx_n_s_NOISE_FUNCTIONS;
  PyObject *__pyx_n_u_Ne;
  PyObject *__pyx_n_u_Neg;
  PyObject *__pyx_n_u_Next;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_n_u_Not;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_u_Pos;
  PyObject *__pyx_n_u_Pow;
  PyObject *__pyx_n_s_Program;
  PyObject *__pyx_n_s_Program___reduce_cython;
  PyObject *__pyx_n_s_Program___setstate_cython;
  PyObject *__pyx_n_s_Program__execute_locals_genexpr;
  PyObject *__pyx_n_s_Program_add;
  PyObject *__pyx_n_s_Program_append;
  PyObject *__pyx_n_s_Program_attributes;
  PyObject *__pyx_n_s_Program_begin_for;
  PyObject *__pyx_n_s_Program_branch_false;
  PyObject *__pyx_n_s_Program_branch_true;
  PyObject *__pyx_n_s_Program_call;
  PyObject *__pyx_n_s_Program_call_fast;
  PyObject *__pyx_n_s_Program_ceil;
  PyObject *__pyx_n_s_Program_compose;
  PyObject *__pyx_n_s_Program_drop;
  PyObject *__pyx_n_s_Program_dup;
  PyObject *__pyx_n_s_Program_end_for;
  PyObject *__pyx_n_s_Program_eq;
  PyObject *__pyx_n_s_Program_execute;
  PyObject *__pyx_n_s_Program_exit;
  PyObject *__pyx_n_s_Program_export;
  PyObject *__pyx_n_s_Program_floor;
  PyObject *__pyx_n_s_Program_floordiv;
  PyObject *__pyx_n_s_Program_fract;
  PyObject *__pyx_n_s_Program_func;
  PyObject *__pyx_n_s_Program_ge;
  PyObject *__pyx_n_s_Program_gt;
  PyObject *__pyx_kp_u_Program_has_not_been_linked;
  PyObject *__pyx_n_s_Program_import;
  PyObject *__pyx_n_s_Program_jump;
  PyObject *__pyx_n_s_Program_label;
  PyObject *__pyx_n_s_Program_le;
  PyObject *__pyx_n_s_Program_link;
  PyObject *__pyx_n_s_Program_literal;
  PyObject *__pyx_n_s_Program_local_drop;
  PyObject *__pyx_n_s_Program_local_load;
  PyObject *__pyx_n_s_Program_local_push;
  PyObject *__pyx_n_s_Program_lookup;
  PyObject *__pyx_n_s_Program_lookup_literal;
  PyObject *__pyx_n_s_Program_lt;
  PyObject *__pyx_n_s_Program_mod;
  PyObject *__pyx_n_s_Program_mul;
  PyObject *__pyx_n_s_Program_mul_add;
  PyObject *__pyx_n_s_Program_ne;
  PyObject *__pyx_n_s_Program_neg;
  PyObject *__pyx_n_s_Program_new_label;
  PyObject *__pyx_n_s_Program_next;
  PyObject *__pyx_n_s_Program_not;
  PyObject *__pyx_n_s_Program_pos;
  PyObject *__pyx_n_s_Program_pow;
  PyObject *__pyx_n_s_Program_range;
  PyObject *__pyx_n_s_Program_run;
  PyObject *__pyx_n_s_Program_set_path;
  PyObject *__pyx_n_s_Program_set_pragma;
  PyObject *__pyx_n_s_Program_set_top;
  PyObject *__pyx_n_s_Program_slice;
  PyObject *__pyx_n_s_Program_slice_literal;
  PyObject *__pyx_n_s_Program_sub;
  PyObject *__pyx_n_s_Program_tag;
  PyObject *__pyx_n_s_Program_truediv;
  PyObject *__pyx_n_s_Program_use_simplifier;
  PyObject *__pyx_n_s_Program_xor;
  PyObject *__pyx_n_u_Range;
  PyObject *__pyx_n_s_RecursionError;
  PyObject *__pyx_kp_u_Recursion_depth_exceeded_for_fun;
  PyObject *__pyx_n_s_STATIC_FUNCTIONS;
  PyObject *__pyx_n_s_SharedCache;
  PyObject *__pyx_n_u_Slice;
  PyObject *__pyx_n_u_SliceLiteral;
  PyObject *__pyx_n_s_StackError;
  PyObject *__pyx_n_s_StackOverflow;
  PyObject *__pyx_n_s_StackUnderflow;
  PyObject *__pyx_kp_u_Stack_cannot_grow_beyond;
  PyObject *__pyx_kp_u_Stack_empty;
  PyObject *__pyx_kp_u_Stack_overflow_in_program;
  PyObject *__pyx_n_u_Sub;
  PyObject *__pyx_n_u_Tag;
  PyObject *__pyx_n_u_TrueDiv;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_u_Unable_to_import;
  PyObject *__pyx_kp_u_Unable_to_import_from;
  PyObject *__pyx_kp_u_Unrecognised_instruction;
  PyObject *__pyx_kp_u_VM_exception_processing;
  PyObject *__pyx_kp_u_VM_execution_statistics;
  PyObject *__pyx_n_s_VectorStack;
  PyObject *__pyx_n_s_VectorStack___reduce_cython;
  PyObject *__pyx_n_s_VectorStack___setstate_cython;
  PyObject *__pyx_n_s_VectorStack_copy;
  PyObject *__pyx_n_s_VectorStack_drop;
  PyObject *__pyx_n_s_VectorStack_peek;
  PyObject *__pyx_n_s_VectorStack_peek_at;
  PyObject *__pyx_n_s_VectorStack_poke;
  PyObject *__pyx_n_s_VectorStack_poke_at;
  PyObject *__pyx_n_s_VectorStack_pop;
  PyObject *__pyx_n_s_VectorStack_pop_composed;
  PyObject *__pyx_n_s_VectorStack_pop_dict;
  PyObject *__pyx_n_s_VectorStack_pop_list;
  PyObject *__pyx_n_s_VectorStack_pop_tuple;
  PyObject *__pyx_n_s_VectorStack_push;
  PyObject *__pyx_n_u_Xor;
  PyObject *__pyx_kp_u__12;
  PyObject *__pyx_n_s__163;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_n_s__21;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_attributes;
  PyObject *__pyx_n_s_begin_for;
  PyObject *__pyx_n_s_branch_false;
  PyObject *__pyx_n_s_branch_true;
  PyObject *__pyx_n_s_builtins;
  PyObject *__pyx_n_s_cache;
  PyObject *__pyx_n_s_call;
  PyObject *__pyx_n_s_call_fast;
  PyObject *__pyx_n_s_ceil;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_code;
  PyObject *__pyx_n_s_compose;
  PyObject *__pyx_n_s_context;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_count;
  PyObject *__pyx_kp_u_d;
  PyObject *__pyx_kp_u_depth_stack;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_drop;
  PyObject *__pyx_n_s_dup;
  PyObject *__pyx_n_s_duration;
  PyObject *__pyx_n_s_dynamic_builtins;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_s_end_for;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_eq;
  PyObject *__pyx_n_s_exception;
  PyObject *__pyx_n_s_execute;
  PyObject *__pyx_n_s_exit;
  PyObject *__pyx_n_s_export;
  PyObject *__pyx_n_u_false;
  PyObject *__pyx_n_s_flitter_language_vm;
  PyObject *__pyx_n_s_floor;
  PyObject *__pyx_n_s_floordiv;
  PyObject *__pyx_n_s_fract;
  PyObject *__pyx_kp_u_from;
  PyObject *__pyx_n_s_func;
  PyObject *__pyx_n_s_function;
  PyObject *__pyx_n_s_functions;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_ge;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get_with_root;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_gt;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_import_2;
  PyObject *__pyx_n_u_inf;
  PyObject *__pyx_n_s_info;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_item;
  PyObject *__pyx_kp_u_item_stack;
  PyObject *__pyx_kp_u_items;
  PyObject *__pyx_n_s_items_2;
  PyObject *__pyx_kp_u_items_below_top_of;
  PyObject *__pyx_kp_u_items_from;
  PyObject *__pyx_n_s_ivalue;
  PyObject *__pyx_n_s_jump;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_label;
  PyObject *__pyx_n_s_le;
  PyObject *__pyx_n_s_link;
  PyObject *__pyx_n_s_literal;
  PyObject *__pyx_n_s_lnames;
  PyObject *__pyx_n_s_local_drop;
  PyObject *__pyx_n_s_local_load;
  PyObject *__pyx_n_s_local_push;
  PyObject *__pyx_n_s_log_vm_stats;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_loguru;
  PyObject *__pyx_n_s_lookup;
  PyObject *__pyx_n_s_lookup_literal;
  PyObject *__pyx_n_s_lt;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_max_size;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_mod;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_mul;
  PyObject *__pyx_n_s_mul_add;
  PyObject *__pyx_n_s_n;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_name_patch;
  PyObject *__pyx_n_s_names;
  PyObject *__pyx_n_u_nan;
  PyObject *__pyx_kp_u_native_funcs;
  PyObject *__pyx_n_s_ncaptures;
  PyObject *__pyx_n_s_ne;
  PyObject *__pyx_n_s_neg;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_new_label;
  PyObject *__pyx_n_s_next;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_noise;
  PyObject *__pyx_n_s_not;
  PyObject *__pyx_n_u_null;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_objptr;
  PyObject *__pyx_n_s_offset;
  PyObject *__pyx_n_s_overhead;
  PyObject *__pyx_n_s_parameters;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_n_s_peek;
  PyObject *__pyx_n_s_peek_at;
  PyObject *__pyx_n_s_per_execution;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_poke;
  PyObject *__pyx_n_s_poke_at;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_pop_composed;
  PyObject *__pyx_n_s_pop_dict;
  PyObject *__pyx_n_s_pop_list;
  PyObject *__pyx_n_s_pop_tuple;
  PyObject *__pyx_n_s_pos;
  PyObject *__pyx_n_s_pow;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_push;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_capi;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_Function;
  PyObject *__pyx_n_s_pyx_unpickle_Instruction;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionFunc;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionInt;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionIntTup;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionJump;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionLabel;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionNode;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionObject;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionStr;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionTuple;
  PyObject *__pyx_n_s_pyx_unpickle_InstructionVector;
  PyObject *__pyx_n_s_pyx_unpickle_LoopSource;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_read_flitter_program;
  PyObject *__pyx_n_s_record_stats;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_n_s_reverse;
  PyObject *__pyx_n_u_root;
  PyObject *__pyx_n_s_run;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_set_path;
  PyObject *__pyx_n_s_set_pragma;
  PyObject *__pyx_n_s_set_top;
  PyObject *__pyx_n_s_setdefault;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_simplify;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_s_slice;
  PyObject *__pyx_n_s_slice_literal;
  PyObject *__pyx_n_s_sort;
  PyObject *__pyx_kp_s_src_flitter_language_vm_pyx;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_static_builtins;
  PyObject *__pyx_n_s_stats;
  PyObject *__pyx_n_s_str___locals_genexpr;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_sub;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_tag;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_top;
  PyObject *__pyx_n_s_total;
  PyObject *__pyx_n_u_true;
  PyObject *__pyx_n_s_truediv;
  PyObject *__pyx_n_s_tvalue;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_use_simplifier;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_vector;
  PyObject *__pyx_n_s_xor;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_10904070;
  PyObject *__pyx_int_19407059;
  PyObject *__pyx_int_29217351;
  PyObject *__pyx_int_45504841;
  PyObject *__pyx_int_59839126;
  PyObject *__pyx_int_61928532;
  PyObject *__pyx_int_63712393;
  PyObject *__pyx_int_71438266;
  PyObject *__pyx_int_72616288;
  PyObject *__pyx_int_73689043;
  PyObject *__pyx_int_90787080;
  PyObject *__pyx_int_107792009;
  PyObject *__pyx_int_109158536;
  PyObject *__pyx_int_138499235;
  PyObject *__pyx_int_163945588;
  PyObject *__pyx_int_176342959;
  PyObject *__pyx_int_184506718;
  PyObject *__pyx_int_184517443;
  PyObject *__pyx_int_186494852;
  PyObject *__pyx_int_200977137;
  PyObject *__pyx_int_202585721;
  PyObject *__pyx_int_210616022;
  PyObject *__pyx_int_229249173;
  PyObject *__pyx_int_231591000;
  PyObject *__pyx_int_236313263;
  PyObject *__pyx_int_242200674;
  PyObject *__pyx_int_259995473;
  int64_t __pyx_k__4;
  int64_t __pyx_k__5;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__63;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__75;
  PyObject *__pyx_tuple__77;
  PyObject *__pyx_tuple__79;
  PyObject *__pyx_tuple__81;
  PyObject *__pyx_tuple__82;
  PyObject *__pyx_tuple__84;
  PyObject *__pyx_tuple__86;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__96;
  PyObject *__pyx_tuple__98;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__132;
  PyObject *__pyx_tuple__134;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__149;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__89;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__95;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__106;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__108;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__111;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__117;
  PyObject *__pyx_codeobj__118;
  PyObject *__pyx_codeobj__119;
  PyObject *__pyx_codeobj__120;
  PyObject *__pyx_codeobj__121;
  PyObject *__pyx_codeobj__122;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__124;
  PyObject *__pyx_codeobj__125;
  PyObject *__pyx_codeobj__126;
  PyObject *__pyx_codeobj__127;
  PyObject *__pyx_codeobj__128;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__133;
  PyObject *__pyx_codeobj__135;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__137;
  PyObject *__pyx_codeobj__138;
  PyObject *__pyx_codeobj__139;
  PyObject *__pyx_codeobj__140;
  PyObject *__pyx_codeobj__141;
  PyObject *__pyx_codeobj__142;
  PyObject *__pyx_codeobj__144;
  PyObject *__pyx_codeobj__146;
  PyObject *__pyx_codeobj__147;
  PyObject *__pyx_codeobj__148;
  PyObject *__pyx_codeobj__150;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__152;
  PyObject *__pyx_codeobj__153;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__155;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__157;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__159;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__161;
  PyObject *__pyx_codeobj__162;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_VectorStack);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_VectorStack);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_Instruction);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_Instruction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionVector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionNode);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionStr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionTuple);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionInt);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionIntTuple);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionLabel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionJump);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionFunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_InstructionObjectInt);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_Program);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_Program);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_Function);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_Function);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_LoopSource);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm_LoopSource);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f);
  Py_CLEAR(clear_module_state->__pyx_n_u_Add);
  Py_CLEAR(clear_module_state->__pyx_n_u_Append);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_u_Attributes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_function_return_lnames);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_function_return_stack);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_lnames);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_root_node);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_stack);
  Py_CLEAR(clear_module_state->__pyx_n_u_BeginFor);
  Py_CLEAR(clear_module_state->__pyx_n_u_BranchFalse);
  Py_CLEAR(clear_module_state->__pyx_n_u_BranchTrue);
  Py_CLEAR(clear_module_state->__pyx_n_u_Call);
  Py_CLEAR(clear_module_state->__pyx_n_u_CallFast);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_drop);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_peek);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_poke);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_pop);
  Py_CLEAR(clear_module_state->__pyx_n_u_Ceil);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Circular_import_of);
  Py_CLEAR(clear_module_state->__pyx_n_u_Compose);
  Py_CLEAR(clear_module_state->__pyx_n_s_DYNAMIC_FUNCTIONS);
  Py_CLEAR(clear_module_state->__pyx_n_u_Drop);
  Py_CLEAR(clear_module_state->__pyx_n_u_Dup);
  Py_CLEAR(clear_module_state->__pyx_n_u_EndFor);
  Py_CLEAR(clear_module_state->__pyx_n_u_Eq);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_calling);
  Py_CLEAR(clear_module_state->__pyx_n_u_Exit);
  Py_CLEAR(clear_module_state->__pyx_n_u_Export);
  Py_CLEAR(clear_module_state->__pyx_n_u_Floor);
  Py_CLEAR(clear_module_state->__pyx_n_u_FloorDiv);
  Py_CLEAR(clear_module_state->__pyx_n_u_Fract);
  Py_CLEAR(clear_module_state->__pyx_n_u_Func);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_u_Ge);
  Py_CLEAR(clear_module_state->__pyx_n_u_Gt);
  Py_CLEAR(clear_module_state->__pyx_n_u_Import);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_CLEAR(clear_module_state->__pyx_n_u_IndexLiteral);
  Py_CLEAR(clear_module_state->__pyx_n_s_Instruction);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionFunc);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionFunc___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionFunc___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionInt);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionIntTuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionIntTuple___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionIntTuple___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionInt___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionInt___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionJump);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionJump___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionJump___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionLabel);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionLabel___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionLabel___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionNode);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionNode___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionNode___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionObjectInt);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionObjectInt___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionObjectInt___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionStr);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionStr___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionStr___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionTuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionTuple___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionTuple___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionVector);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionVector___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstructionVector___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Instruction___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Instruction___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_u_Jump);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Jump_outside_of_program);
  Py_CLEAR(clear_module_state->__pyx_kp_u_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_L_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_L_3);
  Py_CLEAR(clear_module_state->__pyx_n_u_Label);
  Py_CLEAR(clear_module_state->__pyx_n_u_Le);
  Py_CLEAR(clear_module_state->__pyx_n_u_Literal);
  Py_CLEAR(clear_module_state->__pyx_n_u_LiteralNode);
  Py_CLEAR(clear_module_state->__pyx_n_u_LiteralNodes);
  Py_CLEAR(clear_module_state->__pyx_n_u_LocalDrop);
  Py_CLEAR(clear_module_state->__pyx_n_u_LocalLoad);
  Py_CLEAR(clear_module_state->__pyx_n_u_LocalPush);
  Py_CLEAR(clear_module_state->__pyx_n_u_Lookup);
  Py_CLEAR(clear_module_state->__pyx_n_u_LookupLiteral);
  Py_CLEAR(clear_module_state->__pyx_n_s_LoopSource);
  Py_CLEAR(clear_module_state->__pyx_n_s_LoopSource___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LoopSource___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_u_Lt);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Missing_lname);
  Py_CLEAR(clear_module_state->__pyx_n_u_Mod);
  Py_CLEAR(clear_module_state->__pyx_n_u_Mul);
  Py_CLEAR(clear_module_state->__pyx_n_u_MulAdd);
  Py_CLEAR(clear_module_state->__pyx_n_s_NOISE_FUNCTIONS);
  Py_CLEAR(clear_module_state->__pyx_n_u_Ne);
  Py_CLEAR(clear_module_state->__pyx_n_u_Neg);
  Py_CLEAR(clear_module_state->__pyx_n_u_Next);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_n_u_Not);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_u_Pos);
  Py_CLEAR(clear_module_state->__pyx_n_u_Pow);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program__execute_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_begin_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_branch_false);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_branch_true);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_call);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_call_fast);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_compose);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_drop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_dup);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_end_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_eq);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_execute);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_export);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_floordiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_fract);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_ge);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_gt);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Program_has_not_been_linked);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_jump);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_le);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_link);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_local_drop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_local_load);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_local_push);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_lookup);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_lookup_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_lt);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_mod);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_mul);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_mul_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_ne);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_neg);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_new_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_next);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_not);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_pos);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_pow);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_set_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_set_pragma);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_set_top);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_slice);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_slice_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_sub);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_truediv);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_use_simplifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_Program_xor);
  Py_CLEAR(clear_module_state->__pyx_n_u_Range);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecursionError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Recursion_depth_exceeded_for_fun);
  Py_CLEAR(clear_module_state->__pyx_n_s_STATIC_FUNCTIONS);
  Py_CLEAR(clear_module_state->__pyx_n_s_SharedCache);
  Py_CLEAR(clear_module_state->__pyx_n_u_Slice);
  Py_CLEAR(clear_module_state->__pyx_n_u_SliceLiteral);
  Py_CLEAR(clear_module_state->__pyx_n_s_StackError);
  Py_CLEAR(clear_module_state->__pyx_n_s_StackOverflow);
  Py_CLEAR(clear_module_state->__pyx_n_s_StackUnderflow);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Stack_cannot_grow_beyond);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Stack_empty);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Stack_overflow_in_program);
  Py_CLEAR(clear_module_state->__pyx_n_u_Sub);
  Py_CLEAR(clear_module_state->__pyx_n_u_Tag);
  Py_CLEAR(clear_module_state->__pyx_n_u_TrueDiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unable_to_import);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unable_to_import_from);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unrecognised_instruction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_VM_exception_processing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_VM_execution_statistics);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_drop);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_peek);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_peek_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_poke);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_poke_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_pop_composed);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_pop_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_pop_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_pop_tuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_VectorStack_push);
  Py_CLEAR(clear_module_state->__pyx_n_u_Xor);
  Py_CLEAR(clear_module_state->__pyx_kp_u__12);
  Py_CLEAR(clear_module_state->__pyx_n_s__163);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_n_s__21);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_begin_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_branch_false);
  Py_CLEAR(clear_module_state->__pyx_n_s_branch_true);
  Py_CLEAR(clear_module_state->__pyx_n_s_builtins);
  Py_CLEAR(clear_module_state->__pyx_n_s_cache);
  Py_CLEAR(clear_module_state->__pyx_n_s_call);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_fast);
  Py_CLEAR(clear_module_state->__pyx_n_s_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_code);
  Py_CLEAR(clear_module_state->__pyx_n_s_compose);
  Py_CLEAR(clear_module_state->__pyx_n_s_context);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_count);
  Py_CLEAR(clear_module_state->__pyx_kp_u_d);
  Py_CLEAR(clear_module_state->__pyx_kp_u_depth_stack);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_drop);
  Py_CLEAR(clear_module_state->__pyx_n_s_dup);
  Py_CLEAR(clear_module_state->__pyx_n_s_duration);
  Py_CLEAR(clear_module_state->__pyx_n_s_dynamic_builtins);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_end_for);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_eq);
  Py_CLEAR(clear_module_state->__pyx_n_s_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_execute);
  Py_CLEAR(clear_module_state->__pyx_n_s_exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_export);
  Py_CLEAR(clear_module_state->__pyx_n_u_false);
  Py_CLEAR(clear_module_state->__pyx_n_s_flitter_language_vm);
  Py_CLEAR(clear_module_state->__pyx_n_s_floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_floordiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_fract);
  Py_CLEAR(clear_module_state->__pyx_kp_u_from);
  Py_CLEAR(clear_module_state->__pyx_n_s_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_function);
  Py_CLEAR(clear_module_state->__pyx_n_s_functions);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_ge);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_with_root);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_gt);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_import_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_inf);
  Py_CLEAR(clear_module_state->__pyx_n_s_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_item);
  Py_CLEAR(clear_module_state->__pyx_kp_u_item_stack);
  Py_CLEAR(clear_module_state->__pyx_kp_u_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_items_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_items_below_top_of);
  Py_CLEAR(clear_module_state->__pyx_kp_u_items_from);
  Py_CLEAR(clear_module_state->__pyx_n_s_ivalue);
  Py_CLEAR(clear_module_state->__pyx_n_s_jump);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_le);
  Py_CLEAR(clear_module_state->__pyx_n_s_link);
  Py_CLEAR(clear_module_state->__pyx_n_s_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_lnames);
  Py_CLEAR(clear_module_state->__pyx_n_s_local_drop);
  Py_CLEAR(clear_module_state->__pyx_n_s_local_load);
  Py_CLEAR(clear_module_state->__pyx_n_s_local_push);
  Py_CLEAR(clear_module_state->__pyx_n_s_log_vm_stats);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_loguru);
  Py_CLEAR(clear_module_state->__pyx_n_s_lookup);
  Py_CLEAR(clear_module_state->__pyx_n_s_lookup_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_lt);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_max_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_mod);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_mul);
  Py_CLEAR(clear_module_state->__pyx_n_s_mul_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_patch);
  Py_CLEAR(clear_module_state->__pyx_n_s_names);
  Py_CLEAR(clear_module_state->__pyx_n_u_nan);
  Py_CLEAR(clear_module_state->__pyx_kp_u_native_funcs);
  Py_CLEAR(clear_module_state->__pyx_n_s_ncaptures);
  Py_CLEAR(clear_module_state->__pyx_n_s_ne);
  Py_CLEAR(clear_module_state->__pyx_n_s_neg);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_next);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_noise);
  Py_CLEAR(clear_module_state->__pyx_n_s_not);
  Py_CLEAR(clear_module_state->__pyx_n_u_null);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_objptr);
  Py_CLEAR(clear_module_state->__pyx_n_s_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_overhead);
  Py_CLEAR(clear_module_state->__pyx_n_s_parameters);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_peek);
  Py_CLEAR(clear_module_state->__pyx_n_s_peek_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_per_execution);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_poke);
  Py_CLEAR(clear_module_state->__pyx_n_s_poke_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop_composed);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop_tuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_pos);
  Py_CLEAR(clear_module_state->__pyx_n_s_pow);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_push);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_capi);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Function);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Instruction);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionFunc);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionInt);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionIntTup);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionJump);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionLabel);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionNode);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionObject);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionStr);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionTuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InstructionVector);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LoopSource);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_flitter_program);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_stats);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_reverse);
  Py_CLEAR(clear_module_state->__pyx_n_u_root);
  Py_CLEAR(clear_module_state->__pyx_n_s_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_pragma);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_top);
  Py_CLEAR(clear_module_state->__pyx_n_s_setdefault);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_simplify);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_slice);
  Py_CLEAR(clear_module_state->__pyx_n_s_slice_literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_sort);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_flitter_language_vm_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_static_builtins);
  Py_CLEAR(clear_module_state->__pyx_n_s_stats);
  Py_CLEAR(clear_module_state->__pyx_n_s_str___locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_sub);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_top);
  Py_CLEAR(clear_module_state->__pyx_n_s_total);
  Py_CLEAR(clear_module_state->__pyx_n_u_true);
  Py_CLEAR(clear_module_state->__pyx_n_s_truediv);
  Py_CLEAR(clear_module_state->__pyx_n_s_tvalue);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_simplifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_vector);
  Py_CLEAR(clear_module_state->__pyx_n_s_xor);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_10904070);
  Py_CLEAR(clear_module_state->__pyx_int_19407059);
  Py_CLEAR(clear_module_state->__pyx_int_29217351);
  Py_CLEAR(clear_module_state->__pyx_int_45504841);
  Py_CLEAR(clear_module_state->__pyx_int_59839126);
  Py_CLEAR(clear_module_state->__pyx_int_61928532);
  Py_CLEAR(clear_module_state->__pyx_int_63712393);
  Py_CLEAR(clear_module_state->__pyx_int_71438266);
  Py_CLEAR(clear_module_state->__pyx_int_72616288);
  Py_CLEAR(clear_module_state->__pyx_int_73689043);
  Py_CLEAR(clear_module_state->__pyx_int_90787080);
  Py_CLEAR(clear_module_state->__pyx_int_107792009);
  Py_CLEAR(clear_module_state->__pyx_int_109158536);
  Py_CLEAR(clear_module_state->__pyx_int_138499235);
  Py_CLEAR(clear_module_state->__pyx_int_163945588);
  Py_CLEAR(clear_module_state->__pyx_int_176342959);
  Py_CLEAR(clear_module_state->__pyx_int_184506718);
  Py_CLEAR(clear_module_state->__pyx_int_184517443);
  Py_CLEAR(clear_module_state->__pyx_int_186494852);
  Py_CLEAR(clear_module_state->__pyx_int_200977137);
  Py_CLEAR(clear_module_state->__pyx_int_202585721);
  Py_CLEAR(clear_module_state->__pyx_int_210616022);
  Py_CLEAR(clear_module_state->__pyx_int_229249173);
  Py_CLEAR(clear_module_state->__pyx_int_231591000);
  Py_CLEAR(clear_module_state->__pyx_int_236313263);
  Py_CLEAR(clear_module_state->__pyx_int_242200674);
  Py_CLEAR(clear_module_state->__pyx_int_259995473);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__63);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__75);
  Py_CLEAR(clear_module_state->__pyx_tuple__77);
  Py_CLEAR(clear_module_state->__pyx_tuple__79);
  Py_CLEAR(clear_module_state->__pyx_tuple__81);
  Py_CLEAR(clear_module_state->__pyx_tuple__82);
  Py_CLEAR(clear_module_state->__pyx_tuple__84);
  Py_CLEAR(clear_module_state->__pyx_tuple__86);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__96);
  Py_CLEAR(clear_module_state->__pyx_tuple__98);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__132);
  Py_CLEAR(clear_module_state->__pyx_tuple__134);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__149);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__89);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__106);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__108);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__111);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  Py_CLEAR(clear_module_state->__pyx_codeobj__118);
  Py_CLEAR(clear_module_state->__pyx_codeobj__119);
  Py_CLEAR(clear_module_state->__pyx_codeobj__120);
  Py_CLEAR(clear_module_state->__pyx_codeobj__121);
  Py_CLEAR(clear_module_state->__pyx_codeobj__122);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__124);
  Py_CLEAR(clear_module_state->__pyx_codeobj__125);
  Py_CLEAR(clear_module_state->__pyx_codeobj__126);
  Py_CLEAR(clear_module_state->__pyx_codeobj__127);
  Py_CLEAR(clear_module_state->__pyx_codeobj__128);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__133);
  Py_CLEAR(clear_module_state->__pyx_codeobj__135);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__137);
  Py_CLEAR(clear_module_state->__pyx_codeobj__138);
  Py_CLEAR(clear_module_state->__pyx_codeobj__139);
  Py_CLEAR(clear_module_state->__pyx_codeobj__140);
  Py_CLEAR(clear_module_state->__pyx_codeobj__141);
  Py_CLEAR(clear_module_state->__pyx_codeobj__142);
  Py_CLEAR(clear_module_state->__pyx_codeobj__144);
  Py_CLEAR(clear_module_state->__pyx_codeobj__146);
  Py_CLEAR(clear_module_state->__pyx_codeobj__147);
  Py_CLEAR(clear_module_state->__pyx_codeobj__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__150);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__152);
  Py_CLEAR(clear_module_state->__pyx_codeobj__153);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__155);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__157);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__159);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__161);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_VectorStack);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_VectorStack);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_Instruction);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_Instruction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionVector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionNode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionStr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionTuple);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionInt);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionIntTuple);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionLabel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionJump);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionFunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_InstructionObjectInt);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_Program);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_Program);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_Function);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_Function);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_LoopSource);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm_LoopSource);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f);
  Py_VISIT(traverse_module_state->__pyx_n_u_Add);
  Py_VISIT(traverse_module_state->__pyx_n_u_Append);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_u_Attributes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_function_return_lnames);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_function_return_stack);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_lnames);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_root_node);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_stack);
  Py_VISIT(traverse_module_state->__pyx_n_u_BeginFor);
  Py_VISIT(traverse_module_state->__pyx_n_u_BranchFalse);
  Py_VISIT(traverse_module_state->__pyx_n_u_BranchTrue);
  Py_VISIT(traverse_module_state->__pyx_n_u_Call);
  Py_VISIT(traverse_module_state->__pyx_n_u_CallFast);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_drop);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_peek);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_poke);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_pop);
  Py_VISIT(traverse_module_state->__pyx_n_u_Ceil);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Circular_import_of);
  Py_VISIT(traverse_module_state->__pyx_n_u_Compose);
  Py_VISIT(traverse_module_state->__pyx_n_s_DYNAMIC_FUNCTIONS);
  Py_VISIT(traverse_module_state->__pyx_n_u_Drop);
  Py_VISIT(traverse_module_state->__pyx_n_u_Dup);
  Py_VISIT(traverse_module_state->__pyx_n_u_EndFor);
  Py_VISIT(traverse_module_state->__pyx_n_u_Eq);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_calling);
  Py_VISIT(traverse_module_state->__pyx_n_u_Exit);
  Py_VISIT(traverse_module_state->__pyx_n_u_Export);
  Py_VISIT(traverse_module_state->__pyx_n_u_Floor);
  Py_VISIT(traverse_module_state->__pyx_n_u_FloorDiv);
  Py_VISIT(traverse_module_state->__pyx_n_u_Fract);
  Py_VISIT(traverse_module_state->__pyx_n_u_Func);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_u_Ge);
  Py_VISIT(traverse_module_state->__pyx_n_u_Gt);
  Py_VISIT(traverse_module_state->__pyx_n_u_Import);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_VISIT(traverse_module_state->__pyx_n_u_IndexLiteral);
  Py_VISIT(traverse_module_state->__pyx_n_s_Instruction);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionFunc);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionFunc___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionFunc___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionInt);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionIntTuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionIntTuple___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionIntTuple___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionInt___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionInt___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionJump);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionJump___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionJump___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionLabel);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionLabel___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionLabel___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionNode);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionNode___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionNode___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionObjectInt);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionObjectInt___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionObjectInt___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionStr);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionStr___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionStr___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionTuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionTuple___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionTuple___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionVector);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionVector___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstructionVector___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Instruction___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Instruction___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_u_Jump);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Jump_outside_of_program);
  Py_VISIT(traverse_module_state->__pyx_kp_u_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_L_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_L_3);
  Py_VISIT(traverse_module_state->__pyx_n_u_Label);
  Py_VISIT(traverse_module_state->__pyx_n_u_Le);
  Py_VISIT(traverse_module_state->__pyx_n_u_Literal);
  Py_VISIT(traverse_module_state->__pyx_n_u_LiteralNode);
  Py_VISIT(traverse_module_state->__pyx_n_u_LiteralNodes);
  Py_VISIT(traverse_module_state->__pyx_n_u_LocalDrop);
  Py_VISIT(traverse_module_state->__pyx_n_u_LocalLoad);
  Py_VISIT(traverse_module_state->__pyx_n_u_LocalPush);
  Py_VISIT(traverse_module_state->__pyx_n_u_Lookup);
  Py_VISIT(traverse_module_state->__pyx_n_u_LookupLiteral);
  Py_VISIT(traverse_module_state->__pyx_n_s_LoopSource);
  Py_VISIT(traverse_module_state->__pyx_n_s_LoopSource___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LoopSource___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_u_Lt);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Missing_lname);
  Py_VISIT(traverse_module_state->__pyx_n_u_Mod);
  Py_VISIT(traverse_module_state->__pyx_n_u_Mul);
  Py_VISIT(traverse_module_state->__pyx_n_u_MulAdd);
  Py_VISIT(traverse_module_state->__pyx_n_s_NOISE_FUNCTIONS);
  Py_VISIT(traverse_module_state->__pyx_n_u_Ne);
  Py_VISIT(traverse_module_state->__pyx_n_u_Neg);
  Py_VISIT(traverse_module_state->__pyx_n_u_Next);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_n_u_Not);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_u_Pos);
  Py_VISIT(traverse_module_state->__pyx_n_u_Pow);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program__execute_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_begin_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_branch_false);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_branch_true);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_call);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_call_fast);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_compose);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_drop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_dup);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_end_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_eq);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_execute);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_export);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_floordiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_fract);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_ge);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_gt);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Program_has_not_been_linked);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_jump);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_le);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_link);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_local_drop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_local_load);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_local_push);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_lookup);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_lookup_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_lt);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_mod);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_mul);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_mul_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_ne);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_neg);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_new_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_next);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_not);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_pos);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_pow);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_set_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_set_pragma);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_set_top);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_slice);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_slice_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_sub);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_truediv);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_use_simplifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_Program_xor);
  Py_VISIT(traverse_module_state->__pyx_n_u_Range);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecursionError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Recursion_depth_exceeded_for_fun);
  Py_VISIT(traverse_module_state->__pyx_n_s_STATIC_FUNCTIONS);
  Py_VISIT(traverse_module_state->__pyx_n_s_SharedCache);
  Py_VISIT(traverse_module_state->__pyx_n_u_Slice);
  Py_VISIT(traverse_module_state->__pyx_n_u_SliceLiteral);
  Py_VISIT(traverse_module_state->__pyx_n_s_StackError);
  Py_VISIT(traverse_module_state->__pyx_n_s_StackOverflow);
  Py_VISIT(traverse_module_state->__pyx_n_s_StackUnderflow);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Stack_cannot_grow_beyond);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Stack_empty);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Stack_overflow_in_program);
  Py_VISIT(traverse_module_state->__pyx_n_u_Sub);
  Py_VISIT(traverse_module_state->__pyx_n_u_Tag);
  Py_VISIT(traverse_module_state->__pyx_n_u_TrueDiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unable_to_import);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unable_to_import_from);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unrecognised_instruction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_VM_exception_processing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_VM_execution_statistics);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_drop);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_peek);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_peek_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_poke);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_poke_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_pop_composed);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_pop_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_pop_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_pop_tuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_VectorStack_push);
  Py_VISIT(traverse_module_state->__pyx_n_u_Xor);
  Py_VISIT(traverse_module_state->__pyx_kp_u__12);
  Py_VISIT(traverse_module_state->__pyx_n_s__163);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_n_s__21);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_begin_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_branch_false);
  Py_VISIT(traverse_module_state->__pyx_n_s_branch_true);
  Py_VISIT(traverse_module_state->__pyx_n_s_builtins);
  Py_VISIT(traverse_module_state->__pyx_n_s_cache);
  Py_VISIT(traverse_module_state->__pyx_n_s_call);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_fast);
  Py_VISIT(traverse_module_state->__pyx_n_s_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_code);
  Py_VISIT(traverse_module_state->__pyx_n_s_compose);
  Py_VISIT(traverse_module_state->__pyx_n_s_context);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_count);
  Py_VISIT(traverse_module_state->__pyx_kp_u_d);
  Py_VISIT(traverse_module_state->__pyx_kp_u_depth_stack);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_drop);
  Py_VISIT(traverse_module_state->__pyx_n_s_dup);
  Py_VISIT(traverse_module_state->__pyx_n_s_duration);
  Py_VISIT(traverse_module_state->__pyx_n_s_dynamic_builtins);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_end_for);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_eq);
  Py_VISIT(traverse_module_state->__pyx_n_s_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_execute);
  Py_VISIT(traverse_module_state->__pyx_n_s_exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_export);
  Py_VISIT(traverse_module_state->__pyx_n_u_false);
  Py_VISIT(traverse_module_state->__pyx_n_s_flitter_language_vm);
  Py_VISIT(traverse_module_state->__pyx_n_s_floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_floordiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_fract);
  Py_VISIT(traverse_module_state->__pyx_kp_u_from);
  Py_VISIT(traverse_module_state->__pyx_n_s_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_function);
  Py_VISIT(traverse_module_state->__pyx_n_s_functions);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_ge);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_with_root);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_gt);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_import_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_inf);
  Py_VISIT(traverse_module_state->__pyx_n_s_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_item);
  Py_VISIT(traverse_module_state->__pyx_kp_u_item_stack);
  Py_VISIT(traverse_module_state->__pyx_kp_u_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_items_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_items_below_top_of);
  Py_VISIT(traverse_module_state->__pyx_kp_u_items_from);
  Py_VISIT(traverse_module_state->__pyx_n_s_ivalue);
  Py_VISIT(traverse_module_state->__pyx_n_s_jump);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_le);
  Py_VISIT(traverse_module_state->__pyx_n_s_link);
  Py_VISIT(traverse_module_state->__pyx_n_s_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_lnames);
  Py_VISIT(traverse_module_state->__pyx_n_s_local_drop);
  Py_VISIT(traverse_module_state->__pyx_n_s_local_load);
  Py_VISIT(traverse_module_state->__pyx_n_s_local_push);
  Py_VISIT(traverse_module_state->__pyx_n_s_log_vm_stats);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_loguru);
  Py_VISIT(traverse_module_state->__pyx_n_s_lookup);
  Py_VISIT(traverse_module_state->__pyx_n_s_lookup_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_lt);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_max_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_mod);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_mul);
  Py_VISIT(traverse_module_state->__pyx_n_s_mul_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_patch);
  Py_VISIT(traverse_module_state->__pyx_n_s_names);
  Py_VISIT(traverse_module_state->__pyx_n_u_nan);
  Py_VISIT(traverse_module_state->__pyx_kp_u_native_funcs);
  Py_VISIT(traverse_module_state->__pyx_n_s_ncaptures);
  Py_VISIT(traverse_module_state->__pyx_n_s_ne);
  Py_VISIT(traverse_module_state->__pyx_n_s_neg);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_next);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_noise);
  Py_VISIT(traverse_module_state->__pyx_n_s_not);
  Py_VISIT(traverse_module_state->__pyx_n_u_null);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_objptr);
  Py_VISIT(traverse_module_state->__pyx_n_s_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_overhead);
  Py_VISIT(traverse_module_state->__pyx_n_s_parameters);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_peek);
  Py_VISIT(traverse_module_state->__pyx_n_s_peek_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_per_execution);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_poke);
  Py_VISIT(traverse_module_state->__pyx_n_s_poke_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop_composed);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop_tuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_pos);
  Py_VISIT(traverse_module_state->__pyx_n_s_pow);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_push);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_capi);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Function);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Instruction);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionFunc);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionInt);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionIntTup);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionJump);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionLabel);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionNode);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionObject);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionStr);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionTuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InstructionVector);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LoopSource);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_flitter_program);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_stats);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_reverse);
  Py_VISIT(traverse_module_state->__pyx_n_u_root);
  Py_VISIT(traverse_module_state->__pyx_n_s_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_pragma);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_top);
  Py_VISIT(traverse_module_state->__pyx_n_s_setdefault);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_simplify);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_slice);
  Py_VISIT(traverse_module_state->__pyx_n_s_slice_literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_sort);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_flitter_language_vm_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_static_builtins);
  Py_VISIT(traverse_module_state->__pyx_n_s_stats);
  Py_VISIT(traverse_module_state->__pyx_n_s_str___locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_sub);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_top);
  Py_VISIT(traverse_module_state->__pyx_n_s_total);
  Py_VISIT(traverse_module_state->__pyx_n_u_true);
  Py_VISIT(traverse_module_state->__pyx_n_s_truediv);
  Py_VISIT(traverse_module_state->__pyx_n_s_tvalue);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_simplifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_vector);
  Py_VISIT(traverse_module_state->__pyx_n_s_xor);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_10904070);
  Py_VISIT(traverse_module_state->__pyx_int_19407059);
  Py_VISIT(traverse_module_state->__pyx_int_29217351);
  Py_VISIT(traverse_module_state->__pyx_int_45504841);
  Py_VISIT(traverse_module_state->__pyx_int_59839126);
  Py_VISIT(traverse_module_state->__pyx_int_61928532);
  Py_VISIT(traverse_module_state->__pyx_int_63712393);
  Py_VISIT(traverse_module_state->__pyx_int_71438266);
  Py_VISIT(traverse_module_state->__pyx_int_72616288);
  Py_VISIT(traverse_module_state->__pyx_int_73689043);
  Py_VISIT(traverse_module_state->__pyx_int_90787080);
  Py_VISIT(traverse_module_state->__pyx_int_107792009);
  Py_VISIT(traverse_module_state->__pyx_int_109158536);
  Py_VISIT(traverse_module_state->__pyx_int_138499235);
  Py_VISIT(traverse_module_state->__pyx_int_163945588);
  Py_VISIT(traverse_module_state->__pyx_int_176342959);
  Py_VISIT(traverse_module_state->__pyx_int_184506718);
  Py_VISIT(traverse_module_state->__pyx_int_184517443);
  Py_VISIT(traverse_module_state->__pyx_int_186494852);
  Py_VISIT(traverse_module_state->__pyx_int_200977137);
  Py_VISIT(traverse_module_state->__pyx_int_202585721);
  Py_VISIT(traverse_module_state->__pyx_int_210616022);
  Py_VISIT(traverse_module_state->__pyx_int_229249173);
  Py_VISIT(traverse_module_state->__pyx_int_231591000);
  Py_VISIT(traverse_module_state->__pyx_int_236313263);
  Py_VISIT(traverse_module_state->__pyx_int_242200674);
  Py_VISIT(traverse_module_state->__pyx_int_259995473);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__63);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__75);
  Py_VISIT(traverse_module_state->__pyx_tuple__77);
  Py_VISIT(traverse_module_state->__pyx_tuple__79);
  Py_VISIT(traverse_module_state->__pyx_tuple__81);
  Py_VISIT(traverse_module_state->__pyx_tuple__82);
  Py_VISIT(traverse_module_state->__pyx_tuple__84);
  Py_VISIT(traverse_module_state->__pyx_tuple__86);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__96);
  Py_VISIT(traverse_module_state->__pyx_tuple__98);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__132);
  Py_VISIT(traverse_module_state->__pyx_tuple__134);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__149);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__89);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__106);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__108);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__111);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  Py_VISIT(traverse_module_state->__pyx_codeobj__118);
  Py_VISIT(traverse_module_state->__pyx_codeobj__119);
  Py_VISIT(traverse_module_state->__pyx_codeobj__120);
  Py_VISIT(traverse_module_state->__pyx_codeobj__121);
  Py_VISIT(traverse_module_state->__pyx_codeobj__122);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__124);
  Py_VISIT(traverse_module_state->__pyx_codeobj__125);
  Py_VISIT(traverse_module_state->__pyx_codeobj__126);
  Py_VISIT(traverse_module_state->__pyx_codeobj__127);
  Py_VISIT(traverse_module_state->__pyx_codeobj__128);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__133);
  Py_VISIT(traverse_module_state->__pyx_codeobj__135);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__137);
  Py_VISIT(traverse_module_state->__pyx_codeobj__138);
  Py_VISIT(traverse_module_state->__pyx_codeobj__139);
  Py_VISIT(traverse_module_state->__pyx_codeobj__140);
  Py_VISIT(traverse_module_state->__pyx_codeobj__141);
  Py_VISIT(traverse_module_state->__pyx_codeobj__142);
  Py_VISIT(traverse_module_state->__pyx_codeobj__144);
  Py_VISIT(traverse_module_state->__pyx_codeobj__146);
  Py_VISIT(traverse_module_state->__pyx_codeobj__147);
  Py_VISIT(traverse_module_state->__pyx_codeobj__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__150);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__152);
  Py_VISIT(traverse_module_state->__pyx_codeobj__153);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__155);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__157);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__159);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__161);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7flitter_5model_Vector __pyx_mstate_global->__pyx_ptype_7flitter_5model_Vector
#define __pyx_ptype_7flitter_5model_Matrix33 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix33
#define __pyx_ptype_7flitter_5model_Matrix44 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix44
#define __pyx_ptype_7flitter_5model_Quaternion __pyx_mstate_global->__pyx_ptype_7flitter_5model_Quaternion
#define __pyx_ptype_7flitter_5model_Node __pyx_mstate_global->__pyx_ptype_7flitter_5model_Node
#define __pyx_ptype_7flitter_5model_StateDict __pyx_mstate_global->__pyx_ptype_7flitter_5model_StateDict
#define __pyx_ptype_7flitter_5model_Context __pyx_mstate_global->__pyx_ptype_7flitter_5model_Context
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_7flitter_8language_2vm_VectorStack __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_VectorStack
#define __pyx_type_7flitter_8language_2vm_Instruction __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_Instruction
#define __pyx_type_7flitter_8language_2vm_InstructionVector __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionVector
#define __pyx_type_7flitter_8language_2vm_InstructionNode __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionNode
#define __pyx_type_7flitter_8language_2vm_InstructionStr __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionStr
#define __pyx_type_7flitter_8language_2vm_InstructionTuple __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionTuple
#define __pyx_type_7flitter_8language_2vm_InstructionInt __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionInt
#define __pyx_type_7flitter_8language_2vm_InstructionIntTuple __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionIntTuple
#define __pyx_type_7flitter_8language_2vm_InstructionLabel __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionLabel
#define __pyx_type_7flitter_8language_2vm_InstructionJump __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionJump
#define __pyx_type_7flitter_8language_2vm_InstructionFunc __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionFunc
#define __pyx_type_7flitter_8language_2vm_InstructionObjectInt __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_InstructionObjectInt
#define __pyx_type_7flitter_8language_2vm_Program __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_Program
#define __pyx_type_7flitter_8language_2vm_Function __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_Function
#define __pyx_type_7flitter_8language_2vm_LoopSource __pyx_mstate_global->__pyx_type_7flitter_8language_2vm_LoopSource
#define __pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr
#define __pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute __pyx_mstate_global->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute
#define __pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr
#endif
#define __pyx_ptype_7flitter_8language_2vm_VectorStack __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_VectorStack
#define __pyx_ptype_7flitter_8language_2vm_Instruction __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_Instruction
#define __pyx_ptype_7flitter_8language_2vm_InstructionVector __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionVector
#define __pyx_ptype_7flitter_8language_2vm_InstructionNode __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionNode
#define __pyx_ptype_7flitter_8language_2vm_InstructionStr __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionStr
#define __pyx_ptype_7flitter_8language_2vm_InstructionTuple __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionTuple
#define __pyx_ptype_7flitter_8language_2vm_InstructionInt __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionInt
#define __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple
#define __pyx_ptype_7flitter_8language_2vm_InstructionLabel __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionLabel
#define __pyx_ptype_7flitter_8language_2vm_InstructionJump __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionJump
#define __pyx_ptype_7flitter_8language_2vm_InstructionFunc __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionFunc
#define __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt
#define __pyx_ptype_7flitter_8language_2vm_Program __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_Program
#define __pyx_ptype_7flitter_8language_2vm_Function __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_Function
#define __pyx_ptype_7flitter_8language_2vm_LoopSource __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_LoopSource
#define __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr
#define __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute
#define __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f __pyx_mstate_global->__pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f
#define __pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f __pyx_mstate_global->__pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f
#define __pyx_n_u_Add __pyx_mstate_global->__pyx_n_u_Add
#define __pyx_n_u_Append __pyx_mstate_global->__pyx_n_u_Append
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_u_Attributes __pyx_mstate_global->__pyx_n_u_Attributes
#define __pyx_kp_u_Bad_function_return_lnames __pyx_mstate_global->__pyx_kp_u_Bad_function_return_lnames
#define __pyx_kp_u_Bad_function_return_stack __pyx_mstate_global->__pyx_kp_u_Bad_function_return_stack
#define __pyx_kp_u_Bad_lnames __pyx_mstate_global->__pyx_kp_u_Bad_lnames
#define __pyx_kp_u_Bad_root_node __pyx_mstate_global->__pyx_kp_u_Bad_root_node
#define __pyx_kp_u_Bad_stack __pyx_mstate_global->__pyx_kp_u_Bad_stack
#define __pyx_n_u_BeginFor __pyx_mstate_global->__pyx_n_u_BeginFor
#define __pyx_n_u_BranchFalse __pyx_mstate_global->__pyx_n_u_BranchFalse
#define __pyx_n_u_BranchTrue __pyx_mstate_global->__pyx_n_u_BranchTrue
#define __pyx_n_u_Call __pyx_mstate_global->__pyx_n_u_Call
#define __pyx_n_u_CallFast __pyx_mstate_global->__pyx_n_u_CallFast
#define __pyx_kp_u_Cannot_drop __pyx_mstate_global->__pyx_kp_u_Cannot_drop
#define __pyx_kp_u_Cannot_peek __pyx_mstate_global->__pyx_kp_u_Cannot_peek
#define __pyx_kp_u_Cannot_poke __pyx_mstate_global->__pyx_kp_u_Cannot_poke
#define __pyx_kp_u_Cannot_pop __pyx_mstate_global->__pyx_kp_u_Cannot_pop
#define __pyx_n_u_Ceil __pyx_mstate_global->__pyx_n_u_Ceil
#define __pyx_kp_u_Circular_import_of __pyx_mstate_global->__pyx_kp_u_Circular_import_of
#define __pyx_n_u_Compose __pyx_mstate_global->__pyx_n_u_Compose
#define __pyx_n_s_DYNAMIC_FUNCTIONS __pyx_mstate_global->__pyx_n_s_DYNAMIC_FUNCTIONS
#define __pyx_n_u_Drop __pyx_mstate_global->__pyx_n_u_Drop
#define __pyx_n_u_Dup __pyx_mstate_global->__pyx_n_u_Dup
#define __pyx_n_u_EndFor __pyx_mstate_global->__pyx_n_u_EndFor
#define __pyx_n_u_Eq __pyx_mstate_global->__pyx_n_u_Eq
#define __pyx_kp_u_Error_calling __pyx_mstate_global->__pyx_kp_u_Error_calling
#define __pyx_n_u_Exit __pyx_mstate_global->__pyx_n_u_Exit
#define __pyx_n_u_Export __pyx_mstate_global->__pyx_n_u_Export
#define __pyx_n_u_Floor __pyx_mstate_global->__pyx_n_u_Floor
#define __pyx_n_u_FloorDiv __pyx_mstate_global->__pyx_n_u_FloorDiv
#define __pyx_n_u_Fract __pyx_mstate_global->__pyx_n_u_Fract
#define __pyx_n_u_Func __pyx_mstate_global->__pyx_n_u_Func
#define __pyx_n_s_Function __pyx_mstate_global->__pyx_n_s_Function
#define __pyx_n_s_Function___reduce_cython __pyx_mstate_global->__pyx_n_s_Function___reduce_cython
#define __pyx_n_s_Function___setstate_cython __pyx_mstate_global->__pyx_n_s_Function___setstate_cython
#define __pyx_n_u_Ge __pyx_mstate_global->__pyx_n_u_Ge
#define __pyx_n_u_Gt __pyx_mstate_global->__pyx_n_u_Gt
#define __pyx_n_u_Import __pyx_mstate_global->__pyx_n_u_Import
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9
#define __pyx_n_u_IndexLiteral __pyx_mstate_global->__pyx_n_u_IndexLiteral
#define __pyx_n_s_Instruction __pyx_mstate_global->__pyx_n_s_Instruction
#define __pyx_n_s_InstructionFunc __pyx_mstate_global->__pyx_n_s_InstructionFunc
#define __pyx_n_s_InstructionFunc___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionFunc___reduce_cython
#define __pyx_n_s_InstructionFunc___setstate_cytho __pyx_mstate_global->__pyx_n_s_InstructionFunc___setstate_cytho
#define __pyx_n_s_InstructionInt __pyx_mstate_global->__pyx_n_s_InstructionInt
#define __pyx_n_s_InstructionIntTuple __pyx_mstate_global->__pyx_n_s_InstructionIntTuple
#define __pyx_n_s_InstructionIntTuple___reduce_cyt __pyx_mstate_global->__pyx_n_s_InstructionIntTuple___reduce_cyt
#define __pyx_n_s_InstructionIntTuple___setstate_c __pyx_mstate_global->__pyx_n_s_InstructionIntTuple___setstate_c
#define __pyx_n_s_InstructionInt___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionInt___reduce_cython
#define __pyx_n_s_InstructionInt___setstate_cython __pyx_mstate_global->__pyx_n_s_InstructionInt___setstate_cython
#define __pyx_n_s_InstructionJump __pyx_mstate_global->__pyx_n_s_InstructionJump
#define __pyx_n_s_InstructionJump___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionJump___reduce_cython
#define __pyx_n_s_InstructionJump___setstate_cytho __pyx_mstate_global->__pyx_n_s_InstructionJump___setstate_cytho
#define __pyx_n_s_InstructionLabel __pyx_mstate_global->__pyx_n_s_InstructionLabel
#define __pyx_n_s_InstructionLabel___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionLabel___reduce_cython
#define __pyx_n_s_InstructionLabel___setstate_cyth __pyx_mstate_global->__pyx_n_s_InstructionLabel___setstate_cyth
#define __pyx_n_s_InstructionNode __pyx_mstate_global->__pyx_n_s_InstructionNode
#define __pyx_n_s_InstructionNode___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionNode___reduce_cython
#define __pyx_n_s_InstructionNode___setstate_cytho __pyx_mstate_global->__pyx_n_s_InstructionNode___setstate_cytho
#define __pyx_n_s_InstructionObjectInt __pyx_mstate_global->__pyx_n_s_InstructionObjectInt
#define __pyx_n_s_InstructionObjectInt___reduce_cy __pyx_mstate_global->__pyx_n_s_InstructionObjectInt___reduce_cy
#define __pyx_n_s_InstructionObjectInt___setstate __pyx_mstate_global->__pyx_n_s_InstructionObjectInt___setstate
#define __pyx_n_s_InstructionStr __pyx_mstate_global->__pyx_n_s_InstructionStr
#define __pyx_n_s_InstructionStr___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionStr___reduce_cython
#define __pyx_n_s_InstructionStr___setstate_cython __pyx_mstate_global->__pyx_n_s_InstructionStr___setstate_cython
#define __pyx_n_s_InstructionTuple __pyx_mstate_global->__pyx_n_s_InstructionTuple
#define __pyx_n_s_InstructionTuple___reduce_cython __pyx_mstate_global->__pyx_n_s_InstructionTuple___reduce_cython
#define __pyx_n_s_InstructionTuple___setstate_cyth __pyx_mstate_global->__pyx_n_s_InstructionTuple___setstate_cyth
#define __pyx_n_s_InstructionVector __pyx_mstate_global->__pyx_n_s_InstructionVector
#define __pyx_n_s_InstructionVector___reduce_cytho __pyx_mstate_global->__pyx_n_s_InstructionVector___reduce_cytho
#define __pyx_n_s_InstructionVector___setstate_cyt __pyx_mstate_global->__pyx_n_s_InstructionVector___setstate_cyt
#define __pyx_n_s_Instruction___reduce_cython __pyx_mstate_global->__pyx_n_s_Instruction___reduce_cython
#define __pyx_n_s_Instruction___setstate_cython __pyx_mstate_global->__pyx_n_s_Instruction___setstate_cython
#define __pyx_n_u_Jump __pyx_mstate_global->__pyx_n_u_Jump
#define __pyx_kp_u_Jump_outside_of_program __pyx_mstate_global->__pyx_kp_u_Jump_outside_of_program
#define __pyx_kp_u_L __pyx_mstate_global->__pyx_kp_u_L
#define __pyx_kp_u_L_2 __pyx_mstate_global->__pyx_kp_u_L_2
#define __pyx_kp_u_L_3 __pyx_mstate_global->__pyx_kp_u_L_3
#define __pyx_n_u_Label __pyx_mstate_global->__pyx_n_u_Label
#define __pyx_n_u_Le __pyx_mstate_global->__pyx_n_u_Le
#define __pyx_n_u_Literal __pyx_mstate_global->__pyx_n_u_Literal
#define __pyx_n_u_LiteralNode __pyx_mstate_global->__pyx_n_u_LiteralNode
#define __pyx_n_u_LiteralNodes __pyx_mstate_global->__pyx_n_u_LiteralNodes
#define __pyx_n_u_LocalDrop __pyx_mstate_global->__pyx_n_u_LocalDrop
#define __pyx_n_u_LocalLoad __pyx_mstate_global->__pyx_n_u_LocalLoad
#define __pyx_n_u_LocalPush __pyx_mstate_global->__pyx_n_u_LocalPush
#define __pyx_n_u_Lookup __pyx_mstate_global->__pyx_n_u_Lookup
#define __pyx_n_u_LookupLiteral __pyx_mstate_global->__pyx_n_u_LookupLiteral
#define __pyx_n_s_LoopSource __pyx_mstate_global->__pyx_n_s_LoopSource
#define __pyx_n_s_LoopSource___reduce_cython __pyx_mstate_global->__pyx_n_s_LoopSource___reduce_cython
#define __pyx_n_s_LoopSource___setstate_cython __pyx_mstate_global->__pyx_n_s_LoopSource___setstate_cython
#define __pyx_n_u_Lt __pyx_mstate_global->__pyx_n_u_Lt
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_kp_u_Missing_lname __pyx_mstate_global->__pyx_kp_u_Missing_lname
#define __pyx_n_u_Mod __pyx_mstate_global->__pyx_n_u_Mod
#define __pyx_n_u_Mul __pyx_mstate_global->__pyx_n_u_Mul
#define __pyx_n_u_MulAdd __pyx_mstate_global->__pyx_n_u_MulAdd
#define __pyx_n_s_NOISE_FUNCTIONS __pyx_mstate_global->__pyx_n_s_NOISE_FUNCTIONS
#define __pyx_n_u_Ne __pyx_mstate_global->__pyx_n_u_Ne
#define __pyx_n_u_Neg __pyx_mstate_global->__pyx_n_u_Neg
#define __pyx_n_u_Next __pyx_mstate_global->__pyx_n_u_Next
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_n_u_Not __pyx_mstate_global->__pyx_n_u_Not
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_u_Pos __pyx_mstate_global->__pyx_n_u_Pos
#define __pyx_n_u_Pow __pyx_mstate_global->__pyx_n_u_Pow
#define __pyx_n_s_Program __pyx_mstate_global->__pyx_n_s_Program
#define __pyx_n_s_Program___reduce_cython __pyx_mstate_global->__pyx_n_s_Program___reduce_cython
#define __pyx_n_s_Program___setstate_cython __pyx_mstate_global->__pyx_n_s_Program___setstate_cython
#define __pyx_n_s_Program__execute_locals_genexpr __pyx_mstate_global->__pyx_n_s_Program__execute_locals_genexpr
#define __pyx_n_s_Program_add __pyx_mstate_global->__pyx_n_s_Program_add
#define __pyx_n_s_Program_append __pyx_mstate_global->__pyx_n_s_Program_append
#define __pyx_n_s_Program_attributes __pyx_mstate_global->__pyx_n_s_Program_attributes
#define __pyx_n_s_Program_begin_for __pyx_mstate_global->__pyx_n_s_Program_begin_for
#define __pyx_n_s_Program_branch_false __pyx_mstate_global->__pyx_n_s_Program_branch_false
#define __pyx_n_s_Program_branch_true __pyx_mstate_global->__pyx_n_s_Program_branch_true
#define __pyx_n_s_Program_call __pyx_mstate_global->__pyx_n_s_Program_call
#define __pyx_n_s_Program_call_fast __pyx_mstate_global->__pyx_n_s_Program_call_fast
#define __pyx_n_s_Program_ceil __pyx_mstate_global->__pyx_n_s_Program_ceil
#define __pyx_n_s_Program_compose __pyx_mstate_global->__pyx_n_s_Program_compose
#define __pyx_n_s_Program_drop __pyx_mstate_global->__pyx_n_s_Program_drop
#define __pyx_n_s_Program_dup __pyx_mstate_global->__pyx_n_s_Program_dup
#define __pyx_n_s_Program_end_for __pyx_mstate_global->__pyx_n_s_Program_end_for
#define __pyx_n_s_Program_eq __pyx_mstate_global->__pyx_n_s_Program_eq
#define __pyx_n_s_Program_execute __pyx_mstate_global->__pyx_n_s_Program_execute
#define __pyx_n_s_Program_exit __pyx_mstate_global->__pyx_n_s_Program_exit
#define __pyx_n_s_Program_export __pyx_mstate_global->__pyx_n_s_Program_export
#define __pyx_n_s_Program_floor __pyx_mstate_global->__pyx_n_s_Program_floor
#define __pyx_n_s_Program_floordiv __pyx_mstate_global->__pyx_n_s_Program_floordiv
#define __pyx_n_s_Program_fract __pyx_mstate_global->__pyx_n_s_Program_fract
#define __pyx_n_s_Program_func __pyx_mstate_global->__pyx_n_s_Program_func
#define __pyx_n_s_Program_ge __pyx_mstate_global->__pyx_n_s_Program_ge
#define __pyx_n_s_Program_gt __pyx_mstate_global->__pyx_n_s_Program_gt
#define __pyx_kp_u_Program_has_not_been_linked __pyx_mstate_global->__pyx_kp_u_Program_has_not_been_linked
#define __pyx_n_s_Program_import __pyx_mstate_global->__pyx_n_s_Program_import
#define __pyx_n_s_Program_jump __pyx_mstate_global->__pyx_n_s_Program_jump
#define __pyx_n_s_Program_label __pyx_mstate_global->__pyx_n_s_Program_label
#define __pyx_n_s_Program_le __pyx_mstate_global->__pyx_n_s_Program_le
#define __pyx_n_s_Program_link __pyx_mstate_global->__pyx_n_s_Program_link
#define __pyx_n_s_Program_literal __pyx_mstate_global->__pyx_n_s_Program_literal
#define __pyx_n_s_Program_local_drop __pyx_mstate_global->__pyx_n_s_Program_local_drop
#define __pyx_n_s_Program_local_load __pyx_mstate_global->__pyx_n_s_Program_local_load
#define __pyx_n_s_Program_local_push __pyx_mstate_global->__pyx_n_s_Program_local_push
#define __pyx_n_s_Program_lookup __pyx_mstate_global->__pyx_n_s_Program_lookup
#define __pyx_n_s_Program_lookup_literal __pyx_mstate_global->__pyx_n_s_Program_lookup_literal
#define __pyx_n_s_Program_lt __pyx_mstate_global->__pyx_n_s_Program_lt
#define __pyx_n_s_Program_mod __pyx_mstate_global->__pyx_n_s_Program_mod
#define __pyx_n_s_Program_mul __pyx_mstate_global->__pyx_n_s_Program_mul
#define __pyx_n_s_Program_mul_add __pyx_mstate_global->__pyx_n_s_Program_mul_add
#define __pyx_n_s_Program_ne __pyx_mstate_global->__pyx_n_s_Program_ne
#define __pyx_n_s_Program_neg __pyx_mstate_global->__pyx_n_s_Program_neg
#define __pyx_n_s_Program_new_label __pyx_mstate_global->__pyx_n_s_Program_new_label
#define __pyx_n_s_Program_next __pyx_mstate_global->__pyx_n_s_Program_next
#define __pyx_n_s_Program_not __pyx_mstate_global->__pyx_n_s_Program_not
#define __pyx_n_s_Program_pos __pyx_mstate_global->__pyx_n_s_Program_pos
#define __pyx_n_s_Program_pow __pyx_mstate_global->__pyx_n_s_Program_pow
#define __pyx_n_s_Program_range __pyx_mstate_global->__pyx_n_s_Program_range
#define __pyx_n_s_Program_run __pyx_mstate_global->__pyx_n_s_Program_run
#define __pyx_n_s_Program_set_path __pyx_mstate_global->__pyx_n_s_Program_set_path
#define __pyx_n_s_Program_set_pragma __pyx_mstate_global->__pyx_n_s_Program_set_pragma
#define __pyx_n_s_Program_set_top __pyx_mstate_global->__pyx_n_s_Program_set_top
#define __pyx_n_s_Program_slice __pyx_mstate_global->__pyx_n_s_Program_slice
#define __pyx_n_s_Program_slice_literal __pyx_mstate_global->__pyx_n_s_Program_slice_literal
#define __pyx_n_s_Program_sub __pyx_mstate_global->__pyx_n_s_Program_sub
#define __pyx_n_s_Program_tag __pyx_mstate_global->__pyx_n_s_Program_tag
#define __pyx_n_s_Program_truediv __pyx_mstate_global->__pyx_n_s_Program_truediv
#define __pyx_n_s_Program_use_simplifier __pyx_mstate_global->__pyx_n_s_Program_use_simplifier
#define __pyx_n_s_Program_xor __pyx_mstate_global->__pyx_n_s_Program_xor
#define __pyx_n_u_Range __pyx_mstate_global->__pyx_n_u_Range
#define __pyx_n_s_RecursionError __pyx_mstate_global->__pyx_n_s_RecursionError
#define __pyx_kp_u_Recursion_depth_exceeded_for_fun __pyx_mstate_global->__pyx_kp_u_Recursion_depth_exceeded_for_fun
#define __pyx_n_s_STATIC_FUNCTIONS __pyx_mstate_global->__pyx_n_s_STATIC_FUNCTIONS
#define __pyx_n_s_SharedCache __pyx_mstate_global->__pyx_n_s_SharedCache
#define __pyx_n_u_Slice __pyx_mstate_global->__pyx_n_u_Slice
#define __pyx_n_u_SliceLiteral __pyx_mstate_global->__pyx_n_u_SliceLiteral
#define __pyx_n_s_StackError __pyx_mstate_global->__pyx_n_s_StackError
#define __pyx_n_s_StackOverflow __pyx_mstate_global->__pyx_n_s_StackOverflow
#define __pyx_n_s_StackUnderflow __pyx_mstate_global->__pyx_n_s_StackUnderflow
#define __pyx_kp_u_Stack_cannot_grow_beyond __pyx_mstate_global->__pyx_kp_u_Stack_cannot_grow_beyond
#define __pyx_kp_u_Stack_empty __pyx_mstate_global->__pyx_kp_u_Stack_empty
#define __pyx_kp_u_Stack_overflow_in_program __pyx_mstate_global->__pyx_kp_u_Stack_overflow_in_program
#define __pyx_n_u_Sub __pyx_mstate_global->__pyx_n_u_Sub
#define __pyx_n_u_Tag __pyx_mstate_global->__pyx_n_u_Tag
#define __pyx_n_u_TrueDiv __pyx_mstate_global->__pyx_n_u_TrueDiv
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_u_Unable_to_import __pyx_mstate_global->__pyx_kp_u_Unable_to_import
#define __pyx_kp_u_Unable_to_import_from __pyx_mstate_global->__pyx_kp_u_Unable_to_import_from
#define __pyx_kp_u_Unrecognised_instruction __pyx_mstate_global->__pyx_kp_u_Unrecognised_instruction
#define __pyx_kp_u_VM_exception_processing __pyx_mstate_global->__pyx_kp_u_VM_exception_processing
#define __pyx_kp_u_VM_execution_statistics __pyx_mstate_global->__pyx_kp_u_VM_execution_statistics
#define __pyx_n_s_VectorStack __pyx_mstate_global->__pyx_n_s_VectorStack
#define __pyx_n_s_VectorStack___reduce_cython __pyx_mstate_global->__pyx_n_s_VectorStack___reduce_cython
#define __pyx_n_s_VectorStack___setstate_cython __pyx_mstate_global->__pyx_n_s_VectorStack___setstate_cython
#define __pyx_n_s_VectorStack_copy __pyx_mstate_global->__pyx_n_s_VectorStack_copy
#define __pyx_n_s_VectorStack_drop __pyx_mstate_global->__pyx_n_s_VectorStack_drop
#define __pyx_n_s_VectorStack_peek __pyx_mstate_global->__pyx_n_s_VectorStack_peek
#define __pyx_n_s_VectorStack_peek_at __pyx_mstate_global->__pyx_n_s_VectorStack_peek_at
#define __pyx_n_s_VectorStack_poke __pyx_mstate_global->__pyx_n_s_VectorStack_poke
#define __pyx_n_s_VectorStack_poke_at __pyx_mstate_global->__pyx_n_s_VectorStack_poke_at
#define __pyx_n_s_VectorStack_pop __pyx_mstate_global->__pyx_n_s_VectorStack_pop
#define __pyx_n_s_VectorStack_pop_composed __pyx_mstate_global->__pyx_n_s_VectorStack_pop_composed
#define __pyx_n_s_VectorStack_pop_dict __pyx_mstate_global->__pyx_n_s_VectorStack_pop_dict
#define __pyx_n_s_VectorStack_pop_list __pyx_mstate_global->__pyx_n_s_VectorStack_pop_list
#define __pyx_n_s_VectorStack_pop_tuple __pyx_mstate_global->__pyx_n_s_VectorStack_pop_tuple
#define __pyx_n_s_VectorStack_push __pyx_mstate_global->__pyx_n_s_VectorStack_push
#define __pyx_n_u_Xor __pyx_mstate_global->__pyx_n_u_Xor
#define __pyx_kp_u__12 __pyx_mstate_global->__pyx_kp_u__12
#define __pyx_n_s__163 __pyx_mstate_global->__pyx_n_s__163
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_n_s__21 __pyx_mstate_global->__pyx_n_s__21
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_attributes __pyx_mstate_global->__pyx_n_s_attributes
#define __pyx_n_s_begin_for __pyx_mstate_global->__pyx_n_s_begin_for
#define __pyx_n_s_branch_false __pyx_mstate_global->__pyx_n_s_branch_false
#define __pyx_n_s_branch_true __pyx_mstate_global->__pyx_n_s_branch_true
#define __pyx_n_s_builtins __pyx_mstate_global->__pyx_n_s_builtins
#define __pyx_n_s_cache __pyx_mstate_global->__pyx_n_s_cache
#define __pyx_n_s_call __pyx_mstate_global->__pyx_n_s_call
#define __pyx_n_s_call_fast __pyx_mstate_global->__pyx_n_s_call_fast
#define __pyx_n_s_ceil __pyx_mstate_global->__pyx_n_s_ceil
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_code __pyx_mstate_global->__pyx_n_s_code
#define __pyx_n_s_compose __pyx_mstate_global->__pyx_n_s_compose
#define __pyx_n_s_context __pyx_mstate_global->__pyx_n_s_context
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_count __pyx_mstate_global->__pyx_n_s_count
#define __pyx_kp_u_d __pyx_mstate_global->__pyx_kp_u_d
#define __pyx_kp_u_depth_stack __pyx_mstate_global->__pyx_kp_u_depth_stack
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_drop __pyx_mstate_global->__pyx_n_s_drop
#define __pyx_n_s_dup __pyx_mstate_global->__pyx_n_s_dup
#define __pyx_n_s_duration __pyx_mstate_global->__pyx_n_s_duration
#define __pyx_n_s_dynamic_builtins __pyx_mstate_global->__pyx_n_s_dynamic_builtins
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_s_end_for __pyx_mstate_global->__pyx_n_s_end_for
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_eq __pyx_mstate_global->__pyx_n_s_eq
#define __pyx_n_s_exception __pyx_mstate_global->__pyx_n_s_exception
#define __pyx_n_s_execute __pyx_mstate_global->__pyx_n_s_execute
#define __pyx_n_s_exit __pyx_mstate_global->__pyx_n_s_exit
#define __pyx_n_s_export __pyx_mstate_global->__pyx_n_s_export
#define __pyx_n_u_false __pyx_mstate_global->__pyx_n_u_false
#define __pyx_n_s_flitter_language_vm __pyx_mstate_global->__pyx_n_s_flitter_language_vm
#define __pyx_n_s_floor __pyx_mstate_global->__pyx_n_s_floor
#define __pyx_n_s_floordiv __pyx_mstate_global->__pyx_n_s_floordiv
#define __pyx_n_s_fract __pyx_mstate_global->__pyx_n_s_fract
#define __pyx_kp_u_from __pyx_mstate_global->__pyx_kp_u_from
#define __pyx_n_s_func __pyx_mstate_global->__pyx_n_s_func
#define __pyx_n_s_function __pyx_mstate_global->__pyx_n_s_function
#define __pyx_n_s_functions __pyx_mstate_global->__pyx_n_s_functions
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_ge __pyx_mstate_global->__pyx_n_s_ge
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get_with_root __pyx_mstate_global->__pyx_n_s_get_with_root
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_gt __pyx_mstate_global->__pyx_n_s_gt
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_import_2 __pyx_mstate_global->__pyx_n_s_import_2
#define __pyx_n_u_inf __pyx_mstate_global->__pyx_n_u_inf
#define __pyx_n_s_info __pyx_mstate_global->__pyx_n_s_info
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_item __pyx_mstate_global->__pyx_n_s_item
#define __pyx_kp_u_item_stack __pyx_mstate_global->__pyx_kp_u_item_stack
#define __pyx_kp_u_items __pyx_mstate_global->__pyx_kp_u_items
#define __pyx_n_s_items_2 __pyx_mstate_global->__pyx_n_s_items_2
#define __pyx_kp_u_items_below_top_of __pyx_mstate_global->__pyx_kp_u_items_below_top_of
#define __pyx_kp_u_items_from __pyx_mstate_global->__pyx_kp_u_items_from
#define __pyx_n_s_ivalue __pyx_mstate_global->__pyx_n_s_ivalue
#define __pyx_n_s_jump __pyx_mstate_global->__pyx_n_s_jump
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_label __pyx_mstate_global->__pyx_n_s_label
#define __pyx_n_s_le __pyx_mstate_global->__pyx_n_s_le
#define __pyx_n_s_link __pyx_mstate_global->__pyx_n_s_link
#define __pyx_n_s_literal __pyx_mstate_global->__pyx_n_s_literal
#define __pyx_n_s_lnames __pyx_mstate_global->__pyx_n_s_lnames
#define __pyx_n_s_local_drop __pyx_mstate_global->__pyx_n_s_local_drop
#define __pyx_n_s_local_load __pyx_mstate_global->__pyx_n_s_local_load
#define __pyx_n_s_local_push __pyx_mstate_global->__pyx_n_s_local_push
#define __pyx_n_s_log_vm_stats __pyx_mstate_global->__pyx_n_s_log_vm_stats
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_loguru __pyx_mstate_global->__pyx_n_s_loguru
#define __pyx_n_s_lookup __pyx_mstate_global->__pyx_n_s_lookup
#define __pyx_n_s_lookup_literal __pyx_mstate_global->__pyx_n_s_lookup_literal
#define __pyx_n_s_lt __pyx_mstate_global->__pyx_n_s_lt
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_max_size __pyx_mstate_global->__pyx_n_s_max_size
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_mod __pyx_mstate_global->__pyx_n_s_mod
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_mul __pyx_mstate_global->__pyx_n_s_mul
#define __pyx_n_s_mul_add __pyx_mstate_global->__pyx_n_s_mul_add
#define __pyx_n_s_n __pyx_mstate_global->__pyx_n_s_n
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_name_patch __pyx_mstate_global->__pyx_n_s_name_patch
#define __pyx_n_s_names __pyx_mstate_global->__pyx_n_s_names
#define __pyx_n_u_nan __pyx_mstate_global->__pyx_n_u_nan
#define __pyx_kp_u_native_funcs __pyx_mstate_global->__pyx_kp_u_native_funcs
#define __pyx_n_s_ncaptures __pyx_mstate_global->__pyx_n_s_ncaptures
#define __pyx_n_s_ne __pyx_mstate_global->__pyx_n_s_ne
#define __pyx_n_s_neg __pyx_mstate_global->__pyx_n_s_neg
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_new_label __pyx_mstate_global->__pyx_n_s_new_label
#define __pyx_n_s_next __pyx_mstate_global->__pyx_n_s_next
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_noise __pyx_mstate_global->__pyx_n_s_noise
#define __pyx_n_s_not __pyx_mstate_global->__pyx_n_s_not
#define __pyx_n_u_null __pyx_mstate_global->__pyx_n_u_null
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_objptr __pyx_mstate_global->__pyx_n_s_objptr
#define __pyx_n_s_offset __pyx_mstate_global->__pyx_n_s_offset
#define __pyx_n_s_overhead __pyx_mstate_global->__pyx_n_s_overhead
#define __pyx_n_s_parameters __pyx_mstate_global->__pyx_n_s_parameters
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_n_s_peek __pyx_mstate_global->__pyx_n_s_peek
#define __pyx_n_s_peek_at __pyx_mstate_global->__pyx_n_s_peek_at
#define __pyx_n_s_per_execution __pyx_mstate_global->__pyx_n_s_per_execution
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_poke __pyx_mstate_global->__pyx_n_s_poke
#define __pyx_n_s_poke_at __pyx_mstate_global->__pyx_n_s_poke_at
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_pop_composed __pyx_mstate_global->__pyx_n_s_pop_composed
#define __pyx_n_s_pop_dict __pyx_mstate_global->__pyx_n_s_pop_dict
#define __pyx_n_s_pop_list __pyx_mstate_global->__pyx_n_s_pop_list
#define __pyx_n_s_pop_tuple __pyx_mstate_global->__pyx_n_s_pop_tuple
#define __pyx_n_s_pos __pyx_mstate_global->__pyx_n_s_pos
#define __pyx_n_s_pow __pyx_mstate_global->__pyx_n_s_pow
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_push __pyx_mstate_global->__pyx_n_s_push
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_capi __pyx_mstate_global->__pyx_n_s_pyx_capi
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_Function __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Function
#define __pyx_n_s_pyx_unpickle_Instruction __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Instruction
#define __pyx_n_s_pyx_unpickle_InstructionFunc __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionFunc
#define __pyx_n_s_pyx_unpickle_InstructionInt __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionInt
#define __pyx_n_s_pyx_unpickle_InstructionIntTup __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionIntTup
#define __pyx_n_s_pyx_unpickle_InstructionJump __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionJump
#define __pyx_n_s_pyx_unpickle_InstructionLabel __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionLabel
#define __pyx_n_s_pyx_unpickle_InstructionNode __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionNode
#define __pyx_n_s_pyx_unpickle_InstructionObject __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionObject
#define __pyx_n_s_pyx_unpickle_InstructionStr __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionStr
#define __pyx_n_s_pyx_unpickle_InstructionTuple __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionTuple
#define __pyx_n_s_pyx_unpickle_InstructionVector __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InstructionVector
#define __pyx_n_s_pyx_unpickle_LoopSource __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LoopSource
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_read_flitter_program __pyx_mstate_global->__pyx_n_s_read_flitter_program
#define __pyx_n_s_record_stats __pyx_mstate_global->__pyx_n_s_record_stats
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_n_s_reverse __pyx_mstate_global->__pyx_n_s_reverse
#define __pyx_n_u_root __pyx_mstate_global->__pyx_n_u_root
#define __pyx_n_s_run __pyx_mstate_global->__pyx_n_s_run
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_set_path __pyx_mstate_global->__pyx_n_s_set_path
#define __pyx_n_s_set_pragma __pyx_mstate_global->__pyx_n_s_set_pragma
#define __pyx_n_s_set_top __pyx_mstate_global->__pyx_n_s_set_top
#define __pyx_n_s_setdefault __pyx_mstate_global->__pyx_n_s_setdefault
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_simplify __pyx_mstate_global->__pyx_n_s_simplify
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_s_slice __pyx_mstate_global->__pyx_n_s_slice
#define __pyx_n_s_slice_literal __pyx_mstate_global->__pyx_n_s_slice_literal
#define __pyx_n_s_sort __pyx_mstate_global->__pyx_n_s_sort
#define __pyx_kp_s_src_flitter_language_vm_pyx __pyx_mstate_global->__pyx_kp_s_src_flitter_language_vm_pyx
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_static_builtins __pyx_mstate_global->__pyx_n_s_static_builtins
#define __pyx_n_s_stats __pyx_mstate_global->__pyx_n_s_stats
#define __pyx_n_s_str___locals_genexpr __pyx_mstate_global->__pyx_n_s_str___locals_genexpr
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_sub __pyx_mstate_global->__pyx_n_s_sub
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_tag __pyx_mstate_global->__pyx_n_s_tag
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_top __pyx_mstate_global->__pyx_n_s_top
#define __pyx_n_s_total __pyx_mstate_global->__pyx_n_s_total
#define __pyx_n_u_true __pyx_mstate_global->__pyx_n_u_true
#define __pyx_n_s_truediv __pyx_mstate_global->__pyx_n_s_truediv
#define __pyx_n_s_tvalue __pyx_mstate_global->__pyx_n_s_tvalue
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_use_simplifier __pyx_mstate_global->__pyx_n_s_use_simplifier
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_vector __pyx_mstate_global->__pyx_n_s_vector
#define __pyx_n_s_xor __pyx_mstate_global->__pyx_n_s_xor
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_10904070 __pyx_mstate_global->__pyx_int_10904070
#define __pyx_int_19407059 __pyx_mstate_global->__pyx_int_19407059
#define __pyx_int_29217351 __pyx_mstate_global->__pyx_int_29217351
#define __pyx_int_45504841 __pyx_mstate_global->__pyx_int_45504841
#define __pyx_int_59839126 __pyx_mstate_global->__pyx_int_59839126
#define __pyx_int_61928532 __pyx_mstate_global->__pyx_int_61928532
#define __pyx_int_63712393 __pyx_mstate_global->__pyx_int_63712393
#define __pyx_int_71438266 __pyx_mstate_global->__pyx_int_71438266
#define __pyx_int_72616288 __pyx_mstate_global->__pyx_int_72616288
#define __pyx_int_73689043 __pyx_mstate_global->__pyx_int_73689043
#define __pyx_int_90787080 __pyx_mstate_global->__pyx_int_90787080
#define __pyx_int_107792009 __pyx_mstate_global->__pyx_int_107792009
#define __pyx_int_109158536 __pyx_mstate_global->__pyx_int_109158536
#define __pyx_int_138499235 __pyx_mstate_global->__pyx_int_138499235
#define __pyx_int_163945588 __pyx_mstate_global->__pyx_int_163945588
#define __pyx_int_176342959 __pyx_mstate_global->__pyx_int_176342959
#define __pyx_int_184506718 __pyx_mstate_global->__pyx_int_184506718
#define __pyx_int_184517443 __pyx_mstate_global->__pyx_int_184517443
#define __pyx_int_186494852 __pyx_mstate_global->__pyx_int_186494852
#define __pyx_int_200977137 __pyx_mstate_global->__pyx_int_200977137
#define __pyx_int_202585721 __pyx_mstate_global->__pyx_int_202585721
#define __pyx_int_210616022 __pyx_mstate_global->__pyx_int_210616022
#define __pyx_int_229249173 __pyx_mstate_global->__pyx_int_229249173
#define __pyx_int_231591000 __pyx_mstate_global->__pyx_int_231591000
#define __pyx_int_236313263 __pyx_mstate_global->__pyx_int_236313263
#define __pyx_int_242200674 __pyx_mstate_global->__pyx_int_242200674
#define __pyx_int_259995473 __pyx_mstate_global->__pyx_int_259995473
#define __pyx_k__4 __pyx_mstate_global->__pyx_k__4
#define __pyx_k__5 __pyx_mstate_global->__pyx_k__5
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__63 __pyx_mstate_global->__pyx_tuple__63
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__75 __pyx_mstate_global->__pyx_tuple__75
#define __pyx_tuple__77 __pyx_mstate_global->__pyx_tuple__77
#define __pyx_tuple__79 __pyx_mstate_global->__pyx_tuple__79
#define __pyx_tuple__81 __pyx_mstate_global->__pyx_tuple__81
#define __pyx_tuple__82 __pyx_mstate_global->__pyx_tuple__82
#define __pyx_tuple__84 __pyx_mstate_global->__pyx_tuple__84
#define __pyx_tuple__86 __pyx_mstate_global->__pyx_tuple__86
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__96 __pyx_mstate_global->__pyx_tuple__96
#define __pyx_tuple__98 __pyx_mstate_global->__pyx_tuple__98
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__132 __pyx_mstate_global->__pyx_tuple__132
#define __pyx_tuple__134 __pyx_mstate_global->__pyx_tuple__134
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__149 __pyx_mstate_global->__pyx_tuple__149
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__89 __pyx_mstate_global->__pyx_codeobj__89
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__106 __pyx_mstate_global->__pyx_codeobj__106
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__108 __pyx_mstate_global->__pyx_codeobj__108
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__111 __pyx_mstate_global->__pyx_codeobj__111
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#define __pyx_codeobj__118 __pyx_mstate_global->__pyx_codeobj__118
#define __pyx_codeobj__119 __pyx_mstate_global->__pyx_codeobj__119
#define __pyx_codeobj__120 __pyx_mstate_global->__pyx_codeobj__120
#define __pyx_codeobj__121 __pyx_mstate_global->__pyx_codeobj__121
#define __pyx_codeobj__122 __pyx_mstate_global->__pyx_codeobj__122
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__124 __pyx_mstate_global->__pyx_codeobj__124
#define __pyx_codeobj__125 __pyx_mstate_global->__pyx_codeobj__125
#define __pyx_codeobj__126 __pyx_mstate_global->__pyx_codeobj__126
#define __pyx_codeobj__127 __pyx_mstate_global->__pyx_codeobj__127
#define __pyx_codeobj__128 __pyx_mstate_global->__pyx_codeobj__128
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__133 __pyx_mstate_global->__pyx_codeobj__133
#define __pyx_codeobj__135 __pyx_mstate_global->__pyx_codeobj__135
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__137 __pyx_mstate_global->__pyx_codeobj__137
#define __pyx_codeobj__138 __pyx_mstate_global->__pyx_codeobj__138
#define __pyx_codeobj__139 __pyx_mstate_global->__pyx_codeobj__139
#define __pyx_codeobj__140 __pyx_mstate_global->__pyx_codeobj__140
#define __pyx_codeobj__141 __pyx_mstate_global->__pyx_codeobj__141
#define __pyx_codeobj__142 __pyx_mstate_global->__pyx_codeobj__142
#define __pyx_codeobj__144 __pyx_mstate_global->__pyx_codeobj__144
#define __pyx_codeobj__146 __pyx_mstate_global->__pyx_codeobj__146
#define __pyx_codeobj__147 __pyx_mstate_global->__pyx_codeobj__147
#define __pyx_codeobj__148 __pyx_mstate_global->__pyx_codeobj__148
#define __pyx_codeobj__150 __pyx_mstate_global->__pyx_codeobj__150
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__152 __pyx_mstate_global->__pyx_codeobj__152
#define __pyx_codeobj__153 __pyx_mstate_global->__pyx_codeobj__153
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__155 __pyx_mstate_global->__pyx_codeobj__155
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__157 __pyx_mstate_global->__pyx_codeobj__157
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__159 __pyx_mstate_global->__pyx_codeobj__159
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__161 __pyx_mstate_global->__pyx_codeobj__161
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
/* #### Code section: module_code ### */

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_UPDATE(uint64_t __pyx_v__hash, uint64_t __pyx_v_y) {
  uint64_t __pyx_r;

  /* "model.pxd":13
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y             # <<<<<<<<<<<<<<
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 */
  __pyx_v__hash = (__pyx_v__hash ^ __pyx_v_y);

  /* "model.pxd":14
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 */
  __pyx_v__hash = (__pyx_v__hash + ((uint64_t)0x9e3779b97f4a7c15));

  /* "model.pxd":15
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 30));

  /* "model.pxd":16
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0xbf58476d1ce4e5b9));

  /* "model.pxd":17
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 27));

  /* "model.pxd":18
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 31
 *     return _hash
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0x94d049bb133111eb));

  /* "model.pxd":19
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31             # <<<<<<<<<<<<<<
 *     return _hash
 * 
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 31));

  /* "model.pxd":20
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 *     return _hash             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v__hash;
  goto __pyx_L0;

  /* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_STRING(PyObject *__pyx_v_value) {
  void *__pyx_v_data;
  uint64_t __pyx_v_i;
  uint64_t __pyx_v_n;
  uint64_t __pyx_v_kind;
  Py_UCS4 __pyx_v_c;
  uint64_t __pyx_v_y;
  uint64_t __pyx_r;
  uint64_t __pyx_t_1;
  uint64_t __pyx_t_2;
  uint64_t __pyx_t_3;

  /* "model.pxd":25
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)             # <<<<<<<<<<<<<<
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 */
  __pyx_v_data = PyUnicode_DATA(__pyx_v_value);

  /* "model.pxd":26
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)             # <<<<<<<<<<<<<<
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 */
  __pyx_v_n = PyUnicode_GET_LENGTH(__pyx_v_value);
  __pyx_v_kind = PyUnicode_KIND(__pyx_v_value);

  /* "model.pxd":28
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 */
  __pyx_v_y = ((uint64_t)0xcbf29ce484222325);

  /* "model.pxd":29
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "model.pxd":30
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)             # <<<<<<<<<<<<<<
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y
 */
    __pyx_v_c = PyUnicode_READ(__pyx_v_kind, __pyx_v_data, __pyx_v_i);

    /* "model.pxd":31
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)             # <<<<<<<<<<<<<<
 *     return y
 * 
 */
    __pyx_v_y = ((__pyx_v_y ^ ((uint64_t)__pyx_v_c)) * ((uint64_t)0x100000001b3));
  }

  /* "model.pxd":32
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_y;
  goto __pyx_L0;

  /* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/language/vm.pyx":117
 * 
 * 
 * cdef initialize_stats():             # <<<<<<<<<<<<<<
 *     global StatsCount, StatsDuration
 *     cdef int64_t n = OpCode.MAX + 1
 */

static PyObject *__pyx_f_7flitter_8language_2vm_initialize_stats(void) {
  int64_t __pyx_v_n;
  int64_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  __Pyx_RefNannySetupContext("initialize_stats", 1);

  /* "flitter/language/vm.pyx":119
 * cdef initialize_stats():
 *     global StatsCount, StatsDuration
 *     cdef int64_t n = OpCode.MAX + 1             # <<<<<<<<<<<<<<
 *     StatsCount = <int64_t*>PyMem_Malloc(n * sizeof(int64_t))
 *     StatsDuration = <double*>PyMem_Malloc(n * sizeof(double))
 */
  __pyx_v_n = (__pyx_e_7flitter_8language_2vm_MAX + 1);

  /* "flitter/language/vm.pyx":120
 *     global StatsCount, StatsDuration
 *     cdef int64_t n = OpCode.MAX + 1
 *     StatsCount = <int64_t*>PyMem_Malloc(n * sizeof(int64_t))             # <<<<<<<<<<<<<<
 *     StatsDuration = <double*>PyMem_Malloc(n * sizeof(double))
 *     cdef int64_t i
 */
  __pyx_v_7flitter_8language_2vm_StatsCount = ((int64_t *)PyMem_Malloc((__pyx_v_n * (sizeof(int64_t)))));

  /* "flitter/language/vm.pyx":121
 *     cdef int64_t n = OpCode.MAX + 1
 *     StatsCount = <int64_t*>PyMem_Malloc(n * sizeof(int64_t))
 *     StatsDuration = <double*>PyMem_Malloc(n * sizeof(double))             # <<<<<<<<<<<<<<
 *     cdef int64_t i
 *     for i in range(n):
 */
  __pyx_v_7flitter_8language_2vm_StatsDuration = ((double *)PyMem_Malloc((__pyx_v_n * (sizeof(double)))));

  /* "flitter/language/vm.pyx":123
 *     StatsDuration = <double*>PyMem_Malloc(n * sizeof(double))
 *     cdef int64_t i
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         StatsCount[i] = 0
 *         StatsDuration[i] = 0
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/language/vm.pyx":124
 *     cdef int64_t i
 *     for i in range(n):
 *         StatsCount[i] = 0             # <<<<<<<<<<<<<<
 *         StatsDuration[i] = 0
 * 
 */
    (__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_v_i]) = 0;

    /* "flitter/language/vm.pyx":125
 *     for i in range(n):
 *         StatsCount[i] = 0
 *         StatsDuration[i] = 0             # <<<<<<<<<<<<<<
 * 
 * initialize_stats()
 */
    (__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_v_i]) = 0.0;
  }

  /* "flitter/language/vm.pyx":117
 * 
 * 
 * cdef initialize_stats():             # <<<<<<<<<<<<<<
 *     global StatsCount, StatsDuration
 *     cdef int64_t n = OpCode.MAX + 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":131
 * 
 * cdef class Instruction:
 *     def __init__(self, OpCode code):             # <<<<<<<<<<<<<<
 *         self.code = code
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_11Instruction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_11Instruction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 131, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Instruction.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11Instruction___init__(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v_self), __pyx_v_code);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_11Instruction___init__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code) {
  int __pyx_r;

  /* "flitter/language/vm.pyx":132
 * cdef class Instruction:
 *     def __init__(self, OpCode code):
 *         self.code = code             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->code = __pyx_v_code;

  /* "flitter/language/vm.pyx":131
 * 
 * cdef class Instruction:
 *     def __init__(self, OpCode code):             # <<<<<<<<<<<<<<
 *         self.code = code
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "flitter/language/vm.pyx":134
 *         self.code = code
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return OpCodeNames[self.code]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11Instruction_2__str__(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_2__str__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":135
 * 
 *     def __str__(self):
 *         return OpCodeNames[self.code]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 135, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":134
 *         self.code = code
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return OpCodeNames[self.code]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.vm.Instruction.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":91
 * 
 * cdef class Instruction:
 *     cdef readonly OpCode code             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_4code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_4code_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11Instruction_4code___get__(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_4code___get__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Instruction.code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11Instruction_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11Instruction_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11Instruction_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.code,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_90787080);
    __Pyx_GIVEREF(__pyx_int_90787080);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_90787080)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, None), state
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Instruction); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_90787080);
    __Pyx_GIVEREF(__pyx_int_90787080);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_90787080)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Instruction.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11Instruction_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11Instruction_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11Instruction_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Instruction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11Instruction_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11Instruction_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_Instruction__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Instruction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":139
 * 
 * cdef class InstructionVector(Instruction):
 *     def __init__(self, OpCode code, Vector value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_17InstructionVector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_17InstructionVector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 139, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L3_error)
    __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Vector *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 139, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_7flitter_5model_Vector, 1, "value", 0))) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_17InstructionVector___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_self), __pyx_v_code, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_17InstructionVector___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":140
 * cdef class InstructionVector(Instruction):
 *     def __init__(self, OpCode code, Vector value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector))) __PYX_ERR(0, 140, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 140, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":141
 *     def __init__(self, OpCode code, Vector value):
 *         super().__init__(code)
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_value);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->value);
  __Pyx_DECREF((PyObject *)__pyx_v_self->value);
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":139
 * 
 * cdef class InstructionVector(Instruction):
 *     def __init__(self, OpCode code, Vector value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":143
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_17InstructionVector_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":144
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 144, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":143
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":95
 * 
 * cdef class InstructionVector(Instruction):
 *     cdef readonly Vector value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_17InstructionVector_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __pyx_r = ((PyObject *)__pyx_v_self->value);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_17InstructionVector_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->value))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, None), state
 */
  /*else*/ {
    __pyx_t_3 = (((PyObject *)__pyx_v_self->value) != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionVector); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, None), state
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_InstructionVector); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_17InstructionVector_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_17InstructionVector_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionVector__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionVector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":148
 * 
 * cdef class InstructionNode(Instruction):
 *     def __init__(self, OpCode code, Node value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_15InstructionNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_15InstructionNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 148, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 148, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L3_error)
    __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Node *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 148, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_7flitter_5model_Node, 1, "value", 0))) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionNode___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_self), __pyx_v_code, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_15InstructionNode___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, struct __pyx_obj_7flitter_5model_Node *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":149
 * cdef class InstructionNode(Instruction):
 *     def __init__(self, OpCode code, Node value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode))) __PYX_ERR(0, 149, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 149, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":150
 *     def __init__(self, OpCode code, Node value):
 *         super().__init__(code)
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_value);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->value);
  __Pyx_DECREF((PyObject *)__pyx_v_self->value);
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":148
 * 
 * cdef class InstructionNode(Instruction):
 *     def __init__(self, OpCode code, Node value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":152
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionNode_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":153
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 153, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":152
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":99
 * 
 * cdef class InstructionNode(Instruction):
 *     cdef readonly Node value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionNode_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __pyx_r = ((PyObject *)__pyx_v_self->value);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionNode_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->value))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, None), state
 */
  /*else*/ {
    __pyx_t_3 = (((PyObject *)__pyx_v_self->value) != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionNode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, None), state
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_InstructionNode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionNode_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionNode_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":157
 * 
 * cdef class InstructionStr(Instruction):
 *     def __init__(self, OpCode code, str value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_14InstructionStr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_14InstructionStr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 157, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 157, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L3_error)
    __pyx_v_value = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 157, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 1, "value", 1))) __PYX_ERR(0, 157, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionStr___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_self), __pyx_v_code, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_14InstructionStr___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":158
 * cdef class InstructionStr(Instruction):
 *     def __init__(self, OpCode code, str value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr))) __PYX_ERR(0, 158, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 158, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":159
 *     def __init__(self, OpCode code, str value):
 *         super().__init__(code)
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":157
 * 
 * cdef class InstructionStr(Instruction):
 *     def __init__(self, OpCode code, str value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":161
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionStr_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":162
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 162, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->value), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":161
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":103
 * 
 * cdef class InstructionStr(Instruction):
 *     cdef readonly str value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionStr_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->value);
  __pyx_r = __pyx_v_self->value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionStr_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->value);
  __Pyx_GIVEREF(__pyx_v_self->value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->value != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionStr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, None), state
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_InstructionStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionStr_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionStr_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionStr__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionStr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":166
 * 
 * cdef class InstructionTuple(Instruction):
 *     def __init__(self, OpCode code, tuple value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_16InstructionTuple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_16InstructionTuple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 166, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 166, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
    __pyx_v_value = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 166, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyTuple_Type), 1, "value", 1))) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionTuple___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_self), __pyx_v_code, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_16InstructionTuple___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":167
 * cdef class InstructionTuple(Instruction):
 *     def __init__(self, OpCode code, tuple value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple))) __PYX_ERR(0, 167, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 167, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":168
 *     def __init__(self, OpCode code, tuple value):
 *         super().__init__(code)
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":166
 * 
 * cdef class InstructionTuple(Instruction):
 *     def __init__(self, OpCode code, tuple value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":170
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionTuple_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":171
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 171, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->value), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":170
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":107
 * 
 * cdef class InstructionTuple(Instruction):
 *     cdef readonly tuple value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionTuple_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->value);
  __pyx_r = __pyx_v_self->value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionTuple_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->value);
  __Pyx_GIVEREF(__pyx_v_self->value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->value != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionTuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, None), state
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_InstructionTuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_184517443)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionTuple_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionTuple_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionTuple__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionTuple.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":175
 * 
 * cdef class InstructionInt(Instruction):
 *     def __init__(self, OpCode code, int64_t value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_14InstructionInt_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_14InstructionInt_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  int64_t __pyx_v_value;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 175, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 175, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
    __pyx_v_value = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_value == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 175, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionInt___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_self), __pyx_v_code, __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_14InstructionInt___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":176
 * cdef class InstructionInt(Instruction):
 *     def __init__(self, OpCode code, int64_t value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt))) __PYX_ERR(0, 176, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 176, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":177
 *     def __init__(self, OpCode code, int64_t value):
 *         super().__init__(code)
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":175
 * 
 * cdef class InstructionInt(Instruction):
 *     def __init__(self, OpCode code, int64_t value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.value = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":179
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionInt_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":180
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 180, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":179
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":111
 * 
 * cdef class InstructionInt(Instruction):
 *     cdef readonly int64_t value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionInt_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionInt_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_InstructionInt); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_184517443)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, None), state
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionInt); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184517443);
    __Pyx_GIVEREF(__pyx_int_184517443);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_184517443)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14InstructionInt_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14InstructionInt_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionInt__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionInt.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":184
 * 
 * cdef class InstructionIntTuple(Instruction):
 *     def __init__(self, OpCode code, int64_t ivalue, tuple tvalue):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.ivalue = ivalue
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  int64_t __pyx_v_ivalue;
  PyObject *__pyx_v_tvalue = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_ivalue,&__pyx_n_s_tvalue,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ivalue)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 184, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tvalue)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
    __pyx_v_ivalue = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_ivalue == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
    __pyx_v_tvalue = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tvalue), (&PyTuple_Type), 1, "tvalue", 1))) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self), __pyx_v_code, __pyx_v_ivalue, __pyx_v_tvalue);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_ivalue, PyObject *__pyx_v_tvalue) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":185
 * cdef class InstructionIntTuple(Instruction):
 *     def __init__(self, OpCode code, int64_t ivalue, tuple tvalue):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.ivalue = ivalue
 *         self.tvalue = tvalue
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple))) __PYX_ERR(0, 185, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 185, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":186
 *     def __init__(self, OpCode code, int64_t ivalue, tuple tvalue):
 *         super().__init__(code)
 *         self.ivalue = ivalue             # <<<<<<<<<<<<<<
 *         self.tvalue = tvalue
 * 
 */
  __pyx_v_self->ivalue = __pyx_v_ivalue;

  /* "flitter/language/vm.pyx":187
 *         super().__init__(code)
 *         self.ivalue = ivalue
 *         self.tvalue = tvalue             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_INCREF(__pyx_v_tvalue);
  __Pyx_GIVEREF(__pyx_v_tvalue);
  __Pyx_GOTREF(__pyx_v_self->tvalue);
  __Pyx_DECREF(__pyx_v_self->tvalue);
  __pyx_v_self->tvalue = __pyx_v_tvalue;

  /* "flitter/language/vm.pyx":184
 * 
 * cdef class InstructionIntTuple(Instruction):
 *     def __init__(self, OpCode code, int64_t ivalue, tuple tvalue):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.ivalue = ivalue
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":189
 *         self.tvalue = tvalue
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.ivalue!r} {self.tvalue!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":190
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.ivalue!r} {self.tvalue!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 190, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ivalue); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->tvalue), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":189
 *         self.tvalue = tvalue
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.ivalue!r} {self.tvalue!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":115
 * 
 * cdef class InstructionIntTuple(Instruction):
 *     cdef readonly int64_t ivalue             # <<<<<<<<<<<<<<
 *     cdef readonly tuple tvalue
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6ivalue_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6ivalue_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6ivalue___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6ivalue___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ivalue); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.ivalue.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":116
 * cdef class InstructionIntTuple(Instruction):
 *     cdef readonly int64_t ivalue
 *     cdef readonly tuple tvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6tvalue_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6tvalue_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6tvalue___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6tvalue___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->tvalue);
  __pyx_r = __pyx_v_self->tvalue;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.ivalue, self.tvalue)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ivalue); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->tvalue);
  __Pyx_GIVEREF(__pyx_v_self->tvalue);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->tvalue)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.ivalue, self.tvalue)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.ivalue, self.tvalue)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.tvalue is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.ivalue, self.tvalue)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.tvalue is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->tvalue != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.tvalue is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.tvalue is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_InstructionIntTup); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_73689043);
    __Pyx_GIVEREF(__pyx_int_73689043);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_73689043)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.tvalue is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, None), state
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionIntTup); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_73689043);
    __Pyx_GIVEREF(__pyx_int_73689043);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_73689043)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_19InstructionIntTuple_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionIntTuple__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionIntTuple.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":194
 * 
 * cdef class InstructionLabel(Instruction):
 *     def __init__(self, int64_t label):             # <<<<<<<<<<<<<<
 *         super().__init__(OpCode.Label)
 *         self.label = label
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_16InstructionLabel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_16InstructionLabel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int64_t __pyx_v_label;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 194, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 194, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionLabel___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_16InstructionLabel___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self, int64_t __pyx_v_label) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":195
 * cdef class InstructionLabel(Instruction):
 *     def __init__(self, int64_t label):
 *         super().__init__(OpCode.Label)             # <<<<<<<<<<<<<<
 *         self.label = label
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel))) __PYX_ERR(0, 195, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 195, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":196
 *     def __init__(self, int64_t label):
 *         super().__init__(OpCode.Label)
 *         self.label = label             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->label = __pyx_v_label;

  /* "flitter/language/vm.pyx":194
 * 
 * cdef class InstructionLabel(Instruction):
 *     def __init__(self, int64_t label):             # <<<<<<<<<<<<<<
 *         super().__init__(OpCode.Label)
 *         self.label = label
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":198
 *         self.label = label
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'.L{self.label}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionLabel_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":199
 * 
 *     def __str__(self):
 *         return f'.L{self.label}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_L, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":198
 *         self.label = label
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'.L{self.label}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":120
 * 
 * cdef class InstructionLabel(Instruction):
 *     cdef readonly int64_t label             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5label_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionLabel_5label___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_5label___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.label.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionLabel_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.label)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.label)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.label)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.label)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_InstructionLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_163945588);
    __Pyx_GIVEREF(__pyx_int_163945588);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_163945588)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, None), state
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionLabel); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_163945588);
    __Pyx_GIVEREF(__pyx_int_163945588);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_163945588)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16InstructionLabel_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16InstructionLabel_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionLabel__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionLabel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":203
 * 
 * cdef class InstructionJump(Instruction):
 *     def __init__(self, OpCode code, int64_t label):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.label = label
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_15InstructionJump_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_15InstructionJump_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  int64_t __pyx_v_label;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 203, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 203, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 203, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self), __pyx_v_code, __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_15InstructionJump___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_label) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":204
 * cdef class InstructionJump(Instruction):
 *     def __init__(self, OpCode code, int64_t label):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.label = label
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump))) __PYX_ERR(0, 204, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 204, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":205
 *     def __init__(self, OpCode code, int64_t label):
 *         super().__init__(code)
 *         self.label = label             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->label = __pyx_v_label;

  /* "flitter/language/vm.pyx":203
 * 
 * cdef class InstructionJump(Instruction):
 *     def __init__(self, OpCode code, int64_t label):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.label = label
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":207
 *         self.label = label
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.offset:
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":208
 * 
 *     def __str__(self):
 *         if self.offset:             # <<<<<<<<<<<<<<
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
 *         return f'{OpCodeNames[self.code]} .L{self.label}'
 */
  __pyx_t_1 = (__pyx_v_self->offset != 0);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":209
 *     def __str__(self):
 *         if self.offset:
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} .L{self.label}'
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 209, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_L_2);
    __pyx_t_3 += 3;
    __Pyx_GIVEREF(__pyx_kp_u_L_2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u_L_2);
    __pyx_t_5 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__2);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__2);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u__2);
    __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->offset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_kp_u_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__3);
    __pyx_t_3 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__3);
    PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_kp_u__3);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 6, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":208
 * 
 *     def __str__(self):
 *         if self.offset:             # <<<<<<<<<<<<<<
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
 *         return f'{OpCodeNames[self.code]} .L{self.label}'
 */
  }

  /* "flitter/language/vm.pyx":210
 *         if self.offset:
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
 *         return f'{OpCodeNames[self.code]} .L{self.label}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_4 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 210, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_kp_u_L_2);
  __pyx_t_3 += 3;
  __Pyx_GIVEREF(__pyx_kp_u_L_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_L_2);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":207
 *         self.label = label
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.offset:
 *             return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":124
 * 
 * cdef class InstructionJump(Instruction):
 *     cdef readonly int64_t label             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t offset
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_5label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_5label_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump_5label___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_5label___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.label.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":125
 * cdef class InstructionJump(Instruction):
 *     cdef readonly int64_t label
 *     cdef readonly int64_t offset             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_6offset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_6offset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump_6offset___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_6offset___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.offset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.label, self.offset)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->label); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.label, self.offset)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.label, self.offset)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.label, self.offset)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_InstructionJump); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_236313263);
    __Pyx_GIVEREF(__pyx_int_236313263);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_236313263)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, None), state
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_InstructionJump); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_236313263);
    __Pyx_GIVEREF(__pyx_int_236313263);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_236313263)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionJump_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionJump_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionJump__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionJump.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":214
 * 
 * cdef class InstructionFunc(InstructionJump):
 *     def __init__(self, OpCode code, int64_t label, str name, tuple parameters, int64_t ncaptures):             # <<<<<<<<<<<<<<
 *         super().__init__(code, label)
 *         self.name = name
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_15InstructionFunc_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_15InstructionFunc_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  int64_t __pyx_v_label;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parameters = 0;
  int64_t __pyx_v_ncaptures;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_label,&__pyx_n_s_name,&__pyx_n_s_parameters,&__pyx_n_s_ncaptures,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(0, 214, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(0, 214, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parameters)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(0, 214, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ncaptures)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(0, 214, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 214, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
    __pyx_v_name = ((PyObject*)values[2]);
    __pyx_v_parameters = ((PyObject*)values[3]);
    __pyx_v_ncaptures = __Pyx_PyInt_As_int64_t(values[4]); if (unlikely((__pyx_v_ncaptures == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 214, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 214, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parameters), (&PyTuple_Type), 1, "parameters", 1))) __PYX_ERR(0, 214, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self), __pyx_v_code, __pyx_v_label, __pyx_v_name, __pyx_v_parameters, __pyx_v_ncaptures);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_15InstructionFunc___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int64_t __pyx_v_ncaptures) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":215
 * cdef class InstructionFunc(InstructionJump):
 *     def __init__(self, OpCode code, int64_t label, str name, tuple parameters, int64_t ncaptures):
 *         super().__init__(code, label)             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.parameters = parameters
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc))) __PYX_ERR(0, 215, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 215, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":216
 *     def __init__(self, OpCode code, int64_t label, str name, tuple parameters, int64_t ncaptures):
 *         super().__init__(code, label)
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.parameters = parameters
 *         self.ncaptures = ncaptures
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "flitter/language/vm.pyx":217
 *         super().__init__(code, label)
 *         self.name = name
 *         self.parameters = parameters             # <<<<<<<<<<<<<<
 *         self.ncaptures = ncaptures
 * 
 */
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_v_parameters);
  __Pyx_GOTREF(__pyx_v_self->parameters);
  __Pyx_DECREF(__pyx_v_self->parameters);
  __pyx_v_self->parameters = __pyx_v_parameters;

  /* "flitter/language/vm.pyx":218
 *         self.name = name
 *         self.parameters = parameters
 *         self.ncaptures = ncaptures             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->ncaptures = __pyx_v_ncaptures;

  /* "flitter/language/vm.pyx":214
 * 
 * cdef class InstructionFunc(InstructionJump):
 *     def __init__(self, OpCode code, int64_t label, str name, tuple parameters, int64_t ncaptures):             # <<<<<<<<<<<<<<
 *         super().__init__(code, label)
 *         self.name = name
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":220
 *         self.ncaptures = ncaptures
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} L{self.label!r} {self.name!r} {self.parameters!r} {self.ncaptures!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":221
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} L{self.label!r} {self.name!r} {self.parameters!r} {self.ncaptures!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 221, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_L_3);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_L_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_L_3);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->name), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u_);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->parameters), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u_);
  __pyx_t_5 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ncaptures); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 9, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":220
 *         self.ncaptures = ncaptures
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} L{self.label!r} {self.name!r} {self.parameters!r} {self.ncaptures!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":129
 * 
 * cdef class InstructionFunc(InstructionJump):
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 *     cdef readonly tuple parameters
 *     cdef readonly int64_t ncaptures
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_4name___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_4name___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":130
 * cdef class InstructionFunc(InstructionJump):
 *     cdef readonly str name
 *     cdef readonly tuple parameters             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t ncaptures
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_10parameters_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_10parameters_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_10parameters___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_10parameters___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __pyx_r = __pyx_v_self->parameters;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":131
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 *     cdef readonly int64_t ncaptures             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_9ncaptures_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_9ncaptures_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_9ncaptures___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_9ncaptures___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ncaptures); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.ncaptures.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.label, self.name, self.ncaptures, self.offset, self.parameters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.label); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->ncaptures); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.offset); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __Pyx_GIVEREF(__pyx_v_self->parameters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->parameters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.label, self.name, self.ncaptures, self.offset, self.parameters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.label, self.name, self.ncaptures, self.offset, self.parameters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None or self.parameters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.label, self.name, self.ncaptures, self.offset, self.parameters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None or self.parameters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->name != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->parameters != ((PyObject*)Py_None));
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None or self.parameters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None or self.parameters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_InstructionFunc); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_200977137);
    __Pyx_GIVEREF(__pyx_int_200977137);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_200977137)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None or self.parameters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, None), state
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_InstructionFunc); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_200977137);
    __Pyx_GIVEREF(__pyx_int_200977137);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_200977137)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_15InstructionFunc_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_15InstructionFunc_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionFunc__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionFunc.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":225
 * 
 * cdef class InstructionObjectInt(Instruction):
 *     def __init__(self, OpCode code, object obj, int64_t value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.obj = obj
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code;
  PyObject *__pyx_v_obj = 0;
  int64_t __pyx_v_value;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,&__pyx_n_s_obj,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_code)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 225, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 225, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 225, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_code = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L3_error)
    __pyx_v_obj = values[1];
    __pyx_v_value = __Pyx_PyInt_As_int64_t(values[2]); if (unlikely((__pyx_v_value == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 225, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt___init__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self), __pyx_v_code, __pyx_v_obj, __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt___init__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self, enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_v_code, PyObject *__pyx_v_obj, int64_t __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/vm.pyx":226
 * cdef class InstructionObjectInt(Instruction):
 *     def __init__(self, OpCode code, object obj, int64_t value):
 *         super().__init__(code)             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.value = value
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt))) __PYX_ERR(0, 226, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 226, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":227
 *     def __init__(self, OpCode code, object obj, int64_t value):
 *         super().__init__(code)
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "flitter/language/vm.pyx":228
 *         super().__init__(code)
 *         self.obj = obj
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->value = __pyx_v_value;

  /* "flitter/language/vm.pyx":225
 * 
 * cdef class InstructionObjectInt(Instruction):
 *     def __init__(self, OpCode code, object obj, int64_t value):             # <<<<<<<<<<<<<<
 *         super().__init__(code)
 *         self.obj = obj
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":230
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.obj!r} {self.value!r}'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_2__str__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_2__str__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":231
 * 
 *     def __str__(self):
 *         return f'{OpCodeNames[self.code]} {self.obj!r} {self.value!r}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 231, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->obj), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":230
 *         self.value = value
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f'{OpCodeNames[self.code]} {self.obj!r} {self.value!r}'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":135
 * 
 * cdef class InstructionObjectInt(Instruction):
 *     cdef readonly object obj             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3obj_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3obj_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_3obj___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_3obj___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":136
 * cdef class InstructionObjectInt(Instruction):
 *     cdef readonly object obj
 *     cdef readonly int64_t value             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_5value___get__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_5value___get__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.obj, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_v_self->__pyx_base.code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->obj);
  __Pyx_GIVEREF(__pyx_v_self->obj);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->obj)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.obj, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.obj, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.obj is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.obj, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.obj is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->obj != Py_None);
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.obj is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_InstructionObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_59839126);
    __Pyx_GIVEREF(__pyx_int_59839126);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_59839126)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, None), state
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_InstructionObject); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_59839126);
    __Pyx_GIVEREF(__pyx_int_59839126);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_59839126)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20InstructionObjectInt_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionObjectInt__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.InstructionObjectInt.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":247
 * 
 * cdef class VectorStack:
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):             # <<<<<<<<<<<<<<
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_11VectorStack_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_11VectorStack_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int64_t __pyx_v_size;
  int64_t __pyx_v_max_size;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_max_size,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_size);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_max_size);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 247, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_size == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L3_error)
    } else {
      __pyx_v_size = __pyx_k__4;
    }
    if (values[1]) {
      __pyx_v_max_size = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_max_size == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L3_error)
    } else {
      __pyx_v_max_size = __pyx_k__5;
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, __pyx_nargs); __PYX_ERR(0, 247, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack___cinit__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_size, __pyx_v_max_size);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_11VectorStack___cinit__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_size, int64_t __pyx_v_max_size) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/language/vm.pyx":248
 * cdef class VectorStack:
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)             # <<<<<<<<<<<<<<
 *         if self.vectors == NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->vectors = ((PyObject **)PyMem_Malloc(((sizeof(PyObject *)) * __pyx_v_size)));

  /* "flitter/language/vm.pyx":249
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self.size = size
 */
  __pyx_t_1 = (__pyx_v_self->vectors == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":250
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         self.size = size
 *         self.max_size = max_size
 */
    PyErr_NoMemory(); __PYX_ERR(0, 250, __pyx_L1_error)

    /* "flitter/language/vm.pyx":249
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self.size = size
 */
  }

  /* "flitter/language/vm.pyx":251
 *         if self.vectors == NULL:
 *             raise MemoryError()
 *         self.size = size             # <<<<<<<<<<<<<<
 *         self.max_size = max_size
 *         self.top = -1
 */
  __pyx_v_self->size = __pyx_v_size;

  /* "flitter/language/vm.pyx":252
 *             raise MemoryError()
 *         self.size = size
 *         self.max_size = max_size             # <<<<<<<<<<<<<<
 *         self.top = -1
 * 
 */
  __pyx_v_self->max_size = __pyx_v_max_size;

  /* "flitter/language/vm.pyx":253
 *         self.size = size
 *         self.max_size = max_size
 *         self.top = -1             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->top = -1L;

  /* "flitter/language/vm.pyx":247
 * 
 * cdef class VectorStack:
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):             # <<<<<<<<<<<<<<
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/language/vm.pyx":255
 *         self.top = -1
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(self.top+1):
 */

/* Python wrapper */
static void __pyx_pw_7flitter_8language_2vm_11VectorStack_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7flitter_8language_2vm_11VectorStack_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7flitter_8language_2vm_11VectorStack_2__dealloc__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7flitter_8language_2vm_11VectorStack_2__dealloc__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  int64_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "flitter/language/vm.pyx":257
 *     def __dealloc__(self):
 *         cdef int64_t i
 *         for i in range(self.top+1):             # <<<<<<<<<<<<<<
 *             Py_DECREF(<Vector>self.vectors[i])
 *             self.vectors[i] = NULL
 */
  __pyx_t_1 = (__pyx_v_self->top + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/language/vm.pyx":258
 *         cdef int64_t i
 *         for i in range(self.top+1):
 *             Py_DECREF(<Vector>self.vectors[i])             # <<<<<<<<<<<<<<
 *             self.vectors[i] = NULL
 *         PyMem_Free(self.vectors)
 */
    __pyx_t_4 = ((PyObject *)(__pyx_v_self->vectors[__pyx_v_i]));
    __Pyx_INCREF(__pyx_t_4);
    Py_DECREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/vm.pyx":259
 *         for i in range(self.top+1):
 *             Py_DECREF(<Vector>self.vectors[i])
 *             self.vectors[i] = NULL             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.vectors)
 * 
 */
    (__pyx_v_self->vectors[__pyx_v_i]) = NULL;
  }

  /* "flitter/language/vm.pyx":260
 *             Py_DECREF(<Vector>self.vectors[i])
 *             self.vectors[i] = NULL
 *         PyMem_Free(self.vectors)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  PyMem_Free(__pyx_v_self->vectors);

  /* "flitter/language/vm.pyx":255
 *         self.top = -1
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(self.top+1):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":262
 *         PyMem_Free(self.vectors)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.top + 1
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_4__len__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_7flitter_8language_2vm_11VectorStack_4__len__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "flitter/language/vm.pyx":263
 * 
 *     def __len__(self):
 *         return self.top + 1             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_r = (__pyx_v_self->top + 1);
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":262
 *         PyMem_Free(self.vectors)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.top + 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/language/vm.pyx":265
 *         return self.top + 1
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         cdef list items = []
 *         cdef int64_t i
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_7__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_6__str__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_6__str__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_v_items = 0;
  int64_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":266
 * 
 *     def __str__(self):
 *         cdef list items = []             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(self.top+1):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":268
 *         cdef list items = []
 *         cdef int64_t i
 *         for i in range(self.top+1):             # <<<<<<<<<<<<<<
 *             items.append(f"#{self.top+1-i}: {<Vector>self.vectors[i]!r}")
 *         return '\n'.join(items)
 */
  __pyx_t_2 = (__pyx_v_self->top + 1);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "flitter/language/vm.pyx":269
 *         cdef int64_t i
 *         for i in range(self.top+1):
 *             items.append(f"#{self.top+1-i}: {<Vector>self.vectors[i]!r}")             # <<<<<<<<<<<<<<
 *         return '\n'.join(items)
 * 
 */
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_5 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__6);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(((__pyx_v_self->top + 1) - __pyx_v_i)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_5 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)(__pyx_v_self->vectors[__pyx_v_i]))), __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }

  /* "flitter/language/vm.pyx":270
 *         for i in range(self.top+1):
 *             items.append(f"#{self.top+1-i}: {<Vector>self.vectors[i]!r}")
 *         return '\n'.join(items)             # <<<<<<<<<<<<<<
 * 
 *     cpdef VectorStack copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyUnicode_Join(__pyx_kp_u__8, __pyx_v_items); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":265
 *         return self.top + 1
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         cdef list items = []
 *         cdef int64_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":272
 *         return '\n'.join(items)
 * 
 *     cpdef VectorStack copy(self):             # <<<<<<<<<<<<<<
 *         cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)
 *         cdef int64_t i
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_9copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_f_7flitter_8language_2vm_11VectorStack_copy(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_new_stack = 0;
  int64_t __pyx_v_i;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int64_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_9copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 272, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":273
 * 
 *     cpdef VectorStack copy(self):
 *         cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         cdef Vector value
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_7flitter_8language_2vm_VectorStack(((PyTypeObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack), __pyx_t_2, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_new_stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":276
 *         cdef int64_t i
 *         cdef Vector value
 *         for i in range(self.top+1):             # <<<<<<<<<<<<<<
 *             value = (<Vector>self.vectors[i]).copy()
 *             Py_INCREF(value)
 */
  __pyx_t_6 = (__pyx_v_self->top + 1);
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "flitter/language/vm.pyx":277
 *         cdef Vector value
 *         for i in range(self.top+1):
 *             value = (<Vector>self.vectors[i]).copy()             # <<<<<<<<<<<<<<
 *             Py_INCREF(value)
 *             new_stack.vectors[i] = <PyObject*>value
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)(__pyx_v_self->vectors[__pyx_v_i]))->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)(__pyx_v_self->vectors[__pyx_v_i])), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":278
 *         for i in range(self.top+1):
 *             value = (<Vector>self.vectors[i]).copy()
 *             Py_INCREF(value)             # <<<<<<<<<<<<<<
 *             new_stack.vectors[i] = <PyObject*>value
 *         new_stack.top = self.top
 */
    Py_INCREF(((PyObject *)__pyx_v_value));

    /* "flitter/language/vm.pyx":279
 *             value = (<Vector>self.vectors[i]).copy()
 *             Py_INCREF(value)
 *             new_stack.vectors[i] = <PyObject*>value             # <<<<<<<<<<<<<<
 *         new_stack.top = self.top
 *         return new_stack
 */
    (__pyx_v_new_stack->vectors[__pyx_v_i]) = ((PyObject *)__pyx_v_value);
  }

  /* "flitter/language/vm.pyx":280
 *             Py_INCREF(value)
 *             new_stack.vectors[i] = <PyObject*>value
 *         new_stack.top = self.top             # <<<<<<<<<<<<<<
 *         return new_stack
 * 
 */
  __pyx_t_6 = __pyx_v_self->top;
  __pyx_v_new_stack->top = __pyx_t_6;

  /* "flitter/language/vm.pyx":281
 *             new_stack.vectors[i] = <PyObject*>value
 *         new_stack.top = self.top
 *         return new_stack             # <<<<<<<<<<<<<<
 * 
 *     cpdef void drop(self, int64_t count=1):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_stack);
  __pyx_r = __pyx_v_new_stack;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":272
 *         return '\n'.join(items)
 * 
 *     cpdef VectorStack copy(self):             # <<<<<<<<<<<<<<
 *         cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)
 *         cdef int64_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_stack);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_9copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_9copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_9copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_9copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_8copy(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_8copy(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_11VectorStack_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":283
 *         return new_stack
 * 
 *     cpdef void drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         drop(self, count)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_11drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop *__pyx_optional_args) {
  int64_t __pyx_v_count = ((int64_t)1);
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("drop", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_count = __pyx_optional_args->count;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_drop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_11drop)) {
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":284
 * 
 *     cpdef void drop(self, int64_t count=1):
 *         drop(self, count)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void push(self, Vector vector):
 */
  __pyx_f_7flitter_8language_2vm_drop(__pyx_v_self, __pyx_v_count); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 284, __pyx_L1_error)

  /* "flitter/language/vm.pyx":283
 *         return new_stack
 * 
 *     cpdef void drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         drop(self, count)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_11drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_11drop = {"drop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_11drop, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_11drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("drop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "drop") < 0)) __PYX_ERR(0, 283, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L3_error)
    } else {
      __pyx_v_count = ((int64_t)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("drop", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 283, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_10drop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_10drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("drop", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1.__pyx_n = 1;
  __pyx_t_1.count = __pyx_v_count;
  __pyx_vtabptr_7flitter_8language_2vm_VectorStack->drop(__pyx_v_self, 1, &__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L1_error)
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":286
 *         drop(self, count)
 * 
 *     cpdef void push(self, Vector vector):             # <<<<<<<<<<<<<<
 *         push(self, vector)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_13push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("push", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_push); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_13push)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_vector)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":287
 * 
 *     cpdef void push(self, Vector vector):
 *         push(self, vector)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector pop(self):
 */
  __pyx_t_6 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_self, __pyx_v_vector); if (unlikely(__pyx_t_6 == ((int64_t)0))) __PYX_ERR(0, 287, __pyx_L1_error)

  /* "flitter/language/vm.pyx":286
 *         drop(self, count)
 * 
 *     cpdef void push(self, Vector vector):             # <<<<<<<<<<<<<<
 *         push(self, vector)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_13push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_13push = {"push", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_13push, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_13push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("push (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vector,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_vector)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "push") < 0)) __PYX_ERR(0, 286, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_vector = ((struct __pyx_obj_7flitter_5model_Vector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("push", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 286, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vector), __pyx_ptype_7flitter_5model_Vector, 1, "vector", 0))) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_12push(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_vector);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_12push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("push", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7flitter_8language_2vm_11VectorStack_push(__pyx_v_self, __pyx_v_vector, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":289
 *         push(self, vector)
 * 
 *     cpdef Vector pop(self):             # <<<<<<<<<<<<<<
 *         return pop(self)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_15pop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_15pop)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 289, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":290
 * 
 *     cpdef Vector pop(self):
 *         return pop(self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":289
 *         push(self, vector)
 * 
 *     cpdef Vector pop(self):             # <<<<<<<<<<<<<<
 *         return pop(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_15pop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_15pop = {"pop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_15pop, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_15pop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pop", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pop", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_14pop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_14pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_11VectorStack_pop(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":292
 *         return pop(self)
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_tuple(self, count)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_17pop_tuple(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_tuple", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_17pop_tuple)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 292, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_2))) __PYX_ERR(0, 292, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":293
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):
 *         return pop_tuple(self, count)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list pop_list(self, int64_t count):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_pop_tuple(__pyx_v_self, __pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":292
 *         return pop(self)
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_tuple(self, count)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_17pop_tuple(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_17pop_tuple = {"pop_tuple", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_17pop_tuple, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_17pop_tuple(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop_tuple (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 292, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pop_tuple") < 0)) __PYX_ERR(0, 292, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 292, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop_tuple", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 292, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_16pop_tuple(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_16pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_tuple", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_11VectorStack_pop_tuple(__pyx_v_self, __pyx_v_count, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":295
 *         return pop_tuple(self, count)
 * 
 *     cpdef list pop_list(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_list(self, count)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_19pop_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_list", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_19pop_list)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 295, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":296
 * 
 *     cpdef list pop_list(self, int64_t count):
 *         return pop_list(self, count)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict pop_dict(self, tuple keys):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_pop_list(__pyx_v_self, __pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":295
 *         return pop_tuple(self, count)
 * 
 *     cpdef list pop_list(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_list(self, count)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_19pop_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_19pop_list = {"pop_list", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_19pop_list, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_19pop_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop_list (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pop_list") < 0)) __PYX_ERR(0, 295, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop_list", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 295, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_18pop_list(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_18pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_list", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_11VectorStack_pop_list(__pyx_v_self, __pyx_v_count, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":298
 *         return pop_list(self, count)
 * 
 *     cpdef dict pop_dict(self, tuple keys):             # <<<<<<<<<<<<<<
 *         return pop_dict(self, keys)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_21pop_dict(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, PyObject *__pyx_v_keys, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_dict", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_21pop_dict)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_keys};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(0, 298, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":299
 * 
 *     cpdef dict pop_dict(self, tuple keys):
 *         return pop_dict(self, keys)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector pop_composed(self, int64_t count):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_pop_dict(__pyx_v_self, __pyx_v_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":298
 *         return pop_list(self, count)
 * 
 *     cpdef dict pop_dict(self, tuple keys):             # <<<<<<<<<<<<<<
 *         return pop_dict(self, keys)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_21pop_dict(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_21pop_dict = {"pop_dict", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_21pop_dict, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_21pop_dict(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_keys = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop_dict (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keys,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keys)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pop_dict") < 0)) __PYX_ERR(0, 298, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_keys = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop_dict", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 298, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_keys), (&PyTuple_Type), 1, "keys", 1))) __PYX_ERR(0, 298, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_20pop_dict(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_keys);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_20pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, PyObject *__pyx_v_keys) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_dict", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_11VectorStack_pop_dict(__pyx_v_self, __pyx_v_keys, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":301
 *         return pop_dict(self, keys)
 * 
 *     cpdef Vector pop_composed(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_composed(self, count)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_23pop_composed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_composed", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop_composed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_23pop_composed)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 301, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":302
 * 
 *     cpdef Vector pop_composed(self, int64_t count):
 *         return pop_composed(self, count)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector peek(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop_composed(__pyx_v_self, __pyx_v_count)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":301
 *         return pop_dict(self, keys)
 * 
 *     cpdef Vector pop_composed(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_composed(self, count)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_composed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_23pop_composed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_23pop_composed = {"pop_composed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_23pop_composed, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_23pop_composed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop_composed (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pop_composed") < 0)) __PYX_ERR(0, 301, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop_composed", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 301, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_composed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_22pop_composed(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_22pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_composed", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_11VectorStack_pop_composed(__pyx_v_self, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.pop_composed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":304
 *         return pop_composed(self, count)
 * 
 *     cpdef Vector peek(self):             # <<<<<<<<<<<<<<
 *         return peek(self)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_25peek(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_peek); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_25peek)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 304, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":305
 * 
 *     cpdef Vector peek(self):
 *         return peek(self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector peek_at(self, int64_t offset):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":304
 *         return pop_composed(self, count)
 * 
 *     cpdef Vector peek(self):             # <<<<<<<<<<<<<<
 *         return peek(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_25peek(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_25peek = {"peek", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_25peek, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_25peek(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("peek (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("peek", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "peek", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_24peek(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_24peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_11VectorStack_peek(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":307
 *         return peek(self)
 * 
 *     cpdef Vector peek_at(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         return peek_at(self, offset)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_27peek_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_11VectorStack_peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek_at", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_peek_at); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_27peek_at)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 307, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":308
 * 
 *     cpdef Vector peek_at(self, int64_t offset):
 *         return peek_at(self, offset)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void poke(self, Vector vector):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_self, __pyx_v_offset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":307
 *         return peek(self)
 * 
 *     cpdef Vector peek_at(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         return peek_at(self, offset)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.peek_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_27peek_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_27peek_at = {"peek_at", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_27peek_at, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_27peek_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_offset;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("peek_at (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_offset,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_offset)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "peek_at") < 0)) __PYX_ERR(0, 307, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_offset = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_offset == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("peek_at", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 307, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.peek_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_26peek_at(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_offset);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_26peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek_at", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_11VectorStack_peek_at(__pyx_v_self, __pyx_v_offset, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.peek_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":310
 *         return peek_at(self, offset)
 * 
 *     cpdef void poke(self, Vector vector):             # <<<<<<<<<<<<<<
 *         poke(self, vector)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_29poke(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_poke); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_29poke)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_vector)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":311
 * 
 *     cpdef void poke(self, Vector vector):
 *         poke(self, vector)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):
 */
  __pyx_f_7flitter_8language_2vm_poke(__pyx_v_self, __pyx_v_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L1_error)

  /* "flitter/language/vm.pyx":310
 *         return peek_at(self, offset)
 * 
 *     cpdef void poke(self, Vector vector):             # <<<<<<<<<<<<<<
 *         poke(self, vector)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_29poke(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_29poke = {"poke", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_29poke, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_29poke(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poke (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vector,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_vector)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "poke") < 0)) __PYX_ERR(0, 310, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_vector = ((struct __pyx_obj_7flitter_5model_Vector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poke", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 310, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vector), __pyx_ptype_7flitter_5model_Vector, 1, "vector", 0))) __PYX_ERR(0, 310, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_28poke(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_vector);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_28poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7flitter_8language_2vm_11VectorStack_poke(__pyx_v_self, __pyx_v_vector, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":313
 *         poke(self, vector)
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *         poke_at(self, offset, vector)
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_31poke_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7flitter_8language_2vm_11VectorStack_poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke_at", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_poke_at); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_11VectorStack_31poke_at)) {
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, ((PyObject *)__pyx_v_vector)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 313, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":314
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):
 *         poke_at(self, offset, vector)             # <<<<<<<<<<<<<<
 * 
 * cdef int64_t increase(VectorStack stack) except 0:
 */
  __pyx_f_7flitter_8language_2vm_poke_at(__pyx_v_self, __pyx_v_offset, __pyx_v_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 314, __pyx_L1_error)

  /* "flitter/language/vm.pyx":313
 *         poke(self, vector)
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *         poke_at(self, offset, vector)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_31poke_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_31poke_at = {"poke_at", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_31poke_at, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_31poke_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_offset;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poke_at (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_offset,&__pyx_n_s_vector,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_offset)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_vector)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("poke_at", 1, 2, 2, 1); __PYX_ERR(0, 313, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "poke_at") < 0)) __PYX_ERR(0, 313, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_offset = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_offset == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L3_error)
    __pyx_v_vector = ((struct __pyx_obj_7flitter_5model_Vector *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poke_at", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 313, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vector), __pyx_ptype_7flitter_5model_Vector, 1, "vector", 0))) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_30poke_at(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v_offset, __pyx_v_vector);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_30poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, int64_t __pyx_v_offset, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke_at", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7flitter_8language_2vm_11VectorStack_poke_at(__pyx_v_self, __pyx_v_offset, __pyx_v_vector, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.poke_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":17
 *     cdef PyObject** vectors
 *     cdef int64_t top
 *     cdef readonly int64_t size             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t max_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_4size___get__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_4size___get__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":18
 *     cdef int64_t top
 *     cdef readonly int64_t size
 *     cdef readonly int64_t max_size             # <<<<<<<<<<<<<<
 * 
 *     cpdef VectorStack copy(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_8max_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_8max_size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_8max_size___get__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_8max_size___get__(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->max_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.max_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_33__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_33__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_33__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_33__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_32__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_32__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_35__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11VectorStack_35__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_35__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11VectorStack_35__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_11VectorStack_34__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_11VectorStack_34__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.VectorStack.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":316
 *         poke_at(self, offset, vector)
 * 
 * cdef int64_t increase(VectorStack stack) except 0:             # <<<<<<<<<<<<<<
 *     if stack.size == stack.max_size:
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 */

static int64_t __pyx_f_7flitter_8language_2vm_increase(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack) {
  int64_t __pyx_v_new_size;
  int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("increase", 1);

  /* "flitter/language/vm.pyx":317
 * 
 * cdef int64_t increase(VectorStack stack) except 0:
 *     if stack.size == stack.max_size:             # <<<<<<<<<<<<<<
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 */
  __pyx_t_1 = (__pyx_v_stack->size == __pyx_v_stack->max_size);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":318
 * cdef int64_t increase(VectorStack stack) except 0:
 *     if stack.size == stack.max_size:
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")             # <<<<<<<<<<<<<<
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackOverflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Stack_cannot_grow_beyond);
    __pyx_t_5 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_Stack_cannot_grow_beyond);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Stack_cannot_grow_beyond);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_stack->max_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items);
    __pyx_t_5 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_items);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items);
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 318, __pyx_L1_error)

    /* "flitter/language/vm.pyx":317
 * 
 * cdef int64_t increase(VectorStack stack) except 0:
 *     if stack.size == stack.max_size:             # <<<<<<<<<<<<<<
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 */
  }

  /* "flitter/language/vm.pyx":319
 *     if stack.size == stack.max_size:
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)             # <<<<<<<<<<<<<<
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
 *     if stack.vectors == NULL:
 */
  __pyx_t_10 = __pyx_v_stack->max_size;
  __pyx_t_11 = (__pyx_v_stack->size * 2);
  __pyx_t_1 = (__pyx_t_10 < __pyx_t_11);
  if (__pyx_t_1) {
    __pyx_t_12 = __pyx_t_10;
  } else {
    __pyx_t_12 = __pyx_t_11;
  }
  __pyx_v_new_size = __pyx_t_12;

  /* "flitter/language/vm.pyx":320
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)             # <<<<<<<<<<<<<<
 *     if stack.vectors == NULL:
 *         raise MemoryError()
 */
  __pyx_v_stack->vectors = ((PyObject **)PyMem_Realloc(__pyx_v_stack->vectors, ((sizeof(PyObject *)) * __pyx_v_new_size)));

  /* "flitter/language/vm.pyx":321
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
 *     if stack.vectors == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     stack.size = new_size
 */
  __pyx_t_1 = (__pyx_v_stack->vectors == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":322
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
 *     if stack.vectors == NULL:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     stack.size = new_size
 *     return new_size
 */
    PyErr_NoMemory(); __PYX_ERR(0, 322, __pyx_L1_error)

    /* "flitter/language/vm.pyx":321
 *     cdef int64_t new_size = min(stack.size * 2, stack.max_size)
 *     stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
 *     if stack.vectors == NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     stack.size = new_size
 */
  }

  /* "flitter/language/vm.pyx":323
 *     if stack.vectors == NULL:
 *         raise MemoryError()
 *     stack.size = new_size             # <<<<<<<<<<<<<<
 *     return new_size
 * 
 */
  __pyx_v_stack->size = __pyx_v_new_size;

  /* "flitter/language/vm.pyx":324
 *         raise MemoryError()
 *     stack.size = new_size
 *     return new_size             # <<<<<<<<<<<<<<
 * 
 * cdef inline void drop(VectorStack stack, int64_t n):
 */
  __pyx_r = __pyx_v_new_size;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":316
 *         poke_at(self, offset, vector)
 * 
 * cdef int64_t increase(VectorStack stack) except 0:             # <<<<<<<<<<<<<<
 *     if stack.size == stack.max_size:
 *         raise StackOverflow(f"Stack cannot grow beyond {stack.max_size} items")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.increase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":326
 *     return new_size
 * 
 * cdef inline void drop(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if stack.top -n < -1:
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")
 */

static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_drop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_n) {
  int64_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  PyObject *__pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("drop", 1);

  /* "flitter/language/vm.pyx":327
 * 
 * cdef inline void drop(VectorStack stack, int64_t n):
 *     if stack.top -n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")
 *     stack.top -= n
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_n) < -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":328
 * cdef inline void drop(VectorStack stack, int64_t n):
 *     if stack.top -n < -1:
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")             # <<<<<<<<<<<<<<
 *     stack.top -= n
 *     cdef int64_t i
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_drop);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_drop);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_drop);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_from);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_items_from);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_from);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_depth_stack);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_depth_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_depth_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 328, __pyx_L1_error)

    /* "flitter/language/vm.pyx":327
 * 
 * cdef inline void drop(VectorStack stack, int64_t n):
 *     if stack.top -n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")
 *     stack.top -= n
 */
  }

  /* "flitter/language/vm.pyx":329
 *     if stack.top -n < -1:
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")
 *     stack.top -= n             # <<<<<<<<<<<<<<
 *     cdef int64_t i
 *     for i in range(1, n+1):
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - __pyx_v_n);

  /* "flitter/language/vm.pyx":331
 *     stack.top -= n
 *     cdef int64_t i
 *     for i in range(1, n+1):             # <<<<<<<<<<<<<<
 *         Py_DECREF(<Vector>stack.vectors[stack.top+i])
 *         stack.vectors[stack.top+i] = NULL
 */
  __pyx_t_10 = (__pyx_v_n + 1);
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 1; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":332
 *     cdef int64_t i
 *     for i in range(1, n+1):
 *         Py_DECREF(<Vector>stack.vectors[stack.top+i])             # <<<<<<<<<<<<<<
 *         stack.vectors[stack.top+i] = NULL
 * 
 */
    __pyx_t_13 = (__pyx_v_stack->vectors[(__pyx_v_stack->top + __pyx_v_i)]);
    Py_DECREF(((PyObject *)__pyx_t_13));

    /* "flitter/language/vm.pyx":333
 *     for i in range(1, n+1):
 *         Py_DECREF(<Vector>stack.vectors[stack.top+i])
 *         stack.vectors[stack.top+i] = NULL             # <<<<<<<<<<<<<<
 * 
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:
 */
    (__pyx_v_stack->vectors[(__pyx_v_stack->top + __pyx_v_i)]) = NULL;
  }

  /* "flitter/language/vm.pyx":326
 *     return new_size
 * 
 * cdef inline void drop(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if stack.top -n < -1:
 *         raise StackUnderflow(f"Cannot drop {n} items from {stack.top+1} depth stack")
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":335
 *         stack.vectors[stack.top+i] = NULL
 * 
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top + 1 == stack.size:
 */

static CYTHON_INLINE int64_t __pyx_f_7flitter_8language_2vm_push(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  int64_t __pyx_r;
  int __pyx_t_1;
  int64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/language/vm.pyx":336
 * 
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:
 *     assert vector is not None             # <<<<<<<<<<<<<<
 *     if stack.top + 1 == stack.size:
 *         increase(stack)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((PyObject *)__pyx_v_vector) != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 336, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 336, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":337
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:
 *     assert vector is not None
 *     if stack.top + 1 == stack.size:             # <<<<<<<<<<<<<<
 *         increase(stack)
 *     stack.top += 1
 */
  __pyx_t_1 = ((__pyx_v_stack->top + 1) == __pyx_v_stack->size);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":338
 *     assert vector is not None
 *     if stack.top + 1 == stack.size:
 *         increase(stack)             # <<<<<<<<<<<<<<
 *     stack.top += 1
 *     Py_INCREF(vector)
 */
    __pyx_t_2 = __pyx_f_7flitter_8language_2vm_increase(__pyx_v_stack); if (unlikely(__pyx_t_2 == ((int64_t)0))) __PYX_ERR(0, 338, __pyx_L1_error)

    /* "flitter/language/vm.pyx":337
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:
 *     assert vector is not None
 *     if stack.top + 1 == stack.size:             # <<<<<<<<<<<<<<
 *         increase(stack)
 *     stack.top += 1
 */
  }

  /* "flitter/language/vm.pyx":339
 *     if stack.top + 1 == stack.size:
 *         increase(stack)
 *     stack.top += 1             # <<<<<<<<<<<<<<
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top] = <PyObject*>vector
 */
  __pyx_v_stack->top = (__pyx_v_stack->top + 1);

  /* "flitter/language/vm.pyx":340
 *         increase(stack)
 *     stack.top += 1
 *     Py_INCREF(vector)             # <<<<<<<<<<<<<<
 *     stack.vectors[stack.top] = <PyObject*>vector
 *     return stack.size
 */
  Py_INCREF(((PyObject *)__pyx_v_vector));

  /* "flitter/language/vm.pyx":341
 *     stack.top += 1
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top] = <PyObject*>vector             # <<<<<<<<<<<<<<
 *     return stack.size
 * 
 */
  (__pyx_v_stack->vectors[__pyx_v_stack->top]) = ((PyObject *)__pyx_v_vector);

  /* "flitter/language/vm.pyx":342
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top] = <PyObject*>vector
 *     return stack.size             # <<<<<<<<<<<<<<
 * 
 * cdef inline Vector pop(VectorStack stack):
 */
  __pyx_r = __pyx_v_stack->size;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":335
 *         stack.vectors[stack.top+i] = NULL
 * 
 * cdef inline int64_t push(VectorStack stack, Vector vector) except 0:             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top + 1 == stack.size:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/language/vm.pyx":344
 *     return stack.size
 * 
 * cdef inline Vector pop(VectorStack stack):             # <<<<<<<<<<<<<<
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 */

static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_pop(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 1);

  /* "flitter/language/vm.pyx":345
 * 
 * cdef inline Vector pop(VectorStack stack):
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]
 */
  __pyx_t_1 = (__pyx_v_stack->top == -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":346
 * cdef inline Vector pop(VectorStack stack):
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")             # <<<<<<<<<<<<<<
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]
 *     stack.vectors[stack.top] = NULL
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_Stack_empty};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 346, __pyx_L1_error)

    /* "flitter/language/vm.pyx":345
 * 
 * cdef inline Vector pop(VectorStack stack):
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]
 */
  }

  /* "flitter/language/vm.pyx":347
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]             # <<<<<<<<<<<<<<
 *     stack.vectors[stack.top] = NULL
 *     stack.top -= 1
 */
  __pyx_t_2 = ((PyObject *)(__pyx_v_stack->vectors[__pyx_v_stack->top]));
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":348
 *         raise StackUnderflow("Stack empty")
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]
 *     stack.vectors[stack.top] = NULL             # <<<<<<<<<<<<<<
 *     stack.top -= 1
 *     Py_DECREF(vector)
 */
  (__pyx_v_stack->vectors[__pyx_v_stack->top]) = NULL;

  /* "flitter/language/vm.pyx":349
 *     cdef Vector vector = <Vector>stack.vectors[stack.top]
 *     stack.vectors[stack.top] = NULL
 *     stack.top -= 1             # <<<<<<<<<<<<<<
 *     Py_DECREF(vector)
 *     return vector
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - 1);

  /* "flitter/language/vm.pyx":350
 *     stack.vectors[stack.top] = NULL
 *     stack.top -= 1
 *     Py_DECREF(vector)             # <<<<<<<<<<<<<<
 *     return vector
 * 
 */
  Py_DECREF(((PyObject *)__pyx_v_vector));

  /* "flitter/language/vm.pyx":351
 *     stack.top -= 1
 *     Py_DECREF(vector)
 *     return vector             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_vector);
  __pyx_r = __pyx_v_vector;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":344
 *     return stack.size
 * 
 * cdef inline Vector pop(VectorStack stack):             # <<<<<<<<<<<<<<
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_vector);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":353
 *     return vector
 * 
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return ()
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_tuple(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_n) {
  PyObject *__pyx_v_t = 0;
  PyObject *__pyx_v_ptr;
  int64_t __pyx_v_i;
  int64_t __pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_tuple", 1);

  /* "flitter/language/vm.pyx":354
 * 
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return ()
 *     if stack.top - n < -1:
 */
  __pyx_t_1 = (__pyx_v_n == 0);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":355
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):
 *     if n == 0:
 *         return ()             # <<<<<<<<<<<<<<
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":354
 * 
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return ()
 *     if stack.top - n < -1:
 */
  }

  /* "flitter/language/vm.pyx":356
 *     if n == 0:
 *         return ()
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef tuple t = PyTuple_New(n)
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_n) < -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":357
 *         return ()
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")             # <<<<<<<<<<<<<<
 *     cdef tuple t = PyTuple_New(n)
 *     stack.top -= n
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_pop);
    __pyx_t_5 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_pop);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_pop);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_from);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_items_from);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_from);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_depth_stack);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_depth_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_depth_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 357, __pyx_L1_error)

    /* "flitter/language/vm.pyx":356
 *     if n == 0:
 *         return ()
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef tuple t = PyTuple_New(n)
 */
  }

  /* "flitter/language/vm.pyx":358
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef tuple t = PyTuple_New(n)             # <<<<<<<<<<<<<<
 *     stack.top -= n
 *     cdef PyObject* ptr
 */
  __pyx_t_2 = PyTuple_New(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_t = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":359
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef tuple t = PyTuple_New(n)
 *     stack.top -= n             # <<<<<<<<<<<<<<
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - __pyx_v_n);

  /* "flitter/language/vm.pyx":361
 *     stack.top -= n
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 */
  __pyx_v_base = (__pyx_v_stack->top + 1);

  /* "flitter/language/vm.pyx":362
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ptr = stack.vectors[base+i]
 *         PyTuple_SET_ITEM(t, i, <Vector>ptr)
 */
  __pyx_t_10 = __pyx_v_n;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":363
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 */
    __pyx_v_ptr = (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]);

    /* "flitter/language/vm.pyx":364
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 *         PyTuple_SET_ITEM(t, i, <Vector>ptr)             # <<<<<<<<<<<<<<
 *         stack.vectors[base+i] = NULL
 *     return t
 */
    PyTuple_SET_ITEM(__pyx_v_t, __pyx_v_i, ((PyObject *)__pyx_v_ptr));

    /* "flitter/language/vm.pyx":365
 *         ptr = stack.vectors[base+i]
 *         PyTuple_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL             # <<<<<<<<<<<<<<
 *     return t
 * 
 */
    (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]) = NULL;
  }

  /* "flitter/language/vm.pyx":366
 *         PyTuple_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 *     return t             # <<<<<<<<<<<<<<
 * 
 * cdef inline list pop_list(VectorStack stack, int64_t n):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_t);
  __pyx_r = __pyx_v_t;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":353
 *     return vector
 * 
 * cdef inline tuple pop_tuple(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return ()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.pop_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":368
 *     return t
 * 
 * cdef inline list pop_list(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_list(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_n) {
  PyObject *__pyx_v_t = 0;
  PyObject *__pyx_v_ptr;
  int64_t __pyx_v_i;
  int64_t __pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_list", 1);

  /* "flitter/language/vm.pyx":369
 * 
 * cdef inline list pop_list(VectorStack stack, int64_t n):
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef list t = PyList_New(n)
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_n) < -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":370
 * cdef inline list pop_list(VectorStack stack, int64_t n):
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")             # <<<<<<<<<<<<<<
 *     cdef list t = PyList_New(n)
 *     stack.top -= n
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_pop);
    __pyx_t_5 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_pop);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_pop);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_from);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_items_from);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_from);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_depth_stack);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_depth_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_depth_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 370, __pyx_L1_error)

    /* "flitter/language/vm.pyx":369
 * 
 * cdef inline list pop_list(VectorStack stack, int64_t n):
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef list t = PyList_New(n)
 */
  }

  /* "flitter/language/vm.pyx":371
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef list t = PyList_New(n)             # <<<<<<<<<<<<<<
 *     stack.top -= n
 *     cdef PyObject* ptr
 */
  __pyx_t_2 = PyList_New(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_t = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":372
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef list t = PyList_New(n)
 *     stack.top -= n             # <<<<<<<<<<<<<<
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - __pyx_v_n);

  /* "flitter/language/vm.pyx":374
 *     stack.top -= n
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 */
  __pyx_v_base = (__pyx_v_stack->top + 1);

  /* "flitter/language/vm.pyx":375
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ptr = stack.vectors[base+i]
 *         PyList_SET_ITEM(t, i, <Vector>ptr)
 */
  __pyx_t_10 = __pyx_v_n;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":376
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 */
    __pyx_v_ptr = (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]);

    /* "flitter/language/vm.pyx":377
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 *         PyList_SET_ITEM(t, i, <Vector>ptr)             # <<<<<<<<<<<<<<
 *         stack.vectors[base+i] = NULL
 *     return t
 */
    PyList_SET_ITEM(__pyx_v_t, __pyx_v_i, ((PyObject *)__pyx_v_ptr));

    /* "flitter/language/vm.pyx":378
 *         ptr = stack.vectors[base+i]
 *         PyList_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL             # <<<<<<<<<<<<<<
 *     return t
 * 
 */
    (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]) = NULL;
  }

  /* "flitter/language/vm.pyx":379
 *         PyList_SET_ITEM(t, i, <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 *     return t             # <<<<<<<<<<<<<<
 * 
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_t);
  __pyx_r = __pyx_v_t;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":368
 *     return t
 * 
 * cdef inline list pop_list(VectorStack stack, int64_t n):             # <<<<<<<<<<<<<<
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.pop_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":381
 *     return t
 * 
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):             # <<<<<<<<<<<<<<
 *     cdef int64_t n = len(keys)
 *     if stack.top - n < -1:
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_pop_dict(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, PyObject *__pyx_v_keys) {
  int64_t __pyx_v_n;
  PyObject *__pyx_v_t = 0;
  PyObject *__pyx_v_ptr;
  int64_t __pyx_v_i;
  int64_t __pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  PyObject *__pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_dict", 1);

  /* "flitter/language/vm.pyx":382
 * 
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):
 *     cdef int64_t n = len(keys)             # <<<<<<<<<<<<<<
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 */
  if (unlikely(__pyx_v_keys == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 382, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_keys); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 382, __pyx_L1_error)
  __pyx_v_n = __pyx_t_1;

  /* "flitter/language/vm.pyx":383
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):
 *     cdef int64_t n = len(keys)
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef dict t = {}
 */
  __pyx_t_2 = ((__pyx_v_stack->top - __pyx_v_n) < -1L);
  if (unlikely(__pyx_t_2)) {

    /* "flitter/language/vm.pyx":384
 *     cdef int64_t n = len(keys)
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")             # <<<<<<<<<<<<<<
 *     cdef dict t = {}
 *     stack.top -= n
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_pop);
    __pyx_t_1 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_pop);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Cannot_pop);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_from);
    __pyx_t_1 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_items_from);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_items_from);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_depth_stack);
    __pyx_t_1 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_depth_stack);
    PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u_depth_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_5, 5, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 384, __pyx_L1_error)

    /* "flitter/language/vm.pyx":383
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):
 *     cdef int64_t n = len(keys)
 *     if stack.top - n < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef dict t = {}
 */
  }

  /* "flitter/language/vm.pyx":385
 *     if stack.top - n < -1:
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef dict t = {}             # <<<<<<<<<<<<<<
 *     stack.top -= n
 *     cdef PyObject* ptr
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_t = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "flitter/language/vm.pyx":386
 *         raise StackUnderflow(f"Cannot pop {n} items from {stack.top+1} depth stack")
 *     cdef dict t = {}
 *     stack.top -= n             # <<<<<<<<<<<<<<
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - __pyx_v_n);

  /* "flitter/language/vm.pyx":388
 *     stack.top -= n
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 */
  __pyx_v_base = (__pyx_v_stack->top + 1);

  /* "flitter/language/vm.pyx":389
 *     cdef PyObject* ptr
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ptr = stack.vectors[base+i]
 *         PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)
 */
  __pyx_t_10 = __pyx_v_n;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":390
 *     cdef int64_t i, base=stack.top+1
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 */
    __pyx_v_ptr = (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]);

    /* "flitter/language/vm.pyx":391
 *     for i in range(n):
 *         ptr = stack.vectors[base+i]
 *         PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)             # <<<<<<<<<<<<<<
 *         stack.vectors[base+i] = NULL
 *         Py_DECREF(<Vector>ptr)
 */
    __pyx_t_13 = PyTuple_GET_ITEM(__pyx_v_keys, __pyx_v_i);
    __pyx_t_14 = PyDict_SetItem(__pyx_v_t, ((PyObject *)__pyx_t_13), ((PyObject *)__pyx_v_ptr)); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 391, __pyx_L1_error)

    /* "flitter/language/vm.pyx":392
 *         ptr = stack.vectors[base+i]
 *         PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)
 *         stack.vectors[base+i] = NULL             # <<<<<<<<<<<<<<
 *         Py_DECREF(<Vector>ptr)
 *     return t
 */
    (__pyx_v_stack->vectors[(__pyx_v_base + __pyx_v_i)]) = NULL;

    /* "flitter/language/vm.pyx":393
 *         PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)
 *         stack.vectors[base+i] = NULL
 *         Py_DECREF(<Vector>ptr)             # <<<<<<<<<<<<<<
 *     return t
 * 
 */
    Py_DECREF(((PyObject *)__pyx_v_ptr));
  }

  /* "flitter/language/vm.pyx":394
 *         stack.vectors[base+i] = NULL
 *         Py_DECREF(<Vector>ptr)
 *     return t             # <<<<<<<<<<<<<<
 * 
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_t);
  __pyx_r = __pyx_v_t;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":381
 *     return t
 * 
 * cdef inline dict pop_dict(VectorStack stack, tuple keys):             # <<<<<<<<<<<<<<
 *     cdef int64_t n = len(keys)
 *     if stack.top - n < -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.pop_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":396
 *     return t
 * 
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):             # <<<<<<<<<<<<<<
 *     if stack.top - m < -1:
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 */

static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_pop_composed(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_m) {
  int64_t __pyx_v_i;
  int64_t __pyx_v_j;
  int64_t __pyx_v_k;
  int64_t __pyx_v_n;
  int64_t __pyx_v_base;
  int __pyx_v_numeric;
  PyObject *__pyx_v_ptr;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_result = 0;
  PyObject *__pyx_v_vobjects = 0;
  PyObject *__pyx_v_objects = 0;
  PyObject *__pyx_v_itemptr;
  PyObject *__pyx_v_obj = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int64_t __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int64_t __pyx_t_13;
  int64_t __pyx_t_14;
  int64_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_composed", 1);

  /* "flitter/language/vm.pyx":397
 * 
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):
 *     if stack.top - m < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 *     if m == 1:
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_m) < -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":398
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):
 *     if stack.top - m < -1:
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")             # <<<<<<<<<<<<<<
 *     if m == 1:
 *         return pop(stack)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_pop);
    __pyx_t_5 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_pop);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_pop);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_from);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_items_from);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_from);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_depth_stack);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_depth_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_depth_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 398, __pyx_L1_error)

    /* "flitter/language/vm.pyx":397
 * 
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):
 *     if stack.top - m < -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 *     if m == 1:
 */
  }

  /* "flitter/language/vm.pyx":399
 *     if stack.top - m < -1:
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 *     if m == 1:             # <<<<<<<<<<<<<<
 *         return pop(stack)
 *     if m == 0:
 */
  __pyx_t_1 = (__pyx_v_m == 1);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":400
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 *     if m == 1:
 *         return pop(stack)             # <<<<<<<<<<<<<<
 *     if m == 0:
 *         return null_
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":399
 *     if stack.top - m < -1:
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 *     if m == 1:             # <<<<<<<<<<<<<<
 *         return pop(stack)
 *     if m == 0:
 */
  }

  /* "flitter/language/vm.pyx":401
 *     if m == 1:
 *         return pop(stack)
 *     if m == 0:             # <<<<<<<<<<<<<<
 *         return null_
 *     stack.top -= m
 */
  __pyx_t_1 = (__pyx_v_m == 0);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":402
 *         return pop(stack)
 *     if m == 0:
 *         return null_             # <<<<<<<<<<<<<<
 *     stack.top -= m
 *     cdef int64_t i, j=0, k, n=0, base=stack.top+1
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_null_);
    __pyx_r = __pyx_v_7flitter_5model_null_;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":401
 *     if m == 1:
 *         return pop(stack)
 *     if m == 0:             # <<<<<<<<<<<<<<
 *         return null_
 *     stack.top -= m
 */
  }

  /* "flitter/language/vm.pyx":403
 *     if m == 0:
 *         return null_
 *     stack.top -= m             # <<<<<<<<<<<<<<
 *     cdef int64_t i, j=0, k, n=0, base=stack.top+1
 *     cdef bint numeric = True
 */
  __pyx_v_stack->top = (__pyx_v_stack->top - __pyx_v_m);

  /* "flitter/language/vm.pyx":404
 *         return null_
 *     stack.top -= m
 *     cdef int64_t i, j=0, k, n=0, base=stack.top+1             # <<<<<<<<<<<<<<
 *     cdef bint numeric = True
 *     cdef PyObject* ptr
 */
  __pyx_v_j = 0;
  __pyx_v_n = 0;
  __pyx_v_base = (__pyx_v_stack->top + 1);

  /* "flitter/language/vm.pyx":405
 *     stack.top -= m
 *     cdef int64_t i, j=0, k, n=0, base=stack.top+1
 *     cdef bint numeric = True             # <<<<<<<<<<<<<<
 *     cdef PyObject* ptr
 *     for i in range(base, base+m):
 */
  __pyx_v_numeric = 1;

  /* "flitter/language/vm.pyx":407
 *     cdef bint numeric = True
 *     cdef PyObject* ptr
 *     for i in range(base, base+m):             # <<<<<<<<<<<<<<
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is not None:
 */
  __pyx_t_10 = (__pyx_v_base + __pyx_v_m);
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = __pyx_v_base; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":408
 *     cdef PyObject* ptr
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]             # <<<<<<<<<<<<<<
 *         if (<Vector>ptr).objects is not None:
 *             numeric = False
 */
    __pyx_v_ptr = (__pyx_v_stack->vectors[__pyx_v_i]);

    /* "flitter/language/vm.pyx":409
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is not None:             # <<<<<<<<<<<<<<
 *             numeric = False
 *         n += (<Vector>ptr).length
 */
    __pyx_t_1 = (((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->objects != ((PyObject*)Py_None));
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":410
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is not None:
 *             numeric = False             # <<<<<<<<<<<<<<
 *         n += (<Vector>ptr).length
 *     if n == 0:
 */
      __pyx_v_numeric = 0;

      /* "flitter/language/vm.pyx":409
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is not None:             # <<<<<<<<<<<<<<
 *             numeric = False
 *         n += (<Vector>ptr).length
 */
    }

    /* "flitter/language/vm.pyx":411
 *         if (<Vector>ptr).objects is not None:
 *             numeric = False
 *         n += (<Vector>ptr).length             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         for i in range(base, base+m):
 */
    __pyx_v_n = (__pyx_v_n + ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->length);
  }

  /* "flitter/language/vm.pyx":412
 *             numeric = False
 *         n += (<Vector>ptr).length
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]
 */
  __pyx_t_1 = (__pyx_v_n == 0);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":413
 *         n += (<Vector>ptr).length
 *     if n == 0:
 *         for i in range(base, base+m):             # <<<<<<<<<<<<<<
 *             ptr = stack.vectors[i]
 *             Py_DECREF(<object>ptr)
 */
    __pyx_t_10 = (__pyx_v_base + __pyx_v_m);
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_12 = __pyx_v_base; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_i = __pyx_t_12;

      /* "flitter/language/vm.pyx":414
 *     if n == 0:
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]             # <<<<<<<<<<<<<<
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL
 */
      __pyx_v_ptr = (__pyx_v_stack->vectors[__pyx_v_i]);

      /* "flitter/language/vm.pyx":415
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]
 *             Py_DECREF(<object>ptr)             # <<<<<<<<<<<<<<
 *             stack.vectors[i] = NULL
 *         return null_
 */
      Py_DECREF(((PyObject *)__pyx_v_ptr));

      /* "flitter/language/vm.pyx":416
 *             ptr = stack.vectors[i]
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL             # <<<<<<<<<<<<<<
 *         return null_
 *     cdef Vector result = Vector.__new__(Vector)
 */
      (__pyx_v_stack->vectors[__pyx_v_i]) = NULL;
    }

    /* "flitter/language/vm.pyx":417
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL
 *         return null_             # <<<<<<<<<<<<<<
 *     cdef Vector result = Vector.__new__(Vector)
 *     if numeric:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_null_);
    __pyx_r = __pyx_v_7flitter_5model_null_;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":412
 *             numeric = False
 *         n += (<Vector>ptr).length
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]
 */
  }

  /* "flitter/language/vm.pyx":418
 *             stack.vectors[i] = NULL
 *         return null_
 *     cdef Vector result = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *     if numeric:
 *         result.allocate_numbers(n)
 */
  __pyx_t_2 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(0, 418, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":419
 *         return null_
 *     cdef Vector result = Vector.__new__(Vector)
 *     if numeric:             # <<<<<<<<<<<<<<
 *         result.allocate_numbers(n)
 *         for i in range(base, base+m):
 */
  if (__pyx_v_numeric) {

    /* "flitter/language/vm.pyx":420
 *     cdef Vector result = Vector.__new__(Vector)
 *     if numeric:
 *         result.allocate_numbers(n)             # <<<<<<<<<<<<<<
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]
 */
    __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_result->__pyx_vtab)->allocate_numbers(__pyx_v_result, __pyx_v_n); if (unlikely(__pyx_t_10 == ((int64_t)-1L))) __PYX_ERR(0, 420, __pyx_L1_error)

    /* "flitter/language/vm.pyx":421
 *     if numeric:
 *         result.allocate_numbers(n)
 *         for i in range(base, base+m):             # <<<<<<<<<<<<<<
 *             ptr = stack.vectors[i]
 *             for k in range((<Vector>ptr).length):
 */
    __pyx_t_10 = (__pyx_v_base + __pyx_v_m);
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_12 = __pyx_v_base; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_i = __pyx_t_12;

      /* "flitter/language/vm.pyx":422
 *         result.allocate_numbers(n)
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]             # <<<<<<<<<<<<<<
 *             for k in range((<Vector>ptr).length):
 *                 result.numbers[j] = (<Vector>ptr).numbers[k]
 */
      __pyx_v_ptr = (__pyx_v_stack->vectors[__pyx_v_i]);

      /* "flitter/language/vm.pyx":423
 *         for i in range(base, base+m):
 *             ptr = stack.vectors[i]
 *             for k in range((<Vector>ptr).length):             # <<<<<<<<<<<<<<
 *                 result.numbers[j] = (<Vector>ptr).numbers[k]
 *                 j += 1
 */
      __pyx_t_13 = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->length;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "flitter/language/vm.pyx":424
 *             ptr = stack.vectors[i]
 *             for k in range((<Vector>ptr).length):
 *                 result.numbers[j] = (<Vector>ptr).numbers[k]             # <<<<<<<<<<<<<<
 *                 j += 1
 *             Py_DECREF(<object>ptr)
 */
        (__pyx_v_result->numbers[__pyx_v_j]) = (((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->numbers[__pyx_v_k]);

        /* "flitter/language/vm.pyx":425
 *             for k in range((<Vector>ptr).length):
 *                 result.numbers[j] = (<Vector>ptr).numbers[k]
 *                 j += 1             # <<<<<<<<<<<<<<
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL
 */
        __pyx_v_j = (__pyx_v_j + 1);
      }

      /* "flitter/language/vm.pyx":426
 *                 result.numbers[j] = (<Vector>ptr).numbers[k]
 *                 j += 1
 *             Py_DECREF(<object>ptr)             # <<<<<<<<<<<<<<
 *             stack.vectors[i] = NULL
 *         return result
 */
      Py_DECREF(((PyObject *)__pyx_v_ptr));

      /* "flitter/language/vm.pyx":427
 *                 j += 1
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL             # <<<<<<<<<<<<<<
 *         return result
 *     cdef tuple vobjects, objects=PyTuple_New(n)
 */
      (__pyx_v_stack->vectors[__pyx_v_i]) = NULL;
    }

    /* "flitter/language/vm.pyx":428
 *             Py_DECREF(<object>ptr)
 *             stack.vectors[i] = NULL
 *         return result             # <<<<<<<<<<<<<<
 *     cdef tuple vobjects, objects=PyTuple_New(n)
 *     cdef PyObject* itemptr
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":419
 *         return null_
 *     cdef Vector result = Vector.__new__(Vector)
 *     if numeric:             # <<<<<<<<<<<<<<
 *         result.allocate_numbers(n)
 *         for i in range(base, base+m):
 */
  }

  /* "flitter/language/vm.pyx":429
 *             stack.vectors[i] = NULL
 *         return result
 *     cdef tuple vobjects, objects=PyTuple_New(n)             # <<<<<<<<<<<<<<
 *     cdef PyObject* itemptr
 *     cdef object obj
 */
  __pyx_t_2 = PyTuple_New(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_objects = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":432
 *     cdef PyObject* itemptr
 *     cdef object obj
 *     for i in range(base, base+m):             # <<<<<<<<<<<<<<
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is None:
 */
  __pyx_t_10 = (__pyx_v_base + __pyx_v_m);
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = __pyx_v_base; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "flitter/language/vm.pyx":433
 *     cdef object obj
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]             # <<<<<<<<<<<<<<
 *         if (<Vector>ptr).objects is None:
 *             for k in range((<Vector>ptr).length):
 */
    __pyx_v_ptr = (__pyx_v_stack->vectors[__pyx_v_i]);

    /* "flitter/language/vm.pyx":434
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is None:             # <<<<<<<<<<<<<<
 *             for k in range((<Vector>ptr).length):
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
 */
    __pyx_t_1 = (((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->objects == ((PyObject*)Py_None));
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":435
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is None:
 *             for k in range((<Vector>ptr).length):             # <<<<<<<<<<<<<<
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
 *                 Py_INCREF(obj)
 */
      __pyx_t_13 = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->length;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "flitter/language/vm.pyx":436
 *         if (<Vector>ptr).objects is None:
 *             for k in range((<Vector>ptr).length):
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])             # <<<<<<<<<<<<<<
 *                 Py_INCREF(obj)
 *                 PyTuple_SET_ITEM(objects, j, obj)
 */
        __pyx_t_2 = PyFloat_FromDouble((((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->numbers[__pyx_v_k])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "flitter/language/vm.pyx":437
 *             for k in range((<Vector>ptr).length):
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
 *                 Py_INCREF(obj)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(objects, j, obj)
 *                 j += 1
 */
        Py_INCREF(__pyx_v_obj);

        /* "flitter/language/vm.pyx":438
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
 *                 Py_INCREF(obj)
 *                 PyTuple_SET_ITEM(objects, j, obj)             # <<<<<<<<<<<<<<
 *                 j += 1
 *         else:
 */
        PyTuple_SET_ITEM(__pyx_v_objects, __pyx_v_j, __pyx_v_obj);

        /* "flitter/language/vm.pyx":439
 *                 Py_INCREF(obj)
 *                 PyTuple_SET_ITEM(objects, j, obj)
 *                 j += 1             # <<<<<<<<<<<<<<
 *         else:
 *             vobjects = (<Vector>ptr).objects
 */
        __pyx_v_j = (__pyx_v_j + 1);
      }

      /* "flitter/language/vm.pyx":434
 *     for i in range(base, base+m):
 *         ptr = stack.vectors[i]
 *         if (<Vector>ptr).objects is None:             # <<<<<<<<<<<<<<
 *             for k in range((<Vector>ptr).length):
 *                 obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
 */
      goto __pyx_L19;
    }

    /* "flitter/language/vm.pyx":441
 *                 j += 1
 *         else:
 *             vobjects = (<Vector>ptr).objects             # <<<<<<<<<<<<<<
 *             for k in range((<Vector>ptr).length):
 *                 itemptr = PyTuple_GET_ITEM(vobjects, k)
 */
    /*else*/ {
      __pyx_t_2 = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->objects;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_vobjects, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/language/vm.pyx":442
 *         else:
 *             vobjects = (<Vector>ptr).objects
 *             for k in range((<Vector>ptr).length):             # <<<<<<<<<<<<<<
 *                 itemptr = PyTuple_GET_ITEM(vobjects, k)
 *                 Py_INCREF(<object>itemptr)
 */
      __pyx_t_13 = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_ptr)->length;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_k = __pyx_t_15;

        /* "flitter/language/vm.pyx":443
 *             vobjects = (<Vector>ptr).objects
 *             for k in range((<Vector>ptr).length):
 *                 itemptr = PyTuple_GET_ITEM(vobjects, k)             # <<<<<<<<<<<<<<
 *                 Py_INCREF(<object>itemptr)
 *                 PyTuple_SET_ITEM(objects, j, <object>itemptr)
 */
        __pyx_v_itemptr = PyTuple_GET_ITEM(__pyx_v_vobjects, __pyx_v_k);

        /* "flitter/language/vm.pyx":444
 *             for k in range((<Vector>ptr).length):
 *                 itemptr = PyTuple_GET_ITEM(vobjects, k)
 *                 Py_INCREF(<object>itemptr)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(objects, j, <object>itemptr)
 *                 j += 1
 */
        Py_INCREF(((PyObject *)__pyx_v_itemptr));

        /* "flitter/language/vm.pyx":445
 *                 itemptr = PyTuple_GET_ITEM(vobjects, k)
 *                 Py_INCREF(<object>itemptr)
 *                 PyTuple_SET_ITEM(objects, j, <object>itemptr)             # <<<<<<<<<<<<<<
 *                 j += 1
 *         Py_DECREF(<object>ptr)
 */
        PyTuple_SET_ITEM(__pyx_v_objects, __pyx_v_j, ((PyObject *)__pyx_v_itemptr));

        /* "flitter/language/vm.pyx":446
 *                 Py_INCREF(<object>itemptr)
 *                 PyTuple_SET_ITEM(objects, j, <object>itemptr)
 *                 j += 1             # <<<<<<<<<<<<<<
 *         Py_DECREF(<object>ptr)
 *         stack.vectors[i] = NULL
 */
        __pyx_v_j = (__pyx_v_j + 1);
      }
    }
    __pyx_L19:;

    /* "flitter/language/vm.pyx":447
 *                 PyTuple_SET_ITEM(objects, j, <object>itemptr)
 *                 j += 1
 *         Py_DECREF(<object>ptr)             # <<<<<<<<<<<<<<
 *         stack.vectors[i] = NULL
 *     result.objects = objects
 */
    Py_DECREF(((PyObject *)__pyx_v_ptr));

    /* "flitter/language/vm.pyx":448
 *                 j += 1
 *         Py_DECREF(<object>ptr)
 *         stack.vectors[i] = NULL             # <<<<<<<<<<<<<<
 *     result.objects = objects
 *     result.length = n
 */
    (__pyx_v_stack->vectors[__pyx_v_i]) = NULL;
  }

  /* "flitter/language/vm.pyx":449
 *         Py_DECREF(<object>ptr)
 *         stack.vectors[i] = NULL
 *     result.objects = objects             # <<<<<<<<<<<<<<
 *     result.length = n
 *     return result
 */
  __Pyx_INCREF(__pyx_v_objects);
  __Pyx_GIVEREF(__pyx_v_objects);
  __Pyx_GOTREF(__pyx_v_result->objects);
  __Pyx_DECREF(__pyx_v_result->objects);
  __pyx_v_result->objects = __pyx_v_objects;

  /* "flitter/language/vm.pyx":450
 *         stack.vectors[i] = NULL
 *     result.objects = objects
 *     result.length = n             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->length = __pyx_v_n;

  /* "flitter/language/vm.pyx":451
 *     result.objects = objects
 *     result.length = n
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef inline Vector peek(VectorStack stack):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":396
 *     return t
 * 
 * cdef inline Vector pop_composed(VectorStack stack, int64_t m):             # <<<<<<<<<<<<<<
 *     if stack.top - m < -1:
 *         raise StackUnderflow(f"Cannot pop {m} items from {stack.top+1} depth stack")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.pop_composed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_vobjects);
  __Pyx_XDECREF(__pyx_v_objects);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":453
 *     return result
 * 
 * cdef inline Vector peek(VectorStack stack):             # <<<<<<<<<<<<<<
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 */

static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_peek(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek", 1);

  /* "flitter/language/vm.pyx":454
 * 
 * cdef inline Vector peek(VectorStack stack):
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     return <Vector>stack.vectors[stack.top]
 */
  __pyx_t_1 = (__pyx_v_stack->top == -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":455
 * cdef inline Vector peek(VectorStack stack):
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")             # <<<<<<<<<<<<<<
 *     return <Vector>stack.vectors[stack.top]
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_Stack_empty};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 455, __pyx_L1_error)

    /* "flitter/language/vm.pyx":454
 * 
 * cdef inline Vector peek(VectorStack stack):
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     return <Vector>stack.vectors[stack.top]
 */
  }

  /* "flitter/language/vm.pyx":456
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 *     return <Vector>stack.vectors[stack.top]             # <<<<<<<<<<<<<<
 * 
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_5model_Vector *)(__pyx_v_stack->vectors[__pyx_v_stack->top])));
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)(__pyx_v_stack->vectors[__pyx_v_stack->top]));
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":453
 *     return result
 * 
 * cdef inline Vector peek(VectorStack stack):             # <<<<<<<<<<<<<<
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":458
 *     return <Vector>stack.vectors[stack.top]
 * 
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):             # <<<<<<<<<<<<<<
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")
 */

static CYTHON_INLINE struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_2vm_peek_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_offset) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek_at", 1);

  /* "flitter/language/vm.pyx":459
 * 
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):
 *     if stack.top - offset < 0:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")
 *     return <Vector>stack.vectors[stack.top-offset]
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_offset) < 0);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":460
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")             # <<<<<<<<<<<<<<
 *     return <Vector>stack.vectors[stack.top-offset]
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_peek);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_peek);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_peek);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_below_top_of);
    __pyx_t_5 += 20;
    __Pyx_GIVEREF(__pyx_kp_u_items_below_top_of);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_below_top_of);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_item_stack);
    __pyx_t_5 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_item_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_item_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 460, __pyx_L1_error)

    /* "flitter/language/vm.pyx":459
 * 
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):
 *     if stack.top - offset < 0:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")
 *     return <Vector>stack.vectors[stack.top-offset]
 */
  }

  /* "flitter/language/vm.pyx":461
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")
 *     return <Vector>stack.vectors[stack.top-offset]             # <<<<<<<<<<<<<<
 * 
 * cdef inline void poke(VectorStack stack, Vector vector):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_10 = (__pyx_v_stack->vectors[(__pyx_v_stack->top - __pyx_v_offset)]);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10);
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":458
 *     return <Vector>stack.vectors[stack.top]
 * 
 * cdef inline Vector peek_at(VectorStack stack, int64_t offset):             # <<<<<<<<<<<<<<
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot peek {offset} items below top of {stack.top+1} item stack")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.peek_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":463
 *     return <Vector>stack.vectors[stack.top-offset]
 * 
 * cdef inline void poke(VectorStack stack, Vector vector):             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top == -1:
 */

static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_poke(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke", 1);

  /* "flitter/language/vm.pyx":464
 * 
 * cdef inline void poke(VectorStack stack, Vector vector):
 *     assert vector is not None             # <<<<<<<<<<<<<<
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((PyObject *)__pyx_v_vector) != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 464, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 464, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":465
 * cdef inline void poke(VectorStack stack, Vector vector):
 *     assert vector is not None
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     Py_DECREF(<Vector>stack.vectors[stack.top])
 */
  __pyx_t_1 = (__pyx_v_stack->top == -1L);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":466
 *     assert vector is not None
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")             # <<<<<<<<<<<<<<
 *     Py_DECREF(<Vector>stack.vectors[stack.top])
 *     Py_INCREF(vector)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_Stack_empty};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 466, __pyx_L1_error)

    /* "flitter/language/vm.pyx":465
 * cdef inline void poke(VectorStack stack, Vector vector):
 *     assert vector is not None
 *     if stack.top == -1:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow("Stack empty")
 *     Py_DECREF(<Vector>stack.vectors[stack.top])
 */
  }

  /* "flitter/language/vm.pyx":467
 *     if stack.top == -1:
 *         raise StackUnderflow("Stack empty")
 *     Py_DECREF(<Vector>stack.vectors[stack.top])             # <<<<<<<<<<<<<<
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top] = <PyObject*>vector
 */
  __pyx_t_2 = ((PyObject *)(__pyx_v_stack->vectors[__pyx_v_stack->top]));
  __Pyx_INCREF(__pyx_t_2);
  Py_DECREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":468
 *         raise StackUnderflow("Stack empty")
 *     Py_DECREF(<Vector>stack.vectors[stack.top])
 *     Py_INCREF(vector)             # <<<<<<<<<<<<<<
 *     stack.vectors[stack.top] = <PyObject*>vector
 * 
 */
  Py_INCREF(((PyObject *)__pyx_v_vector));

  /* "flitter/language/vm.pyx":469
 *     Py_DECREF(<Vector>stack.vectors[stack.top])
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top] = <PyObject*>vector             # <<<<<<<<<<<<<<
 * 
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):
 */
  (__pyx_v_stack->vectors[__pyx_v_stack->top]) = ((PyObject *)__pyx_v_vector);

  /* "flitter/language/vm.pyx":463
 *     return <Vector>stack.vectors[stack.top-offset]
 * 
 * cdef inline void poke(VectorStack stack, Vector vector):             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top == -1:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.poke", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":471
 *     stack.vectors[stack.top] = <PyObject*>vector
 * 
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top - offset < 0:
 */

static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_poke_at(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_offset, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poke_at", 1);

  /* "flitter/language/vm.pyx":472
 * 
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):
 *     assert vector is not None             # <<<<<<<<<<<<<<
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((PyObject *)__pyx_v_vector) != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 472, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 472, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":473
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):
 *     assert vector is not None
 *     if stack.top - offset < 0:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])
 */
  __pyx_t_1 = ((__pyx_v_stack->top - __pyx_v_offset) < 0);
  if (unlikely(__pyx_t_1)) {

    /* "flitter/language/vm.pyx":474
 *     assert vector is not None
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")             # <<<<<<<<<<<<<<
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])
 *     Py_INCREF(vector)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_StackUnderflow); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_poke);
    __pyx_t_5 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_poke);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Cannot_poke);
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_below_top_of);
    __pyx_t_5 += 20;
    __Pyx_GIVEREF(__pyx_kp_u_items_below_top_of);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_items_below_top_of);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_stack->top + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_item_stack);
    __pyx_t_5 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_item_stack);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_item_stack);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 474, __pyx_L1_error)

    /* "flitter/language/vm.pyx":473
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):
 *     assert vector is not None
 *     if stack.top - offset < 0:             # <<<<<<<<<<<<<<
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])
 */
  }

  /* "flitter/language/vm.pyx":475
 *     if stack.top - offset < 0:
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])             # <<<<<<<<<<<<<<
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top-offset] = <PyObject*>vector
 */
  __pyx_t_10 = (__pyx_v_stack->vectors[(__pyx_v_stack->top - __pyx_v_offset)]);
  Py_DECREF(((PyObject *)__pyx_t_10));

  /* "flitter/language/vm.pyx":476
 *         raise StackUnderflow(f"Cannot poke {offset} items below top of {stack.top+1} item stack")
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])
 *     Py_INCREF(vector)             # <<<<<<<<<<<<<<
 *     stack.vectors[stack.top-offset] = <PyObject*>vector
 * 
 */
  Py_INCREF(((PyObject *)__pyx_v_vector));

  /* "flitter/language/vm.pyx":477
 *     Py_DECREF(<Vector>stack.vectors[stack.top-offset])
 *     Py_INCREF(vector)
 *     stack.vectors[stack.top-offset] = <PyObject*>vector             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_stack->vectors[(__pyx_v_stack->top - __pyx_v_offset)]) = ((PyObject *)__pyx_v_vector);

  /* "flitter/language/vm.pyx":471
 *     stack.vectors[stack.top] = <PyObject*>vector
 * 
 * cdef inline void poke_at(VectorStack stack, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *     assert vector is not None
 *     if stack.top - offset < 0:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.poke_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":492
 *     cdef readonly int64_t hash
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self.hash == 0:
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_7flitter_8language_2vm_8Function_1__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_7flitter_8language_2vm_8Function_1__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function___hash__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_7flitter_8language_2vm_8Function___hash__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_hash_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 1);

  /* "flitter/language/vm.pyx":493
 * 
 *     def __hash__(self):
 *         if self.hash == 0:             # <<<<<<<<<<<<<<
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)
 *         return self.hash
 */
  __pyx_t_1 = (__pyx_v_self->hash == 0);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":494
 *     def __hash__(self):
 *         if self.hash == 0:
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)             # <<<<<<<<<<<<<<
 *         return self.hash
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self->program)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->address); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Xor(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_v_self->defaults;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = PyObject_Hash(__pyx_t_3); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_FromHash_t(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Xor(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_v_self->captures;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = PyObject_Hash(__pyx_t_3); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_FromHash_t(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Xor(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_self->hash = __pyx_t_6;

    /* "flitter/language/vm.pyx":493
 * 
 *     def __hash__(self):
 *         if self.hash == 0:             # <<<<<<<<<<<<<<
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)
 *         return self.hash
 */
  }

  /* "flitter/language/vm.pyx":495
 *         if self.hash == 0:
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)
 *         return self.hash             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, Context context, *args, **kwargs):
 */
  __pyx_r = __pyx_v_self->hash;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":492
 *     cdef readonly int64_t hash
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self.hash == 0:
 *             self.hash = id(self.program) ^ self.address ^ hash(self.defaults) ^ hash(self.captures)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Function.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":497
 *         return self.hash
 * 
 *     def __call__(self, Context context, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.call_depth == MAX_CALL_DEPTH:
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_context = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __pyx_v_args = __Pyx_ArgsSlice_VARARGS(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, used_pos_args, "__call__") < 0)) __PYX_ERR(0, 497, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_context = ((struct __pyx_obj_7flitter_5model_Context *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 1, 1, __pyx_nargs); __PYX_ERR(0, 497, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("flitter.language.vm.Function.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_7flitter_5model_Context, 1, "context", 0))) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_2__call__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self), __pyx_v_context, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_2__call__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  int64_t __pyx_v_i;
  int64_t __pyx_v_lnames_top;
  int64_t __pyx_v_stack_top;
  int64_t __pyx_v_k;
  int64_t __pyx_v_m;
  int64_t __pyx_v_n;
  PyObject *__pyx_v_objptr;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_lnames = 0;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int64_t __pyx_t_8;
  int64_t __pyx_t_9;
  int64_t __pyx_t_10;
  PyObject *__pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 1);

  /* "flitter/language/vm.pyx":498
 * 
 *     def __call__(self, Context context, *args, **kwargs):
 *         if self.call_depth == MAX_CALL_DEPTH:             # <<<<<<<<<<<<<<
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 *             return null_
 */
  __pyx_t_1 = (__pyx_v_self->call_depth == __pyx_v_7flitter_8language_2vm_MAX_CALL_DEPTH);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":499
 *     def __call__(self, Context context, *args, **kwargs):
 *         if self.call_depth == MAX_CALL_DEPTH:
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")             # <<<<<<<<<<<<<<
 *             return null_
 *         cdef int64_t i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)
 */
    __pyx_t_2 = __pyx_v_context->errors;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Recursion_depth_exceeded_for_fun);
    __pyx_t_4 += 35;
    __Pyx_GIVEREF(__pyx_kp_u_Recursion_depth_exceeded_for_fun);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Recursion_depth_exceeded_for_fun);
    __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_self->__name__); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__9);
    __pyx_t_4 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__9);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__9);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = PySet_Add(__pyx_t_2, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flitter/language/vm.pyx":500
 *         if self.call_depth == MAX_CALL_DEPTH:
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 *             return null_             # <<<<<<<<<<<<<<
 *         cdef int64_t i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)
 *         cdef PyObject* objptr
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_null_);
    __pyx_r = ((PyObject *)__pyx_v_7flitter_5model_null_);
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":498
 * 
 *     def __call__(self, Context context, *args, **kwargs):
 *         if self.call_depth == MAX_CALL_DEPTH:             # <<<<<<<<<<<<<<
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 *             return null_
 */
  }

  /* "flitter/language/vm.pyx":501
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 *             return null_
 *         cdef int64_t i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)             # <<<<<<<<<<<<<<
 *         cdef PyObject* objptr
 *         if context.state is None:
 */
  __pyx_t_6 = __pyx_v_self->captures;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_v_k = PyTuple_GET_SIZE(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_v_self->parameters;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_v_m = PyTuple_GET_SIZE(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_n = PyTuple_GET_SIZE(__pyx_v_args);

  /* "flitter/language/vm.pyx":503
 *         cdef int64_t i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)
 *         cdef PyObject* objptr
 *         if context.state is None:             # <<<<<<<<<<<<<<
 *             context.state = StateDict()
 *         cdef VectorStack lnames = context.lnames
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_context->state) == Py_None);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":504
 *         cdef PyObject* objptr
 *         if context.state is None:
 *             context.state = StateDict()             # <<<<<<<<<<<<<<
 *         cdef VectorStack lnames = context.lnames
 *         if lnames is None:
 */
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_5model_StateDict)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF((PyObject *)__pyx_v_context->state);
    __Pyx_DECREF((PyObject *)__pyx_v_context->state);
    __pyx_v_context->state = ((struct __pyx_obj_7flitter_5model_StateDict *)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "flitter/language/vm.pyx":503
 *         cdef int64_t i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)
 *         cdef PyObject* objptr
 *         if context.state is None:             # <<<<<<<<<<<<<<
 *             context.state = StateDict()
 *         cdef VectorStack lnames = context.lnames
 */
  }

  /* "flitter/language/vm.pyx":505
 *         if context.state is None:
 *             context.state = StateDict()
 *         cdef VectorStack lnames = context.lnames             # <<<<<<<<<<<<<<
 *         if lnames is None:
 *             lnames = context.lnames = VectorStack()
 */
  if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 505, __pyx_L1_error)
  __pyx_t_6 = __pyx_v_context->lnames;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_v_lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":506
 *             context.state = StateDict()
 *         cdef VectorStack lnames = context.lnames
 *         if lnames is None:             # <<<<<<<<<<<<<<
 *             lnames = context.lnames = VectorStack()
 *         lnames_top = lnames.top
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_lnames) == Py_None);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":507
 *         cdef VectorStack lnames = context.lnames
 *         if lnames is None:
 *             lnames = context.lnames = VectorStack()             # <<<<<<<<<<<<<<
 *         lnames_top = lnames.top
 *         cdef VectorStack stack = context.stack
 */
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_lnames, ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6));
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_context->lnames);
    __Pyx_DECREF(__pyx_v_context->lnames);
    __pyx_v_context->lnames = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flitter/language/vm.pyx":506
 *             context.state = StateDict()
 *         cdef VectorStack lnames = context.lnames
 *         if lnames is None:             # <<<<<<<<<<<<<<
 *             lnames = context.lnames = VectorStack()
 *         lnames_top = lnames.top
 */
  }

  /* "flitter/language/vm.pyx":508
 *         if lnames is None:
 *             lnames = context.lnames = VectorStack()
 *         lnames_top = lnames.top             # <<<<<<<<<<<<<<
 *         cdef VectorStack stack = context.stack
 *         if stack is None:
 */
  __pyx_t_8 = __pyx_v_lnames->top;
  __pyx_v_lnames_top = __pyx_t_8;

  /* "flitter/language/vm.pyx":509
 *             lnames = context.lnames = VectorStack()
 *         lnames_top = lnames.top
 *         cdef VectorStack stack = context.stack             # <<<<<<<<<<<<<<
 *         if stack is None:
 *             stack = context.stack = VectorStack()
 */
  if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 509, __pyx_L1_error)
  __pyx_t_6 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_v_stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":510
 *         lnames_top = lnames.top
 *         cdef VectorStack stack = context.stack
 *         if stack is None:             # <<<<<<<<<<<<<<
 *             stack = context.stack = VectorStack()
 *         stack_top = stack.top
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_stack) == Py_None);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":511
 *         cdef VectorStack stack = context.stack
 *         if stack is None:
 *             stack = context.stack = VectorStack()             # <<<<<<<<<<<<<<
 *         stack_top = stack.top
 *         for i in range(k):
 */
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_stack, ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6));
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_v_context->stack);
    __Pyx_DECREF(__pyx_v_context->stack);
    __pyx_v_context->stack = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flitter/language/vm.pyx":510
 *         lnames_top = lnames.top
 *         cdef VectorStack stack = context.stack
 *         if stack is None:             # <<<<<<<<<<<<<<
 *             stack = context.stack = VectorStack()
 *         stack_top = stack.top
 */
  }

  /* "flitter/language/vm.pyx":512
 *         if stack is None:
 *             stack = context.stack = VectorStack()
 *         stack_top = stack.top             # <<<<<<<<<<<<<<
 *         for i in range(k):
 *             push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))
 */
  __pyx_t_8 = __pyx_v_stack->top;
  __pyx_v_stack_top = __pyx_t_8;

  /* "flitter/language/vm.pyx":513
 *             stack = context.stack = VectorStack()
 *         stack_top = stack.top
 *         for i in range(k):             # <<<<<<<<<<<<<<
 *             push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))
 *         push(lnames, self.vself)
 */
  __pyx_t_8 = __pyx_v_k;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "flitter/language/vm.pyx":514
 *         stack_top = stack.top
 *         for i in range(k):
 *             push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))             # <<<<<<<<<<<<<<
 *         push(lnames, self.vself)
 *         for i in range(m):
 */
    __pyx_t_6 = __pyx_v_self->captures;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_v_i);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_12 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_11)); if (unlikely(__pyx_t_12 == ((int64_t)0))) __PYX_ERR(0, 514, __pyx_L1_error)
  }

  /* "flitter/language/vm.pyx":515
 *         for i in range(k):
 *             push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))
 *         push(lnames, self.vself)             # <<<<<<<<<<<<<<
 *         for i in range(m):
 *             if i < n:
 */
  __pyx_t_6 = ((PyObject *)__pyx_v_self->vself);
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":516
 *             push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))
 *         push(lnames, self.vself)
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if i < n:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 */
  __pyx_t_8 = __pyx_v_m;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "flitter/language/vm.pyx":517
 *         push(lnames, self.vself)
 *         for i in range(m):
 *             if i < n:             # <<<<<<<<<<<<<<
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:
 */
    __pyx_t_1 = (__pyx_v_i < __pyx_v_n);
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":518
 *         for i in range(m):
 *             if i < n:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))             # <<<<<<<<<<<<<<
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:
 *                 push(lnames, <Vector>objptr)
 */
      __pyx_t_11 = PyTuple_GET_ITEM(__pyx_v_args, __pyx_v_i);
      __pyx_t_12 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_11)); if (unlikely(__pyx_t_12 == ((int64_t)0))) __PYX_ERR(0, 518, __pyx_L1_error)

      /* "flitter/language/vm.pyx":517
 *         push(lnames, self.vself)
 *         for i in range(m):
 *             if i < n:             # <<<<<<<<<<<<<<
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:
 */
      goto __pyx_L11;
    }

    /* "flitter/language/vm.pyx":519
 *             if i < n:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:             # <<<<<<<<<<<<<<
 *                 push(lnames, <Vector>objptr)
 *             else:
 */
    __pyx_t_13 = (__pyx_v_kwargs != ((PyObject*)Py_None));
    if (__pyx_t_13) {
    } else {
      __pyx_t_1 = __pyx_t_13;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_6 = __pyx_v_self->parameters;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_v_i);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_14 = PyDict_GetItem(__pyx_v_kwargs, ((PyObject *)__pyx_t_11));
    __pyx_v_objptr = __pyx_t_14;
    __pyx_t_13 = (__pyx_t_14 != NULL);
    __pyx_t_1 = __pyx_t_13;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":520
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:
 *                 push(lnames, <Vector>objptr)             # <<<<<<<<<<<<<<
 *             else:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(self.defaults, i))
 */
      __pyx_t_12 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_objptr)); if (unlikely(__pyx_t_12 == ((int64_t)0))) __PYX_ERR(0, 520, __pyx_L1_error)

      /* "flitter/language/vm.pyx":519
 *             if i < n:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
 *             elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:             # <<<<<<<<<<<<<<
 *                 push(lnames, <Vector>objptr)
 *             else:
 */
      goto __pyx_L11;
    }

    /* "flitter/language/vm.pyx":522
 *                 push(lnames, <Vector>objptr)
 *             else:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(self.defaults, i))             # <<<<<<<<<<<<<<
 *         self.call_depth += 1
 *         try:
 */
    /*else*/ {
      __pyx_t_6 = __pyx_v_self->defaults;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_v_i);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_12 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_14)); if (unlikely(__pyx_t_12 == ((int64_t)0))) __PYX_ERR(0, 522, __pyx_L1_error)
    }
    __pyx_L11:;
  }

  /* "flitter/language/vm.pyx":523
 *             else:
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(self.defaults, i))
 *         self.call_depth += 1             # <<<<<<<<<<<<<<
 *         try:
 *             self.program._execute(context, self.address, self.record_stats)
 */
  __pyx_v_self->call_depth = (__pyx_v_self->call_depth + 1);

  /* "flitter/language/vm.pyx":524
 *                 push(lnames, <Vector>PyTuple_GET_ITEM(self.defaults, i))
 *         self.call_depth += 1
 *         try:             # <<<<<<<<<<<<<<
 *             self.program._execute(context, self.address, self.record_stats)
 *         finally:
 */
  /*try:*/ {

    /* "flitter/language/vm.pyx":525
 *         self.call_depth += 1
 *         try:
 *             self.program._execute(context, self.address, self.record_stats)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.call_depth -= 1
 */
    ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_self->program->__pyx_vtab)->_execute(__pyx_v_self->program, __pyx_v_context, __pyx_v_self->address, __pyx_v_self->record_stats); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L15_error)
  }

  /* "flitter/language/vm.pyx":527
 *             self.program._execute(context, self.address, self.record_stats)
 *         finally:
 *             self.call_depth -= 1             # <<<<<<<<<<<<<<
 *         drop(lnames, k + 1 + m)
 *         cdef Vector result = pop(stack)
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_v_self->call_depth = (__pyx_v_self->call_depth - 1);
      goto __pyx_L16;
    }
    __pyx_L15_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __pyx_t_7 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        __pyx_v_self->call_depth = (__pyx_v_self->call_depth - 1);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
      }
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
    }
    __pyx_L16:;
  }

  /* "flitter/language/vm.pyx":528
 *         finally:
 *             self.call_depth -= 1
 *         drop(lnames, k + 1 + m)             # <<<<<<<<<<<<<<
 *         cdef Vector result = pop(stack)
 *         assert stack.top == stack_top, "Bad function return stack"
 */
  __pyx_f_7flitter_8language_2vm_drop(__pyx_v_lnames, ((__pyx_v_k + 1) + __pyx_v_m)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)

  /* "flitter/language/vm.pyx":529
 *             self.call_depth -= 1
 *         drop(lnames, k + 1 + m)
 *         cdef Vector result = pop(stack)             # <<<<<<<<<<<<<<
 *         assert stack.top == stack_top, "Bad function return stack"
 *         assert lnames.top == lnames_top, "Bad function return lnames"
 */
  __pyx_t_6 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_result = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":530
 *         drop(lnames, k + 1 + m)
 *         cdef Vector result = pop(stack)
 *         assert stack.top == stack_top, "Bad function return stack"             # <<<<<<<<<<<<<<
 *         assert lnames.top == lnames_top, "Bad function return lnames"
 *         return result
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_stack->top == __pyx_v_stack_top);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_function_return_stack, 0, 0);
      __PYX_ERR(0, 530, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 530, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":531
 *         cdef Vector result = pop(stack)
 *         assert stack.top == stack_top, "Bad function return stack"
 *         assert lnames.top == lnames_top, "Bad function return lnames"             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_lnames->top == __pyx_v_lnames_top);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_function_return_lnames, 0, 0);
      __PYX_ERR(0, 531, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 531, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":532
 *         assert stack.top == stack_top, "Bad function return stack"
 *         assert lnames.top == lnames_top, "Bad function return lnames"
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":497
 *         return self.hash
 * 
 *     def __call__(self, Context context, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.call_depth == MAX_CALL_DEPTH:
 *             PySet_Add(context.errors, f"Recursion depth exceeded for func '{self.__name__}'")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.vm.Function.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_lnames);
  __Pyx_XDECREF((PyObject *)__pyx_v_stack);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":481
 * 
 * cdef class Function:
 *     cdef readonly str __name__             # <<<<<<<<<<<<<<
 *     cdef readonly Vector vself
 *     cdef readonly tuple parameters
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8__name___1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8__name___1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_8__name_____get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8__name_____get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__name__);
  __pyx_r = __pyx_v_self->__name__;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":482
 * cdef class Function:
 *     cdef readonly str __name__
 *     cdef readonly Vector vself             # <<<<<<<<<<<<<<
 *     cdef readonly tuple parameters
 *     cdef readonly tuple defaults
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_5vself_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_5vself_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_5vself___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_5vself___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->vself);
  __pyx_r = ((PyObject *)__pyx_v_self->vself);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":483
 *     cdef readonly str __name__
 *     cdef readonly Vector vself
 *     cdef readonly tuple parameters             # <<<<<<<<<<<<<<
 *     cdef readonly tuple defaults
 *     cdef readonly Program program
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_10parameters_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_10parameters_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_10parameters___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_10parameters___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __pyx_r = __pyx_v_self->parameters;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":484
 *     cdef readonly Vector vself
 *     cdef readonly tuple parameters
 *     cdef readonly tuple defaults             # <<<<<<<<<<<<<<
 *     cdef readonly Program program
 *     cdef readonly int64_t address
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8defaults_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8defaults_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_8defaults___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8defaults___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->defaults);
  __pyx_r = __pyx_v_self->defaults;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":485
 *     cdef readonly tuple parameters
 *     cdef readonly tuple defaults
 *     cdef readonly Program program             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t address
 *     cdef readonly bint record_stats
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7program_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7program_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_7program___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_7program___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->program);
  __pyx_r = ((PyObject *)__pyx_v_self->program);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":486
 *     cdef readonly tuple defaults
 *     cdef readonly Program program
 *     cdef readonly int64_t address             # <<<<<<<<<<<<<<
 *     cdef readonly bint record_stats
 *     cdef readonly tuple captures
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7address_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7address_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_7address___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_7address___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->address); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Function.address.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":487
 *     cdef readonly Program program
 *     cdef readonly int64_t address
 *     cdef readonly bint record_stats             # <<<<<<<<<<<<<<
 *     cdef readonly tuple captures
 *     cdef readonly int64_t call_depth
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_12record_stats_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_12record_stats_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_12record_stats___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_12record_stats___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_stats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Function.record_stats.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":488
 *     cdef readonly int64_t address
 *     cdef readonly bint record_stats
 *     cdef readonly tuple captures             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t call_depth
 *     cdef readonly int64_t hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8captures_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_8captures_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_8captures___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_8captures___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->captures);
  __pyx_r = __pyx_v_self->captures;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":489
 *     cdef readonly bint record_stats
 *     cdef readonly tuple captures
 *     cdef readonly int64_t call_depth             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t hash
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_10call_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_10call_depth_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_10call_depth___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_10call_depth___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->call_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Function.call_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":490
 *     cdef readonly tuple captures
 *     cdef readonly int64_t call_depth
 *     cdef readonly int64_t hash             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_4hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_4hash_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_4hash___get__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_4hash___get__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Function.hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_8Function_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_8Function_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_4__reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.__name__, self.address, self.call_depth, self.captures, self.defaults, self.hash, self.parameters, self.program, self.record_stats, self.vself)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->address); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->call_depth); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->hash); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->record_stats); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(10); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->__name__);
  __Pyx_GIVEREF(__pyx_v_self->__name__);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->__name__)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->captures);
  __Pyx_GIVEREF(__pyx_v_self->captures);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_self->captures)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->defaults);
  __Pyx_GIVEREF(__pyx_v_self->defaults);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->defaults)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __Pyx_GIVEREF(__pyx_v_self->parameters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_v_self->parameters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->program);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->program);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 7, ((PyObject *)__pyx_v_self->program))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->vself);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->vself);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 9, ((PyObject *)__pyx_v_self->vself))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.__name__, self.address, self.call_depth, self.captures, self.defaults, self.hash, self.parameters, self.program, self.record_stats, self.vself)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.__name__, self.address, self.call_depth, self.captures, self.defaults, self.hash, self.parameters, self.program, self.record_stats, self.vself)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.__name__ is not None or self.captures is not None or self.defaults is not None or self.parameters is not None or self.program is not None or self.vself is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.__name__, self.address, self.call_depth, self.captures, self.defaults, self.hash, self.parameters, self.program, self.record_stats, self.vself)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.__name__ is not None or self.captures is not None or self.defaults is not None or self.parameters is not None or self.program is not None or self.vself is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->__name__ != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->captures != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->defaults != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->parameters != ((PyObject*)Py_None));
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (((PyObject *)__pyx_v_self->program) != Py_None);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (((PyObject *)__pyx_v_self->vself) != Py_None);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.__name__ is not None or self.captures is not None or self.defaults is not None or self.parameters is not None or self.program is not None or self.vself is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.__name__ is not None or self.captures is not None or self.defaults is not None or self.parameters is not None or self.program is not None or self.vself is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Function); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_10904070);
    __Pyx_GIVEREF(__pyx_int_10904070);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_10904070)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.__name__ is not None or self.captures is not None or self.defaults is not None or self.parameters is not None or self.program is not None or self.vself is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, None), state
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Function); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_10904070);
    __Pyx_GIVEREF(__pyx_int_10904070);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_10904070)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Function.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_8Function_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_8Function_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_8Function_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Function.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8Function_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8Function_6__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_Function__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Function.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_10LoopSource_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_10LoopSource_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_10LoopSource_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_10LoopSource_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_10LoopSource___reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_10LoopSource___reduce_cython__(struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.iterations, self.position, self.source, self.stride)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->iterations); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->position); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->stride); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->source);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->source);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_self->source))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.iterations, self.position, self.source, self.stride)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.iterations, self.position, self.source, self.stride)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.source is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.iterations, self.position, self.source, self.stride)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.source is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, None), state
 */
  /*else*/ {
    __pyx_t_5 = (((PyObject *)__pyx_v_self->source) != Py_None);
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.source is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_LoopSource); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_45504841);
    __Pyx_GIVEREF(__pyx_int_45504841);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_45504841)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, None), state
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_LoopSource); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_45504841);
    __Pyx_GIVEREF(__pyx_int_45504841);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_45504841)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.LoopSource.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_10LoopSource_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_10LoopSource_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_10LoopSource_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_10LoopSource_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.LoopSource.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_10LoopSource_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_10LoopSource_2__setstate_cython__(struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_LoopSource__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.LoopSource.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":542
 * 
 * 
 * def log_vm_stats():             # <<<<<<<<<<<<<<
 *     cdef list stats = []
 *     cdef double duration, total=0
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_1log_vm_stats(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_1log_vm_stats = {"log_vm_stats", (PyCFunction)__pyx_pw_7flitter_8language_2vm_1log_vm_stats, METH_NOARGS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_1log_vm_stats(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("log_vm_stats (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_log_vm_stats(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_log_vm_stats(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_stats = 0;
  double __pyx_v_duration;
  double __pyx_v_total;
  int64_t __pyx_v_count;
  int64_t __pyx_v_i;
  double __pyx_v_start;
  double __pyx_v_end;
  double __pyx_v_overhead;
  double __pyx_v_per_execution;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  int __pyx_t_4;
  double __pyx_t_5;
  long __pyx_t_6;
  double __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_11;
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  unsigned int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("log_vm_stats", 1);

  /* "flitter/language/vm.pyx":543
 * 
 * def log_vm_stats():
 *     cdef list stats = []             # <<<<<<<<<<<<<<
 *     cdef double duration, total=0
 *     cdef int64_t count
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_stats = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":544
 * def log_vm_stats():
 *     cdef list stats = []
 *     cdef double duration, total=0             # <<<<<<<<<<<<<<
 *     cdef int64_t count
 *     cdef int64_t i
 */
  __pyx_v_total = 0.0;

  /* "flitter/language/vm.pyx":548
 *     cdef int64_t i
 *     cdef double start, end, overhead, per_execution
 *     start = perf_counter()             # <<<<<<<<<<<<<<
 *     for count in range(10000):
 *         end = perf_counter()
 */
  __pyx_v_start = perf_counter();

  /* "flitter/language/vm.pyx":549
 *     cdef double start, end, overhead, per_execution
 *     start = perf_counter()
 *     for count in range(10000):             # <<<<<<<<<<<<<<
 *         end = perf_counter()
 *         StatsCount[<int64_t>OpCode.MAX] += 1
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 0x2710; __pyx_t_2+=1) {
    __pyx_v_count = __pyx_t_2;

    /* "flitter/language/vm.pyx":550
 *     start = perf_counter()
 *     for count in range(10000):
 *         end = perf_counter()             # <<<<<<<<<<<<<<
 *         StatsCount[<int64_t>OpCode.MAX] += 1
 *         StatsDuration[<int64_t>OpCode.MAX] += end
 */
    __pyx_v_end = perf_counter();

    /* "flitter/language/vm.pyx":551
 *     for count in range(10000):
 *         end = perf_counter()
 *         StatsCount[<int64_t>OpCode.MAX] += 1             # <<<<<<<<<<<<<<
 *         StatsDuration[<int64_t>OpCode.MAX] += end
 *     overhead = (end - start) / 10000
 */
    __pyx_t_3 = ((int64_t)__pyx_e_7flitter_8language_2vm_MAX);
    (__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_t_3]) = ((__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_t_3]) + 1);

    /* "flitter/language/vm.pyx":552
 *         end = perf_counter()
 *         StatsCount[<int64_t>OpCode.MAX] += 1
 *         StatsDuration[<int64_t>OpCode.MAX] += end             # <<<<<<<<<<<<<<
 *     overhead = (end - start) / 10000
 *     if CallOutCount:
 */
    __pyx_t_3 = ((int64_t)__pyx_e_7flitter_8language_2vm_MAX);
    (__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_t_3]) = ((__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_t_3]) + __pyx_v_end);
  }

  /* "flitter/language/vm.pyx":553
 *         StatsCount[<int64_t>OpCode.MAX] += 1
 *         StatsDuration[<int64_t>OpCode.MAX] += end
 *     overhead = (end - start) / 10000             # <<<<<<<<<<<<<<
 *     if CallOutCount:
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)
 */
  __pyx_v_overhead = ((__pyx_v_end - __pyx_v_start) / 10000.0);

  /* "flitter/language/vm.pyx":554
 *         StatsDuration[<int64_t>OpCode.MAX] += end
 *     overhead = (end - start) / 10000
 *     if CallOutCount:             # <<<<<<<<<<<<<<
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)
 *         stats.append((duration, CallOutCount, '(native funcs)'))
 */
  __pyx_t_4 = (__pyx_v_7flitter_8language_2vm_CallOutCount != 0);
  if (__pyx_t_4) {

    /* "flitter/language/vm.pyx":555
 *     overhead = (end - start) / 10000
 *     if CallOutCount:
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)             # <<<<<<<<<<<<<<
 *         stats.append((duration, CallOutCount, '(native funcs)'))
 *         total += duration
 */
    __pyx_t_5 = (__pyx_v_7flitter_8language_2vm_CallOutDuration - (__pyx_v_7flitter_8language_2vm_CallOutCount * __pyx_v_overhead));
    __pyx_t_6 = 0;
    __pyx_t_4 = (__pyx_t_5 > __pyx_t_6);
    if (__pyx_t_4) {
      __pyx_t_7 = __pyx_t_5;
    } else {
      __pyx_t_7 = __pyx_t_6;
    }
    __pyx_v_duration = __pyx_t_7;

    /* "flitter/language/vm.pyx":556
 *     if CallOutCount:
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)
 *         stats.append((duration, CallOutCount, '(native funcs)'))             # <<<<<<<<<<<<<<
 *         total += duration
 *     for i in range(OpCode.MAX):
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyInt_From_int64_t(__pyx_v_7flitter_8language_2vm_CallOutCount); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8)) __PYX_ERR(0, 556, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_u_native_funcs);
    __Pyx_GIVEREF(__pyx_kp_u_native_funcs);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_native_funcs)) __PYX_ERR(0, 556, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_8 = 0;
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_stats, __pyx_t_9); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "flitter/language/vm.pyx":557
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)
 *         stats.append((duration, CallOutCount, '(native funcs)'))
 *         total += duration             # <<<<<<<<<<<<<<
 *     for i in range(OpCode.MAX):
 *         if StatsCount[i]:
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_duration);

    /* "flitter/language/vm.pyx":554
 *         StatsDuration[<int64_t>OpCode.MAX] += end
 *     overhead = (end - start) / 10000
 *     if CallOutCount:             # <<<<<<<<<<<<<<
 *         duration = max(0, CallOutDuration - CallOutCount*overhead)
 *         stats.append((duration, CallOutCount, '(native funcs)'))
 */
  }

  /* "flitter/language/vm.pyx":558
 *         stats.append((duration, CallOutCount, '(native funcs)'))
 *         total += duration
 *     for i in range(OpCode.MAX):             # <<<<<<<<<<<<<<
 *         if StatsCount[i]:
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
 */
  __pyx_t_11 = __pyx_e_7flitter_8language_2vm_MAX;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_12; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "flitter/language/vm.pyx":559
 *         total += duration
 *     for i in range(OpCode.MAX):
 *         if StatsCount[i]:             # <<<<<<<<<<<<<<
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))
 */
    __pyx_t_4 = ((__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_v_i]) != 0);
    if (__pyx_t_4) {

      /* "flitter/language/vm.pyx":560
 *     for i in range(OpCode.MAX):
 *         if StatsCount[i]:
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)             # <<<<<<<<<<<<<<
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))
 *             total += duration
 */
      __pyx_t_7 = ((__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_v_i]) - ((__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_v_i]) * __pyx_v_overhead));
      __pyx_t_6 = 0;
      __pyx_t_4 = (__pyx_t_7 > __pyx_t_6);
      if (__pyx_t_4) {
        __pyx_t_5 = __pyx_t_7;
      } else {
        __pyx_t_5 = __pyx_t_6;
      }
      __pyx_v_duration = __pyx_t_5;

      /* "flitter/language/vm.pyx":561
 *         if StatsCount[i]:
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))             # <<<<<<<<<<<<<<
 *             total += duration
 *     stats.sort(reverse=True)
 */
      __pyx_t_9 = PyFloat_FromDouble(__pyx_v_duration); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = __Pyx_PyInt_From_int64_t((__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_v_i])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(__pyx_v_7flitter_8language_2vm_OpCodeNames == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 561, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_OpCodeNames, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_9);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9)) __PYX_ERR(0, 561, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8)) __PYX_ERR(0, 561, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_13);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_13)) __PYX_ERR(0, 561, __pyx_L1_error);
      __pyx_t_9 = 0;
      __pyx_t_8 = 0;
      __pyx_t_13 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_stats, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":562
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))
 *             total += duration             # <<<<<<<<<<<<<<
 *     stats.sort(reverse=True)
 *     cdef str name
 */
      __pyx_v_total = (__pyx_v_total + __pyx_v_duration);

      /* "flitter/language/vm.pyx":559
 *         total += duration
 *     for i in range(OpCode.MAX):
 *         if StatsCount[i]:             # <<<<<<<<<<<<<<
 *             duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))
 */
    }
  }

  /* "flitter/language/vm.pyx":563
 *             stats.append((duration, StatsCount[i], OpCodeNames[i]))
 *             total += duration
 *     stats.sort(reverse=True)             # <<<<<<<<<<<<<<
 *     cdef str name
 *     logger.info("VM execution statistics:")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stats, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 563, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_13); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "flitter/language/vm.pyx":565
 *     stats.sort(reverse=True)
 *     cdef str name
 *     logger.info("VM execution statistics:")             # <<<<<<<<<<<<<<
 *     for duration, count, name in stats:
 *         per_execution = duration / count * 1e6
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_logger); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = NULL;
  __pyx_t_14 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_14 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_VM_execution_statistics};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "flitter/language/vm.pyx":566
 *     cdef str name
 *     logger.info("VM execution statistics:")
 *     for duration, count, name in stats:             # <<<<<<<<<<<<<<
 *         per_execution = duration / count * 1e6
 *         if per_execution < 1:
 */
  __pyx_t_8 = __pyx_v_stats; __Pyx_INCREF(__pyx_t_8);
  __pyx_t_15 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 566, __pyx_L1_error)
      #endif
      if (__pyx_t_15 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_15); __Pyx_INCREF(__pyx_t_1); __pyx_t_15++; if (unlikely((0 < 0))) __PYX_ERR(0, 566, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 566, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_13 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_13 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_16);
      #else
      __pyx_t_13 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_16 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_17 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_18 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_17);
      index = 0; __pyx_t_13 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_13)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_13);
      index = 1; __pyx_t_9 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_9)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_16 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_16)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_16);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_17), 3) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
      __pyx_t_18 = NULL;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_18 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 566, __pyx_L1_error)
      __pyx_L12_unpacking_done:;
    }
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_13); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_int64_t(__pyx_t_9); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (!(likely(PyUnicode_CheckExact(__pyx_t_16))||((__pyx_t_16) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_16))) __PYX_ERR(0, 566, __pyx_L1_error)
    __pyx_v_duration = __pyx_t_5;
    __pyx_v_count = __pyx_t_2;
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_16));
    __pyx_t_16 = 0;

    /* "flitter/language/vm.pyx":567
 *     logger.info("VM execution statistics:")
 *     for duration, count, name in stats:
 *         per_execution = duration / count * 1e6             # <<<<<<<<<<<<<<
 *         if per_execution < 1:
 *             logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,
 */
    if (unlikely(__pyx_v_count == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 567, __pyx_L1_error)
    }
    __pyx_v_per_execution = ((__pyx_v_duration / ((double)__pyx_v_count)) * 1e6);

    /* "flitter/language/vm.pyx":568
 *     for duration, count, name in stats:
 *         per_execution = duration / count * 1e6
 *         if per_execution < 1:             # <<<<<<<<<<<<<<
 *             logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,
 *                         duration, 100*duration/total)
 */
    __pyx_t_4 = (__pyx_v_per_execution < 1.0);
    if (__pyx_t_4) {

      /* "flitter/language/vm.pyx":569
 *         per_execution = duration / count * 1e6
 *         if per_execution < 1:
 *             logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,             # <<<<<<<<<<<<<<
 *                         duration, 100*duration/total)
 *         else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_logger); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_info); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_13 = PyFloat_FromDouble((__pyx_v_per_execution * 1000.0)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "flitter/language/vm.pyx":570
 *         if per_execution < 1:
 *             logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,
 *                         duration, 100*duration/total)             # <<<<<<<<<<<<<<
 *         else:
 *             logger.info("- {:15s} {:9d} x {:8.3f}s = {:7.3f}s ({:4.1f}%)", name, count, per_execution,
 */
      __pyx_t_17 = PyFloat_FromDouble(__pyx_v_duration); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_5 = (100.0 * __pyx_v_duration);
      if (unlikely(__pyx_v_total == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 570, __pyx_L1_error)
      }
      __pyx_t_19 = PyFloat_FromDouble((__pyx_t_5 / __pyx_v_total)); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_20 = NULL;
      __pyx_t_14 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_20)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_20);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_14 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[7] = {__pyx_t_20, __pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f, __pyx_v_name, __pyx_t_16, __pyx_t_13, __pyx_t_17, __pyx_t_19};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_14, 6+__pyx_t_14);
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":568
 *     for duration, count, name in stats:
 *         per_execution = duration / count * 1e6
 *         if per_execution < 1:             # <<<<<<<<<<<<<<
 *             logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,
 *                         duration, 100*duration/total)
 */
      goto __pyx_L13;
    }

    /* "flitter/language/vm.pyx":572
 *                         duration, 100*duration/total)
 *         else:
 *             logger.info("- {:15s} {:9d} x {:8.3f}s = {:7.3f}s ({:4.1f}%)", name, count, per_execution,             # <<<<<<<<<<<<<<
 *                         duration, 100*duration/total)
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_logger); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_info); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_17 = PyFloat_FromDouble(__pyx_v_per_execution); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);

      /* "flitter/language/vm.pyx":573
 *         else:
 *             logger.info("- {:15s} {:9d} x {:8.3f}s = {:7.3f}s ({:4.1f}%)", name, count, per_execution,
 *                         duration, 100*duration/total)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_13 = PyFloat_FromDouble(__pyx_v_duration); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_5 = (100.0 * __pyx_v_duration);
      if (unlikely(__pyx_v_total == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "float division");
        __PYX_ERR(0, 573, __pyx_L1_error)
      }
      __pyx_t_16 = PyFloat_FromDouble((__pyx_t_5 / __pyx_v_total)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_20 = NULL;
      __pyx_t_14 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_19))) {
        __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_19);
        if (likely(__pyx_t_20)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
          __Pyx_INCREF(__pyx_t_20);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_19, function);
          __pyx_t_14 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[7] = {__pyx_t_20, __pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f, __pyx_v_name, __pyx_t_9, __pyx_t_17, __pyx_t_13, __pyx_t_16};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_19, __pyx_callargs+1-__pyx_t_14, 6+__pyx_t_14);
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L13:;

    /* "flitter/language/vm.pyx":566
 *     cdef str name
 *     logger.info("VM execution statistics:")
 *     for duration, count, name in stats:             # <<<<<<<<<<<<<<
 *         per_execution = duration / count * 1e6
 *         if per_execution < 1:
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "flitter/language/vm.pyx":542
 * 
 * 
 * def log_vm_stats():             # <<<<<<<<<<<<<<
 *     cdef list stats = []
 *     cdef double duration, total=0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("flitter.language.vm.log_vm_stats", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stats);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":576
 * 
 * 
 * cdef inline void call_helper(Context context, VectorStack stack, object function, tuple args, dict kwargs, bint record_stats, double* duration):             # <<<<<<<<<<<<<<
 *     global CallOutDuration, CallOutCount
 *     cdef int64_t i, n=PyTuple_GET_SIZE(args)
 */

static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_call_helper(struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, PyObject *__pyx_v_function, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, int __pyx_v_record_stats, double *__pyx_v_duration) {
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  double __pyx_v_call_duration;
  PyObject *__pyx_v_context_args = 0;
  int __pyx_v_is_func;
  PyObject *__pyx_v_obj;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  int64_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_helper", 0);
  __Pyx_INCREF(__pyx_v_args);

  /* "flitter/language/vm.pyx":578
 * cdef inline void call_helper(Context context, VectorStack stack, object function, tuple args, dict kwargs, bint record_stats, double* duration):
 *     global CallOutDuration, CallOutCount
 *     cdef int64_t i, n=PyTuple_GET_SIZE(args)             # <<<<<<<<<<<<<<
 *     cdef double call_duration = 0
 *     cdef tuple context_args
 */
  __pyx_v_n = PyTuple_GET_SIZE(__pyx_v_args);

  /* "flitter/language/vm.pyx":579
 *     global CallOutDuration, CallOutCount
 *     cdef int64_t i, n=PyTuple_GET_SIZE(args)
 *     cdef double call_duration = 0             # <<<<<<<<<<<<<<
 *     cdef tuple context_args
 *     cdef bint is_func = type(function) is Function
 */
  __pyx_v_call_duration = 0.0;

  /* "flitter/language/vm.pyx":581
 *     cdef double call_duration = 0
 *     cdef tuple context_args
 *     cdef bint is_func = type(function) is Function             # <<<<<<<<<<<<<<
 *     if is_func or PyObject_HasAttrString(function, ContextFunc):
 *         context_args = PyTuple_New(n+1)
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_function)) == ((PyObject *)__pyx_ptype_7flitter_8language_2vm_Function));
  __pyx_v_is_func = __pyx_t_1;

  /* "flitter/language/vm.pyx":582
 *     cdef tuple context_args
 *     cdef bint is_func = type(function) is Function
 *     if is_func or PyObject_HasAttrString(function, ContextFunc):             # <<<<<<<<<<<<<<
 *         context_args = PyTuple_New(n+1)
 *         Py_INCREF(context)
 */
  if (!__pyx_v_is_func) {
  } else {
    __pyx_t_1 = __pyx_v_is_func;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyObject_HasAttrString(__pyx_v_function, __pyx_v_7flitter_8language_2vm_ContextFunc);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":583
 *     cdef bint is_func = type(function) is Function
 *     if is_func or PyObject_HasAttrString(function, ContextFunc):
 *         context_args = PyTuple_New(n+1)             # <<<<<<<<<<<<<<
 *         Py_INCREF(context)
 *         PyTuple_SET_ITEM(context_args, 0, context)
 */
    __pyx_t_3 = PyTuple_New((__pyx_v_n + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_context_args = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "flitter/language/vm.pyx":584
 *     if is_func or PyObject_HasAttrString(function, ContextFunc):
 *         context_args = PyTuple_New(n+1)
 *         Py_INCREF(context)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(context_args, 0, context)
 *         for i in range(n):
 */
    Py_INCREF(((PyObject *)__pyx_v_context));

    /* "flitter/language/vm.pyx":585
 *         context_args = PyTuple_New(n+1)
 *         Py_INCREF(context)
 *         PyTuple_SET_ITEM(context_args, 0, context)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obj = PyTuple_GET_ITEM(args, i)
 */
    PyTuple_SET_ITEM(__pyx_v_context_args, 0, ((PyObject *)__pyx_v_context));

    /* "flitter/language/vm.pyx":586
 *         Py_INCREF(context)
 *         PyTuple_SET_ITEM(context_args, 0, context)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obj = PyTuple_GET_ITEM(args, i)
 *             Py_INCREF(<object>obj)
 */
    __pyx_t_4 = __pyx_v_n;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "flitter/language/vm.pyx":587
 *         PyTuple_SET_ITEM(context_args, 0, context)
 *         for i in range(n):
 *             obj = PyTuple_GET_ITEM(args, i)             # <<<<<<<<<<<<<<
 *             Py_INCREF(<object>obj)
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)
 */
      __pyx_v_obj = PyTuple_GET_ITEM(__pyx_v_args, __pyx_v_i);

      /* "flitter/language/vm.pyx":588
 *         for i in range(n):
 *             obj = PyTuple_GET_ITEM(args, i)
 *             Py_INCREF(<object>obj)             # <<<<<<<<<<<<<<
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)
 *         args = context_args
 */
      Py_INCREF(((PyObject *)__pyx_v_obj));

      /* "flitter/language/vm.pyx":589
 *             obj = PyTuple_GET_ITEM(args, i)
 *             Py_INCREF(<object>obj)
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)             # <<<<<<<<<<<<<<
 *         args = context_args
 *     if record_stats:
 */
      PyTuple_SET_ITEM(__pyx_v_context_args, (__pyx_v_i + 1), ((PyObject *)__pyx_v_obj));
    }

    /* "flitter/language/vm.pyx":590
 *             Py_INCREF(<object>obj)
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)
 *         args = context_args             # <<<<<<<<<<<<<<
 *     if record_stats:
 *         call_duration = -perf_counter()
 */
    __Pyx_INCREF(__pyx_v_context_args);
    __Pyx_DECREF_SET(__pyx_v_args, __pyx_v_context_args);

    /* "flitter/language/vm.pyx":582
 *     cdef tuple context_args
 *     cdef bint is_func = type(function) is Function
 *     if is_func or PyObject_HasAttrString(function, ContextFunc):             # <<<<<<<<<<<<<<
 *         context_args = PyTuple_New(n+1)
 *         Py_INCREF(context)
 */
  }

  /* "flitter/language/vm.pyx":591
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)
 *         args = context_args
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         call_duration = -perf_counter()
 *     try:
 */
  if (__pyx_v_record_stats) {

    /* "flitter/language/vm.pyx":592
 *         args = context_args
 *     if record_stats:
 *         call_duration = -perf_counter()             # <<<<<<<<<<<<<<
 *     try:
 *         if kwargs is None:
 */
    __pyx_v_call_duration = (-perf_counter());

    /* "flitter/language/vm.pyx":591
 *             PyTuple_SET_ITEM(context_args, i+1, <object>obj)
 *         args = context_args
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         call_duration = -perf_counter()
 *     try:
 */
  }

  /* "flitter/language/vm.pyx":593
 *     if record_stats:
 *         call_duration = -perf_counter()
 *     try:             # <<<<<<<<<<<<<<
 *         if kwargs is None:
 *             push(stack, PyObject_CallObject(function, args))
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "flitter/language/vm.pyx":594
 *         call_duration = -perf_counter()
 *     try:
 *         if kwargs is None:             # <<<<<<<<<<<<<<
 *             push(stack, PyObject_CallObject(function, args))
 *         else:
 */
      __pyx_t_1 = (__pyx_v_kwargs == ((PyObject*)Py_None));
      if (__pyx_t_1) {

        /* "flitter/language/vm.pyx":595
 *     try:
 *         if kwargs is None:
 *             push(stack, PyObject_CallObject(function, args))             # <<<<<<<<<<<<<<
 *         else:
 *             push(stack, PyObject_Call(function, args, kwargs))
 */
        __pyx_t_3 = PyObject_CallObject(__pyx_v_function, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 595, __pyx_L9_error)
        __pyx_t_4 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3)); if (unlikely(__pyx_t_4 == ((int64_t)0))) __PYX_ERR(0, 595, __pyx_L9_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "flitter/language/vm.pyx":594
 *         call_duration = -perf_counter()
 *     try:
 *         if kwargs is None:             # <<<<<<<<<<<<<<
 *             push(stack, PyObject_CallObject(function, args))
 *         else:
 */
        goto __pyx_L15;
      }

      /* "flitter/language/vm.pyx":597
 *             push(stack, PyObject_CallObject(function, args))
 *         else:
 *             push(stack, PyObject_Call(function, args, kwargs))             # <<<<<<<<<<<<<<
 *     except (RecursionError, AssertionError, StackError):
 *         raise
 */
      /*else*/ {
        __pyx_t_3 = PyObject_Call(__pyx_v_function, __pyx_v_args, __pyx_v_kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 597, __pyx_L9_error)
        __pyx_t_4 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3)); if (unlikely(__pyx_t_4 == ((int64_t)0))) __PYX_ERR(0, 597, __pyx_L9_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_L15:;

      /* "flitter/language/vm.pyx":593
 *     if record_stats:
 *         call_duration = -perf_counter()
 *     try:             # <<<<<<<<<<<<<<
 *         if kwargs is None:
 *             push(stack, PyObject_CallObject(function, args))
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/vm.pyx":598
 *         else:
 *             push(stack, PyObject_Call(function, args, kwargs))
 *     except (RecursionError, AssertionError, StackError):             # <<<<<<<<<<<<<<
 *         raise
 *     except Exception as exc:
 */
    __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_10, &__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 598, __pyx_L11_except_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_StackError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 598, __pyx_L11_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_12) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_builtin_AssertionError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_ErrRestore(__pyx_t_3, __pyx_t_10, __pyx_t_11);
    __pyx_t_3 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
    if (__pyx_t_14) {
      __Pyx_AddTraceback("flitter.language.vm.call_helper", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_10, &__pyx_t_3) < 0) __PYX_ERR(0, 598, __pyx_L11_except_error)
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "flitter/language/vm.pyx":599
 *             push(stack, PyObject_Call(function, args, kwargs))
 *     except (RecursionError, AssertionError, StackError):
 *         raise             # <<<<<<<<<<<<<<
 *     except Exception as exc:
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 */
      __Pyx_GIVEREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_11, __pyx_t_10, __pyx_t_3);
      __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(0, 599, __pyx_L11_except_error)
    }

    /* "flitter/language/vm.pyx":600
 *     except (RecursionError, AssertionError, StackError):
 *         raise
 *     except Exception as exc:             # <<<<<<<<<<<<<<
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 *         push(stack, null_)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_14) {
      __Pyx_AddTraceback("flitter.language.vm.call_helper", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_10, &__pyx_t_11) < 0) __PYX_ERR(0, 600, __pyx_L11_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_10);
      __pyx_v_exc = __pyx_t_10;
      /*try:*/ {

        /* "flitter/language/vm.pyx":601
 *         raise
 *     except Exception as exc:
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")             # <<<<<<<<<<<<<<
 *         push(stack, null_)
 *     if record_stats:
 */
        __pyx_t_13 = __pyx_v_context->errors;
        __Pyx_INCREF(__pyx_t_13);
        __pyx_t_12 = PyTuple_New(4); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_15 = 0;
        __pyx_t_16 = 127;
        __Pyx_INCREF(__pyx_kp_u_Error_calling);
        __pyx_t_15 += 14;
        __Pyx_GIVEREF(__pyx_kp_u_Error_calling);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_u_Error_calling);
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_function, __pyx_n_s_name_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_18 = __Pyx_PyObject_FormatSimple(__pyx_t_17, __pyx_empty_unicode); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) : __pyx_t_16;
        __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18);
        __Pyx_GIVEREF(__pyx_t_18);
        PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_18);
        __pyx_t_18 = 0;
        __Pyx_INCREF(__pyx_kp_u__7);
        __pyx_t_15 += 2;
        __Pyx_GIVEREF(__pyx_kp_u__7);
        PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_u__7);
        __pyx_t_18 = __Pyx_PyObject_Unicode(__pyx_v_exc); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) : __pyx_t_16;
        __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18);
        __Pyx_GIVEREF(__pyx_t_18);
        PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_t_18);
        __pyx_t_18 = 0;
        __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_12, 4, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_14 = PySet_Add(__pyx_t_13, __pyx_t_18); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 601, __pyx_L23_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

        /* "flitter/language/vm.pyx":602
 *     except Exception as exc:
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 *         push(stack, null_)             # <<<<<<<<<<<<<<
 *     if record_stats:
 *         call_duration += perf_counter()
 */
        __pyx_t_18 = ((PyObject *)__pyx_v_7flitter_5model_null_);
        __Pyx_INCREF(__pyx_t_18);
        __pyx_t_4 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_18)); if (unlikely(__pyx_t_4 == ((int64_t)0))) __PYX_ERR(0, 602, __pyx_L23_error)
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      }

      /* "flitter/language/vm.pyx":600
 *     except (RecursionError, AssertionError, StackError):
 *         raise
 *     except Exception as exc:             # <<<<<<<<<<<<<<
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 *         push(stack, null_)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L24;
        }
        __pyx_L23_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_24, &__pyx_t_25, &__pyx_t_26);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_24);
          __Pyx_XGOTREF(__pyx_t_25);
          __Pyx_XGOTREF(__pyx_t_26);
          __pyx_t_14 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_XGIVEREF(__pyx_t_26);
            __Pyx_ExceptionReset(__pyx_t_24, __pyx_t_25, __pyx_t_26);
          }
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_XGIVEREF(__pyx_t_22);
          __Pyx_XGIVEREF(__pyx_t_23);
          __Pyx_ErrRestore(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
          goto __pyx_L11_except_error;
        }
        __pyx_L24:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L10_exception_handled;
    }
    goto __pyx_L11_except_error;

    /* "flitter/language/vm.pyx":593
 *     if record_stats:
 *         call_duration = -perf_counter()
 *     try:             # <<<<<<<<<<<<<<
 *         if kwargs is None:
 *             push(stack, PyObject_CallObject(function, args))
 */
    __pyx_L11_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L10_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L14_try_end:;
  }

  /* "flitter/language/vm.pyx":603
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 *         push(stack, null_)
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         call_duration += perf_counter()
 *         duration[0] = duration[0] - call_duration
 */
  if (__pyx_v_record_stats) {

    /* "flitter/language/vm.pyx":604
 *         push(stack, null_)
 *     if record_stats:
 *         call_duration += perf_counter()             # <<<<<<<<<<<<<<
 *         duration[0] = duration[0] - call_duration
 *         if not is_func:
 */
    __pyx_v_call_duration = (__pyx_v_call_duration + perf_counter());

    /* "flitter/language/vm.pyx":605
 *     if record_stats:
 *         call_duration += perf_counter()
 *         duration[0] = duration[0] - call_duration             # <<<<<<<<<<<<<<
 *         if not is_func:
 *             CallOutDuration += call_duration
 */
    (__pyx_v_duration[0]) = ((__pyx_v_duration[0]) - __pyx_v_call_duration);

    /* "flitter/language/vm.pyx":606
 *         call_duration += perf_counter()
 *         duration[0] = duration[0] - call_duration
 *         if not is_func:             # <<<<<<<<<<<<<<
 *             CallOutDuration += call_duration
 *             CallOutCount += 1
 */
    __pyx_t_1 = (!__pyx_v_is_func);
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":607
 *         duration[0] = duration[0] - call_duration
 *         if not is_func:
 *             CallOutDuration += call_duration             # <<<<<<<<<<<<<<
 *             CallOutCount += 1
 * 
 */
      __pyx_v_7flitter_8language_2vm_CallOutDuration = (__pyx_v_7flitter_8language_2vm_CallOutDuration + __pyx_v_call_duration);

      /* "flitter/language/vm.pyx":608
 *         if not is_func:
 *             CallOutDuration += call_duration
 *             CallOutCount += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_7flitter_8language_2vm_CallOutCount = (__pyx_v_7flitter_8language_2vm_CallOutCount + 1);

      /* "flitter/language/vm.pyx":606
 *         call_duration += perf_counter()
 *         duration[0] = duration[0] - call_duration
 *         if not is_func:             # <<<<<<<<<<<<<<
 *             CallOutDuration += call_duration
 *             CallOutCount += 1
 */
    }

    /* "flitter/language/vm.pyx":603
 *         PySet_Add(context.errors, f"Error calling {function.__name__}: {str(exc)}")
 *         push(stack, null_)
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         call_duration += perf_counter()
 *         duration[0] = duration[0] - call_duration
 */
  }

  /* "flitter/language/vm.pyx":576
 * 
 * 
 * cdef inline void call_helper(Context context, VectorStack stack, object function, tuple args, dict kwargs, bint record_stats, double* duration):             # <<<<<<<<<<<<<<
 *     global CallOutDuration, CallOutCount
 *     cdef int64_t i, n=PyTuple_GET_SIZE(args)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("flitter.language.vm.call_helper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context_args);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":611
 * 
 * 
 * cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration, bint simplify):             # <<<<<<<<<<<<<<
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)
 *     if program is None:
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_import_module(struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, PyObject *__pyx_v_filename, int __pyx_v_record_stats, double *__pyx_v_duration, int __pyx_v_simplify) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program = 0;
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_import_context = 0;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack = 0;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_lnames = 0;
  int64_t __pyx_v_stack_top;
  int64_t __pyx_v_lnames_top;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_t_8;
  int64_t __pyx_t_9;
  long __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_module", 1);

  /* "flitter/language/vm.pyx":612
 * 
 * cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration, bint simplify):
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)             # <<<<<<<<<<<<<<
 *     if program is None:
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SharedCache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_with_root); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_filename, __pyx_v_context->path};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_read_flitter_program); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_simplify); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_simplify, __pyx_t_2) < 0) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 612, __pyx_L1_error)
  __pyx_v_program = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":613
 * cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration, bint simplify):
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)
 *     if program is None:             # <<<<<<<<<<<<<<
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")
 *         return None
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_program) == Py_None);
  if (__pyx_t_5) {

    /* "flitter/language/vm.pyx":614
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)
 *     if program is None:
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")             # <<<<<<<<<<<<<<
 *         return None
 *     cdef Context import_context = context
 */
    __pyx_t_2 = __pyx_v_context->errors;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_Unable_to_import_from);
    __pyx_t_6 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_Unable_to_import_from);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Unable_to_import_from);
    __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u__9);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__9);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__9);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = PySet_Add(__pyx_t_2, __pyx_t_3); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/vm.pyx":615
 *     if program is None:
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")
 *         return None             # <<<<<<<<<<<<<<
 *     cdef Context import_context = context
 *     while import_context is not None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":613
 * cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration, bint simplify):
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)
 *     if program is None:             # <<<<<<<<<<<<<<
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")
 *         return None
 */
  }

  /* "flitter/language/vm.pyx":616
 *         PySet_Add(context.errors, f"Unable to import from '{filename}'")
 *         return None
 *     cdef Context import_context = context             # <<<<<<<<<<<<<<
 *     while import_context is not None:
 *         if import_context.path is program.path:
 */
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __pyx_v_import_context = __pyx_v_context;

  /* "flitter/language/vm.pyx":617
 *         return None
 *     cdef Context import_context = context
 *     while import_context is not None:             # <<<<<<<<<<<<<<
 *         if import_context.path is program.path:
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")
 */
  while (1) {
    __pyx_t_5 = (((PyObject *)__pyx_v_import_context) != Py_None);
    if (!__pyx_t_5) break;

    /* "flitter/language/vm.pyx":618
 *     cdef Context import_context = context
 *     while import_context is not None:
 *         if import_context.path is program.path:             # <<<<<<<<<<<<<<
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")
 *             return None
 */
    __pyx_t_5 = (__pyx_v_import_context->path == __pyx_v_program->path);
    if (__pyx_t_5) {

      /* "flitter/language/vm.pyx":619
 *     while import_context is not None:
 *         if import_context.path is program.path:
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")             # <<<<<<<<<<<<<<
 *             return None
 *         import_context = import_context.parent
 */
      __pyx_t_3 = __pyx_v_context->errors;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = 0;
      __pyx_t_7 = 127;
      __Pyx_INCREF(__pyx_kp_u_Circular_import_of);
      __pyx_t_6 += 20;
      __Pyx_GIVEREF(__pyx_kp_u_Circular_import_of);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Circular_import_of);
      __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_7;
      __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__9);
      __pyx_t_6 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__9);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__9);
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = PySet_Add(__pyx_t_3, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":620
 *         if import_context.path is program.path:
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")
 *             return None             # <<<<<<<<<<<<<<
 *         import_context = import_context.parent
 *     context.errors.update(program.compiler_errors)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "flitter/language/vm.pyx":618
 *     cdef Context import_context = context
 *     while import_context is not None:
 *         if import_context.path is program.path:             # <<<<<<<<<<<<<<
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")
 *             return None
 */
    }

    /* "flitter/language/vm.pyx":621
 *             PySet_Add(context.errors, f"Circular import of '{filename}'")
 *             return None
 *         import_context = import_context.parent             # <<<<<<<<<<<<<<
 *     context.errors.update(program.compiler_errors)
 *     cdef VectorStack stack=context.stack, lnames=context.lnames
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_import_context->parent);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_import_context, ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_1));
    __pyx_t_1 = 0;
  }

  /* "flitter/language/vm.pyx":622
 *             return None
 *         import_context = import_context.parent
 *     context.errors.update(program.compiler_errors)             # <<<<<<<<<<<<<<
 *     cdef VectorStack stack=context.stack, lnames=context.lnames
 *     cdef int64_t stack_top=stack.top, lnames_top=lnames.top
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->errors, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_program->compiler_errors};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":623
 *         import_context = import_context.parent
 *     context.errors.update(program.compiler_errors)
 *     cdef VectorStack stack=context.stack, lnames=context.lnames             # <<<<<<<<<<<<<<
 *     cdef int64_t stack_top=stack.top, lnames_top=lnames.top
 *     import_context = Context.__new__(Context)
 */
  if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 623, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 623, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_context->lnames;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":624
 *     context.errors.update(program.compiler_errors)
 *     cdef VectorStack stack=context.stack, lnames=context.lnames
 *     cdef int64_t stack_top=stack.top, lnames_top=lnames.top             # <<<<<<<<<<<<<<
 *     import_context = Context.__new__(Context)
 *     import_context.parent = context
 */
  __pyx_t_9 = __pyx_v_stack->top;
  __pyx_v_stack_top = __pyx_t_9;
  __pyx_t_9 = __pyx_v_lnames->top;
  __pyx_v_lnames_top = __pyx_t_9;

  /* "flitter/language/vm.pyx":625
 *     cdef VectorStack stack=context.stack, lnames=context.lnames
 *     cdef int64_t stack_top=stack.top, lnames_top=lnames.top
 *     import_context = Context.__new__(Context)             # <<<<<<<<<<<<<<
 *     import_context.parent = context
 *     import_context.errors = context.errors
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Context), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_5model_Context)))) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_import_context, ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":626
 *     cdef int64_t stack_top=stack.top, lnames_top=lnames.top
 *     import_context = Context.__new__(Context)
 *     import_context.parent = context             # <<<<<<<<<<<<<<
 *     import_context.errors = context.errors
 *     import_context.logs = context.logs
 */
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __Pyx_GIVEREF((PyObject *)__pyx_v_context);
  __Pyx_GOTREF((PyObject *)__pyx_v_import_context->parent);
  __Pyx_DECREF((PyObject *)__pyx_v_import_context->parent);
  __pyx_v_import_context->parent = __pyx_v_context;

  /* "flitter/language/vm.pyx":627
 *     import_context = Context.__new__(Context)
 *     import_context.parent = context
 *     import_context.errors = context.errors             # <<<<<<<<<<<<<<
 *     import_context.logs = context.logs
 *     import_context.state = StateDict()
 */
  __pyx_t_1 = __pyx_v_context->errors;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_import_context->errors);
  __Pyx_DECREF(__pyx_v_import_context->errors);
  __pyx_v_import_context->errors = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":628
 *     import_context.parent = context
 *     import_context.errors = context.errors
 *     import_context.logs = context.logs             # <<<<<<<<<<<<<<
 *     import_context.state = StateDict()
 *     import_context.names = {}
 */
  __pyx_t_1 = __pyx_v_context->logs;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_import_context->logs);
  __Pyx_DECREF(__pyx_v_import_context->logs);
  __pyx_v_import_context->logs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":629
 *     import_context.errors = context.errors
 *     import_context.logs = context.logs
 *     import_context.state = StateDict()             # <<<<<<<<<<<<<<
 *     import_context.names = {}
 *     import_context.exports = {}
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_5model_StateDict)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_import_context->state);
  __Pyx_DECREF((PyObject *)__pyx_v_import_context->state);
  __pyx_v_import_context->state = ((struct __pyx_obj_7flitter_5model_StateDict *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":630
 *     import_context.logs = context.logs
 *     import_context.state = StateDict()
 *     import_context.names = {}             # <<<<<<<<<<<<<<
 *     import_context.exports = {}
 *     import_context.path = program.path
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_import_context->names);
  __Pyx_DECREF(__pyx_v_import_context->names);
  __pyx_v_import_context->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":631
 *     import_context.state = StateDict()
 *     import_context.names = {}
 *     import_context.exports = {}             # <<<<<<<<<<<<<<
 *     import_context.path = program.path
 *     import_context.stack = stack
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_import_context->exports);
  __Pyx_DECREF(__pyx_v_import_context->exports);
  __pyx_v_import_context->exports = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":632
 *     import_context.names = {}
 *     import_context.exports = {}
 *     import_context.path = program.path             # <<<<<<<<<<<<<<
 *     import_context.stack = stack
 *     import_context.lnames = lnames
 */
  __pyx_t_1 = __pyx_v_program->path;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_import_context->path);
  __Pyx_DECREF(__pyx_v_import_context->path);
  __pyx_v_import_context->path = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":633
 *     import_context.exports = {}
 *     import_context.path = program.path
 *     import_context.stack = stack             # <<<<<<<<<<<<<<
 *     import_context.lnames = lnames
 *     push(stack, Vector(Node('root')))
 */
  __Pyx_INCREF((PyObject *)__pyx_v_stack);
  __Pyx_GIVEREF((PyObject *)__pyx_v_stack);
  __Pyx_GOTREF(__pyx_v_import_context->stack);
  __Pyx_DECREF(__pyx_v_import_context->stack);
  __pyx_v_import_context->stack = ((PyObject *)__pyx_v_stack);

  /* "flitter/language/vm.pyx":634
 *     import_context.path = program.path
 *     import_context.stack = stack
 *     import_context.lnames = lnames             # <<<<<<<<<<<<<<
 *     push(stack, Vector(Node('root')))
 *     if record_stats:
 */
  __Pyx_INCREF((PyObject *)__pyx_v_lnames);
  __Pyx_GIVEREF((PyObject *)__pyx_v_lnames);
  __Pyx_GOTREF(__pyx_v_import_context->lnames);
  __Pyx_DECREF(__pyx_v_import_context->lnames);
  __pyx_v_import_context->lnames = ((PyObject *)__pyx_v_lnames);

  /* "flitter/language/vm.pyx":635
 *     import_context.stack = stack
 *     import_context.lnames = lnames
 *     push(stack, Vector(Node('root')))             # <<<<<<<<<<<<<<
 *     if record_stats:
 *         duration[0] += perf_counter()
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_5model_Node), __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3)); if (unlikely(__pyx_t_9 == ((int64_t)0))) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/vm.pyx":636
 *     import_context.lnames = lnames
 *     push(stack, Vector(Node('root')))
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         duration[0] += perf_counter()
 *     program._execute(import_context, 0, record_stats)
 */
  if (__pyx_v_record_stats) {

    /* "flitter/language/vm.pyx":637
 *     push(stack, Vector(Node('root')))
 *     if record_stats:
 *         duration[0] += perf_counter()             # <<<<<<<<<<<<<<
 *     program._execute(import_context, 0, record_stats)
 *     if record_stats:
 */
    __pyx_t_10 = 0;
    (__pyx_v_duration[__pyx_t_10]) = ((__pyx_v_duration[__pyx_t_10]) + perf_counter());

    /* "flitter/language/vm.pyx":636
 *     import_context.lnames = lnames
 *     push(stack, Vector(Node('root')))
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         duration[0] += perf_counter()
 *     program._execute(import_context, 0, record_stats)
 */
  }

  /* "flitter/language/vm.pyx":638
 *     if record_stats:
 *         duration[0] += perf_counter()
 *     program._execute(import_context, 0, record_stats)             # <<<<<<<<<<<<<<
 *     if record_stats:
 *         duration[0] -= perf_counter()
 */
  ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->_execute(__pyx_v_program, __pyx_v_import_context, 0, __pyx_v_record_stats); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 638, __pyx_L1_error)

  /* "flitter/language/vm.pyx":639
 *         duration[0] += perf_counter()
 *     program._execute(import_context, 0, record_stats)
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         duration[0] -= perf_counter()
 *     drop(stack, 1)
 */
  if (__pyx_v_record_stats) {

    /* "flitter/language/vm.pyx":640
 *     program._execute(import_context, 0, record_stats)
 *     if record_stats:
 *         duration[0] -= perf_counter()             # <<<<<<<<<<<<<<
 *     drop(stack, 1)
 *     assert stack.top == stack_top, "Bad stack"
 */
    __pyx_t_10 = 0;
    (__pyx_v_duration[__pyx_t_10]) = ((__pyx_v_duration[__pyx_t_10]) - perf_counter());

    /* "flitter/language/vm.pyx":639
 *         duration[0] += perf_counter()
 *     program._execute(import_context, 0, record_stats)
 *     if record_stats:             # <<<<<<<<<<<<<<
 *         duration[0] -= perf_counter()
 *     drop(stack, 1)
 */
  }

  /* "flitter/language/vm.pyx":641
 *     if record_stats:
 *         duration[0] -= perf_counter()
 *     drop(stack, 1)             # <<<<<<<<<<<<<<
 *     assert stack.top == stack_top, "Bad stack"
 *     assert lnames.top == lnames_top, "Bad lnames"
 */
  __pyx_f_7flitter_8language_2vm_drop(__pyx_v_stack, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 641, __pyx_L1_error)

  /* "flitter/language/vm.pyx":642
 *         duration[0] -= perf_counter()
 *     drop(stack, 1)
 *     assert stack.top == stack_top, "Bad stack"             # <<<<<<<<<<<<<<
 *     assert lnames.top == lnames_top, "Bad lnames"
 *     return import_context.exports
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_5 = (__pyx_v_stack->top == __pyx_v_stack_top);
    if (unlikely(!__pyx_t_5)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_stack, 0, 0);
      __PYX_ERR(0, 642, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 642, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":643
 *     drop(stack, 1)
 *     assert stack.top == stack_top, "Bad stack"
 *     assert lnames.top == lnames_top, "Bad lnames"             # <<<<<<<<<<<<<<
 *     return import_context.exports
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_5 = (__pyx_v_lnames->top == __pyx_v_lnames_top);
    if (unlikely(!__pyx_t_5)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_lnames, 0, 0);
      __PYX_ERR(0, 643, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 643, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":644
 *     assert stack.top == stack_top, "Bad stack"
 *     assert lnames.top == lnames_top, "Bad lnames"
 *     return import_context.exports             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_import_context->exports);
  __pyx_r = __pyx_v_import_context->exports;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":611
 * 
 * 
 * cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration, bint simplify):             # <<<<<<<<<<<<<<
 *     cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program(simplify=simplify)
 *     if program is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.vm.import_module", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_program);
  __Pyx_XDECREF((PyObject *)__pyx_v_import_context);
  __Pyx_XDECREF((PyObject *)__pyx_v_stack);
  __Pyx_XDECREF((PyObject *)__pyx_v_lnames);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":647
 * 
 * 
 * cdef inline void execute_append(VectorStack stack, int64_t count):             # <<<<<<<<<<<<<<
 *     cdef Vector nodes_vec = peek_at(stack, count)
 *     cdef Vector children
 */

static CYTHON_INLINE void __pyx_f_7flitter_8language_2vm_execute_append(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, int64_t __pyx_v_count) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes_vec = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_children = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_j;
  int64_t __pyx_v_k;
  int64_t __pyx_v_m;
  int64_t __pyx_v_n;
  int64_t __pyx_v_o;
  PyObject *__pyx_v_nodes = 0;
  PyObject *__pyx_v_nodeptr;
  PyObject *__pyx_v_childrenptr;
  PyObject *__pyx_v_objptr;
  PyObject *__pyx_v_dest = 0;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;
  int64_t __pyx_t_9;
  int64_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_append", 1);

  /* "flitter/language/vm.pyx":648
 * 
 * cdef inline void execute_append(VectorStack stack, int64_t count):
 *     cdef Vector nodes_vec = peek_at(stack, count)             # <<<<<<<<<<<<<<
 *     cdef Vector children
 *     cdef int64_t i, j, k, m=nodes_vec.length, n, o
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_stack, __pyx_v_count)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nodes_vec = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":650
 *     cdef Vector nodes_vec = peek_at(stack, count)
 *     cdef Vector children
 *     cdef int64_t i, j, k, m=nodes_vec.length, n, o             # <<<<<<<<<<<<<<
 *     cdef tuple nodes = nodes_vec.objects
 *     cdef PyObject* nodeptr
 */
  __pyx_t_2 = __pyx_v_nodes_vec->length;
  __pyx_v_m = __pyx_t_2;

  /* "flitter/language/vm.pyx":651
 *     cdef Vector children
 *     cdef int64_t i, j, k, m=nodes_vec.length, n, o
 *     cdef tuple nodes = nodes_vec.objects             # <<<<<<<<<<<<<<
 *     cdef PyObject* nodeptr
 *     cdef PyObject* childrenptr
 */
  __pyx_t_1 = __pyx_v_nodes_vec->objects;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":656
 *     cdef PyObject* objptr
 *     cdef tuple dest
 *     if nodes is not None:             # <<<<<<<<<<<<<<
 *         for i in range(m):
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 */
  __pyx_t_3 = (__pyx_v_nodes != ((PyObject*)Py_None));
  if (__pyx_t_3) {

    /* "flitter/language/vm.pyx":657
 *     cdef tuple dest
 *     if nodes is not None:
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>nodeptr) is not Node:
 */
    __pyx_t_2 = __pyx_v_m;
    __pyx_t_4 = __pyx_t_2;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/language/vm.pyx":658
 *     if nodes is not None:
 *         for i in range(m):
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)             # <<<<<<<<<<<<<<
 *             if type(<object>nodeptr) is not Node:
 *                 continue
 */
      __pyx_v_nodeptr = PyTuple_GET_ITEM(__pyx_v_nodes, __pyx_v_i);

      /* "flitter/language/vm.pyx":659
 *         for i in range(m):
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>nodeptr) is not Node:             # <<<<<<<<<<<<<<
 *                 continue
 *             if nodeptr.ob_refcnt > 1:
 */
      __pyx_t_3 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_nodeptr))) != ((PyObject *)__pyx_ptype_7flitter_5model_Node));
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":660
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>nodeptr) is not Node:
 *                 continue             # <<<<<<<<<<<<<<
 *             if nodeptr.ob_refcnt > 1:
 *                 node = (<Node>nodeptr).copy()
 */
        goto __pyx_L4_continue;

        /* "flitter/language/vm.pyx":659
 *         for i in range(m):
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>nodeptr) is not Node:             # <<<<<<<<<<<<<<
 *                 continue
 *             if nodeptr.ob_refcnt > 1:
 */
      }

      /* "flitter/language/vm.pyx":661
 *             if type(<object>nodeptr) is not Node:
 *                 continue
 *             if nodeptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *                 node = (<Node>nodeptr).copy()
 *                 Py_DECREF(<Node>nodeptr)
 */
      __pyx_t_3 = (__pyx_v_nodeptr->ob_refcnt > 1);
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":662
 *                 continue
 *             if nodeptr.ob_refcnt > 1:
 *                 node = (<Node>nodeptr).copy()             # <<<<<<<<<<<<<<
 *                 Py_DECREF(<Node>nodeptr)
 *                 Py_INCREF(node)
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":663
 *             if nodeptr.ob_refcnt > 1:
 *                 node = (<Node>nodeptr).copy()
 *                 Py_DECREF(<Node>nodeptr)             # <<<<<<<<<<<<<<
 *                 Py_INCREF(node)
 *                 PyTuple_SET_ITEM(nodes, i, node)
 */
        Py_DECREF(((PyObject *)__pyx_v_nodeptr));

        /* "flitter/language/vm.pyx":664
 *                 node = (<Node>nodeptr).copy()
 *                 Py_DECREF(<Node>nodeptr)
 *                 Py_INCREF(node)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(nodes, i, node)
 *                 nodeptr = <PyObject*>node
 */
        Py_INCREF(((PyObject *)__pyx_v_node));

        /* "flitter/language/vm.pyx":665
 *                 Py_DECREF(<Node>nodeptr)
 *                 Py_INCREF(node)
 *                 PyTuple_SET_ITEM(nodes, i, node)             # <<<<<<<<<<<<<<
 *                 nodeptr = <PyObject*>node
 *             childrenptr = <PyObject*>((<Node>nodeptr)._children)
 */
        PyTuple_SET_ITEM(__pyx_v_nodes, __pyx_v_i, ((PyObject *)__pyx_v_node));

        /* "flitter/language/vm.pyx":666
 *                 Py_INCREF(node)
 *                 PyTuple_SET_ITEM(nodes, i, node)
 *                 nodeptr = <PyObject*>node             # <<<<<<<<<<<<<<
 *             childrenptr = <PyObject*>((<Node>nodeptr)._children)
 *             n = o = PyTuple_GET_SIZE(<tuple>childrenptr)
 */
        __pyx_v_nodeptr = ((PyObject *)__pyx_v_node);

        /* "flitter/language/vm.pyx":661
 *             if type(<object>nodeptr) is not Node:
 *                 continue
 *             if nodeptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *                 node = (<Node>nodeptr).copy()
 *                 Py_DECREF(<Node>nodeptr)
 */
      }

      /* "flitter/language/vm.pyx":667
 *                 PyTuple_SET_ITEM(nodes, i, node)
 *                 nodeptr = <PyObject*>node
 *             childrenptr = <PyObject*>((<Node>nodeptr)._children)             # <<<<<<<<<<<<<<
 *             n = o = PyTuple_GET_SIZE(<tuple>childrenptr)
 *             for j in range(count-1, -1, -1):
 */
      __pyx_v_childrenptr = ((PyObject *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);

      /* "flitter/language/vm.pyx":668
 *                 nodeptr = <PyObject*>node
 *             childrenptr = <PyObject*>((<Node>nodeptr)._children)
 *             n = o = PyTuple_GET_SIZE(<tuple>childrenptr)             # <<<<<<<<<<<<<<
 *             for j in range(count-1, -1, -1):
 *                 n += peek_at(stack, j).length
 */
      __pyx_t_6 = PyTuple_GET_SIZE(((PyObject *)__pyx_v_childrenptr));
      __pyx_v_n = __pyx_t_6;
      __pyx_v_o = __pyx_t_6;

      /* "flitter/language/vm.pyx":669
 *             childrenptr = <PyObject*>((<Node>nodeptr)._children)
 *             n = o = PyTuple_GET_SIZE(<tuple>childrenptr)
 *             for j in range(count-1, -1, -1):             # <<<<<<<<<<<<<<
 *                 n += peek_at(stack, j).length
 *             dest = PyTuple_New(n)
 */
      for (__pyx_t_7 = (__pyx_v_count - 1); __pyx_t_7 > -1L; __pyx_t_7-=1) {
        __pyx_v_j = __pyx_t_7;

        /* "flitter/language/vm.pyx":670
 *             n = o = PyTuple_GET_SIZE(<tuple>childrenptr)
 *             for j in range(count-1, -1, -1):
 *                 n += peek_at(stack, j).length             # <<<<<<<<<<<<<<
 *             dest = PyTuple_New(n)
 *             for j in range(o):
 */
        __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_stack, __pyx_v_j)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_n = (__pyx_v_n + ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->length);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "flitter/language/vm.pyx":671
 *             for j in range(count-1, -1, -1):
 *                 n += peek_at(stack, j).length
 *             dest = PyTuple_New(n)             # <<<<<<<<<<<<<<
 *             for j in range(o):
 *                 objptr = PyTuple_GET_ITEM(<tuple>childrenptr, j)
 */
      __pyx_t_1 = PyTuple_New(__pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dest, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":672
 *                 n += peek_at(stack, j).length
 *             dest = PyTuple_New(n)
 *             for j in range(o):             # <<<<<<<<<<<<<<
 *                 objptr = PyTuple_GET_ITEM(<tuple>childrenptr, j)
 *                 Py_INCREF(<Node>objptr)
 */
      __pyx_t_7 = __pyx_v_o;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_j = __pyx_t_9;

        /* "flitter/language/vm.pyx":673
 *             dest = PyTuple_New(n)
 *             for j in range(o):
 *                 objptr = PyTuple_GET_ITEM(<tuple>childrenptr, j)             # <<<<<<<<<<<<<<
 *                 Py_INCREF(<Node>objptr)
 *                 PyTuple_SET_ITEM(dest, j, <Node>objptr)
 */
        __pyx_v_objptr = PyTuple_GET_ITEM(((PyObject *)__pyx_v_childrenptr), __pyx_v_j);

        /* "flitter/language/vm.pyx":674
 *             for j in range(o):
 *                 objptr = PyTuple_GET_ITEM(<tuple>childrenptr, j)
 *                 Py_INCREF(<Node>objptr)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(dest, j, <Node>objptr)
 *             for j in range(count-1, -1, -1):
 */
        Py_INCREF(((PyObject *)__pyx_v_objptr));

        /* "flitter/language/vm.pyx":675
 *                 objptr = PyTuple_GET_ITEM(<tuple>childrenptr, j)
 *                 Py_INCREF(<Node>objptr)
 *                 PyTuple_SET_ITEM(dest, j, <Node>objptr)             # <<<<<<<<<<<<<<
 *             for j in range(count-1, -1, -1):
 *                 children = peek_at(stack, j)
 */
        PyTuple_SET_ITEM(__pyx_v_dest, __pyx_v_j, ((PyObject *)__pyx_v_objptr));
      }

      /* "flitter/language/vm.pyx":676
 *                 Py_INCREF(<Node>objptr)
 *                 PyTuple_SET_ITEM(dest, j, <Node>objptr)
 *             for j in range(count-1, -1, -1):             # <<<<<<<<<<<<<<
 *                 children = peek_at(stack, j)
 *                 childrenptr = <PyObject*>children.objects
 */
      for (__pyx_t_7 = (__pyx_v_count - 1); __pyx_t_7 > -1L; __pyx_t_7-=1) {
        __pyx_v_j = __pyx_t_7;

        /* "flitter/language/vm.pyx":677
 *                 PyTuple_SET_ITEM(dest, j, <Node>objptr)
 *             for j in range(count-1, -1, -1):
 *                 children = peek_at(stack, j)             # <<<<<<<<<<<<<<
 *                 childrenptr = <PyObject*>children.objects
 *                 if childrenptr != <PyObject*>None:
 */
        __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_stack, __pyx_v_j)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_children, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":678
 *             for j in range(count-1, -1, -1):
 *                 children = peek_at(stack, j)
 *                 childrenptr = <PyObject*>children.objects             # <<<<<<<<<<<<<<
 *                 if childrenptr != <PyObject*>None:
 *                     for k in range(children.length):
 */
        __pyx_v_childrenptr = ((PyObject *)__pyx_v_children->objects);

        /* "flitter/language/vm.pyx":679
 *                 children = peek_at(stack, j)
 *                 childrenptr = <PyObject*>children.objects
 *                 if childrenptr != <PyObject*>None:             # <<<<<<<<<<<<<<
 *                     for k in range(children.length):
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 */
        __pyx_t_3 = (__pyx_v_childrenptr != ((PyObject *)Py_None));
        if (__pyx_t_3) {

          /* "flitter/language/vm.pyx":680
 *                 childrenptr = <PyObject*>children.objects
 *                 if childrenptr != <PyObject*>None:
 *                     for k in range(children.length):             # <<<<<<<<<<<<<<
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 *                         if type(<object>objptr) is Node:
 */
          __pyx_t_8 = __pyx_v_children->length;
          __pyx_t_9 = __pyx_t_8;
          for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
            __pyx_v_k = __pyx_t_10;

            /* "flitter/language/vm.pyx":681
 *                 if childrenptr != <PyObject*>None:
 *                     for k in range(children.length):
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)             # <<<<<<<<<<<<<<
 *                         if type(<object>objptr) is Node:
 *                             Py_INCREF(<Node>objptr)
 */
            __pyx_v_objptr = PyTuple_GET_ITEM(((PyObject *)__pyx_v_childrenptr), __pyx_v_k);

            /* "flitter/language/vm.pyx":682
 *                     for k in range(children.length):
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 *                         if type(<object>objptr) is Node:             # <<<<<<<<<<<<<<
 *                             Py_INCREF(<Node>objptr)
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 */
            __pyx_t_3 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_objptr))) == ((PyObject *)__pyx_ptype_7flitter_5model_Node));
            if (__pyx_t_3) {

              /* "flitter/language/vm.pyx":683
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 *                         if type(<object>objptr) is Node:
 *                             Py_INCREF(<Node>objptr)             # <<<<<<<<<<<<<<
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 *                             o += 1
 */
              Py_INCREF(((PyObject *)__pyx_v_objptr));

              /* "flitter/language/vm.pyx":684
 *                         if type(<object>objptr) is Node:
 *                             Py_INCREF(<Node>objptr)
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)             # <<<<<<<<<<<<<<
 *                             o += 1
 *             if o < n:
 */
              PyTuple_SET_ITEM(__pyx_v_dest, __pyx_v_o, ((PyObject *)__pyx_v_objptr));

              /* "flitter/language/vm.pyx":685
 *                             Py_INCREF(<Node>objptr)
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 *                             o += 1             # <<<<<<<<<<<<<<
 *             if o < n:
 *                 for k in range(o, n):
 */
              __pyx_v_o = (__pyx_v_o + 1);

              /* "flitter/language/vm.pyx":682
 *                     for k in range(children.length):
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 *                         if type(<object>objptr) is Node:             # <<<<<<<<<<<<<<
 *                             Py_INCREF(<Node>objptr)
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 */
            }
          }

          /* "flitter/language/vm.pyx":679
 *                 children = peek_at(stack, j)
 *                 childrenptr = <PyObject*>children.objects
 *                 if childrenptr != <PyObject*>None:             # <<<<<<<<<<<<<<
 *                     for k in range(children.length):
 *                         objptr = PyTuple_GET_ITEM(<tuple>childrenptr, k)
 */
        }
      }

      /* "flitter/language/vm.pyx":686
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 *                             o += 1
 *             if o < n:             # <<<<<<<<<<<<<<
 *                 for k in range(o, n):
 *                     Py_INCREF(None)
 */
      __pyx_t_3 = (__pyx_v_o < __pyx_v_n);
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":687
 *                             o += 1
 *             if o < n:
 *                 for k in range(o, n):             # <<<<<<<<<<<<<<
 *                     Py_INCREF(None)
 *                     PyTuple_SET_ITEM(dest, k, None)
 */
        __pyx_t_7 = __pyx_v_n;
        __pyx_t_8 = __pyx_t_7;
        for (__pyx_t_9 = __pyx_v_o; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
          __pyx_v_k = __pyx_t_9;

          /* "flitter/language/vm.pyx":688
 *             if o < n:
 *                 for k in range(o, n):
 *                     Py_INCREF(None)             # <<<<<<<<<<<<<<
 *                     PyTuple_SET_ITEM(dest, k, None)
 *                 if o:
 */
          Py_INCREF(Py_None);

          /* "flitter/language/vm.pyx":689
 *                 for k in range(o, n):
 *                     Py_INCREF(None)
 *                     PyTuple_SET_ITEM(dest, k, None)             # <<<<<<<<<<<<<<
 *                 if o:
 *                     (<Node>nodeptr)._children = PyTuple_GetSlice(dest, 0, o)
 */
          PyTuple_SET_ITEM(__pyx_v_dest, __pyx_v_k, Py_None);
        }

        /* "flitter/language/vm.pyx":690
 *                     Py_INCREF(None)
 *                     PyTuple_SET_ITEM(dest, k, None)
 *                 if o:             # <<<<<<<<<<<<<<
 *                     (<Node>nodeptr)._children = PyTuple_GetSlice(dest, 0, o)
 *                 else:
 */
        __pyx_t_3 = (__pyx_v_o != 0);
        if (__pyx_t_3) {

          /* "flitter/language/vm.pyx":691
 *                     PyTuple_SET_ITEM(dest, k, None)
 *                 if o:
 *                     (<Node>nodeptr)._children = PyTuple_GetSlice(dest, 0, o)             # <<<<<<<<<<<<<<
 *                 else:
 *                     (<Node>nodeptr)._children = ()
 */
          __pyx_t_1 = PyTuple_GetSlice(__pyx_v_dest, 0, __pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_1);
          __Pyx_GOTREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
          __Pyx_DECREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
          ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children = ((PyObject*)__pyx_t_1);
          __pyx_t_1 = 0;

          /* "flitter/language/vm.pyx":690
 *                     Py_INCREF(None)
 *                     PyTuple_SET_ITEM(dest, k, None)
 *                 if o:             # <<<<<<<<<<<<<<
 *                     (<Node>nodeptr)._children = PyTuple_GetSlice(dest, 0, o)
 *                 else:
 */
          goto __pyx_L21;
        }

        /* "flitter/language/vm.pyx":693
 *                     (<Node>nodeptr)._children = PyTuple_GetSlice(dest, 0, o)
 *                 else:
 *                     (<Node>nodeptr)._children = ()             # <<<<<<<<<<<<<<
 *             else:
 *                 (<Node>nodeptr)._children = dest
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_empty_tuple);
          __Pyx_GIVEREF(__pyx_empty_tuple);
          __Pyx_GOTREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
          __Pyx_DECREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
          ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children = __pyx_empty_tuple;
        }
        __pyx_L21:;

        /* "flitter/language/vm.pyx":686
 *                             PyTuple_SET_ITEM(dest, o, <Node>objptr)
 *                             o += 1
 *             if o < n:             # <<<<<<<<<<<<<<
 *                 for k in range(o, n):
 *                     Py_INCREF(None)
 */
        goto __pyx_L18;
      }

      /* "flitter/language/vm.pyx":695
 *                     (<Node>nodeptr)._children = ()
 *             else:
 *                 (<Node>nodeptr)._children = dest             # <<<<<<<<<<<<<<
 *     drop(stack, count)
 * 
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_dest);
        __Pyx_GIVEREF(__pyx_v_dest);
        __Pyx_GOTREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
        __Pyx_DECREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children);
        ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_children = __pyx_v_dest;
      }
      __pyx_L18:;
      __pyx_L4_continue:;
    }

    /* "flitter/language/vm.pyx":656
 *     cdef PyObject* objptr
 *     cdef tuple dest
 *     if nodes is not None:             # <<<<<<<<<<<<<<
 *         for i in range(m):
 *             nodeptr = PyTuple_GET_ITEM(nodes, i)
 */
  }

  /* "flitter/language/vm.pyx":696
 *             else:
 *                 (<Node>nodeptr)._children = dest
 *     drop(stack, count)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_7flitter_8language_2vm_drop(__pyx_v_stack, __pyx_v_count); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L1_error)

  /* "flitter/language/vm.pyx":647
 * 
 * 
 * cdef inline void execute_append(VectorStack stack, int64_t count):             # <<<<<<<<<<<<<<
 *     cdef Vector nodes_vec = peek_at(stack, count)
 *     cdef Vector children
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.execute_append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes_vec);
  __Pyx_XDECREF((PyObject *)__pyx_v_children);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_dest);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pyx":699
 * 
 * 
 * cdef inline execute_attributes(VectorStack stack, tuple names):             # <<<<<<<<<<<<<<
 *     cdef int64_t n = PyTuple_GET_SIZE(names)
 *     cdef Vector nodes_vec = peek_at(stack, n)
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_execute_attributes(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, PyObject *__pyx_v_names) {
  int64_t __pyx_v_n;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes_vec = 0;
  PyObject *__pyx_v_nodes = 0;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_copy = 0;
  PyObject *__pyx_v_attributes = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_j;
  int64_t __pyx_v_m;
  PyObject *__pyx_v_nodeptr;
  PyObject *__pyx_v_attrptr;
  PyObject *__pyx_v_nameptr;
  PyObject *__pyx_v_valptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  int64_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_attributes", 1);

  /* "flitter/language/vm.pyx":700
 * 
 * cdef inline execute_attributes(VectorStack stack, tuple names):
 *     cdef int64_t n = PyTuple_GET_SIZE(names)             # <<<<<<<<<<<<<<
 *     cdef Vector nodes_vec = peek_at(stack, n)
 *     cdef tuple nodes = nodes_vec.objects
 */
  __pyx_v_n = PyTuple_GET_SIZE(__pyx_v_names);

  /* "flitter/language/vm.pyx":701
 * cdef inline execute_attributes(VectorStack stack, tuple names):
 *     cdef int64_t n = PyTuple_GET_SIZE(names)
 *     cdef Vector nodes_vec = peek_at(stack, n)             # <<<<<<<<<<<<<<
 *     cdef tuple nodes = nodes_vec.objects
 *     if nodes is None:
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_stack, __pyx_v_n)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nodes_vec = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":702
 *     cdef int64_t n = PyTuple_GET_SIZE(names)
 *     cdef Vector nodes_vec = peek_at(stack, n)
 *     cdef tuple nodes = nodes_vec.objects             # <<<<<<<<<<<<<<
 *     if nodes is None:
 *         drop(stack, n)
 */
  __pyx_t_1 = __pyx_v_nodes_vec->objects;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":703
 *     cdef Vector nodes_vec = peek_at(stack, n)
 *     cdef tuple nodes = nodes_vec.objects
 *     if nodes is None:             # <<<<<<<<<<<<<<
 *         drop(stack, n)
 *         return
 */
  __pyx_t_2 = (__pyx_v_nodes == ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":704
 *     cdef tuple nodes = nodes_vec.objects
 *     if nodes is None:
 *         drop(stack, n)             # <<<<<<<<<<<<<<
 *         return
 *     cdef Node copy
 */
    __pyx_f_7flitter_8language_2vm_drop(__pyx_v_stack, __pyx_v_n); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L1_error)

    /* "flitter/language/vm.pyx":705
 *     if nodes is None:
 *         drop(stack, n)
 *         return             # <<<<<<<<<<<<<<
 *     cdef Node copy
 *     cdef dict attributes
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "flitter/language/vm.pyx":703
 *     cdef Vector nodes_vec = peek_at(stack, n)
 *     cdef tuple nodes = nodes_vec.objects
 *     if nodes is None:             # <<<<<<<<<<<<<<
 *         drop(stack, n)
 *         return
 */
  }

  /* "flitter/language/vm.pyx":708
 *     cdef Node copy
 *     cdef dict attributes
 *     cdef int64_t i, j, m=nodes_vec.length             # <<<<<<<<<<<<<<
 *     cdef PyObject* nodeptr
 *     cdef PyObject* attrptr
 */
  __pyx_t_3 = __pyx_v_nodes_vec->length;
  __pyx_v_m = __pyx_t_3;

  /* "flitter/language/vm.pyx":713
 *     cdef PyObject* nameptr
 *     cdef PyObject* valptr
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         nodeptr = PyTuple_GET_ITEM(nodes, i)
 *         if type(<object>nodeptr) is not Node:
 */
  __pyx_t_3 = __pyx_v_m;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "flitter/language/vm.pyx":714
 *     cdef PyObject* valptr
 *     for i in range(m):
 *         nodeptr = PyTuple_GET_ITEM(nodes, i)             # <<<<<<<<<<<<<<
 *         if type(<object>nodeptr) is not Node:
 *             continue
 */
    __pyx_v_nodeptr = PyTuple_GET_ITEM(__pyx_v_nodes, __pyx_v_i);

    /* "flitter/language/vm.pyx":715
 *     for i in range(m):
 *         nodeptr = PyTuple_GET_ITEM(nodes, i)
 *         if type(<object>nodeptr) is not Node:             # <<<<<<<<<<<<<<
 *             continue
 *         if nodeptr.ob_refcnt > 1:
 */
    __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_nodeptr))) != ((PyObject *)__pyx_ptype_7flitter_5model_Node));
    if (__pyx_t_2) {

      /* "flitter/language/vm.pyx":716
 *         nodeptr = PyTuple_GET_ITEM(nodes, i)
 *         if type(<object>nodeptr) is not Node:
 *             continue             # <<<<<<<<<<<<<<
 *         if nodeptr.ob_refcnt > 1:
 *             copy = (<Node>nodeptr).copy()
 */
      goto __pyx_L4_continue;

      /* "flitter/language/vm.pyx":715
 *     for i in range(m):
 *         nodeptr = PyTuple_GET_ITEM(nodes, i)
 *         if type(<object>nodeptr) is not Node:             # <<<<<<<<<<<<<<
 *             continue
 *         if nodeptr.ob_refcnt > 1:
 */
    }

    /* "flitter/language/vm.pyx":717
 *         if type(<object>nodeptr) is not Node:
 *             continue
 *         if nodeptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *             copy = (<Node>nodeptr).copy()
 *             Py_INCREF(copy)
 */
    __pyx_t_2 = (__pyx_v_nodeptr->ob_refcnt > 1);
    if (__pyx_t_2) {

      /* "flitter/language/vm.pyx":718
 *             continue
 *         if nodeptr.ob_refcnt > 1:
 *             copy = (<Node>nodeptr).copy()             # <<<<<<<<<<<<<<
 *             Py_INCREF(copy)
 *             PyTuple_SET_ITEM(nodes, i, copy)
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_copy, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":719
 *         if nodeptr.ob_refcnt > 1:
 *             copy = (<Node>nodeptr).copy()
 *             Py_INCREF(copy)             # <<<<<<<<<<<<<<
 *             PyTuple_SET_ITEM(nodes, i, copy)
 *             Py_DECREF(<Node>nodeptr)
 */
      Py_INCREF(((PyObject *)__pyx_v_copy));

      /* "flitter/language/vm.pyx":720
 *             copy = (<Node>nodeptr).copy()
 *             Py_INCREF(copy)
 *             PyTuple_SET_ITEM(nodes, i, copy)             # <<<<<<<<<<<<<<
 *             Py_DECREF(<Node>nodeptr)
 *             nodeptr = <PyObject*>copy
 */
      PyTuple_SET_ITEM(__pyx_v_nodes, __pyx_v_i, ((PyObject *)__pyx_v_copy));

      /* "flitter/language/vm.pyx":721
 *             Py_INCREF(copy)
 *             PyTuple_SET_ITEM(nodes, i, copy)
 *             Py_DECREF(<Node>nodeptr)             # <<<<<<<<<<<<<<
 *             nodeptr = <PyObject*>copy
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes
 */
      Py_DECREF(((PyObject *)__pyx_v_nodeptr));

      /* "flitter/language/vm.pyx":722
 *             PyTuple_SET_ITEM(nodes, i, copy)
 *             Py_DECREF(<Node>nodeptr)
 *             nodeptr = <PyObject*>copy             # <<<<<<<<<<<<<<
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         if (<Node>nodeptr)._attributes_shared:
 */
      __pyx_v_nodeptr = ((PyObject *)__pyx_v_copy);

      /* "flitter/language/vm.pyx":717
 *         if type(<object>nodeptr) is not Node:
 *             continue
 *         if nodeptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *             copy = (<Node>nodeptr).copy()
 *             Py_INCREF(copy)
 */
    }

    /* "flitter/language/vm.pyx":723
 *             Py_DECREF(<Node>nodeptr)
 *             nodeptr = <PyObject*>copy
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes             # <<<<<<<<<<<<<<
 *         if (<Node>nodeptr)._attributes_shared:
 *             attributes = PyDict_Copy(<object>attrptr)
 */
    __pyx_v_attrptr = ((PyObject *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);

    /* "flitter/language/vm.pyx":724
 *             nodeptr = <PyObject*>copy
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         if (<Node>nodeptr)._attributes_shared:             # <<<<<<<<<<<<<<
 *             attributes = PyDict_Copy(<object>attrptr)
 *             (<Node>nodeptr)._attributes = attributes
 */
    if (((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes_shared) {

      /* "flitter/language/vm.pyx":725
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         if (<Node>nodeptr)._attributes_shared:
 *             attributes = PyDict_Copy(<object>attrptr)             # <<<<<<<<<<<<<<
 *             (<Node>nodeptr)._attributes = attributes
 *             (<Node>nodeptr)._attributes_shared = False
 */
      __pyx_t_1 = PyDict_Copy(((PyObject *)__pyx_v_attrptr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_attributes, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":726
 *         if (<Node>nodeptr)._attributes_shared:
 *             attributes = PyDict_Copy(<object>attrptr)
 *             (<Node>nodeptr)._attributes = attributes             # <<<<<<<<<<<<<<
 *             (<Node>nodeptr)._attributes_shared = False
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes
 */
      __Pyx_INCREF(__pyx_v_attributes);
      __Pyx_GIVEREF(__pyx_v_attributes);
      __Pyx_GOTREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);
      __Pyx_DECREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);
      ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes = __pyx_v_attributes;

      /* "flitter/language/vm.pyx":727
 *             attributes = PyDict_Copy(<object>attrptr)
 *             (<Node>nodeptr)._attributes = attributes
 *             (<Node>nodeptr)._attributes_shared = False             # <<<<<<<<<<<<<<
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         elif attrptr is <PyObject*>None:
 */
      ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes_shared = 0;

      /* "flitter/language/vm.pyx":728
 *             (<Node>nodeptr)._attributes = attributes
 *             (<Node>nodeptr)._attributes_shared = False
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes             # <<<<<<<<<<<<<<
 *         elif attrptr is <PyObject*>None:
 *             attributes = PyDict_New()
 */
      __pyx_v_attrptr = ((PyObject *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);

      /* "flitter/language/vm.pyx":724
 *             nodeptr = <PyObject*>copy
 *         attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         if (<Node>nodeptr)._attributes_shared:             # <<<<<<<<<<<<<<
 *             attributes = PyDict_Copy(<object>attrptr)
 *             (<Node>nodeptr)._attributes = attributes
 */
      goto __pyx_L8;
    }

    /* "flitter/language/vm.pyx":729
 *             (<Node>nodeptr)._attributes_shared = False
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         elif attrptr is <PyObject*>None:             # <<<<<<<<<<<<<<
 *             attributes = PyDict_New()
 *             attrptr = <PyObject*>attributes
 */
    __pyx_t_2 = (__pyx_v_attrptr == ((PyObject *)Py_None));
    if (__pyx_t_2) {

      /* "flitter/language/vm.pyx":730
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         elif attrptr is <PyObject*>None:
 *             attributes = PyDict_New()             # <<<<<<<<<<<<<<
 *             attrptr = <PyObject*>attributes
 *             (<Node>nodeptr)._attributes = attributes
 */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_attributes, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":731
 *         elif attrptr is <PyObject*>None:
 *             attributes = PyDict_New()
 *             attrptr = <PyObject*>attributes             # <<<<<<<<<<<<<<
 *             (<Node>nodeptr)._attributes = attributes
 *         for j in range(n):
 */
      __pyx_v_attrptr = ((PyObject *)__pyx_v_attributes);

      /* "flitter/language/vm.pyx":732
 *             attributes = PyDict_New()
 *             attrptr = <PyObject*>attributes
 *             (<Node>nodeptr)._attributes = attributes             # <<<<<<<<<<<<<<
 *         for j in range(n):
 *             valptr = stack.vectors[stack.top-(n-1-j)]
 */
      __Pyx_INCREF(__pyx_v_attributes);
      __Pyx_GIVEREF(__pyx_v_attributes);
      __Pyx_GOTREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);
      __Pyx_DECREF(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes);
      ((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_nodeptr)->_attributes = __pyx_v_attributes;

      /* "flitter/language/vm.pyx":729
 *             (<Node>nodeptr)._attributes_shared = False
 *             attrptr = <PyObject*>(<Node>nodeptr)._attributes
 *         elif attrptr is <PyObject*>None:             # <<<<<<<<<<<<<<
 *             attributes = PyDict_New()
 *             attrptr = <PyObject*>attributes
 */
    }
    __pyx_L8:;

    /* "flitter/language/vm.pyx":733
 *             attrptr = <PyObject*>attributes
 *             (<Node>nodeptr)._attributes = attributes
 *         for j in range(n):             # <<<<<<<<<<<<<<
 *             valptr = stack.vectors[stack.top-(n-1-j)]
 *             nameptr = PyTuple_GET_ITEM(names, j)
 */
    __pyx_t_6 = __pyx_v_n;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_j = __pyx_t_8;

      /* "flitter/language/vm.pyx":734
 *             (<Node>nodeptr)._attributes = attributes
 *         for j in range(n):
 *             valptr = stack.vectors[stack.top-(n-1-j)]             # <<<<<<<<<<<<<<
 *             nameptr = PyTuple_GET_ITEM(names, j)
 *             if (<Vector>valptr).length:
 */
      __pyx_v_valptr = (__pyx_v_stack->vectors[(__pyx_v_stack->top - ((__pyx_v_n - 1) - __pyx_v_j))]);

      /* "flitter/language/vm.pyx":735
 *         for j in range(n):
 *             valptr = stack.vectors[stack.top-(n-1-j)]
 *             nameptr = PyTuple_GET_ITEM(names, j)             # <<<<<<<<<<<<<<
 *             if (<Vector>valptr).length:
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 */
      __pyx_v_nameptr = PyTuple_GET_ITEM(__pyx_v_names, __pyx_v_j);

      /* "flitter/language/vm.pyx":736
 *             valptr = stack.vectors[stack.top-(n-1-j)]
 *             nameptr = PyTuple_GET_ITEM(names, j)
 *             if (<Vector>valptr).length:             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:
 */
      __pyx_t_2 = (((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_valptr)->length != 0);
      if (__pyx_t_2) {

        /* "flitter/language/vm.pyx":737
 *             nameptr = PyTuple_GET_ITEM(names, j)
 *             if (<Vector>valptr).length:
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)             # <<<<<<<<<<<<<<
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:
 *                 PyDict_DelItem(<object>attrptr, <object>nameptr)
 */
        __pyx_t_9 = PyDict_SetItem(((PyObject *)__pyx_v_attrptr), ((PyObject *)__pyx_v_nameptr), ((PyObject *)__pyx_v_valptr)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 737, __pyx_L1_error)

        /* "flitter/language/vm.pyx":736
 *             valptr = stack.vectors[stack.top-(n-1-j)]
 *             nameptr = PyTuple_GET_ITEM(names, j)
 *             if (<Vector>valptr).length:             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:
 */
        goto __pyx_L11;
      }

      /* "flitter/language/vm.pyx":738
 *             if (<Vector>valptr).length:
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_DelItem(<object>attrptr, <object>nameptr)
 *     drop(stack, n)
 */
      __pyx_t_2 = (PyDict_GetItem(((PyObject *)__pyx_v_attrptr), ((PyObject *)__pyx_v_nameptr)) != NULL);
      if (__pyx_t_2) {

        /* "flitter/language/vm.pyx":739
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:
 *                 PyDict_DelItem(<object>attrptr, <object>nameptr)             # <<<<<<<<<<<<<<
 *     drop(stack, n)
 * 
 */
        __pyx_t_9 = PyDict_DelItem(((PyObject *)__pyx_v_attrptr), ((PyObject *)__pyx_v_nameptr)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 739, __pyx_L1_error)

        /* "flitter/language/vm.pyx":738
 *             if (<Vector>valptr).length:
 *                 PyDict_SetItem(<object>attrptr, <object>nameptr, <Vector>valptr)
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:             # <<<<<<<<<<<<<<
 *                 PyDict_DelItem(<object>attrptr, <object>nameptr)
 *     drop(stack, n)
 */
      }
      __pyx_L11:;
    }
    __pyx_L4_continue:;
  }

  /* "flitter/language/vm.pyx":740
 *             elif PyDict_GetItem(<object>attrptr, <object>nameptr) != NULL:
 *                 PyDict_DelItem(<object>attrptr, <object>nameptr)
 *     drop(stack, n)             # <<<<<<<<<<<<<<
 * 
 * cdef inline execute_tag(VectorStack stack, str name):
 */
  __pyx_f_7flitter_8language_2vm_drop(__pyx_v_stack, __pyx_v_n); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L1_error)

  /* "flitter/language/vm.pyx":699
 * 
 * 
 * cdef inline execute_attributes(VectorStack stack, tuple names):             # <<<<<<<<<<<<<<
 *     cdef int64_t n = PyTuple_GET_SIZE(names)
 *     cdef Vector nodes_vec = peek_at(stack, n)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.execute_attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes_vec);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_copy);
  __Pyx_XDECREF(__pyx_v_attributes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":742
 *     drop(stack, n)
 * 
 * cdef inline execute_tag(VectorStack stack, str name):             # <<<<<<<<<<<<<<
 *     cdef Vector nodes_vec = peek(stack)
 *     cdef int64_t i, m=nodes_vec.length
 */

static CYTHON_INLINE PyObject *__pyx_f_7flitter_8language_2vm_execute_tag(struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack, PyObject *__pyx_v_name) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes_vec = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_m;
  PyObject *__pyx_v_nodes = 0;
  PyObject *__pyx_v_objptr;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_tag", 1);

  /* "flitter/language/vm.pyx":743
 * 
 * cdef inline execute_tag(VectorStack stack, str name):
 *     cdef Vector nodes_vec = peek(stack)             # <<<<<<<<<<<<<<
 *     cdef int64_t i, m=nodes_vec.length
 *     cdef tuple nodes = nodes_vec.objects
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nodes_vec = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":744
 * cdef inline execute_tag(VectorStack stack, str name):
 *     cdef Vector nodes_vec = peek(stack)
 *     cdef int64_t i, m=nodes_vec.length             # <<<<<<<<<<<<<<
 *     cdef tuple nodes = nodes_vec.objects
 *     cdef PyObject* objptr
 */
  __pyx_t_2 = __pyx_v_nodes_vec->length;
  __pyx_v_m = __pyx_t_2;

  /* "flitter/language/vm.pyx":745
 *     cdef Vector nodes_vec = peek(stack)
 *     cdef int64_t i, m=nodes_vec.length
 *     cdef tuple nodes = nodes_vec.objects             # <<<<<<<<<<<<<<
 *     cdef PyObject* objptr
 *     cdef Node node
 */
  __pyx_t_1 = __pyx_v_nodes_vec->objects;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_nodes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":748
 *     cdef PyObject* objptr
 *     cdef Node node
 *     if nodes is not None:             # <<<<<<<<<<<<<<
 *         for i in range(m):
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 */
  __pyx_t_3 = (__pyx_v_nodes != ((PyObject*)Py_None));
  if (__pyx_t_3) {

    /* "flitter/language/vm.pyx":749
 *     cdef Node node
 *     if nodes is not None:
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>objptr) is not Node:
 */
    __pyx_t_2 = __pyx_v_m;
    __pyx_t_4 = __pyx_t_2;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/language/vm.pyx":750
 *     if nodes is not None:
 *         for i in range(m):
 *             objptr = PyTuple_GET_ITEM(nodes, i)             # <<<<<<<<<<<<<<
 *             if type(<object>objptr) is not Node:
 *                 continue
 */
      __pyx_v_objptr = PyTuple_GET_ITEM(__pyx_v_nodes, __pyx_v_i);

      /* "flitter/language/vm.pyx":751
 *         for i in range(m):
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>objptr) is not Node:             # <<<<<<<<<<<<<<
 *                 continue
 *             if objptr.ob_refcnt > 1:
 */
      __pyx_t_3 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_objptr))) != ((PyObject *)__pyx_ptype_7flitter_5model_Node));
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":752
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>objptr) is not Node:
 *                 continue             # <<<<<<<<<<<<<<
 *             if objptr.ob_refcnt > 1:
 *                 node = (<Node>objptr).copy()
 */
        goto __pyx_L4_continue;

        /* "flitter/language/vm.pyx":751
 *         for i in range(m):
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 *             if type(<object>objptr) is not Node:             # <<<<<<<<<<<<<<
 *                 continue
 *             if objptr.ob_refcnt > 1:
 */
      }

      /* "flitter/language/vm.pyx":753
 *             if type(<object>objptr) is not Node:
 *                 continue
 *             if objptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *                 node = (<Node>objptr).copy()
 *                 Py_DECREF(<Node>objptr)
 */
      __pyx_t_3 = (__pyx_v_objptr->ob_refcnt > 1);
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":754
 *                 continue
 *             if objptr.ob_refcnt > 1:
 *                 node = (<Node>objptr).copy()             # <<<<<<<<<<<<<<
 *                 Py_DECREF(<Node>objptr)
 *                 Py_INCREF(node)
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_objptr)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_objptr), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":755
 *             if objptr.ob_refcnt > 1:
 *                 node = (<Node>objptr).copy()
 *                 Py_DECREF(<Node>objptr)             # <<<<<<<<<<<<<<
 *                 Py_INCREF(node)
 *                 PyTuple_SET_ITEM(nodes, i, node)
 */
        Py_DECREF(((PyObject *)__pyx_v_objptr));

        /* "flitter/language/vm.pyx":756
 *                 node = (<Node>objptr).copy()
 *                 Py_DECREF(<Node>objptr)
 *                 Py_INCREF(node)             # <<<<<<<<<<<<<<
 *                 PyTuple_SET_ITEM(nodes, i, node)
 *             else:
 */
        Py_INCREF(((PyObject *)__pyx_v_node));

        /* "flitter/language/vm.pyx":757
 *                 Py_DECREF(<Node>objptr)
 *                 Py_INCREF(node)
 *                 PyTuple_SET_ITEM(nodes, i, node)             # <<<<<<<<<<<<<<
 *             else:
 *                 node = <Node>objptr
 */
        PyTuple_SET_ITEM(__pyx_v_nodes, __pyx_v_i, ((PyObject *)__pyx_v_node));

        /* "flitter/language/vm.pyx":753
 *             if type(<object>objptr) is not Node:
 *                 continue
 *             if objptr.ob_refcnt > 1:             # <<<<<<<<<<<<<<
 *                 node = (<Node>objptr).copy()
 *                 Py_DECREF(<Node>objptr)
 */
        goto __pyx_L7;
      }

      /* "flitter/language/vm.pyx":759
 *                 PyTuple_SET_ITEM(nodes, i, node)
 *             else:
 *                 node = <Node>objptr             # <<<<<<<<<<<<<<
 *             if (<Node>node)._tags is None:
 *                 (<Node>node)._tags = set()
 */
      /*else*/ {
        __pyx_t_1 = ((PyObject *)__pyx_v_objptr);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_1));
        __pyx_t_1 = 0;
      }
      __pyx_L7:;

      /* "flitter/language/vm.pyx":760
 *             else:
 *                 node = <Node>objptr
 *             if (<Node>node)._tags is None:             # <<<<<<<<<<<<<<
 *                 (<Node>node)._tags = set()
 *             PySet_Add((<Node>node)._tags, name)
 */
      __pyx_t_3 = (__pyx_v_node->_tags == ((PyObject*)Py_None));
      if (__pyx_t_3) {

        /* "flitter/language/vm.pyx":761
 *                 node = <Node>objptr
 *             if (<Node>node)._tags is None:
 *                 (<Node>node)._tags = set()             # <<<<<<<<<<<<<<
 *             PySet_Add((<Node>node)._tags, name)
 * 
 */
        __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_v_node->_tags);
        __Pyx_DECREF(__pyx_v_node->_tags);
        __pyx_v_node->_tags = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":760
 *             else:
 *                 node = <Node>objptr
 *             if (<Node>node)._tags is None:             # <<<<<<<<<<<<<<
 *                 (<Node>node)._tags = set()
 *             PySet_Add((<Node>node)._tags, name)
 */
      }

      /* "flitter/language/vm.pyx":762
 *             if (<Node>node)._tags is None:
 *                 (<Node>node)._tags = set()
 *             PySet_Add((<Node>node)._tags, name)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_1 = __pyx_v_node->_tags;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = PySet_Add(__pyx_t_1, __pyx_v_name); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 762, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_L4_continue:;
    }

    /* "flitter/language/vm.pyx":748
 *     cdef PyObject* objptr
 *     cdef Node node
 *     if nodes is not None:             # <<<<<<<<<<<<<<
 *         for i in range(m):
 *             objptr = PyTuple_GET_ITEM(nodes, i)
 */
  }

  /* "flitter/language/vm.pyx":742
 *     drop(stack, n)
 * 
 * cdef inline execute_tag(VectorStack stack, str name):             # <<<<<<<<<<<<<<
 *     cdef Vector nodes_vec = peek(stack)
 *     cdef int64_t i, m=nodes_vec.length
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.execute_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes_vec);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":766
 * 
 * cdef class Program:
 *     def __cinit__(self, lnames=()):             # <<<<<<<<<<<<<<
 *         self.pragmas = {}
 *         self.instructions = []
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_2vm_7Program_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_2vm_7Program_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_lnames = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lnames,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_empty_tuple));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lnames);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 766, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lnames = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 766, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program___cinit__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_lnames);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_2vm_7Program___cinit__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_lnames) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/language/vm.pyx":767
 * cdef class Program:
 *     def __cinit__(self, lnames=()):
 *         self.pragmas = {}             # <<<<<<<<<<<<<<
 *         self.instructions = []
 *         self.initial_lnames = lnames
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pragmas);
  __Pyx_DECREF(__pyx_v_self->pragmas);
  __pyx_v_self->pragmas = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":768
 *     def __cinit__(self, lnames=()):
 *         self.pragmas = {}
 *         self.instructions = []             # <<<<<<<<<<<<<<
 *         self.initial_lnames = lnames
 *         self.linked = False
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->instructions);
  __Pyx_DECREF(__pyx_v_self->instructions);
  __pyx_v_self->instructions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":769
 *         self.pragmas = {}
 *         self.instructions = []
 *         self.initial_lnames = lnames             # <<<<<<<<<<<<<<
 *         self.linked = False
 *         self.next_label = 1
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_lnames))||((__pyx_v_lnames) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_lnames))) __PYX_ERR(0, 769, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_lnames;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->initial_lnames);
  __Pyx_DECREF(__pyx_v_self->initial_lnames);
  __pyx_v_self->initial_lnames = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":770
 *         self.instructions = []
 *         self.initial_lnames = lnames
 *         self.linked = False             # <<<<<<<<<<<<<<
 *         self.next_label = 1
 *         self.simplify = True
 */
  __pyx_v_self->linked = 0;

  /* "flitter/language/vm.pyx":771
 *         self.initial_lnames = lnames
 *         self.linked = False
 *         self.next_label = 1             # <<<<<<<<<<<<<<
 *         self.simplify = True
 *         self.compiler_errors = set()
 */
  __pyx_v_self->next_label = 1;

  /* "flitter/language/vm.pyx":772
 *         self.linked = False
 *         self.next_label = 1
 *         self.simplify = True             # <<<<<<<<<<<<<<
 *         self.compiler_errors = set()
 * 
 */
  __pyx_v_self->simplify = 1;

  /* "flitter/language/vm.pyx":773
 *         self.next_label = 1
 *         self.simplify = True
 *         self.compiler_errors = set()             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->compiler_errors);
  __Pyx_DECREF(__pyx_v_self->compiler_errors);
  __pyx_v_self->compiler_errors = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":766
 * 
 * cdef class Program:
 *     def __cinit__(self, lnames=()):             # <<<<<<<<<<<<<<
 *         self.pragmas = {}
 *         self.instructions = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":775
 *         self.compiler_errors = set()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.instructions)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_7flitter_8language_2vm_7Program_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_7flitter_8language_2vm_7Program_3__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_2__len__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_7flitter_8language_2vm_7Program_2__len__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 1);

  /* "flitter/language/vm.pyx":776
 * 
 *     def __len__(self):
 *         return len(self.instructions)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_1 = __pyx_v_self->instructions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 776, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":775
 *         self.compiler_errors = set()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.instructions)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":778
 *         return len(self.instructions)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '\n'.join(str(instruction) for instruction in self.instructions)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_5__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_4__str__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7flitter_8language_2vm_7Program_7__str___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/language/vm.pyx":779
 * 
 *     def __str__(self):
 *         return '\n'.join(str(instruction) for instruction in self.instructions)             # <<<<<<<<<<<<<<
 * 
 *     cdef Instruction last_instruction(self):
 */

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct__genexpr(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 779, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_8language_2vm_7Program_7__str___2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_str___locals_genexpr, __pyx_n_s_flitter_language_vm); if (unlikely(!gen)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_8language_2vm_7Program_7__str___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 779, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 779, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 779, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 779, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 779, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_instruction);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_instruction, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_cur_scope->__pyx_v_instruction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":778
 *         return len(self.instructions)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '\n'.join(str(instruction) for instruction in self.instructions)
 * 
 */

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_4__str__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_gb_7flitter_8language_2vm_7Program_7__str___2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "flitter/language/vm.pyx":779
 * 
 *     def __str__(self):
 *         return '\n'.join(str(instruction) for instruction in self.instructions)             # <<<<<<<<<<<<<<
 * 
 *     cdef Instruction last_instruction(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_7flitter_8language_2vm_7Program_7__str___genexpr(NULL, __pyx_v_self->instructions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__8, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":778
 *         return len(self.instructions)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '\n'.join(str(instruction) for instruction in self.instructions)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.vm.Program.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_7flitter_8language_2vm_7Program_7__str___2generator);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":781
 *         return '\n'.join(str(instruction) for instruction in self.instructions)
 * 
 *     cdef Instruction last_instruction(self):             # <<<<<<<<<<<<<<
 *         cdef int64_t n=len(self.instructions)
 *         return self.instructions[n-1]
 */

static struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_f_7flitter_8language_2vm_7Program_last_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  int64_t __pyx_v_n;
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("last_instruction", 1);

  /* "flitter/language/vm.pyx":782
 * 
 *     cdef Instruction last_instruction(self):
 *         cdef int64_t n=len(self.instructions)             # <<<<<<<<<<<<<<
 *         return self.instructions[n-1]
 * 
 */
  __pyx_t_1 = __pyx_v_self->instructions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 782, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n = __pyx_t_2;

  /* "flitter/language/vm.pyx":783
 *     cdef Instruction last_instruction(self):
 *         cdef int64_t n=len(self.instructions)
 *         return self.instructions[n-1]             # <<<<<<<<<<<<<<
 * 
 *     cdef Instruction pop_instruction(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 783, __pyx_L1_error)
  }
  __pyx_t_3 = (__pyx_v_n - 1);
  if (!(likely(((PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_t_3)) == Py_None) || likely(__Pyx_TypeTest(PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_t_3), __pyx_ptype_7flitter_8language_2vm_Instruction))))) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_t_3));
  __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_t_3));
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":781
 *         return '\n'.join(str(instruction) for instruction in self.instructions)
 * 
 *     cdef Instruction last_instruction(self):             # <<<<<<<<<<<<<<
 *         cdef int64_t n=len(self.instructions)
 *         return self.instructions[n-1]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.last_instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":785
 *         return self.instructions[n-1]
 * 
 *     cdef Instruction pop_instruction(self):             # <<<<<<<<<<<<<<
 *         return self.instructions.pop()
 * 
 */

static struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_f_7flitter_8language_2vm_7Program_pop_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_instruction", 1);

  /* "flitter/language/vm.pyx":786
 * 
 *     cdef Instruction pop_instruction(self):
 *         return self.instructions.pop()             # <<<<<<<<<<<<<<
 * 
 *     cdef Program push_instruction(self, Instruction instruction):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 786, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_self->instructions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_2vm_Instruction))))) __PYX_ERR(0, 786, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":785
 *         return self.instructions[n-1]
 * 
 *     cdef Instruction pop_instruction(self):             # <<<<<<<<<<<<<<
 *         return self.instructions.pop()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.pop_instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":788
 *         return self.instructions.pop()
 * 
 *     cdef Program push_instruction(self, Instruction instruction):             # <<<<<<<<<<<<<<
 *         self.instructions.append(instruction)
 *         return self
 */

static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_push_instruction(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instruction) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("push_instruction", 1);

  /* "flitter/language/vm.pyx":789
 * 
 *     cdef Program push_instruction(self, Instruction instruction):
 *         self.instructions.append(instruction)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 789, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_Append(__pyx_v_self->instructions, ((PyObject *)__pyx_v_instruction)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "flitter/language/vm.pyx":790
 *     cdef Program push_instruction(self, Instruction instruction):
 *         self.instructions.append(instruction)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def set_path(self, object path):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":788
 *         return self.instructions.pop()
 * 
 *     cdef Program push_instruction(self, Instruction instruction):             # <<<<<<<<<<<<<<
 *         self.instructions.append(instruction)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program.push_instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":792
 *         return self
 * 
 *     def set_path(self, object path):             # <<<<<<<<<<<<<<
 *         self.path = path
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_7set_path(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_7set_path = {"set_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_7set_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_7set_path(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_path = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_path (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 792, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_path") < 0)) __PYX_ERR(0, 792, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_path = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_path", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 792, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.set_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_6set_path(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_path);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_6set_path(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_path", 1);

  /* "flitter/language/vm.pyx":793
 * 
 *     def set_path(self, object path):
 *         self.path = path             # <<<<<<<<<<<<<<
 * 
 *     def set_top(self, object top):
 */
  __Pyx_INCREF(__pyx_v_path);
  __Pyx_GIVEREF(__pyx_v_path);
  __Pyx_GOTREF(__pyx_v_self->path);
  __Pyx_DECREF(__pyx_v_self->path);
  __pyx_v_self->path = __pyx_v_path;

  /* "flitter/language/vm.pyx":792
 *         return self
 * 
 *     def set_path(self, object path):             # <<<<<<<<<<<<<<
 *         self.path = path
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":795
 *         self.path = path
 * 
 *     def set_top(self, object top):             # <<<<<<<<<<<<<<
 *         self.top = top
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_9set_top(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_9set_top = {"set_top", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_9set_top, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_9set_top(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_top = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_top (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_top,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_top)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 795, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_top") < 0)) __PYX_ERR(0, 795, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_top = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_top", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 795, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.set_top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_8set_top(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_top);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8set_top(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_top) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_top", 1);

  /* "flitter/language/vm.pyx":796
 * 
 *     def set_top(self, object top):
 *         self.top = top             # <<<<<<<<<<<<<<
 * 
 *     def set_pragma(self, str name, Vector value):
 */
  __Pyx_INCREF(__pyx_v_top);
  __Pyx_GIVEREF(__pyx_v_top);
  __Pyx_GOTREF(__pyx_v_self->top);
  __Pyx_DECREF(__pyx_v_self->top);
  __pyx_v_self->top = __pyx_v_top;

  /* "flitter/language/vm.pyx":795
 *         self.path = path
 * 
 *     def set_top(self, object top):             # <<<<<<<<<<<<<<
 *         self.top = top
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":798
 *         self.top = top
 * 
 *     def set_pragma(self, str name, Vector value):             # <<<<<<<<<<<<<<
 *         self.pragmas[name] = value
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_11set_pragma(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_11set_pragma = {"set_pragma", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_11set_pragma, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_11set_pragma(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pragma (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_pragma", 1, 2, 2, 1); __PYX_ERR(0, 798, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_pragma") < 0)) __PYX_ERR(0, 798, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Vector *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_pragma", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 798, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.set_pragma", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 798, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_7flitter_5model_Vector, 1, "value", 0))) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_10set_pragma(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_name, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_10set_pragma(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_pragma", 1);

  /* "flitter/language/vm.pyx":799
 * 
 *     def set_pragma(self, str name, Vector value):
 *         self.pragmas[name] = value             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->pragmas == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 799, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->pragmas, __pyx_v_name, ((PyObject *)__pyx_v_value)) < 0))) __PYX_ERR(0, 799, __pyx_L1_error)

  /* "flitter/language/vm.pyx":800
 *     def set_pragma(self, str name, Vector value):
 *         self.pragmas[name] = value
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":798
 *         self.top = top
 * 
 *     def set_pragma(self, str name, Vector value):             # <<<<<<<<<<<<<<
 *         self.pragmas[name] = value
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program.set_pragma", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":802
 *         return self
 * 
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_13execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7flitter_8language_2vm_7Program_12execute, "This is a test-harness function. Do not use.");
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_13execute = {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_13execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7flitter_8language_2vm_7Program_12execute};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_13execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_context = 0;
  PyObject *__pyx_v_lnames = 0;
  int __pyx_v_record_stats;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_lnames,&__pyx_n_s_record_stats,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 802, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lnames);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 802, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record_stats);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 802, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "execute") < 0)) __PYX_ERR(0, 802, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_7flitter_5model_Context *)values[0]);
    __pyx_v_lnames = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_record_stats = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_record_stats == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 802, __pyx_L3_error)
    } else {
      __pyx_v_record_stats = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 802, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_7flitter_5model_Context, 1, "context", 0))) __PYX_ERR(0, 802, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lnames), (&PyList_Type), 1, "lnames", 1))) __PYX_ERR(0, 802, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_12execute(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_context, __pyx_v_lnames, __pyx_v_record_stats);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_12execute(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, PyObject *__pyx_v_lnames, int __pyx_v_record_stats) {
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute", 1);

  /* "flitter/language/vm.pyx":804
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()             # <<<<<<<<<<<<<<
 *         if not self.linked:
 *             self.link()
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->initial_lnames, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 804, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 804, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":805
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 *         if not self.linked:             # <<<<<<<<<<<<<<
 *             self.link()
 *         if context.lnames is None:
 */
  __pyx_t_2 = (!__pyx_v_self->linked);
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":806
 *         assert self.initial_lnames == ()
 *         if not self.linked:
 *             self.link()             # <<<<<<<<<<<<<<
 *         if context.lnames is None:
 *             context.lnames = VectorStack()
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_self->__pyx_vtab)->link(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":805
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 *         if not self.linked:             # <<<<<<<<<<<<<<
 *             self.link()
 *         if context.lnames is None:
 */
  }

  /* "flitter/language/vm.pyx":807
 *         if not self.linked:
 *             self.link()
 *         if context.lnames is None:             # <<<<<<<<<<<<<<
 *             context.lnames = VectorStack()
 *         if lnames:
 */
  __pyx_t_2 = (__pyx_v_context->lnames == Py_None);
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":808
 *             self.link()
 *         if context.lnames is None:
 *             context.lnames = VectorStack()             # <<<<<<<<<<<<<<
 *         if lnames:
 *             for item in lnames:
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_context->lnames);
    __Pyx_DECREF(__pyx_v_context->lnames);
    __pyx_v_context->lnames = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":807
 *         if not self.linked:
 *             self.link()
 *         if context.lnames is None:             # <<<<<<<<<<<<<<
 *             context.lnames = VectorStack()
 *         if lnames:
 */
  }

  /* "flitter/language/vm.pyx":809
 *         if context.lnames is None:
 *             context.lnames = VectorStack()
 *         if lnames:             # <<<<<<<<<<<<<<
 *             for item in lnames:
 *                 context.lnames.push(Vector._coerce(item))
 */
  __pyx_t_2 = (__pyx_v_lnames != Py_None)&&(PyList_GET_SIZE(__pyx_v_lnames) != 0);
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":810
 *             context.lnames = VectorStack()
 *         if lnames:
 *             for item in lnames:             # <<<<<<<<<<<<<<
 *                 context.lnames.push(Vector._coerce(item))
 *         if context.stack is None:
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 810, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_lnames; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 810, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 810, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "flitter/language/vm.pyx":811
 *         if lnames:
 *             for item in lnames:
 *                 context.lnames.push(Vector._coerce(item))             # <<<<<<<<<<<<<<
 *         if context.stack is None:
 *             context.stack = VectorStack()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->lnames, __pyx_n_s_push); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_coerce(__pyx_v_item)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 811, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "flitter/language/vm.pyx":810
 *             context.lnames = VectorStack()
 *         if lnames:
 *             for item in lnames:             # <<<<<<<<<<<<<<
 *                 context.lnames.push(Vector._coerce(item))
 *         if context.stack is None:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":809
 *         if context.lnames is None:
 *             context.lnames = VectorStack()
 *         if lnames:             # <<<<<<<<<<<<<<
 *             for item in lnames:
 *                 context.lnames.push(Vector._coerce(item))
 */
  }

  /* "flitter/language/vm.pyx":812
 *             for item in lnames:
 *                 context.lnames.push(Vector._coerce(item))
 *         if context.stack is None:             # <<<<<<<<<<<<<<
 *             context.stack = VectorStack()
 *         self._execute(context, 0, record_stats)
 */
  __pyx_t_2 = (__pyx_v_context->stack == Py_None);
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":813
 *                 context.lnames.push(Vector._coerce(item))
 *         if context.stack is None:
 *             context.stack = VectorStack()             # <<<<<<<<<<<<<<
 *         self._execute(context, 0, record_stats)
 *         if lnames is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_context->stack);
    __Pyx_DECREF(__pyx_v_context->stack);
    __pyx_v_context->stack = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":812
 *             for item in lnames:
 *                 context.lnames.push(Vector._coerce(item))
 *         if context.stack is None:             # <<<<<<<<<<<<<<
 *             context.stack = VectorStack()
 *         self._execute(context, 0, record_stats)
 */
  }

  /* "flitter/language/vm.pyx":814
 *         if context.stack is None:
 *             context.stack = VectorStack()
 *         self._execute(context, 0, record_stats)             # <<<<<<<<<<<<<<
 *         if lnames is not None:
 *             lnames.clear()
 */
  ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_self->__pyx_vtab)->_execute(__pyx_v_self, __pyx_v_context, 0, __pyx_v_record_stats); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 814, __pyx_L1_error)

  /* "flitter/language/vm.pyx":815
 *             context.stack = VectorStack()
 *         self._execute(context, 0, record_stats)
 *         if lnames is not None:             # <<<<<<<<<<<<<<
 *             lnames.clear()
 *             while len(context.lnames):
 */
  __pyx_t_2 = (__pyx_v_lnames != ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "flitter/language/vm.pyx":816
 *         self._execute(context, 0, record_stats)
 *         if lnames is not None:
 *             lnames.clear()             # <<<<<<<<<<<<<<
 *             while len(context.lnames):
 *                 lnames.insert(0, context.lnames.pop())
 */
    __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_clear, __pyx_v_lnames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/vm.pyx":817
 *         if lnames is not None:
 *             lnames.clear()
 *             while len(context.lnames):             # <<<<<<<<<<<<<<
 *                 lnames.insert(0, context.lnames.pop())
 *         return context.stack.pop_list(len(context.stack))
 */
    while (1) {
      __pyx_t_1 = __pyx_v_context->lnames;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 817, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (!__pyx_t_2) break;

      /* "flitter/language/vm.pyx":818
 *             lnames.clear()
 *             while len(context.lnames):
 *                 lnames.insert(0, context.lnames.pop())             # <<<<<<<<<<<<<<
 *         return context.stack.pop_list(len(context.stack))
 * 
 */
      if (unlikely(__pyx_v_lnames == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "insert");
        __PYX_ERR(0, 818, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_context->lnames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = PyList_Insert(__pyx_v_lnames, 0, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 818, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "flitter/language/vm.pyx":815
 *             context.stack = VectorStack()
 *         self._execute(context, 0, record_stats)
 *         if lnames is not None:             # <<<<<<<<<<<<<<
 *             lnames.clear()
 *             while len(context.lnames):
 */
  }

  /* "flitter/language/vm.pyx":819
 *             while len(context.lnames):
 *                 lnames.insert(0, context.lnames.pop())
 *         return context.stack.pop_list(len(context.stack))             # <<<<<<<<<<<<<<
 * 
 *     def run(self, Context context, bint record_stats=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->stack, __pyx_n_s_pop_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":802
 *         return self
 * 
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.Program.execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":821
 *         return context.stack.pop_list(len(context.stack))
 * 
 *     def run(self, Context context, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         if context.stack is None:
 *             if self.stack is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_15run(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_15run = {"run", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_15run, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_15run(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_context = 0;
  int __pyx_v_record_stats;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,&__pyx_n_s_record_stats,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record_stats);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "run") < 0)) __PYX_ERR(0, 821, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = ((struct __pyx_obj_7flitter_5model_Context *)values[0]);
    if (values[1]) {
      __pyx_v_record_stats = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_record_stats == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L3_error)
    } else {
      __pyx_v_record_stats = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 821, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_ptype_7flitter_5model_Context, 1, "context", 0))) __PYX_ERR(0, 821, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_14run(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_context, __pyx_v_record_stats);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_14run(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, int __pyx_v_record_stats) {
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  PyObject *__pyx_v_objptr;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7;
  int64_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run", 1);

  /* "flitter/language/vm.pyx":822
 * 
 *     def run(self, Context context, bint record_stats=False):
 *         if context.stack is None:             # <<<<<<<<<<<<<<
 *             if self.stack is None:
 *                 self.stack = VectorStack.__new__(VectorStack)
 */
  __pyx_t_1 = (__pyx_v_context->stack == Py_None);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":823
 *     def run(self, Context context, bint record_stats=False):
 *         if context.stack is None:
 *             if self.stack is None:             # <<<<<<<<<<<<<<
 *                 self.stack = VectorStack.__new__(VectorStack)
 *             context.stack = self.stack
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->stack) == Py_None);
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":824
 *         if context.stack is None:
 *             if self.stack is None:
 *                 self.stack = VectorStack.__new__(VectorStack)             # <<<<<<<<<<<<<<
 *             context.stack = self.stack
 *         if context.lnames is None:
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_7flitter_8language_2vm_VectorStack(((PyTypeObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
      __Pyx_GIVEREF((PyObject *)__pyx_t_2);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->stack);
      __Pyx_DECREF((PyObject *)__pyx_v_self->stack);
      __pyx_v_self->stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "flitter/language/vm.pyx":823
 *     def run(self, Context context, bint record_stats=False):
 *         if context.stack is None:
 *             if self.stack is None:             # <<<<<<<<<<<<<<
 *                 self.stack = VectorStack.__new__(VectorStack)
 *             context.stack = self.stack
 */
    }

    /* "flitter/language/vm.pyx":825
 *             if self.stack is None:
 *                 self.stack = VectorStack.__new__(VectorStack)
 *             context.stack = self.stack             # <<<<<<<<<<<<<<
 *         if context.lnames is None:
 *             if self.lnames is None:
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->stack);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_context->stack);
    __Pyx_DECREF(__pyx_v_context->stack);
    __pyx_v_context->stack = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "flitter/language/vm.pyx":822
 * 
 *     def run(self, Context context, bint record_stats=False):
 *         if context.stack is None:             # <<<<<<<<<<<<<<
 *             if self.stack is None:
 *                 self.stack = VectorStack.__new__(VectorStack)
 */
  }

  /* "flitter/language/vm.pyx":826
 *                 self.stack = VectorStack.__new__(VectorStack)
 *             context.stack = self.stack
 *         if context.lnames is None:             # <<<<<<<<<<<<<<
 *             if self.lnames is None:
 *                 self.lnames = VectorStack.__new__(VectorStack)
 */
  __pyx_t_1 = (__pyx_v_context->lnames == Py_None);
  if (__pyx_t_1) {

    /* "flitter/language/vm.pyx":827
 *             context.stack = self.stack
 *         if context.lnames is None:
 *             if self.lnames is None:             # <<<<<<<<<<<<<<
 *                 self.lnames = VectorStack.__new__(VectorStack)
 *             context.lnames = self.lnames
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->lnames) == Py_None);
    if (__pyx_t_1) {

      /* "flitter/language/vm.pyx":828
 *         if context.lnames is None:
 *             if self.lnames is None:
 *                 self.lnames = VectorStack.__new__(VectorStack)             # <<<<<<<<<<<<<<
 *             context.lnames = self.lnames
 *         context.path = self.path
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_7flitter_8language_2vm_VectorStack(((PyTypeObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
      __Pyx_GIVEREF((PyObject *)__pyx_t_2);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->lnames);
      __Pyx_DECREF((PyObject *)__pyx_v_self->lnames);
      __pyx_v_self->lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "flitter/language/vm.pyx":827
 *             context.stack = self.stack
 *         if context.lnames is None:
 *             if self.lnames is None:             # <<<<<<<<<<<<<<
 *                 self.lnames = VectorStack.__new__(VectorStack)
 *             context.lnames = self.lnames
 */
    }

    /* "flitter/language/vm.pyx":829
 *             if self.lnames is None:
 *                 self.lnames = VectorStack.__new__(VectorStack)
 *             context.lnames = self.lnames             # <<<<<<<<<<<<<<
 *         context.path = self.path
 *         cdef int64_t i, n=PyTuple_GET_SIZE(self.initial_lnames)
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->lnames);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_context->lnames);
    __Pyx_DECREF(__pyx_v_context->lnames);
    __pyx_v_context->lnames = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "flitter/language/vm.pyx":826
 *                 self.stack = VectorStack.__new__(VectorStack)
 *             context.stack = self.stack
 *         if context.lnames is None:             # <<<<<<<<<<<<<<
 *             if self.lnames is None:
 *                 self.lnames = VectorStack.__new__(VectorStack)
 */
  }

  /* "flitter/language/vm.pyx":830
 *                 self.lnames = VectorStack.__new__(VectorStack)
 *             context.lnames = self.lnames
 *         context.path = self.path             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=PyTuple_GET_SIZE(self.initial_lnames)
 *         cdef PyObject* objptr
 */
  __pyx_t_2 = __pyx_v_self->path;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_context->path);
  __Pyx_DECREF(__pyx_v_context->path);
  __pyx_v_context->path = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":831
 *             context.lnames = self.lnames
 *         context.path = self.path
 *         cdef int64_t i, n=PyTuple_GET_SIZE(self.initial_lnames)             # <<<<<<<<<<<<<<
 *         cdef PyObject* objptr
 *         for i in range(n):
 */
  __pyx_t_2 = __pyx_v_self->initial_lnames;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_n = PyTuple_GET_SIZE(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":833
 *         cdef int64_t i, n=PyTuple_GET_SIZE(self.initial_lnames)
 *         cdef PyObject* objptr
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             objptr = PyDict_GetItem(context.names, <object>PyTuple_GET_ITEM(self.initial_lnames, i))
 *             assert objptr != NULL, "Missing lname"
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "flitter/language/vm.pyx":834
 *         cdef PyObject* objptr
 *         for i in range(n):
 *             objptr = PyDict_GetItem(context.names, <object>PyTuple_GET_ITEM(self.initial_lnames, i))             # <<<<<<<<<<<<<<
 *             assert objptr != NULL, "Missing lname"
 *             push(context.lnames, <Vector>objptr)
 */
    __pyx_t_2 = __pyx_v_context->names;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = __pyx_v_self->initial_lnames;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_v_i);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_objptr = PyDict_GetItem(__pyx_t_2, ((PyObject *)__pyx_t_7));
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/vm.pyx":835
 *         for i in range(n):
 *             objptr = PyDict_GetItem(context.names, <object>PyTuple_GET_ITEM(self.initial_lnames, i))
 *             assert objptr != NULL, "Missing lname"             # <<<<<<<<<<<<<<
 *             push(context.lnames, <Vector>objptr)
 *         push(context.stack, Vector(context.root))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_objptr != NULL);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Missing_lname, 0, 0);
        __PYX_ERR(0, 835, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 835, __pyx_L1_error)
    #endif

    /* "flitter/language/vm.pyx":836
 *             objptr = PyDict_GetItem(context.names, <object>PyTuple_GET_ITEM(self.initial_lnames, i))
 *             assert objptr != NULL, "Missing lname"
 *             push(context.lnames, <Vector>objptr)             # <<<<<<<<<<<<<<
 *         push(context.stack, Vector(context.root))
 *         try:
 */
    if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 836, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_context->lnames;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_2), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_objptr)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 836, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "flitter/language/vm.pyx":837
 *             assert objptr != NULL, "Missing lname"
 *             push(context.lnames, <Vector>objptr)
 *         push(context.stack, Vector(context.root))             # <<<<<<<<<<<<<<
 *         try:
 *             self._execute(context, 0, record_stats)
 */
  if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 837, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_5model_Vector), ((PyObject *)__pyx_v_context->root)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __pyx_f_7flitter_8language_2vm_push(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_2), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)); if (unlikely(__pyx_t_3 == ((int64_t)0))) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":838
 *             push(context.lnames, <Vector>objptr)
 *         push(context.stack, Vector(context.root))
 *         try:             # <<<<<<<<<<<<<<
 *             self._execute(context, 0, record_stats)
 *         except StackOverflow:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "flitter/language/vm.pyx":839
 *         push(context.stack, Vector(context.root))
 *         try:
 *             self._execute(context, 0, record_stats)             # <<<<<<<<<<<<<<
 *         except StackOverflow:
 *             context.errors.add("Stack overflow in program")
 */
      ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_self->__pyx_vtab)->_execute(__pyx_v_self, __pyx_v_context, 0, __pyx_v_record_stats); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L9_error)

      /* "flitter/language/vm.pyx":838
 *             push(context.lnames, <Vector>objptr)
 *         push(context.stack, Vector(context.root))
 *         try:             # <<<<<<<<<<<<<<
 *             self._execute(context, 0, record_stats)
 *         except StackOverflow:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flitter/language/vm.pyx":840
 *         try:
 *             self._execute(context, 0, record_stats)
 *         except StackOverflow:             # <<<<<<<<<<<<<<
 *             context.errors.add("Stack overflow in program")
 *             drop(context.lnames, len(context.lnames) - n)
 */
    __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_2, &__pyx_t_12);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_StackOverflow); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 840, __pyx_L11_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_13);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_ErrRestore(__pyx_t_6, __pyx_t_2, __pyx_t_12);
    __pyx_t_6 = 0; __pyx_t_2 = 0; __pyx_t_12 = 0;
    if (__pyx_t_14) {
      __Pyx_AddTraceback("flitter.language.vm.Program.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_2, &__pyx_t_6) < 0) __PYX_ERR(0, 840, __pyx_L11_except_error)
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "flitter/language/vm.pyx":841
 *             self._execute(context, 0, record_stats)
 *         except StackOverflow:
 *             context.errors.add("Stack overflow in program")             # <<<<<<<<<<<<<<
 *             drop(context.lnames, len(context.lnames) - n)
 *             drop(context.stack, len(context.stack) - 1)
 */
      if (unlikely(__pyx_v_context->errors == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 841, __pyx_L11_except_error)
      }
      __pyx_t_15 = PySet_Add(__pyx_v_context->errors, __pyx_kp_u_Stack_overflow_in_program); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 841, __pyx_L11_except_error)

      /* "flitter/language/vm.pyx":842
 *         except StackOverflow:
 *             context.errors.add("Stack overflow in program")
 *             drop(context.lnames, len(context.lnames) - n)             # <<<<<<<<<<<<<<
 *             drop(context.stack, len(context.stack) - 1)
 *             return context
 */
      if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 842, __pyx_L11_except_error)
      __pyx_t_13 = __pyx_v_context->lnames;
      __Pyx_INCREF(__pyx_t_13);
      __pyx_t_16 = __pyx_v_context->lnames;
      __Pyx_INCREF(__pyx_t_16);
      __pyx_t_17 = PyObject_Length(__pyx_t_16); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 842, __pyx_L11_except_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_f_7flitter_8language_2vm_drop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_13), (__pyx_t_17 - __pyx_v_n)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L11_except_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "flitter/language/vm.pyx":843
 *             context.errors.add("Stack overflow in program")
 *             drop(context.lnames, len(context.lnames) - n)
 *             drop(context.stack, len(context.stack) - 1)             # <<<<<<<<<<<<<<
 *             return context
 *         assert (<VectorStack>context.lnames).top == n-1, "Bad lnames"
 */
      if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 843, __pyx_L11_except_error)
      __pyx_t_13 = __pyx_v_context->stack;
      __Pyx_INCREF(__pyx_t_13);
      __pyx_t_16 = __pyx_v_context->stack;
      __Pyx_INCREF(__pyx_t_16);
      __pyx_t_17 = PyObject_Length(__pyx_t_16); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 843, __pyx_L11_except_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_f_7flitter_8language_2vm_drop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_13), (__pyx_t_17 - 1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L11_except_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "flitter/language/vm.pyx":844
 *             drop(context.lnames, len(context.lnames) - n)
 *             drop(context.stack, len(context.stack) - 1)
 *             return context             # <<<<<<<<<<<<<<
 *         assert (<VectorStack>context.lnames).top == n-1, "Bad lnames"
 *         drop(context.lnames, n)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_context);
      __pyx_r = ((PyObject *)__pyx_v_context);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L12_except_return;
    }
    goto __pyx_L11_except_error;

    /* "flitter/language/vm.pyx":838
 *             push(context.lnames, <Vector>objptr)
 *         push(context.stack, Vector(context.root))
 *         try:             # <<<<<<<<<<<<<<
 *             self._execute(context, 0, record_stats)
 *         except StackOverflow:
 */
    __pyx_L11_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L12_except_return:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L0;
    __pyx_L14_try_end:;
  }

  /* "flitter/language/vm.pyx":845
 *             drop(context.stack, len(context.stack) - 1)
 *             return context
 *         assert (<VectorStack>context.lnames).top == n-1, "Bad lnames"             # <<<<<<<<<<<<<<
 *         drop(context.lnames, n)
 *         assert (<VectorStack>context.stack).top == 0, "Bad stack"
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_context->lnames)->top == (__pyx_v_n - 1));
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_lnames, 0, 0);
      __PYX_ERR(0, 845, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 845, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":846
 *             return context
 *         assert (<VectorStack>context.lnames).top == n-1, "Bad lnames"
 *         drop(context.lnames, n)             # <<<<<<<<<<<<<<
 *         assert (<VectorStack>context.stack).top == 0, "Bad stack"
 *         cdef Vector result = pop(context.stack)
 */
  if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 846, __pyx_L1_error)
  __pyx_t_6 = __pyx_v_context->lnames;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_f_7flitter_8language_2vm_drop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6), __pyx_v_n); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flitter/language/vm.pyx":847
 *         assert (<VectorStack>context.lnames).top == n-1, "Bad lnames"
 *         drop(context.lnames, n)
 *         assert (<VectorStack>context.stack).top == 0, "Bad stack"             # <<<<<<<<<<<<<<
 *         cdef Vector result = pop(context.stack)
 *         assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_v_context->stack)->top == 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_stack, 0, 0);
      __PYX_ERR(0, 847, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 847, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":848
 *         drop(context.lnames, n)
 *         assert (<VectorStack>context.stack).top == 0, "Bad stack"
 *         cdef Vector result = pop(context.stack)             # <<<<<<<<<<<<<<
 *         assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"
 *         context.root = result.objects[0]
 */
  if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 848, __pyx_L1_error)
  __pyx_t_6 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_2 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_6))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_result = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":849
 *         assert (<VectorStack>context.stack).top == 0, "Bad stack"
 *         cdef Vector result = pop(context.stack)
 *         assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"             # <<<<<<<<<<<<<<
 *         context.root = result.objects[0]
 *         return context
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_18 = (__pyx_v_result->length == 1);
    if (__pyx_t_18) {
    } else {
      __pyx_t_1 = __pyx_t_18;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_18 = (__pyx_v_result->objects != ((PyObject*)Py_None));
    if (__pyx_t_18) {
    } else {
      __pyx_t_1 = __pyx_t_18;
      goto __pyx_L17_bool_binop_done;
    }
    if (unlikely(__pyx_v_result->objects == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 849, __pyx_L1_error)
    }
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v_result->objects, 0);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_18 = __Pyx_TypeCheck(__pyx_t_2, __pyx_ptype_7flitter_5model_Node); 
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_18;
    __pyx_L17_bool_binop_done:;
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Bad_root_node, 0, 0);
      __PYX_ERR(0, 849, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 849, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":850
 *         cdef Vector result = pop(context.stack)
 *         assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"
 *         context.root = result.objects[0]             # <<<<<<<<<<<<<<
 *         return context
 * 
 */
  if (unlikely(__pyx_v_result->objects == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 850, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v_result->objects, 0)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v_result->objects, 0), __pyx_ptype_7flitter_5model_Node))))) __PYX_ERR(0, 850, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v_result->objects, 0);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_context->root);
  __Pyx_DECREF((PyObject *)__pyx_v_context->root);
  __pyx_v_context->root = ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":851
 *         assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"
 *         context.root = result.objects[0]
 *         return context             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program link(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __pyx_r = ((PyObject *)__pyx_v_context);
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":821
 *         return context.stack.pop_list(len(context.stack))
 * 
 *     def run(self, Context context, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         if context.stack is None:
 *             if self.stack is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("flitter.language.vm.Program.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":853
 *         return context
 * 
 *     cpdef Program link(self):             # <<<<<<<<<<<<<<
 *         cdef Instruction instruction
 *         cdef int64_t label, address, target
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_17link(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_link(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instruction = 0;
  int64_t __pyx_v_label;
  int64_t __pyx_v_address;
  int64_t __pyx_v_target;
  PyObject *__pyx_v_addresses = 0;
  struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v_jump = 0;
  PyObject *__pyx_v_jumps = 0;
  PyObject *__pyx_v_labels = 0;
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("link", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_link); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_17link)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 853, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":858
 *         cdef list addresses
 *         cdef InstructionJump jump
 *         cdef dict jumps={}, labels={}             # <<<<<<<<<<<<<<
 *         for address, instruction in enumerate(self.instructions):
 *             if instruction.code == OpCode.Label:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_jumps = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_labels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":859
 *         cdef InstructionJump jump
 *         cdef dict jumps={}, labels={}
 *         for address, instruction in enumerate(self.instructions):             # <<<<<<<<<<<<<<
 *             if instruction.code == OpCode.Label:
 *                 labels[(<InstructionLabel>instruction).label] = address
 */
  __pyx_t_6 = 0;
  __pyx_t_1 = __pyx_v_self->instructions; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_7 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
      #endif
      if (__pyx_t_7 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Instruction))))) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_instruction, ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_2));
    __pyx_t_2 = 0;
    __pyx_v_address = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "flitter/language/vm.pyx":860
 *         cdef dict jumps={}, labels={}
 *         for address, instruction in enumerate(self.instructions):
 *             if instruction.code == OpCode.Label:             # <<<<<<<<<<<<<<
 *                 labels[(<InstructionLabel>instruction).label] = address
 *             elif isinstance(instruction, InstructionJump):
 */
    __pyx_t_8 = (__pyx_v_instruction->code == __pyx_e_7flitter_8language_2vm_Label);
    if (__pyx_t_8) {

      /* "flitter/language/vm.pyx":861
 *         for address, instruction in enumerate(self.instructions):
 *             if instruction.code == OpCode.Label:
 *                 labels[(<InstructionLabel>instruction).label] = address             # <<<<<<<<<<<<<<
 *             elif isinstance(instruction, InstructionJump):
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
 */
      __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_address); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 861, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v_instruction)->label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely((PyDict_SetItem(__pyx_v_labels, __pyx_t_3, __pyx_t_2) < 0))) __PYX_ERR(0, 861, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "flitter/language/vm.pyx":860
 *         cdef dict jumps={}, labels={}
 *         for address, instruction in enumerate(self.instructions):
 *             if instruction.code == OpCode.Label:             # <<<<<<<<<<<<<<
 *                 labels[(<InstructionLabel>instruction).label] = address
 *             elif isinstance(instruction, InstructionJump):
 */
      goto __pyx_L5;
    }

    /* "flitter/language/vm.pyx":862
 *             if instruction.code == OpCode.Label:
 *                 labels[(<InstructionLabel>instruction).label] = address
 *             elif isinstance(instruction, InstructionJump):             # <<<<<<<<<<<<<<
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
 *         for label, addresses in jumps.items():
 */
    __pyx_t_8 = __Pyx_TypeCheck(((PyObject *)__pyx_v_instruction), __pyx_ptype_7flitter_8language_2vm_InstructionJump); 
    if (__pyx_t_8) {

      /* "flitter/language/vm.pyx":863
 *                 labels[(<InstructionLabel>instruction).label] = address
 *             elif isinstance(instruction, InstructionJump):
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)             # <<<<<<<<<<<<<<
 *         for label, addresses in jumps.items():
 *             target = labels[label]
 */
      __pyx_t_2 = __Pyx_PyInt_From_int64_t(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_instruction)->label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyDict_SetDefault(__pyx_v_jumps, __pyx_t_2, __pyx_t_3, -1L); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_t_4 == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 863, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_address); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __Pyx_PyList_Append(((PyObject*)__pyx_t_4), __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flitter/language/vm.pyx":862
 *             if instruction.code == OpCode.Label:
 *                 labels[(<InstructionLabel>instruction).label] = address
 *             elif isinstance(instruction, InstructionJump):             # <<<<<<<<<<<<<<
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
 *         for label, addresses in jumps.items():
 */
    }
    __pyx_L5:;

    /* "flitter/language/vm.pyx":859
 *         cdef InstructionJump jump
 *         cdef dict jumps={}, labels={}
 *         for address, instruction in enumerate(self.instructions):             # <<<<<<<<<<<<<<
 *             if instruction.code == OpCode.Label:
 *                 labels[(<InstructionLabel>instruction).label] = address
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":864
 *             elif isinstance(instruction, InstructionJump):
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
 *         for label, addresses in jumps.items():             # <<<<<<<<<<<<<<
 *             target = labels[label]
 *             for address in addresses:
 */
  __pyx_t_7 = 0;
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_jumps, 1, __pyx_n_s_items_2, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_7, &__pyx_t_3, &__pyx_t_4, NULL, __pyx_t_11);
    if (unlikely(__pyx_t_12 == 0)) break;
    if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_4))) __PYX_ERR(0, 864, __pyx_L1_error)
    __pyx_v_label = __pyx_t_6;
    __Pyx_XDECREF_SET(__pyx_v_addresses, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "flitter/language/vm.pyx":865
 *                 (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
 *         for label, addresses in jumps.items():
 *             target = labels[label]             # <<<<<<<<<<<<<<
 *             for address in addresses:
 *                 jump = self.instructions[address]
 */
    __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_labels, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_target = __pyx_t_6;

    /* "flitter/language/vm.pyx":866
 *         for label, addresses in jumps.items():
 *             target = labels[label]
 *             for address in addresses:             # <<<<<<<<<<<<<<
 *                 jump = self.instructions[address]
 *                 jump.offset = target - address
 */
    if (unlikely(__pyx_v_addresses == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 866, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_v_addresses; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_13 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 866, __pyx_L1_error)
        #endif
        if (__pyx_t_13 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_4); __pyx_t_13++; if (unlikely((0 < 0))) __PYX_ERR(0, 866, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 866, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_address = __pyx_t_6;

      /* "flitter/language/vm.pyx":867
 *             target = labels[label]
 *             for address in addresses:
 *                 jump = self.instructions[address]             # <<<<<<<<<<<<<<
 *                 jump.offset = target - address
 *         self.linked = True
 */
      if (unlikely(__pyx_v_self->instructions == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 867, __pyx_L1_error)
      }
      if (!(likely(((PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_v_address)) == Py_None) || likely(__Pyx_TypeTest(PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_v_address), __pyx_ptype_7flitter_8language_2vm_InstructionJump))))) __PYX_ERR(0, 867, __pyx_L1_error)
      __pyx_t_4 = PyList_GET_ITEM(__pyx_v_self->instructions, __pyx_v_address);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_jump, ((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/vm.pyx":868
 *             for address in addresses:
 *                 jump = self.instructions[address]
 *                 jump.offset = target - address             # <<<<<<<<<<<<<<
 *         self.linked = True
 *         return self
 */
      __pyx_v_jump->offset = (__pyx_v_target - __pyx_v_address);

      /* "flitter/language/vm.pyx":866
 *         for label, addresses in jumps.items():
 *             target = labels[label]
 *             for address in addresses:             # <<<<<<<<<<<<<<
 *                 jump = self.instructions[address]
 *                 jump.offset = target - address
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":869
 *                 jump = self.instructions[address]
 *                 jump.offset = target - address
 *         self.linked = True             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_v_self->linked = 1;

  /* "flitter/language/vm.pyx":870
 *                 jump.offset = target - address
 *         self.linked = True
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef void use_simplifier(self, bint simplify):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":853
 *         return context
 * 
 *     cpdef Program link(self):             # <<<<<<<<<<<<<<
 *         cdef Instruction instruction
 *         cdef int64_t label, address, target
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.link", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_instruction);
  __Pyx_XDECREF(__pyx_v_addresses);
  __Pyx_XDECREF((PyObject *)__pyx_v_jump);
  __Pyx_XDECREF(__pyx_v_jumps);
  __Pyx_XDECREF(__pyx_v_labels);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_17link(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_17link = {"link", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_17link, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_17link(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("link (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("link", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "link", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_16link(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_16link(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("link", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_link(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.link", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":872
 *         return self
 * 
 *     cpdef void use_simplifier(self, bint simplify):             # <<<<<<<<<<<<<<
 *         self.simplify = simplify
 * 
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_19use_simplifier(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7flitter_8language_2vm_7Program_use_simplifier(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_v_simplify, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("use_simplifier", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_use_simplifier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_19use_simplifier)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_simplify); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 872, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":873
 * 
 *     cpdef void use_simplifier(self, bint simplify):
 *         self.simplify = simplify             # <<<<<<<<<<<<<<
 * 
 *     cpdef int64_t new_label(self):
 */
  __pyx_v_self->simplify = __pyx_v_simplify;

  /* "flitter/language/vm.pyx":872
 *         return self
 * 
 *     cpdef void use_simplifier(self, bint simplify):             # <<<<<<<<<<<<<<
 *         self.simplify = simplify
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.use_simplifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_19use_simplifier(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_19use_simplifier = {"use_simplifier", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_19use_simplifier, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_19use_simplifier(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_simplify;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("use_simplifier (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_simplify,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_simplify)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "use_simplifier") < 0)) __PYX_ERR(0, 872, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_simplify = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_simplify == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("use_simplifier", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 872, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.use_simplifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_18use_simplifier(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_simplify);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_18use_simplifier(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_v_simplify) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("use_simplifier", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7flitter_8language_2vm_7Program_use_simplifier(__pyx_v_self, __pyx_v_simplify, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.use_simplifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":875
 *         self.simplify = simplify
 * 
 *     cpdef int64_t new_label(self):             # <<<<<<<<<<<<<<
 *         cdef object label = self.next_label
 *         self.next_label += 1
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_21new_label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int64_t __pyx_f_7flitter_8language_2vm_7Program_new_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_label = 0;
  int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_label", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_new_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_21new_label)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 875, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 875, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":876
 * 
 *     cpdef int64_t new_label(self):
 *         cdef object label = self.next_label             # <<<<<<<<<<<<<<
 *         self.next_label += 1
 *         return label
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->next_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_label = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":877
 *     cpdef int64_t new_label(self):
 *         cdef object label = self.next_label
 *         self.next_label += 1             # <<<<<<<<<<<<<<
 *         return label
 * 
 */
  __pyx_v_self->next_label = (__pyx_v_self->next_label + 1);

  /* "flitter/language/vm.pyx":878
 *         cdef object label = self.next_label
 *         self.next_label += 1
 *         return label             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program dup(self):
 */
  __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_v_label); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":875
 *         self.simplify = simplify
 * 
 *     cpdef int64_t new_label(self):             # <<<<<<<<<<<<<<
 *         cdef object label = self.next_label
 *         self.next_label += 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.new_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_label);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_21new_label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_21new_label = {"new_label", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_21new_label, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_21new_label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("new_label (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("new_label", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "new_label", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_20new_label(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_20new_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_label", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_8language_2vm_7Program_new_label(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 875, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.vm.Program.new_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":880
 *         return label
 * 
 *     cpdef Program dup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Dup))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_23dup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_dup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dup", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_23dup)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 880, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":881
 * 
 *     cpdef Program dup(self):
 *         self.instructions.append(Instruction(OpCode.Dup))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 881, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Dup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":882
 *     cpdef Program dup(self):
 *         self.instructions.append(Instruction(OpCode.Dup))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program drop(self, int64_t count=1):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":880
 *         return label
 * 
 *     cpdef Program dup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Dup))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.dup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_23dup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_23dup = {"dup", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_23dup, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_23dup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("dup", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "dup", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_22dup(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_22dup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dup", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_dup(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.dup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":884
 *         return self
 * 
 *     cpdef Program drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_25drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop *__pyx_optional_args) {
  int64_t __pyx_v_count = ((int64_t)1);
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("drop", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_count = __pyx_optional_args->count;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_drop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_25drop)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 884, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":885
 * 
 *     cpdef Program drop(self, int64_t count=1):
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 885, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Drop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":886
 *     cpdef Program drop(self, int64_t count=1):
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program label(self, int64_t label):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":884
 *         return self
 * 
 *     cpdef Program drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_25drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_25drop = {"drop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_25drop, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_25drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("drop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "drop") < 0)) __PYX_ERR(0, 884, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 884, __pyx_L3_error)
    } else {
      __pyx_v_count = ((int64_t)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("drop", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 884, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_24drop(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_24drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("drop", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.count = __pyx_v_count;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_8language_2vm_Program->drop(__pyx_v_self, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":888
 *         return self
 * 
 *     cpdef Program label(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionLabel(label))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_27label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("label", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_27label)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 888, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":889
 * 
 *     cpdef Program label(self, int64_t label):
 *         self.instructions.append(InstructionLabel(label))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 889, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":890
 *     cpdef Program label(self, int64_t label):
 *         self.instructions.append(InstructionLabel(label))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program jump(self, int64_t label):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":888
 *         return self
 * 
 *     cpdef Program label(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionLabel(label))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_27label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_27label = {"label", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_27label, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_27label(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("label (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "label") < 0)) __PYX_ERR(0, 888, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("label", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 888, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_26label(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_26label(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("label", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_label(__pyx_v_self, __pyx_v_label, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":892
 *         return self
 * 
 *     cpdef Program jump(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_29jump(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_jump(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jump", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_jump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_29jump)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 892, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 892, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":893
 * 
 *     cpdef Program jump(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 893, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Jump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":894
 *     cpdef Program jump(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program branch_true(self, int64_t label):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":892
 *         return self
 * 
 *     cpdef Program jump(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.jump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_29jump(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_29jump = {"jump", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_29jump, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_29jump(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("jump (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 892, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "jump") < 0)) __PYX_ERR(0, 892, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 892, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("jump", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 892, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.jump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_28jump(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_28jump(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("jump", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_jump(__pyx_v_self, __pyx_v_label, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.jump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":896
 *         return self
 * 
 *     cpdef Program branch_true(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_31branch_true(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_branch_true(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("branch_true", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_branch_true); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_31branch_true)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 896, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":897
 * 
 *     cpdef Program branch_true(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 897, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BranchTrue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":898
 *     cpdef Program branch_true(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program branch_false(self, int64_t label):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":896
 *         return self
 * 
 *     cpdef Program branch_true(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_true", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_31branch_true(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_31branch_true = {"branch_true", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_31branch_true, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_31branch_true(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("branch_true (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 896, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "branch_true") < 0)) __PYX_ERR(0, 896, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 896, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("branch_true", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 896, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_true", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_30branch_true(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_30branch_true(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("branch_true", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_branch_true(__pyx_v_self, __pyx_v_label, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_true", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":900
 *         return self
 * 
 *     cpdef Program branch_false(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_33branch_false(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_branch_false(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("branch_false", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_branch_false); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_33branch_false)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 900, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 900, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":901
 * 
 *     cpdef Program branch_false(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 901, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BranchFalse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":902
 *     cpdef Program branch_false(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program import_(self, tuple names):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":900
 *         return self
 * 
 *     cpdef Program branch_false(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_false", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_33branch_false(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_33branch_false = {"branch_false", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_33branch_false, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_33branch_false(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("branch_false (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "branch_false") < 0)) __PYX_ERR(0, 900, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("branch_false", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 900, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_false", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_32branch_false(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_32branch_false(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("branch_false", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_branch_false(__pyx_v_self, __pyx_v_label, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.branch_false", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":904
 *         return self
 * 
 *     cpdef Program import_(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_35import_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_import_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_import); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_35import_)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_names};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 904, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":905
 * 
 *     cpdef Program import_(self, tuple names):
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 905, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Import); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_names)) __PYX_ERR(0, 905, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":906
 *     cpdef Program import_(self, tuple names):
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program literal(self, value):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":904
 *         return self
 * 
 *     cpdef Program import_(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.import_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_35import_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_35import_ = {"import_", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_35import_, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_35import_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_names = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_names,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "import_") < 0)) __PYX_ERR(0, 904, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_names = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("import_", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 904, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.import_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 904, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_34import_(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_names);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_34import_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_import_(__pyx_v_self, __pyx_v_names, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.import_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":908
 *         return self
 * 
 *     cpdef Program literal(self, value):             # <<<<<<<<<<<<<<
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_37literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_vector = 0;
  PyObject *__pyx_v_obj = 0;
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("literal", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_37literal)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 908, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":909
 * 
 *     cpdef Program literal(self, value):
 *         cdef Vector vector = Vector._coerce(value)             # <<<<<<<<<<<<<<
 *         cdef object obj
 *         if vector.objects is not None:
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_coerce(__pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":911
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 *         if vector.objects is not None:             # <<<<<<<<<<<<<<
 *             if vector.length == 1:
 *                 obj = vector.objects[0]
 */
  __pyx_t_6 = (__pyx_v_vector->objects != ((PyObject*)Py_None));
  if (__pyx_t_6) {

    /* "flitter/language/vm.pyx":912
 *         cdef object obj
 *         if vector.objects is not None:
 *             if vector.length == 1:             # <<<<<<<<<<<<<<
 *                 obj = vector.objects[0]
 *                 if type(obj) is Node:
 */
    __pyx_t_6 = (__pyx_v_vector->length == 1);
    if (__pyx_t_6) {

      /* "flitter/language/vm.pyx":913
 *         if vector.objects is not None:
 *             if vector.length == 1:
 *                 obj = vector.objects[0]             # <<<<<<<<<<<<<<
 *                 if type(obj) is Node:
 *                     self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))
 */
      if (unlikely(__pyx_v_vector->objects == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 913, __pyx_L1_error)
      }
      __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_vector->objects, 0);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_obj = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "flitter/language/vm.pyx":914
 *             if vector.length == 1:
 *                 obj = vector.objects[0]
 *                 if type(obj) is Node:             # <<<<<<<<<<<<<<
 *                     self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))
 *                     return self
 */
      __pyx_t_6 = (((PyObject *)Py_TYPE(__pyx_v_obj)) == ((PyObject *)__pyx_ptype_7flitter_5model_Node));
      if (__pyx_t_6) {

        /* "flitter/language/vm.pyx":915
 *                 obj = vector.objects[0]
 *                 if type(obj) is Node:
 *                     self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))             # <<<<<<<<<<<<<<
 *                     return self
 *             else:
 */
        if (unlikely(__pyx_v_self->instructions == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(0, 915, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LiteralNode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_1);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj));
        __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj));
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_obj)) __PYX_ERR(0, 915, __pyx_L1_error);
        __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":916
 *                 if type(obj) is Node:
 *                     self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))
 *                     return self             # <<<<<<<<<<<<<<
 *             else:
 *                 for obj in vector.objects:
 */
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF((PyObject *)__pyx_v_self);
        __pyx_r = __pyx_v_self;
        goto __pyx_L0;

        /* "flitter/language/vm.pyx":914
 *             if vector.length == 1:
 *                 obj = vector.objects[0]
 *                 if type(obj) is Node:             # <<<<<<<<<<<<<<
 *                     self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))
 *                     return self
 */
      }

      /* "flitter/language/vm.pyx":912
 *         cdef object obj
 *         if vector.objects is not None:
 *             if vector.length == 1:             # <<<<<<<<<<<<<<
 *                 obj = vector.objects[0]
 *                 if type(obj) is Node:
 */
      goto __pyx_L4;
    }

    /* "flitter/language/vm.pyx":918
 *                     return self
 *             else:
 *                 for obj in vector.objects:             # <<<<<<<<<<<<<<
 *                     if type(obj) is Node:
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 */
    /*else*/ {
      if (unlikely(__pyx_v_vector->objects == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 918, __pyx_L1_error)
      }
      __pyx_t_1 = __pyx_v_vector->objects; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 918, __pyx_L1_error)
          #endif
          if (__pyx_t_8 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 918, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "flitter/language/vm.pyx":919
 *             else:
 *                 for obj in vector.objects:
 *                     if type(obj) is Node:             # <<<<<<<<<<<<<<
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 *                         return self
 */
        __pyx_t_6 = (((PyObject *)Py_TYPE(__pyx_v_obj)) == ((PyObject *)__pyx_ptype_7flitter_5model_Node));
        if (__pyx_t_6) {

          /* "flitter/language/vm.pyx":920
 *                 for obj in vector.objects:
 *                     if type(obj) is Node:
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))             # <<<<<<<<<<<<<<
 *                         return self
 *         self.instructions.append(InstructionVector(OpCode.Literal, vector))
 */
          if (unlikely(__pyx_v_self->instructions == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
            __PYX_ERR(0, 920, __pyx_L1_error)
          }
          __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LiteralNodes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_2);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error);
          __Pyx_INCREF(__pyx_v_vector->objects);
          __Pyx_GIVEREF(__pyx_v_vector->objects);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_vector->objects)) __PYX_ERR(0, 920, __pyx_L1_error);
          __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 920, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "flitter/language/vm.pyx":921
 *                     if type(obj) is Node:
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 *                         return self             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionVector(OpCode.Literal, vector))
 *         return self
 */
          __Pyx_XDECREF((PyObject *)__pyx_r);
          __Pyx_INCREF((PyObject *)__pyx_v_self);
          __pyx_r = __pyx_v_self;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L0;

          /* "flitter/language/vm.pyx":919
 *             else:
 *                 for obj in vector.objects:
 *                     if type(obj) is Node:             # <<<<<<<<<<<<<<
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 *                         return self
 */
        }

        /* "flitter/language/vm.pyx":918
 *                     return self
 *             else:
 *                 for obj in vector.objects:             # <<<<<<<<<<<<<<
 *                     if type(obj) is Node:
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L4:;

    /* "flitter/language/vm.pyx":911
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 *         if vector.objects is not None:             # <<<<<<<<<<<<<<
 *             if vector.length == 1:
 *                 obj = vector.objects[0]
 */
  }

  /* "flitter/language/vm.pyx":922
 *                         self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
 *                         return self
 *         self.instructions.append(InstructionVector(OpCode.Literal, vector))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 922, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_vector);
  __Pyx_GIVEREF((PyObject *)__pyx_v_vector);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_vector))) __PYX_ERR(0, 922, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":923
 *                         return self
 *         self.instructions.append(InstructionVector(OpCode.Literal, vector))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program local_push(self, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":908
 *         return self
 * 
 *     cpdef Program literal(self, value):             # <<<<<<<<<<<<<<
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_vector);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_37literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_37literal = {"literal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_37literal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_37literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("literal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 908, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "literal") < 0)) __PYX_ERR(0, 908, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("literal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 908, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_36literal(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_36literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("literal", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_literal(__pyx_v_self, __pyx_v_value, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":925
 *         return self
 * 
 *     cpdef Program local_push(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_39local_push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_push(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_push", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_local_push); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_39local_push)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 925, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 925, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 925, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":926
 * 
 *     cpdef Program local_push(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 926, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalPush); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":927
 *     cpdef Program local_push(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program local_load(self, int64_t offset):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":925
 *         return self
 * 
 *     cpdef Program local_push(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_39local_push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_39local_push = {"local_push", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_39local_push, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_39local_push(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("local_push (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 925, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "local_push") < 0)) __PYX_ERR(0, 925, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 925, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("local_push", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 925, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.local_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_38local_push(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_38local_push(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_push", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_local_push(__pyx_v_self, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":929
 *         return self
 * 
 *     cpdef Program local_load(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_41local_load(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_load(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_offset, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_load", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_local_load); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_41local_load)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 929, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":930
 * 
 *     cpdef Program local_load(self, int64_t offset):
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 930, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalLoad); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":931
 *     cpdef Program local_load(self, int64_t offset):
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program local_drop(self, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":929
 *         return self
 * 
 *     cpdef Program local_load(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_41local_load(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_41local_load = {"local_load", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_41local_load, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_41local_load(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_offset;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("local_load (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_offset,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_offset)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 929, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "local_load") < 0)) __PYX_ERR(0, 929, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_offset = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_offset == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 929, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("local_load", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 929, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.local_load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_40local_load(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_offset);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_40local_load(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_offset) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_load", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_local_load(__pyx_v_self, __pyx_v_offset, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":933
 *         return self
 * 
 *     cpdef Program local_drop(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_43local_drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_local_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_drop", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_local_drop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_43local_drop)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 933, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":934
 * 
 *     cpdef Program local_drop(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 934, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalDrop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 934, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":935
 *     cpdef Program local_drop(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program lookup(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":933
 *         return self
 * 
 *     cpdef Program local_drop(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_43local_drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_43local_drop = {"local_drop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_43local_drop, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_43local_drop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("local_drop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "local_drop") < 0)) __PYX_ERR(0, 933, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("local_drop", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 933, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.local_drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_42local_drop(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_42local_drop(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("local_drop", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_local_drop(__pyx_v_self, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.local_drop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":937
 *         return self
 * 
 *     cpdef Program lookup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lookup))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_45lookup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lookup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lookup", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lookup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_45lookup)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 937, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":938
 * 
 *     cpdef Program lookup(self):
 *         self.instructions.append(Instruction(OpCode.Lookup))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 938, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Lookup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":939
 *     cpdef Program lookup(self):
 *         self.instructions.append(Instruction(OpCode.Lookup))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program lookup_literal(self, Vector value):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":937
 *         return self
 * 
 *     cpdef Program lookup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lookup))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_45lookup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_45lookup = {"lookup", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_45lookup, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_45lookup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lookup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("lookup", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "lookup", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_44lookup(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_44lookup(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lookup", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_lookup(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":941
 *         return self
 * 
 *     cpdef Program lookup_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_47lookup_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lookup_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lookup_literal", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lookup_literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_47lookup_literal)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_value)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 941, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":942
 * 
 *     cpdef Program lookup_literal(self, Vector value):
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 942, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LookupLiteral); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_value))) __PYX_ERR(0, 942, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":943
 *     cpdef Program lookup_literal(self, Vector value):
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program range(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":941
 *         return self
 * 
 *     cpdef Program lookup_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.lookup_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_47lookup_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_47lookup_literal = {"lookup_literal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_47lookup_literal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_47lookup_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lookup_literal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 941, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lookup_literal") < 0)) __PYX_ERR(0, 941, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Vector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lookup_literal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 941, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.lookup_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_7flitter_5model_Vector, 1, "value", 0))) __PYX_ERR(0, 941, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_46lookup_literal(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_46lookup_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lookup_literal", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_lookup_literal(__pyx_v_self, __pyx_v_value, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.lookup_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":945
 *         return self
 * 
 *     cpdef Program range(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Range))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_49range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_range(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("range", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_range); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_49range)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 945, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":946
 * 
 *     cpdef Program range(self):
 *         self.instructions.append(Instruction(OpCode.Range))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 946, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Range); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":947
 *     cpdef Program range(self):
 *         self.instructions.append(Instruction(OpCode.Range))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program neg(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":945
 *         return self
 * 
 *     cpdef Program range(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Range))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_49range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_49range = {"range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_49range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_49range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("range (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("range", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "range", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_48range(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_48range(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("range", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_range(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":949
 *         return self
 * 
 *     cpdef Program neg(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Neg))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_51neg(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_neg(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("neg", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_neg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_51neg)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 949, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 949, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":950
 * 
 *     cpdef Program neg(self):
 *         self.instructions.append(Instruction(OpCode.Neg))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 950, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Neg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":951
 *     cpdef Program neg(self):
 *         self.instructions.append(Instruction(OpCode.Neg))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program pos(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":949
 *         return self
 * 
 *     cpdef Program neg(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Neg))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.neg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_51neg(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_51neg = {"neg", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_51neg, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_51neg(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("neg (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("neg", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "neg", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_50neg(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_50neg(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("neg", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_neg(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.neg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":953
 *         return self
 * 
 *     cpdef Program pos(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pos))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_53pos(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_pos(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pos", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_53pos)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 953, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 953, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":954
 * 
 *     cpdef Program pos(self):
 *         self.instructions.append(Instruction(OpCode.Pos))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 954, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":955
 *     cpdef Program pos(self):
 *         self.instructions.append(Instruction(OpCode.Pos))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program ceil(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":953
 *         return self
 * 
 *     cpdef Program pos(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pos))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_53pos(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_53pos = {"pos", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_53pos, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_53pos(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pos (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pos", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pos", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_52pos(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_52pos(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pos", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_pos(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":957
 *         return self
 * 
 *     cpdef Program ceil(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ceil))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_55ceil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ceil(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ceil", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ceil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_55ceil)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 957, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":958
 * 
 *     cpdef Program ceil(self):
 *         self.instructions.append(Instruction(OpCode.Ceil))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 958, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ceil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":959
 *     cpdef Program ceil(self):
 *         self.instructions.append(Instruction(OpCode.Ceil))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program floor(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":957
 *         return self
 * 
 *     cpdef Program ceil(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ceil))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.ceil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_55ceil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_55ceil = {"ceil", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_55ceil, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_55ceil(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ceil (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("ceil", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "ceil", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_54ceil(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_54ceil(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ceil", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_ceil(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.ceil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":961
 *         return self
 * 
 *     cpdef Program floor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Floor))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_57floor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_floor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("floor", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_floor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_57floor)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 961, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 961, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":962
 * 
 *     cpdef Program floor(self):
 *         self.instructions.append(Instruction(OpCode.Floor))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 962, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Floor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":963
 *     cpdef Program floor(self):
 *         self.instructions.append(Instruction(OpCode.Floor))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program fract(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":961
 *         return self
 * 
 *     cpdef Program floor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Floor))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.floor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_57floor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_57floor = {"floor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_57floor, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_57floor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("floor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("floor", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "floor", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_56floor(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_56floor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("floor", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_floor(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.floor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":965
 *         return self
 * 
 *     cpdef Program fract(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Fract))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_59fract(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_fract(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fract", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fract); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_59fract)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 965, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":966
 * 
 *     cpdef Program fract(self):
 *         self.instructions.append(Instruction(OpCode.Fract))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 966, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Fract); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":967
 *     cpdef Program fract(self):
 *         self.instructions.append(Instruction(OpCode.Fract))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program not_(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":965
 *         return self
 * 
 *     cpdef Program fract(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Fract))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.fract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_59fract(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_59fract = {"fract", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_59fract, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_59fract(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fract (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("fract", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "fract", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_58fract(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_58fract(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fract", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_fract(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.fract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":969
 *         return self
 * 
 *     cpdef Program not_(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Not))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_61not_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_not_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("not_", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_not); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_61not_)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 969, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 969, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":970
 * 
 *     cpdef Program not_(self):
 *         self.instructions.append(Instruction(OpCode.Not))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 970, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Not); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":971
 *     cpdef Program not_(self):
 *         self.instructions.append(Instruction(OpCode.Not))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program add(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":969
 *         return self
 * 
 *     cpdef Program not_(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Not))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.not_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_61not_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_61not_ = {"not_", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_61not_, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_61not_(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("not_ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("not_", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "not_", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_60not_(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_60not_(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("not_", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_not_(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.not_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":973
 *         return self
 * 
 *     cpdef Program add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Add))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_63add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_63add)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 973, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":974
 * 
 *     cpdef Program add(self):
 *         self.instructions.append(Instruction(OpCode.Add))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 974, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":975
 *     cpdef Program add(self):
 *         self.instructions.append(Instruction(OpCode.Add))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program sub(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":973
 *         return self
 * 
 *     cpdef Program add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Add))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_63add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_63add = {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_63add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_63add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("add", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "add", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_62add(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_62add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_add(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":977
 *         return self
 * 
 *     cpdef Program sub(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Sub))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_65sub(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_sub(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sub", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sub); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_65sub)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 977, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":978
 * 
 *     cpdef Program sub(self):
 *         self.instructions.append(Instruction(OpCode.Sub))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 978, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Sub); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":979
 *     cpdef Program sub(self):
 *         self.instructions.append(Instruction(OpCode.Sub))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program mul(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":977
 *         return self
 * 
 *     cpdef Program sub(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Sub))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.sub", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_65sub(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_65sub = {"sub", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_65sub, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_65sub(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sub (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("sub", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sub", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_64sub(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_64sub(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sub", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_sub(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.sub", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":981
 *         return self
 * 
 *     cpdef Program mul(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mul))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_67mul(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mul(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mul); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_67mul)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 981, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":982
 * 
 *     cpdef Program mul(self):
 *         self.instructions.append(Instruction(OpCode.Mul))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 982, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Mul); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":983
 *     cpdef Program mul(self):
 *         self.instructions.append(Instruction(OpCode.Mul))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program mul_add(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":981
 *         return self
 * 
 *     cpdef Program mul(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mul))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.mul", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_67mul(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_67mul = {"mul", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_67mul, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_67mul(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("mul", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "mul", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_66mul(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_66mul(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_mul(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.mul", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":985
 *         return self
 * 
 *     cpdef Program mul_add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.MulAdd))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_69mul_add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mul_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul_add", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mul_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_69mul_add)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 985, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":986
 * 
 *     cpdef Program mul_add(self):
 *         self.instructions.append(Instruction(OpCode.MulAdd))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 986, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_MulAdd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":987
 *     cpdef Program mul_add(self):
 *         self.instructions.append(Instruction(OpCode.MulAdd))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program truediv(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":985
 *         return self
 * 
 *     cpdef Program mul_add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.MulAdd))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.mul_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_69mul_add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_69mul_add = {"mul_add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_69mul_add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_69mul_add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul_add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("mul_add", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "mul_add", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_68mul_add(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_68mul_add(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul_add", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_mul_add(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.mul_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":989
 *         return self
 * 
 *     cpdef Program truediv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.TrueDiv))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_71truediv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_truediv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("truediv", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_truediv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_71truediv)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 989, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 989, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":990
 * 
 *     cpdef Program truediv(self):
 *         self.instructions.append(Instruction(OpCode.TrueDiv))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 990, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_TrueDiv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":991
 *     cpdef Program truediv(self):
 *         self.instructions.append(Instruction(OpCode.TrueDiv))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program floordiv(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":989
 *         return self
 * 
 *     cpdef Program truediv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.TrueDiv))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.truediv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_71truediv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_71truediv = {"truediv", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_71truediv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_71truediv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("truediv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("truediv", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "truediv", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_70truediv(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_70truediv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("truediv", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_truediv(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.truediv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":993
 *         return self
 * 
 *     cpdef Program floordiv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.FloorDiv))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_73floordiv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_floordiv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("floordiv", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_floordiv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_73floordiv)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 993, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":994
 * 
 *     cpdef Program floordiv(self):
 *         self.instructions.append(Instruction(OpCode.FloorDiv))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 994, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_FloorDiv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":995
 *     cpdef Program floordiv(self):
 *         self.instructions.append(Instruction(OpCode.FloorDiv))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program mod(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":993
 *         return self
 * 
 *     cpdef Program floordiv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.FloorDiv))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.floordiv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_73floordiv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_73floordiv = {"floordiv", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_73floordiv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_73floordiv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("floordiv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("floordiv", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "floordiv", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_72floordiv(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_72floordiv(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("floordiv", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_floordiv(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.floordiv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":997
 *         return self
 * 
 *     cpdef Program mod(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mod))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_75mod(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_mod(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mod", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mod); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_75mod)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 997, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":998
 * 
 *     cpdef Program mod(self):
 *         self.instructions.append(Instruction(OpCode.Mod))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 998, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Mod); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":999
 *     cpdef Program mod(self):
 *         self.instructions.append(Instruction(OpCode.Mod))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program pow(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":997
 *         return self
 * 
 *     cpdef Program mod(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mod))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.mod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_75mod(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_75mod = {"mod", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_75mod, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_75mod(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mod (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("mod", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "mod", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_74mod(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_74mod(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mod", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_mod(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.mod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1001
 *         return self
 * 
 *     cpdef Program pow(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pow))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_77pow(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_pow(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pow", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pow); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_77pow)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1001, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1002
 * 
 *     cpdef Program pow(self):
 *         self.instructions.append(Instruction(OpCode.Pow))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1002, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Pow); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1003
 *     cpdef Program pow(self):
 *         self.instructions.append(Instruction(OpCode.Pow))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program eq(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1001
 *         return self
 * 
 *     cpdef Program pow(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pow))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.pow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_77pow(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_77pow = {"pow", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_77pow, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_77pow(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pow (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pow", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pow", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_76pow(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_76pow(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pow", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_pow(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.pow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1005
 *         return self
 * 
 *     cpdef Program eq(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Eq))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_79eq(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_eq(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eq", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_79eq)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1005, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1006
 * 
 *     cpdef Program eq(self):
 *         self.instructions.append(Instruction(OpCode.Eq))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1006, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1007
 *     cpdef Program eq(self):
 *         self.instructions.append(Instruction(OpCode.Eq))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program ne(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1005
 *         return self
 * 
 *     cpdef Program eq(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Eq))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_79eq(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_79eq = {"eq", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_79eq, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_79eq(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eq (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("eq", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "eq", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_78eq(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_78eq(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eq", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_eq(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1009
 *         return self
 * 
 *     cpdef Program ne(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ne))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_81ne(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ne(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ne", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ne); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_81ne)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1009, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1010
 * 
 *     cpdef Program ne(self):
 *         self.instructions.append(Instruction(OpCode.Ne))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1010, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ne); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1011
 *     cpdef Program ne(self):
 *         self.instructions.append(Instruction(OpCode.Ne))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program gt(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1009
 *         return self
 * 
 *     cpdef Program ne(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ne))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.ne", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_81ne(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_81ne = {"ne", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_81ne, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_81ne(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ne (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("ne", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "ne", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_80ne(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_80ne(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ne", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_ne(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.ne", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1013
 *         return self
 * 
 *     cpdef Program gt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Gt))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_83gt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_gt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gt", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_83gt)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1013, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1014
 * 
 *     cpdef Program gt(self):
 *         self.instructions.append(Instruction(OpCode.Gt))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1014, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Gt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1015
 *     cpdef Program gt(self):
 *         self.instructions.append(Instruction(OpCode.Gt))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program lt(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1013
 *         return self
 * 
 *     cpdef Program gt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Gt))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.gt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_83gt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_83gt = {"gt", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_83gt, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_83gt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gt (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("gt", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "gt", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_82gt(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_82gt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gt", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_gt(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.gt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1017
 *         return self
 * 
 *     cpdef Program lt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lt))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_85lt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_lt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lt", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_85lt)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1017, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1018
 * 
 *     cpdef Program lt(self):
 *         self.instructions.append(Instruction(OpCode.Lt))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1018, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Lt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1019
 *     cpdef Program lt(self):
 *         self.instructions.append(Instruction(OpCode.Lt))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program ge(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1017
 *         return self
 * 
 *     cpdef Program lt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lt))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.lt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_85lt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_85lt = {"lt", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_85lt, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_85lt(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lt (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("lt", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "lt", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_84lt(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_84lt(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lt", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_lt(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.lt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1021
 *         return self
 * 
 *     cpdef Program ge(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ge))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_87ge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_ge(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ge", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_87ge)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1021, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1022
 * 
 *     cpdef Program ge(self):
 *         self.instructions.append(Instruction(OpCode.Ge))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1022, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1023
 *     cpdef Program ge(self):
 *         self.instructions.append(Instruction(OpCode.Ge))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program le(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1021
 *         return self
 * 
 *     cpdef Program ge(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ge))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.ge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_87ge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_87ge = {"ge", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_87ge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_87ge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ge (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("ge", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "ge", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_86ge(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_86ge(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ge", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_ge(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.ge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1025
 *         return self
 * 
 *     cpdef Program le(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Le))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_89le(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_le(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("le", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_le); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_89le)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1025, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1026
 * 
 *     cpdef Program le(self):
 *         self.instructions.append(Instruction(OpCode.Le))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1026, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Le); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1027
 *     cpdef Program le(self):
 *         self.instructions.append(Instruction(OpCode.Le))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program xor(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1025
 *         return self
 * 
 *     cpdef Program le(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Le))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.le", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_89le(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_89le = {"le", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_89le, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_89le(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("le (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("le", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "le", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_88le(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_88le(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("le", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_le(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.le", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1029
 *         return self
 * 
 *     cpdef Program xor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Xor))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_91xor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_xor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("xor", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_xor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_91xor)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1029, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1030
 * 
 *     cpdef Program xor(self):
 *         self.instructions.append(Instruction(OpCode.Xor))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1030, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Xor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1031
 *     cpdef Program xor(self):
 *         self.instructions.append(Instruction(OpCode.Xor))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program slice(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1029
 *         return self
 * 
 *     cpdef Program xor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Xor))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.xor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_91xor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_91xor = {"xor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_91xor, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_91xor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("xor", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "xor", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_90xor(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_90xor(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("xor", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_xor(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.xor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1033
 *         return self
 * 
 *     cpdef Program slice(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Slice))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_93slice(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_slice(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("slice", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slice); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_93slice)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1033, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1034
 * 
 *     cpdef Program slice(self):
 *         self.instructions.append(Instruction(OpCode.Slice))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1034, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Slice); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1035
 *     cpdef Program slice(self):
 *         self.instructions.append(Instruction(OpCode.Slice))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program slice_literal(self, Vector value):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1033
 *         return self
 * 
 *     cpdef Program slice(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Slice))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_93slice(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_93slice = {"slice", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_93slice, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_93slice(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("slice (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("slice", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "slice", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_92slice(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_92slice(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("slice", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_slice(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1037
 *         return self
 * 
 *     cpdef Program slice_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         if value.length == 1 and value.numbers != NULL:
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_95slice_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_slice_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("slice_literal", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slice_literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_95slice_literal)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_value)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1037, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1038
 * 
 *     cpdef Program slice_literal(self, Vector value):
 *         if value.length == 1 and value.numbers != NULL:             # <<<<<<<<<<<<<<
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 *         else:
 */
  __pyx_t_7 = (__pyx_v_value->length == 1);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_value->numbers != NULL);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "flitter/language/vm.pyx":1039
 *     cpdef Program slice_literal(self, Vector value):
 *         if value.length == 1 and value.numbers != NULL:
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))             # <<<<<<<<<<<<<<
 *         else:
 *             self.instructions.append(InstructionVector(OpCode.SliceLiteral, value))
 */
    if (unlikely(__pyx_v_self->instructions == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 1039, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_IndexLiteral); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_From_int64_t(((int64_t)floor((__pyx_v_value->numbers[0])))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/vm.pyx":1038
 * 
 *     cpdef Program slice_literal(self, Vector value):
 *         if value.length == 1 and value.numbers != NULL:             # <<<<<<<<<<<<<<
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 *         else:
 */
    goto __pyx_L3;
  }

  /* "flitter/language/vm.pyx":1041
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 *         else:
 *             self.instructions.append(InstructionVector(OpCode.SliceLiteral, value))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    if (unlikely(__pyx_v_self->instructions == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 1041, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_SliceLiteral); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_value);
    __Pyx_GIVEREF((PyObject *)__pyx_v_value);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_value))) __PYX_ERR(0, 1041, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/vm.pyx":1042
 *         else:
 *             self.instructions.append(InstructionVector(OpCode.SliceLiteral, value))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1037
 *         return self
 * 
 *     cpdef Program slice_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         if value.length == 1 and value.numbers != NULL:
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.slice_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_95slice_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_95slice_literal = {"slice_literal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_95slice_literal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_95slice_literal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("slice_literal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "slice_literal") < 0)) __PYX_ERR(0, 1037, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Vector *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("slice_literal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1037, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.slice_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), __pyx_ptype_7flitter_5model_Vector, 1, "value", 0))) __PYX_ERR(0, 1037, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_94slice_literal(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_94slice_literal(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("slice_literal", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_slice_literal(__pyx_v_self, __pyx_v_value, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.slice_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1044
 *         return self
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_97call(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_call(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_call *__pyx_optional_args) {
  PyObject *__pyx_v_names = ((PyObject*)Py_None);
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_names = __pyx_optional_args->names;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_97call)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_names};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1044, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1045
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1045, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Call); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_names)) __PYX_ERR(0, 1045, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1046
 *     cpdef Program call(self, int64_t count, tuple names=None):
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program call_fast(self, function, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1044
 *         return self
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_97call(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_97call = {"call", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_97call, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_97call(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  PyObject *__pyx_v_names = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,&__pyx_n_s_names,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "call") < 0)) __PYX_ERR(0, 1044, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_names = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1044, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_96call(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count, __pyx_v_names);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_96call(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, PyObject *__pyx_v_names) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_call __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.names = __pyx_v_names;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_8language_2vm_Program->call(__pyx_v_self, __pyx_v_count, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1048
 *         return self
 * 
 *     cpdef Program call_fast(self, function, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_99call_fast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_call_fast(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_function, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_fast", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_fast); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_99call_fast)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1048, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_function, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1048, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1048, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1049
 * 
 *     cpdef Program call_fast(self, function, int64_t count):
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1049, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_CallFast); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_function);
  __Pyx_GIVEREF(__pyx_v_function);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_function)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1050
 *     cpdef Program call_fast(self, function, int64_t count):
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program tag(self, str name):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1048
 *         return self
 * 
 *     cpdef Program call_fast(self, function, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.call_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_99call_fast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_99call_fast = {"call_fast", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_99call_fast, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_99call_fast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_function = 0;
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_fast (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_function,&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_function)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("call_fast", 1, 2, 2, 1); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "call_fast") < 0)) __PYX_ERR(0, 1048, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_function = values[0];
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_fast", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1048, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.call_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_98call_fast(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_function, __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_98call_fast(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_function, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_fast", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_call_fast(__pyx_v_self, __pyx_v_function, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.call_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1052
 *         return self
 * 
 *     cpdef Program tag(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_101tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_tag(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tag", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_101tag)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1052, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1053
 * 
 *     cpdef Program tag(self, str name):
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1053, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_name)) __PYX_ERR(0, 1053, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1054
 *     cpdef Program tag(self, str name):
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program attributes(self, tuple names):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1052
 *         return self
 * 
 *     cpdef Program tag(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_101tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_101tag = {"tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_101tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_101tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1052, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "tag") < 0)) __PYX_ERR(0, 1052, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_name = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("tag", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1052, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1052, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_100tag(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_100tag(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tag", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_tag(__pyx_v_self, __pyx_v_name, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1056
 *         return self
 * 
 *     cpdef Program attributes(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_103attributes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_attributes(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("attributes", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_attributes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_103attributes)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_names};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1056, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1057
 * 
 *     cpdef Program attributes(self, tuple names):
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1057, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Attributes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_names)) __PYX_ERR(0, 1057, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1058
 *     cpdef Program attributes(self, tuple names):
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program append(self, int64_t count=1):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1056
 *         return self
 * 
 *     cpdef Program attributes(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_103attributes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_103attributes = {"attributes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_103attributes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_103attributes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_names = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("attributes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_names,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1056, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "attributes") < 0)) __PYX_ERR(0, 1056, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_names = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("attributes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1056, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 1056, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_102attributes(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_names);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_102attributes(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_names) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("attributes", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_attributes(__pyx_v_self, __pyx_v_names, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1060
 *         return self
 * 
 *     cpdef Program append(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Append, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_105append(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_append(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_append *__pyx_optional_args) {
  int64_t __pyx_v_count = ((int64_t)1);
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_count = __pyx_optional_args->count;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_105append)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1060, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1061
 * 
 *     cpdef Program append(self, int64_t count=1):
 *         self.instructions.append(InstructionInt(OpCode.Append, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1061, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1062
 *     cpdef Program append(self, int64_t count=1):
 *         self.instructions.append(InstructionInt(OpCode.Append, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program compose(self, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1060
 *         return self
 * 
 *     cpdef Program append(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Append, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_105append(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_105append = {"append", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_105append, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_105append(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1060, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "append") < 0)) __PYX_ERR(0, 1060, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1060, __pyx_L3_error)
    } else {
      __pyx_v_count = ((int64_t)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("append", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1060, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_104append(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_104append(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_append __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.count = __pyx_v_count;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_8language_2vm_Program->append(__pyx_v_self, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1064
 *         return self
 * 
 *     cpdef Program compose(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_107compose(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_compose(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compose", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_compose); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_107compose)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1064, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1064, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1065
 * 
 *     cpdef Program compose(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1065, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Compose); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1066
 *     cpdef Program compose(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program begin_for(self, int64_t count):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1064
 *         return self
 * 
 *     cpdef Program compose(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.compose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_107compose(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_107compose = {"compose", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_107compose, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_107compose(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compose (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1064, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "compose") < 0)) __PYX_ERR(0, 1064, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1064, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compose", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1064, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.compose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_106compose(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_106compose(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compose", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_compose(__pyx_v_self, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.compose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1068
 *         return self
 * 
 *     cpdef Program begin_for(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_109begin_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_begin_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("begin_for", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_begin_for); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_109begin_for)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1068, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1068, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1069
 * 
 *     cpdef Program begin_for(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1069, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BeginFor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1070
 *     cpdef Program begin_for(self, int64_t count):
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program next(self, int64_t label):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1068
 *         return self
 * 
 *     cpdef Program begin_for(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.begin_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_109begin_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_109begin_for = {"begin_for", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_109begin_for, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_109begin_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_count;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("begin_for (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_count,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_count)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "begin_for") < 0)) __PYX_ERR(0, 1068, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_count = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_count == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("begin_for", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1068, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.begin_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_108begin_for(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_count);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_108begin_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_count) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("begin_for", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_begin_for(__pyx_v_self, __pyx_v_count, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.begin_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1072
 *         return self
 * 
 *     cpdef Program next(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Next, label))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_111next(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_next(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_next); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_111next)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1072, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1073
 * 
 *     cpdef Program next(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.Next, label))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1073, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Next); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1074
 *     cpdef Program next(self, int64_t label):
 *         self.instructions.append(InstructionJump(OpCode.Next, label))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program end_for(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1072
 *         return self
 * 
 *     cpdef Program next(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Next, label))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.vm.Program.next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_111next(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_111next = {"next", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_111next, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_111next(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("next (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "next") < 0)) __PYX_ERR(0, 1072, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("next", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1072, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_110next(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_110next(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_next(__pyx_v_self, __pyx_v_label, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1076
 *         return self
 * 
 *     cpdef Program end_for(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.EndFor))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_113end_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_end_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("end_for", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end_for); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_113end_for)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1076, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1077
 * 
 *     cpdef Program end_for(self):
 *         self.instructions.append(Instruction(OpCode.EndFor))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1077, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_EndFor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1078
 *     cpdef Program end_for(self):
 *         self.instructions.append(Instruction(OpCode.EndFor))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program export(self, str name):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1076
 *         return self
 * 
 *     cpdef Program end_for(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.EndFor))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.end_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_113end_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_113end_for = {"end_for", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_113end_for, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_113end_for(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("end_for (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("end_for", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "end_for", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_112end_for(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_112end_for(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("end_for", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_end_for(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.end_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1080
 *         return self
 * 
 *     cpdef Program export(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Export, name))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_115export(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_export(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("export", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_export); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_115export)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1080, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1080, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1081
 * 
 *     cpdef Program export(self, str name):
 *         self.instructions.append(InstructionStr(OpCode.Export, name))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1081, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Export); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_name)) __PYX_ERR(0, 1081, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1082
 *     cpdef Program export(self, str name):
 *         self.instructions.append(InstructionStr(OpCode.Export, name))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1080
 *         return self
 * 
 *     cpdef Program export(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Export, name))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.export", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_115export(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_115export = {"export", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_115export, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_115export(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("export (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "export") < 0)) __PYX_ERR(0, 1080, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_name = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("export", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1080, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.export", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1080, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_114export(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_114export(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("export", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_export(__pyx_v_self, __pyx_v_name, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.export", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1084
 *         return self
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_117func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_func(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_func *__pyx_optional_args) {
  int64_t __pyx_v_ncaptures = ((int64_t)0);
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("func", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_ncaptures = __pyx_optional_args->ncaptures;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_117func)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_ncaptures); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_t_3, __pyx_v_name, __pyx_v_parameters, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 4+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1084, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1085
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1085, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Func); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_From_int64_t(__pyx_v_ncaptures); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_name)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_v_parameters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_parameters)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":1086
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef Program exit(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1084
 *         return self
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.vm.Program.func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_117func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_117func = {"func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_117func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_117func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_label;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parameters = 0;
  int64_t __pyx_v_ncaptures;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label,&__pyx_n_s_name,&__pyx_n_s_parameters,&__pyx_n_s_ncaptures,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("func", 0, 3, 4, 1); __PYX_ERR(0, 1084, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parameters)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("func", 0, 3, 4, 2); __PYX_ERR(0, 1084, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ncaptures);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "func") < 0)) __PYX_ERR(0, 1084, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_label = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_label == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
    __pyx_v_name = ((PyObject*)values[1]);
    __pyx_v_parameters = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_ncaptures = __Pyx_PyInt_As_int64_t(values[3]); if (unlikely((__pyx_v_ncaptures == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
    } else {
      __pyx_v_ncaptures = ((int64_t)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("func", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 1084, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1084, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parameters), (&PyTuple_Type), 1, "parameters", 1))) __PYX_ERR(0, 1084, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_116func(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v_label, __pyx_v_name, __pyx_v_parameters, __pyx_v_ncaptures);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_116func(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int64_t __pyx_v_label, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, int64_t __pyx_v_ncaptures) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_func __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("func", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.ncaptures = __pyx_v_ncaptures;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_8language_2vm_Program->func(__pyx_v_self, __pyx_v_label, __pyx_v_name, __pyx_v_parameters, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1088
 *         return self
 * 
 *     cpdef Program exit(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Exit))
 *         return self
 */

static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_119exit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_f_7flitter_8language_2vm_7Program_exit(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("exit", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_8language_2vm_7Program_119exit)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(0, 1088, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/language/vm.pyx":1089
 * 
 *     cpdef Program exit(self):
 *         self.instructions.append(Instruction(OpCode.Exit))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (unlikely(__pyx_v_self->instructions == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1089, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->instructions, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":1090
 *     cpdef Program exit(self):
 *         self.instructions.append(Instruction(OpCode.Exit))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "flitter/language/vm.pyx":1088
 *         return self
 * 
 *     cpdef Program exit(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Exit))
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.Program.exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_119exit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_119exit = {"exit", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_119exit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_119exit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("exit", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "exit", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_118exit(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_118exit(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("exit", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_7Program_exit(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7flitter_8language_2vm_7Program_8_execute_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/language/vm.pyx":1454
 *             if instruction is not None:
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))             # <<<<<<<<<<<<<<
 *             raise
 * 
 */

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8_execute_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_2_genexpr(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1454, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_8language_2vm_7Program_8_execute_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Program__execute_locals_genexpr, __pyx_n_s_flitter_language_vm); if (unlikely(!gen)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program._execute.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_8language_2vm_7Program_8_execute_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1454, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 1454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1454, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1454, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1454, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_instruction);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_instruction, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_cur_scope->__pyx_v_instruction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pyx":1092
 *         return self
 * 
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats):             # <<<<<<<<<<<<<<
 *         global CallOutCount, CallOutDuration
 *         cdef VectorStack stack=context.stack, lnames=context.lnames
 */

static void __pyx_f_7flitter_8language_2vm_7Program__execute(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context, int64_t __pyx_v_pc, int __pyx_v_record_stats) {
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *__pyx_cur_scope;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_stack = 0;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *__pyx_v_lnames = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  int64_t __pyx_v_program_end;
  PyObject *__pyx_v_exports = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_loop_sources = 0;
  struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v_loop_source = 0;
  double __pyx_v_duration;
  double __pyx_v_call_duration;
  PyObject *__pyx_v_instructions = 0;
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instruction = 0;
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_names = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_nodes = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_r1 = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_r2 = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_r3 = 0;
  PyObject *__pyx_v_import_names = 0;
  PyObject *__pyx_v_kwargs = 0;
  struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v_function = 0;
  PyObject *__pyx_v_objptr;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_gb_7flitter_8language_2vm_7Program_8_execute_2generator1 = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7;
  int64_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int64_t __pyx_t_13;
  Py_UCS4 __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_t_24;
  char const *__pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  char const *__pyx_t_32;
  int __pyx_t_33;
  unsigned int __pyx_t_34;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_execute", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_1__execute(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1092, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_pc = __pyx_v_pc;

  /* "flitter/language/vm.pyx":1094
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats):
 *         global CallOutCount, CallOutDuration
 *         cdef VectorStack stack=context.stack, lnames=context.lnames             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n, program_end=len(self.instructions)
 *         cdef dict exports=context.exports, state=context.state._state
 */
  if (!(likely(((__pyx_v_context->stack) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->stack, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_context->stack;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (!(likely(((__pyx_v_context->lnames) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_context->lnames, __pyx_ptype_7flitter_8language_2vm_VectorStack))))) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_context->lnames;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1095
 *         global CallOutCount, CallOutDuration
 *         cdef VectorStack stack=context.stack, lnames=context.lnames
 *         cdef int64_t i, n, program_end=len(self.instructions)             # <<<<<<<<<<<<<<
 *         cdef dict exports=context.exports, state=context.state._state
 *         cdef list loop_sources=[]
 */
  __pyx_t_1 = __pyx_v_self->instructions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1095, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_program_end = __pyx_t_2;

  /* "flitter/language/vm.pyx":1096
 *         cdef VectorStack stack=context.stack, lnames=context.lnames
 *         cdef int64_t i, n, program_end=len(self.instructions)
 *         cdef dict exports=context.exports, state=context.state._state             # <<<<<<<<<<<<<<
 *         cdef list loop_sources=[]
 *         cdef LoopSource loop_source = None
 */
  __pyx_t_1 = __pyx_v_context->exports;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_exports = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_v_context->state->_state;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1097
 *         cdef int64_t i, n, program_end=len(self.instructions)
 *         cdef dict exports=context.exports, state=context.state._state
 *         cdef list loop_sources=[]             # <<<<<<<<<<<<<<
 *         cdef LoopSource loop_source = None
 *         cdef double duration, call_duration
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_loop_sources = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1098
 *         cdef dict exports=context.exports, state=context.state._state
 *         cdef list loop_sources=[]
 *         cdef LoopSource loop_source = None             # <<<<<<<<<<<<<<
 *         cdef double duration, call_duration
 * 
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_loop_source = ((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)Py_None);

  /* "flitter/language/vm.pyx":1101
 *         cdef double duration, call_duration
 * 
 *         cdef list instructions=self.instructions             # <<<<<<<<<<<<<<
 *         cdef Instruction instruction=None
 *         cdef str filename
 */
  __pyx_t_1 = __pyx_v_self->instructions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_instructions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/vm.pyx":1102
 * 
 *         cdef list instructions=self.instructions
 *         cdef Instruction instruction=None             # <<<<<<<<<<<<<<
 *         cdef str filename
 *         cdef tuple names, args, nodes
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_instruction = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)Py_None);

  /* "flitter/language/vm.pyx":1110
 *         cdef PyObject* objptr
 * 
 *         assert self.linked, "Program has not been linked"             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    if (unlikely(!__pyx_v_self->linked)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Program_has_not_been_linked, 0, 0);
      __PYX_ERR(0, 1110, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 1110, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":1112
 *         assert self.linked, "Program has not been linked"
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while 0 <= pc < program_end:
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "flitter/language/vm.pyx":1113
 * 
 *         try:
 *             while 0 <= pc < program_end:             # <<<<<<<<<<<<<<
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 *                 pc += 1
 */
      while (1) {
        __pyx_t_6 = (0 <= __pyx_cur_scope->__pyx_v_pc);
        if (__pyx_t_6) {
          __pyx_t_6 = (__pyx_cur_scope->__pyx_v_pc < __pyx_v_program_end);
        }
        if (!__pyx_t_6) break;

        /* "flitter/language/vm.pyx":1114
 *         try:
 *             while 0 <= pc < program_end:
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)             # <<<<<<<<<<<<<<
 *                 pc += 1
 *                 if record_stats:
 */
        __pyx_t_7 = PyList_GET_ITEM(__pyx_v_instructions, __pyx_cur_scope->__pyx_v_pc);
        __pyx_t_1 = ((PyObject *)__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_instruction, ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/vm.pyx":1115
 *             while 0 <= pc < program_end:
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 *                 pc += 1             # <<<<<<<<<<<<<<
 *                 if record_stats:
 *                     duration = -perf_counter()
 */
        __pyx_cur_scope->__pyx_v_pc = (__pyx_cur_scope->__pyx_v_pc + 1);

        /* "flitter/language/vm.pyx":1116
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 *                 pc += 1
 *                 if record_stats:             # <<<<<<<<<<<<<<
 *                     duration = -perf_counter()
 * 
 */
        if (__pyx_v_record_stats) {

          /* "flitter/language/vm.pyx":1117
 *                 pc += 1
 *                 if record_stats:
 *                     duration = -perf_counter()             # <<<<<<<<<<<<<<
 * 
 *                 if instruction.code == OpCode.Dup:
 */
          __pyx_v_duration = (-perf_counter());

          /* "flitter/language/vm.pyx":1116
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 *                 pc += 1
 *                 if record_stats:             # <<<<<<<<<<<<<<
 *                     duration = -perf_counter()
 * 
 */
        }

        /* "flitter/language/vm.pyx":1119
 *                     duration = -perf_counter()
 * 
 *                 if instruction.code == OpCode.Dup:             # <<<<<<<<<<<<<<
 *                     push(stack, peek(stack))
 * 
 */
        switch (__pyx_v_instruction->code) {
          case __pyx_e_7flitter_8language_2vm_Dup:

          /* "flitter/language/vm.pyx":1120
 * 
 *                 if instruction.code == OpCode.Dup:
 *                     push(stack, peek(stack))             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Drop:
 */
          __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1120, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "flitter/language/vm.pyx":1119
 *                     duration = -perf_counter()
 * 
 *                 if instruction.code == OpCode.Dup:             # <<<<<<<<<<<<<<
 *                     push(stack, peek(stack))
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Drop:

          /* "flitter/language/vm.pyx":1123
 * 
 *                 elif instruction.code == OpCode.Drop:
 *                     drop(stack, (<InstructionInt>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Label:
 */
          __pyx_f_7flitter_8language_2vm_drop(__pyx_v_stack, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1122
 *                     push(stack, peek(stack))
 * 
 *                 elif instruction.code == OpCode.Drop:             # <<<<<<<<<<<<<<
 *                     drop(stack, (<InstructionInt>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Label:

          /* "flitter/language/vm.pyx":1125
 *                     drop(stack, (<InstructionInt>instruction).value)
 * 
 *                 elif instruction.code == OpCode.Label:             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Jump:

          /* "flitter/language/vm.pyx":1129
 * 
 *                 elif instruction.code == OpCode.Jump:
 *                     pc += (<InstructionJump>instruction).offset             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Exit:
 */
          __pyx_cur_scope->__pyx_v_pc = (__pyx_cur_scope->__pyx_v_pc + ((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_instruction)->offset);

          /* "flitter/language/vm.pyx":1128
 *                     pass
 * 
 *                 elif instruction.code == OpCode.Jump:             # <<<<<<<<<<<<<<
 *                     pc += (<InstructionJump>instruction).offset
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Exit:

          /* "flitter/language/vm.pyx":1132
 * 
 *                 elif instruction.code == OpCode.Exit:
 *                     pc = program_end             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.BranchTrue:
 */
          __pyx_cur_scope->__pyx_v_pc = __pyx_v_program_end;

          /* "flitter/language/vm.pyx":1131
 *                     pc += (<InstructionJump>instruction).offset
 * 
 *                 elif instruction.code == OpCode.Exit:             # <<<<<<<<<<<<<<
 *                     pc = program_end
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_BranchTrue:

          /* "flitter/language/vm.pyx":1135
 * 
 *                 elif instruction.code == OpCode.BranchTrue:
 *                     if pop(stack).as_bool():             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 * 
 */
          __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1135, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1136
 *                 elif instruction.code == OpCode.BranchTrue:
 *                     if pop(stack).as_bool():
 *                         pc += (<InstructionJump>instruction).offset             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.BranchFalse:
 */
            __pyx_cur_scope->__pyx_v_pc = (__pyx_cur_scope->__pyx_v_pc + ((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_instruction)->offset);

            /* "flitter/language/vm.pyx":1135
 * 
 *                 elif instruction.code == OpCode.BranchTrue:
 *                     if pop(stack).as_bool():             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 * 
 */
          }

          /* "flitter/language/vm.pyx":1134
 *                     pc = program_end
 * 
 *                 elif instruction.code == OpCode.BranchTrue:             # <<<<<<<<<<<<<<
 *                     if pop(stack).as_bool():
 *                         pc += (<InstructionJump>instruction).offset
 */
          break;
          case __pyx_e_7flitter_8language_2vm_BranchFalse:

          /* "flitter/language/vm.pyx":1139
 * 
 *                 elif instruction.code == OpCode.BranchFalse:
 *                     if not pop(stack).as_bool():             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 * 
 */
          __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = (!__pyx_t_6);
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1140
 *                 elif instruction.code == OpCode.BranchFalse:
 *                     if not pop(stack).as_bool():
 *                         pc += (<InstructionJump>instruction).offset             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Import:
 */
            __pyx_cur_scope->__pyx_v_pc = (__pyx_cur_scope->__pyx_v_pc + ((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_instruction)->offset);

            /* "flitter/language/vm.pyx":1139
 * 
 *                 elif instruction.code == OpCode.BranchFalse:
 *                     if not pop(stack).as_bool():             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 * 
 */
          }

          /* "flitter/language/vm.pyx":1138
 *                         pc += (<InstructionJump>instruction).offset
 * 
 *                 elif instruction.code == OpCode.BranchFalse:             # <<<<<<<<<<<<<<
 *                     if not pop(stack).as_bool():
 *                         pc += (<InstructionJump>instruction).offset
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Import:

          /* "flitter/language/vm.pyx":1143
 * 
 *                 elif instruction.code == OpCode.Import:
 *                     filename = pop(stack).as_string()             # <<<<<<<<<<<<<<
 *                     names = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(names)
 */
          __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->__pyx_vtab)->as_string(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1143, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_filename, ((PyObject*)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1144
 *                 elif instruction.code == OpCode.Import:
 *                     filename = pop(stack).as_string()
 *                     names = (<InstructionTuple>instruction).value             # <<<<<<<<<<<<<<
 *                     n = PyTuple_GET_SIZE(names)
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)
 */
          __pyx_t_10 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_instruction)->value;
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_names, ((PyObject*)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1145
 *                     filename = pop(stack).as_string()
 *                     names = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(names)             # <<<<<<<<<<<<<<
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)
 *                     if import_names is not None:
 */
          __pyx_v_n = PyTuple_GET_SIZE(__pyx_v_names);

          /* "flitter/language/vm.pyx":1146
 *                     names = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(names)
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)             # <<<<<<<<<<<<<<
 *                     if import_names is not None:
 *                         for i in range(n):
 */
          __pyx_t_10 = __pyx_f_7flitter_8language_2vm_import_module(__pyx_v_context, __pyx_v_filename, __pyx_v_record_stats, (&__pyx_v_duration), __pyx_v_self->simplify); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1146, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_import_names, ((PyObject*)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1147
 *                     n = PyTuple_GET_SIZE(names)
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)
 *                     if import_names is not None:             # <<<<<<<<<<<<<<
 *                         for i in range(n):
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 */
          __pyx_t_9 = (__pyx_v_import_names != ((PyObject*)Py_None));
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1148
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)
 *                     if import_names is not None:
 *                         for i in range(n):             # <<<<<<<<<<<<<<
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 *                             if objptr != NULL:
 */
            __pyx_t_8 = __pyx_v_n;
            __pyx_t_11 = __pyx_t_8;
            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
              __pyx_v_i = __pyx_t_12;

              /* "flitter/language/vm.pyx":1149
 *                     if import_names is not None:
 *                         for i in range(n):
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))             # <<<<<<<<<<<<<<
 *                             if objptr != NULL:
 *                                 push(lnames, <Vector>objptr)
 */
              __pyx_t_7 = PyTuple_GET_ITEM(__pyx_v_names, __pyx_v_i);
              __pyx_v_objptr = PyDict_GetItem(__pyx_v_import_names, ((PyObject *)__pyx_t_7));

              /* "flitter/language/vm.pyx":1150
 *                         for i in range(n):
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 *                             if objptr != NULL:             # <<<<<<<<<<<<<<
 *                                 push(lnames, <Vector>objptr)
 *                             else:
 */
              __pyx_t_9 = (__pyx_v_objptr != NULL);
              if (__pyx_t_9) {

                /* "flitter/language/vm.pyx":1151
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 *                             if objptr != NULL:
 *                                 push(lnames, <Vector>objptr)             # <<<<<<<<<<<<<<
 *                             else:
 *                                 PySet_Add(context.errors, f"Unable to import '{<str>PyTuple_GET_ITEM(names, i)}' from '{filename}'")
 */
                __pyx_t_13 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_objptr)); if (unlikely(__pyx_t_13 == ((int64_t)0))) __PYX_ERR(0, 1151, __pyx_L3_error)

                /* "flitter/language/vm.pyx":1150
 *                         for i in range(n):
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 *                             if objptr != NULL:             # <<<<<<<<<<<<<<
 *                                 push(lnames, <Vector>objptr)
 *                             else:
 */
                goto __pyx_L17;
              }

              /* "flitter/language/vm.pyx":1153
 *                                 push(lnames, <Vector>objptr)
 *                             else:
 *                                 PySet_Add(context.errors, f"Unable to import '{<str>PyTuple_GET_ITEM(names, i)}' from '{filename}'")             # <<<<<<<<<<<<<<
 *                                 push(lnames, null_)
 *                     else:
 */
              /*else*/ {
                __pyx_t_10 = __pyx_v_context->errors;
                __Pyx_INCREF(__pyx_t_10);
                __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L3_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_2 = 0;
                __pyx_t_14 = 127;
                __Pyx_INCREF(__pyx_kp_u_Unable_to_import);
                __pyx_t_2 += 18;
                __Pyx_GIVEREF(__pyx_kp_u_Unable_to_import);
                PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Unable_to_import);
                __pyx_t_7 = PyTuple_GET_ITEM(__pyx_v_names, __pyx_v_i);
                __pyx_t_15 = __Pyx_PyUnicode_Unicode(((PyObject *)__pyx_t_7)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1153, __pyx_L3_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_14;
                __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_15);
                __pyx_t_15 = 0;
                __Pyx_INCREF(__pyx_kp_u_from);
                __pyx_t_2 += 8;
                __Pyx_GIVEREF(__pyx_kp_u_from);
                PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_from);
                __pyx_t_15 = __Pyx_PyUnicode_Unicode(__pyx_v_filename); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1153, __pyx_L3_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_14;
                __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_15);
                __pyx_t_15 = 0;
                __Pyx_INCREF(__pyx_kp_u__9);
                __pyx_t_2 += 1;
                __Pyx_GIVEREF(__pyx_kp_u__9);
                PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__9);
                __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1153, __pyx_L3_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_t_16 = PySet_Add(__pyx_t_10, __pyx_t_15); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1153, __pyx_L3_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                /* "flitter/language/vm.pyx":1154
 *                             else:
 *                                 PySet_Add(context.errors, f"Unable to import '{<str>PyTuple_GET_ITEM(names, i)}' from '{filename}'")
 *                                 push(lnames, null_)             # <<<<<<<<<<<<<<
 *                     else:
 *                         for i in range(n):
 */
                __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_null_);
                __Pyx_INCREF(__pyx_t_15);
                __pyx_t_13 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_13 == ((int64_t)0))) __PYX_ERR(0, 1154, __pyx_L3_error)
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              }
              __pyx_L17:;
            }

            /* "flitter/language/vm.pyx":1147
 *                     n = PyTuple_GET_SIZE(names)
 *                     import_names = import_module(context, filename, record_stats, &duration, self.simplify)
 *                     if import_names is not None:             # <<<<<<<<<<<<<<
 *                         for i in range(n):
 *                             objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
 */
            goto __pyx_L14;
          }

          /* "flitter/language/vm.pyx":1156
 *                                 push(lnames, null_)
 *                     else:
 *                         for i in range(n):             # <<<<<<<<<<<<<<
 *                             push(lnames, null_)
 *                     filename = names = import_names = None
 */
          /*else*/ {
            __pyx_t_8 = __pyx_v_n;
            __pyx_t_11 = __pyx_t_8;
            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
              __pyx_v_i = __pyx_t_12;

              /* "flitter/language/vm.pyx":1157
 *                     else:
 *                         for i in range(n):
 *                             push(lnames, null_)             # <<<<<<<<<<<<<<
 *                     filename = names = import_names = None
 * 
 */
              __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_null_);
              __Pyx_INCREF(__pyx_t_15);
              __pyx_t_13 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_13 == ((int64_t)0))) __PYX_ERR(0, 1157, __pyx_L3_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
          }
          __pyx_L14:;

          /* "flitter/language/vm.pyx":1158
 *                         for i in range(n):
 *                             push(lnames, null_)
 *                     filename = names = import_names = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Literal:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_filename, ((PyObject*)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_names, ((PyObject*)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_import_names, ((PyObject*)Py_None));

          /* "flitter/language/vm.pyx":1142
 *                         pc += (<InstructionJump>instruction).offset
 * 
 *                 elif instruction.code == OpCode.Import:             # <<<<<<<<<<<<<<
 *                     filename = pop(stack).as_string()
 *                     names = (<InstructionTuple>instruction).value
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Literal:

          /* "flitter/language/vm.pyx":1161
 * 
 *                 elif instruction.code == OpCode.Literal:
 *                     push(stack, (<InstructionVector>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LiteralNode:
 */
          __pyx_t_15 = ((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instruction)->value);
          __Pyx_INCREF(__pyx_t_15);
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1161, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1160
 *                     filename = names = import_names = None
 * 
 *                 elif instruction.code == OpCode.Literal:             # <<<<<<<<<<<<<<
 *                     push(stack, (<InstructionVector>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LiteralNode:

          /* "flitter/language/vm.pyx":1164
 * 
 *                 elif instruction.code == OpCode.LiteralNode:
 *                     r1 = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *                     r1.objects = ((<InstructionNode>instruction).value,)
 *                     r1.length = 1
 */
          __pyx_t_15 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1164, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (!(likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(0, 1164, __pyx_L3_error)
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1165
 *                 elif instruction.code == OpCode.LiteralNode:
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = ((<InstructionNode>instruction).value,)             # <<<<<<<<<<<<<<
 *                     r1.length = 1
 *                     push(stack, r1)
 */
          __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1165, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_instruction)->value);
          __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_instruction)->value);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v_instruction)->value))) __PYX_ERR(0, 1165, __pyx_L3_error);
          __Pyx_GIVEREF(__pyx_t_15);
          __Pyx_GOTREF(__pyx_v_r1->objects);
          __Pyx_DECREF(__pyx_v_r1->objects);
          __pyx_v_r1->objects = ((PyObject*)__pyx_t_15);
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1166
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = ((<InstructionNode>instruction).value,)
 *                     r1.length = 1             # <<<<<<<<<<<<<<
 *                     push(stack, r1)
 *                     r1 = None
 */
          __pyx_v_r1->length = 1;

          /* "flitter/language/vm.pyx":1167
 *                     r1.objects = ((<InstructionNode>instruction).value,)
 *                     r1.length = 1
 *                     push(stack, r1)             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, __pyx_v_r1); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1167, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1168
 *                     r1.length = 1
 *                     push(stack, r1)
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LiteralNodes:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1163
 *                     push(stack, (<InstructionVector>instruction).value)
 * 
 *                 elif instruction.code == OpCode.LiteralNode:             # <<<<<<<<<<<<<<
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = ((<InstructionNode>instruction).value,)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LiteralNodes:

          /* "flitter/language/vm.pyx":1171
 * 
 *                 elif instruction.code == OpCode.LiteralNodes:
 *                     nodes = (<InstructionTuple>instruction).value             # <<<<<<<<<<<<<<
 *                     n = PyTuple_GET_SIZE(nodes)
 *                     r1 = Vector.__new__(Vector)
 */
          __pyx_t_15 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_instruction)->value;
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_nodes, ((PyObject*)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1172
 *                 elif instruction.code == OpCode.LiteralNodes:
 *                     nodes = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(nodes)             # <<<<<<<<<<<<<<
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = PyTuple_New(n)
 */
          __pyx_v_n = PyTuple_GET_SIZE(__pyx_v_nodes);

          /* "flitter/language/vm.pyx":1173
 *                     nodes = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(nodes)
 *                     r1 = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *                     r1.objects = PyTuple_New(n)
 *                     for i in range(n):
 */
          __pyx_t_15 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1173, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (!(likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(0, 1173, __pyx_L3_error)
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1174
 *                     n = PyTuple_GET_SIZE(nodes)
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = PyTuple_New(n)             # <<<<<<<<<<<<<<
 *                     for i in range(n):
 *                         objptr = PyTuple_GET_ITEM(nodes, i)
 */
          __pyx_t_15 = PyTuple_New(__pyx_v_n); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1174, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_GIVEREF(__pyx_t_15);
          __Pyx_GOTREF(__pyx_v_r1->objects);
          __Pyx_DECREF(__pyx_v_r1->objects);
          __pyx_v_r1->objects = ((PyObject*)__pyx_t_15);
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1175
 *                     r1 = Vector.__new__(Vector)
 *                     r1.objects = PyTuple_New(n)
 *                     for i in range(n):             # <<<<<<<<<<<<<<
 *                         objptr = PyTuple_GET_ITEM(nodes, i)
 *                         Py_INCREF(<object>objptr)
 */
          __pyx_t_8 = __pyx_v_n;
          __pyx_t_11 = __pyx_t_8;
          for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
            __pyx_v_i = __pyx_t_12;

            /* "flitter/language/vm.pyx":1176
 *                     r1.objects = PyTuple_New(n)
 *                     for i in range(n):
 *                         objptr = PyTuple_GET_ITEM(nodes, i)             # <<<<<<<<<<<<<<
 *                         Py_INCREF(<object>objptr)
 *                         PyTuple_SET_ITEM(r1.objects, i, <object>objptr)
 */
            __pyx_v_objptr = PyTuple_GET_ITEM(__pyx_v_nodes, __pyx_v_i);

            /* "flitter/language/vm.pyx":1177
 *                     for i in range(n):
 *                         objptr = PyTuple_GET_ITEM(nodes, i)
 *                         Py_INCREF(<object>objptr)             # <<<<<<<<<<<<<<
 *                         PyTuple_SET_ITEM(r1.objects, i, <object>objptr)
 *                     r1.length = n
 */
            Py_INCREF(((PyObject *)__pyx_v_objptr));

            /* "flitter/language/vm.pyx":1178
 *                         objptr = PyTuple_GET_ITEM(nodes, i)
 *                         Py_INCREF(<object>objptr)
 *                         PyTuple_SET_ITEM(r1.objects, i, <object>objptr)             # <<<<<<<<<<<<<<
 *                     r1.length = n
 *                     push(stack, r1)
 */
            __pyx_t_15 = __pyx_v_r1->objects;
            __Pyx_INCREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_15, __pyx_v_i, ((PyObject *)__pyx_v_objptr));
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }

          /* "flitter/language/vm.pyx":1179
 *                         Py_INCREF(<object>objptr)
 *                         PyTuple_SET_ITEM(r1.objects, i, <object>objptr)
 *                     r1.length = n             # <<<<<<<<<<<<<<
 *                     push(stack, r1)
 *                     r1 = nodes = None
 */
          __pyx_v_r1->length = __pyx_v_n;

          /* "flitter/language/vm.pyx":1180
 *                         PyTuple_SET_ITEM(r1.objects, i, <object>objptr)
 *                     r1.length = n
 *                     push(stack, r1)             # <<<<<<<<<<<<<<
 *                     r1 = nodes = None
 * 
 */
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, __pyx_v_r1); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1180, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1181
 *                     r1.length = n
 *                     push(stack, r1)
 *                     r1 = nodes = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LocalDrop:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_nodes, ((PyObject*)Py_None));

          /* "flitter/language/vm.pyx":1170
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.LiteralNodes:             # <<<<<<<<<<<<<<
 *                     nodes = (<InstructionTuple>instruction).value
 *                     n = PyTuple_GET_SIZE(nodes)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LocalDrop:

          /* "flitter/language/vm.pyx":1184
 * 
 *                 elif instruction.code == OpCode.LocalDrop:
 *                     drop(lnames, (<InstructionInt>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LocalLoad:
 */
          __pyx_f_7flitter_8language_2vm_drop(__pyx_v_lnames, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1183
 *                     r1 = nodes = None
 * 
 *                 elif instruction.code == OpCode.LocalDrop:             # <<<<<<<<<<<<<<
 *                     drop(lnames, (<InstructionInt>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LocalLoad:

          /* "flitter/language/vm.pyx":1187
 * 
 *                 elif instruction.code == OpCode.LocalLoad:
 *                     r1 = peek_at(lnames, (<InstructionInt>instruction).value)             # <<<<<<<<<<<<<<
 *                     push(stack, r1.copy() if r1.objects is not None else r1)
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek_at(__pyx_v_lnames, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1187, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1188
 *                 elif instruction.code == OpCode.LocalLoad:
 *                     r1 = peek_at(lnames, (<InstructionInt>instruction).value)
 *                     push(stack, r1.copy() if r1.objects is not None else r1)             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_9 = (__pyx_v_r1->objects != ((PyObject*)Py_None));
          if (__pyx_t_9) {
            __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_r1->__pyx_vtab)->copy(__pyx_v_r1, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1188, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_15 = __pyx_t_10;
            __pyx_t_10 = 0;
          } else {
            __Pyx_INCREF((PyObject *)__pyx_v_r1);
            __pyx_t_15 = ((PyObject *)__pyx_v_r1);
          }
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1188, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1189
 *                     r1 = peek_at(lnames, (<InstructionInt>instruction).value)
 *                     push(stack, r1.copy() if r1.objects is not None else r1)
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LocalPush:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1186
 *                     drop(lnames, (<InstructionInt>instruction).value)
 * 
 *                 elif instruction.code == OpCode.LocalLoad:             # <<<<<<<<<<<<<<
 *                     r1 = peek_at(lnames, (<InstructionInt>instruction).value)
 *                     push(stack, r1.copy() if r1.objects is not None else r1)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LocalPush:

          /* "flitter/language/vm.pyx":1192
 * 
 *                 elif instruction.code == OpCode.LocalPush:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     n = (<InstructionInt>instruction).value
 *                     if n == 1:
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1192, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1193
 *                 elif instruction.code == OpCode.LocalPush:
 *                     r1 = pop(stack)
 *                     n = (<InstructionInt>instruction).value             # <<<<<<<<<<<<<<
 *                     if n == 1:
 *                         push(lnames, r1)
 */
          __pyx_t_8 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value;
          __pyx_v_n = __pyx_t_8;

          /* "flitter/language/vm.pyx":1194
 *                     r1 = pop(stack)
 *                     n = (<InstructionInt>instruction).value
 *                     if n == 1:             # <<<<<<<<<<<<<<
 *                         push(lnames, r1)
 *                     else:
 */
          __pyx_t_9 = (__pyx_v_n == 1);
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1195
 *                     n = (<InstructionInt>instruction).value
 *                     if n == 1:
 *                         push(lnames, r1)             # <<<<<<<<<<<<<<
 *                     else:
 *                         for i in range(n):
 */
            __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, __pyx_v_r1); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1195, __pyx_L3_error)

            /* "flitter/language/vm.pyx":1194
 *                     r1 = pop(stack)
 *                     n = (<InstructionInt>instruction).value
 *                     if n == 1:             # <<<<<<<<<<<<<<
 *                         push(lnames, r1)
 *                     else:
 */
            goto __pyx_L22;
          }

          /* "flitter/language/vm.pyx":1197
 *                         push(lnames, r1)
 *                     else:
 *                         for i in range(n):             # <<<<<<<<<<<<<<
 *                             push(lnames, r1.item(i))
 *                     r1 = None
 */
          /*else*/ {
            __pyx_t_8 = __pyx_v_n;
            __pyx_t_11 = __pyx_t_8;
            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
              __pyx_v_i = __pyx_t_12;

              /* "flitter/language/vm.pyx":1198
 *                     else:
 *                         for i in range(n):
 *                             push(lnames, r1.item(i))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
              __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_r1->__pyx_vtab)->item(__pyx_v_r1, __pyx_v_i, 0)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1198, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_13 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_13 == ((int64_t)0))) __PYX_ERR(0, 1198, __pyx_L3_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
          }
          __pyx_L22:;

          /* "flitter/language/vm.pyx":1199
 *                         for i in range(n):
 *                             push(lnames, r1.item(i))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Lookup:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1191
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.LocalPush:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     n = (<InstructionInt>instruction).value
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Lookup:

          /* "flitter/language/vm.pyx":1202
 * 
 *                 elif instruction.code == OpCode.Lookup:
 *                     objptr = PyDict_GetItem(state, peek(stack))             # <<<<<<<<<<<<<<
 *                     if objptr != NULL:
 *                         poke(stack, <Vector>objptr)
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1202, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_v_objptr = PyDict_GetItem(__pyx_v_state, __pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1203
 *                 elif instruction.code == OpCode.Lookup:
 *                     objptr = PyDict_GetItem(state, peek(stack))
 *                     if objptr != NULL:             # <<<<<<<<<<<<<<
 *                         poke(stack, <Vector>objptr)
 *                     else:
 */
          __pyx_t_9 = (__pyx_v_objptr != NULL);
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1204
 *                     objptr = PyDict_GetItem(state, peek(stack))
 *                     if objptr != NULL:
 *                         poke(stack, <Vector>objptr)             # <<<<<<<<<<<<<<
 *                     else:
 *                         poke(stack, null_)
 */
            __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_objptr)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1204, __pyx_L3_error)

            /* "flitter/language/vm.pyx":1203
 *                 elif instruction.code == OpCode.Lookup:
 *                     objptr = PyDict_GetItem(state, peek(stack))
 *                     if objptr != NULL:             # <<<<<<<<<<<<<<
 *                         poke(stack, <Vector>objptr)
 *                     else:
 */
            goto __pyx_L25;
          }

          /* "flitter/language/vm.pyx":1206
 *                         poke(stack, <Vector>objptr)
 *                     else:
 *                         poke(stack, null_)             # <<<<<<<<<<<<<<
 *                     objptr = NULL
 * 
 */
          /*else*/ {
            __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_null_);
            __Pyx_INCREF(__pyx_t_15);
            __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __pyx_L25:;

          /* "flitter/language/vm.pyx":1207
 *                     else:
 *                         poke(stack, null_)
 *                     objptr = NULL             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.LookupLiteral:
 */
          __pyx_v_objptr = NULL;

          /* "flitter/language/vm.pyx":1201
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Lookup:             # <<<<<<<<<<<<<<
 *                     objptr = PyDict_GetItem(state, peek(stack))
 *                     if objptr != NULL:
 */
          break;
          case __pyx_e_7flitter_8language_2vm_LookupLiteral:

          /* "flitter/language/vm.pyx":1210
 * 
 *                 elif instruction.code == OpCode.LookupLiteral:
 *                     objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)             # <<<<<<<<<<<<<<
 *                     if objptr != NULL:
 *                         push(stack, <Vector>objptr)
 */
          __pyx_t_15 = ((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instruction)->value);
          __Pyx_INCREF(__pyx_t_15);
          __pyx_v_objptr = PyDict_GetItem(__pyx_v_state, __pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1211
 *                 elif instruction.code == OpCode.LookupLiteral:
 *                     objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)
 *                     if objptr != NULL:             # <<<<<<<<<<<<<<
 *                         push(stack, <Vector>objptr)
 *                     else:
 */
          __pyx_t_9 = (__pyx_v_objptr != NULL);
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1212
 *                     objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)
 *                     if objptr != NULL:
 *                         push(stack, <Vector>objptr)             # <<<<<<<<<<<<<<
 *                     else:
 *                         push(stack, null_)
 */
            __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_objptr)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1212, __pyx_L3_error)

            /* "flitter/language/vm.pyx":1211
 *                 elif instruction.code == OpCode.LookupLiteral:
 *                     objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)
 *                     if objptr != NULL:             # <<<<<<<<<<<<<<
 *                         push(stack, <Vector>objptr)
 *                     else:
 */
            goto __pyx_L26;
          }

          /* "flitter/language/vm.pyx":1214
 *                         push(stack, <Vector>objptr)
 *                     else:
 *                         push(stack, null_)             # <<<<<<<<<<<<<<
 *                     objptr = NULL
 * 
 */
          /*else*/ {
            __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_null_);
            __Pyx_INCREF(__pyx_t_15);
            __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1214, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __pyx_L26:;

          /* "flitter/language/vm.pyx":1215
 *                     else:
 *                         push(stack, null_)
 *                     objptr = NULL             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Range:
 */
          __pyx_v_objptr = NULL;

          /* "flitter/language/vm.pyx":1209
 *                     objptr = NULL
 * 
 *                 elif instruction.code == OpCode.LookupLiteral:             # <<<<<<<<<<<<<<
 *                     objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)
 *                     if objptr != NULL:
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Range:

          /* "flitter/language/vm.pyx":1218
 * 
 *                 elif instruction.code == OpCode.Range:
 *                     r3 = pop(stack)             # <<<<<<<<<<<<<<
 *                     r2 = pop(stack)
 *                     r1 = Vector.__new__(Vector)
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1218, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r3, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1219
 *                 elif instruction.code == OpCode.Range:
 *                     r3 = pop(stack)
 *                     r2 = pop(stack)             # <<<<<<<<<<<<<<
 *                     r1 = Vector.__new__(Vector)
 *                     r1.fill_range(peek(stack), r2, r3)
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1219, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1220
 *                     r3 = pop(stack)
 *                     r2 = pop(stack)
 *                     r1 = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *                     r1.fill_range(peek(stack), r2, r3)
 *                     poke(stack, r1)
 */
          __pyx_t_15 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1220, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (!(likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(0, 1220, __pyx_L3_error)
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1221
 *                     r2 = pop(stack)
 *                     r1 = Vector.__new__(Vector)
 *                     r1.fill_range(peek(stack), r2, r3)             # <<<<<<<<<<<<<<
 *                     poke(stack, r1)
 *                     r1 = r2 = r3 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1221, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_r1->__pyx_vtab)->fill_range(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r2, __pyx_v_r3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1221, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1222
 *                     r1 = Vector.__new__(Vector)
 *                     r1.fill_range(peek(stack), r2, r3)
 *                     poke(stack, r1)             # <<<<<<<<<<<<<<
 *                     r1 = r2 = r3 = None
 * 
 */
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, __pyx_v_r1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1222, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1223
 *                     r1.fill_range(peek(stack), r2, r3)
 *                     poke(stack, r1)
 *                     r1 = r2 = r3 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Neg:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r3, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1217
 *                     objptr = NULL
 * 
 *                 elif instruction.code == OpCode.Range:             # <<<<<<<<<<<<<<
 *                     r3 = pop(stack)
 *                     r2 = pop(stack)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Neg:

          /* "flitter/language/vm.pyx":1226
 * 
 *                 elif instruction.code == OpCode.Neg:
 *                     poke(stack, peek(stack).neg())             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Pos:
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1226, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->neg(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15))); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1226, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1226, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1225
 *                     r1 = r2 = r3 = None
 * 
 *                 elif instruction.code == OpCode.Neg:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).neg())
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Pos:

          /* "flitter/language/vm.pyx":1229
 * 
 *                 elif instruction.code == OpCode.Pos:
 *                     if peek(stack).objects is not None:             # <<<<<<<<<<<<<<
 *                         poke(stack, null_)
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1229, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_9 = (((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->objects != ((PyObject*)Py_None));
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_9) {

            /* "flitter/language/vm.pyx":1230
 *                 elif instruction.code == OpCode.Pos:
 *                     if peek(stack).objects is not None:
 *                         poke(stack, null_)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Ceil:
 */
            __pyx_t_10 = ((PyObject *)__pyx_v_7flitter_5model_null_);
            __Pyx_INCREF(__pyx_t_10);
            __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1230, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "flitter/language/vm.pyx":1229
 * 
 *                 elif instruction.code == OpCode.Pos:
 *                     if peek(stack).objects is not None:             # <<<<<<<<<<<<<<
 *                         poke(stack, null_)
 * 
 */
          }

          /* "flitter/language/vm.pyx":1228
 *                     poke(stack, peek(stack).neg())
 * 
 *                 elif instruction.code == OpCode.Pos:             # <<<<<<<<<<<<<<
 *                     if peek(stack).objects is not None:
 *                         poke(stack, null_)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Ceil:

          /* "flitter/language/vm.pyx":1233
 * 
 *                 elif instruction.code == OpCode.Ceil:
 *                     poke(stack, peek(stack).ceil())             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Floor:
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1233, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->ceil(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10))); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1233, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1233, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1232
 *                         poke(stack, null_)
 * 
 *                 elif instruction.code == OpCode.Ceil:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).ceil())
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Floor:

          /* "flitter/language/vm.pyx":1236
 * 
 *                 elif instruction.code == OpCode.Floor:
 *                     poke(stack, peek(stack).floor())             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Fract:
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1236, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->floor(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15))); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1236, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1236, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1235
 *                     poke(stack, peek(stack).ceil())
 * 
 *                 elif instruction.code == OpCode.Floor:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).floor())
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Fract:

          /* "flitter/language/vm.pyx":1239
 * 
 *                 elif instruction.code == OpCode.Fract:
 *                     poke(stack, peek(stack).fract())             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Not:
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1239, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->fract(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), 0)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1239, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1239, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1238
 *                     poke(stack, peek(stack).floor())
 * 
 *                 elif instruction.code == OpCode.Fract:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).fract())
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Not:

          /* "flitter/language/vm.pyx":1242
 * 
 *                 elif instruction.code == OpCode.Not:
 *                     poke(stack, false_ if peek(stack).as_bool() else true_)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Add:
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1242, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_9 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_9) {
            __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_false_);
            __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_false_);
          } else {
            __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_true_);
            __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_true_);
          }
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1241
 *                     poke(stack, peek(stack).fract())
 * 
 *                 elif instruction.code == OpCode.Not:             # <<<<<<<<<<<<<<
 *                     poke(stack, false_ if peek(stack).as_bool() else true_)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Add:

          /* "flitter/language/vm.pyx":1245
 * 
 *                 elif instruction.code == OpCode.Add:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).add(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1245, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1246
 *                 elif instruction.code == OpCode.Add:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).add(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1246, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->add(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1246, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1246, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1247
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).add(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Sub:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1244
 *                     poke(stack, false_ if peek(stack).as_bool() else true_)
 * 
 *                 elif instruction.code == OpCode.Add:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).add(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Sub:

          /* "flitter/language/vm.pyx":1250
 * 
 *                 elif instruction.code == OpCode.Sub:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).sub(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1250, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1251
 *                 elif instruction.code == OpCode.Sub:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).sub(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1251, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->sub(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1251, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1251, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1252
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).sub(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Mul:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1249
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Sub:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).sub(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Mul:

          /* "flitter/language/vm.pyx":1255
 * 
 *                 elif instruction.code == OpCode.Mul:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).mul(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1255, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1256
 *                 elif instruction.code == OpCode.Mul:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1256, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->mul(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1256, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1257
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.MulAdd:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1254
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Mul:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_MulAdd:

          /* "flitter/language/vm.pyx":1260
 * 
 *                 elif instruction.code == OpCode.MulAdd:
 *                     r2 = pop(stack)             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul_add(r1, r2))
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1260, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1261
 *                 elif instruction.code == OpCode.MulAdd:
 *                     r2 = pop(stack)
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).mul_add(r1, r2))
 *                     r1 = r2 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1261, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1262
 *                     r2 = pop(stack)
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul_add(r1, r2))             # <<<<<<<<<<<<<<
 *                     r1 = r2 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1262, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->mul_add(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1, __pyx_v_r2, 0)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1262, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1263
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mul_add(r1, r2))
 *                     r1 = r2 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.TrueDiv:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1259
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.MulAdd:             # <<<<<<<<<<<<<<
 *                     r2 = pop(stack)
 *                     r1 = pop(stack)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_TrueDiv:

          /* "flitter/language/vm.pyx":1266
 * 
 *                 elif instruction.code == OpCode.TrueDiv:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).truediv(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1266, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1267
 *                 elif instruction.code == OpCode.TrueDiv:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).truediv(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1267, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->truediv(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1267, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1267, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1268
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).truediv(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.FloorDiv:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1265
 *                     r1 = r2 = None
 * 
 *                 elif instruction.code == OpCode.TrueDiv:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).truediv(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_FloorDiv:

          /* "flitter/language/vm.pyx":1271
 * 
 *                 elif instruction.code == OpCode.FloorDiv:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).floordiv(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1271, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1272
 *                 elif instruction.code == OpCode.FloorDiv:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).floordiv(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1272, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->floordiv(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1272, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1272, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1273
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).floordiv(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Mod:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1270
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.FloorDiv:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).floordiv(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Mod:

          /* "flitter/language/vm.pyx":1276
 * 
 *                 elif instruction.code == OpCode.Mod:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).mod(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1276, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1277
 *                 elif instruction.code == OpCode.Mod:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mod(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1277, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->mod(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1277, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1277, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1278
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mod(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Pow:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1275
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Mod:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).mod(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Pow:

          /* "flitter/language/vm.pyx":1281
 * 
 *                 elif instruction.code == OpCode.Pow:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).pow(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1281, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1282
 *                 elif instruction.code == OpCode.Pow:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).pow(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1282, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->pow(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1282, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1282, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1283
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).pow(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Eq:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1280
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Pow:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).pow(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Eq:

          /* "flitter/language/vm.pyx":1286
 * 
 *                 elif instruction.code == OpCode.Eq:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).eq(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1286, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1287
 *                 elif instruction.code == OpCode.Eq:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).eq(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1287, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->eq(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1287, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1287, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1288
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).eq(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Ne:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1285
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Eq:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).eq(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Ne:

          /* "flitter/language/vm.pyx":1291
 * 
 *                 elif instruction.code == OpCode.Ne:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).ne(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1291, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1292
 *                 elif instruction.code == OpCode.Ne:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ne(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1292, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->ne(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1292, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1292, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1293
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ne(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Gt:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1290
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Ne:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ne(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Gt:

          /* "flitter/language/vm.pyx":1296
 * 
 *                 elif instruction.code == OpCode.Gt:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).gt(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1296, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1297
 *                 elif instruction.code == OpCode.Gt:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).gt(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1297, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->gt(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1297, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1298
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).gt(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Lt:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1295
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Gt:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).gt(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Lt:

          /* "flitter/language/vm.pyx":1301
 * 
 *                 elif instruction.code == OpCode.Lt:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).lt(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1301, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1302
 *                 elif instruction.code == OpCode.Lt:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).lt(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1302, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->lt(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1302, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1302, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1303
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).lt(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Ge:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1300
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Lt:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).lt(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Ge:

          /* "flitter/language/vm.pyx":1306
 * 
 *                 elif instruction.code == OpCode.Ge:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).ge(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1306, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1307
 *                 elif instruction.code == OpCode.Ge:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ge(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1307, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->ge(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1307, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1308
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ge(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Le:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1305
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Ge:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).ge(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Le:

          /* "flitter/language/vm.pyx":1311
 * 
 *                 elif instruction.code == OpCode.Le:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).le(r1))
 *                     r1 = None
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1311, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1312
 *                 elif instruction.code == OpCode.Le:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).le(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1312, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->le(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), __pyx_v_r1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1312, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1312, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1313
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).le(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Xor:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1310
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Le:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).le(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Xor:

          /* "flitter/language/vm.pyx":1316
 * 
 *                 elif instruction.code == OpCode.Xor:
 *                     r2 = pop(stack)             # <<<<<<<<<<<<<<
 *                     r1 = peek(stack)
 *                     if not r1.as_bool():
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1316, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1317
 *                 elif instruction.code == OpCode.Xor:
 *                     r2 = pop(stack)
 *                     r1 = peek(stack)             # <<<<<<<<<<<<<<
 *                     if not r1.as_bool():
 *                         poke(stack, r2)
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1317, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1318
 *                     r2 = pop(stack)
 *                     r1 = peek(stack)
 *                     if not r1.as_bool():             # <<<<<<<<<<<<<<
 *                         poke(stack, r2)
 *                     elif r2.as_bool():
 */
          __pyx_t_9 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_r1->__pyx_vtab)->as_bool(__pyx_v_r1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1318, __pyx_L3_error)
          __pyx_t_6 = (!__pyx_t_9);
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1319
 *                     r1 = peek(stack)
 *                     if not r1.as_bool():
 *                         poke(stack, r2)             # <<<<<<<<<<<<<<
 *                     elif r2.as_bool():
 *                         poke(stack, false_)
 */
            __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, __pyx_v_r2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L3_error)

            /* "flitter/language/vm.pyx":1318
 *                     r2 = pop(stack)
 *                     r1 = peek(stack)
 *                     if not r1.as_bool():             # <<<<<<<<<<<<<<
 *                         poke(stack, r2)
 *                     elif r2.as_bool():
 */
            goto __pyx_L28;
          }

          /* "flitter/language/vm.pyx":1320
 *                     if not r1.as_bool():
 *                         poke(stack, r2)
 *                     elif r2.as_bool():             # <<<<<<<<<<<<<<
 *                         poke(stack, false_)
 *                     r1 = r2 = None
 */
          __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_r2->__pyx_vtab)->as_bool(__pyx_v_r2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1321
 *                         poke(stack, r2)
 *                     elif r2.as_bool():
 *                         poke(stack, false_)             # <<<<<<<<<<<<<<
 *                     r1 = r2 = None
 * 
 */
            __pyx_t_15 = ((PyObject *)__pyx_v_7flitter_5model_false_);
            __Pyx_INCREF(__pyx_t_15);
            __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

            /* "flitter/language/vm.pyx":1320
 *                     if not r1.as_bool():
 *                         poke(stack, r2)
 *                     elif r2.as_bool():             # <<<<<<<<<<<<<<
 *                         poke(stack, false_)
 *                     r1 = r2 = None
 */
          }
          __pyx_L28:;

          /* "flitter/language/vm.pyx":1322
 *                     elif r2.as_bool():
 *                         poke(stack, false_)
 *                     r1 = r2 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Slice:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r2, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1315
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.Xor:             # <<<<<<<<<<<<<<
 *                     r2 = pop(stack)
 *                     r1 = peek(stack)
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Slice:

          /* "flitter/language/vm.pyx":1325
 * 
 *                 elif instruction.code == OpCode.Slice:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).slice(r1))
 *                     r1 = None
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1325, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1326
 *                 elif instruction.code == OpCode.Slice:
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).slice(r1))             # <<<<<<<<<<<<<<
 *                     r1 = None
 * 
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1326, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)->__pyx_vtab)->slice(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), __pyx_v_r1, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1326, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1327
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).slice(r1))
 *                     r1 = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.SliceLiteral:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));

          /* "flitter/language/vm.pyx":1324
 *                     r1 = r2 = None
 * 
 *                 elif instruction.code == OpCode.Slice:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     poke(stack, peek(stack).slice(r1))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_SliceLiteral:

          /* "flitter/language/vm.pyx":1330
 * 
 *                 elif instruction.code == OpCode.SliceLiteral:
 *                     poke(stack, peek(stack).slice((<InstructionVector>instruction).value))             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.IndexLiteral:
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1330, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = ((PyObject *)((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instruction)->value);
          __Pyx_INCREF(__pyx_t_15);
          __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)->__pyx_vtab)->slice(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1330, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1330, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "flitter/language/vm.pyx":1329
 *                     r1 = None
 * 
 *                 elif instruction.code == OpCode.SliceLiteral:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).slice((<InstructionVector>instruction).value))
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_IndexLiteral:

          /* "flitter/language/vm.pyx":1333
 * 
 *                 elif instruction.code == OpCode.IndexLiteral:
 *                     poke(stack, peek(stack).item((<InstructionInt>instruction).value))             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Call:
 */
          __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1333, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_15 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->__pyx_vtab)->item(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value, 0)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1333, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1333, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1332
 *                     poke(stack, peek(stack).slice((<InstructionVector>instruction).value))
 * 
 *                 elif instruction.code == OpCode.IndexLiteral:             # <<<<<<<<<<<<<<
 *                     poke(stack, peek(stack).item((<InstructionInt>instruction).value))
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Call:

          /* "flitter/language/vm.pyx":1336
 * 
 *                 elif instruction.code == OpCode.Call:
 *                     r1 = pop(stack)             # <<<<<<<<<<<<<<
 *                     if (<InstructionIntTuple>instruction).tvalue is not None:
 *                         kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)
 */
          __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1336, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "flitter/language/vm.pyx":1337
 *                 elif instruction.code == OpCode.Call:
 *                     r1 = pop(stack)
 *                     if (<InstructionIntTuple>instruction).tvalue is not None:             # <<<<<<<<<<<<<<
 *                         kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)
 *                     else:
 */
          __pyx_t_6 = (((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_instruction)->tvalue != ((PyObject*)Py_None));
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1338
 *                     r1 = pop(stack)
 *                     if (<InstructionIntTuple>instruction).tvalue is not None:
 *                         kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)             # <<<<<<<<<<<<<<
 *                     else:
 *                         kwargs = None
 */
            __pyx_t_15 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_instruction)->tvalue;
            __Pyx_INCREF(__pyx_t_15);
            __pyx_t_1 = __pyx_f_7flitter_8language_2vm_pop_dict(__pyx_v_stack, ((PyObject*)__pyx_t_15)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF_SET(__pyx_v_kwargs, ((PyObject*)__pyx_t_1));
            __pyx_t_1 = 0;

            /* "flitter/language/vm.pyx":1337
 *                 elif instruction.code == OpCode.Call:
 *                     r1 = pop(stack)
 *                     if (<InstructionIntTuple>instruction).tvalue is not None:             # <<<<<<<<<<<<<<
 *                         kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)
 *                     else:
 */
            goto __pyx_L29;
          }

          /* "flitter/language/vm.pyx":1340
 *                         kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)
 *                     else:
 *                         kwargs = None             # <<<<<<<<<<<<<<
 *                     n = (<InstructionIntTuple>instruction).ivalue
 *                     args = pop_tuple(stack, n) if n else ()
 */
          /*else*/ {
            __Pyx_INCREF(Py_None);
            __Pyx_XDECREF_SET(__pyx_v_kwargs, ((PyObject*)Py_None));
          }
          __pyx_L29:;

          /* "flitter/language/vm.pyx":1341
 *                     else:
 *                         kwargs = None
 *                     n = (<InstructionIntTuple>instruction).ivalue             # <<<<<<<<<<<<<<
 *                     args = pop_tuple(stack, n) if n else ()
 *                     if r1.objects is not None:
 */
          __pyx_t_8 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v_instruction)->ivalue;
          __pyx_v_n = __pyx_t_8;

          /* "flitter/language/vm.pyx":1342
 *                         kwargs = None
 *                     n = (<InstructionIntTuple>instruction).ivalue
 *                     args = pop_tuple(stack, n) if n else ()             # <<<<<<<<<<<<<<
 *                     if r1.objects is not None:
 *                         for i in range(r1.length):
 */
          __pyx_t_6 = (__pyx_v_n != 0);
          if (__pyx_t_6) {
            __pyx_t_15 = __pyx_f_7flitter_8language_2vm_pop_tuple(__pyx_v_stack, __pyx_v_n); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1342, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_1 = __pyx_t_15;
            __pyx_t_15 = 0;
          } else {
            __Pyx_INCREF(__pyx_empty_tuple);
            __pyx_t_1 = __pyx_empty_tuple;
          }
          __Pyx_XDECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_1));
          __pyx_t_1 = 0;

          /* "flitter/language/vm.pyx":1343
 *                     n = (<InstructionIntTuple>instruction).ivalue
 *                     args = pop_tuple(stack, n) if n else ()
 *                     if r1.objects is not None:             # <<<<<<<<<<<<<<
 *                         for i in range(r1.length):
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 */
          __pyx_t_6 = (__pyx_v_r1->objects != ((PyObject*)Py_None));
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1344
 *                     args = pop_tuple(stack, n) if n else ()
 *                     if r1.objects is not None:
 *                         for i in range(r1.length):             # <<<<<<<<<<<<<<
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 *                         if r1.length > 1:
 */
            __pyx_t_8 = __pyx_v_r1->length;
            __pyx_t_11 = __pyx_t_8;
            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
              __pyx_v_i = __pyx_t_12;

              /* "flitter/language/vm.pyx":1345
 *                     if r1.objects is not None:
 *                         for i in range(r1.length):
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)             # <<<<<<<<<<<<<<
 *                         if r1.length > 1:
 *                             push(stack, pop_composed(stack, r1.length))
 */
              __pyx_t_1 = __pyx_v_r1->objects;
              __Pyx_INCREF(__pyx_t_1);
              __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_v_i);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_f_7flitter_8language_2vm_call_helper(__pyx_v_context, __pyx_v_stack, ((PyObject *)__pyx_t_7), __pyx_v_args, __pyx_v_kwargs, __pyx_v_record_stats, (&__pyx_v_duration)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1345, __pyx_L3_error)
            }

            /* "flitter/language/vm.pyx":1346
 *                         for i in range(r1.length):
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 *                         if r1.length > 1:             # <<<<<<<<<<<<<<
 *                             push(stack, pop_composed(stack, r1.length))
 *                     else:
 */
            __pyx_t_6 = (__pyx_v_r1->length > 1);
            if (__pyx_t_6) {

              /* "flitter/language/vm.pyx":1347
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 *                         if r1.length > 1:
 *                             push(stack, pop_composed(stack, r1.length))             # <<<<<<<<<<<<<<
 *                     else:
 *                         push(stack, null_)
 */
              __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop_composed(__pyx_v_stack, __pyx_v_r1->length)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1347, __pyx_L3_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "flitter/language/vm.pyx":1346
 *                         for i in range(r1.length):
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 *                         if r1.length > 1:             # <<<<<<<<<<<<<<
 *                             push(stack, pop_composed(stack, r1.length))
 *                     else:
 */
            }

            /* "flitter/language/vm.pyx":1343
 *                     n = (<InstructionIntTuple>instruction).ivalue
 *                     args = pop_tuple(stack, n) if n else ()
 *                     if r1.objects is not None:             # <<<<<<<<<<<<<<
 *                         for i in range(r1.length):
 *                             call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
 */
            goto __pyx_L30;
          }

          /* "flitter/language/vm.pyx":1349
 *                             push(stack, pop_composed(stack, r1.length))
 *                     else:
 *                         push(stack, null_)             # <<<<<<<<<<<<<<
 *                     r1 = kwargs = args = None
 * 
 */
          /*else*/ {
            __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_null_);
            __Pyx_INCREF(__pyx_t_1);
            __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1349, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_L30:;

          /* "flitter/language/vm.pyx":1350
 *                     else:
 *                         push(stack, null_)
 *                     r1 = kwargs = args = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.CallFast:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_r1, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_kwargs, ((PyObject*)Py_None));
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));

          /* "flitter/language/vm.pyx":1335
 *                     poke(stack, peek(stack).item((<InstructionInt>instruction).value))
 * 
 *                 elif instruction.code == OpCode.Call:             # <<<<<<<<<<<<<<
 *                     r1 = pop(stack)
 *                     if (<InstructionIntTuple>instruction).tvalue is not None:
 */
          break;
          case __pyx_e_7flitter_8language_2vm_CallFast:

          /* "flitter/language/vm.pyx":1353
 * 
 *                 elif instruction.code == OpCode.CallFast:
 *                     n = (<InstructionObjectInt>instruction).value             # <<<<<<<<<<<<<<
 *                     if n == 1:
 *                         if record_stats:
 */
          __pyx_t_8 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_instruction)->value;
          __pyx_v_n = __pyx_t_8;

          /* "flitter/language/vm.pyx":1354
 *                 elif instruction.code == OpCode.CallFast:
 *                     n = (<InstructionObjectInt>instruction).value
 *                     if n == 1:             # <<<<<<<<<<<<<<
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 */
          __pyx_t_6 = (__pyx_v_n == 1);
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1355
 *                     n = (<InstructionObjectInt>instruction).value
 *                     if n == 1:
 *                         if record_stats:             # <<<<<<<<<<<<<<
 *                             call_duration = -perf_counter()
 *                         try:
 */
            if (__pyx_v_record_stats) {

              /* "flitter/language/vm.pyx":1356
 *                     if n == 1:
 *                         if record_stats:
 *                             call_duration = -perf_counter()             # <<<<<<<<<<<<<<
 *                         try:
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 */
              __pyx_v_call_duration = (-perf_counter());

              /* "flitter/language/vm.pyx":1355
 *                     n = (<InstructionObjectInt>instruction).value
 *                     if n == 1:
 *                         if record_stats:             # <<<<<<<<<<<<<<
 *                             call_duration = -perf_counter()
 *                         try:
 */
            }

            /* "flitter/language/vm.pyx":1357
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              /*try:*/ {

                /* "flitter/language/vm.pyx":1358
 *                             call_duration = -perf_counter()
 *                         try:
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))             # <<<<<<<<<<<<<<
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 */
                __pyx_t_1 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_instruction)->obj;
                __Pyx_INCREF(__pyx_t_1);
                __pyx_t_15 = ((PyObject *)__pyx_f_7flitter_8language_2vm_peek(__pyx_v_stack)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1358, __pyx_L36_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_10 = PyObject_CallOneArg(__pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1358, __pyx_L36_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 1358, __pyx_L36_error)
                __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1358, __pyx_L36_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "flitter/language/vm.pyx":1357
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:
 */
              }
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              goto __pyx_L43_try_end;
              __pyx_L36_error:;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;

              /* "flitter/language/vm.pyx":1359
 *                         try:
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:             # <<<<<<<<<<<<<<
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             poke(stack, null_)
 */
              __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
              if (__pyx_t_16) {
                __Pyx_AddTraceback("flitter.language.vm.Program._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_15, &__pyx_t_1) < 0) __PYX_ERR(0, 1359, __pyx_L38_except_error)
                __Pyx_XGOTREF(__pyx_t_10);
                __Pyx_XGOTREF(__pyx_t_15);
                __Pyx_XGOTREF(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_15);
                __pyx_v_exc = __pyx_t_15;
                /*try:*/ {

                  /* "flitter/language/vm.pyx":1360
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")             # <<<<<<<<<<<<<<
 *                             poke(stack, null_)
 *                     else:
 */
                  __pyx_t_20 = __pyx_v_context->errors;
                  __Pyx_INCREF(__pyx_t_20);
                  __pyx_t_21 = PyTuple_New(4); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_GOTREF(__pyx_t_21);
                  __pyx_t_2 = 0;
                  __pyx_t_14 = 127;
                  __Pyx_INCREF(__pyx_kp_u_Error_calling);
                  __pyx_t_2 += 14;
                  __Pyx_GIVEREF(__pyx_kp_u_Error_calling);
                  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_kp_u_Error_calling);
                  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_instruction)->obj, __pyx_n_s_name_2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __pyx_t_23 = __Pyx_PyObject_FormatSimple(__pyx_t_22, __pyx_empty_unicode); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_GOTREF(__pyx_t_23);
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_23) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_23) : __pyx_t_14;
                  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_23);
                  __Pyx_GIVEREF(__pyx_t_23);
                  PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_23);
                  __pyx_t_23 = 0;
                  __Pyx_INCREF(__pyx_kp_u__7);
                  __pyx_t_2 += 2;
                  __Pyx_GIVEREF(__pyx_kp_u__7);
                  PyTuple_SET_ITEM(__pyx_t_21, 2, __pyx_kp_u__7);
                  __pyx_t_23 = __Pyx_PyObject_Unicode(__pyx_v_exc); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_GOTREF(__pyx_t_23);
                  __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_23) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_23) : __pyx_t_14;
                  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_23);
                  __Pyx_GIVEREF(__pyx_t_23);
                  PyTuple_SET_ITEM(__pyx_t_21, 3, __pyx_t_23);
                  __pyx_t_23 = 0;
                  __pyx_t_23 = __Pyx_PyUnicode_Join(__pyx_t_21, 4, __pyx_t_2, __pyx_t_14); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_GOTREF(__pyx_t_23);
                  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                  __pyx_t_16 = PySet_Add(__pyx_t_20, __pyx_t_23); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1360, __pyx_L49_error)
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

                  /* "flitter/language/vm.pyx":1361
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             poke(stack, null_)             # <<<<<<<<<<<<<<
 *                     else:
 *                         args = pop_tuple(stack, n)
 */
                  __pyx_t_23 = ((PyObject *)__pyx_v_7flitter_5model_null_);
                  __Pyx_INCREF(__pyx_t_23);
                  __pyx_f_7flitter_8language_2vm_poke(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_23)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L49_error)
                  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                }

                /* "flitter/language/vm.pyx":1359
 *                         try:
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:             # <<<<<<<<<<<<<<
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             poke(stack, null_)
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                    goto __pyx_L50;
                  }
                  __pyx_L49_error:;
                  /*exception exit:*/{
                    __Pyx_PyThreadState_declare
                    __Pyx_PyThreadState_assign
                    __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0; __pyx_t_29 = 0; __pyx_t_30 = 0; __pyx_t_31 = 0;
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                    if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_29, &__pyx_t_30, &__pyx_t_31);
                    if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_26, &__pyx_t_27, &__pyx_t_28) < 0)) __Pyx_ErrFetch(&__pyx_t_26, &__pyx_t_27, &__pyx_t_28);
                    __Pyx_XGOTREF(__pyx_t_26);
                    __Pyx_XGOTREF(__pyx_t_27);
                    __Pyx_XGOTREF(__pyx_t_28);
                    __Pyx_XGOTREF(__pyx_t_29);
                    __Pyx_XGOTREF(__pyx_t_30);
                    __Pyx_XGOTREF(__pyx_t_31);
                    __pyx_t_16 = __pyx_lineno; __pyx_t_24 = __pyx_clineno; __pyx_t_25 = __pyx_filename;
                    {
                      __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                    }
                    if (PY_MAJOR_VERSION >= 3) {
                      __Pyx_XGIVEREF(__pyx_t_29);
                      __Pyx_XGIVEREF(__pyx_t_30);
                      __Pyx_XGIVEREF(__pyx_t_31);
                      __Pyx_ExceptionReset(__pyx_t_29, __pyx_t_30, __pyx_t_31);
                    }
                    __Pyx_XGIVEREF(__pyx_t_26);
                    __Pyx_XGIVEREF(__pyx_t_27);
                    __Pyx_XGIVEREF(__pyx_t_28);
                    __Pyx_ErrRestore(__pyx_t_26, __pyx_t_27, __pyx_t_28);
                    __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0; __pyx_t_29 = 0; __pyx_t_30 = 0; __pyx_t_31 = 0;
                    __pyx_lineno = __pyx_t_16; __pyx_clineno = __pyx_t_24; __pyx_filename = __pyx_t_25;
                    goto __pyx_L38_except_error;
                  }
                  __pyx_L50:;
                }
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                goto __pyx_L37_exception_handled;
              }
              goto __pyx_L38_except_error;

              /* "flitter/language/vm.pyx":1357
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
 *                         except Exception as exc:
 */
              __pyx_L38_except_error:;
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              goto __pyx_L3_error;
              __pyx_L37_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_L43_try_end:;
            }

            /* "flitter/language/vm.pyx":1354
 *                 elif instruction.code == OpCode.CallFast:
 *                     n = (<InstructionObjectInt>instruction).value
 *                     if n == 1:             # <<<<<<<<<<<<<<
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 */
            goto __pyx_L34;
          }

          /* "flitter/language/vm.pyx":1363
 *                             poke(stack, null_)
 *                     else:
 *                         args = pop_tuple(stack, n)             # <<<<<<<<<<<<<<
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 */
          /*else*/ {
            __pyx_t_1 = __pyx_f_7flitter_8language_2vm_pop_tuple(__pyx_v_stack, __pyx_v_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1363, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_XDECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_1));
            __pyx_t_1 = 0;

            /* "flitter/language/vm.pyx":1364
 *                     else:
 *                         args = pop_tuple(stack, n)
 *                         if record_stats:             # <<<<<<<<<<<<<<
 *                             call_duration = -perf_counter()
 *                         try:
 */
            if (__pyx_v_record_stats) {

              /* "flitter/language/vm.pyx":1365
 *                         args = pop_tuple(stack, n)
 *                         if record_stats:
 *                             call_duration = -perf_counter()             # <<<<<<<<<<<<<<
 *                         try:
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 */
              __pyx_v_call_duration = (-perf_counter());

              /* "flitter/language/vm.pyx":1364
 *                     else:
 *                         args = pop_tuple(stack, n)
 *                         if record_stats:             # <<<<<<<<<<<<<<
 *                             call_duration = -perf_counter()
 *                         try:
 */
            }

            /* "flitter/language/vm.pyx":1366
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_17);
              /*try:*/ {

                /* "flitter/language/vm.pyx":1367
 *                             call_duration = -perf_counter()
 *                         try:
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))             # <<<<<<<<<<<<<<
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 */
                __pyx_t_1 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_instruction)->obj;
                __Pyx_INCREF(__pyx_t_1);
                __pyx_t_15 = PyObject_CallObject(__pyx_t_1, __pyx_v_args); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1367, __pyx_L56_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (!(likely(((__pyx_t_15) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 1367, __pyx_L56_error)
                __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_15)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1367, __pyx_L56_error)
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                /* "flitter/language/vm.pyx":1366
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:
 */
              }
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              goto __pyx_L63_try_end;
              __pyx_L56_error:;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
              __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;

              /* "flitter/language/vm.pyx":1368
 *                         try:
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:             # <<<<<<<<<<<<<<
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             push(stack, null_)
 */
              __pyx_t_24 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
              if (__pyx_t_24) {
                __Pyx_AddTraceback("flitter.language.vm.Program._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_15, &__pyx_t_1, &__pyx_t_10) < 0) __PYX_ERR(0, 1368, __pyx_L58_except_error)
                __Pyx_XGOTREF(__pyx_t_15);
                __Pyx_XGOTREF(__pyx_t_1);
                __Pyx_XGOTREF(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_1);
                __pyx_v_exc = __pyx_t_1;
                /*try:*/ {

                  /* "flitter/language/vm.pyx":1369
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")             # <<<<<<<<<<<<<<
 *                             push(stack, null_)
 *                         args = None
 */
                  __pyx_t_23 = __pyx_v_context->errors;
                  __Pyx_INCREF(__pyx_t_23);
                  __pyx_t_20 = PyTuple_New(4); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  __pyx_t_2 = 0;
                  __pyx_t_14 = 127;
                  __Pyx_INCREF(__pyx_kp_u_Error_calling);
                  __pyx_t_2 += 14;
                  __Pyx_GIVEREF(__pyx_kp_u_Error_calling);
                  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_kp_u_Error_calling);
                  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v_instruction)->obj, __pyx_n_s_name_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_21);
                  __pyx_t_22 = __Pyx_PyObject_FormatSimple(__pyx_t_21, __pyx_empty_unicode); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                  __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_22) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_22) : __pyx_t_14;
                  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_22);
                  __Pyx_GIVEREF(__pyx_t_22);
                  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_22);
                  __pyx_t_22 = 0;
                  __Pyx_INCREF(__pyx_kp_u__7);
                  __pyx_t_2 += 2;
                  __Pyx_GIVEREF(__pyx_kp_u__7);
                  PyTuple_SET_ITEM(__pyx_t_20, 2, __pyx_kp_u__7);
                  __pyx_t_22 = __Pyx_PyObject_Unicode(__pyx_v_exc); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_22) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_22) : __pyx_t_14;
                  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_22);
                  __Pyx_GIVEREF(__pyx_t_22);
                  PyTuple_SET_ITEM(__pyx_t_20, 3, __pyx_t_22);
                  __pyx_t_22 = 0;
                  __pyx_t_22 = __Pyx_PyUnicode_Join(__pyx_t_20, 4, __pyx_t_2, __pyx_t_14); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __pyx_t_24 = PySet_Add(__pyx_t_23, __pyx_t_22); if (unlikely(__pyx_t_24 == ((int)-1))) __PYX_ERR(0, 1369, __pyx_L69_error)
                  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                  /* "flitter/language/vm.pyx":1370
 *                         except Exception as exc:
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             push(stack, null_)             # <<<<<<<<<<<<<<
 *                         args = None
 *                     if record_stats:
 */
                  __pyx_t_22 = ((PyObject *)__pyx_v_7flitter_5model_null_);
                  __Pyx_INCREF(__pyx_t_22);
                  __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_22)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1370, __pyx_L69_error)
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                }

                /* "flitter/language/vm.pyx":1368
 *                         try:
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:             # <<<<<<<<<<<<<<
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             push(stack, null_)
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                    goto __pyx_L70;
                  }
                  __pyx_L69_error:;
                  /*exception exit:*/{
                    __Pyx_PyThreadState_declare
                    __Pyx_PyThreadState_assign
                    __pyx_t_31 = 0; __pyx_t_30 = 0; __pyx_t_29 = 0; __pyx_t_28 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0;
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                    if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_28, &__pyx_t_27, &__pyx_t_26);
                    if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_31, &__pyx_t_30, &__pyx_t_29) < 0)) __Pyx_ErrFetch(&__pyx_t_31, &__pyx_t_30, &__pyx_t_29);
                    __Pyx_XGOTREF(__pyx_t_31);
                    __Pyx_XGOTREF(__pyx_t_30);
                    __Pyx_XGOTREF(__pyx_t_29);
                    __Pyx_XGOTREF(__pyx_t_28);
                    __Pyx_XGOTREF(__pyx_t_27);
                    __Pyx_XGOTREF(__pyx_t_26);
                    __pyx_t_24 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_32 = __pyx_filename;
                    {
                      __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                    }
                    if (PY_MAJOR_VERSION >= 3) {
                      __Pyx_XGIVEREF(__pyx_t_28);
                      __Pyx_XGIVEREF(__pyx_t_27);
                      __Pyx_XGIVEREF(__pyx_t_26);
                      __Pyx_ExceptionReset(__pyx_t_28, __pyx_t_27, __pyx_t_26);
                    }
                    __Pyx_XGIVEREF(__pyx_t_31);
                    __Pyx_XGIVEREF(__pyx_t_30);
                    __Pyx_XGIVEREF(__pyx_t_29);
                    __Pyx_ErrRestore(__pyx_t_31, __pyx_t_30, __pyx_t_29);
                    __pyx_t_31 = 0; __pyx_t_30 = 0; __pyx_t_29 = 0; __pyx_t_28 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0;
                    __pyx_lineno = __pyx_t_24; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_32;
                    goto __pyx_L58_except_error;
                  }
                  __pyx_L70:;
                }
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                goto __pyx_L57_exception_handled;
              }
              goto __pyx_L58_except_error;

              /* "flitter/language/vm.pyx":1366
 *                         if record_stats:
 *                             call_duration = -perf_counter()
 *                         try:             # <<<<<<<<<<<<<<
 *                             push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
 *                         except Exception as exc:
 */
              __pyx_L58_except_error:;
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
              goto __pyx_L3_error;
              __pyx_L57_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_18, __pyx_t_17);
              __pyx_L63_try_end:;
            }

            /* "flitter/language/vm.pyx":1371
 *                             PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj.__name__}: {str(exc)}")
 *                             push(stack, null_)
 *                         args = None             # <<<<<<<<<<<<<<
 *                     if record_stats:
 *                         call_duration += perf_counter()
 */
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));
          }
          __pyx_L34:;

          /* "flitter/language/vm.pyx":1372
 *                             push(stack, null_)
 *                         args = None
 *                     if record_stats:             # <<<<<<<<<<<<<<
 *                         call_duration += perf_counter()
 *                         duration -= call_duration
 */
          if (__pyx_v_record_stats) {

            /* "flitter/language/vm.pyx":1373
 *                         args = None
 *                     if record_stats:
 *                         call_duration += perf_counter()             # <<<<<<<<<<<<<<
 *                         duration -= call_duration
 *                         CallOutDuration += call_duration
 */
            __pyx_v_call_duration = (__pyx_v_call_duration + perf_counter());

            /* "flitter/language/vm.pyx":1374
 *                     if record_stats:
 *                         call_duration += perf_counter()
 *                         duration -= call_duration             # <<<<<<<<<<<<<<
 *                         CallOutDuration += call_duration
 *                         CallOutCount += 1
 */
            __pyx_v_duration = (__pyx_v_duration - __pyx_v_call_duration);

            /* "flitter/language/vm.pyx":1375
 *                         call_duration += perf_counter()
 *                         duration -= call_duration
 *                         CallOutDuration += call_duration             # <<<<<<<<<<<<<<
 *                         CallOutCount += 1
 * 
 */
            __pyx_v_7flitter_8language_2vm_CallOutDuration = (__pyx_v_7flitter_8language_2vm_CallOutDuration + __pyx_v_call_duration);

            /* "flitter/language/vm.pyx":1376
 *                         duration -= call_duration
 *                         CallOutDuration += call_duration
 *                         CallOutCount += 1             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Func:
 */
            __pyx_v_7flitter_8language_2vm_CallOutCount = (__pyx_v_7flitter_8language_2vm_CallOutCount + 1);

            /* "flitter/language/vm.pyx":1372
 *                             push(stack, null_)
 *                         args = None
 *                     if record_stats:             # <<<<<<<<<<<<<<
 *                         call_duration += perf_counter()
 *                         duration -= call_duration
 */
          }

          /* "flitter/language/vm.pyx":1352
 *                     r1 = kwargs = args = None
 * 
 *                 elif instruction.code == OpCode.CallFast:             # <<<<<<<<<<<<<<
 *                     n = (<InstructionObjectInt>instruction).value
 *                     if n == 1:
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Func:

          /* "flitter/language/vm.pyx":1379
 * 
 *                 elif instruction.code == OpCode.Func:
 *                     function = Function.__new__(Function)             # <<<<<<<<<<<<<<
 *                     function.__name__ = (<InstructionFunc>instruction).name
 *                     function.parameters = (<InstructionFunc>instruction).parameters
 */
          __pyx_t_10 = ((PyObject *)__pyx_tp_new_7flitter_8language_2vm_Function(((PyTypeObject *)__pyx_ptype_7flitter_8language_2vm_Function), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1379, __pyx_L3_error)
          __Pyx_GOTREF((PyObject *)__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_function, ((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1380
 *                 elif instruction.code == OpCode.Func:
 *                     function = Function.__new__(Function)
 *                     function.__name__ = (<InstructionFunc>instruction).name             # <<<<<<<<<<<<<<
 *                     function.parameters = (<InstructionFunc>instruction).parameters
 *                     function.program = self
 */
          __pyx_t_10 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_instruction)->name;
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_v_function->__name__);
          __Pyx_DECREF(__pyx_v_function->__name__);
          __pyx_v_function->__name__ = ((PyObject*)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1381
 *                     function = Function.__new__(Function)
 *                     function.__name__ = (<InstructionFunc>instruction).name
 *                     function.parameters = (<InstructionFunc>instruction).parameters             # <<<<<<<<<<<<<<
 *                     function.program = self
 *                     function.address = (<InstructionFunc>instruction).offset + pc
 */
          __pyx_t_10 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_instruction)->parameters;
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_v_function->parameters);
          __Pyx_DECREF(__pyx_v_function->parameters);
          __pyx_v_function->parameters = ((PyObject*)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1382
 *                     function.__name__ = (<InstructionFunc>instruction).name
 *                     function.parameters = (<InstructionFunc>instruction).parameters
 *                     function.program = self             # <<<<<<<<<<<<<<
 *                     function.address = (<InstructionFunc>instruction).offset + pc
 *                     function.record_stats = record_stats
 */
          __Pyx_INCREF((PyObject *)__pyx_v_self);
          __Pyx_GIVEREF((PyObject *)__pyx_v_self);
          __Pyx_GOTREF((PyObject *)__pyx_v_function->program);
          __Pyx_DECREF((PyObject *)__pyx_v_function->program);
          __pyx_v_function->program = __pyx_v_self;

          /* "flitter/language/vm.pyx":1383
 *                     function.parameters = (<InstructionFunc>instruction).parameters
 *                     function.program = self
 *                     function.address = (<InstructionFunc>instruction).offset + pc             # <<<<<<<<<<<<<<
 *                     function.record_stats = record_stats
 *                     n = PyTuple_GET_SIZE(function.parameters)
 */
          __pyx_v_function->address = (((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_instruction)->__pyx_base.offset + __pyx_cur_scope->__pyx_v_pc);

          /* "flitter/language/vm.pyx":1384
 *                     function.program = self
 *                     function.address = (<InstructionFunc>instruction).offset + pc
 *                     function.record_stats = record_stats             # <<<<<<<<<<<<<<
 *                     n = PyTuple_GET_SIZE(function.parameters)
 *                     function.defaults = pop_tuple(stack, n)
 */
          __pyx_v_function->record_stats = __pyx_v_record_stats;

          /* "flitter/language/vm.pyx":1385
 *                     function.address = (<InstructionFunc>instruction).offset + pc
 *                     function.record_stats = record_stats
 *                     n = PyTuple_GET_SIZE(function.parameters)             # <<<<<<<<<<<<<<
 *                     function.defaults = pop_tuple(stack, n)
 *                     function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)
 */
          __pyx_t_10 = __pyx_v_function->parameters;
          __Pyx_INCREF(__pyx_t_10);
          __pyx_v_n = PyTuple_GET_SIZE(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1386
 *                     function.record_stats = record_stats
 *                     n = PyTuple_GET_SIZE(function.parameters)
 *                     function.defaults = pop_tuple(stack, n)             # <<<<<<<<<<<<<<
 *                     function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)
 *                     function.vself = Vector(function)
 */
          __pyx_t_10 = __pyx_f_7flitter_8language_2vm_pop_tuple(__pyx_v_stack, __pyx_v_n); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1386, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_v_function->defaults);
          __Pyx_DECREF(__pyx_v_function->defaults);
          __pyx_v_function->defaults = ((PyObject*)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1387
 *                     n = PyTuple_GET_SIZE(function.parameters)
 *                     function.defaults = pop_tuple(stack, n)
 *                     function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)             # <<<<<<<<<<<<<<
 *                     function.vself = Vector(function)
 *                     push(stack, function.vself)
 */
          __pyx_t_10 = __pyx_f_7flitter_8language_2vm_pop_tuple(__pyx_v_stack, ((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v_instruction)->ncaptures); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1387, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_v_function->captures);
          __Pyx_DECREF(__pyx_v_function->captures);
          __pyx_v_function->captures = ((PyObject*)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1388
 *                     function.defaults = pop_tuple(stack, n)
 *                     function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)
 *                     function.vself = Vector(function)             # <<<<<<<<<<<<<<
 *                     push(stack, function.vself)
 *                     function = None
 */
          __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_5model_Vector), ((PyObject *)__pyx_v_function)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1388, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF((PyObject *)__pyx_v_function->vself);
          __Pyx_DECREF((PyObject *)__pyx_v_function->vself);
          __pyx_v_function->vself = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1389
 *                     function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)
 *                     function.vself = Vector(function)
 *                     push(stack, function.vself)             # <<<<<<<<<<<<<<
 *                     function = None
 * 
 */
          __pyx_t_10 = ((PyObject *)__pyx_v_function->vself);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1389, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1390
 *                     function.vself = Vector(function)
 *                     push(stack, function.vself)
 *                     function = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Tag:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_function, ((struct __pyx_obj_7flitter_8language_2vm_Function *)Py_None));

          /* "flitter/language/vm.pyx":1378
 *                         CallOutCount += 1
 * 
 *                 elif instruction.code == OpCode.Func:             # <<<<<<<<<<<<<<
 *                     function = Function.__new__(Function)
 *                     function.__name__ = (<InstructionFunc>instruction).name
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Tag:

          /* "flitter/language/vm.pyx":1393
 * 
 *                 elif instruction.code == OpCode.Tag:
 *                     execute_tag(stack, (<InstructionStr>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Attributes:
 */
          __pyx_t_10 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_instruction)->value;
          __Pyx_INCREF(__pyx_t_10);
          __pyx_t_1 = __pyx_f_7flitter_8language_2vm_execute_tag(__pyx_v_stack, ((PyObject*)__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "flitter/language/vm.pyx":1392
 *                     function = None
 * 
 *                 elif instruction.code == OpCode.Tag:             # <<<<<<<<<<<<<<
 *                     execute_tag(stack, (<InstructionStr>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Attributes:

          /* "flitter/language/vm.pyx":1396
 * 
 *                 elif instruction.code == OpCode.Attributes:
 *                     execute_attributes(stack, (<InstructionTuple>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Append:
 */
          __pyx_t_1 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v_instruction)->value;
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_10 = __pyx_f_7flitter_8language_2vm_execute_attributes(__pyx_v_stack, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1396, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1395
 *                     execute_tag(stack, (<InstructionStr>instruction).value)
 * 
 *                 elif instruction.code == OpCode.Attributes:             # <<<<<<<<<<<<<<
 *                     execute_attributes(stack, (<InstructionTuple>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Append:

          /* "flitter/language/vm.pyx":1399
 * 
 *                 elif instruction.code == OpCode.Append:
 *                     execute_append(stack, (<InstructionInt>instruction).value)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Compose:
 */
          __pyx_f_7flitter_8language_2vm_execute_append(__pyx_v_stack, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1398
 *                     execute_attributes(stack, (<InstructionTuple>instruction).value)
 * 
 *                 elif instruction.code == OpCode.Append:             # <<<<<<<<<<<<<<
 *                     execute_append(stack, (<InstructionInt>instruction).value)
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Compose:

          /* "flitter/language/vm.pyx":1402
 * 
 *                 elif instruction.code == OpCode.Compose:
 *                     push(stack, pop_composed(stack, (<InstructionInt>instruction).value))             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.BeginFor:
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop_composed(__pyx_v_stack, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1402, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1402, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1401
 *                     execute_append(stack, (<InstructionInt>instruction).value)
 * 
 *                 elif instruction.code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *                     push(stack, pop_composed(stack, (<InstructionInt>instruction).value))
 * 
 */
          break;
          case __pyx_e_7flitter_8language_2vm_BeginFor:

          /* "flitter/language/vm.pyx":1405
 * 
 *                 elif instruction.code == OpCode.BeginFor:
 *                     n = (<InstructionInt>instruction).value             # <<<<<<<<<<<<<<
 *                     if loop_source is not None:
 *                         loop_sources.append(loop_source)
 */
          __pyx_t_8 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instruction)->value;
          __pyx_v_n = __pyx_t_8;

          /* "flitter/language/vm.pyx":1406
 *                 elif instruction.code == OpCode.BeginFor:
 *                     n = (<InstructionInt>instruction).value
 *                     if loop_source is not None:             # <<<<<<<<<<<<<<
 *                         loop_sources.append(loop_source)
 *                     loop_source = LoopSource.__new__(LoopSource)
 */
          __pyx_t_6 = (((PyObject *)__pyx_v_loop_source) != Py_None);
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1407
 *                     n = (<InstructionInt>instruction).value
 *                     if loop_source is not None:
 *                         loop_sources.append(loop_source)             # <<<<<<<<<<<<<<
 *                     loop_source = LoopSource.__new__(LoopSource)
 *                     loop_source.source = pop(stack)
 */
            __pyx_t_33 = __Pyx_PyList_Append(__pyx_v_loop_sources, ((PyObject *)__pyx_v_loop_source)); if (unlikely(__pyx_t_33 == ((int)-1))) __PYX_ERR(0, 1407, __pyx_L3_error)

            /* "flitter/language/vm.pyx":1406
 *                 elif instruction.code == OpCode.BeginFor:
 *                     n = (<InstructionInt>instruction).value
 *                     if loop_source is not None:             # <<<<<<<<<<<<<<
 *                         loop_sources.append(loop_source)
 *                     loop_source = LoopSource.__new__(LoopSource)
 */
          }

          /* "flitter/language/vm.pyx":1408
 *                     if loop_source is not None:
 *                         loop_sources.append(loop_source)
 *                     loop_source = LoopSource.__new__(LoopSource)             # <<<<<<<<<<<<<<
 *                     loop_source.source = pop(stack)
 *                     loop_source.stride = n
 */
          __pyx_t_10 = ((PyObject *)__pyx_tp_new_7flitter_8language_2vm_LoopSource(((PyTypeObject *)__pyx_ptype_7flitter_8language_2vm_LoopSource), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1408, __pyx_L3_error)
          __Pyx_GOTREF((PyObject *)__pyx_t_10);
          __Pyx_DECREF_SET(__pyx_v_loop_source, ((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1409
 *                         loop_sources.append(loop_source)
 *                     loop_source = LoopSource.__new__(LoopSource)
 *                     loop_source.source = pop(stack)             # <<<<<<<<<<<<<<
 *                     loop_source.stride = n
 *                     loop_source.position = 0
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1409, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_GOTREF((PyObject *)__pyx_v_loop_source->source);
          __Pyx_DECREF((PyObject *)__pyx_v_loop_source->source);
          __pyx_v_loop_source->source = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10);
          __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1410
 *                     loop_source = LoopSource.__new__(LoopSource)
 *                     loop_source.source = pop(stack)
 *                     loop_source.stride = n             # <<<<<<<<<<<<<<
 *                     loop_source.position = 0
 *                     loop_source.iterations = 0
 */
          __pyx_v_loop_source->stride = __pyx_v_n;

          /* "flitter/language/vm.pyx":1411
 *                     loop_source.source = pop(stack)
 *                     loop_source.stride = n
 *                     loop_source.position = 0             # <<<<<<<<<<<<<<
 *                     loop_source.iterations = 0
 *                     for i in range(n):
 */
          __pyx_v_loop_source->position = 0;

          /* "flitter/language/vm.pyx":1412
 *                     loop_source.stride = n
 *                     loop_source.position = 0
 *                     loop_source.iterations = 0             # <<<<<<<<<<<<<<
 *                     for i in range(n):
 *                         push(lnames, null_)
 */
          __pyx_v_loop_source->iterations = 0;

          /* "flitter/language/vm.pyx":1413
 *                     loop_source.position = 0
 *                     loop_source.iterations = 0
 *                     for i in range(n):             # <<<<<<<<<<<<<<
 *                         push(lnames, null_)
 * 
 */
          __pyx_t_8 = __pyx_v_n;
          __pyx_t_11 = __pyx_t_8;
          for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
            __pyx_v_i = __pyx_t_12;

            /* "flitter/language/vm.pyx":1414
 *                     loop_source.iterations = 0
 *                     for i in range(n):
 *                         push(lnames, null_)             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Next:
 */
            __pyx_t_10 = ((PyObject *)__pyx_v_7flitter_5model_null_);
            __Pyx_INCREF(__pyx_t_10);
            __pyx_t_13 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_lnames, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(__pyx_t_13 == ((int64_t)0))) __PYX_ERR(0, 1414, __pyx_L3_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }

          /* "flitter/language/vm.pyx":1404
 *                     push(stack, pop_composed(stack, (<InstructionInt>instruction).value))
 * 
 *                 elif instruction.code == OpCode.BeginFor:             # <<<<<<<<<<<<<<
 *                     n = (<InstructionInt>instruction).value
 *                     if loop_source is not None:
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Next:

          /* "flitter/language/vm.pyx":1417
 * 
 *                 elif instruction.code == OpCode.Next:
 *                     if loop_source.position >= loop_source.source.length:             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 *                     else:
 */
          __pyx_t_6 = (__pyx_v_loop_source->position >= __pyx_v_loop_source->source->length);
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1418
 *                 elif instruction.code == OpCode.Next:
 *                     if loop_source.position >= loop_source.source.length:
 *                         pc += (<InstructionJump>instruction).offset             # <<<<<<<<<<<<<<
 *                     else:
 *                         n = loop_source.stride
 */
            __pyx_cur_scope->__pyx_v_pc = (__pyx_cur_scope->__pyx_v_pc + ((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v_instruction)->offset);

            /* "flitter/language/vm.pyx":1417
 * 
 *                 elif instruction.code == OpCode.Next:
 *                     if loop_source.position >= loop_source.source.length:             # <<<<<<<<<<<<<<
 *                         pc += (<InstructionJump>instruction).offset
 *                     else:
 */
            goto __pyx_L79;
          }

          /* "flitter/language/vm.pyx":1420
 *                         pc += (<InstructionJump>instruction).offset
 *                     else:
 *                         n = loop_source.stride             # <<<<<<<<<<<<<<
 *                         for i in range(n-1, -1, -1):
 *                             if loop_source.position >= loop_source.source.length:
 */
          /*else*/ {
            __pyx_t_8 = __pyx_v_loop_source->stride;
            __pyx_v_n = __pyx_t_8;

            /* "flitter/language/vm.pyx":1421
 *                     else:
 *                         n = loop_source.stride
 *                         for i in range(n-1, -1, -1):             # <<<<<<<<<<<<<<
 *                             if loop_source.position >= loop_source.source.length:
 *                                 poke_at(lnames, i, null_)
 */
            for (__pyx_t_8 = (__pyx_v_n - 1); __pyx_t_8 > -1L; __pyx_t_8-=1) {
              __pyx_v_i = __pyx_t_8;

              /* "flitter/language/vm.pyx":1422
 *                         n = loop_source.stride
 *                         for i in range(n-1, -1, -1):
 *                             if loop_source.position >= loop_source.source.length:             # <<<<<<<<<<<<<<
 *                                 poke_at(lnames, i, null_)
 *                             else:
 */
              __pyx_t_6 = (__pyx_v_loop_source->position >= __pyx_v_loop_source->source->length);
              if (__pyx_t_6) {

                /* "flitter/language/vm.pyx":1423
 *                         for i in range(n-1, -1, -1):
 *                             if loop_source.position >= loop_source.source.length:
 *                                 poke_at(lnames, i, null_)             # <<<<<<<<<<<<<<
 *                             else:
 *                                 poke_at(lnames, i, loop_source.source.item(loop_source.position))
 */
                __pyx_t_10 = ((PyObject *)__pyx_v_7flitter_5model_null_);
                __Pyx_INCREF(__pyx_t_10);
                __pyx_f_7flitter_8language_2vm_poke_at(__pyx_v_lnames, __pyx_v_i, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1423, __pyx_L3_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "flitter/language/vm.pyx":1422
 *                         n = loop_source.stride
 *                         for i in range(n-1, -1, -1):
 *                             if loop_source.position >= loop_source.source.length:             # <<<<<<<<<<<<<<
 *                                 poke_at(lnames, i, null_)
 *                             else:
 */
                goto __pyx_L82;
              }

              /* "flitter/language/vm.pyx":1425
 *                                 poke_at(lnames, i, null_)
 *                             else:
 *                                 poke_at(lnames, i, loop_source.source.item(loop_source.position))             # <<<<<<<<<<<<<<
 *                             loop_source.position += 1
 *                         loop_source.iterations += 1
 */
              /*else*/ {
                __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_loop_source->source->__pyx_vtab)->item(__pyx_v_loop_source->source, __pyx_v_loop_source->position, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1425, __pyx_L3_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_f_7flitter_8language_2vm_poke_at(__pyx_v_lnames, __pyx_v_i, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __pyx_L82:;

              /* "flitter/language/vm.pyx":1426
 *                             else:
 *                                 poke_at(lnames, i, loop_source.source.item(loop_source.position))
 *                             loop_source.position += 1             # <<<<<<<<<<<<<<
 *                         loop_source.iterations += 1
 * 
 */
              __pyx_v_loop_source->position = (__pyx_v_loop_source->position + 1);
            }

            /* "flitter/language/vm.pyx":1427
 *                                 poke_at(lnames, i, loop_source.source.item(loop_source.position))
 *                             loop_source.position += 1
 *                         loop_source.iterations += 1             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.EndFor:
 */
            __pyx_v_loop_source->iterations = (__pyx_v_loop_source->iterations + 1);
          }
          __pyx_L79:;

          /* "flitter/language/vm.pyx":1416
 *                         push(lnames, null_)
 * 
 *                 elif instruction.code == OpCode.Next:             # <<<<<<<<<<<<<<
 *                     if loop_source.position >= loop_source.source.length:
 *                         pc += (<InstructionJump>instruction).offset
 */
          break;
          case __pyx_e_7flitter_8language_2vm_EndFor:

          /* "flitter/language/vm.pyx":1430
 * 
 *                 elif instruction.code == OpCode.EndFor:
 *                     drop(lnames, loop_source.stride)             # <<<<<<<<<<<<<<
 *                     push(stack, pop_composed(stack, loop_source.iterations))
 *                     if loop_sources:
 */
          __pyx_f_7flitter_8language_2vm_drop(__pyx_v_lnames, __pyx_v_loop_source->stride); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)

          /* "flitter/language/vm.pyx":1431
 *                 elif instruction.code == OpCode.EndFor:
 *                     drop(lnames, loop_source.stride)
 *                     push(stack, pop_composed(stack, loop_source.iterations))             # <<<<<<<<<<<<<<
 *                     if loop_sources:
 *                         loop_source = <LoopSource>loop_sources.pop()
 */
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop_composed(__pyx_v_stack, __pyx_v_loop_source->iterations)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1431, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = __pyx_f_7flitter_8language_2vm_push(__pyx_v_stack, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(__pyx_t_8 == ((int64_t)0))) __PYX_ERR(0, 1431, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1432
 *                     drop(lnames, loop_source.stride)
 *                     push(stack, pop_composed(stack, loop_source.iterations))
 *                     if loop_sources:             # <<<<<<<<<<<<<<
 *                         loop_source = <LoopSource>loop_sources.pop()
 *                     else:
 */
          __pyx_t_6 = (PyList_GET_SIZE(__pyx_v_loop_sources) != 0);
          if (__pyx_t_6) {

            /* "flitter/language/vm.pyx":1433
 *                     push(stack, pop_composed(stack, loop_source.iterations))
 *                     if loop_sources:
 *                         loop_source = <LoopSource>loop_sources.pop()             # <<<<<<<<<<<<<<
 *                     else:
 *                         loop_source = None
 */
            __pyx_t_10 = __Pyx_PyList_Pop(__pyx_v_loop_sources); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1433, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_1 = __pyx_t_10;
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF_SET(__pyx_v_loop_source, ((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)__pyx_t_1));
            __pyx_t_1 = 0;

            /* "flitter/language/vm.pyx":1432
 *                     drop(lnames, loop_source.stride)
 *                     push(stack, pop_composed(stack, loop_source.iterations))
 *                     if loop_sources:             # <<<<<<<<<<<<<<
 *                         loop_source = <LoopSource>loop_sources.pop()
 *                     else:
 */
            goto __pyx_L83;
          }

          /* "flitter/language/vm.pyx":1435
 *                         loop_source = <LoopSource>loop_sources.pop()
 *                     else:
 *                         loop_source = None             # <<<<<<<<<<<<<<
 * 
 *                 elif instruction.code == OpCode.Export:
 */
          /*else*/ {
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_loop_source, ((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)Py_None));
          }
          __pyx_L83:;

          /* "flitter/language/vm.pyx":1429
 *                         loop_source.iterations += 1
 * 
 *                 elif instruction.code == OpCode.EndFor:             # <<<<<<<<<<<<<<
 *                     drop(lnames, loop_source.stride)
 *                     push(stack, pop_composed(stack, loop_source.iterations))
 */
          break;
          case __pyx_e_7flitter_8language_2vm_Export:

          /* "flitter/language/vm.pyx":1438
 * 
 *                 elif instruction.code == OpCode.Export:
 *                     PyDict_SetItem(exports, (<InstructionStr>instruction).value, pop(stack))             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
          __pyx_t_1 = ((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v_instruction)->value;
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_10 = ((PyObject *)__pyx_f_7flitter_8language_2vm_pop(__pyx_v_stack)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1438, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_16 = PyDict_SetItem(__pyx_v_exports, __pyx_t_1, __pyx_t_10); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 1438, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/language/vm.pyx":1437
 *                         loop_source = None
 * 
 *                 elif instruction.code == OpCode.Export:             # <<<<<<<<<<<<<<
 *                     PyDict_SetItem(exports, (<InstructionStr>instruction).value, pop(stack))
 * 
 */
          break;
          default:

          /* "flitter/language/vm.pyx":1441
 * 
 *                 else:
 *                     raise AssertionError(f"Unrecognised instruction: {instruction}")             # <<<<<<<<<<<<<<
 * 
 *                 if record_stats:
 */
          __pyx_t_10 = __Pyx_PyObject_FormatSimple(((PyObject *)__pyx_v_instruction), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1441, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Unrecognised_instruction, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1441, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1441, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_Raise(__pyx_t_10, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __PYX_ERR(0, 1441, __pyx_L3_error)
          break;
        }

        /* "flitter/language/vm.pyx":1443
 *                     raise AssertionError(f"Unrecognised instruction: {instruction}")
 * 
 *                 if record_stats:             # <<<<<<<<<<<<<<
 *                     duration += perf_counter()
 *                     StatsCount[<int64_t>instruction.code] += 1
 */
        if (__pyx_v_record_stats) {

          /* "flitter/language/vm.pyx":1444
 * 
 *                 if record_stats:
 *                     duration += perf_counter()             # <<<<<<<<<<<<<<
 *                     StatsCount[<int64_t>instruction.code] += 1
 *                     StatsDuration[<int64_t>instruction.code] += duration
 */
          __pyx_v_duration = (__pyx_v_duration + perf_counter());

          /* "flitter/language/vm.pyx":1445
 *                 if record_stats:
 *                     duration += perf_counter()
 *                     StatsCount[<int64_t>instruction.code] += 1             # <<<<<<<<<<<<<<
 *                     StatsDuration[<int64_t>instruction.code] += duration
 * 
 */
          __pyx_t_8 = ((int64_t)__pyx_v_instruction->code);
          (__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_t_8]) = ((__pyx_v_7flitter_8language_2vm_StatsCount[__pyx_t_8]) + 1);

          /* "flitter/language/vm.pyx":1446
 *                     duration += perf_counter()
 *                     StatsCount[<int64_t>instruction.code] += 1
 *                     StatsDuration[<int64_t>instruction.code] += duration             # <<<<<<<<<<<<<<
 * 
 *         except StackOverflow:
 */
          __pyx_t_8 = ((int64_t)__pyx_v_instruction->code);
          (__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_t_8]) = ((__pyx_v_7flitter_8language_2vm_StatsDuration[__pyx_t_8]) + __pyx_v_duration);

          /* "flitter/language/vm.pyx":1443
 *                     raise AssertionError(f"Unrecognised instruction: {instruction}")
 * 
 *                 if record_stats:             # <<<<<<<<<<<<<<
 *                     duration += perf_counter()
 *                     StatsCount[<int64_t>instruction.code] += 1
 */
        }
      }

      /* "flitter/language/vm.pyx":1112
 *         assert self.linked, "Program has not been linked"
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while 0 <= pc < program_end:
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "flitter/language/vm.pyx":1448
 *                     StatsDuration[<int64_t>instruction.code] += duration
 * 
 *         except StackOverflow:             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_1, &__pyx_t_15);
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_StackOverflow); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1448, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_16 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_10, __pyx_t_22);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_ErrRestore(__pyx_t_10, __pyx_t_1, __pyx_t_15);
    __pyx_t_10 = 0; __pyx_t_1 = 0; __pyx_t_15 = 0;
    if (__pyx_t_16) {
      __Pyx_AddTraceback("flitter.language.vm.Program._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_15, &__pyx_t_1, &__pyx_t_10) < 0) __PYX_ERR(0, 1448, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_10);

      /* "flitter/language/vm.pyx":1449
 * 
 *         except StackOverflow:
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         except Exception:
 */
      __Pyx_GIVEREF(__pyx_t_15);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestoreWithState(__pyx_t_15, __pyx_t_1, __pyx_t_10);
      __pyx_t_15 = 0; __pyx_t_1 = 0; __pyx_t_10 = 0; 
      __PYX_ERR(0, 1449, __pyx_L5_except_error)
    }

    /* "flitter/language/vm.pyx":1451
 *             raise
 * 
 *         except Exception:             # <<<<<<<<<<<<<<
 *             if instruction is not None:
 *                 logger.exception("VM exception processing:\n{} <--",
 */
    __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_16) {
      __Pyx_AddTraceback("flitter.language.vm.Program._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_1, &__pyx_t_15) < 0) __PYX_ERR(0, 1451, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_15);

      /* "flitter/language/vm.pyx":1452
 * 
 *         except Exception:
 *             if instruction is not None:             # <<<<<<<<<<<<<<
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))
 */
      __pyx_t_6 = (((PyObject *)__pyx_v_instruction) != Py_None);
      if (__pyx_t_6) {

        /* "flitter/language/vm.pyx":1453
 *         except Exception:
 *             if instruction is not None:
 *                 logger.exception("VM exception processing:\n{} <--",             # <<<<<<<<<<<<<<
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))
 *             raise
 */
        __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_logger); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1453, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_23);
        __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_n_s_exception); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1453, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

        /* "flitter/language/vm.pyx":1454
 *             if instruction is not None:
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
        if (unlikely(__pyx_v_self->instructions == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1454, __pyx_L5_except_error)
        }
        __pyx_t_23 = __Pyx_PyList_GetSlice(__pyx_v_self->instructions, (__pyx_cur_scope->__pyx_v_pc - 5), __pyx_cur_scope->__pyx_v_pc); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1454, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_23);
        __pyx_t_21 = __pyx_pf_7flitter_8language_2vm_7Program_8_execute_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_23); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1454, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        __pyx_t_23 = __Pyx_Generator_Next(__pyx_t_21); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1454, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_23);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_21 = PyUnicode_Join(__pyx_kp_u__8, __pyx_t_23); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1454, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        __pyx_t_23 = NULL;
        __pyx_t_34 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_20))) {
          __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_20);
          if (likely(__pyx_t_23)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
            __Pyx_INCREF(__pyx_t_23);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_20, function);
            __pyx_t_34 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_23, __pyx_kp_u_VM_exception_processing, __pyx_t_21};
          __pyx_t_22 = __Pyx_PyObject_FastCall(__pyx_t_20, __pyx_callargs+1-__pyx_t_34, 2+__pyx_t_34);
          __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1453, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_22);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        }
        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

        /* "flitter/language/vm.pyx":1452
 * 
 *         except Exception:
 *             if instruction is not None:             # <<<<<<<<<<<<<<
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))
 */
      }

      /* "flitter/language/vm.pyx":1455
 *                 logger.exception("VM exception processing:\n{} <--",
 *                                  "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         assert pc == program_end, "Jump outside of program"
 */
      __Pyx_GIVEREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ErrRestoreWithState(__pyx_t_10, __pyx_t_1, __pyx_t_15);
      __pyx_t_10 = 0; __pyx_t_1 = 0; __pyx_t_15 = 0; 
      __PYX_ERR(0, 1455, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "flitter/language/vm.pyx":1112
 *         assert self.linked, "Program has not been linked"
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while 0 <= pc < program_end:
 *                 instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "flitter/language/vm.pyx":1457
 *             raise
 * 
 *         assert pc == program_end, "Jump outside of program"             # <<<<<<<<<<<<<<
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_6 = (__pyx_cur_scope->__pyx_v_pc == __pyx_v_program_end);
    if (unlikely(!__pyx_t_6)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Jump_outside_of_program, 0, 0);
      __PYX_ERR(0, 1457, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 1457, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":1092
 *         return self
 * 
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats):             # <<<<<<<<<<<<<<
 *         global CallOutCount, CallOutDuration
 *         cdef VectorStack stack=context.stack, lnames=context.lnames
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("flitter.language.vm.Program._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stack);
  __Pyx_XDECREF((PyObject *)__pyx_v_lnames);
  __Pyx_XDECREF(__pyx_v_exports);
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v_loop_sources);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop_source);
  __Pyx_XDECREF(__pyx_v_instructions);
  __Pyx_XDECREF((PyObject *)__pyx_v_instruction);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_r1);
  __Pyx_XDECREF((PyObject *)__pyx_v_r2);
  __Pyx_XDECREF((PyObject *)__pyx_v_r3);
  __Pyx_XDECREF(__pyx_v_import_names);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XDECREF((PyObject *)__pyx_v_function);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_gb_7flitter_8language_2vm_7Program_8_execute_2generator1);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/vm.pxd":140
 * 
 * cdef class Program:
 *     cdef readonly dict pragmas             # <<<<<<<<<<<<<<
 *     cdef readonly list instructions
 *     cdef bint linked
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_7pragmas_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_7pragmas_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_7pragmas___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_7pragmas___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->pragmas);
  __pyx_r = __pyx_v_self->pragmas;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":141
 * cdef class Program:
 *     cdef readonly dict pragmas
 *     cdef readonly list instructions             # <<<<<<<<<<<<<<
 *     cdef bint linked
 *     cdef readonly object path
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_12instructions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_12instructions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_12instructions___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_12instructions___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->instructions);
  __pyx_r = __pyx_v_self->instructions;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":143
 *     cdef readonly list instructions
 *     cdef bint linked
 *     cdef readonly object path             # <<<<<<<<<<<<<<
 *     cdef readonly object top
 *     cdef readonly tuple initial_lnames
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_4path_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_4path_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_4path___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_4path___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->path);
  __pyx_r = __pyx_v_self->path;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":144
 *     cdef bint linked
 *     cdef readonly object path
 *     cdef readonly object top             # <<<<<<<<<<<<<<
 *     cdef readonly tuple initial_lnames
 *     cdef readonly VectorStack stack
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_3top_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_3top_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_3top___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_3top___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->top);
  __pyx_r = __pyx_v_self->top;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":145
 *     cdef readonly object path
 *     cdef readonly object top
 *     cdef readonly tuple initial_lnames             # <<<<<<<<<<<<<<
 *     cdef readonly VectorStack stack
 *     cdef readonly VectorStack lnames
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_14initial_lnames_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_14initial_lnames_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_14initial_lnames___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_14initial_lnames___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->initial_lnames);
  __pyx_r = __pyx_v_self->initial_lnames;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":146
 *     cdef readonly object top
 *     cdef readonly tuple initial_lnames
 *     cdef readonly VectorStack stack             # <<<<<<<<<<<<<<
 *     cdef readonly VectorStack lnames
 *     cdef readonly set compiler_errors
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_5stack_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_5stack_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_5stack___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_5stack___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->stack);
  __pyx_r = ((PyObject *)__pyx_v_self->stack);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":147
 *     cdef readonly tuple initial_lnames
 *     cdef readonly VectorStack stack
 *     cdef readonly VectorStack lnames             # <<<<<<<<<<<<<<
 *     cdef readonly set compiler_errors
 *     cdef readonly bint simplify
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_6lnames_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_6lnames_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_6lnames___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_6lnames___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->lnames);
  __pyx_r = ((PyObject *)__pyx_v_self->lnames);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":148
 *     cdef readonly VectorStack stack
 *     cdef readonly VectorStack lnames
 *     cdef readonly set compiler_errors             # <<<<<<<<<<<<<<
 *     cdef readonly bint simplify
 *     cdef int64_t next_label
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_15compiler_errors_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_15compiler_errors_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_15compiler_errors___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_15compiler_errors___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->compiler_errors);
  __pyx_r = __pyx_v_self->compiler_errors;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/vm.pxd":149
 *     cdef readonly VectorStack lnames
 *     cdef readonly set compiler_errors
 *     cdef readonly bint simplify             # <<<<<<<<<<<<<<
 *     cdef int64_t next_label
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_8simplify_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_8simplify_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_8simplify___get__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_8simplify___get__(struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->simplify); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.vm.Program.simplify.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_121__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_121__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_121__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_121__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_120__reduce_cython__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_120__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_123__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7Program_123__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_123__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7Program_123__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.Program.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_7Program_122__setstate_cython__(((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_7Program_122__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.vm.Program.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Instruction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_3__pyx_unpickle_Instruction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_3__pyx_unpickle_Instruction = {"__pyx_unpickle_Instruction", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_3__pyx_unpickle_Instruction, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_3__pyx_unpickle_Instruction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Instruction (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Instruction", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Instruction", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Instruction") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Instruction", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_2__pyx_unpickle_Instruction(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_2__pyx_unpickle_Instruction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Instruction", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5694d08, 0xe6fb062, 0xc133679):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__11, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5694d08, 0xe6fb062, 0xc133679):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 *     __pyx_result = Instruction.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x5694d08, 0xe6fb062, 0xc133679):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Instruction.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5694d08, 0xe6fb062, 0xc133679):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 *     __pyx_result = Instruction.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 *     __pyx_result = Instruction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Instruction.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_Instruction__set_state(((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 *     __pyx_result = Instruction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Instruction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_Instruction__set_state(struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Instruction__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->code = __pyx_t_1;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Instruction__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionVector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_5__pyx_unpickle_InstructionVector(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_5__pyx_unpickle_InstructionVector = {"__pyx_unpickle_InstructionVector", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_5__pyx_unpickle_InstructionVector, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_5__pyx_unpickle_InstructionVector(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionVector (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionVector", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionVector", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionVector") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionVector", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_4__pyx_unpickle_InstructionVector(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_4__pyx_unpickle_InstructionVector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionVector", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionVector.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionVector.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionVector.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionVector.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionVector.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionVector__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionVector.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionVector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionVector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionVector__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionVector *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionVector__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->value);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionVector__set_state(<InstructionVector> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionVector__set_state(InstructionVector __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionVector__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_7__pyx_unpickle_InstructionNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_7__pyx_unpickle_InstructionNode = {"__pyx_unpickle_InstructionNode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7__pyx_unpickle_InstructionNode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_7__pyx_unpickle_InstructionNode(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionNode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionNode", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionNode", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionNode") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionNode", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_6__pyx_unpickle_InstructionNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_6__pyx_unpickle_InstructionNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionNode", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionNode.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionNode__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionNode__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionNode__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1), __pyx_ptype_7flitter_5model_Node))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->value);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionStr(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_9__pyx_unpickle_InstructionStr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_9__pyx_unpickle_InstructionStr = {"__pyx_unpickle_InstructionStr", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_9__pyx_unpickle_InstructionStr, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_9__pyx_unpickle_InstructionStr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionStr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionStr", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionStr", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionStr") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionStr", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_8__pyx_unpickle_InstructionStr(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_8__pyx_unpickle_InstructionStr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionStr", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionStr.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionStr.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionStr.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionStr.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionStr.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionStr__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionStr.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionStr(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionStr__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionStr *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionStr__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyUnicode_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->value);
  __Pyx_DECREF(__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionStr__set_state(<InstructionStr> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionStr__set_state(InstructionStr __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionStr__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_11__pyx_unpickle_InstructionTuple(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_11__pyx_unpickle_InstructionTuple = {"__pyx_unpickle_InstructionTuple", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11__pyx_unpickle_InstructionTuple, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_11__pyx_unpickle_InstructionTuple(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionTuple (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionTuple", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionTuple", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionTuple") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionTuple", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionTuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_10__pyx_unpickle_InstructionTuple(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_10__pyx_unpickle_InstructionTuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionTuple", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionTuple__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionTuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionTuple__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionTuple__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->value);
  __Pyx_DECREF(__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionTuple__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionInt(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_13__pyx_unpickle_InstructionInt(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_13__pyx_unpickle_InstructionInt = {"__pyx_unpickle_InstructionInt", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_13__pyx_unpickle_InstructionInt, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_13__pyx_unpickle_InstructionInt(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionInt (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionInt", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionInt", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionInt") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionInt", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionInt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_12__pyx_unpickle_InstructionInt(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_12__pyx_unpickle_InstructionInt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionInt", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionInt.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionInt.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaff8343, 0x4420fba, 0x4540960):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionInt.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionInt.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionInt.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionInt__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaff8343, 0x4420fba, 0x4540960) = (code, value))" % __pyx_checksum
 *     __pyx_result = InstructionInt.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionInt(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionInt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionInt__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionInt__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->value = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionInt__set_state(<InstructionInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionInt__set_state(InstructionInt __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionInt__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionIntTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_15__pyx_unpickle_InstructionIntTuple(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_15__pyx_unpickle_InstructionIntTuple = {"__pyx_unpickle_InstructionIntTuple", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15__pyx_unpickle_InstructionIntTuple, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_15__pyx_unpickle_InstructionIntTuple(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionIntTuple (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionIntTuple", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionIntTuple", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionIntTuple") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionIntTuple", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionIntTuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_14__pyx_unpickle_InstructionIntTuple(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_14__pyx_unpickle_InstructionIntTuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionIntTuple", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x46467d3, 0x66cc689, 0x681a088):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__14, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x46467d3, 0x66cc689, 0x681a088):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x46467d3, 0x66cc689, 0x681a088):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x46467d3, 0x66cc689, 0x681a088):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionIntTuple__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x46467d3, 0x66cc689, 0x681a088) = (code, ivalue, tvalue))" % __pyx_checksum
 *     __pyx_result = InstructionIntTuple.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionIntTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionIntTuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionIntTuple__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionIntTuple__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->ivalue = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->tvalue);
  __Pyx_DECREF(__pyx_v___pyx_result->tvalue);
  __pyx_v___pyx_result->tvalue = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 3);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionIntTuple__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionLabel(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_17__pyx_unpickle_InstructionLabel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_17__pyx_unpickle_InstructionLabel = {"__pyx_unpickle_InstructionLabel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_17__pyx_unpickle_InstructionLabel, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_17__pyx_unpickle_InstructionLabel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionLabel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionLabel", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionLabel", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionLabel") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionLabel", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionLabel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_16__pyx_unpickle_InstructionLabel(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_16__pyx_unpickle_InstructionLabel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionLabel", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9c59c74, 0x3b0f454, 0xc8dbed6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__15, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9c59c74, 0x3b0f454, 0xc8dbed6):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9c59c74, 0x3b0f454, 0xc8dbed6):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9c59c74, 0x3b0f454, 0xc8dbed6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionLabel__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9c59c74, 0x3b0f454, 0xc8dbed6) = (code, label))" % __pyx_checksum
 *     __pyx_result = InstructionLabel.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionLabel(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionLabel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionLabel__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionLabel__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->label = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 2);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionLabel__set_state(<InstructionLabel> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionLabel__set_state(InstructionLabel __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionLabel__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionJump(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_19__pyx_unpickle_InstructionJump(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_19__pyx_unpickle_InstructionJump = {"__pyx_unpickle_InstructionJump", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_19__pyx_unpickle_InstructionJump, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_19__pyx_unpickle_InstructionJump(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionJump (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionJump", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionJump", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionJump") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionJump", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionJump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_18__pyx_unpickle_InstructionJump(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_18__pyx_unpickle_InstructionJump(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionJump", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe15daaf, 0xa82c7af, 0x12820d3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__16, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe15daaf, 0xa82c7af, 0x12820d3):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 *     __pyx_result = InstructionJump.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe15daaf, 0xa82c7af, 0x12820d3):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionJump.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe15daaf, 0xa82c7af, 0x12820d3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 *     __pyx_result = InstructionJump.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 *     __pyx_result = InstructionJump.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionJump.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionJump__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionJump *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe15daaf, 0xa82c7af, 0x12820d3) = (code, label, offset))" % __pyx_checksum
 *     __pyx_result = InstructionJump.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionJump(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionJump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionJump__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionJump *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionJump__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->label = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->offset = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 3);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionJump__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionFunc(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_21__pyx_unpickle_InstructionFunc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_21__pyx_unpickle_InstructionFunc = {"__pyx_unpickle_InstructionFunc", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_21__pyx_unpickle_InstructionFunc, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_21__pyx_unpickle_InstructionFunc(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionFunc (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionFunc", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionFunc", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionFunc") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionFunc", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionFunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_20__pyx_unpickle_InstructionFunc(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_20__pyx_unpickle_InstructionFunc(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionFunc", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaaaf1, 0xf7f3751, 0x1bdd247):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaaaf1, 0xf7f3751, 0x1bdd247):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xbfaaaf1, 0xf7f3751, 0x1bdd247):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaaaf1, 0xf7f3751, 0x1bdd247):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionFunc__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaaaf1, 0xf7f3751, 0x1bdd247) = (code, label, name, ncaptures, offset, parameters))" % __pyx_checksum
 *     __pyx_result = InstructionFunc.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionFunc(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionFunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionFunc__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  int64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionFunc__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.label = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyUnicode_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->ncaptures = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.offset = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->parameters);
  __Pyx_DECREF(__pyx_v___pyx_result->parameters);
  __pyx_v___pyx_result->parameters = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 6);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionFunc__set_state(<InstructionFunc> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionFunc__set_state(InstructionFunc __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.name = __pyx_state[2]; __pyx_result.ncaptures = __pyx_state[3]; __pyx_result.offset = __pyx_state[4]; __pyx_result.parameters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionFunc__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InstructionObjectInt(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_23__pyx_unpickle_InstructionObjectInt(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_23__pyx_unpickle_InstructionObjectInt = {"__pyx_unpickle_InstructionObjectInt", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_23__pyx_unpickle_InstructionObjectInt, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_23__pyx_unpickle_InstructionObjectInt(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionObjectInt (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionObjectInt", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionObjectInt", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InstructionObjectInt") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InstructionObjectInt", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionObjectInt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_22__pyx_unpickle_InstructionObjectInt(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_22__pyx_unpickle_InstructionObjectInt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionObjectInt", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3911296, 0x84154a3, 0xb1daf84):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3911296, 0x84154a3, 0xb1daf84):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3911296, 0x84154a3, 0xb1daf84):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3911296, 0x84154a3, 0xb1daf84):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionObjectInt__set_state(((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3911296, 0x84154a3, 0xb1daf84) = (code, obj, value))" % __pyx_checksum
 *     __pyx_result = InstructionObjectInt.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionObjectInt(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionObjectInt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_InstructionObjectInt__set_state(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7flitter_8language_2vm_OpCode __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int64_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InstructionObjectInt__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = ((enum __pyx_t_7flitter_8language_2vm_OpCode)__Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base.code = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->obj);
  __Pyx_DECREF(__pyx_v___pyx_result->obj);
  __pyx_v___pyx_result->obj = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_3 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->value = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 3);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_InstructionObjectInt__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Function(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_25__pyx_unpickle_Function(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_25__pyx_unpickle_Function = {"__pyx_unpickle_Function", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_25__pyx_unpickle_Function, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_25__pyx_unpickle_Function(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Function") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_24__pyx_unpickle_Function(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_24__pyx_unpickle_Function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0a66206, 0xaff595e, 0x3cc2c89):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__19, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0a66206, 0xaff595e, 0x3cc2c89):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x0a66206, 0xaff595e, 0x3cc2c89):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0a66206, 0xaff595e, 0x3cc2c89):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Function), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_Function__set_state(((struct __pyx_obj_7flitter_8language_2vm_Function *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0a66206, 0xaff595e, 0x3cc2c89) = (__name__, address, call_depth, captures, defaults, hash, parameters, program, record_stats, vself))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Function(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_Function__set_state(struct __pyx_obj_7flitter_8language_2vm_Function *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyUnicode_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__name__);
  __Pyx_DECREF(__pyx_v___pyx_result->__name__);
  __pyx_v___pyx_result->__name__ = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->address = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->call_depth = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 3);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->captures);
  __Pyx_DECREF(__pyx_v___pyx_result->captures);
  __pyx_v___pyx_result->captures = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 4);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->defaults);
  __Pyx_DECREF(__pyx_v___pyx_result->defaults);
  __pyx_v___pyx_result->defaults = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->hash = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 6);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->parameters);
  __Pyx_DECREF(__pyx_v___pyx_result->parameters);
  __pyx_v___pyx_result->parameters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 7)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 7), __pyx_ptype_7flitter_8language_2vm_Program))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 7);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->program);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->program);
  __pyx_v___pyx_result->program = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(PyTuple_GET_ITEM(__pyx_v___pyx_state, 8)); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->record_stats = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 9)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 9), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 9);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->vself);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->vself);
  __pyx_v___pyx_result->vself = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 10);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 10)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.__name__ = __pyx_state[0]; __pyx_result.address = __pyx_state[1]; __pyx_result.call_depth = __pyx_state[2]; __pyx_result.captures = __pyx_state[3]; __pyx_result.defaults = __pyx_state[4]; __pyx_result.hash = __pyx_state[5]; __pyx_result.parameters = __pyx_state[6]; __pyx_result.program = __pyx_state[7]; __pyx_result.record_stats = __pyx_state[8]; __pyx_result.vself = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_Function__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LoopSource(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_2vm_27__pyx_unpickle_LoopSource(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_2vm_27__pyx_unpickle_LoopSource = {"__pyx_unpickle_LoopSource", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_27__pyx_unpickle_LoopSource, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_2vm_27__pyx_unpickle_LoopSource(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LoopSource (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LoopSource", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LoopSource", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LoopSource") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LoopSource", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_LoopSource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_2vm_26__pyx_unpickle_LoopSource(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_2vm_26__pyx_unpickle_LoopSource(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LoopSource", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2b65949, 0xdaa1095, 0xdcdcc58):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__20, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2b65949, 0xdaa1095, 0xdcdcc58):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 *     __pyx_result = LoopSource.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x2b65949, 0xdaa1095, 0xdcdcc58):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LoopSource.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2b65949, 0xdaa1095, 0xdcdcc58):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 *     __pyx_result = LoopSource.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_2vm_LoopSource), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 *     __pyx_result = LoopSource.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LoopSource.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_2vm___pyx_unpickle_LoopSource__set_state(((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2b65949, 0xdaa1095, 0xdcdcc58) = (iterations, position, source, stride))" % __pyx_checksum
 *     __pyx_result = LoopSource.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LoopSource(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_LoopSource", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_2vm___pyx_unpickle_LoopSource__set_state(struct __pyx_obj_7flitter_8language_2vm_LoopSource *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LoopSource__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->iterations = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->position = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->source);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->source);
  __pyx_v___pyx_result->source = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->stride = __pyx_t_1;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 4);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LoopSource__set_state(<LoopSource> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LoopSource__set_state(LoopSource __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.iterations = __pyx_state[0]; __pyx_result.position = __pyx_state[1]; __pyx_result.source = __pyx_state[2]; __pyx_result.stride = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.vm.__pyx_unpickle_LoopSource__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack __pyx_vtable_7flitter_8language_2vm_VectorStack;

static PyObject *__pyx_tp_new_7flitter_8language_2vm_VectorStack(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_8language_2vm_VectorStack;
  if (unlikely(__pyx_pw_7flitter_8language_2vm_11VectorStack_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_VectorStack(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_VectorStack) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7flitter_8language_2vm_11VectorStack_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_11VectorStack_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_11VectorStack_4size_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_11VectorStack_max_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_11VectorStack_8max_size_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_VectorStack[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_33__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11VectorStack_35__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_VectorStack[] = {
  {(char *)"size", __pyx_getprop_7flitter_8language_2vm_11VectorStack_size, 0, (char *)0, 0},
  {(char *)"max_size", __pyx_getprop_7flitter_8language_2vm_11VectorStack_max_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_VectorStack_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_VectorStack},
  {Py_sq_length, (void *)__pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__},
  {Py_mp_length, (void *)__pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_11VectorStack_7__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_VectorStack},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_VectorStack},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_VectorStack},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_VectorStack_spec = {
  "flitter.language.vm.VectorStack",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_VectorStack),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_VectorStack_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_VectorStack = {
  __pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_VectorStack = {
  __pyx_pw_7flitter_8language_2vm_11VectorStack_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_7flitter_8language_2vm_VectorStack = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""VectorStack", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_VectorStack), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_VectorStack, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_VectorStack, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_VectorStack, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_11VectorStack_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_VectorStack, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_VectorStack, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_VectorStack, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_Instruction(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_Instruction) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_11Instruction_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_11Instruction_4code_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_Instruction[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11Instruction_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_11Instruction_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_Instruction[] = {
  {(char *)"code", __pyx_getprop_7flitter_8language_2vm_11Instruction_code, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_Instruction_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_Instruction},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_11Instruction_3__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_Instruction},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_Instruction},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_11Instruction_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Instruction},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_Instruction_spec = {
  "flitter.language.vm.Instruction",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Instruction),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_Instruction_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_Instruction = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""Instruction", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Instruction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_11Instruction_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_Instruction, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_Instruction, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_11Instruction_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Instruction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionVector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionVector *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)o);
  p->value = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionVector(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionVector *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionVector) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->value);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionVector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionVector *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) ? ((__pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionVector)); if (e) return e;
  if (p->value) {
    e = (*v)(((PyObject *)p->value), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionVector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionVector *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) { if (__pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear) __pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionVector);
  tmp = ((PyObject*)p->value);
  p->value = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_17InstructionVector_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_17InstructionVector_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionVector[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionVector[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_17InstructionVector_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionVector_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionVector},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_17InstructionVector_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionVector},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionVector},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionVector},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionVector},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_17InstructionVector_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionVector},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionVector_spec = {
  "flitter.language.vm.InstructionVector",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionVector),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionVector_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionVector = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionVector", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionVector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionVector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_17InstructionVector_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionVector, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionVector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionVector, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionVector, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_17InstructionVector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionVector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionNode(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionNode *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)o);
  p->value = ((struct __pyx_obj_7flitter_5model_Node *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionNode(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionNode *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionNode) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->value);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionNode *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) ? ((__pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionNode)); if (e) return e;
  if (p->value) {
    e = (*v)(((PyObject *)p->value), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionNode *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionNode *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) { if (__pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear) __pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionNode);
  tmp = ((PyObject*)p->value);
  p->value = ((struct __pyx_obj_7flitter_5model_Node *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionNode_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionNode_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionNode[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionNode[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_15InstructionNode_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionNode_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionNode},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionNode_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionNode},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionNode},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionNode},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionNode},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionNode_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionNode},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionNode_spec = {
  "flitter.language.vm.InstructionNode",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionNode),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionNode_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionNode", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionNode, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_15InstructionNode_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionNode, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_15InstructionNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionStr(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionStr *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)o);
  p->value = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionStr(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionStr *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionStr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionStr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->value);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_14InstructionStr_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_14InstructionStr_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionStr[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionStr[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_14InstructionStr_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionStr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionStr},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_14InstructionStr_3__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionStr},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionStr},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_14InstructionStr_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionStr},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionStr_spec = {
  "flitter.language.vm.InstructionStr",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionStr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_InstructionStr_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionStr = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionStr", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionStr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionStr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_14InstructionStr_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionStr, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionStr, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_14InstructionStr_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionStr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionTuple(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)o);
  p->value = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionTuple(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionTuple) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->value);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionTuple(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) ? ((__pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionTuple)); if (e) return e;
  if (p->value) {
    e = (*v)(p->value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionTuple(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionTuple *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) { if (__pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear) __pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionTuple);
  tmp = ((PyObject*)p->value);
  p->value = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_16InstructionTuple_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_16InstructionTuple_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionTuple[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionTuple[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_16InstructionTuple_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionTuple_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionTuple},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionTuple},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionTuple},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionTuple},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionTuple},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_16InstructionTuple_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionTuple},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionTuple_spec = {
  "flitter.language.vm.InstructionTuple",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionTuple_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionTuple = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionTuple", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionTuple, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_16InstructionTuple_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionTuple, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionTuple, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionTuple, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionTuple, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_16InstructionTuple_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionTuple, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_getprop_7flitter_8language_2vm_14InstructionInt_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_14InstructionInt_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionInt[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionInt[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_14InstructionInt_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionInt_slots[] = {
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_14InstructionInt_3__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionInt},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionInt},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_14InstructionInt_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Instruction},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionInt_spec = {
  "flitter.language.vm.InstructionInt",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionInt),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_InstructionInt_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionInt = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionInt", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionInt), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_14InstructionInt_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionInt, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionInt, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_14InstructionInt_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Instruction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionIntTuple(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)o);
  p->tvalue = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionIntTuple(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionIntTuple) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->tvalue);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionIntTuple(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) ? ((__pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionIntTuple)); if (e) return e;
  if (p->tvalue) {
    e = (*v)(p->tvalue, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionIntTuple(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) { if (__pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear) __pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionIntTuple);
  tmp = ((PyObject*)p->tvalue);
  p->tvalue = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_19InstructionIntTuple_ivalue(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6ivalue_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_19InstructionIntTuple_tvalue(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_6tvalue_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionIntTuple[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionIntTuple[] = {
  {(char *)"ivalue", __pyx_getprop_7flitter_8language_2vm_19InstructionIntTuple_ivalue, 0, (char *)0, 0},
  {(char *)"tvalue", __pyx_getprop_7flitter_8language_2vm_19InstructionIntTuple_tvalue, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionIntTuple_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionIntTuple},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionIntTuple},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionIntTuple},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionIntTuple},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionIntTuple},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionIntTuple},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionIntTuple_spec = {
  "flitter.language.vm.InstructionIntTuple",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionIntTuple_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionIntTuple = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionIntTuple", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionIntTuple, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionIntTuple, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionIntTuple, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionIntTuple, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionIntTuple, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_19InstructionIntTuple_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionIntTuple, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_getprop_7flitter_8language_2vm_16InstructionLabel_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_16InstructionLabel_5label_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionLabel[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionLabel[] = {
  {(char *)"label", __pyx_getprop_7flitter_8language_2vm_16InstructionLabel_label, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionLabel_slots[] = {
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_3__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionLabel},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionLabel},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_16InstructionLabel_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Instruction},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionLabel_spec = {
  "flitter.language.vm.InstructionLabel",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_InstructionLabel_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionLabel = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionLabel", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_16InstructionLabel_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionLabel, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionLabel, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_16InstructionLabel_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Instruction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionJump_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionJump_5label_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionJump_offset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionJump_6offset_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionJump[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionJump[] = {
  {(char *)"label", __pyx_getprop_7flitter_8language_2vm_15InstructionJump_label, 0, (char *)0, 0},
  {(char *)"offset", __pyx_getprop_7flitter_8language_2vm_15InstructionJump_offset, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionJump_slots[] = {
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionJump_3__str__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionJump},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionJump},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionJump_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Instruction},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionJump_spec = {
  "flitter.language.vm.InstructionJump",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionJump),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_2vm_InstructionJump_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionJump = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionJump", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionJump), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_15InstructionJump_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionJump, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionJump, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_15InstructionJump_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Instruction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionFunc(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionFunc(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionFunc) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->parameters);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionFunc(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_InstructionJump)) ? ((__pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionFunc)); if (e) return e;
  if (p->parameters) {
    e = (*v)(p->parameters, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionFunc(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionFunc *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_InstructionJump)) { if (__pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_clear) __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionFunc);
  tmp = ((PyObject*)p->parameters);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionFunc_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionFunc_4name_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionFunc_parameters(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionFunc_10parameters_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_15InstructionFunc_ncaptures(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_15InstructionFunc_9ncaptures_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionFunc[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionFunc[] = {
  {(char *)"name", __pyx_getprop_7flitter_8language_2vm_15InstructionFunc_name, 0, (char *)0, 0},
  {(char *)"parameters", __pyx_getprop_7flitter_8language_2vm_15InstructionFunc_parameters, 0, (char *)0, 0},
  {(char *)"ncaptures", __pyx_getprop_7flitter_8language_2vm_15InstructionFunc_ncaptures, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionFunc_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionFunc},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionFunc},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionFunc},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionFunc},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionFunc},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_15InstructionFunc_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionFunc},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionFunc_spec = {
  "flitter.language.vm.InstructionFunc",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionFunc_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionFunc = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionFunc", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionFunc, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_15InstructionFunc_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionFunc, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionFunc, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionFunc, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionFunc, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_15InstructionFunc_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionFunc, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_InstructionObjectInt(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_2vm_Instruction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)o);
  p->obj = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_InstructionObjectInt(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_InstructionObjectInt) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->obj);
  __pyx_tp_dealloc_7flitter_8language_2vm_Instruction(o);
}

static int __pyx_tp_traverse_7flitter_8language_2vm_InstructionObjectInt(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) ? ((__pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse) ? __pyx_ptype_7flitter_8language_2vm_Instruction->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_2vm_InstructionObjectInt)); if (e) return e;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_InstructionObjectInt(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *p = (struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt *)o;
  if (likely(__pyx_ptype_7flitter_8language_2vm_Instruction)) { if (__pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear) __pyx_ptype_7flitter_8language_2vm_Instruction->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_2vm_InstructionObjectInt);
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_20InstructionObjectInt_obj(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3obj_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_20InstructionObjectInt_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_InstructionObjectInt[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_InstructionObjectInt[] = {
  {(char *)"obj", __pyx_getprop_7flitter_8language_2vm_20InstructionObjectInt_obj, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_7flitter_8language_2vm_20InstructionObjectInt_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_InstructionObjectInt_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_InstructionObjectInt},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_InstructionObjectInt},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_InstructionObjectInt},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_InstructionObjectInt},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_InstructionObjectInt},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_InstructionObjectInt},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_InstructionObjectInt_spec = {
  "flitter.language.vm.InstructionObjectInt",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_InstructionObjectInt_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_InstructionObjectInt = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""InstructionObjectInt", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_InstructionObjectInt, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_InstructionObjectInt, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_InstructionObjectInt, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_InstructionObjectInt, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_InstructionObjectInt, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_2vm_20InstructionObjectInt_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_InstructionObjectInt, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_2vm_Program __pyx_vtable_7flitter_8language_2vm_Program;

static PyObject *__pyx_tp_new_7flitter_8language_2vm_Program(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_Program *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_2vm_Program *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_8language_2vm_Program;
  p->pragmas = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->instructions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->path = Py_None; Py_INCREF(Py_None);
  p->top = Py_None; Py_INCREF(Py_None);
  p->initial_lnames = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)Py_None); Py_INCREF(Py_None);
  p->lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)Py_None); Py_INCREF(Py_None);
  p->compiler_errors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_8language_2vm_7Program_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_Program(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_Program *p = (struct __pyx_obj_7flitter_8language_2vm_Program *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_Program) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->pragmas);
  Py_CLEAR(p->instructions);
  Py_CLEAR(p->path);
  Py_CLEAR(p->top);
  Py_CLEAR(p->initial_lnames);
  Py_CLEAR(p->stack);
  Py_CLEAR(p->lnames);
  Py_CLEAR(p->compiler_errors);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_2vm_Program(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_Program *p = (struct __pyx_obj_7flitter_8language_2vm_Program *)o;
  if (p->pragmas) {
    e = (*v)(p->pragmas, a); if (e) return e;
  }
  if (p->instructions) {
    e = (*v)(p->instructions, a); if (e) return e;
  }
  if (p->path) {
    e = (*v)(p->path, a); if (e) return e;
  }
  if (p->top) {
    e = (*v)(p->top, a); if (e) return e;
  }
  if (p->initial_lnames) {
    e = (*v)(p->initial_lnames, a); if (e) return e;
  }
  if (p->stack) {
    e = (*v)(((PyObject *)p->stack), a); if (e) return e;
  }
  if (p->lnames) {
    e = (*v)(((PyObject *)p->lnames), a); if (e) return e;
  }
  if (p->compiler_errors) {
    e = (*v)(p->compiler_errors, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_Program(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_Program *p = (struct __pyx_obj_7flitter_8language_2vm_Program *)o;
  tmp = ((PyObject*)p->pragmas);
  p->pragmas = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->instructions);
  p->instructions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->path);
  p->path = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->top);
  p->top = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->initial_lnames);
  p->initial_lnames = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->stack);
  p->stack = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->lnames);
  p->lnames = ((struct __pyx_obj_7flitter_8language_2vm_VectorStack *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->compiler_errors);
  p->compiler_errors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_pragmas(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_7pragmas_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_instructions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_12instructions_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_path(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_4path_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_top(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_3top_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_initial_lnames(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_14initial_lnames_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_stack(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_5stack_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_lnames(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_6lnames_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_compiler_errors(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_15compiler_errors_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_7Program_simplify(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_7Program_8simplify_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_Program[] = {
  {"set_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_7set_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_top", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_9set_top, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_pragma", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_11set_pragma, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_13execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7flitter_8language_2vm_7Program_12execute},
  {"run", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_15run, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_121__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_7Program_123__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_Program[] = {
  {(char *)"pragmas", __pyx_getprop_7flitter_8language_2vm_7Program_pragmas, 0, (char *)0, 0},
  {(char *)"instructions", __pyx_getprop_7flitter_8language_2vm_7Program_instructions, 0, (char *)0, 0},
  {(char *)"path", __pyx_getprop_7flitter_8language_2vm_7Program_path, 0, (char *)0, 0},
  {(char *)"top", __pyx_getprop_7flitter_8language_2vm_7Program_top, 0, (char *)0, 0},
  {(char *)"initial_lnames", __pyx_getprop_7flitter_8language_2vm_7Program_initial_lnames, 0, (char *)0, 0},
  {(char *)"stack", __pyx_getprop_7flitter_8language_2vm_7Program_stack, 0, (char *)0, 0},
  {(char *)"lnames", __pyx_getprop_7flitter_8language_2vm_7Program_lnames, 0, (char *)0, 0},
  {(char *)"compiler_errors", __pyx_getprop_7flitter_8language_2vm_7Program_compiler_errors, 0, (char *)0, 0},
  {(char *)"simplify", __pyx_getprop_7flitter_8language_2vm_7Program_simplify, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_Program_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_Program},
  {Py_sq_length, (void *)__pyx_pw_7flitter_8language_2vm_7Program_3__len__},
  {Py_mp_length, (void *)__pyx_pw_7flitter_8language_2vm_7Program_3__len__},
  {Py_tp_str, (void *)__pyx_pw_7flitter_8language_2vm_7Program_5__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_Program},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_Program},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_Program},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_Program},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Program},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_Program_spec = {
  "flitter.language.vm.Program",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Program),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_Program_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Program = {
  __pyx_pw_7flitter_8language_2vm_7Program_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Program = {
  __pyx_pw_7flitter_8language_2vm_7Program_3__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_7flitter_8language_2vm_Program = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""Program", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Program), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Program, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Program, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Program, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7flitter_8language_2vm_7Program_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_Program, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_Program, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_Program, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_Program, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Program, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_Function(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_Function *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_2vm_Function *)o);
  p->__name__ = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->vself = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->defaults = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->program = ((struct __pyx_obj_7flitter_8language_2vm_Program *)Py_None); Py_INCREF(Py_None);
  p->captures = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_Function(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_Function *p = (struct __pyx_obj_7flitter_8language_2vm_Function *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_Function) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__name__);
  Py_CLEAR(p->vself);
  Py_CLEAR(p->parameters);
  Py_CLEAR(p->defaults);
  Py_CLEAR(p->program);
  Py_CLEAR(p->captures);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_2vm_Function(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_Function *p = (struct __pyx_obj_7flitter_8language_2vm_Function *)o;
  if (p->vself) {
    e = (*v)(((PyObject *)p->vself), a); if (e) return e;
  }
  if (p->parameters) {
    e = (*v)(p->parameters, a); if (e) return e;
  }
  if (p->defaults) {
    e = (*v)(p->defaults, a); if (e) return e;
  }
  if (p->program) {
    e = (*v)(((PyObject *)p->program), a); if (e) return e;
  }
  if (p->captures) {
    e = (*v)(p->captures, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_Function(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_Function *p = (struct __pyx_obj_7flitter_8language_2vm_Function *)o;
  tmp = ((PyObject*)p->vself);
  p->vself = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parameters);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->defaults);
  p->defaults = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->program);
  p->program = ((struct __pyx_obj_7flitter_8language_2vm_Program *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->captures);
  p->captures = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function___name__(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_8__name___1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_vself(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_5vself_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_parameters(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_10parameters_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_defaults(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_8defaults_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_program(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_7program_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_address(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_7address_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_record_stats(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_12record_stats_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_captures(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_8captures_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_call_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_10call_depth_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_2vm_8Function_hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_2vm_8Function_4hash_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_Function[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_8Function_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_8Function_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_2vm_Function[] = {
  {(char *)"__name__", __pyx_getprop_7flitter_8language_2vm_8Function___name__, 0, (char *)0, 0},
  {(char *)"vself", __pyx_getprop_7flitter_8language_2vm_8Function_vself, 0, (char *)0, 0},
  {(char *)"parameters", __pyx_getprop_7flitter_8language_2vm_8Function_parameters, 0, (char *)0, 0},
  {(char *)"defaults", __pyx_getprop_7flitter_8language_2vm_8Function_defaults, 0, (char *)0, 0},
  {(char *)"program", __pyx_getprop_7flitter_8language_2vm_8Function_program, 0, (char *)0, 0},
  {(char *)"address", __pyx_getprop_7flitter_8language_2vm_8Function_address, 0, (char *)0, 0},
  {(char *)"record_stats", __pyx_getprop_7flitter_8language_2vm_8Function_record_stats, 0, (char *)0, 0},
  {(char *)"captures", __pyx_getprop_7flitter_8language_2vm_8Function_captures, 0, (char *)0, 0},
  {(char *)"call_depth", __pyx_getprop_7flitter_8language_2vm_8Function_call_depth, 0, (char *)0, 0},
  {(char *)"hash", __pyx_getprop_7flitter_8language_2vm_8Function_hash, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_Function_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_Function},
  {Py_tp_hash, (void *)__pyx_pw_7flitter_8language_2vm_8Function_1__hash__},
  {Py_tp_call, (void *)__pyx_pw_7flitter_8language_2vm_8Function_3__call__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_Function},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_Function},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_Function},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_2vm_Function},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_Function},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_Function_spec = {
  "flitter.language.vm.Function",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Function),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_Function_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_Function = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""Function", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_Function), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_Function, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_7flitter_8language_2vm_8Function_1__hash__, /*tp_hash*/
  __pyx_pw_7flitter_8language_2vm_8Function_3__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_Function, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_Function, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_Function, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_2vm_Function, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_Function, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm_LoopSource(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_2vm_LoopSource *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_2vm_LoopSource *)o);
  p->source = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm_LoopSource(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm_LoopSource *p = (struct __pyx_obj_7flitter_8language_2vm_LoopSource *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm_LoopSource) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->source);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_2vm_LoopSource(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm_LoopSource *p = (struct __pyx_obj_7flitter_8language_2vm_LoopSource *)o;
  if (p->source) {
    e = (*v)(((PyObject *)p->source), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_2vm_LoopSource(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_2vm_LoopSource *p = (struct __pyx_obj_7flitter_8language_2vm_LoopSource *)o;
  tmp = ((PyObject*)p->source);
  p->source = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_8language_2vm_LoopSource[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_10LoopSource_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_2vm_10LoopSource_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm_LoopSource_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm_LoopSource},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm_LoopSource},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_2vm_LoopSource},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_2vm_LoopSource},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm_LoopSource},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm_LoopSource_spec = {
  "flitter.language.vm.LoopSource",
  sizeof(struct __pyx_obj_7flitter_8language_2vm_LoopSource),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_2vm_LoopSource_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm_LoopSource = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""LoopSource", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm_LoopSource), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm_LoopSource, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm_LoopSource, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_2vm_LoopSource, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_2vm_LoopSource, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm_LoopSource, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_7flitter_8language_2vm___pyx_scope_struct__genexpr = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct__genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct__genexpr[--__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct__genexpr(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *p = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct__genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_instruction);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct__genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_7flitter_8language_2vm___pyx_scope_struct__genexpr[__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct__genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *p = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_instruction) {
    e = (*v)(p->__pyx_v_instruction, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct__genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct__genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct__genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr_spec = {
  "flitter.language.vm.__pyx_scope_struct__genexpr",
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct__genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_1__execute[8];
static int __pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_1__execute = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_1__execute(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_1__execute > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute)))) {
    o = (PyObject*)__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_1__execute[--__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_1__execute];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_1__execute(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_1__execute) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_1__execute < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute)))) {
    __pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_1__execute[__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_1__execute++] = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_1__execute},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_1__execute},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute_spec = {
  "flitter.language.vm.__pyx_scope_struct_1__execute",
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""__pyx_scope_struct_1__execute", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_1__execute), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_1__execute, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_1__execute, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_2_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_2_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_2_genexpr[--__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_2_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_instruction);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_2_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_7flitter_8language_2vm___pyx_scope_struct_2_genexpr[__pyx_freecount_7flitter_8language_2vm___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_instruction) {
    e = (*v)(p->__pyx_v_instruction, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_2_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct_2_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_2_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr_spec = {
  "flitter.language.vm.__pyx_scope_struct_2_genexpr",
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.vm.""__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_2vm___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_2vm___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_2vm___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_2vm___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_kp_u_15s_9d_x_8_0f_ns_7_3f_s_4_1f, __pyx_k_15s_9d_x_8_0f_ns_7_3f_s_4_1f, sizeof(__pyx_k_15s_9d_x_8_0f_ns_7_3f_s_4_1f), 0, 1, 0, 0},
    {&__pyx_kp_u_15s_9d_x_8_3f_s_7_3f_s_4_1f, __pyx_k_15s_9d_x_8_3f_s_7_3f_s_4_1f, sizeof(__pyx_k_15s_9d_x_8_3f_s_7_3f_s_4_1f), 0, 1, 0, 0},
    {&__pyx_n_u_Add, __pyx_k_Add, sizeof(__pyx_k_Add), 0, 1, 0, 1},
    {&__pyx_n_u_Append, __pyx_k_Append, sizeof(__pyx_k_Append), 0, 1, 0, 1},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_u_Attributes, __pyx_k_Attributes, sizeof(__pyx_k_Attributes), 0, 1, 0, 1},
    {&__pyx_kp_u_Bad_function_return_lnames, __pyx_k_Bad_function_return_lnames, sizeof(__pyx_k_Bad_function_return_lnames), 0, 1, 0, 0},
    {&__pyx_kp_u_Bad_function_return_stack, __pyx_k_Bad_function_return_stack, sizeof(__pyx_k_Bad_function_return_stack), 0, 1, 0, 0},
    {&__pyx_kp_u_Bad_lnames, __pyx_k_Bad_lnames, sizeof(__pyx_k_Bad_lnames), 0, 1, 0, 0},
    {&__pyx_kp_u_Bad_root_node, __pyx_k_Bad_root_node, sizeof(__pyx_k_Bad_root_node), 0, 1, 0, 0},
    {&__pyx_kp_u_Bad_stack, __pyx_k_Bad_stack, sizeof(__pyx_k_Bad_stack), 0, 1, 0, 0},
    {&__pyx_n_u_BeginFor, __pyx_k_BeginFor, sizeof(__pyx_k_BeginFor), 0, 1, 0, 1},
    {&__pyx_n_u_BranchFalse, __pyx_k_BranchFalse, sizeof(__pyx_k_BranchFalse), 0, 1, 0, 1},
    {&__pyx_n_u_BranchTrue, __pyx_k_BranchTrue, sizeof(__pyx_k_BranchTrue), 0, 1, 0, 1},
    {&__pyx_n_u_Call, __pyx_k_Call, sizeof(__pyx_k_Call), 0, 1, 0, 1},
    {&__pyx_n_u_CallFast, __pyx_k_CallFast, sizeof(__pyx_k_CallFast), 0, 1, 0, 1},
    {&__pyx_kp_u_Cannot_drop, __pyx_k_Cannot_drop, sizeof(__pyx_k_Cannot_drop), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_peek, __pyx_k_Cannot_peek, sizeof(__pyx_k_Cannot_peek), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_poke, __pyx_k_Cannot_poke, sizeof(__pyx_k_Cannot_poke), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_pop, __pyx_k_Cannot_pop, sizeof(__pyx_k_Cannot_pop), 0, 1, 0, 0},
    {&__pyx_n_u_Ceil, __pyx_k_Ceil, sizeof(__pyx_k_Ceil), 0, 1, 0, 1},
    {&__pyx_kp_u_Circular_import_of, __pyx_k_Circular_import_of, sizeof(__pyx_k_Circular_import_of), 0, 1, 0, 0},
    {&__pyx_n_u_Compose, __pyx_k_Compose, sizeof(__pyx_k_Compose), 0, 1, 0, 1},
    {&__pyx_n_s_DYNAMIC_FUNCTIONS, __pyx_k_DYNAMIC_FUNCTIONS, sizeof(__pyx_k_DYNAMIC_FUNCTIONS), 0, 0, 1, 1},
    {&__pyx_n_u_Drop, __pyx_k_Drop, sizeof(__pyx_k_Drop), 0, 1, 0, 1},
    {&__pyx_n_u_Dup, __pyx_k_Dup, sizeof(__pyx_k_Dup), 0, 1, 0, 1},
    {&__pyx_n_u_EndFor, __pyx_k_EndFor, sizeof(__pyx_k_EndFor), 0, 1, 0, 1},
    {&__pyx_n_u_Eq, __pyx_k_Eq, sizeof(__pyx_k_Eq), 0, 1, 0, 1},
    {&__pyx_kp_u_Error_calling, __pyx_k_Error_calling, sizeof(__pyx_k_Error_calling), 0, 1, 0, 0},
    {&__pyx_n_u_Exit, __pyx_k_Exit, sizeof(__pyx_k_Exit), 0, 1, 0, 1},
    {&__pyx_n_u_Export, __pyx_k_Export, sizeof(__pyx_k_Export), 0, 1, 0, 1},
    {&__pyx_n_u_Floor, __pyx_k_Floor, sizeof(__pyx_k_Floor), 0, 1, 0, 1},
    {&__pyx_n_u_FloorDiv, __pyx_k_FloorDiv, sizeof(__pyx_k_FloorDiv), 0, 1, 0, 1},
    {&__pyx_n_u_Fract, __pyx_k_Fract, sizeof(__pyx_k_Fract), 0, 1, 0, 1},
    {&__pyx_n_u_Func, __pyx_k_Func, sizeof(__pyx_k_Func), 0, 1, 0, 1},
    {&__pyx_n_s_Function, __pyx_k_Function, sizeof(__pyx_k_Function), 0, 0, 1, 1},
    {&__pyx_n_s_Function___reduce_cython, __pyx_k_Function___reduce_cython, sizeof(__pyx_k_Function___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Function___setstate_cython, __pyx_k_Function___setstate_cython, sizeof(__pyx_k_Function___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_u_Ge, __pyx_k_Ge, sizeof(__pyx_k_Ge), 0, 1, 0, 1},
    {&__pyx_n_u_Gt, __pyx_k_Gt, sizeof(__pyx_k_Gt), 0, 1, 0, 1},
    {&__pyx_n_u_Import, __pyx_k_Import, sizeof(__pyx_k_Import), 0, 1, 0, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_k_Incompatible_checksums_0x_x_vs_0_8, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_8), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_k_Incompatible_checksums_0x_x_vs_0_9, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_9), 0, 0, 1, 0},
    {&__pyx_n_u_IndexLiteral, __pyx_k_IndexLiteral, sizeof(__pyx_k_IndexLiteral), 0, 1, 0, 1},
    {&__pyx_n_s_Instruction, __pyx_k_Instruction, sizeof(__pyx_k_Instruction), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionFunc, __pyx_k_InstructionFunc, sizeof(__pyx_k_InstructionFunc), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionFunc___reduce_cython, __pyx_k_InstructionFunc___reduce_cython, sizeof(__pyx_k_InstructionFunc___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionFunc___setstate_cytho, __pyx_k_InstructionFunc___setstate_cytho, sizeof(__pyx_k_InstructionFunc___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionInt, __pyx_k_InstructionInt, sizeof(__pyx_k_InstructionInt), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionIntTuple, __pyx_k_InstructionIntTuple, sizeof(__pyx_k_InstructionIntTuple), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionIntTuple___reduce_cyt, __pyx_k_InstructionIntTuple___reduce_cyt, sizeof(__pyx_k_InstructionIntTuple___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionIntTuple___setstate_c, __pyx_k_InstructionIntTuple___setstate_c, sizeof(__pyx_k_InstructionIntTuple___setstate_c), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionInt___reduce_cython, __pyx_k_InstructionInt___reduce_cython, sizeof(__pyx_k_InstructionInt___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionInt___setstate_cython, __pyx_k_InstructionInt___setstate_cython, sizeof(__pyx_k_InstructionInt___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionJump, __pyx_k_InstructionJump, sizeof(__pyx_k_InstructionJump), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionJump___reduce_cython, __pyx_k_InstructionJump___reduce_cython, sizeof(__pyx_k_InstructionJump___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionJump___setstate_cytho, __pyx_k_InstructionJump___setstate_cytho, sizeof(__pyx_k_InstructionJump___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionLabel, __pyx_k_InstructionLabel, sizeof(__pyx_k_InstructionLabel), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionLabel___reduce_cython, __pyx_k_InstructionLabel___reduce_cython, sizeof(__pyx_k_InstructionLabel___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionLabel___setstate_cyth, __pyx_k_InstructionLabel___setstate_cyth, sizeof(__pyx_k_InstructionLabel___setstate_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionNode, __pyx_k_InstructionNode, sizeof(__pyx_k_InstructionNode), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionNode___reduce_cython, __pyx_k_InstructionNode___reduce_cython, sizeof(__pyx_k_InstructionNode___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionNode___setstate_cytho, __pyx_k_InstructionNode___setstate_cytho, sizeof(__pyx_k_InstructionNode___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionObjectInt, __pyx_k_InstructionObjectInt, sizeof(__pyx_k_InstructionObjectInt), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionObjectInt___reduce_cy, __pyx_k_InstructionObjectInt___reduce_cy, sizeof(__pyx_k_InstructionObjectInt___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionObjectInt___setstate, __pyx_k_InstructionObjectInt___setstate, sizeof(__pyx_k_InstructionObjectInt___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionStr, __pyx_k_InstructionStr, sizeof(__pyx_k_InstructionStr), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionStr___reduce_cython, __pyx_k_InstructionStr___reduce_cython, sizeof(__pyx_k_InstructionStr___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionStr___setstate_cython, __pyx_k_InstructionStr___setstate_cython, sizeof(__pyx_k_InstructionStr___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionTuple, __pyx_k_InstructionTuple, sizeof(__pyx_k_InstructionTuple), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionTuple___reduce_cython, __pyx_k_InstructionTuple___reduce_cython, sizeof(__pyx_k_InstructionTuple___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionTuple___setstate_cyth, __pyx_k_InstructionTuple___setstate_cyth, sizeof(__pyx_k_InstructionTuple___setstate_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionVector, __pyx_k_InstructionVector, sizeof(__pyx_k_InstructionVector), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionVector___reduce_cytho, __pyx_k_InstructionVector___reduce_cytho, sizeof(__pyx_k_InstructionVector___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_InstructionVector___setstate_cyt, __pyx_k_InstructionVector___setstate_cyt, sizeof(__pyx_k_InstructionVector___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_Instruction___reduce_cython, __pyx_k_Instruction___reduce_cython, sizeof(__pyx_k_Instruction___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Instruction___setstate_cython, __pyx_k_Instruction___setstate_cython, sizeof(__pyx_k_Instruction___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_u_Jump, __pyx_k_Jump, sizeof(__pyx_k_Jump), 0, 1, 0, 1},
    {&__pyx_kp_u_Jump_outside_of_program, __pyx_k_Jump_outside_of_program, sizeof(__pyx_k_Jump_outside_of_program), 0, 1, 0, 0},
    {&__pyx_kp_u_L, __pyx_k_L, sizeof(__pyx_k_L), 0, 1, 0, 0},
    {&__pyx_kp_u_L_2, __pyx_k_L_2, sizeof(__pyx_k_L_2), 0, 1, 0, 0},
    {&__pyx_kp_u_L_3, __pyx_k_L_3, sizeof(__pyx_k_L_3), 0, 1, 0, 0},
    {&__pyx_n_u_Label, __pyx_k_Label, sizeof(__pyx_k_Label), 0, 1, 0, 1},
    {&__pyx_n_u_Le, __pyx_k_Le, sizeof(__pyx_k_Le), 0, 1, 0, 1},
    {&__pyx_n_u_Literal, __pyx_k_Literal, sizeof(__pyx_k_Literal), 0, 1, 0, 1},
    {&__pyx_n_u_LiteralNode, __pyx_k_LiteralNode, sizeof(__pyx_k_LiteralNode), 0, 1, 0, 1},
    {&__pyx_n_u_LiteralNodes, __pyx_k_LiteralNodes, sizeof(__pyx_k_LiteralNodes), 0, 1, 0, 1},
    {&__pyx_n_u_LocalDrop, __pyx_k_LocalDrop, sizeof(__pyx_k_LocalDrop), 0, 1, 0, 1},
    {&__pyx_n_u_LocalLoad, __pyx_k_LocalLoad, sizeof(__pyx_k_LocalLoad), 0, 1, 0, 1},
    {&__pyx_n_u_LocalPush, __pyx_k_LocalPush, sizeof(__pyx_k_LocalPush), 0, 1, 0, 1},
    {&__pyx_n_u_Lookup, __pyx_k_Lookup, sizeof(__pyx_k_Lookup), 0, 1, 0, 1},
    {&__pyx_n_u_LookupLiteral, __pyx_k_LookupLiteral, sizeof(__pyx_k_LookupLiteral), 0, 1, 0, 1},
    {&__pyx_n_s_LoopSource, __pyx_k_LoopSource, sizeof(__pyx_k_LoopSource), 0, 0, 1, 1},
    {&__pyx_n_s_LoopSource___reduce_cython, __pyx_k_LoopSource___reduce_cython, sizeof(__pyx_k_LoopSource___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LoopSource___setstate_cython, __pyx_k_LoopSource___setstate_cython, sizeof(__pyx_k_LoopSource___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_u_Lt, __pyx_k_Lt, sizeof(__pyx_k_Lt), 0, 1, 0, 1},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_kp_u_Missing_lname, __pyx_k_Missing_lname, sizeof(__pyx_k_Missing_lname), 0, 1, 0, 0},
    {&__pyx_n_u_Mod, __pyx_k_Mod, sizeof(__pyx_k_Mod), 0, 1, 0, 1},
    {&__pyx_n_u_Mul, __pyx_k_Mul, sizeof(__pyx_k_Mul), 0, 1, 0, 1},
    {&__pyx_n_u_MulAdd, __pyx_k_MulAdd, sizeof(__pyx_k_MulAdd), 0, 1, 0, 1},
    {&__pyx_n_s_NOISE_FUNCTIONS, __pyx_k_NOISE_FUNCTIONS, sizeof(__pyx_k_NOISE_FUNCTIONS), 0, 0, 1, 1},
    {&__pyx_n_u_Ne, __pyx_k_Ne, sizeof(__pyx_k_Ne), 0, 1, 0, 1},
    {&__pyx_n_u_Neg, __pyx_k_Neg, sizeof(__pyx_k_Neg), 0, 1, 0, 1},
    {&__pyx_n_u_Next, __pyx_k_Next, sizeof(__pyx_k_Next), 0, 1, 0, 1},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_n_u_Not, __pyx_k_Not, sizeof(__pyx_k_Not), 0, 1, 0, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_n_u_Pos, __pyx_k_Pos, sizeof(__pyx_k_Pos), 0, 1, 0, 1},
    {&__pyx_n_u_Pow, __pyx_k_Pow, sizeof(__pyx_k_Pow), 0, 1, 0, 1},
    {&__pyx_n_s_Program, __pyx_k_Program, sizeof(__pyx_k_Program), 0, 0, 1, 1},
    {&__pyx_n_s_Program___reduce_cython, __pyx_k_Program___reduce_cython, sizeof(__pyx_k_Program___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Program___setstate_cython, __pyx_k_Program___setstate_cython, sizeof(__pyx_k_Program___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Program__execute_locals_genexpr, __pyx_k_Program__execute_locals_genexpr, sizeof(__pyx_k_Program__execute_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_Program_add, __pyx_k_Program_add, sizeof(__pyx_k_Program_add), 0, 0, 1, 1},
    {&__pyx_n_s_Program_append, __pyx_k_Program_append, sizeof(__pyx_k_Program_append), 0, 0, 1, 1},
    {&__pyx_n_s_Program_attributes, __pyx_k_Program_attributes, sizeof(__pyx_k_Program_attributes), 0, 0, 1, 1},
    {&__pyx_n_s_Program_begin_for, __pyx_k_Program_begin_for, sizeof(__pyx_k_Program_begin_for), 0, 0, 1, 1},
    {&__pyx_n_s_Program_branch_false, __pyx_k_Program_branch_false, sizeof(__pyx_k_Program_branch_false), 0, 0, 1, 1},
    {&__pyx_n_s_Program_branch_true, __pyx_k_Program_branch_true, sizeof(__pyx_k_Program_branch_true), 0, 0, 1, 1},
    {&__pyx_n_s_Program_call, __pyx_k_Program_call, sizeof(__pyx_k_Program_call), 0, 0, 1, 1},
    {&__pyx_n_s_Program_call_fast, __pyx_k_Program_call_fast, sizeof(__pyx_k_Program_call_fast), 0, 0, 1, 1},
    {&__pyx_n_s_Program_ceil, __pyx_k_Program_ceil, sizeof(__pyx_k_Program_ceil), 0, 0, 1, 1},
    {&__pyx_n_s_Program_compose, __pyx_k_Program_compose, sizeof(__pyx_k_Program_compose), 0, 0, 1, 1},
    {&__pyx_n_s_Program_drop, __pyx_k_Program_drop, sizeof(__pyx_k_Program_drop), 0, 0, 1, 1},
    {&__pyx_n_s_Program_dup, __pyx_k_Program_dup, sizeof(__pyx_k_Program_dup), 0, 0, 1, 1},
    {&__pyx_n_s_Program_end_for, __pyx_k_Program_end_for, sizeof(__pyx_k_Program_end_for), 0, 0, 1, 1},
    {&__pyx_n_s_Program_eq, __pyx_k_Program_eq, sizeof(__pyx_k_Program_eq), 0, 0, 1, 1},
    {&__pyx_n_s_Program_execute, __pyx_k_Program_execute, sizeof(__pyx_k_Program_execute), 0, 0, 1, 1},
    {&__pyx_n_s_Program_exit, __pyx_k_Program_exit, sizeof(__pyx_k_Program_exit), 0, 0, 1, 1},
    {&__pyx_n_s_Program_export, __pyx_k_Program_export, sizeof(__pyx_k_Program_export), 0, 0, 1, 1},
    {&__pyx_n_s_Program_floor, __pyx_k_Program_floor, sizeof(__pyx_k_Program_floor), 0, 0, 1, 1},
    {&__pyx_n_s_Program_floordiv, __pyx_k_Program_floordiv, sizeof(__pyx_k_Program_floordiv), 0, 0, 1, 1},
    {&__pyx_n_s_Program_fract, __pyx_k_Program_fract, sizeof(__pyx_k_Program_fract), 0, 0, 1, 1},
    {&__pyx_n_s_Program_func, __pyx_k_Program_func, sizeof(__pyx_k_Program_func), 0, 0, 1, 1},
    {&__pyx_n_s_Program_ge, __pyx_k_Program_ge, sizeof(__pyx_k_Program_ge), 0, 0, 1, 1},
    {&__pyx_n_s_Program_gt, __pyx_k_Program_gt, sizeof(__pyx_k_Program_gt), 0, 0, 1, 1},
    {&__pyx_kp_u_Program_has_not_been_linked, __pyx_k_Program_has_not_been_linked, sizeof(__pyx_k_Program_has_not_been_linked), 0, 1, 0, 0},
    {&__pyx_n_s_Program_import, __pyx_k_Program_import, sizeof(__pyx_k_Program_import), 0, 0, 1, 1},
    {&__pyx_n_s_Program_jump, __pyx_k_Program_jump, sizeof(__pyx_k_Program_jump), 0, 0, 1, 1},
    {&__pyx_n_s_Program_label, __pyx_k_Program_label, sizeof(__pyx_k_Program_label), 0, 0, 1, 1},
    {&__pyx_n_s_Program_le, __pyx_k_Program_le, sizeof(__pyx_k_Program_le), 0, 0, 1, 1},
    {&__pyx_n_s_Program_link, __pyx_k_Program_link, sizeof(__pyx_k_Program_link), 0, 0, 1, 1},
    {&__pyx_n_s_Program_literal, __pyx_k_Program_literal, sizeof(__pyx_k_Program_literal), 0, 0, 1, 1},
    {&__pyx_n_s_Program_local_drop, __pyx_k_Program_local_drop, sizeof(__pyx_k_Program_local_drop), 0, 0, 1, 1},
    {&__pyx_n_s_Program_local_load, __pyx_k_Program_local_load, sizeof(__pyx_k_Program_local_load), 0, 0, 1, 1},
    {&__pyx_n_s_Program_local_push, __pyx_k_Program_local_push, sizeof(__pyx_k_Program_local_push), 0, 0, 1, 1},
    {&__pyx_n_s_Program_lookup, __pyx_k_Program_lookup, sizeof(__pyx_k_Program_lookup), 0, 0, 1, 1},
    {&__pyx_n_s_Program_lookup_literal, __pyx_k_Program_lookup_literal, sizeof(__pyx_k_Program_lookup_literal), 0, 0, 1, 1},
    {&__pyx_n_s_Program_lt, __pyx_k_Program_lt, sizeof(__pyx_k_Program_lt), 0, 0, 1, 1},
    {&__pyx_n_s_Program_mod, __pyx_k_Program_mod, sizeof(__pyx_k_Program_mod), 0, 0, 1, 1},
    {&__pyx_n_s_Program_mul, __pyx_k_Program_mul, sizeof(__pyx_k_Program_mul), 0, 0, 1, 1},
    {&__pyx_n_s_Program_mul_add, __pyx_k_Program_mul_add, sizeof(__pyx_k_Program_mul_add), 0, 0, 1, 1},
    {&__pyx_n_s_Program_ne, __pyx_k_Program_ne, sizeof(__pyx_k_Program_ne), 0, 0, 1, 1},
    {&__pyx_n_s_Program_neg, __pyx_k_Program_neg, sizeof(__pyx_k_Program_neg), 0, 0, 1, 1},
    {&__pyx_n_s_Program_new_label, __pyx_k_Program_new_label, sizeof(__pyx_k_Program_new_label), 0, 0, 1, 1},
    {&__pyx_n_s_Program_next, __pyx_k_Program_next, sizeof(__pyx_k_Program_next), 0, 0, 1, 1},
    {&__pyx_n_s_Program_not, __pyx_k_Program_not, sizeof(__pyx_k_Program_not), 0, 0, 1, 1},
    {&__pyx_n_s_Program_pos, __pyx_k_Program_pos, sizeof(__pyx_k_Program_pos), 0, 0, 1, 1},
    {&__pyx_n_s_Program_pow, __pyx_k_Program_pow, sizeof(__pyx_k_Program_pow), 0, 0, 1, 1},
    {&__pyx_n_s_Program_range, __pyx_k_Program_range, sizeof(__pyx_k_Program_range), 0, 0, 1, 1},
    {&__pyx_n_s_Program_run, __pyx_k_Program_run, sizeof(__pyx_k_Program_run), 0, 0, 1, 1},
    {&__pyx_n_s_Program_set_path, __pyx_k_Program_set_path, sizeof(__pyx_k_Program_set_path), 0, 0, 1, 1},
    {&__pyx_n_s_Program_set_pragma, __pyx_k_Program_set_pragma, sizeof(__pyx_k_Program_set_pragma), 0, 0, 1, 1},
    {&__pyx_n_s_Program_set_top, __pyx_k_Program_set_top, sizeof(__pyx_k_Program_set_top), 0, 0, 1, 1},
    {&__pyx_n_s_Program_slice, __pyx_k_Program_slice, sizeof(__pyx_k_Program_slice), 0, 0, 1, 1},
    {&__pyx_n_s_Program_slice_literal, __pyx_k_Program_slice_literal, sizeof(__pyx_k_Program_slice_literal), 0, 0, 1, 1},
    {&__pyx_n_s_Program_sub, __pyx_k_Program_sub, sizeof(__pyx_k_Program_sub), 0, 0, 1, 1},
    {&__pyx_n_s_Program_tag, __pyx_k_Program_tag, sizeof(__pyx_k_Program_tag), 0, 0, 1, 1},
    {&__pyx_n_s_Program_truediv, __pyx_k_Program_truediv, sizeof(__pyx_k_Program_truediv), 0, 0, 1, 1},
    {&__pyx_n_s_Program_use_simplifier, __pyx_k_Program_use_simplifier, sizeof(__pyx_k_Program_use_simplifier), 0, 0, 1, 1},
    {&__pyx_n_s_Program_xor, __pyx_k_Program_xor, sizeof(__pyx_k_Program_xor), 0, 0, 1, 1},
    {&__pyx_n_u_Range, __pyx_k_Range, sizeof(__pyx_k_Range), 0, 1, 0, 1},
    {&__pyx_n_s_RecursionError, __pyx_k_RecursionError, sizeof(__pyx_k_RecursionError), 0, 0, 1, 1},
    {&__pyx_kp_u_Recursion_depth_exceeded_for_fun, __pyx_k_Recursion_depth_exceeded_for_fun, sizeof(__pyx_k_Recursion_depth_exceeded_for_fun), 0, 1, 0, 0},
    {&__pyx_n_s_STATIC_FUNCTIONS, __pyx_k_STATIC_FUNCTIONS, sizeof(__pyx_k_STATIC_FUNCTIONS), 0, 0, 1, 1},
    {&__pyx_n_s_SharedCache, __pyx_k_SharedCache, sizeof(__pyx_k_SharedCache), 0, 0, 1, 1},
    {&__pyx_n_u_Slice, __pyx_k_Slice, sizeof(__pyx_k_Slice), 0, 1, 0, 1},
    {&__pyx_n_u_SliceLiteral, __pyx_k_SliceLiteral, sizeof(__pyx_k_SliceLiteral), 0, 1, 0, 1},
    {&__pyx_n_s_StackError, __pyx_k_StackError, sizeof(__pyx_k_StackError), 0, 0, 1, 1},
    {&__pyx_n_s_StackOverflow, __pyx_k_StackOverflow, sizeof(__pyx_k_StackOverflow), 0, 0, 1, 1},
    {&__pyx_n_s_StackUnderflow, __pyx_k_StackUnderflow, sizeof(__pyx_k_StackUnderflow), 0, 0, 1, 1},
    {&__pyx_kp_u_Stack_cannot_grow_beyond, __pyx_k_Stack_cannot_grow_beyond, sizeof(__pyx_k_Stack_cannot_grow_beyond), 0, 1, 0, 0},
    {&__pyx_kp_u_Stack_empty, __pyx_k_Stack_empty, sizeof(__pyx_k_Stack_empty), 0, 1, 0, 0},
    {&__pyx_kp_u_Stack_overflow_in_program, __pyx_k_Stack_overflow_in_program, sizeof(__pyx_k_Stack_overflow_in_program), 0, 1, 0, 0},
    {&__pyx_n_u_Sub, __pyx_k_Sub, sizeof(__pyx_k_Sub), 0, 1, 0, 1},
    {&__pyx_n_u_Tag, __pyx_k_Tag, sizeof(__pyx_k_Tag), 0, 1, 0, 1},
    {&__pyx_n_u_TrueDiv, __pyx_k_TrueDiv, sizeof(__pyx_k_TrueDiv), 0, 1, 0, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Unable_to_import, __pyx_k_Unable_to_import, sizeof(__pyx_k_Unable_to_import), 0, 1, 0, 0},
    {&__pyx_kp_u_Unable_to_import_from, __pyx_k_Unable_to_import_from, sizeof(__pyx_k_Unable_to_import_from), 0, 1, 0, 0},
    {&__pyx_kp_u_Unrecognised_instruction, __pyx_k_Unrecognised_instruction, sizeof(__pyx_k_Unrecognised_instruction), 0, 1, 0, 0},
    {&__pyx_kp_u_VM_exception_processing, __pyx_k_VM_exception_processing, sizeof(__pyx_k_VM_exception_processing), 0, 1, 0, 0},
    {&__pyx_kp_u_VM_execution_statistics, __pyx_k_VM_execution_statistics, sizeof(__pyx_k_VM_execution_statistics), 0, 1, 0, 0},
    {&__pyx_n_s_VectorStack, __pyx_k_VectorStack, sizeof(__pyx_k_VectorStack), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack___reduce_cython, __pyx_k_VectorStack___reduce_cython, sizeof(__pyx_k_VectorStack___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack___setstate_cython, __pyx_k_VectorStack___setstate_cython, sizeof(__pyx_k_VectorStack___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_copy, __pyx_k_VectorStack_copy, sizeof(__pyx_k_VectorStack_copy), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_drop, __pyx_k_VectorStack_drop, sizeof(__pyx_k_VectorStack_drop), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_peek, __pyx_k_VectorStack_peek, sizeof(__pyx_k_VectorStack_peek), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_peek_at, __pyx_k_VectorStack_peek_at, sizeof(__pyx_k_VectorStack_peek_at), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_poke, __pyx_k_VectorStack_poke, sizeof(__pyx_k_VectorStack_poke), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_poke_at, __pyx_k_VectorStack_poke_at, sizeof(__pyx_k_VectorStack_poke_at), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_pop, __pyx_k_VectorStack_pop, sizeof(__pyx_k_VectorStack_pop), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_pop_composed, __pyx_k_VectorStack_pop_composed, sizeof(__pyx_k_VectorStack_pop_composed), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_pop_dict, __pyx_k_VectorStack_pop_dict, sizeof(__pyx_k_VectorStack_pop_dict), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_pop_list, __pyx_k_VectorStack_pop_list, sizeof(__pyx_k_VectorStack_pop_list), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_pop_tuple, __pyx_k_VectorStack_pop_tuple, sizeof(__pyx_k_VectorStack_pop_tuple), 0, 0, 1, 1},
    {&__pyx_n_s_VectorStack_push, __pyx_k_VectorStack_push, sizeof(__pyx_k_VectorStack_push), 0, 0, 1, 1},
    {&__pyx_n_u_Xor, __pyx_k_Xor, sizeof(__pyx_k_Xor), 0, 1, 0, 1},
    {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
    {&__pyx_n_s__163, __pyx_k__163, sizeof(__pyx_k__163), 0, 0, 1, 1},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_n_s__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 0, 1, 1},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_attributes, __pyx_k_attributes, sizeof(__pyx_k_attributes), 0, 0, 1, 1},
    {&__pyx_n_s_begin_for, __pyx_k_begin_for, sizeof(__pyx_k_begin_for), 0, 0, 1, 1},
    {&__pyx_n_s_branch_false, __pyx_k_branch_false, sizeof(__pyx_k_branch_false), 0, 0, 1, 1},
    {&__pyx_n_s_branch_true, __pyx_k_branch_true, sizeof(__pyx_k_branch_true), 0, 0, 1, 1},
    {&__pyx_n_s_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
    {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
    {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
    {&__pyx_n_s_call_fast, __pyx_k_call_fast, sizeof(__pyx_k_call_fast), 0, 0, 1, 1},
    {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
    {&__pyx_n_s_compose, __pyx_k_compose, sizeof(__pyx_k_compose), 0, 0, 1, 1},
    {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
    {&__pyx_kp_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 0},
    {&__pyx_kp_u_depth_stack, __pyx_k_depth_stack, sizeof(__pyx_k_depth_stack), 0, 1, 0, 0},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
    {&__pyx_n_s_drop, __pyx_k_drop, sizeof(__pyx_k_drop), 0, 0, 1, 1},
    {&__pyx_n_s_dup, __pyx_k_dup, sizeof(__pyx_k_dup), 0, 0, 1, 1},
    {&__pyx_n_s_duration, __pyx_k_duration, sizeof(__pyx_k_duration), 0, 0, 1, 1},
    {&__pyx_n_s_dynamic_builtins, __pyx_k_dynamic_builtins, sizeof(__pyx_k_dynamic_builtins), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
    {&__pyx_n_s_end_for, __pyx_k_end_for, sizeof(__pyx_k_end_for), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
    {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
    {&__pyx_n_s_execute, __pyx_k_execute, sizeof(__pyx_k_execute), 0, 0, 1, 1},
    {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
    {&__pyx_n_s_export, __pyx_k_export, sizeof(__pyx_k_export), 0, 0, 1, 1},
    {&__pyx_n_u_false, __pyx_k_false, sizeof(__pyx_k_false), 0, 1, 0, 1},
    {&__pyx_n_s_flitter_language_vm, __pyx_k_flitter_language_vm, sizeof(__pyx_k_flitter_language_vm), 0, 0, 1, 1},
    {&__pyx_n_s_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
    {&__pyx_n_s_floordiv, __pyx_k_floordiv, sizeof(__pyx_k_floordiv), 0, 0, 1, 1},
    {&__pyx_n_s_fract, __pyx_k_fract, sizeof(__pyx_k_fract), 0, 0, 1, 1},
    {&__pyx_kp_u_from, __pyx_k_from, sizeof(__pyx_k_from), 0, 1, 0, 0},
    {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
    {&__pyx_n_s_function, __pyx_k_function, sizeof(__pyx_k_function), 0, 0, 1, 1},
    {&__pyx_n_s_functions, __pyx_k_functions, sizeof(__pyx_k_functions), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_ge, __pyx_k_ge, sizeof(__pyx_k_ge), 0, 0, 1, 1},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get_with_root, __pyx_k_get_with_root, sizeof(__pyx_k_get_with_root), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_gt, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_import_2, __pyx_k_import_2, sizeof(__pyx_k_import_2), 0, 0, 1, 1},
    {&__pyx_n_u_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 1, 0, 1},
    {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_item, __pyx_k_item, sizeof(__pyx_k_item), 0, 0, 1, 1},
    {&__pyx_kp_u_item_stack, __pyx_k_item_stack, sizeof(__pyx_k_item_stack), 0, 1, 0, 0},
    {&__pyx_kp_u_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 1, 0, 0},
    {&__pyx_n_s_items_2, __pyx_k_items_2, sizeof(__pyx_k_items_2), 0, 0, 1, 1},
    {&__pyx_kp_u_items_below_top_of, __pyx_k_items_below_top_of, sizeof(__pyx_k_items_below_top_of), 0, 1, 0, 0},
    {&__pyx_kp_u_items_from, __pyx_k_items_from, sizeof(__pyx_k_items_from), 0, 1, 0, 0},
    {&__pyx_n_s_ivalue, __pyx_k_ivalue, sizeof(__pyx_k_ivalue), 0, 0, 1, 1},
    {&__pyx_n_s_jump, __pyx_k_jump, sizeof(__pyx_k_jump), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
    {&__pyx_n_s_le, __pyx_k_le, sizeof(__pyx_k_le), 0, 0, 1, 1},
    {&__pyx_n_s_link, __pyx_k_link, sizeof(__pyx_k_link), 0, 0, 1, 1},
    {&__pyx_n_s_literal, __pyx_k_literal, sizeof(__pyx_k_literal), 0, 0, 1, 1},
    {&__pyx_n_s_lnames, __pyx_k_lnames, sizeof(__pyx_k_lnames), 0, 0, 1, 1},
    {&__pyx_n_s_local_drop, __pyx_k_local_drop, sizeof(__pyx_k_local_drop), 0, 0, 1, 1},
    {&__pyx_n_s_local_load, __pyx_k_local_load, sizeof(__pyx_k_local_load), 0, 0, 1, 1},
    {&__pyx_n_s_local_push, __pyx_k_local_push, sizeof(__pyx_k_local_push), 0, 0, 1, 1},
    {&__pyx_n_s_log_vm_stats, __pyx_k_log_vm_stats, sizeof(__pyx_k_log_vm_stats), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_loguru, __pyx_k_loguru, sizeof(__pyx_k_loguru), 0, 0, 1, 1},
    {&__pyx_n_s_lookup, __pyx_k_lookup, sizeof(__pyx_k_lookup), 0, 0, 1, 1},
    {&__pyx_n_s_lookup_literal, __pyx_k_lookup_literal, sizeof(__pyx_k_lookup_literal), 0, 0, 1, 1},
    {&__pyx_n_s_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_max_size, __pyx_k_max_size, sizeof(__pyx_k_max_size), 0, 0, 1, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_mod, __pyx_k_mod, sizeof(__pyx_k_mod), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_mul, __pyx_k_mul, sizeof(__pyx_k_mul), 0, 0, 1, 1},
    {&__pyx_n_s_mul_add, __pyx_k_mul_add, sizeof(__pyx_k_mul_add), 0, 0, 1, 1},
    {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_name_patch, __pyx_k_name_patch, sizeof(__pyx_k_name_patch), 0, 0, 1, 1},
    {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
    {&__pyx_n_u_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 1, 0, 1},
    {&__pyx_kp_u_native_funcs, __pyx_k_native_funcs, sizeof(__pyx_k_native_funcs), 0, 1, 0, 0},
    {&__pyx_n_s_ncaptures, __pyx_k_ncaptures, sizeof(__pyx_k_ncaptures), 0, 0, 1, 1},
    {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
    {&__pyx_n_s_neg, __pyx_k_neg, sizeof(__pyx_k_neg), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_new_label, __pyx_k_new_label, sizeof(__pyx_k_new_label), 0, 0, 1, 1},
    {&__pyx_n_s_next, __pyx_k_next, sizeof(__pyx_k_next), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_noise, __pyx_k_noise, sizeof(__pyx_k_noise), 0, 0, 1, 1},
    {&__pyx_n_s_not, __pyx_k_not, sizeof(__pyx_k_not), 0, 0, 1, 1},
    {&__pyx_n_u_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 1, 0, 1},
    {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
    {&__pyx_n_s_objptr, __pyx_k_objptr, sizeof(__pyx_k_objptr), 0, 0, 1, 1},
    {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
    {&__pyx_n_s_overhead, __pyx_k_overhead, sizeof(__pyx_k_overhead), 0, 0, 1, 1},
    {&__pyx_n_s_parameters, __pyx_k_parameters, sizeof(__pyx_k_parameters), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_n_s_peek, __pyx_k_peek, sizeof(__pyx_k_peek), 0, 0, 1, 1},
    {&__pyx_n_s_peek_at, __pyx_k_peek_at, sizeof(__pyx_k_peek_at), 0, 0, 1, 1},
    {&__pyx_n_s_per_execution, __pyx_k_per_execution, sizeof(__pyx_k_per_execution), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_poke, __pyx_k_poke, sizeof(__pyx_k_poke), 0, 0, 1, 1},
    {&__pyx_n_s_poke_at, __pyx_k_poke_at, sizeof(__pyx_k_poke_at), 0, 0, 1, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_pop_composed, __pyx_k_pop_composed, sizeof(__pyx_k_pop_composed), 0, 0, 1, 1},
    {&__pyx_n_s_pop_dict, __pyx_k_pop_dict, sizeof(__pyx_k_pop_dict), 0, 0, 1, 1},
    {&__pyx_n_s_pop_list, __pyx_k_pop_list, sizeof(__pyx_k_pop_list), 0, 0, 1, 1},
    {&__pyx_n_s_pop_tuple, __pyx_k_pop_tuple, sizeof(__pyx_k_pop_tuple), 0, 0, 1, 1},
    {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
    {&__pyx_n_s_pow, __pyx_k_pow, sizeof(__pyx_k_pow), 0, 0, 1, 1},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_push, __pyx_k_push, sizeof(__pyx_k_push), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Function, __pyx_k_pyx_unpickle_Function, sizeof(__pyx_k_pyx_unpickle_Function), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Instruction, __pyx_k_pyx_unpickle_Instruction, sizeof(__pyx_k_pyx_unpickle_Instruction), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionFunc, __pyx_k_pyx_unpickle_InstructionFunc, sizeof(__pyx_k_pyx_unpickle_InstructionFunc), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionInt, __pyx_k_pyx_unpickle_InstructionInt, sizeof(__pyx_k_pyx_unpickle_InstructionInt), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionIntTup, __pyx_k_pyx_unpickle_InstructionIntTup, sizeof(__pyx_k_pyx_unpickle_InstructionIntTup), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionJump, __pyx_k_pyx_unpickle_InstructionJump, sizeof(__pyx_k_pyx_unpickle_InstructionJump), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionLabel, __pyx_k_pyx_unpickle_InstructionLabel, sizeof(__pyx_k_pyx_unpickle_InstructionLabel), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionNode, __pyx_k_pyx_unpickle_InstructionNode, sizeof(__pyx_k_pyx_unpickle_InstructionNode), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionObject, __pyx_k_pyx_unpickle_InstructionObject, sizeof(__pyx_k_pyx_unpickle_InstructionObject), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionStr, __pyx_k_pyx_unpickle_InstructionStr, sizeof(__pyx_k_pyx_unpickle_InstructionStr), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionTuple, __pyx_k_pyx_unpickle_InstructionTuple, sizeof(__pyx_k_pyx_unpickle_InstructionTuple), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InstructionVector, __pyx_k_pyx_unpickle_InstructionVector, sizeof(__pyx_k_pyx_unpickle_InstructionVector), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LoopSource, __pyx_k_pyx_unpickle_LoopSource, sizeof(__pyx_k_pyx_unpickle_LoopSource), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_read_flitter_program, __pyx_k_read_flitter_program, sizeof(__pyx_k_read_flitter_program), 0, 0, 1, 1},
    {&__pyx_n_s_record_stats, __pyx_k_record_stats, sizeof(__pyx_k_record_stats), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
    {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
    {&__pyx_n_u_root, __pyx_k_root, sizeof(__pyx_k_root), 0, 1, 0, 1},
    {&__pyx_n_s_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_set_path, __pyx_k_set_path, sizeof(__pyx_k_set_path), 0, 0, 1, 1},
    {&__pyx_n_s_set_pragma, __pyx_k_set_pragma, sizeof(__pyx_k_set_pragma), 0, 0, 1, 1},
    {&__pyx_n_s_set_top, __pyx_k_set_top, sizeof(__pyx_k_set_top), 0, 0, 1, 1},
    {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_simplify, __pyx_k_simplify, sizeof(__pyx_k_simplify), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_s_slice, __pyx_k_slice, sizeof(__pyx_k_slice), 0, 0, 1, 1},
    {&__pyx_n_s_slice_literal, __pyx_k_slice_literal, sizeof(__pyx_k_slice_literal), 0, 0, 1, 1},
    {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
    {&__pyx_kp_s_src_flitter_language_vm_pyx, __pyx_k_src_flitter_language_vm_pyx, sizeof(__pyx_k_src_flitter_language_vm_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_static_builtins, __pyx_k_static_builtins, sizeof(__pyx_k_static_builtins), 0, 0, 1, 1},
    {&__pyx_n_s_stats, __pyx_k_stats, sizeof(__pyx_k_stats), 0, 0, 1, 1},
    {&__pyx_n_s_str___locals_genexpr, __pyx_k_str___locals_genexpr, sizeof(__pyx_k_str___locals_genexpr), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_sub, __pyx_k_sub, sizeof(__pyx_k_sub), 0, 0, 1, 1},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_tag, __pyx_k_tag, sizeof(__pyx_k_tag), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
    {&__pyx_n_s_total, __pyx_k_total, sizeof(__pyx_k_total), 0, 0, 1, 1},
    {&__pyx_n_u_true, __pyx_k_true, sizeof(__pyx_k_true), 0, 1, 0, 1},
    {&__pyx_n_s_truediv, __pyx_k_truediv, sizeof(__pyx_k_truediv), 0, 0, 1, 1},
    {&__pyx_n_s_tvalue, __pyx_k_tvalue, sizeof(__pyx_k_tvalue), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_use_simplifier, __pyx_k_use_simplifier, sizeof(__pyx_k_use_simplifier), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_vector, __pyx_k_vector, sizeof(__pyx_k_vector), 0, 0, 1, 1},
    {&__pyx_n_s_xor, __pyx_k_xor, sizeof(__pyx_k_xor), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 250, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 336, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 494, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 859, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "flitter/language/vm.pyx":635
 *     import_context.stack = stack
 *     import_context.lnames = lnames
 *     push(stack, Vector(Node('root')))             # <<<<<<<<<<<<<<
 *     if record_stats:
 *         duration[0] += perf_counter()
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_n_u_root); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5694d08, 0xe6fb062, 0xc133679):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5694d08, 0xe6fb062, 0xc133679) = (code))" % __pyx_checksum
 */
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_90787080, __pyx_int_242200674, __pyx_int_202585721); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_int_184517443, __pyx_int_71438266, __pyx_int_72616288); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_73689043, __pyx_int_107792009, __pyx_int_109158536); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_int_163945588, __pyx_int_61928532, __pyx_int_210616022); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_tuple__16 = PyTuple_Pack(3, __pyx_int_236313263, __pyx_int_176342959, __pyx_int_19407059); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_int_200977137, __pyx_int_259995473, __pyx_int_29217351); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_int_59839126, __pyx_int_138499235, __pyx_int_186494852); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_tuple__19 = PyTuple_Pack(3, __pyx_int_10904070, __pyx_int_184506718, __pyx_int_63712393); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_tuple__20 = PyTuple_Pack(3, __pyx_int_45504841, __pyx_int_229249173, __pyx_int_231591000); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__22 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)
 */
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)
 */
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)
 */
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "flitter/language/vm.pyx":272
 *         return '\n'.join(items)
 * 
 *     cpdef VectorStack copy(self):             # <<<<<<<<<<<<<<
 *         cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)
 *         cdef int64_t i
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_copy, 272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "flitter/language/vm.pyx":283
 *         return new_stack
 * 
 *     cpdef void drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         drop(self, count)
 * 
 */
  __pyx_tuple__48 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_count); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_drop, 283, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 283, __pyx_L1_error)
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "flitter/language/vm.pyx":286
 *         drop(self, count)
 * 
 *     cpdef void push(self, Vector vector):             # <<<<<<<<<<<<<<
 *         push(self, vector)
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_vector); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_push, 286, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 286, __pyx_L1_error)

  /* "flitter/language/vm.pyx":289
 *         push(self, vector)
 * 
 *     cpdef Vector pop(self):             # <<<<<<<<<<<<<<
 *         return pop(self)
 * 
 */
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pop, 289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 289, __pyx_L1_error)

  /* "flitter/language/vm.pyx":292
 *         return pop(self)
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_tuple(self, count)
 * 
 */
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pop_tuple, 292, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 292, __pyx_L1_error)

  /* "flitter/language/vm.pyx":295
 *         return pop_tuple(self, count)
 * 
 *     cpdef list pop_list(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_list(self, count)
 * 
 */
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pop_list, 295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 295, __pyx_L1_error)

  /* "flitter/language/vm.pyx":298
 *         return pop_list(self, count)
 * 
 *     cpdef dict pop_dict(self, tuple keys):             # <<<<<<<<<<<<<<
 *         return pop_dict(self, keys)
 * 
 */
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_keys); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pop_dict, 298, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 298, __pyx_L1_error)

  /* "flitter/language/vm.pyx":301
 *         return pop_dict(self, keys)
 * 
 *     cpdef Vector pop_composed(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_composed(self, count)
 * 
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pop_composed, 301, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 301, __pyx_L1_error)

  /* "flitter/language/vm.pyx":304
 *         return pop_composed(self, count)
 * 
 *     cpdef Vector peek(self):             # <<<<<<<<<<<<<<
 *         return peek(self)
 * 
 */
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_peek, 304, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 304, __pyx_L1_error)

  /* "flitter/language/vm.pyx":307
 *         return peek(self)
 * 
 *     cpdef Vector peek_at(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         return peek_at(self, offset)
 * 
 */
  __pyx_tuple__60 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_offset); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_peek_at, 307, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 307, __pyx_L1_error)

  /* "flitter/language/vm.pyx":310
 *         return peek_at(self, offset)
 * 
 *     cpdef void poke(self, Vector vector):             # <<<<<<<<<<<<<<
 *         poke(self, vector)
 * 
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_poke, 310, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 310, __pyx_L1_error)

  /* "flitter/language/vm.pyx":313
 *         poke(self, vector)
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *         poke_at(self, offset, vector)
 * 
 */
  __pyx_tuple__63 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_offset, __pyx_n_s_vector); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_poke_at, 313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 313, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "flitter/language/vm.pyx":542
 * 
 * 
 * def log_vm_stats():             # <<<<<<<<<<<<<<
 *     cdef list stats = []
 *     cdef double duration, total=0
 */
  __pyx_tuple__71 = PyTuple_Pack(10, __pyx_n_s_stats, __pyx_n_s_duration, __pyx_n_s_total, __pyx_n_s_count, __pyx_n_s_i, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_overhead, __pyx_n_s_per_execution, __pyx_n_s_name); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_log_vm_stats, 542, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 542, __pyx_L1_error)

  /* "flitter/language/vm.pyx":792
 *         return self
 * 
 *     def set_path(self, object path):             # <<<<<<<<<<<<<<
 *         self.path = path
 * 
 */
  __pyx_tuple__73 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_path); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_set_path, 792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 792, __pyx_L1_error)

  /* "flitter/language/vm.pyx":795
 *         self.path = path
 * 
 *     def set_top(self, object top):             # <<<<<<<<<<<<<<
 *         self.top = top
 * 
 */
  __pyx_tuple__75 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_top); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_set_top, 795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "flitter/language/vm.pyx":798
 *         self.top = top
 * 
 *     def set_pragma(self, str name, Vector value):             # <<<<<<<<<<<<<<
 *         self.pragmas[name] = value
 *         return self
 */
  __pyx_tuple__77 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_value); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_set_pragma, 798, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 798, __pyx_L1_error)

  /* "flitter/language/vm.pyx":802
 *         return self
 * 
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 */
  __pyx_tuple__79 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_context, __pyx_n_s_lnames, __pyx_n_s_record_stats, __pyx_n_s_item); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_execute, 802, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 802, __pyx_L1_error)
  __pyx_tuple__81 = PyTuple_Pack(2, Py_None, Py_False); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "flitter/language/vm.pyx":821
 *         return context.stack.pop_list(len(context.stack))
 * 
 *     def run(self, Context context, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         if context.stack is None:
 *             if self.stack is None:
 */
  __pyx_tuple__82 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_context, __pyx_n_s_record_stats, __pyx_n_s_i, __pyx_n_s_n, __pyx_n_s_objptr, __pyx_n_s_result); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_run, 821, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 821, __pyx_L1_error)
  __pyx_tuple__84 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "flitter/language/vm.pyx":853
 *         return context
 * 
 *     cpdef Program link(self):             # <<<<<<<<<<<<<<
 *         cdef Instruction instruction
 *         cdef int64_t label, address, target
 */
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_link, 853, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 853, __pyx_L1_error)

  /* "flitter/language/vm.pyx":872
 *         return self
 * 
 *     cpdef void use_simplifier(self, bint simplify):             # <<<<<<<<<<<<<<
 *         self.simplify = simplify
 * 
 */
  __pyx_tuple__86 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_simplify); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__86, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_use_simplifier, 872, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 872, __pyx_L1_error)

  /* "flitter/language/vm.pyx":875
 *         self.simplify = simplify
 * 
 *     cpdef int64_t new_label(self):             # <<<<<<<<<<<<<<
 *         cdef object label = self.next_label
 *         self.next_label += 1
 */
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_new_label, 875, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 875, __pyx_L1_error)

  /* "flitter/language/vm.pyx":880
 *         return label
 * 
 *     cpdef Program dup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Dup))
 *         return self
 */
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_dup, 880, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 880, __pyx_L1_error)

  /* "flitter/language/vm.pyx":884
 *         return self
 * 
 *     cpdef Program drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))
 *         return self
 */
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_drop, 884, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 884, __pyx_L1_error)

  /* "flitter/language/vm.pyx":888
 *         return self
 * 
 *     cpdef Program label(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionLabel(label))
 *         return self
 */
  __pyx_tuple__91 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_label); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_label, 888, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 888, __pyx_L1_error)

  /* "flitter/language/vm.pyx":892
 *         return self
 * 
 *     cpdef Program jump(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))
 *         return self
 */
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_jump, 892, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 892, __pyx_L1_error)

  /* "flitter/language/vm.pyx":896
 *         return self
 * 
 *     cpdef Program branch_true(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))
 *         return self
 */
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_branch_true, 896, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 896, __pyx_L1_error)

  /* "flitter/language/vm.pyx":900
 *         return self
 * 
 *     cpdef Program branch_false(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))
 *         return self
 */
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_branch_false, 900, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 900, __pyx_L1_error)

  /* "flitter/language/vm.pyx":904
 *         return self
 * 
 *     cpdef Program import_(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))
 *         return self
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_names); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_import, 904, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 904, __pyx_L1_error)

  /* "flitter/language/vm.pyx":908
 *         return self
 * 
 *     cpdef Program literal(self, value):             # <<<<<<<<<<<<<<
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_value); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_literal, 908, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 908, __pyx_L1_error)

  /* "flitter/language/vm.pyx":925
 *         return self
 * 
 *     cpdef Program local_push(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))
 *         return self
 */
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_local_push, 925, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 925, __pyx_L1_error)

  /* "flitter/language/vm.pyx":929
 *         return self
 * 
 *     cpdef Program local_load(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))
 *         return self
 */
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_local_load, 929, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 929, __pyx_L1_error)

  /* "flitter/language/vm.pyx":933
 *         return self
 * 
 *     cpdef Program local_drop(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))
 *         return self
 */
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_local_drop, 933, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 933, __pyx_L1_error)

  /* "flitter/language/vm.pyx":937
 *         return self
 * 
 *     cpdef Program lookup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lookup))
 *         return self
 */
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_lookup, 937, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 937, __pyx_L1_error)

  /* "flitter/language/vm.pyx":941
 *         return self
 * 
 *     cpdef Program lookup_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))
 *         return self
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_lookup_literal, 941, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 941, __pyx_L1_error)

  /* "flitter/language/vm.pyx":945
 *         return self
 * 
 *     cpdef Program range(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Range))
 *         return self
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_range, 945, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 945, __pyx_L1_error)

  /* "flitter/language/vm.pyx":949
 *         return self
 * 
 *     cpdef Program neg(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Neg))
 *         return self
 */
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_neg, 949, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 949, __pyx_L1_error)

  /* "flitter/language/vm.pyx":953
 *         return self
 * 
 *     cpdef Program pos(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pos))
 *         return self
 */
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pos, 953, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 953, __pyx_L1_error)

  /* "flitter/language/vm.pyx":957
 *         return self
 * 
 *     cpdef Program ceil(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ceil))
 *         return self
 */
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_ceil, 957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 957, __pyx_L1_error)

  /* "flitter/language/vm.pyx":961
 *         return self
 * 
 *     cpdef Program floor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Floor))
 *         return self
 */
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_floor, 961, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 961, __pyx_L1_error)

  /* "flitter/language/vm.pyx":965
 *         return self
 * 
 *     cpdef Program fract(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Fract))
 *         return self
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_fract, 965, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 965, __pyx_L1_error)

  /* "flitter/language/vm.pyx":969
 *         return self
 * 
 *     cpdef Program not_(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Not))
 *         return self
 */
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_not, 969, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 969, __pyx_L1_error)

  /* "flitter/language/vm.pyx":973
 *         return self
 * 
 *     cpdef Program add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Add))
 *         return self
 */
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_add, 973, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 973, __pyx_L1_error)

  /* "flitter/language/vm.pyx":977
 *         return self
 * 
 *     cpdef Program sub(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Sub))
 *         return self
 */
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_sub, 977, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 977, __pyx_L1_error)

  /* "flitter/language/vm.pyx":981
 *         return self
 * 
 *     cpdef Program mul(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mul))
 *         return self
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_mul, 981, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 981, __pyx_L1_error)

  /* "flitter/language/vm.pyx":985
 *         return self
 * 
 *     cpdef Program mul_add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.MulAdd))
 *         return self
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_mul_add, 985, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 985, __pyx_L1_error)

  /* "flitter/language/vm.pyx":989
 *         return self
 * 
 *     cpdef Program truediv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.TrueDiv))
 *         return self
 */
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_truediv, 989, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 989, __pyx_L1_error)

  /* "flitter/language/vm.pyx":993
 *         return self
 * 
 *     cpdef Program floordiv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.FloorDiv))
 *         return self
 */
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_floordiv, 993, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(0, 993, __pyx_L1_error)

  /* "flitter/language/vm.pyx":997
 *         return self
 * 
 *     cpdef Program mod(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mod))
 *         return self
 */
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_mod, 997, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 997, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1001
 *         return self
 * 
 *     cpdef Program pow(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pow))
 *         return self
 */
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_pow, 1001, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 1001, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1005
 *         return self
 * 
 *     cpdef Program eq(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Eq))
 *         return self
 */
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_eq, 1005, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(0, 1005, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1009
 *         return self
 * 
 *     cpdef Program ne(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ne))
 *         return self
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_ne, 1009, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 1009, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1013
 *         return self
 * 
 *     cpdef Program gt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Gt))
 *         return self
 */
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_gt, 1013, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 1013, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1017
 *         return self
 * 
 *     cpdef Program lt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lt))
 *         return self
 */
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_lt, 1017, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 1017, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1021
 *         return self
 * 
 *     cpdef Program ge(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ge))
 *         return self
 */
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_ge, 1021, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 1021, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1025
 *         return self
 * 
 *     cpdef Program le(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Le))
 *         return self
 */
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_le, 1025, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(0, 1025, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1029
 *         return self
 * 
 *     cpdef Program xor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Xor))
 *         return self
 */
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_xor, 1029, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(0, 1029, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1033
 *         return self
 * 
 *     cpdef Program slice(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Slice))
 *         return self
 */
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_slice, 1033, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(0, 1033, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1037
 *         return self
 * 
 *     cpdef Program slice_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         if value.length == 1 and value.numbers != NULL:
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 */
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_slice_literal, 1037, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 1037, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1044
 *         return self
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))
 *         return self
 */
  __pyx_tuple__129 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_count, __pyx_n_s_names); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_call, 1044, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_tuple__131 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "flitter/language/vm.pyx":1048
 *         return self
 * 
 *     cpdef Program call_fast(self, function, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))
 *         return self
 */
  __pyx_tuple__132 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_function, __pyx_n_s_count); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_call_fast, 1048, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(0, 1048, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1052
 *         return self
 * 
 *     cpdef Program tag(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))
 *         return self
 */
  __pyx_tuple__134 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_tag, 1052, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(0, 1052, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1056
 *         return self
 * 
 *     cpdef Program attributes(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))
 *         return self
 */
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_attributes, 1056, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 1056, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1060
 *         return self
 * 
 *     cpdef Program append(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Append, count))
 *         return self
 */
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_append, 1060, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(0, 1060, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1064
 *         return self
 * 
 *     cpdef Program compose(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))
 *         return self
 */
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_compose, 1064, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 1064, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1068
 *         return self
 * 
 *     cpdef Program begin_for(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))
 *         return self
 */
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_begin_for, 1068, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(0, 1068, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1072
 *         return self
 * 
 *     cpdef Program next(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Next, label))
 *         return self
 */
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_next, 1072, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 1072, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1076
 *         return self
 * 
 *     cpdef Program end_for(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.EndFor))
 *         return self
 */
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_end_for, 1076, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(0, 1076, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1080
 *         return self
 * 
 *     cpdef Program export(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Export, name))
 *         return self
 */
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_export, 1080, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 1080, __pyx_L1_error)

  /* "flitter/language/vm.pyx":1084
 *         return self
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))
 *         return self
 */
  __pyx_tuple__143 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_label, __pyx_n_s_name, __pyx_n_s_parameters, __pyx_n_s_ncaptures); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_func, 1084, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __pyx_tuple__145 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "flitter/language/vm.pyx":1088
 *         return self
 * 
 *     cpdef Program exit(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Exit))
 *         return self
 */
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_vm_pyx, __pyx_n_s_exit, 1088, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 1088, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Instruction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__149 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Instruction, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionVector, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionNode, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionStr, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionTuple, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionInt, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionIntTup, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__157 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionLabel, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__157)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionJump, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__159 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionFunc, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__159)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InstructionObject, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__161 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Function, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__161)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LoopSource, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_setdefault.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_setdefault.method_name = &__pyx_n_s_setdefault;
  __pyx_umethod_PyDict_Type_update.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_update.method_name = &__pyx_n_s_update;
  __pyx_umethod_PyList_Type_clear.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyList_Type_clear.method_name = &__pyx_n_s_clear;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyList_Type_pop.method_name = &__pyx_n_s_pop;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10904070 = PyInt_FromLong(10904070L); if (unlikely(!__pyx_int_10904070)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_19407059 = PyInt_FromLong(19407059L); if (unlikely(!__pyx_int_19407059)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_29217351 = PyInt_FromLong(29217351L); if (unlikely(!__pyx_int_29217351)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_45504841 = PyInt_FromLong(45504841L); if (unlikely(!__pyx_int_45504841)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_59839126 = PyInt_FromLong(59839126L); if (unlikely(!__pyx_int_59839126)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_61928532 = PyInt_FromLong(61928532L); if (unlikely(!__pyx_int_61928532)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_63712393 = PyInt_FromLong(63712393L); if (unlikely(!__pyx_int_63712393)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_71438266 = PyInt_FromLong(71438266L); if (unlikely(!__pyx_int_71438266)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_72616288 = PyInt_FromLong(72616288L); if (unlikely(!__pyx_int_72616288)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_73689043 = PyInt_FromLong(73689043L); if (unlikely(!__pyx_int_73689043)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_90787080 = PyInt_FromLong(90787080L); if (unlikely(!__pyx_int_90787080)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_107792009 = PyInt_FromLong(107792009L); if (unlikely(!__pyx_int_107792009)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_109158536 = PyInt_FromLong(109158536L); if (unlikely(!__pyx_int_109158536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_138499235 = PyInt_FromLong(138499235L); if (unlikely(!__pyx_int_138499235)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_163945588 = PyInt_FromLong(163945588L); if (unlikely(!__pyx_int_163945588)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_176342959 = PyInt_FromLong(176342959L); if (unlikely(!__pyx_int_176342959)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184506718 = PyInt_FromLong(184506718L); if (unlikely(!__pyx_int_184506718)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184517443 = PyInt_FromLong(184517443L); if (unlikely(!__pyx_int_184517443)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_186494852 = PyInt_FromLong(186494852L); if (unlikely(!__pyx_int_186494852)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_200977137 = PyInt_FromLong(200977137L); if (unlikely(!__pyx_int_200977137)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202585721 = PyInt_FromLong(202585721L); if (unlikely(!__pyx_int_202585721)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_210616022 = PyInt_FromLong(210616022L); if (unlikely(!__pyx_int_210616022)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_229249173 = PyInt_FromLong(229249173L); if (unlikely(!__pyx_int_229249173)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_231591000 = PyInt_FromLong(231591000L); if (unlikely(!__pyx_int_231591000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_236313263 = PyInt_FromLong(236313263L); if (unlikely(!__pyx_int_236313263)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_242200674 = PyInt_FromLong(242200674L); if (unlikely(!__pyx_int_242200674)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_259995473 = PyInt_FromLong(259995473L); if (unlikely(!__pyx_int_259995473)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_7flitter_8language_2vm_static_builtins = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_8language_2vm_dynamic_builtins = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_8language_2vm_builtins = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_8language_2vm_all_builtins = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_8language_2vm_OpCodeNames = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_static_builtins, (void *)&__pyx_v_7flitter_8language_2vm_static_builtins, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_dynamic_builtins, (void *)&__pyx_v_7flitter_8language_2vm_dynamic_builtins, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_builtins, (void *)&__pyx_v_7flitter_8language_2vm_builtins, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_7flitter_8language_2vm_VectorStack = &__pyx_vtable_7flitter_8language_2vm_VectorStack;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.copy = (struct __pyx_obj_7flitter_8language_2vm_VectorStack *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_copy;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.drop = (void (*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop *__pyx_optional_args))__pyx_f_7flitter_8language_2vm_11VectorStack_drop;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.push = (void (*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_push;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.pop = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_pop;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.pop_tuple = (PyObject *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_pop_tuple;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.pop_list = (PyObject *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_pop_list;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.pop_dict = (PyObject *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_pop_dict;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.pop_composed = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_pop_composed;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.peek = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_peek;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.peek_at = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_peek_at;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.poke = (void (*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_poke;
  __pyx_vtable_7flitter_8language_2vm_VectorStack.poke_at = (void (*)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_11VectorStack_poke_at;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm_VectorStack = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_VectorStack_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm_VectorStack)) __PYX_ERR(0, 246, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_VectorStack_spec, __pyx_ptype_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_VectorStack = &__pyx_type_7flitter_8language_2vm_VectorStack;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_VectorStack->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_VectorStack->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_VectorStack->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_VectorStack->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_vtabptr_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_VectorStack, (PyObject *) __pyx_ptype_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_VectorStack) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm_Instruction = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_Instruction_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm_Instruction)) __PYX_ERR(0, 130, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_Instruction_spec, __pyx_ptype_7flitter_8language_2vm_Instruction) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_Instruction = &__pyx_type_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_Instruction) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_Instruction->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_Instruction->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_Instruction->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_Instruction->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Instruction, (PyObject *) __pyx_ptype_7flitter_8language_2vm_Instruction) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_Instruction) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionVector = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionVector_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionVector)) __PYX_ERR(0, 138, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionVector_spec, __pyx_ptype_7flitter_8language_2vm_InstructionVector) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionVector = &__pyx_type_7flitter_8language_2vm_InstructionVector;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionVector->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionVector) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionVector->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionVector->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionVector->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionVector->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionVector, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionVector) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionVector) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionNode = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionNode_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionNode)) __PYX_ERR(0, 147, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionNode_spec, __pyx_ptype_7flitter_8language_2vm_InstructionNode) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionNode = &__pyx_type_7flitter_8language_2vm_InstructionNode;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionNode->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionNode) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionNode->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionNode->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionNode->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionNode->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionNode, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionNode) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionNode) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionStr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionStr_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionStr)) __PYX_ERR(0, 156, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionStr_spec, __pyx_ptype_7flitter_8language_2vm_InstructionStr) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionStr = &__pyx_type_7flitter_8language_2vm_InstructionStr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionStr->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionStr) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionStr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionStr->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionStr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionStr->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionStr, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionStr) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionStr) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionTuple = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionTuple_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionTuple)) __PYX_ERR(0, 165, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionTuple_spec, __pyx_ptype_7flitter_8language_2vm_InstructionTuple) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionTuple = &__pyx_type_7flitter_8language_2vm_InstructionTuple;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionTuple->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionTuple) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionTuple->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionTuple->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionTuple->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionTuple->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionTuple, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionTuple) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionTuple) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionInt = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionInt_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionInt)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionInt_spec, __pyx_ptype_7flitter_8language_2vm_InstructionInt) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionInt = &__pyx_type_7flitter_8language_2vm_InstructionInt;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionInt->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionInt) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionInt->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionInt->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionInt->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionInt->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionInt, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionInt) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionInt) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionIntTuple_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple)) __PYX_ERR(0, 183, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionIntTuple_spec, __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple = &__pyx_type_7flitter_8language_2vm_InstructionIntTuple;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionIntTuple, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionLabel = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionLabel_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionLabel)) __PYX_ERR(0, 193, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionLabel_spec, __pyx_ptype_7flitter_8language_2vm_InstructionLabel) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionLabel = &__pyx_type_7flitter_8language_2vm_InstructionLabel;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionLabel->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionLabel) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionLabel->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionLabel->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionLabel->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionLabel->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionLabel, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionLabel) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionLabel) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionJump = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionJump_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionJump)) __PYX_ERR(0, 202, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionJump_spec, __pyx_ptype_7flitter_8language_2vm_InstructionJump) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionJump = &__pyx_type_7flitter_8language_2vm_InstructionJump;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionJump) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionJump->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionJump, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionJump) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionJump) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionFunc = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionFunc_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionFunc)) __PYX_ERR(0, 213, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionFunc_spec, __pyx_ptype_7flitter_8language_2vm_InstructionFunc) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionFunc = &__pyx_type_7flitter_8language_2vm_InstructionFunc;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionFunc->tp_base = __pyx_ptype_7flitter_8language_2vm_InstructionJump;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionFunc) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionFunc->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionFunc->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionFunc->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionFunc->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionFunc, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionFunc) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionFunc) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_InstructionObjectInt_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt)) __PYX_ERR(0, 224, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_InstructionObjectInt_spec, __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt = &__pyx_type_7flitter_8language_2vm_InstructionObjectInt;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt->tp_base = __pyx_ptype_7flitter_8language_2vm_Instruction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InstructionObjectInt, (PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_2vm_Program = &__pyx_vtable_7flitter_8language_2vm_Program;
  __pyx_vtable_7flitter_8language_2vm_Program.last_instruction = (struct __pyx_obj_7flitter_8language_2vm_Instruction *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_2vm_7Program_last_instruction;
  __pyx_vtable_7flitter_8language_2vm_Program.pop_instruction = (struct __pyx_obj_7flitter_8language_2vm_Instruction *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_2vm_7Program_pop_instruction;
  __pyx_vtable_7flitter_8language_2vm_Program.push_instruction = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_8language_2vm_Instruction *))__pyx_f_7flitter_8language_2vm_7Program_push_instruction;
  __pyx_vtable_7flitter_8language_2vm_Program.link = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_link;
  __pyx_vtable_7flitter_8language_2vm_Program.use_simplifier = (void (*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_use_simplifier;
  __pyx_vtable_7flitter_8language_2vm_Program.new_label = (int64_t (*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_new_label;
  __pyx_vtable_7flitter_8language_2vm_Program.dup = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_dup;
  __pyx_vtable_7flitter_8language_2vm_Program.drop = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop *__pyx_optional_args))__pyx_f_7flitter_8language_2vm_7Program_drop;
  __pyx_vtable_7flitter_8language_2vm_Program.label = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_label;
  __pyx_vtable_7flitter_8language_2vm_Program.jump = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_jump;
  __pyx_vtable_7flitter_8language_2vm_Program.branch_true = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_branch_true;
  __pyx_vtable_7flitter_8language_2vm_Program.branch_false = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_branch_false;
  __pyx_vtable_7flitter_8language_2vm_Program.import_ = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_import_;
  __pyx_vtable_7flitter_8language_2vm_Program.literal = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_literal;
  __pyx_vtable_7flitter_8language_2vm_Program.local_push = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_local_push;
  __pyx_vtable_7flitter_8language_2vm_Program.local_load = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_local_load;
  __pyx_vtable_7flitter_8language_2vm_Program.local_drop = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_local_drop;
  __pyx_vtable_7flitter_8language_2vm_Program.lookup = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_lookup;
  __pyx_vtable_7flitter_8language_2vm_Program.lookup_literal = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_lookup_literal;
  __pyx_vtable_7flitter_8language_2vm_Program.range = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_range;
  __pyx_vtable_7flitter_8language_2vm_Program.neg = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_neg;
  __pyx_vtable_7flitter_8language_2vm_Program.pos = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_pos;
  __pyx_vtable_7flitter_8language_2vm_Program.ceil = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_ceil;
  __pyx_vtable_7flitter_8language_2vm_Program.floor = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_floor;
  __pyx_vtable_7flitter_8language_2vm_Program.fract = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_fract;
  __pyx_vtable_7flitter_8language_2vm_Program.not_ = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_not_;
  __pyx_vtable_7flitter_8language_2vm_Program.add = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_add;
  __pyx_vtable_7flitter_8language_2vm_Program.sub = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_sub;
  __pyx_vtable_7flitter_8language_2vm_Program.mul = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_mul;
  __pyx_vtable_7flitter_8language_2vm_Program.mul_add = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_mul_add;
  __pyx_vtable_7flitter_8language_2vm_Program.truediv = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_truediv;
  __pyx_vtable_7flitter_8language_2vm_Program.floordiv = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_floordiv;
  __pyx_vtable_7flitter_8language_2vm_Program.mod = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_mod;
  __pyx_vtable_7flitter_8language_2vm_Program.pow = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_pow;
  __pyx_vtable_7flitter_8language_2vm_Program.eq = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_eq;
  __pyx_vtable_7flitter_8language_2vm_Program.ne = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_ne;
  __pyx_vtable_7flitter_8language_2vm_Program.gt = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_gt;
  __pyx_vtable_7flitter_8language_2vm_Program.lt = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_lt;
  __pyx_vtable_7flitter_8language_2vm_Program.ge = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_ge;
  __pyx_vtable_7flitter_8language_2vm_Program.le = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_le;
  __pyx_vtable_7flitter_8language_2vm_Program.xor = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_xor;
  __pyx_vtable_7flitter_8language_2vm_Program.slice = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_slice;
  __pyx_vtable_7flitter_8language_2vm_Program.slice_literal = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_slice_literal;
  __pyx_vtable_7flitter_8language_2vm_Program.call = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_call *__pyx_optional_args))__pyx_f_7flitter_8language_2vm_7Program_call;
  __pyx_vtable_7flitter_8language_2vm_Program.call_fast = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_call_fast;
  __pyx_vtable_7flitter_8language_2vm_Program.tag = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_tag;
  __pyx_vtable_7flitter_8language_2vm_Program.attributes = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_attributes;
  __pyx_vtable_7flitter_8language_2vm_Program.append = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_append *__pyx_optional_args))__pyx_f_7flitter_8language_2vm_7Program_append;
  __pyx_vtable_7flitter_8language_2vm_Program.compose = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_compose;
  __pyx_vtable_7flitter_8language_2vm_Program.begin_for = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_begin_for;
  __pyx_vtable_7flitter_8language_2vm_Program.next = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_next;
  __pyx_vtable_7flitter_8language_2vm_Program.end_for = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_end_for;
  __pyx_vtable_7flitter_8language_2vm_Program.export = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_export;
  __pyx_vtable_7flitter_8language_2vm_Program.func = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_func *__pyx_optional_args))__pyx_f_7flitter_8language_2vm_7Program_func;
  __pyx_vtable_7flitter_8language_2vm_Program.exit = (struct __pyx_obj_7flitter_8language_2vm_Program *(*)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch))__pyx_f_7flitter_8language_2vm_7Program_exit;
  __pyx_vtable_7flitter_8language_2vm_Program._execute = (void (*)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Context *, int64_t, int))__pyx_f_7flitter_8language_2vm_7Program__execute;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm_Program = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_Program_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm_Program)) __PYX_ERR(0, 765, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_Program_spec, __pyx_ptype_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_Program = &__pyx_type_7flitter_8language_2vm_Program;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_Program->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_Program->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_Program->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_Program->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_2vm_Program, __pyx_vtabptr_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Program, (PyObject *) __pyx_ptype_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_Program) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm_Function = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_Function_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm_Function)) __PYX_ERR(0, 480, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_Function_spec, __pyx_ptype_7flitter_8language_2vm_Function) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_Function = &__pyx_type_7flitter_8language_2vm_Function;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_Function) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_Function->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_Function->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_Function->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_Function->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Function, (PyObject *) __pyx_ptype_7flitter_8language_2vm_Function) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_Function) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm_LoopSource = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm_LoopSource_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm_LoopSource)) __PYX_ERR(0, 535, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm_LoopSource_spec, __pyx_ptype_7flitter_8language_2vm_LoopSource) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm_LoopSource = &__pyx_type_7flitter_8language_2vm_LoopSource;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm_LoopSource) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm_LoopSource->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm_LoopSource->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm_LoopSource->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm_LoopSource->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LoopSource, (PyObject *) __pyx_ptype_7flitter_8language_2vm_LoopSource) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_2vm_LoopSource) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr)) __PYX_ERR(0, 779, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr_spec, __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 779, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr = &__pyx_type_7flitter_8language_2vm___pyx_scope_struct__genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 779, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct__genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute)) __PYX_ERR(0, 1092, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute_spec, __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute) < 0) __PYX_ERR(0, 1092, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute = &__pyx_type_7flitter_8language_2vm___pyx_scope_struct_1__execute;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute) < 0) __PYX_ERR(0, 1092, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_1__execute->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr)) __PYX_ERR(0, 1454, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr_spec, __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 1454, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr = &__pyx_type_7flitter_8language_2vm___pyx_scope_struct_2_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 1454, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr->tp_dictoffset && __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_2vm___pyx_scope_struct_2_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_5model_Vector = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Vector", sizeof(struct __pyx_obj_7flitter_5model_Vector), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Vector),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Vector) __PYX_ERR(7, 35, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Vector = (struct __pyx_vtabstruct_7flitter_5model_Vector*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Vector); if (unlikely(!__pyx_vtabptr_7flitter_5model_Vector)) __PYX_ERR(7, 35, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix33 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix33", sizeof(struct __pyx_obj_7flitter_5model_Matrix33), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix33),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix33) __PYX_ERR(7, 107, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix33 = (struct __pyx_vtabstruct_7flitter_5model_Matrix33*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix33); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix33)) __PYX_ERR(7, 107, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix44 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix44", sizeof(struct __pyx_obj_7flitter_5model_Matrix44), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix44),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix44) __PYX_ERR(7, 128, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix44 = (struct __pyx_vtabstruct_7flitter_5model_Matrix44*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix44); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix44)) __PYX_ERR(7, 128, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Quaternion = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Quaternion", sizeof(struct __pyx_obj_7flitter_5model_Quaternion), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Quaternion),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Quaternion) __PYX_ERR(7, 178, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Quaternion = (struct __pyx_vtabstruct_7flitter_5model_Quaternion*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Quaternion); if (unlikely(!__pyx_vtabptr_7flitter_5model_Quaternion)) __PYX_ERR(7, 178, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Node = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Node", sizeof(struct __pyx_obj_7flitter_5model_Node), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Node),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Node) __PYX_ERR(7, 198, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Node = (struct __pyx_vtabstruct_7flitter_5model_Node*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Node); if (unlikely(!__pyx_vtabptr_7flitter_5model_Node)) __PYX_ERR(7, 198, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_StateDict = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "StateDict", sizeof(struct __pyx_obj_7flitter_5model_StateDict), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_StateDict),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_StateDict) __PYX_ERR(7, 222, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_StateDict = (struct __pyx_vtabstruct_7flitter_5model_StateDict*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_StateDict); if (unlikely(!__pyx_vtabptr_7flitter_5model_StateDict)) __PYX_ERR(7, 222, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Context = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Context", sizeof(struct __pyx_obj_7flitter_5model_Context), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Context),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Context) __PYX_ERR(7, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "HASH_START", (void **)&__pyx_vp_7flitter_5model_HASH_START, "uint64_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "null_", (void **)&__pyx_vp_7flitter_5model_null_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "true_", (void **)&__pyx_vp_7flitter_5model_true_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "false_", (void **)&__pyx_vp_7flitter_5model_false_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "minusone_", (void **)&__pyx_vp_7flitter_5model_minusone_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "inf_", (void **)&__pyx_vp_7flitter_5model_inf_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "nan_", (void **)&__pyx_vp_7flitter_5model_nan_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_vm(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_vm},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "vm",
      __pyx_k_Flitter_language_stack_based_vi, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initvm(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initvm(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_vm(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_vm(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_vm(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'vm' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("vm", __pyx_methods, __pyx_k_Flitter_language_stack_based_vi, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "vm" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_vm(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flitter__language__vm) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flitter.language.vm")) {
      if (unlikely((PyDict_SetItemString(modules, "flitter.language.vm", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely((__Pyx_modinit_variable_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "flitter/language/vm.pyx":7
 * """
 * 
 * from loguru import logger             # <<<<<<<<<<<<<<
 * 
 * from .. import name_patch
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(0, 7, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_loguru, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/vm.pyx":9
 * from loguru import logger
 * 
 * from .. import name_patch             # <<<<<<<<<<<<<<
 * from ..cache import SharedCache
 * from .functions import STATIC_FUNCTIONS, DYNAMIC_FUNCTIONS
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_name_patch);
  __Pyx_GIVEREF(__pyx_n_s_name_patch);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_name_patch)) __PYX_ERR(0, 9, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s__21, __pyx_t_3, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_name_patch, __pyx_t_3) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":10
 * 
 * from .. import name_patch
 * from ..cache import SharedCache             # <<<<<<<<<<<<<<
 * from .functions import STATIC_FUNCTIONS, DYNAMIC_FUNCTIONS
 * from ..model cimport StateDict, null_, true_, false_, inf_, nan_
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_SharedCache);
  __Pyx_GIVEREF(__pyx_n_s_SharedCache);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_SharedCache)) __PYX_ERR(0, 10, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_cache, __pyx_t_2, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_SharedCache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SharedCache, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/vm.pyx":11
 * from .. import name_patch
 * from ..cache import SharedCache
 * from .functions import STATIC_FUNCTIONS, DYNAMIC_FUNCTIONS             # <<<<<<<<<<<<<<
 * from ..model cimport StateDict, null_, true_, false_, inf_, nan_
 * from ..timer cimport perf_counter
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_STATIC_FUNCTIONS);
  __Pyx_GIVEREF(__pyx_n_s_STATIC_FUNCTIONS);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_STATIC_FUNCTIONS)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_DYNAMIC_FUNCTIONS);
  __Pyx_GIVEREF(__pyx_n_s_DYNAMIC_FUNCTIONS);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DYNAMIC_FUNCTIONS)) __PYX_ERR(0, 11, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_functions, __pyx_t_3, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_STATIC_FUNCTIONS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STATIC_FUNCTIONS, __pyx_t_3) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DYNAMIC_FUNCTIONS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DYNAMIC_FUNCTIONS, __pyx_t_3) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/vm.pyx":14
 * from ..model cimport StateDict, null_, true_, false_, inf_, nan_
 * from ..timer cimport perf_counter
 * from .noise import NOISE_FUNCTIONS             # <<<<<<<<<<<<<<
 * 
 * from libc.math cimport floor as c_floor
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_NOISE_FUNCTIONS);
  __Pyx_GIVEREF(__pyx_n_s_NOISE_FUNCTIONS);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_NOISE_FUNCTIONS)) __PYX_ERR(0, 14, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_noise, __pyx_t_2, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_NOISE_FUNCTIONS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NOISE_FUNCTIONS, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/vm.pyx":32
 * 
 * 
 * logger = name_patch(logger, __name__)             # <<<<<<<<<<<<<<
 * 
 * cdef const char* ContextFunc = "context_func\0"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4)) __PYX_ERR(0, 32, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_4) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":34
 * logger = name_patch(logger, __name__)
 * 
 * cdef const char* ContextFunc = "context_func\0"             # <<<<<<<<<<<<<<
 * 
 * cdef int64_t DEFAULT_STACK_SIZE = 32
 */
  __pyx_v_7flitter_8language_2vm_ContextFunc = ((char const *)"context_func\000");

  /* "flitter/language/vm.pyx":36
 * cdef const char* ContextFunc = "context_func\0"
 * 
 * cdef int64_t DEFAULT_STACK_SIZE = 32             # <<<<<<<<<<<<<<
 * cdef int64_t MAX_STACK_SIZE = 1 << 20
 * cdef int64_t MAX_CALL_DEPTH = 500
 */
  __pyx_v_7flitter_8language_2vm_DEFAULT_STACK_SIZE = 32;

  /* "flitter/language/vm.pyx":37
 * 
 * cdef int64_t DEFAULT_STACK_SIZE = 32
 * cdef int64_t MAX_STACK_SIZE = 1 << 20             # <<<<<<<<<<<<<<
 * cdef int64_t MAX_CALL_DEPTH = 500
 * 
 */
  __pyx_v_7flitter_8language_2vm_MAX_STACK_SIZE = 0x100000;

  /* "flitter/language/vm.pyx":38
 * cdef int64_t DEFAULT_STACK_SIZE = 32
 * cdef int64_t MAX_STACK_SIZE = 1 << 20
 * cdef int64_t MAX_CALL_DEPTH = 500             # <<<<<<<<<<<<<<
 * 
 * cdef dict dynamic_builtins = DYNAMIC_FUNCTIONS
 */
  __pyx_v_7flitter_8language_2vm_MAX_CALL_DEPTH = 0x1F4;

  /* "flitter/language/vm.pyx":40
 * cdef int64_t MAX_CALL_DEPTH = 500
 * 
 * cdef dict dynamic_builtins = DYNAMIC_FUNCTIONS             # <<<<<<<<<<<<<<
 * cdef dict static_builtins = {
 *     'true': true_,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DYNAMIC_FUNCTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyDict_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_4))) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_7flitter_8language_2vm_dynamic_builtins);
  __Pyx_DECREF_SET(__pyx_v_7flitter_8language_2vm_dynamic_builtins, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":42
 * cdef dict dynamic_builtins = DYNAMIC_FUNCTIONS
 * cdef dict static_builtins = {
 *     'true': true_,             # <<<<<<<<<<<<<<
 *     'false': false_,
 *     'null': null_,
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_true, ((PyObject *)__pyx_v_7flitter_5model_true_)) < 0) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "flitter/language/vm.pyx":43
 * cdef dict static_builtins = {
 *     'true': true_,
 *     'false': false_,             # <<<<<<<<<<<<<<
 *     'null': null_,
 *     'inf': inf_,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_false, ((PyObject *)__pyx_v_7flitter_5model_false_)) < 0) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "flitter/language/vm.pyx":44
 *     'true': true_,
 *     'false': false_,
 *     'null': null_,             # <<<<<<<<<<<<<<
 *     'inf': inf_,
 *     'nan': nan_,
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_null, ((PyObject *)__pyx_v_7flitter_5model_null_)) < 0) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "flitter/language/vm.pyx":45
 *     'false': false_,
 *     'null': null_,
 *     'inf': inf_,             # <<<<<<<<<<<<<<
 *     'nan': nan_,
 * }
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_inf, ((PyObject *)__pyx_v_7flitter_5model_inf_)) < 0) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "flitter/language/vm.pyx":46
 *     'null': null_,
 *     'inf': inf_,
 *     'nan': nan_,             # <<<<<<<<<<<<<<
 * }
 * static_builtins.update(STATIC_FUNCTIONS)
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_nan, ((PyObject *)__pyx_v_7flitter_5model_nan_)) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_7flitter_8language_2vm_static_builtins);
  __Pyx_DECREF_SET(__pyx_v_7flitter_8language_2vm_static_builtins, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":48
 *     'nan': nan_,
 * }
 * static_builtins.update(STATIC_FUNCTIONS)             # <<<<<<<<<<<<<<
 * static_builtins.update(NOISE_FUNCTIONS)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_STATIC_FUNCTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_7flitter_8language_2vm_static_builtins, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":49
 * }
 * static_builtins.update(STATIC_FUNCTIONS)
 * static_builtins.update(NOISE_FUNCTIONS)             # <<<<<<<<<<<<<<
 * 
 * cdef dict all_builtins = {}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_NOISE_FUNCTIONS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_7flitter_8language_2vm_static_builtins, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":51
 * static_builtins.update(NOISE_FUNCTIONS)
 * 
 * cdef dict all_builtins = {}             # <<<<<<<<<<<<<<
 * all_builtins.update(dynamic_builtins)
 * all_builtins.update(static_builtins)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_7flitter_8language_2vm_all_builtins);
  __Pyx_DECREF_SET(__pyx_v_7flitter_8language_2vm_all_builtins, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":52
 * 
 * cdef dict all_builtins = {}
 * all_builtins.update(dynamic_builtins)             # <<<<<<<<<<<<<<
 * all_builtins.update(static_builtins)
 * 
 */
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_7flitter_8language_2vm_all_builtins, __pyx_v_7flitter_8language_2vm_dynamic_builtins); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":53
 * cdef dict all_builtins = {}
 * all_builtins.update(dynamic_builtins)
 * all_builtins.update(static_builtins)             # <<<<<<<<<<<<<<
 * 
 * cdef int64_t* StatsCount = NULL
 */
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_7flitter_8language_2vm_all_builtins, __pyx_v_7flitter_8language_2vm_static_builtins); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":55
 * all_builtins.update(static_builtins)
 * 
 * cdef int64_t* StatsCount = NULL             # <<<<<<<<<<<<<<
 * cdef double* StatsDuration = NULL
 * cdef double CallOutDuration = 0
 */
  __pyx_v_7flitter_8language_2vm_StatsCount = NULL;

  /* "flitter/language/vm.pyx":56
 * 
 * cdef int64_t* StatsCount = NULL
 * cdef double* StatsDuration = NULL             # <<<<<<<<<<<<<<
 * cdef double CallOutDuration = 0
 * cdef int64_t CallOutCount = 0
 */
  __pyx_v_7flitter_8language_2vm_StatsDuration = NULL;

  /* "flitter/language/vm.pyx":57
 * cdef int64_t* StatsCount = NULL
 * cdef double* StatsDuration = NULL
 * cdef double CallOutDuration = 0             # <<<<<<<<<<<<<<
 * cdef int64_t CallOutCount = 0
 * 
 */
  __pyx_v_7flitter_8language_2vm_CallOutDuration = 0.0;

  /* "flitter/language/vm.pyx":58
 * cdef double* StatsDuration = NULL
 * cdef double CallOutDuration = 0
 * cdef int64_t CallOutCount = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_7flitter_8language_2vm_CallOutCount = 0;

  /* "flitter/language/vm.pyx":62
 * 
 * cdef dict OpCodeNames = {
 *     OpCode.Add: 'Add',             # <<<<<<<<<<<<<<
 *     OpCode.Append: 'Append',
 *     OpCode.Attributes: 'Attributes',
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(52); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Add); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Add) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":63
 * cdef dict OpCodeNames = {
 *     OpCode.Add: 'Add',
 *     OpCode.Append: 'Append',             # <<<<<<<<<<<<<<
 *     OpCode.Attributes: 'Attributes',
 *     OpCode.BeginFor: 'BeginFor',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Append); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Append) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":64
 *     OpCode.Add: 'Add',
 *     OpCode.Append: 'Append',
 *     OpCode.Attributes: 'Attributes',             # <<<<<<<<<<<<<<
 *     OpCode.BeginFor: 'BeginFor',
 *     OpCode.BranchFalse: 'BranchFalse',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Attributes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Attributes) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":65
 *     OpCode.Append: 'Append',
 *     OpCode.Attributes: 'Attributes',
 *     OpCode.BeginFor: 'BeginFor',             # <<<<<<<<<<<<<<
 *     OpCode.BranchFalse: 'BranchFalse',
 *     OpCode.BranchTrue: 'BranchTrue',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BeginFor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_BeginFor) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":66
 *     OpCode.Attributes: 'Attributes',
 *     OpCode.BeginFor: 'BeginFor',
 *     OpCode.BranchFalse: 'BranchFalse',             # <<<<<<<<<<<<<<
 *     OpCode.BranchTrue: 'BranchTrue',
 *     OpCode.Call: 'Call',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BranchFalse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_BranchFalse) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":67
 *     OpCode.BeginFor: 'BeginFor',
 *     OpCode.BranchFalse: 'BranchFalse',
 *     OpCode.BranchTrue: 'BranchTrue',             # <<<<<<<<<<<<<<
 *     OpCode.Call: 'Call',
 *     OpCode.CallFast: 'CallFast',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_BranchTrue); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_BranchTrue) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":68
 *     OpCode.BranchFalse: 'BranchFalse',
 *     OpCode.BranchTrue: 'BranchTrue',
 *     OpCode.Call: 'Call',             # <<<<<<<<<<<<<<
 *     OpCode.CallFast: 'CallFast',
 *     OpCode.Ceil: 'Ceil',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Call); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Call) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":69
 *     OpCode.BranchTrue: 'BranchTrue',
 *     OpCode.Call: 'Call',
 *     OpCode.CallFast: 'CallFast',             # <<<<<<<<<<<<<<
 *     OpCode.Ceil: 'Ceil',
 *     OpCode.Compose: 'Compose',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_CallFast); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_CallFast) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":70
 *     OpCode.Call: 'Call',
 *     OpCode.CallFast: 'CallFast',
 *     OpCode.Ceil: 'Ceil',             # <<<<<<<<<<<<<<
 *     OpCode.Compose: 'Compose',
 *     OpCode.Drop: 'Drop',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ceil); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Ceil) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":71
 *     OpCode.CallFast: 'CallFast',
 *     OpCode.Ceil: 'Ceil',
 *     OpCode.Compose: 'Compose',             # <<<<<<<<<<<<<<
 *     OpCode.Drop: 'Drop',
 *     OpCode.Dup: 'Dup',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Compose); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Compose) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":72
 *     OpCode.Ceil: 'Ceil',
 *     OpCode.Compose: 'Compose',
 *     OpCode.Drop: 'Drop',             # <<<<<<<<<<<<<<
 *     OpCode.Dup: 'Dup',
 *     OpCode.EndFor: 'EndFor',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Drop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Drop) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":73
 *     OpCode.Compose: 'Compose',
 *     OpCode.Drop: 'Drop',
 *     OpCode.Dup: 'Dup',             # <<<<<<<<<<<<<<
 *     OpCode.EndFor: 'EndFor',
 *     OpCode.Eq: 'Eq',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Dup); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Dup) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":74
 *     OpCode.Drop: 'Drop',
 *     OpCode.Dup: 'Dup',
 *     OpCode.EndFor: 'EndFor',             # <<<<<<<<<<<<<<
 *     OpCode.Eq: 'Eq',
 *     OpCode.Floor: 'Floor',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_EndFor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_EndFor) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":75
 *     OpCode.Dup: 'Dup',
 *     OpCode.EndFor: 'EndFor',
 *     OpCode.Eq: 'Eq',             # <<<<<<<<<<<<<<
 *     OpCode.Floor: 'Floor',
 *     OpCode.FloorDiv: 'FloorDiv',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Eq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Eq) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":76
 *     OpCode.EndFor: 'EndFor',
 *     OpCode.Eq: 'Eq',
 *     OpCode.Floor: 'Floor',             # <<<<<<<<<<<<<<
 *     OpCode.FloorDiv: 'FloorDiv',
 *     OpCode.Fract: 'Fract',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Floor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Floor) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":77
 *     OpCode.Eq: 'Eq',
 *     OpCode.Floor: 'Floor',
 *     OpCode.FloorDiv: 'FloorDiv',             # <<<<<<<<<<<<<<
 *     OpCode.Fract: 'Fract',
 *     OpCode.Func: 'Func',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_FloorDiv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_FloorDiv) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":78
 *     OpCode.Floor: 'Floor',
 *     OpCode.FloorDiv: 'FloorDiv',
 *     OpCode.Fract: 'Fract',             # <<<<<<<<<<<<<<
 *     OpCode.Func: 'Func',
 *     OpCode.Exit: 'Exit',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Fract); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Fract) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":79
 *     OpCode.FloorDiv: 'FloorDiv',
 *     OpCode.Fract: 'Fract',
 *     OpCode.Func: 'Func',             # <<<<<<<<<<<<<<
 *     OpCode.Exit: 'Exit',
 *     OpCode.Ge: 'Ge',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Func); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Func) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":80
 *     OpCode.Fract: 'Fract',
 *     OpCode.Func: 'Func',
 *     OpCode.Exit: 'Exit',             # <<<<<<<<<<<<<<
 *     OpCode.Ge: 'Ge',
 *     OpCode.Gt: 'Gt',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Exit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Exit) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":81
 *     OpCode.Func: 'Func',
 *     OpCode.Exit: 'Exit',
 *     OpCode.Ge: 'Ge',             # <<<<<<<<<<<<<<
 *     OpCode.Gt: 'Gt',
 *     OpCode.Import: 'Import',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Ge) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":82
 *     OpCode.Exit: 'Exit',
 *     OpCode.Ge: 'Ge',
 *     OpCode.Gt: 'Gt',             # <<<<<<<<<<<<<<
 *     OpCode.Import: 'Import',
 *     OpCode.IndexLiteral: 'IndexLiteral',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Gt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Gt) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":83
 *     OpCode.Ge: 'Ge',
 *     OpCode.Gt: 'Gt',
 *     OpCode.Import: 'Import',             # <<<<<<<<<<<<<<
 *     OpCode.IndexLiteral: 'IndexLiteral',
 *     OpCode.Jump: 'Jump',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Import); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Import) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":84
 *     OpCode.Gt: 'Gt',
 *     OpCode.Import: 'Import',
 *     OpCode.IndexLiteral: 'IndexLiteral',             # <<<<<<<<<<<<<<
 *     OpCode.Jump: 'Jump',
 *     OpCode.Label: 'Label',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_IndexLiteral); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_IndexLiteral) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":85
 *     OpCode.Import: 'Import',
 *     OpCode.IndexLiteral: 'IndexLiteral',
 *     OpCode.Jump: 'Jump',             # <<<<<<<<<<<<<<
 *     OpCode.Label: 'Label',
 *     OpCode.Le: 'Le',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Jump); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Jump) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":86
 *     OpCode.IndexLiteral: 'IndexLiteral',
 *     OpCode.Jump: 'Jump',
 *     OpCode.Label: 'Label',             # <<<<<<<<<<<<<<
 *     OpCode.Le: 'Le',
 *     OpCode.Literal: 'Literal',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Label); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Label) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":87
 *     OpCode.Jump: 'Jump',
 *     OpCode.Label: 'Label',
 *     OpCode.Le: 'Le',             # <<<<<<<<<<<<<<
 *     OpCode.Literal: 'Literal',
 *     OpCode.LiteralNode: 'LiteralNode',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Le); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Le) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":88
 *     OpCode.Label: 'Label',
 *     OpCode.Le: 'Le',
 *     OpCode.Literal: 'Literal',             # <<<<<<<<<<<<<<
 *     OpCode.LiteralNode: 'LiteralNode',
 *     OpCode.LiteralNodes: 'LiteralNodes',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Literal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Literal) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":89
 *     OpCode.Le: 'Le',
 *     OpCode.Literal: 'Literal',
 *     OpCode.LiteralNode: 'LiteralNode',             # <<<<<<<<<<<<<<
 *     OpCode.LiteralNodes: 'LiteralNodes',
 *     OpCode.LocalDrop: 'LocalDrop',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LiteralNode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LiteralNode) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":90
 *     OpCode.Literal: 'Literal',
 *     OpCode.LiteralNode: 'LiteralNode',
 *     OpCode.LiteralNodes: 'LiteralNodes',             # <<<<<<<<<<<<<<
 *     OpCode.LocalDrop: 'LocalDrop',
 *     OpCode.LocalLoad: 'LocalLoad',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LiteralNodes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LiteralNodes) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":91
 *     OpCode.LiteralNode: 'LiteralNode',
 *     OpCode.LiteralNodes: 'LiteralNodes',
 *     OpCode.LocalDrop: 'LocalDrop',             # <<<<<<<<<<<<<<
 *     OpCode.LocalLoad: 'LocalLoad',
 *     OpCode.LocalPush: 'LocalPush',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalDrop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LocalDrop) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":92
 *     OpCode.LiteralNodes: 'LiteralNodes',
 *     OpCode.LocalDrop: 'LocalDrop',
 *     OpCode.LocalLoad: 'LocalLoad',             # <<<<<<<<<<<<<<
 *     OpCode.LocalPush: 'LocalPush',
 *     OpCode.Lookup: 'Lookup',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalLoad); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LocalLoad) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":93
 *     OpCode.LocalDrop: 'LocalDrop',
 *     OpCode.LocalLoad: 'LocalLoad',
 *     OpCode.LocalPush: 'LocalPush',             # <<<<<<<<<<<<<<
 *     OpCode.Lookup: 'Lookup',
 *     OpCode.LookupLiteral: 'LookupLiteral',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LocalPush); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LocalPush) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":94
 *     OpCode.LocalLoad: 'LocalLoad',
 *     OpCode.LocalPush: 'LocalPush',
 *     OpCode.Lookup: 'Lookup',             # <<<<<<<<<<<<<<
 *     OpCode.LookupLiteral: 'LookupLiteral',
 *     OpCode.Lt: 'Lt',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Lookup); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Lookup) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":95
 *     OpCode.LocalPush: 'LocalPush',
 *     OpCode.Lookup: 'Lookup',
 *     OpCode.LookupLiteral: 'LookupLiteral',             # <<<<<<<<<<<<<<
 *     OpCode.Lt: 'Lt',
 *     OpCode.Mod: 'Mod',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_LookupLiteral); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_LookupLiteral) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":96
 *     OpCode.Lookup: 'Lookup',
 *     OpCode.LookupLiteral: 'LookupLiteral',
 *     OpCode.Lt: 'Lt',             # <<<<<<<<<<<<<<
 *     OpCode.Mod: 'Mod',
 *     OpCode.Mul: 'Mul',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Lt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Lt) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":97
 *     OpCode.LookupLiteral: 'LookupLiteral',
 *     OpCode.Lt: 'Lt',
 *     OpCode.Mod: 'Mod',             # <<<<<<<<<<<<<<
 *     OpCode.Mul: 'Mul',
 *     OpCode.MulAdd: 'MulAdd',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Mod); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Mod) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":98
 *     OpCode.Lt: 'Lt',
 *     OpCode.Mod: 'Mod',
 *     OpCode.Mul: 'Mul',             # <<<<<<<<<<<<<<
 *     OpCode.MulAdd: 'MulAdd',
 *     OpCode.Ne: 'Ne',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Mul); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Mul) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":99
 *     OpCode.Mod: 'Mod',
 *     OpCode.Mul: 'Mul',
 *     OpCode.MulAdd: 'MulAdd',             # <<<<<<<<<<<<<<
 *     OpCode.Ne: 'Ne',
 *     OpCode.Neg: 'Neg',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_MulAdd); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_MulAdd) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":100
 *     OpCode.Mul: 'Mul',
 *     OpCode.MulAdd: 'MulAdd',
 *     OpCode.Ne: 'Ne',             # <<<<<<<<<<<<<<
 *     OpCode.Neg: 'Neg',
 *     OpCode.Next: 'Next',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Ne); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Ne) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":101
 *     OpCode.MulAdd: 'MulAdd',
 *     OpCode.Ne: 'Ne',
 *     OpCode.Neg: 'Neg',             # <<<<<<<<<<<<<<
 *     OpCode.Next: 'Next',
 *     OpCode.Not: 'Not',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Neg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Neg) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":102
 *     OpCode.Ne: 'Ne',
 *     OpCode.Neg: 'Neg',
 *     OpCode.Next: 'Next',             # <<<<<<<<<<<<<<
 *     OpCode.Not: 'Not',
 *     OpCode.Pos: 'Pos',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Next); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Next) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":103
 *     OpCode.Neg: 'Neg',
 *     OpCode.Next: 'Next',
 *     OpCode.Not: 'Not',             # <<<<<<<<<<<<<<
 *     OpCode.Pos: 'Pos',
 *     OpCode.Pow: 'Pow',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Not); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Not) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":104
 *     OpCode.Next: 'Next',
 *     OpCode.Not: 'Not',
 *     OpCode.Pos: 'Pos',             # <<<<<<<<<<<<<<
 *     OpCode.Pow: 'Pow',
 *     OpCode.Range: 'Range',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Pos) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":105
 *     OpCode.Not: 'Not',
 *     OpCode.Pos: 'Pos',
 *     OpCode.Pow: 'Pow',             # <<<<<<<<<<<<<<
 *     OpCode.Range: 'Range',
 *     OpCode.Slice: 'Slice',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Pow); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Pow) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":106
 *     OpCode.Pos: 'Pos',
 *     OpCode.Pow: 'Pow',
 *     OpCode.Range: 'Range',             # <<<<<<<<<<<<<<
 *     OpCode.Slice: 'Slice',
 *     OpCode.SliceLiteral: 'SliceLiteral',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Range); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Range) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":107
 *     OpCode.Pow: 'Pow',
 *     OpCode.Range: 'Range',
 *     OpCode.Slice: 'Slice',             # <<<<<<<<<<<<<<
 *     OpCode.SliceLiteral: 'SliceLiteral',
 *     OpCode.Export: 'Export',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Slice); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Slice) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":108
 *     OpCode.Range: 'Range',
 *     OpCode.Slice: 'Slice',
 *     OpCode.SliceLiteral: 'SliceLiteral',             # <<<<<<<<<<<<<<
 *     OpCode.Export: 'Export',
 *     OpCode.Sub: 'Sub',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_SliceLiteral); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_SliceLiteral) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":109
 *     OpCode.Slice: 'Slice',
 *     OpCode.SliceLiteral: 'SliceLiteral',
 *     OpCode.Export: 'Export',             # <<<<<<<<<<<<<<
 *     OpCode.Sub: 'Sub',
 *     OpCode.Tag: 'Tag',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Export); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Export) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":110
 *     OpCode.SliceLiteral: 'SliceLiteral',
 *     OpCode.Export: 'Export',
 *     OpCode.Sub: 'Sub',             # <<<<<<<<<<<<<<
 *     OpCode.Tag: 'Tag',
 *     OpCode.TrueDiv: 'TrueDiv',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Sub); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Sub) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":111
 *     OpCode.Export: 'Export',
 *     OpCode.Sub: 'Sub',
 *     OpCode.Tag: 'Tag',             # <<<<<<<<<<<<<<
 *     OpCode.TrueDiv: 'TrueDiv',
 *     OpCode.Xor: 'Xor',
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Tag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Tag) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":112
 *     OpCode.Sub: 'Sub',
 *     OpCode.Tag: 'Tag',
 *     OpCode.TrueDiv: 'TrueDiv',             # <<<<<<<<<<<<<<
 *     OpCode.Xor: 'Xor',
 * }
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_TrueDiv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_TrueDiv) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":113
 *     OpCode.Tag: 'Tag',
 *     OpCode.TrueDiv: 'TrueDiv',
 *     OpCode.Xor: 'Xor',             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(__pyx_e_7flitter_8language_2vm_Xor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_n_u_Xor) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XGOTREF(__pyx_v_7flitter_8language_2vm_OpCodeNames);
  __Pyx_DECREF_SET(__pyx_v_7flitter_8language_2vm_OpCodeNames, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/vm.pyx":127
 *         StatsDuration[i] = 0
 * 
 * initialize_stats()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_f_7flitter_8language_2vm_initialize_stats(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11Instruction_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Instruction___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Instruction);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Instruction, (type(self), 0x5694d08, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Instruction__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11Instruction_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Instruction___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Instruction, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Instruction);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_17InstructionVector_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionVector___reduce_cytho, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionVector);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionVector, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionVector__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_17InstructionVector_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionVector___setstate_cyt, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionVector, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionVector);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionNode_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionNode___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionNode);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionNode, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionNode__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionNode_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionNode___setstate_cytho, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionNode, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionNode);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_14InstructionStr_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionStr___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionStr);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionStr, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionStr__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_14InstructionStr_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionStr___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionStr, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionStr);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_16InstructionTuple_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionTuple___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionTuple);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionTuple, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionTuple__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_16InstructionTuple_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionTuple___setstate_cyth, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionTuple, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionTuple);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_14InstructionInt_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionInt___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionInt);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionInt, (type(self), 0xaff8343, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionInt__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_14InstructionInt_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionInt___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionInt, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionInt);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_19InstructionIntTuple_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionIntTuple___reduce_cyt, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionIntTuple, (type(self), 0x46467d3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionIntTuple__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_19InstructionIntTuple_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionIntTuple___setstate_c, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_16InstructionLabel_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionLabel___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionLabel);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionLabel, (type(self), 0x9c59c74, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionLabel__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_16InstructionLabel_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionLabel___setstate_cyth, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionLabel, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionLabel);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionJump_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionJump___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionJump);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionJump, (type(self), 0xe15daaf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionJump__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionJump_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionJump___setstate_cytho, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionJump, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionJump);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionFunc_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionFunc___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionFunc);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionFunc, (type(self), 0xbfaaaf1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionFunc__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15InstructionFunc_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionFunc___setstate_cytho, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionFunc, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionFunc);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_20InstructionObjectInt_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionObjectInt___reduce_cy, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InstructionObjectInt, (type(self), 0x3911296, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InstructionObjectInt__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_20InstructionObjectInt_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InstructionObjectInt___setstate, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);

  /* "flitter/language/vm.pyx":234
 * 
 * 
 * class StackError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])))) __PYX_ERR(0, 234, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_5, __pyx_n_s_StackError, __pyx_n_s_StackError, (PyObject *) NULL, __pyx_n_s_flitter_language_vm, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_5 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 234, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_StackError, __pyx_t_5, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StackError, __pyx_t_4) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":238
 * 
 * 
 * class StackUnderflow(StackError):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_StackError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(0, 238, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_5, __pyx_n_s_StackUnderflow, __pyx_n_s_StackUnderflow, (PyObject *) NULL, __pyx_n_s_flitter_language_vm, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_5 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 238, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_StackUnderflow, __pyx_t_5, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StackUnderflow, __pyx_t_3) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":242
 * 
 * 
 * class StackOverflow(StackError):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_StackError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5)) __PYX_ERR(0, 242, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_5, __pyx_n_s_StackOverflow, __pyx_n_s_StackOverflow, (PyObject *) NULL, __pyx_n_s_flitter_language_vm, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_5 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 242, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_StackOverflow, __pyx_t_5, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StackOverflow, __pyx_t_2) < 0) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":247
 * 
 * cdef class VectorStack:
 *     def __cinit__(self, int64_t size=DEFAULT_STACK_SIZE, int64_t max_size=MAX_STACK_SIZE):             # <<<<<<<<<<<<<<
 *         self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
 *         if self.vectors == NULL:
 */
  __pyx_k__4 = __pyx_v_7flitter_8language_2vm_DEFAULT_STACK_SIZE;
  __pyx_k__5 = __pyx_v_7flitter_8language_2vm_MAX_STACK_SIZE;

  /* "flitter/language/vm.pyx":272
 *         return '\n'.join(items)
 * 
 *     cpdef VectorStack copy(self):             # <<<<<<<<<<<<<<
 *         cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)
 *         cdef int64_t i
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_9copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_copy, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_copy, __pyx_t_5) < 0) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":283
 *         return new_stack
 * 
 *     cpdef void drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         drop(self, count)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_11drop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_drop, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__50);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_drop, __pyx_t_5) < 0) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":286
 *         drop(self, count)
 * 
 *     cpdef void push(self, Vector vector):             # <<<<<<<<<<<<<<
 *         push(self, vector)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_13push, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_push, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_push, __pyx_t_5) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":289
 *         push(self, vector)
 * 
 *     cpdef Vector pop(self):             # <<<<<<<<<<<<<<
 *         return pop(self)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_15pop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_pop, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_pop, __pyx_t_5) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":292
 *         return pop(self)
 * 
 *     cpdef tuple pop_tuple(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_tuple(self, count)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_17pop_tuple, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_pop_tuple, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_pop_tuple, __pyx_t_5) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":295
 *         return pop_tuple(self, count)
 * 
 *     cpdef list pop_list(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_list(self, count)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_19pop_list, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_pop_list, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_pop_list, __pyx_t_5) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":298
 *         return pop_list(self, count)
 * 
 *     cpdef dict pop_dict(self, tuple keys):             # <<<<<<<<<<<<<<
 *         return pop_dict(self, keys)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_21pop_dict, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_pop_dict, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_pop_dict, __pyx_t_5) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":301
 *         return pop_dict(self, keys)
 * 
 *     cpdef Vector pop_composed(self, int64_t count):             # <<<<<<<<<<<<<<
 *         return pop_composed(self, count)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_23pop_composed, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_pop_composed, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_pop_composed, __pyx_t_5) < 0) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":304
 *         return pop_composed(self, count)
 * 
 *     cpdef Vector peek(self):             # <<<<<<<<<<<<<<
 *         return peek(self)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_25peek, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_peek, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_peek, __pyx_t_5) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":307
 *         return peek(self)
 * 
 *     cpdef Vector peek_at(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         return peek_at(self, offset)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_27peek_at, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_peek_at, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_peek_at, __pyx_t_5) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":310
 *         return peek_at(self, offset)
 * 
 *     cpdef void poke(self, Vector vector):             # <<<<<<<<<<<<<<
 *         poke(self, vector)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_29poke, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_poke, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_poke, __pyx_t_5) < 0) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "flitter/language/vm.pyx":313
 *         poke(self, vector)
 * 
 *     cpdef void poke_at(self, int64_t offset, Vector vector):             # <<<<<<<<<<<<<<
 *         poke_at(self, offset, vector)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_31poke_at, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack_poke_at, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_VectorStack, __pyx_n_s_poke_at, __pyx_t_5) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_VectorStack);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_33__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11VectorStack_35__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_VectorStack___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_8Function_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Function___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Function, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Function);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0x0a66206, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_8Function_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Function___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Function, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Function);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_10LoopSource_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LoopSource___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_LoopSource, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_LoopSource);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LoopSource, (type(self), 0x2b65949, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LoopSource__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_10LoopSource_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LoopSource___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_LoopSource, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_LoopSource);

  /* "flitter/language/vm.pyx":542
 * 
 * 
 * def log_vm_stats():             # <<<<<<<<<<<<<<
 *     cdef list stats = []
 *     cdef double duration, total=0
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_1log_vm_stats, 0, __pyx_n_s_log_vm_stats, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_log_vm_stats, __pyx_t_5) < 0) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":792
 *         return self
 * 
 *     def set_path(self, object path):             # <<<<<<<<<<<<<<
 *         self.path = path
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_7set_path, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_set_path, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_set_path, __pyx_t_5) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":795
 *         self.path = path
 * 
 *     def set_top(self, object top):             # <<<<<<<<<<<<<<
 *         self.top = top
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_9set_top, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_set_top, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_set_top, __pyx_t_5) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":798
 *         self.top = top
 * 
 *     def set_pragma(self, str name, Vector value):             # <<<<<<<<<<<<<<
 *         self.pragmas[name] = value
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_11set_pragma, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_set_pragma, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_set_pragma, __pyx_t_5) < 0) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":802
 *         return self
 * 
 *     def execute(self, Context context, list lnames=None, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         """This is a test-harness function. Do not use."""
 *         assert self.initial_lnames == ()
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_13execute, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_execute, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__81);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_execute, __pyx_t_5) < 0) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":821
 *         return context.stack.pop_list(len(context.stack))
 * 
 *     def run(self, Context context, bint record_stats=False):             # <<<<<<<<<<<<<<
 *         if context.stack is None:
 *             if self.stack is None:
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_15run, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_run, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__84);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_run, __pyx_t_5) < 0) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":853
 *         return context
 * 
 *     cpdef Program link(self):             # <<<<<<<<<<<<<<
 *         cdef Instruction instruction
 *         cdef int64_t label, address, target
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_17link, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_link, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_link, __pyx_t_5) < 0) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":872
 *         return self
 * 
 *     cpdef void use_simplifier(self, bint simplify):             # <<<<<<<<<<<<<<
 *         self.simplify = simplify
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_19use_simplifier, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_use_simplifier, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_use_simplifier, __pyx_t_5) < 0) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":875
 *         self.simplify = simplify
 * 
 *     cpdef int64_t new_label(self):             # <<<<<<<<<<<<<<
 *         cdef object label = self.next_label
 *         self.next_label += 1
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_21new_label, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_new_label, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_new_label, __pyx_t_5) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":880
 *         return label
 * 
 *     cpdef Program dup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Dup))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_23dup, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_dup, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_dup, __pyx_t_5) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":884
 *         return self
 * 
 *     cpdef Program drop(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Drop, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_25drop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_drop, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__50);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_drop, __pyx_t_5) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":888
 *         return self
 * 
 *     cpdef Program label(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionLabel(label))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_27label, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_label, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_label, __pyx_t_5) < 0) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":892
 *         return self
 * 
 *     cpdef Program jump(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Jump, label))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_29jump, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_jump, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_jump, __pyx_t_5) < 0) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":896
 *         return self
 * 
 *     cpdef Program branch_true(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchTrue, label))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_31branch_true, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_branch_true, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_branch_true, __pyx_t_5) < 0) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":900
 *         return self
 * 
 *     cpdef Program branch_false(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.BranchFalse, label))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_33branch_false, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_branch_false, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_branch_false, __pyx_t_5) < 0) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":904
 *         return self
 * 
 *     cpdef Program import_(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Import, names))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_35import_, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_import, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_import, __pyx_t_5) < 0) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":908
 *         return self
 * 
 *     cpdef Program literal(self, value):             # <<<<<<<<<<<<<<
 *         cdef Vector vector = Vector._coerce(value)
 *         cdef object obj
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_37literal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_literal, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_literal, __pyx_t_5) < 0) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":925
 *         return self
 * 
 *     cpdef Program local_push(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalPush, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_39local_push, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_local_push, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_local_push, __pyx_t_5) < 0) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":929
 *         return self
 * 
 *     cpdef Program local_load(self, int64_t offset):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_41local_load, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_local_load, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_local_load, __pyx_t_5) < 0) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":933
 *         return self
 * 
 *     cpdef Program local_drop(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.LocalDrop, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_43local_drop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_local_drop, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_local_drop, __pyx_t_5) < 0) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":937
 *         return self
 * 
 *     cpdef Program lookup(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lookup))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_45lookup, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_lookup, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_lookup, __pyx_t_5) < 0) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":941
 *         return self
 * 
 *     cpdef Program lookup_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_47lookup_literal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_lookup_literal, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_lookup_literal, __pyx_t_5) < 0) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":945
 *         return self
 * 
 *     cpdef Program range(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Range))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_49range, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_range, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_range, __pyx_t_5) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":949
 *         return self
 * 
 *     cpdef Program neg(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Neg))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_51neg, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_neg, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_neg, __pyx_t_5) < 0) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":953
 *         return self
 * 
 *     cpdef Program pos(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pos))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_53pos, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_pos, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_pos, __pyx_t_5) < 0) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":957
 *         return self
 * 
 *     cpdef Program ceil(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ceil))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_55ceil, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_ceil, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_ceil, __pyx_t_5) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":961
 *         return self
 * 
 *     cpdef Program floor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Floor))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_57floor, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_floor, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_floor, __pyx_t_5) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":965
 *         return self
 * 
 *     cpdef Program fract(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Fract))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_59fract, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_fract, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_fract, __pyx_t_5) < 0) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":969
 *         return self
 * 
 *     cpdef Program not_(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Not))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_61not_, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_not, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_not, __pyx_t_5) < 0) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":973
 *         return self
 * 
 *     cpdef Program add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Add))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_63add, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_add, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_add, __pyx_t_5) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":977
 *         return self
 * 
 *     cpdef Program sub(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Sub))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_65sub, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_sub, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_sub, __pyx_t_5) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":981
 *         return self
 * 
 *     cpdef Program mul(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mul))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_67mul, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_mul, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_mul, __pyx_t_5) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":985
 *         return self
 * 
 *     cpdef Program mul_add(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.MulAdd))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_69mul_add, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_mul_add, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_mul_add, __pyx_t_5) < 0) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":989
 *         return self
 * 
 *     cpdef Program truediv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.TrueDiv))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_71truediv, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_truediv, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_truediv, __pyx_t_5) < 0) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":993
 *         return self
 * 
 *     cpdef Program floordiv(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.FloorDiv))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_73floordiv, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_floordiv, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_floordiv, __pyx_t_5) < 0) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":997
 *         return self
 * 
 *     cpdef Program mod(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Mod))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_75mod, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_mod, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_mod, __pyx_t_5) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1001
 *         return self
 * 
 *     cpdef Program pow(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Pow))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_77pow, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_pow, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__119)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_pow, __pyx_t_5) < 0) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1005
 *         return self
 * 
 *     cpdef Program eq(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Eq))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_79eq, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_eq, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_eq, __pyx_t_5) < 0) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1009
 *         return self
 * 
 *     cpdef Program ne(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ne))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_81ne, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_ne, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_ne, __pyx_t_5) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1013
 *         return self
 * 
 *     cpdef Program gt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Gt))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_83gt, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_gt, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_gt, __pyx_t_5) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1017
 *         return self
 * 
 *     cpdef Program lt(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Lt))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_85lt, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_lt, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_lt, __pyx_t_5) < 0) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1021
 *         return self
 * 
 *     cpdef Program ge(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Ge))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_87ge, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_ge, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_ge, __pyx_t_5) < 0) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1025
 *         return self
 * 
 *     cpdef Program le(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Le))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_89le, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_le, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_le, __pyx_t_5) < 0) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1029
 *         return self
 * 
 *     cpdef Program xor(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Xor))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_91xor, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_xor, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_xor, __pyx_t_5) < 0) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1033
 *         return self
 * 
 *     cpdef Program slice(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Slice))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_93slice, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_slice, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_slice, __pyx_t_5) < 0) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1037
 *         return self
 * 
 *     cpdef Program slice_literal(self, Vector value):             # <<<<<<<<<<<<<<
 *         if value.length == 1 and value.numbers != NULL:
 *             self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int64_t>c_floor(value.numbers[0])))
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_95slice_literal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_slice_literal, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_slice_literal, __pyx_t_5) < 0) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1044
 *         return self
 * 
 *     cpdef Program call(self, int64_t count, tuple names=None):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_97call, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_call, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__131);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_call, __pyx_t_5) < 0) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1048
 *         return self
 * 
 *     cpdef Program call_fast(self, function, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_99call_fast, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_call_fast, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__133)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_call_fast, __pyx_t_5) < 0) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1052
 *         return self
 * 
 *     cpdef Program tag(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Tag, name))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_101tag, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_tag, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_tag, __pyx_t_5) < 0) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1056
 *         return self
 * 
 *     cpdef Program attributes(self, tuple names):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionTuple(OpCode.Attributes, names))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_103attributes, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_attributes, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_attributes, __pyx_t_5) < 0) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1060
 *         return self
 * 
 *     cpdef Program append(self, int64_t count=1):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Append, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_105append, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_append, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__137)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__50);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_append, __pyx_t_5) < 0) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1064
 *         return self
 * 
 *     cpdef Program compose(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.Compose, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_107compose, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_compose, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_compose, __pyx_t_5) < 0) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1068
 *         return self
 * 
 *     cpdef Program begin_for(self, int64_t count):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionInt(OpCode.BeginFor, count))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_109begin_for, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_begin_for, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_begin_for, __pyx_t_5) < 0) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1072
 *         return self
 * 
 *     cpdef Program next(self, int64_t label):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionJump(OpCode.Next, label))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_111next, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_next, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_next, __pyx_t_5) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1076
 *         return self
 * 
 *     cpdef Program end_for(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.EndFor))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_113end_for, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_end_for, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_end_for, __pyx_t_5) < 0) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1080
 *         return self
 * 
 *     cpdef Program export(self, str name):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionStr(OpCode.Export, name))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_115export, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_export, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_export, __pyx_t_5) < 0) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1084
 *         return self
 * 
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=0):             # <<<<<<<<<<<<<<
 *         self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_117func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_func, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__145);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_func, __pyx_t_5) < 0) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "flitter/language/vm.pyx":1088
 *         return self
 * 
 *     cpdef Program exit(self):             # <<<<<<<<<<<<<<
 *         self.instructions.append(Instruction(OpCode.Exit))
 *         return self
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_119exit, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program_exit, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program, __pyx_n_s_exit, __pyx_t_5) < 0) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_2vm_Program);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_121__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program___reduce_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7Program_123__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Program___setstate_cython, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Instruction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_3__pyx_unpickle_Instruction, 0, __pyx_n_s_pyx_unpickle_Instruction, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Instruction, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Instruction__set_state(<Instruction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Instruction__set_state(Instruction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_5__pyx_unpickle_InstructionVector, 0, __pyx_n_s_pyx_unpickle_InstructionVector, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionVector, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_7__pyx_unpickle_InstructionNode, 0, __pyx_n_s_pyx_unpickle_InstructionNode, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionNode, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionNode__set_state(<InstructionNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionNode__set_state(InstructionNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_9__pyx_unpickle_InstructionStr, 0, __pyx_n_s_pyx_unpickle_InstructionStr, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionStr, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_11__pyx_unpickle_InstructionTuple, 0, __pyx_n_s_pyx_unpickle_InstructionTuple, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionTuple, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionTuple__set_state(<InstructionTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionTuple__set_state(InstructionTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.value = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_13__pyx_unpickle_InstructionInt, 0, __pyx_n_s_pyx_unpickle_InstructionInt, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionInt, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionIntTuple(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_15__pyx_unpickle_InstructionIntTuple, 0, __pyx_n_s_pyx_unpickle_InstructionIntTup, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionIntTup, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionIntTuple__set_state(<InstructionIntTuple> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionIntTuple__set_state(InstructionIntTuple __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.ivalue = __pyx_state[1]; __pyx_result.tvalue = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_17__pyx_unpickle_InstructionLabel, 0, __pyx_n_s_pyx_unpickle_InstructionLabel, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__157)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionLabel, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionJump(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_19__pyx_unpickle_InstructionJump, 0, __pyx_n_s_pyx_unpickle_InstructionJump, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionJump, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionJump__set_state(<InstructionJump> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionJump__set_state(InstructionJump __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.label = __pyx_state[1]; __pyx_result.offset = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_21__pyx_unpickle_InstructionFunc, 0, __pyx_n_s_pyx_unpickle_InstructionFunc, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__159)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionFunc, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InstructionObjectInt(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_23__pyx_unpickle_InstructionObjectInt, 0, __pyx_n_s_pyx_unpickle_InstructionObject, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InstructionObject, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_InstructionObjectInt__set_state(<InstructionObjectInt> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InstructionObjectInt__set_state(InstructionObjectInt __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_25__pyx_unpickle_Function, 0, __pyx_n_s_pyx_unpickle_Function, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__161)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Function, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LoopSource(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_2vm_27__pyx_unpickle_LoopSource, 0, __pyx_n_s_pyx_unpickle_LoopSource, NULL, __pyx_n_s_flitter_language_vm, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LoopSource, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/language/vm.pyx":1
 * # cython: wraparound=False, boundscheck=False             # <<<<<<<<<<<<<<
 * 
 * """
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init flitter.language.vm", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flitter.language.vm");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (__Pyx_IS_TYPE(L, &PySet_Type)) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       int is_safe_type) {
    PyObject* value;
    CYTHON_MAYBE_UNUSED_VAR(is_safe_type);
#if PY_VERSION_HEX >= 0x030400A0
    if ((1)) {
        value = PyDict_SetDefault(d, key, default_value);
        if (unlikely(!value)) return NULL;
        Py_INCREF(value);
#else
    if (is_safe_type == 1 || (is_safe_type == -1 &&
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
            (PyUnicode_CheckExact(key) || PyString_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItemWithError(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#else
            (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#endif
#endif
    } else {
        value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_setdefault, d, key, default_value);
    }
    return value;
}

/* RaiseNoneIterError */
    static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
    static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
    static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
    #if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* SliceTupleAndList */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (length <= 0) {
        return PyList_New(0);
    }
    return __Pyx_PyList_FromArray(((PyListObject*)src)->ob_item + start, length);
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    return __Pyx_PyTuple_FromArray(((PyTupleObject*)src)->ob_item + start, length);
}
#endif

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import_2);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__12);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
    static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
    static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpTraverse */
    static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
    static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* FixUpExtensionType */
    #if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* ValidateBasesTuple */
    #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
    static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
    static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
    static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CallUnboundCMethod1 */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* FetchSharedCythonModule */
    static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
    static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
    #if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
    #if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
    static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3UpdateBases */
    static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_MACROS
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args */
    static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectLookupSpecial */
    #if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
    static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7flitter_8language_2vm_OpCode(enum __pyx_t_7flitter_8language_2vm_OpCode value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_7flitter_8language_2vm_OpCode neg_one = (enum __pyx_t_7flitter_8language_2vm_OpCode) -1, const_zero = (enum __pyx_t_7flitter_8language_2vm_OpCode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
    static CYTHON_INLINE enum __pyx_t_7flitter_8language_2vm_OpCode __Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_7flitter_8language_2vm_OpCode neg_one = (enum __pyx_t_7flitter_8language_2vm_OpCode) -1, const_zero = (enum __pyx_t_7flitter_8language_2vm_OpCode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7flitter_8language_2vm_OpCode) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        enum __pyx_t_7flitter_8language_2vm_OpCode val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7flitter_8language_2vm_OpCode(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) >= 2 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) >= 3 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) >= 4 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((enum __pyx_t_7flitter_8language_2vm_OpCode)-1)*(((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) ((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((enum __pyx_t_7flitter_8language_2vm_OpCode)-1)*(((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) ((((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) (((enum __pyx_t_7flitter_8language_2vm_OpCode)-1)*(((((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7flitter_8language_2vm_OpCode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum __pyx_t_7flitter_8language_2vm_OpCode) ((((((((((enum __pyx_t_7flitter_8language_2vm_OpCode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7flitter_8language_2vm_OpCode)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7flitter_8language_2vm_OpCode, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(enum __pyx_t_7flitter_8language_2vm_OpCode) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7flitter_8language_2vm_OpCode, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        enum __pyx_t_7flitter_8language_2vm_OpCode val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
#endif
        if (unlikely(ret))
            return (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7flitter_8language_2vm_OpCode");
    return (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7flitter_8language_2vm_OpCode");
    return (enum __pyx_t_7flitter_8language_2vm_OpCode) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
    static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
    static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
    #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__163);
    }
    return name;
}
#endif

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* PyObjectCallMethod1 */
    #if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
    #include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
    static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
    #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
    static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
    static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* PyObjectSetAttrStr */
    #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
    static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
    #ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_11
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_11
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
    #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
