/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "flitter.language.tree",
        "sources": [
            "src/flitter/language/tree.pyx"
        ]
    },
    "module_name": "flitter.language.tree"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flitter__language__tree
#define __PYX_HAVE_API__flitter__language__tree
/* Early includes */
#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/flitter/language/tree.pyx",
  "contextvars.pxd",
  "<stringsource>",
  "src/flitter/model.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "src/flitter/language/vm.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7flitter_5model_Vector;
struct __pyx_obj_7flitter_5model_Matrix33;
struct __pyx_obj_7flitter_5model_Matrix44;
struct __pyx_obj_7flitter_5model_Quaternion;
struct __pyx_obj_7flitter_5model_Node;
struct __pyx_obj_7flitter_5model_StateDict;
struct __pyx_obj_7flitter_5model_Context;
struct __pyx_obj_7flitter_8language_2vm_VectorStack;
struct __pyx_obj_7flitter_8language_2vm_Instruction;
struct __pyx_obj_7flitter_8language_2vm_InstructionVector;
struct __pyx_obj_7flitter_8language_2vm_InstructionNode;
struct __pyx_obj_7flitter_8language_2vm_InstructionStr;
struct __pyx_obj_7flitter_8language_2vm_InstructionTuple;
struct __pyx_obj_7flitter_8language_2vm_InstructionInt;
struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple;
struct __pyx_obj_7flitter_8language_2vm_InstructionLabel;
struct __pyx_obj_7flitter_8language_2vm_InstructionJump;
struct __pyx_obj_7flitter_8language_2vm_InstructionFunc;
struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt;
struct __pyx_obj_7flitter_8language_2vm_Program;
struct __pyx_obj_7flitter_8language_4tree_Expression;
struct __pyx_obj_7flitter_8language_4tree_Top;
struct __pyx_obj_7flitter_8language_4tree_Export;
struct __pyx_obj_7flitter_8language_4tree_Import;
struct __pyx_obj_7flitter_8language_4tree_Sequence;
struct __pyx_obj_7flitter_8language_4tree_Literal;
struct __pyx_obj_7flitter_8language_4tree_Name;
struct __pyx_obj_7flitter_8language_4tree_Lookup;
struct __pyx_obj_7flitter_8language_4tree_Range;
struct __pyx_obj_7flitter_8language_4tree_UnaryOperation;
struct __pyx_obj_7flitter_8language_4tree_Negative;
struct __pyx_obj_7flitter_8language_4tree_Positive;
struct __pyx_obj_7flitter_8language_4tree_Not;
struct __pyx_obj_7flitter_8language_4tree_Ceil;
struct __pyx_obj_7flitter_8language_4tree_Floor;
struct __pyx_obj_7flitter_8language_4tree_Fract;
struct __pyx_obj_7flitter_8language_4tree_BinaryOperation;
struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation;
struct __pyx_obj_7flitter_8language_4tree_Add;
struct __pyx_obj_7flitter_8language_4tree_Subtract;
struct __pyx_obj_7flitter_8language_4tree_Multiply;
struct __pyx_obj_7flitter_8language_4tree_Divide;
struct __pyx_obj_7flitter_8language_4tree_FloorDivide;
struct __pyx_obj_7flitter_8language_4tree_Modulo;
struct __pyx_obj_7flitter_8language_4tree_Power;
struct __pyx_obj_7flitter_8language_4tree_Comparison;
struct __pyx_obj_7flitter_8language_4tree_EqualTo;
struct __pyx_obj_7flitter_8language_4tree_NotEqualTo;
struct __pyx_obj_7flitter_8language_4tree_LessThan;
struct __pyx_obj_7flitter_8language_4tree_GreaterThan;
struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo;
struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo;
struct __pyx_obj_7flitter_8language_4tree_And;
struct __pyx_obj_7flitter_8language_4tree_Or;
struct __pyx_obj_7flitter_8language_4tree_Xor;
struct __pyx_obj_7flitter_8language_4tree_Slice;
struct __pyx_obj_7flitter_8language_4tree_Call;
struct __pyx_obj_7flitter_8language_4tree_NodeModifier;
struct __pyx_obj_7flitter_8language_4tree_Tag;
struct __pyx_obj_7flitter_8language_4tree_Attributes;
struct __pyx_obj_7flitter_8language_4tree_Append;
struct __pyx_obj_7flitter_8language_4tree_Binding;
struct __pyx_obj_7flitter_8language_4tree_PolyBinding;
struct __pyx_obj_7flitter_8language_4tree_Let;
struct __pyx_obj_7flitter_8language_4tree_For;
struct __pyx_obj_7flitter_8language_4tree_IfCondition;
struct __pyx_obj_7flitter_8language_4tree_IfElse;
struct __pyx_obj_7flitter_8language_4tree_Function;
struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr;
struct __pyx_opt_args_7flitter_5model_6Vector_match;
struct __pyx_opt_args_7flitter_5model_4Node_get;

/* "model.pxd":63
 *     cdef str as_string(self)
 *     cpdef uint64_t hash(self, bint floor_floats)
 *     cpdef object match(self, int64_t n=?, type t=?, default=?)             # <<<<<<<<<<<<<<
 *     cdef str repr(self)
 *     cdef Vector neg(self)
 */
struct __pyx_opt_args_7flitter_5model_6Vector_match {
  int __pyx_n;
  int64_t n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};

/* "model.pxd":212
 *     cpdef void append_vector(self, Vector nodes)
 *     cdef bint _equals(self, Node other)
 *     cpdef object get(self, str name, int n=?, type t=?, object default=?)             # <<<<<<<<<<<<<<
 *     cdef Vector get_vec(self, str name, Vector default)
 *     cdef Vector get_fvec(self, str name, int n, Vector default)
 */
struct __pyx_opt_args_7flitter_5model_4Node_get {
  int __pyx_n;
  int n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_call;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_append;
struct __pyx_opt_args_7flitter_8language_2vm_7Program_func;

/* "vm.pxd":34
 * 
 * 
 * cdef enum OpCode:             # <<<<<<<<<<<<<<
 *     Add
 *     Append
 */
enum __pyx_t_7flitter_8language_2vm_OpCode {
  __pyx_e_7flitter_8language_2vm_Add,
  __pyx_e_7flitter_8language_2vm_Append,
  __pyx_e_7flitter_8language_2vm_Attributes,
  __pyx_e_7flitter_8language_2vm_BeginFor,
  __pyx_e_7flitter_8language_2vm_BranchFalse,
  __pyx_e_7flitter_8language_2vm_BranchTrue,
  __pyx_e_7flitter_8language_2vm_Call,
  __pyx_e_7flitter_8language_2vm_CallFast,
  __pyx_e_7flitter_8language_2vm_Ceil,
  __pyx_e_7flitter_8language_2vm_Compose,
  __pyx_e_7flitter_8language_2vm_Drop,
  __pyx_e_7flitter_8language_2vm_Dup,
  __pyx_e_7flitter_8language_2vm_EndFor,
  __pyx_e_7flitter_8language_2vm_Eq,
  __pyx_e_7flitter_8language_2vm_Floor,
  __pyx_e_7flitter_8language_2vm_FloorDiv,
  __pyx_e_7flitter_8language_2vm_Fract,
  __pyx_e_7flitter_8language_2vm_Func,
  __pyx_e_7flitter_8language_2vm_Exit,
  __pyx_e_7flitter_8language_2vm_Ge,
  __pyx_e_7flitter_8language_2vm_Gt,
  __pyx_e_7flitter_8language_2vm_Import,
  __pyx_e_7flitter_8language_2vm_IndexLiteral,
  __pyx_e_7flitter_8language_2vm_Jump,
  __pyx_e_7flitter_8language_2vm_Label,
  __pyx_e_7flitter_8language_2vm_Le,
  __pyx_e_7flitter_8language_2vm_Literal,
  __pyx_e_7flitter_8language_2vm_LiteralNode,
  __pyx_e_7flitter_8language_2vm_LiteralNodes,
  __pyx_e_7flitter_8language_2vm_LocalDrop,
  __pyx_e_7flitter_8language_2vm_LocalLoad,
  __pyx_e_7flitter_8language_2vm_LocalPush,
  __pyx_e_7flitter_8language_2vm_Lookup,
  __pyx_e_7flitter_8language_2vm_LookupLiteral,
  __pyx_e_7flitter_8language_2vm_Lt,
  __pyx_e_7flitter_8language_2vm_Mod,
  __pyx_e_7flitter_8language_2vm_Mul,
  __pyx_e_7flitter_8language_2vm_MulAdd,
  __pyx_e_7flitter_8language_2vm_Ne,
  __pyx_e_7flitter_8language_2vm_Neg,
  __pyx_e_7flitter_8language_2vm_Next,
  __pyx_e_7flitter_8language_2vm_Not,
  __pyx_e_7flitter_8language_2vm_Pos,
  __pyx_e_7flitter_8language_2vm_Pow,
  __pyx_e_7flitter_8language_2vm_Range,
  __pyx_e_7flitter_8language_2vm_Slice,
  __pyx_e_7flitter_8language_2vm_SliceLiteral,
  __pyx_e_7flitter_8language_2vm_Export,
  __pyx_e_7flitter_8language_2vm_Sub,
  __pyx_e_7flitter_8language_2vm_Tag,
  __pyx_e_7flitter_8language_2vm_TrueDiv,
  __pyx_e_7flitter_8language_2vm_Xor,
  __pyx_e_7flitter_8language_2vm_MAX
};

/* "vm.pxd":21
 * 
 *     cpdef VectorStack copy(self)
 *     cpdef void drop(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef void push(self, Vector vector)
 *     cpdef Vector pop(self)
 */
struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop {
  int __pyx_n;
  int64_t count;
};

/* "vm.pxd":159
 *     cpdef int64_t new_label(self)
 *     cpdef Program dup(self)
 *     cpdef Program drop(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef Program label(self, int64_t label)
 *     cpdef Program jump(self, int64_t label)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop {
  int __pyx_n;
  int64_t count;
};

/* "vm.pxd":195
 *     cpdef Program slice(self)
 *     cpdef Program slice_literal(self, Vector value)
 *     cpdef Program call(self, int64_t count, tuple names=?)             # <<<<<<<<<<<<<<
 *     cpdef Program call_fast(self, function, int64_t count)
 *     cpdef Program tag(self, str name)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_call {
  int __pyx_n;
  PyObject *names;
};

/* "vm.pxd":199
 *     cpdef Program tag(self, str name)
 *     cpdef Program attributes(self, tuple names)
 *     cpdef Program append(self, int64_t count=?)             # <<<<<<<<<<<<<<
 *     cpdef Program compose(self, int64_t count)
 *     cpdef Program begin_for(self, int64_t count)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_append {
  int __pyx_n;
  int64_t count;
};

/* "vm.pxd":205
 *     cpdef Program end_for(self)
 *     cpdef Program export(self, str name)
 *     cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=?)             # <<<<<<<<<<<<<<
 *     cpdef Program exit(self)
 *     cdef void _execute(self, Context context, int64_t pc, bint record_stats)
 */
struct __pyx_opt_args_7flitter_8language_2vm_7Program_func {
  int __pyx_n;
  int64_t ncaptures;
};

/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */
struct __pyx_obj_7flitter_5model_Vector {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtab;
  int64_t length;
  PyObject *objects;
  double *numbers;
  double _numbers[16];
  uint64_t _hash;
};


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix33 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":128
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix44 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":178
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */
struct __pyx_obj_7flitter_5model_Quaternion {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":198
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */
struct __pyx_obj_7flitter_5model_Node {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtab;
  PyObject *kind;
  PyObject *_tags;
  PyObject *_attributes;
  int _attributes_shared;
  PyObject *_children;
};


/* "model.pxd":222
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */
struct __pyx_obj_7flitter_5model_StateDict {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtab;
  PyObject *_changed_keys;
  PyObject *_state;
};


/* "model.pxd":231
 * 
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 *     cdef readonly dict names
 *     cdef readonly set captures
 */
struct __pyx_obj_7flitter_5model_Context {
  PyObject_HEAD
  PyObject *names;
  PyObject *captures;
  struct __pyx_obj_7flitter_5model_StateDict *state;
  struct __pyx_obj_7flitter_5model_Node *root;
  PyObject *path;
  struct __pyx_obj_7flitter_5model_Context *parent;
  PyObject *exports;
  PyObject *errors;
  PyObject *logs;
  PyObject *references;
  PyObject *stack;
  PyObject *lnames;
  PyObject *dependencies;
};


/* "vm.pxd":14
 * 
 * 
 * cdef class VectorStack:             # <<<<<<<<<<<<<<
 *     cdef PyObject** vectors
 *     cdef int64_t top
 */
struct __pyx_obj_7flitter_8language_2vm_VectorStack {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack *__pyx_vtab;
  PyObject **vectors;
  int64_t top;
  int64_t size;
  int64_t max_size;
};


/* "vm.pxd":90
 * 
 * 
 * cdef class Instruction:             # <<<<<<<<<<<<<<
 *     cdef readonly OpCode code
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_Instruction {
  PyObject_HEAD
  enum __pyx_t_7flitter_8language_2vm_OpCode code;
};


/* "vm.pxd":94
 * 
 * 
 * cdef class InstructionVector(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly Vector value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionVector {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *value;
};


/* "vm.pxd":98
 * 
 * 
 * cdef class InstructionNode(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly Node value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionNode {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  struct __pyx_obj_7flitter_5model_Node *value;
};


/* "vm.pxd":102
 * 
 * 
 * cdef class InstructionStr(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly str value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionStr {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *value;
};


/* "vm.pxd":106
 * 
 * 
 * cdef class InstructionTuple(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionTuple {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *value;
};


/* "vm.pxd":110
 * 
 * 
 * cdef class InstructionInt(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t value
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionInt {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t value;
};


/* "vm.pxd":114
 * 
 * 
 * cdef class InstructionIntTuple(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t ivalue
 *     cdef readonly tuple tvalue
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t ivalue;
  PyObject *tvalue;
};


/* "vm.pxd":119
 * 
 * 
 * cdef class InstructionLabel(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t label
 * 
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionLabel {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t label;
};


/* "vm.pxd":123
 * 
 * 
 * cdef class InstructionJump(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t label
 *     cdef readonly int64_t offset
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionJump {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  int64_t label;
  int64_t offset;
};


/* "vm.pxd":128
 * 
 * 
 * cdef class InstructionFunc(InstructionJump):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionFunc {
  struct __pyx_obj_7flitter_8language_2vm_InstructionJump __pyx_base;
  PyObject *name;
  PyObject *parameters;
  int64_t ncaptures;
};


/* "vm.pxd":134
 * 
 * 
 * cdef class InstructionObjectInt(Instruction):             # <<<<<<<<<<<<<<
 *     cdef readonly object obj
 *     cdef readonly int64_t value
 */
struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt {
  struct __pyx_obj_7flitter_8language_2vm_Instruction __pyx_base;
  PyObject *obj;
  int64_t value;
};


/* "vm.pxd":139
 * 
 * 
 * cdef class Program:             # <<<<<<<<<<<<<<
 *     cdef readonly dict pragmas
 *     cdef readonly list instructions
 */
struct __pyx_obj_7flitter_8language_2vm_Program {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_8language_2vm_Program *__pyx_vtab;
  PyObject *pragmas;
  PyObject *instructions;
  int linked;
  PyObject *path;
  PyObject *top;
  PyObject *initial_lnames;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *stack;
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *lnames;
  PyObject *compiler_errors;
  int simplify;
  int64_t next_label;
};


/* "flitter/language/tree.pyx":49
 * 
 * 
 * cdef class Expression:             # <<<<<<<<<<<<<<
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 */
struct __pyx_obj_7flitter_8language_4tree_Expression {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *__pyx_vtab;
};


/* "flitter/language/tree.pyx":92
 * 
 * 
 * cdef class Top(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple pragmas
 *     cdef readonly Expression body
 */
struct __pyx_obj_7flitter_8language_4tree_Top {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *pragmas;
  struct __pyx_obj_7flitter_8language_4tree_Expression *body;
  PyObject *dependencies;
};


/* "flitter/language/tree.pyx":126
 * 
 * 
 * cdef class Export(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly dict static_exports
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Export {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *static_exports;
};


/* "flitter/language/tree.pyx":166
 * 
 * 
 * cdef class Import(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple names
 *     cdef readonly Expression filename
 */
struct __pyx_obj_7flitter_8language_4tree_Import {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *names;
  struct __pyx_obj_7flitter_8language_4tree_Expression *filename;
  struct __pyx_obj_7flitter_8language_4tree_Expression *expr;
};


/* "flitter/language/tree.pyx":234
 * 
 * 
 * cdef class Sequence(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple expressions
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Sequence {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *expressions;
};


/* "flitter/language/tree.pyx":276
 * 
 * 
 * cdef class Literal(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Vector value
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Literal {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *value;
};


/* "flitter/language/tree.pyx":292
 * 
 * 
 * cdef class Name(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Name {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *name;
};


/* "flitter/language/tree.pyx":338
 * 
 * 
 * cdef class Lookup(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression key
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Lookup {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *key;
};


/* "flitter/language/tree.pyx":366
 * 
 * 
 * cdef class Range(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression start
 *     cdef readonly Expression stop
 */
struct __pyx_obj_7flitter_8language_4tree_Range {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *start;
  struct __pyx_obj_7flitter_8language_4tree_Expression *stop;
  struct __pyx_obj_7flitter_8language_4tree_Expression *step;
};


/* "flitter/language/tree.pyx":399
 * 
 * 
 * cdef class UnaryOperation(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_UnaryOperation {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *expr;
};


/* "flitter/language/tree.pyx":416
 * 
 * 
 * cdef class Negative(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.neg()
 */
struct __pyx_obj_7flitter_8language_4tree_Negative {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":451
 * 
 * 
 * cdef class Positive(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.pos()
 */
struct __pyx_obj_7flitter_8language_4tree_Positive {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":466
 * 
 * 
 * cdef class Not(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.not_()
 */
struct __pyx_obj_7flitter_8language_4tree_Not {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":479
 * 
 * 
 * cdef class Ceil(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.ceil()
 */
struct __pyx_obj_7flitter_8language_4tree_Ceil {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":492
 * 
 * 
 * cdef class Floor(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.floor()
 */
struct __pyx_obj_7flitter_8language_4tree_Floor {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":505
 * 
 * 
 * cdef class Fract(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.fract()
 */
struct __pyx_obj_7flitter_8language_4tree_Fract {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":518
 * 
 * 
 * cdef class BinaryOperation(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression left
 *     cdef readonly Expression right
 */
struct __pyx_obj_7flitter_8language_4tree_BinaryOperation {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *left;
  struct __pyx_obj_7flitter_8language_4tree_Expression *right;
};


/* "flitter/language/tree.pyx":570
 * 
 * 
 * cdef class MathsBinaryOperation(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 */
struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":581
 * 
 * 
 * cdef class Add(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.add(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Add {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":608
 * 
 * 
 * cdef class Subtract(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.sub(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Subtract {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":629
 * 
 * 
 * cdef class Multiply(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mul(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Multiply {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":663
 * 
 * 
 * cdef class Divide(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.truediv(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Divide {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":676
 * 
 * 
 * cdef class FloorDivide(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.floordiv(right)
 */
struct __pyx_obj_7flitter_8language_4tree_FloorDivide {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":688
 * 
 * 
 * cdef class Modulo(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mod(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Modulo {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":700
 * 
 * 
 * cdef class Power(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.pow(right)
 */
struct __pyx_obj_7flitter_8language_4tree_Power {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":718
 * 
 * 
 * cdef class Comparison(BinaryOperation):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Comparison {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":722
 * 
 * 
 * cdef class EqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.eq(right)
 */
struct __pyx_obj_7flitter_8language_4tree_EqualTo {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":730
 * 
 * 
 * cdef class NotEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ne(right)
 */
struct __pyx_obj_7flitter_8language_4tree_NotEqualTo {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":738
 * 
 * 
 * cdef class LessThan(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.lt(right)
 */
struct __pyx_obj_7flitter_8language_4tree_LessThan {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":746
 * 
 * 
 * cdef class GreaterThan(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.gt(right)
 */
struct __pyx_obj_7flitter_8language_4tree_GreaterThan {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":754
 * 
 * 
 * cdef class LessThanOrEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.le(right)
 */
struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":762
 * 
 * 
 * cdef class GreaterThanOrEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ge(right)
 */
struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo {
  struct __pyx_obj_7flitter_8language_4tree_Comparison __pyx_base;
};


/* "flitter/language/tree.pyx":770
 * 
 * 
 * cdef class And(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 */
struct __pyx_obj_7flitter_8language_4tree_And {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":792
 * 
 * 
 * cdef class Or(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 */
struct __pyx_obj_7flitter_8language_4tree_Or {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":814
 * 
 * 
 * cdef class Xor(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.xor()
 */
struct __pyx_obj_7flitter_8language_4tree_Xor {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation __pyx_base;
};


/* "flitter/language/tree.pyx":834
 * 
 * 
 * cdef class Slice(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 *     cdef readonly Expression index
 */
struct __pyx_obj_7flitter_8language_4tree_Slice {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *expr;
  struct __pyx_obj_7flitter_8language_4tree_Expression *index;
};


/* "flitter/language/tree.pyx":867
 * 
 * 
 * cdef class Call(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression function
 *     cdef readonly tuple args
 */
struct __pyx_obj_7flitter_8language_4tree_Call {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *function;
  PyObject *args;
  PyObject *keyword_args;
};


/* "flitter/language/tree.pyx":975
 * 
 * 
 * cdef class NodeModifier(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression node
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_NodeModifier {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *node;
};


/* "flitter/language/tree.pyx":979
 * 
 * 
 * cdef class Tag(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly str tag
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Tag {
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier __pyx_base;
  PyObject *tag;
};


/* "flitter/language/tree.pyx":1018
 * 
 * 
 * cdef class Attributes(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple bindings
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Attributes {
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier __pyx_base;
  PyObject *bindings;
};


/* "flitter/language/tree.pyx":1079
 * 
 * 
 * cdef class Append(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression children
 * 
 */
struct __pyx_obj_7flitter_8language_4tree_Append {
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier __pyx_base;
  struct __pyx_obj_7flitter_8language_4tree_Expression *children;
};


/* "flitter/language/tree.pyx":1135
 * 
 * 
 * cdef class Binding:             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 *     cdef readonly Expression expr
 */
struct __pyx_obj_7flitter_8language_4tree_Binding {
  PyObject_HEAD
  PyObject *name;
  struct __pyx_obj_7flitter_8language_4tree_Expression *expr;
};


/* "flitter/language/tree.pyx":1147
 * 
 * 
 * cdef class PolyBinding:             # <<<<<<<<<<<<<<
 *     cdef readonly tuple names
 *     cdef readonly Expression expr
 */
struct __pyx_obj_7flitter_8language_4tree_PolyBinding {
  PyObject_HEAD
  PyObject *names;
  struct __pyx_obj_7flitter_8language_4tree_Expression *expr;
};


/* "flitter/language/tree.pyx":1159
 * 
 * 
 * cdef class Let(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple bindings
 *     cdef readonly Expression body
 */
struct __pyx_obj_7flitter_8language_4tree_Let {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *bindings;
  struct __pyx_obj_7flitter_8language_4tree_Expression *body;
};


/* "flitter/language/tree.pyx":1275
 * 
 * 
 * cdef class For(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple names
 *     cdef readonly Expression source
 */
struct __pyx_obj_7flitter_8language_4tree_For {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *names;
  struct __pyx_obj_7flitter_8language_4tree_Expression *source;
  struct __pyx_obj_7flitter_8language_4tree_Expression *body;
};


/* "flitter/language/tree.pyx":1335
 * 
 * 
 * cdef class IfCondition:             # <<<<<<<<<<<<<<
 *     cdef readonly Expression condition
 *     cdef readonly Expression then
 */
struct __pyx_obj_7flitter_8language_4tree_IfCondition {
  PyObject_HEAD
  struct __pyx_obj_7flitter_8language_4tree_Expression *condition;
  struct __pyx_obj_7flitter_8language_4tree_Expression *then;
};


/* "flitter/language/tree.pyx":1347
 * 
 * 
 * cdef class IfElse(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple tests
 *     cdef readonly Expression else_
 */
struct __pyx_obj_7flitter_8language_4tree_IfElse {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *tests;
  struct __pyx_obj_7flitter_8language_4tree_Expression *else_;
};


/* "flitter/language/tree.pyx":1403
 * 
 * 
 * cdef class Function(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 */
struct __pyx_obj_7flitter_8language_4tree_Function {
  struct __pyx_obj_7flitter_8language_4tree_Expression __pyx_base;
  PyObject *name;
  PyObject *parameters;
  struct __pyx_obj_7flitter_8language_4tree_Expression *body;
  PyObject *captures;
  int inlineable;
  int recursive;
};


/* "flitter/language/tree.pyx":221
 *         cdef Expression expr = self.expr
 *         if let_names:
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)             # <<<<<<<<<<<<<<
 *         expr = expr._simplify(context)
 *         context.names = saved
 */
struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_name;
  PyObject *__pyx_v_value;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};



/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */

struct __pyx_vtabstruct_7flitter_5model_Vector {
  struct __pyx_obj_7flitter_5model_Vector *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_compose)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_symbol)(PyObject *);
  int64_t (*allocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  void (*deallocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*copy)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*intern)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*fill_range)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*isinstance)(struct __pyx_obj_7flitter_5model_Vector *, PyObject *, int __pyx_skip_dispatch);
  int (*is_finite)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*as_bool)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*as_double)(struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*as_integer)(struct __pyx_obj_7flitter_5model_Vector *);
  PyObject *(*as_string)(struct __pyx_obj_7flitter_5model_Vector *);
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Vector *, int, int __pyx_skip_dispatch);
  PyObject *(*match)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_6Vector_match *__pyx_optional_args);
  PyObject *(*repr)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*neg)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pos)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*abs)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ceil)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floor)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*fract)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*round)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  struct __pyx_obj_7flitter_5model_Vector *(*add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul_add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*sub)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*truediv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floordiv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mod)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pow)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*eq)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ne)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*gt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ge)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*lt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*le)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*compare)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*slice)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*item)(struct __pyx_obj_7flitter_5model_Vector *, int64_t, int __pyx_skip_dispatch);
  double (*squared_sum)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*normalize)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*dot)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*cross)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*clamp)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  double (*minimum)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*maximum)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*concat)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtabptr_7flitter_5model_Vector;


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix33 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_rotate)(double);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Matrix33 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix33 *__pyx_vtabptr_7flitter_5model_Matrix33;


/* "model.pxd":128
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix44 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_project)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_ortho)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_look)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_x)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_y)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_z)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_x)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_y)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_z)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*immul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse_transpose_matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix44 *__pyx_vtabptr_7flitter_5model_Matrix44;


/* "model.pxd":178
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */

struct __pyx_vtabstruct_7flitter_5model_Quaternion {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Quaternion *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_euler)(struct __pyx_obj_7flitter_5model_Vector *, double);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_between)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*copy)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*mmul)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *);
  struct __pyx_obj_7flitter_5model_Vector *(*conjugate)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*normalize)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*inverse)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*exponent)(struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*slerp)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Quaternion *__pyx_vtabptr_7flitter_5model_Quaternion;


/* "model.pxd":198
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */

struct __pyx_vtabstruct_7flitter_5model_Node {
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Node *);
  struct __pyx_obj_7flitter_5model_Node *(*copy)(struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*add_tag)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch);
  void (*set_attribute)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*append)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*append_vector)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*_equals)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *);
  PyObject *(*get)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_4Node_get *__pyx_optional_args);
  struct __pyx_obj_7flitter_5model_Vector *(*get_vec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*get_fvec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int, struct __pyx_obj_7flitter_5model_Vector *);
  double (*get_float)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, double);
  int64_t (*get_int)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int64_t);
  int (*get_bool)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
  PyObject *(*get_str)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, PyObject *);
  void (*repr)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
};
static struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtabptr_7flitter_5model_Node;


/* "model.pxd":222
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */

struct __pyx_vtabstruct_7flitter_5model_StateDict {
  struct __pyx_obj_7flitter_5model_Vector *(*get_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
  void (*set_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*contains)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
};
static struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtabptr_7flitter_5model_StateDict;


/* "vm.pxd":14
 * 
 * 
 * cdef class VectorStack:             # <<<<<<<<<<<<<<
 *     cdef PyObject** vectors
 *     cdef int64_t top
 */

struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack {
  struct __pyx_obj_7flitter_8language_2vm_VectorStack *(*copy)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  void (*drop)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_11VectorStack_drop *__pyx_optional_args);
  void (*push)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*pop)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  PyObject *(*pop_tuple)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*pop_list)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*pop_dict)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*pop_composed)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*peek)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*peek_at)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, int __pyx_skip_dispatch);
  void (*poke)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*poke_at)(struct __pyx_obj_7flitter_8language_2vm_VectorStack *, int64_t, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack *__pyx_vtabptr_7flitter_8language_2vm_VectorStack;


/* "vm.pxd":139
 * 
 * 
 * cdef class Program:             # <<<<<<<<<<<<<<
 *     cdef readonly dict pragmas
 *     cdef readonly list instructions
 */

struct __pyx_vtabstruct_7flitter_8language_2vm_Program {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *(*last_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *);
  struct __pyx_obj_7flitter_8language_2vm_Instruction *(*pop_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*push_instruction)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_8language_2vm_Instruction *);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*link)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  void (*use_simplifier)(struct __pyx_obj_7flitter_8language_2vm_Program *, int, int __pyx_skip_dispatch);
  int64_t (*new_label)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*dup)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*drop)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_drop *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*label)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*jump)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*branch_true)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*branch_false)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*import_)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_push)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_load)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*local_drop)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lookup)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lookup_literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*range)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*neg)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*pos)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ceil)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*floor)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*fract)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*not_)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*add)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*sub)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mul)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mul_add)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*truediv)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*floordiv)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*mod)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*pow)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*eq)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ne)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*gt)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*lt)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*ge)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*le)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*xor)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*slice)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*slice_literal)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*call)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_call *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*call_fast)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*tag)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*attributes)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*append)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_append *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*compose)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*begin_for)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*next)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*end_for)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*export)(struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*func)(struct __pyx_obj_7flitter_8language_2vm_Program *, int64_t, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_8language_2vm_7Program_func *__pyx_optional_args);
  struct __pyx_obj_7flitter_8language_2vm_Program *(*exit)(struct __pyx_obj_7flitter_8language_2vm_Program *, int __pyx_skip_dispatch);
  void (*_execute)(struct __pyx_obj_7flitter_8language_2vm_Program *, struct __pyx_obj_7flitter_5model_Context *, int64_t, int);
};
static struct __pyx_vtabstruct_7flitter_8language_2vm_Program *__pyx_vtabptr_7flitter_8language_2vm_Program;


/* "flitter/language/tree.pyx":49
 * 
 * 
 * cdef class Expression:             # <<<<<<<<<<<<<<
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Expression {
  void (*_compile)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *);
  struct __pyx_obj_7flitter_8language_4tree_Expression *(*_simplify)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *);
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *__pyx_vtabptr_7flitter_8language_4tree_Expression;


/* "flitter/language/tree.pyx":92
 * 
 * 
 * cdef class Top(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple pragmas
 *     cdef readonly Expression body
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Top {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Top *__pyx_vtabptr_7flitter_8language_4tree_Top;


/* "flitter/language/tree.pyx":126
 * 
 * 
 * cdef class Export(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly dict static_exports
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Export {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Export *__pyx_vtabptr_7flitter_8language_4tree_Export;


/* "flitter/language/tree.pyx":166
 * 
 * 
 * cdef class Import(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple names
 *     cdef readonly Expression filename
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Import {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Import *__pyx_vtabptr_7flitter_8language_4tree_Import;


/* "flitter/language/tree.pyx":234
 * 
 * 
 * cdef class Sequence(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple expressions
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Sequence {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Sequence *__pyx_vtabptr_7flitter_8language_4tree_Sequence;


/* "flitter/language/tree.pyx":276
 * 
 * 
 * cdef class Literal(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Vector value
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Literal {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Literal *__pyx_vtabptr_7flitter_8language_4tree_Literal;


/* "flitter/language/tree.pyx":292
 * 
 * 
 * cdef class Name(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Name {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Name *__pyx_vtabptr_7flitter_8language_4tree_Name;


/* "flitter/language/tree.pyx":338
 * 
 * 
 * cdef class Lookup(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression key
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Lookup {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Lookup *__pyx_vtabptr_7flitter_8language_4tree_Lookup;


/* "flitter/language/tree.pyx":366
 * 
 * 
 * cdef class Range(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression start
 *     cdef readonly Expression stop
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Range {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Range *__pyx_vtabptr_7flitter_8language_4tree_Range;


/* "flitter/language/tree.pyx":399
 * 
 * 
 * cdef class UnaryOperation(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
  void (*_compile_op)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *);
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;


/* "flitter/language/tree.pyx":416
 * 
 * 
 * cdef class Negative(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.neg()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Negative {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Negative *__pyx_vtabptr_7flitter_8language_4tree_Negative;


/* "flitter/language/tree.pyx":451
 * 
 * 
 * cdef class Positive(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.pos()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Positive {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Positive *__pyx_vtabptr_7flitter_8language_4tree_Positive;


/* "flitter/language/tree.pyx":466
 * 
 * 
 * cdef class Not(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.not_()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Not {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Not *__pyx_vtabptr_7flitter_8language_4tree_Not;


/* "flitter/language/tree.pyx":479
 * 
 * 
 * cdef class Ceil(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.ceil()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Ceil {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Ceil *__pyx_vtabptr_7flitter_8language_4tree_Ceil;


/* "flitter/language/tree.pyx":492
 * 
 * 
 * cdef class Floor(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.floor()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Floor {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Floor *__pyx_vtabptr_7flitter_8language_4tree_Floor;


/* "flitter/language/tree.pyx":505
 * 
 * 
 * cdef class Fract(UnaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.fract()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Fract {
  struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Fract *__pyx_vtabptr_7flitter_8language_4tree_Fract;


/* "flitter/language/tree.pyx":518
 * 
 * 
 * cdef class BinaryOperation(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression left
 *     cdef readonly Expression right
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
  void (*_compile_op)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *);
  struct __pyx_obj_7flitter_5model_Vector *(*op)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_8language_4tree_Expression *(*constant_left)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_8language_4tree_Expression *);
  struct __pyx_obj_7flitter_8language_4tree_Expression *(*constant_right)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_8language_4tree_Expression *(*additional_rules)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_4tree_Expression *);
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;


/* "flitter/language/tree.pyx":570
 * 
 * 
 * cdef class MathsBinaryOperation(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation {
  struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;


/* "flitter/language/tree.pyx":581
 * 
 * 
 * cdef class Add(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.add(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Add {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Add *__pyx_vtabptr_7flitter_8language_4tree_Add;


/* "flitter/language/tree.pyx":608
 * 
 * 
 * cdef class Subtract(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.sub(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Subtract {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Subtract *__pyx_vtabptr_7flitter_8language_4tree_Subtract;


/* "flitter/language/tree.pyx":629
 * 
 * 
 * cdef class Multiply(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mul(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Multiply {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Multiply *__pyx_vtabptr_7flitter_8language_4tree_Multiply;


/* "flitter/language/tree.pyx":663
 * 
 * 
 * cdef class Divide(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.truediv(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Divide {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Divide *__pyx_vtabptr_7flitter_8language_4tree_Divide;


/* "flitter/language/tree.pyx":676
 * 
 * 
 * cdef class FloorDivide(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.floordiv(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_FloorDivide {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_FloorDivide *__pyx_vtabptr_7flitter_8language_4tree_FloorDivide;


/* "flitter/language/tree.pyx":688
 * 
 * 
 * cdef class Modulo(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mod(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Modulo {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Modulo *__pyx_vtabptr_7flitter_8language_4tree_Modulo;


/* "flitter/language/tree.pyx":700
 * 
 * 
 * cdef class Power(MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.pow(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Power {
  struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Power *__pyx_vtabptr_7flitter_8language_4tree_Power;


/* "flitter/language/tree.pyx":718
 * 
 * 
 * cdef class Comparison(BinaryOperation):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison {
  struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison *__pyx_vtabptr_7flitter_8language_4tree_Comparison;


/* "flitter/language/tree.pyx":722
 * 
 * 
 * cdef class EqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.eq(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_EqualTo {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_EqualTo *__pyx_vtabptr_7flitter_8language_4tree_EqualTo;


/* "flitter/language/tree.pyx":730
 * 
 * 
 * cdef class NotEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ne(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_NotEqualTo {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_NotEqualTo *__pyx_vtabptr_7flitter_8language_4tree_NotEqualTo;


/* "flitter/language/tree.pyx":738
 * 
 * 
 * cdef class LessThan(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.lt(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_LessThan {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_LessThan *__pyx_vtabptr_7flitter_8language_4tree_LessThan;


/* "flitter/language/tree.pyx":746
 * 
 * 
 * cdef class GreaterThan(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.gt(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThan {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThan *__pyx_vtabptr_7flitter_8language_4tree_GreaterThan;


/* "flitter/language/tree.pyx":754
 * 
 * 
 * cdef class LessThanOrEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.le(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_LessThanOrEqualTo {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_vtabptr_7flitter_8language_4tree_LessThanOrEqualTo;


/* "flitter/language/tree.pyx":762
 * 
 * 
 * cdef class GreaterThanOrEqualTo(Comparison):             # <<<<<<<<<<<<<<
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ge(right)
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThanOrEqualTo {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_vtabptr_7flitter_8language_4tree_GreaterThanOrEqualTo;


/* "flitter/language/tree.pyx":770
 * 
 * 
 * cdef class And(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_And {
  struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_And *__pyx_vtabptr_7flitter_8language_4tree_And;


/* "flitter/language/tree.pyx":792
 * 
 * 
 * cdef class Or(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Or {
  struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Or *__pyx_vtabptr_7flitter_8language_4tree_Or;


/* "flitter/language/tree.pyx":814
 * 
 * 
 * cdef class Xor(BinaryOperation):             # <<<<<<<<<<<<<<
 *     cdef void _compile_op(self, Program program):
 *         program.xor()
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Xor {
  struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Xor *__pyx_vtabptr_7flitter_8language_4tree_Xor;


/* "flitter/language/tree.pyx":834
 * 
 * 
 * cdef class Slice(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 *     cdef readonly Expression index
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Slice {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Slice *__pyx_vtabptr_7flitter_8language_4tree_Slice;


/* "flitter/language/tree.pyx":867
 * 
 * 
 * cdef class Call(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression function
 *     cdef readonly tuple args
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Call {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Call *__pyx_vtabptr_7flitter_8language_4tree_Call;


/* "flitter/language/tree.pyx":975
 * 
 * 
 * cdef class NodeModifier(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression node
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier *__pyx_vtabptr_7flitter_8language_4tree_NodeModifier;


/* "flitter/language/tree.pyx":979
 * 
 * 
 * cdef class Tag(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly str tag
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Tag {
  struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Tag *__pyx_vtabptr_7flitter_8language_4tree_Tag;


/* "flitter/language/tree.pyx":1018
 * 
 * 
 * cdef class Attributes(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple bindings
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Attributes {
  struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Attributes *__pyx_vtabptr_7flitter_8language_4tree_Attributes;


/* "flitter/language/tree.pyx":1079
 * 
 * 
 * cdef class Append(NodeModifier):             # <<<<<<<<<<<<<<
 *     cdef readonly Expression children
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Append {
  struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Append *__pyx_vtabptr_7flitter_8language_4tree_Append;


/* "flitter/language/tree.pyx":1159
 * 
 * 
 * cdef class Let(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple bindings
 *     cdef readonly Expression body
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Let {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Let *__pyx_vtabptr_7flitter_8language_4tree_Let;


/* "flitter/language/tree.pyx":1275
 * 
 * 
 * cdef class For(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple names
 *     cdef readonly Expression source
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_For {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_For *__pyx_vtabptr_7flitter_8language_4tree_For;


/* "flitter/language/tree.pyx":1347
 * 
 * 
 * cdef class IfElse(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly tuple tests
 *     cdef readonly Expression else_
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_IfElse {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_IfElse *__pyx_vtabptr_7flitter_8language_4tree_IfElse;


/* "flitter/language/tree.pyx":1403
 * 
 * 
 * cdef class Function(Expression):             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 */

struct __pyx_vtabstruct_7flitter_8language_4tree_Function {
  struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_4tree_Function *__pyx_vtabptr_7flitter_8language_4tree_Function;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* py_set_discard.proto */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key);

/* CBIntToPyUnicode.proto */
#define __Pyx_PyUnicode_FromBInt_bint(value)\
    ((value) ? __Pyx_NewRef(__pyx_n_u_True) : __Pyx_NewRef(__pyx_n_u_False))

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static void __pyx_f_7flitter_8language_4tree_10Expression__compile(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, CYTHON_UNUSED PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_10Expression__simplify(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Top__compile(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Top__simplify(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Export__compile(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Export__simplify(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Import__compile(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Import__simplify(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Sequence__compile(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Sequence__simplify(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_7Literal__compile(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, CYTHON_UNUSED PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_7Literal__simplify(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_4Name__compile(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Name__simplify(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Lookup__compile(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Lookup__simplify(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_5Range__compile(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Range__simplify(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_14UnaryOperation__compile(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static void __pyx_f_7flitter_8language_4tree_14UnaryOperation__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Negative__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Negative__simplify(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Positive__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Positive__simplify(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Not__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Not__simplify(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_4Ceil__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Ceil__simplify(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_5Floor__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Floor__simplify(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_5Fract__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Fract__simplify(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_15BinaryOperation__compile(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static void __pyx_f_7flitter_8language_4tree_15BinaryOperation__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation__simplify(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_15BinaryOperation_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_20MathsBinaryOperation__simplify(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_3Add_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Add__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_constant_right(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8Subtract_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Subtract__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8Multiply_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Multiply__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Multiply_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Multiply_constant_right(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_6Divide_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Divide__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Divide_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_11FloorDivide_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_11FloorDivide__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_11FloorDivide_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_6Modulo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Modulo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Modulo_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_5Power_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_5Power__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Power_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_7EqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_7EqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_10NotEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_10NotEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8LessThan_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8LessThan__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_11GreaterThan_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_11GreaterThan__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right); /* proto*/
static void __pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3And__compile(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3And__simplify(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_2Or__compile(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_2Or__simplify(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Xor__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Xor__simplify(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_5Slice__compile(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Slice__simplify(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_4Call__compile(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Call__simplify(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Tag__compile(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Tag__simplify(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_10Attributes__compile(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_10Attributes__simplify(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6Append__compile(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Append__simplify(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3Let__compile(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Let__simplify(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_3For__compile(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3For__simplify(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_6IfElse__compile(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6IfElse__simplify(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/
static void __pyx_f_7flitter_8language_4tree_8Function__compile(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames); /* proto*/
static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Function__simplify(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "cython" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "flitter.model" */
static uint64_t *__pyx_vp_7flitter_5model_HASH_START = 0;
#define __pyx_v_7flitter_5model_HASH_START (*__pyx_vp_7flitter_5model_HASH_START)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_null_ = 0;
#define __pyx_v_7flitter_5model_null_ (*__pyx_vp_7flitter_5model_null_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_true_ = 0;
#define __pyx_v_7flitter_5model_true_ (*__pyx_vp_7flitter_5model_true_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_false_ = 0;
#define __pyx_v_7flitter_5model_false_ (*__pyx_vp_7flitter_5model_false_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_minusone_ = 0;
#define __pyx_v_7flitter_5model_minusone_ (*__pyx_vp_7flitter_5model_minusone_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_inf_ = 0;
#define __pyx_v_7flitter_5model_inf_ (*__pyx_vp_7flitter_5model_inf_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_nan_ = 0;
#define __pyx_v_7flitter_5model_nan_ (*__pyx_vp_7flitter_5model_nan_)

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "flitter.language.vm" */
static PyObject **__pyx_vp_7flitter_8language_2vm_static_builtins = 0;
#define __pyx_v_7flitter_8language_2vm_static_builtins (*__pyx_vp_7flitter_8language_2vm_static_builtins)
static PyObject **__pyx_vp_7flitter_8language_2vm_dynamic_builtins = 0;
#define __pyx_v_7flitter_8language_2vm_dynamic_builtins (*__pyx_vp_7flitter_8language_2vm_dynamic_builtins)
static PyObject **__pyx_vp_7flitter_8language_2vm_builtins = 0;
#define __pyx_v_7flitter_8language_2vm_builtins (*__pyx_vp_7flitter_8language_2vm_builtins)

/* Module declarations from "flitter.language.tree" */
static struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_7flitter_8language_4tree_NoOp = 0;
static int __pyx_f_7flitter_8language_4tree_sequence_pack(PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Expression__set_state(struct __pyx_obj_7flitter_8language_4tree_Expression *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Top__set_state(struct __pyx_obj_7flitter_8language_4tree_Top *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Export__set_state(struct __pyx_obj_7flitter_8language_4tree_Export *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Import__set_state(struct __pyx_obj_7flitter_8language_4tree_Import *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Sequence__set_state(struct __pyx_obj_7flitter_8language_4tree_Sequence *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Literal__set_state(struct __pyx_obj_7flitter_8language_4tree_Literal *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Name__set_state(struct __pyx_obj_7flitter_8language_4tree_Name *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Lookup__set_state(struct __pyx_obj_7flitter_8language_4tree_Lookup *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Range__set_state(struct __pyx_obj_7flitter_8language_4tree_Range *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_UnaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Negative__set_state(struct __pyx_obj_7flitter_8language_4tree_Negative *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Positive__set_state(struct __pyx_obj_7flitter_8language_4tree_Positive *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Not__set_state(struct __pyx_obj_7flitter_8language_4tree_Not *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Ceil__set_state(struct __pyx_obj_7flitter_8language_4tree_Ceil *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Floor__set_state(struct __pyx_obj_7flitter_8language_4tree_Floor *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Fract__set_state(struct __pyx_obj_7flitter_8language_4tree_Fract *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_BinaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_MathsBinaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Add__set_state(struct __pyx_obj_7flitter_8language_4tree_Add *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Subtract__set_state(struct __pyx_obj_7flitter_8language_4tree_Subtract *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Multiply__set_state(struct __pyx_obj_7flitter_8language_4tree_Multiply *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Divide__set_state(struct __pyx_obj_7flitter_8language_4tree_Divide *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_FloorDivide__set_state(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Modulo__set_state(struct __pyx_obj_7flitter_8language_4tree_Modulo *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Power__set_state(struct __pyx_obj_7flitter_8language_4tree_Power *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Comparison__set_state(struct __pyx_obj_7flitter_8language_4tree_Comparison *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_EqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_EqualTo *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_NotEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThan__set_state(struct __pyx_obj_7flitter_8language_4tree_LessThan *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThan__set_state(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThanOrEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThanOrEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_And__set_state(struct __pyx_obj_7flitter_8language_4tree_And *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Or__set_state(struct __pyx_obj_7flitter_8language_4tree_Or *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Xor__set_state(struct __pyx_obj_7flitter_8language_4tree_Xor *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Slice__set_state(struct __pyx_obj_7flitter_8language_4tree_Slice *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Call__set_state(struct __pyx_obj_7flitter_8language_4tree_Call *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_NodeModifier__set_state(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Tag__set_state(struct __pyx_obj_7flitter_8language_4tree_Tag *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Attributes__set_state(struct __pyx_obj_7flitter_8language_4tree_Attributes *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Append__set_state(struct __pyx_obj_7flitter_8language_4tree_Append *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Binding__set_state(struct __pyx_obj_7flitter_8language_4tree_Binding *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_PolyBinding__set_state(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Let__set_state(struct __pyx_obj_7flitter_8language_4tree_Let *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_For__set_state(struct __pyx_obj_7flitter_8language_4tree_For *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_IfCondition__set_state(struct __pyx_obj_7flitter_8language_4tree_IfCondition *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_IfElse__set_state(struct __pyx_obj_7flitter_8language_4tree_IfElse *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Function__set_state(struct __pyx_obj_7flitter_8language_4tree_Function *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "flitter.language.tree"
extern int __pyx_module_is_main_flitter__language__tree;
int __pyx_module_is_main_flitter__language__tree = 0;

/* Implementation of "flitter.language.tree" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_reversed;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_AssertionError;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = ", ";
static const char __pyx_k_Or[] = "Or";
static const char __pyx_k__2[] = ")";
static const char __pyx_k__3[] = "'";
static const char __pyx_k__4[] = "(";
static const char __pyx_k__6[] = ": ";
static const char __pyx_k__8[] = ".";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_Add[] = "Add";
static const char __pyx_k_And[] = "And";
static const char __pyx_k_For[] = "For(";
static const char __pyx_k_Let[] = "Let(";
static const char __pyx_k_Not[] = "Not";
static const char __pyx_k_Tag[] = "Tag(";
static const char __pyx_k_Top[] = "Top(";
static const char __pyx_k_Xor[] = "Xor";
static const char __pyx_k__32[] = "";
static const char __pyx_k_exc[] = "exc";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_opt[] = "opt";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_tag[] = "tag";
static const char __pyx_k_Call[] = "Call(";
static const char __pyx_k_Ceil[] = "Ceil";
static const char __pyx_k_Name[] = "Name(";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_True[] = "True";
static const char __pyx_k__186[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_body[] = "body";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_else[] = "else_";
static const char __pyx_k_expr[] = "expr";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_then[] = "then";
static const char __pyx_k_False[] = "False";
static const char __pyx_k_Floor[] = "Floor";
static const char __pyx_k_For_2[] = "For";
static const char __pyx_k_Fract[] = "Fract";
static const char __pyx_k_Let_2[] = "Let";
static const char __pyx_k_Power[] = "Power";
static const char __pyx_k_Range[] = "Range(";
static const char __pyx_k_Slice[] = "Slice(";
static const char __pyx_k_Tag_2[] = "Tag";
static const char __pyx_k_Top_2[] = "Top";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_fract[] = "fract";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_tests[] = "tests";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_Append[] = "Append";
static const char __pyx_k_Call_2[] = "Call";
static const char __pyx_k_Divide[] = "Divide";
static const char __pyx_k_Export[] = "Export(";
static const char __pyx_k_IfElse[] = "IfElse(";
static const char __pyx_k_Import[] = "Import(";
static const char __pyx_k_Lookup[] = "Lookup(";
static const char __pyx_k_Modulo[] = "Modulo";
static const char __pyx_k_Name_2[] = "Name";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_loguru[] = "loguru";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_static[] = "static";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_Binding[] = "Binding(";
static const char __pyx_k_EqualTo[] = "EqualTo";
static const char __pyx_k_Literal[] = "Literal(";
static const char __pyx_k_Range_2[] = "Range";
static const char __pyx_k_Slice_2[] = "Slice";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_dynamic[] = "dynamic";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_pragmas[] = "pragmas";
static const char __pyx_k_program[] = "program";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Export_2[] = "Export";
static const char __pyx_k_Function[] = "Function(";
static const char __pyx_k_IfElse_2[] = "IfElse";
static const char __pyx_k_Import_2[] = "Import";
static const char __pyx_k_LessThan[] = "LessThan";
static const char __pyx_k_Lookup_2[] = "Lookup";
static const char __pyx_k_Multiply[] = "Multiply";
static const char __pyx_k_Negative[] = "Negative";
static const char __pyx_k_Positive[] = "Positive";
static const char __pyx_k_Sequence[] = "Sequence(";
static const char __pyx_k_Subtract[] = "Subtract";
static const char __pyx_k_bindings[] = "bindings";
static const char __pyx_k_captures[] = "captures";
static const char __pyx_k_children[] = "children";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_function[] = "function";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_reversed[] = "reversed";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_simplify[] = "simplify";
static const char __pyx_k_Binding_2[] = "Binding";
static const char __pyx_k_Literal_2[] = "Literal";
static const char __pyx_k_condition[] = "condition";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_recursive[] = "recursive";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_Attributes[] = "Attributes";
static const char __pyx_k_Comparison[] = "Comparison";
static const char __pyx_k_Expression[] = "Expression";
static const char __pyx_k_Function_2[] = "Function";
static const char __pyx_k_NotEqualTo[] = "NotEqualTo";
static const char __pyx_k_Sequence_2[] = "Sequence";
static const char __pyx_k_inlineable[] = "inlineable";
static const char __pyx_k_log_errors[] = "log_errors";
static const char __pyx_k_name_patch[] = "name_patch";
static const char __pyx_k_parameters[] = "parameters";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_set_pragma[] = "set_pragma";
static const char __pyx_k_FloorDivide[] = "FloorDivide";
static const char __pyx_k_GreaterThan[] = "GreaterThan";
static const char __pyx_k_IfCondition[] = "IfCondition(";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_PolyBinding[] = "PolyBinding(";
static const char __pyx_k_SharedCache[] = "SharedCache";
static const char __pyx_k_expressions[] = "expressions";
static const char __pyx_k_NodeModifier[] = "NodeModifier";
static const char __pyx_k_Unbound_name[] = "Unbound name '";
static const char __pyx_k_context_func[] = "context_func";
static const char __pyx_k_context_vars[] = "context_vars";
static const char __pyx_k_dependencies[] = "dependencies";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_keyword_args[] = "keyword_args";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_Error_calling[] = "Error calling ";
static const char __pyx_k_IfCondition_2[] = "IfCondition";
static const char __pyx_k_PolyBinding_2[] = "PolyBinding";
static const char __pyx_k_get_with_root[] = "get_with_root";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_Compiler_error[] = "Compiler error: {}";
static const char __pyx_k_UnaryOperation[] = "UnaryOperation";
static const char __pyx_k_initial_errors[] = "initial_errors";
static const char __pyx_k_initial_lnames[] = "initial_lnames";
static const char __pyx_k_return_context[] = "return_context";
static const char __pyx_k_static_exports[] = "static_exports";
static const char __pyx_k_BinaryOperation[] = "BinaryOperation";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_pyx_unpickle_Or[] = "__pyx_unpickle_Or";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_Simplifier_error[] = "Simplifier error: {}";
static const char __pyx_k_pyx_unpickle_Add[] = "__pyx_unpickle_Add";
static const char __pyx_k_pyx_unpickle_And[] = "__pyx_unpickle_And";
static const char __pyx_k_pyx_unpickle_For[] = "__pyx_unpickle_For";
static const char __pyx_k_pyx_unpickle_Let[] = "__pyx_unpickle_Let";
static const char __pyx_k_pyx_unpickle_Not[] = "__pyx_unpickle_Not";
static const char __pyx_k_pyx_unpickle_Tag[] = "__pyx_unpickle_Tag";
static const char __pyx_k_pyx_unpickle_Top[] = "__pyx_unpickle_Top";
static const char __pyx_k_pyx_unpickle_Xor[] = "__pyx_unpickle_Xor";
static const char __pyx_k_read_flitter_top[] = "read_flitter_top";
static const char __pyx_k_LessThanOrEqualTo[] = "LessThanOrEqualTo";
static const char __pyx_k_difference_update[] = "difference_update";
static const char __pyx_k_pyx_unpickle_Call[] = "__pyx_unpickle_Call";
static const char __pyx_k_pyx_unpickle_Ceil[] = "__pyx_unpickle_Ceil";
static const char __pyx_k_pyx_unpickle_Name[] = "__pyx_unpickle_Name";
static const char __pyx_k_Circular_import_of[] = "Circular import of '";
static const char __pyx_k_Expression_compile[] = "Expression.compile";
static const char __pyx_k_Or___reduce_cython[] = "Or.__reduce_cython__";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_pyx_unpickle_Floor[] = "__pyx_unpickle_Floor";
static const char __pyx_k_pyx_unpickle_Fract[] = "__pyx_unpickle_Fract";
static const char __pyx_k_pyx_unpickle_Power[] = "__pyx_unpickle_Power";
static const char __pyx_k_pyx_unpickle_Range[] = "__pyx_unpickle_Range";
static const char __pyx_k_pyx_unpickle_Slice[] = "__pyx_unpickle_Slice";
static const char __pyx_k_Add___reduce_cython[] = "Add.__reduce_cython__";
static const char __pyx_k_And___reduce_cython[] = "And.__reduce_cython__";
static const char __pyx_k_Expression_simplify[] = "Expression.simplify";
static const char __pyx_k_For___reduce_cython[] = "For.__reduce_cython__";
static const char __pyx_k_Let___reduce_cython[] = "Let.__reduce_cython__";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_Not___reduce_cython[] = "Not.__reduce_cython__";
static const char __pyx_k_Tag___reduce_cython[] = "Tag.__reduce_cython__";
static const char __pyx_k_Top___reduce_cython[] = "Top.__reduce_cython__";
static const char __pyx_k_Xor___reduce_cython[] = "Xor.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_Append[] = "__pyx_unpickle_Append";
static const char __pyx_k_pyx_unpickle_Divide[] = "__pyx_unpickle_Divide";
static const char __pyx_k_pyx_unpickle_Export[] = "__pyx_unpickle_Export";
static const char __pyx_k_pyx_unpickle_IfElse[] = "__pyx_unpickle_IfElse";
static const char __pyx_k_pyx_unpickle_Import[] = "__pyx_unpickle_Import";
static const char __pyx_k_pyx_unpickle_Lookup[] = "__pyx_unpickle_Lookup";
static const char __pyx_k_pyx_unpickle_Modulo[] = "__pyx_unpickle_Modulo";
static const char __pyx_k_Call___reduce_cython[] = "Call.__reduce_cython__";
static const char __pyx_k_Ceil___reduce_cython[] = "Ceil.__reduce_cython__";
static const char __pyx_k_GreaterThanOrEqualTo[] = "GreaterThanOrEqualTo";
static const char __pyx_k_MathsBinaryOperation[] = "MathsBinaryOperation";
static const char __pyx_k_Name___reduce_cython[] = "Name.__reduce_cython__";
static const char __pyx_k_Or___setstate_cython[] = "Or.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Binding[] = "__pyx_unpickle_Binding";
static const char __pyx_k_pyx_unpickle_EqualTo[] = "__pyx_unpickle_EqualTo";
static const char __pyx_k_pyx_unpickle_Literal[] = "__pyx_unpickle_Literal";
static const char __pyx_k_Add___setstate_cython[] = "Add.__setstate_cython__";
static const char __pyx_k_And___setstate_cython[] = "And.__setstate_cython__";
static const char __pyx_k_Floor___reduce_cython[] = "Floor.__reduce_cython__";
static const char __pyx_k_For___setstate_cython[] = "For.__setstate_cython__";
static const char __pyx_k_Fract___reduce_cython[] = "Fract.__reduce_cython__";
static const char __pyx_k_Let___setstate_cython[] = "Let.__setstate_cython__";
static const char __pyx_k_Not___setstate_cython[] = "Not.__setstate_cython__";
static const char __pyx_k_Power___reduce_cython[] = "Power.__reduce_cython__";
static const char __pyx_k_Range___reduce_cython[] = "Range.__reduce_cython__";
static const char __pyx_k_Slice___reduce_cython[] = "Slice.__reduce_cython__";
static const char __pyx_k_Tag___setstate_cython[] = "Tag.__setstate_cython__";
static const char __pyx_k_Top___setstate_cython[] = "Top.__setstate_cython__";
static const char __pyx_k_Xor___setstate_cython[] = "Xor.__setstate_cython__";
static const char __pyx_k_flitter_language_tree[] = "flitter.language.tree";
static const char __pyx_k_pyx_unpickle_Function[] = "__pyx_unpickle_Function";
static const char __pyx_k_pyx_unpickle_LessThan[] = "__pyx_unpickle_LessThan";
static const char __pyx_k_pyx_unpickle_Multiply[] = "__pyx_unpickle_Multiply";
static const char __pyx_k_pyx_unpickle_Negative[] = "__pyx_unpickle_Negative";
static const char __pyx_k_pyx_unpickle_Positive[] = "__pyx_unpickle_Positive";
static const char __pyx_k_pyx_unpickle_Sequence[] = "__pyx_unpickle_Sequence";
static const char __pyx_k_pyx_unpickle_Subtract[] = "__pyx_unpickle_Subtract";
static const char __pyx_k_Append___reduce_cython[] = "Append.__reduce_cython__";
static const char __pyx_k_Call___setstate_cython[] = "Call.__setstate_cython__";
static const char __pyx_k_Ceil___setstate_cython[] = "Ceil.__setstate_cython__";
static const char __pyx_k_Divide___reduce_cython[] = "Divide.__reduce_cython__";
static const char __pyx_k_Export___reduce_cython[] = "Export.__reduce_cython__";
static const char __pyx_k_IfElse___reduce_cython[] = "IfElse.__reduce_cython__";
static const char __pyx_k_Import___reduce_cython[] = "Import.__reduce_cython__";
static const char __pyx_k_Lookup___reduce_cython[] = "Lookup.__reduce_cython__";
static const char __pyx_k_Modulo___reduce_cython[] = "Modulo.__reduce_cython__";
static const char __pyx_k_Name___setstate_cython[] = "Name.__setstate_cython__";
static const char __pyx_k_Binding___reduce_cython[] = "Binding.__reduce_cython__";
static const char __pyx_k_EqualTo___reduce_cython[] = "EqualTo.__reduce_cython__";
static const char __pyx_k_Floor___setstate_cython[] = "Floor.__setstate_cython__";
static const char __pyx_k_Fract___setstate_cython[] = "Fract.__setstate_cython__";
static const char __pyx_k_Literal___reduce_cython[] = "Literal.__reduce_cython__";
static const char __pyx_k_Power___setstate_cython[] = "Power.__setstate_cython__";
static const char __pyx_k_Range___setstate_cython[] = "Range.__setstate_cython__";
static const char __pyx_k_Slice___setstate_cython[] = "Slice.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Attributes[] = "__pyx_unpickle_Attributes";
static const char __pyx_k_pyx_unpickle_Comparison[] = "__pyx_unpickle_Comparison";
static const char __pyx_k_pyx_unpickle_Expression[] = "__pyx_unpickle_Expression";
static const char __pyx_k_pyx_unpickle_NotEqualTo[] = "__pyx_unpickle_NotEqualTo";
static const char __pyx_k_Append___setstate_cython[] = "Append.__setstate_cython__";
static const char __pyx_k_Divide___setstate_cython[] = "Divide.__setstate_cython__";
static const char __pyx_k_Export___setstate_cython[] = "Export.__setstate_cython__";
static const char __pyx_k_Function___reduce_cython[] = "Function.__reduce_cython__";
static const char __pyx_k_IfElse___setstate_cython[] = "IfElse.__setstate_cython__";
static const char __pyx_k_Import___setstate_cython[] = "Import.__setstate_cython__";
static const char __pyx_k_LessThan___reduce_cython[] = "LessThan.__reduce_cython__";
static const char __pyx_k_Lookup___setstate_cython[] = "Lookup.__setstate_cython__";
static const char __pyx_k_Modulo___setstate_cython[] = "Modulo.__setstate_cython__";
static const char __pyx_k_Multiply___reduce_cython[] = "Multiply.__reduce_cython__";
static const char __pyx_k_Negative___reduce_cython[] = "Negative.__reduce_cython__";
static const char __pyx_k_Positive___reduce_cython[] = "Positive.__reduce_cython__";
static const char __pyx_k_Sequence___reduce_cython[] = "Sequence.__reduce_cython__";
static const char __pyx_k_Subtract___reduce_cython[] = "Subtract.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_FloorDivide[] = "__pyx_unpickle_FloorDivide";
static const char __pyx_k_pyx_unpickle_GreaterThan[] = "__pyx_unpickle_GreaterThan";
static const char __pyx_k_pyx_unpickle_IfCondition[] = "__pyx_unpickle_IfCondition";
static const char __pyx_k_pyx_unpickle_PolyBinding[] = "__pyx_unpickle_PolyBinding";
static const char __pyx_k_Binding___setstate_cython[] = "Binding.__setstate_cython__";
static const char __pyx_k_EqualTo___setstate_cython[] = "EqualTo.__setstate_cython__";
static const char __pyx_k_Literal___setstate_cython[] = "Literal.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_NodeModifier[] = "__pyx_unpickle_NodeModifier";
static const char __pyx_k_Attributes___reduce_cython[] = "Attributes.__reduce_cython__";
static const char __pyx_k_Comparison___reduce_cython[] = "Comparison.__reduce_cython__";
static const char __pyx_k_Expression___reduce_cython[] = "Expression.__reduce_cython__";
static const char __pyx_k_Function___setstate_cython[] = "Function.__setstate_cython__";
static const char __pyx_k_LessThan___setstate_cython[] = "LessThan.__setstate_cython__";
static const char __pyx_k_Multiply___setstate_cython[] = "Multiply.__setstate_cython__";
static const char __pyx_k_Negative___setstate_cython[] = "Negative.__setstate_cython__";
static const char __pyx_k_NotEqualTo___reduce_cython[] = "NotEqualTo.__reduce_cython__";
static const char __pyx_k_Positive___setstate_cython[] = "Positive.__setstate_cython__";
static const char __pyx_k_Sequence___setstate_cython[] = "Sequence.__setstate_cython__";
static const char __pyx_k_Subtract___setstate_cython[] = "Subtract.__setstate_cython__";
static const char __pyx_k_Unable_to_simplify_program[] = "Unable to simplify program";
static const char __pyx_k_FloorDivide___reduce_cython[] = "FloorDivide.__reduce_cython__";
static const char __pyx_k_GreaterThan___reduce_cython[] = "GreaterThan.__reduce_cython__";
static const char __pyx_k_IfCondition___reduce_cython[] = "IfCondition.__reduce_cython__";
static const char __pyx_k_PolyBinding___reduce_cython[] = "PolyBinding.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_UnaryOperation[] = "__pyx_unpickle_UnaryOperation";
static const char __pyx_k_Attributes___setstate_cython[] = "Attributes.__setstate_cython__";
static const char __pyx_k_Comparison___setstate_cython[] = "Comparison.__setstate_cython__";
static const char __pyx_k_Expression___setstate_cython[] = "Expression.__setstate_cython__";
static const char __pyx_k_NodeModifier___reduce_cython[] = "NodeModifier.__reduce_cython__";
static const char __pyx_k_NotEqualTo___setstate_cython[] = "NotEqualTo.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_BinaryOperation[] = "__pyx_unpickle_BinaryOperation";
static const char __pyx_k_FloorDivide___setstate_cython[] = "FloorDivide.__setstate_cython__";
static const char __pyx_k_GreaterThan___setstate_cython[] = "GreaterThan.__setstate_cython__";
static const char __pyx_k_IfCondition___setstate_cython[] = "IfCondition.__setstate_cython__";
static const char __pyx_k_PolyBinding___setstate_cython[] = "PolyBinding.__setstate_cython__";
static const char __pyx_k_src_flitter_language_tree_pyx[] = "src/flitter/language/tree.pyx";
static const char __pyx_k_NodeModifier___setstate_cython[] = "NodeModifier.__setstate_cython__";
static const char __pyx_k_UnaryOperation___reduce_cython[] = "UnaryOperation.__reduce_cython__";
static const char __pyx_k_pyx_unpickle_GreaterThanOrEqua[] = "__pyx_unpickle_GreaterThanOrEqualTo";
static const char __pyx_k_pyx_unpickle_LessThanOrEqualTo[] = "__pyx_unpickle_LessThanOrEqualTo";
static const char __pyx_k_pyx_unpickle_MathsBinaryOperat[] = "__pyx_unpickle_MathsBinaryOperation";
static const char __pyx_k_BinaryOperation___reduce_cython[] = "BinaryOperation.__reduce_cython__";
static const char __pyx_k_GreaterThanOrEqualTo___setstate[] = "GreaterThanOrEqualTo.__setstate_cython__";
static const char __pyx_k_Import__simplify_locals_genexpr[] = "Import._simplify.<locals>.genexpr";
static const char __pyx_k_Language_abstract_syntax_tree_T[] = "\nLanguage abstract syntax tree\n\nThe tree supports reasonably sophisticated simplification that will reduce\nthe tree down to a \"simpler\" form by propogating constants. As this can result\nin unrolling loops, simpler does not necessarily mean \"smaller.\"\n";
static const char __pyx_k_MathsBinaryOperation___setstate[] = "MathsBinaryOperation.__setstate_cython__";
static const char __pyx_k_BinaryOperation___setstate_cytho[] = "BinaryOperation.__setstate_cython__";
static const char __pyx_k_GreaterThanOrEqualTo___reduce_cy[] = "GreaterThanOrEqualTo.__reduce_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())";
static const char __pyx_k_LessThanOrEqualTo___reduce_cytho[] = "LessThanOrEqualTo.__reduce_cython__";
static const char __pyx_k_LessThanOrEqualTo___setstate_cyt[] = "LessThanOrEqualTo.__setstate_cython__";
static const char __pyx_k_MathsBinaryOperation___reduce_cy[] = "MathsBinaryOperation.__reduce_cython__";
static const char __pyx_k_UnaryOperation___setstate_cython[] = "UnaryOperation.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_8[] = "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_9[] = "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_10[] = "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_11[] = "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_12[] = "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_13[] = "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_14[] = "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_15[] = "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_16[] = "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_17[] = "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_18[] = "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_19[] = "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_20[] = "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_21[] = "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_22[] = "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_23[] = "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_24[] = "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_compile(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, PyObject *__pyx_v_initial_lnames, PyObject *__pyx_v_initial_errors, int __pyx_v_log_errors); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_2simplify(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state, PyObject *__pyx_v_static, PyObject *__pyx_v_dynamic, struct __pyx_obj_7flitter_5model_Context *__pyx_v_parent, PyObject *__pyx_v_path, int __pyx_v_return_context); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_3Top___init__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, PyObject *__pyx_v_pragmas, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body, PyObject *__pyx_v_dependencies); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_7pragmas___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_12dependencies___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_6Export___init__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, PyObject *__pyx_v_static_exports); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_14static_exports___get__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_6Import___init__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_filename, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_9_simplify_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_5names___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_8filename___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_8Sequence___init__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, PyObject *__pyx_v_expressions); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_11expressions___get__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_7Literal___init__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_5value___get__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_4Name___init__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_6Lookup___init__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_3key___get__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_5Range___init__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_start, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_stop, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_step); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_5start___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4stop___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4step___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_14UnaryOperation___init__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_2__repr__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Negative___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Negative_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Positive___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Positive_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Not___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Not_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Ceil___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Ceil_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Floor___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Floor_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Fract___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Fract_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_15BinaryOperation___init__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_2__repr__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_4left___get__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_5right___get__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Add___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Add_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Subtract___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Subtract_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Multiply___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Multiply_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Divide___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Divide_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11FloorDivide___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11FloorDivide_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Modulo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Modulo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Power___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Power_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Comparison___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Comparison *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Comparison_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Comparison *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7EqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7EqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10NotEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10NotEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8LessThan___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8LessThan_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11GreaterThan___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11GreaterThan_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3And___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3And_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_2Or___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_2Or_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Xor___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Xor_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_5Slice___init__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_5index___get__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_4Call___init__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_function, PyObject *__pyx_v_args, PyObject *__pyx_v_keyword_args); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_8function___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_4args___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_12keyword_args___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier_4node___get__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_3Tag___init__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_3tag___get__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_10Attributes___init__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, PyObject *__pyx_v_bindings); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_8bindings___get__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_6Append___init__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_children); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_8children___get__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_7Binding___init__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_11PolyBinding___init__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_2__repr__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_5names___get__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_3Let___init__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, PyObject *__pyx_v_bindings, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_8bindings___get__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_3For___init__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_source, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_2__repr__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_5names___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_6source___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_4body___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_3For_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_11IfCondition___init__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_condition, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_then); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_2__repr__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_9condition___get__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_4then___get__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_6IfElse___init__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, PyObject *__pyx_v_tests, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_else_); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_2__repr__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_5tests___get__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_5else____get__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_8language_4tree_8Function___init__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body, PyObject *__pyx_v_captures, int __pyx_v_inlineable, int __pyx_v_recursive); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_10parameters___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_8captures___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_10inlineable___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_9recursive___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree___pyx_unpickle_Expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_2__pyx_unpickle_Top(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_4__pyx_unpickle_Export(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_6__pyx_unpickle_Import(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_8__pyx_unpickle_Sequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_10__pyx_unpickle_Literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_12__pyx_unpickle_Name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_14__pyx_unpickle_Lookup(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_16__pyx_unpickle_Range(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_18__pyx_unpickle_UnaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_20__pyx_unpickle_Negative(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_22__pyx_unpickle_Positive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_24__pyx_unpickle_Not(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_26__pyx_unpickle_Ceil(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_28__pyx_unpickle_Floor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_30__pyx_unpickle_Fract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_32__pyx_unpickle_BinaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_34__pyx_unpickle_MathsBinaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_36__pyx_unpickle_Add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_38__pyx_unpickle_Subtract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_40__pyx_unpickle_Multiply(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_42__pyx_unpickle_Divide(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_44__pyx_unpickle_FloorDivide(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_46__pyx_unpickle_Modulo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_48__pyx_unpickle_Power(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_50__pyx_unpickle_Comparison(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_52__pyx_unpickle_EqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_54__pyx_unpickle_NotEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_56__pyx_unpickle_LessThan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_58__pyx_unpickle_GreaterThan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_60__pyx_unpickle_LessThanOrEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_62__pyx_unpickle_GreaterThanOrEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_64__pyx_unpickle_And(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_66__pyx_unpickle_Or(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_68__pyx_unpickle_Xor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_70__pyx_unpickle_Slice(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_72__pyx_unpickle_Call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_74__pyx_unpickle_NodeModifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_76__pyx_unpickle_Tag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_78__pyx_unpickle_Attributes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_80__pyx_unpickle_Append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_82__pyx_unpickle_Binding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_84__pyx_unpickle_PolyBinding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_86__pyx_unpickle_Let(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_88__pyx_unpickle_For(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_90__pyx_unpickle_IfCondition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_92__pyx_unpickle_IfElse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_8language_4tree_94__pyx_unpickle_Function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Expression(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Top(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Export(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Import(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Sequence(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Literal(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Name(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Lookup(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Range(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_UnaryOperation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Negative(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Positive(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Not(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Ceil(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Floor(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Fract(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_BinaryOperation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Add(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Subtract(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Multiply(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Divide(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_FloorDivide(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Modulo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Power(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Comparison(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_EqualTo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_NotEqualTo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_LessThan(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_GreaterThan(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_LessThanOrEqualTo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_GreaterThanOrEqualTo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_And(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Or(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Xor(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Slice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Call(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_NodeModifier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Tag(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Attributes(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Append(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Binding(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_PolyBinding(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Let(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_For(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_IfCondition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_IfElse(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree_Function(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_8language_4tree___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_update = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7flitter_5model_Vector;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix33;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix44;
  PyTypeObject *__pyx_ptype_7flitter_5model_Quaternion;
  PyTypeObject *__pyx_ptype_7flitter_5model_Node;
  PyTypeObject *__pyx_ptype_7flitter_5model_StateDict;
  PyTypeObject *__pyx_ptype_7flitter_5model_Context;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_VectorStack;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_Instruction;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionVector;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionNode;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionStr;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionTuple;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionInt;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionLabel;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionJump;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionFunc;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt;
  PyTypeObject *__pyx_ptype_7flitter_8language_2vm_Program;
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_7flitter_8language_4tree_Expression;
  PyObject *__pyx_type_7flitter_8language_4tree_Top;
  PyObject *__pyx_type_7flitter_8language_4tree_Export;
  PyObject *__pyx_type_7flitter_8language_4tree_Import;
  PyObject *__pyx_type_7flitter_8language_4tree_Sequence;
  PyObject *__pyx_type_7flitter_8language_4tree_Literal;
  PyObject *__pyx_type_7flitter_8language_4tree_Name;
  PyObject *__pyx_type_7flitter_8language_4tree_Lookup;
  PyObject *__pyx_type_7flitter_8language_4tree_Range;
  PyObject *__pyx_type_7flitter_8language_4tree_UnaryOperation;
  PyObject *__pyx_type_7flitter_8language_4tree_Negative;
  PyObject *__pyx_type_7flitter_8language_4tree_Positive;
  PyObject *__pyx_type_7flitter_8language_4tree_Not;
  PyObject *__pyx_type_7flitter_8language_4tree_Ceil;
  PyObject *__pyx_type_7flitter_8language_4tree_Floor;
  PyObject *__pyx_type_7flitter_8language_4tree_Fract;
  PyObject *__pyx_type_7flitter_8language_4tree_BinaryOperation;
  PyObject *__pyx_type_7flitter_8language_4tree_MathsBinaryOperation;
  PyObject *__pyx_type_7flitter_8language_4tree_Add;
  PyObject *__pyx_type_7flitter_8language_4tree_Subtract;
  PyObject *__pyx_type_7flitter_8language_4tree_Multiply;
  PyObject *__pyx_type_7flitter_8language_4tree_Divide;
  PyObject *__pyx_type_7flitter_8language_4tree_FloorDivide;
  PyObject *__pyx_type_7flitter_8language_4tree_Modulo;
  PyObject *__pyx_type_7flitter_8language_4tree_Power;
  PyObject *__pyx_type_7flitter_8language_4tree_Comparison;
  PyObject *__pyx_type_7flitter_8language_4tree_EqualTo;
  PyObject *__pyx_type_7flitter_8language_4tree_NotEqualTo;
  PyObject *__pyx_type_7flitter_8language_4tree_LessThan;
  PyObject *__pyx_type_7flitter_8language_4tree_GreaterThan;
  PyObject *__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo;
  PyObject *__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo;
  PyObject *__pyx_type_7flitter_8language_4tree_And;
  PyObject *__pyx_type_7flitter_8language_4tree_Or;
  PyObject *__pyx_type_7flitter_8language_4tree_Xor;
  PyObject *__pyx_type_7flitter_8language_4tree_Slice;
  PyObject *__pyx_type_7flitter_8language_4tree_Call;
  PyObject *__pyx_type_7flitter_8language_4tree_NodeModifier;
  PyObject *__pyx_type_7flitter_8language_4tree_Tag;
  PyObject *__pyx_type_7flitter_8language_4tree_Attributes;
  PyObject *__pyx_type_7flitter_8language_4tree_Append;
  PyObject *__pyx_type_7flitter_8language_4tree_Binding;
  PyObject *__pyx_type_7flitter_8language_4tree_PolyBinding;
  PyObject *__pyx_type_7flitter_8language_4tree_Let;
  PyObject *__pyx_type_7flitter_8language_4tree_For;
  PyObject *__pyx_type_7flitter_8language_4tree_IfCondition;
  PyObject *__pyx_type_7flitter_8language_4tree_IfElse;
  PyObject *__pyx_type_7flitter_8language_4tree_Function;
  PyObject *__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr;
  #endif
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Expression;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Top;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Export;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Import;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Sequence;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Literal;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Name;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Lookup;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Range;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Negative;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Positive;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Not;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Ceil;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Floor;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Fract;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Add;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Subtract;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Multiply;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Divide;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_FloorDivide;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Modulo;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Power;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Comparison;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_EqualTo;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_NotEqualTo;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_LessThan;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_GreaterThan;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_And;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Or;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Xor;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Slice;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Call;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_NodeModifier;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Tag;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Attributes;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Append;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Binding;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_PolyBinding;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Let;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_For;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_IfCondition;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_IfElse;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree_Function;
  PyTypeObject *__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_n_s_Add;
  PyObject *__pyx_n_s_Add___reduce_cython;
  PyObject *__pyx_n_s_Add___setstate_cython;
  PyObject *__pyx_n_s_And;
  PyObject *__pyx_n_s_And___reduce_cython;
  PyObject *__pyx_n_s_And___setstate_cython;
  PyObject *__pyx_n_s_Append;
  PyObject *__pyx_n_s_Append___reduce_cython;
  PyObject *__pyx_n_s_Append___setstate_cython;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_Attributes;
  PyObject *__pyx_n_s_Attributes___reduce_cython;
  PyObject *__pyx_n_s_Attributes___setstate_cython;
  PyObject *__pyx_n_s_BinaryOperation;
  PyObject *__pyx_n_s_BinaryOperation___reduce_cython;
  PyObject *__pyx_n_s_BinaryOperation___setstate_cytho;
  PyObject *__pyx_kp_u_Binding;
  PyObject *__pyx_n_s_Binding_2;
  PyObject *__pyx_n_s_Binding___reduce_cython;
  PyObject *__pyx_n_s_Binding___setstate_cython;
  PyObject *__pyx_kp_u_Call;
  PyObject *__pyx_n_s_Call_2;
  PyObject *__pyx_n_s_Call___reduce_cython;
  PyObject *__pyx_n_s_Call___setstate_cython;
  PyObject *__pyx_n_s_Ceil;
  PyObject *__pyx_n_s_Ceil___reduce_cython;
  PyObject *__pyx_n_s_Ceil___setstate_cython;
  PyObject *__pyx_kp_u_Circular_import_of;
  PyObject *__pyx_n_s_Comparison;
  PyObject *__pyx_n_s_Comparison___reduce_cython;
  PyObject *__pyx_n_s_Comparison___setstate_cython;
  PyObject *__pyx_kp_u_Compiler_error;
  PyObject *__pyx_n_s_Divide;
  PyObject *__pyx_n_s_Divide___reduce_cython;
  PyObject *__pyx_n_s_Divide___setstate_cython;
  PyObject *__pyx_n_s_EqualTo;
  PyObject *__pyx_n_s_EqualTo___reduce_cython;
  PyObject *__pyx_n_s_EqualTo___setstate_cython;
  PyObject *__pyx_kp_u_Error_calling;
  PyObject *__pyx_kp_u_Export;
  PyObject *__pyx_n_s_Export_2;
  PyObject *__pyx_n_s_Export___reduce_cython;
  PyObject *__pyx_n_s_Export___setstate_cython;
  PyObject *__pyx_n_s_Expression;
  PyObject *__pyx_n_s_Expression___reduce_cython;
  PyObject *__pyx_n_s_Expression___setstate_cython;
  PyObject *__pyx_n_s_Expression_compile;
  PyObject *__pyx_n_s_Expression_simplify;
  PyObject *__pyx_n_u_False;
  PyObject *__pyx_n_s_Floor;
  PyObject *__pyx_n_s_FloorDivide;
  PyObject *__pyx_n_s_FloorDivide___reduce_cython;
  PyObject *__pyx_n_s_FloorDivide___setstate_cython;
  PyObject *__pyx_n_s_Floor___reduce_cython;
  PyObject *__pyx_n_s_Floor___setstate_cython;
  PyObject *__pyx_kp_u_For;
  PyObject *__pyx_n_s_For_2;
  PyObject *__pyx_n_s_For___reduce_cython;
  PyObject *__pyx_n_s_For___setstate_cython;
  PyObject *__pyx_n_s_Fract;
  PyObject *__pyx_n_s_Fract___reduce_cython;
  PyObject *__pyx_n_s_Fract___setstate_cython;
  PyObject *__pyx_kp_u_Function;
  PyObject *__pyx_n_s_Function_2;
  PyObject *__pyx_n_s_Function___reduce_cython;
  PyObject *__pyx_n_s_Function___setstate_cython;
  PyObject *__pyx_n_s_GreaterThan;
  PyObject *__pyx_n_s_GreaterThanOrEqualTo;
  PyObject *__pyx_n_s_GreaterThanOrEqualTo___reduce_cy;
  PyObject *__pyx_n_s_GreaterThanOrEqualTo___setstate;
  PyObject *__pyx_n_s_GreaterThan___reduce_cython;
  PyObject *__pyx_n_s_GreaterThan___setstate_cython;
  PyObject *__pyx_kp_u_IfCondition;
  PyObject *__pyx_n_s_IfCondition_2;
  PyObject *__pyx_n_s_IfCondition___reduce_cython;
  PyObject *__pyx_n_s_IfCondition___setstate_cython;
  PyObject *__pyx_kp_u_IfElse;
  PyObject *__pyx_n_s_IfElse_2;
  PyObject *__pyx_n_s_IfElse___reduce_cython;
  PyObject *__pyx_n_s_IfElse___setstate_cython;
  PyObject *__pyx_kp_u_Import;
  PyObject *__pyx_n_s_Import_2;
  PyObject *__pyx_n_s_Import___reduce_cython;
  PyObject *__pyx_n_s_Import___setstate_cython;
  PyObject *__pyx_n_s_Import__simplify_locals_genexpr;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9;
  PyObject *__pyx_n_s_LessThan;
  PyObject *__pyx_n_s_LessThanOrEqualTo;
  PyObject *__pyx_n_s_LessThanOrEqualTo___reduce_cytho;
  PyObject *__pyx_n_s_LessThanOrEqualTo___setstate_cyt;
  PyObject *__pyx_n_s_LessThan___reduce_cython;
  PyObject *__pyx_n_s_LessThan___setstate_cython;
  PyObject *__pyx_kp_u_Let;
  PyObject *__pyx_n_s_Let_2;
  PyObject *__pyx_n_s_Let___reduce_cython;
  PyObject *__pyx_n_s_Let___setstate_cython;
  PyObject *__pyx_kp_u_Literal;
  PyObject *__pyx_n_s_Literal_2;
  PyObject *__pyx_n_s_Literal___reduce_cython;
  PyObject *__pyx_n_s_Literal___setstate_cython;
  PyObject *__pyx_kp_u_Lookup;
  PyObject *__pyx_n_s_Lookup_2;
  PyObject *__pyx_n_s_Lookup___reduce_cython;
  PyObject *__pyx_n_s_Lookup___setstate_cython;
  PyObject *__pyx_n_s_MathsBinaryOperation;
  PyObject *__pyx_n_s_MathsBinaryOperation___reduce_cy;
  PyObject *__pyx_n_s_MathsBinaryOperation___setstate;
  PyObject *__pyx_n_s_Modulo;
  PyObject *__pyx_n_s_Modulo___reduce_cython;
  PyObject *__pyx_n_s_Modulo___setstate_cython;
  PyObject *__pyx_n_s_Multiply;
  PyObject *__pyx_n_s_Multiply___reduce_cython;
  PyObject *__pyx_n_s_Multiply___setstate_cython;
  PyObject *__pyx_kp_u_Name;
  PyObject *__pyx_n_s_Name_2;
  PyObject *__pyx_n_s_Name___reduce_cython;
  PyObject *__pyx_n_s_Name___setstate_cython;
  PyObject *__pyx_n_s_Negative;
  PyObject *__pyx_n_s_Negative___reduce_cython;
  PyObject *__pyx_n_s_Negative___setstate_cython;
  PyObject *__pyx_n_s_NodeModifier;
  PyObject *__pyx_n_s_NodeModifier___reduce_cython;
  PyObject *__pyx_n_s_NodeModifier___setstate_cython;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_n_s_Not;
  PyObject *__pyx_n_s_NotEqualTo;
  PyObject *__pyx_n_s_NotEqualTo___reduce_cython;
  PyObject *__pyx_n_s_NotEqualTo___setstate_cython;
  PyObject *__pyx_n_s_NotImplementedError;
  PyObject *__pyx_n_s_Not___reduce_cython;
  PyObject *__pyx_n_s_Not___setstate_cython;
  PyObject *__pyx_n_s_Or;
  PyObject *__pyx_n_s_Or___reduce_cython;
  PyObject *__pyx_n_s_Or___setstate_cython;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_u_PolyBinding;
  PyObject *__pyx_n_s_PolyBinding_2;
  PyObject *__pyx_n_s_PolyBinding___reduce_cython;
  PyObject *__pyx_n_s_PolyBinding___setstate_cython;
  PyObject *__pyx_n_s_Positive;
  PyObject *__pyx_n_s_Positive___reduce_cython;
  PyObject *__pyx_n_s_Positive___setstate_cython;
  PyObject *__pyx_n_s_Power;
  PyObject *__pyx_n_s_Power___reduce_cython;
  PyObject *__pyx_n_s_Power___setstate_cython;
  PyObject *__pyx_kp_u_Range;
  PyObject *__pyx_n_s_Range_2;
  PyObject *__pyx_n_s_Range___reduce_cython;
  PyObject *__pyx_n_s_Range___setstate_cython;
  PyObject *__pyx_kp_u_Sequence;
  PyObject *__pyx_n_s_Sequence_2;
  PyObject *__pyx_n_s_Sequence___reduce_cython;
  PyObject *__pyx_n_s_Sequence___setstate_cython;
  PyObject *__pyx_n_s_SharedCache;
  PyObject *__pyx_kp_u_Simplifier_error;
  PyObject *__pyx_kp_u_Slice;
  PyObject *__pyx_n_s_Slice_2;
  PyObject *__pyx_n_s_Slice___reduce_cython;
  PyObject *__pyx_n_s_Slice___setstate_cython;
  PyObject *__pyx_n_s_Subtract;
  PyObject *__pyx_n_s_Subtract___reduce_cython;
  PyObject *__pyx_n_s_Subtract___setstate_cython;
  PyObject *__pyx_kp_u_Tag;
  PyObject *__pyx_n_s_Tag_2;
  PyObject *__pyx_n_s_Tag___reduce_cython;
  PyObject *__pyx_n_s_Tag___setstate_cython;
  PyObject *__pyx_kp_u_Top;
  PyObject *__pyx_n_s_Top_2;
  PyObject *__pyx_n_s_Top___reduce_cython;
  PyObject *__pyx_n_s_Top___setstate_cython;
  PyObject *__pyx_n_u_True;
  PyObject *__pyx_kp_u_Unable_to_simplify_program;
  PyObject *__pyx_n_s_UnaryOperation;
  PyObject *__pyx_n_s_UnaryOperation___reduce_cython;
  PyObject *__pyx_n_s_UnaryOperation___setstate_cython;
  PyObject *__pyx_kp_u_Unbound_name;
  PyObject *__pyx_n_s_Xor;
  PyObject *__pyx_n_s_Xor___reduce_cython;
  PyObject *__pyx_n_s_Xor___setstate_cython;
  PyObject *__pyx_n_s__186;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_n_s__32;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_bindings;
  PyObject *__pyx_n_s_body;
  PyObject *__pyx_n_s_cache;
  PyObject *__pyx_n_s_captures;
  PyObject *__pyx_n_u_ceil;
  PyObject *__pyx_n_s_children;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_compile;
  PyObject *__pyx_n_s_condition;
  PyObject *__pyx_n_s_context;
  PyObject *__pyx_n_u_context_func;
  PyObject *__pyx_n_s_context_vars;
  PyObject *__pyx_n_s_dependencies;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_s_difference_update;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_dynamic;
  PyObject *__pyx_n_s_else;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_s_exc;
  PyObject *__pyx_n_s_exception;
  PyObject *__pyx_n_s_expr;
  PyObject *__pyx_n_s_expressions;
  PyObject *__pyx_n_s_filename;
  PyObject *__pyx_n_s_flitter_language_tree;
  PyObject *__pyx_n_u_floor;
  PyObject *__pyx_n_u_fract;
  PyObject *__pyx_n_s_function;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_with_root;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_initial_errors;
  PyObject *__pyx_n_s_initial_lnames;
  PyObject *__pyx_n_s_inlineable;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_keyword_args;
  PyObject *__pyx_n_s_left;
  PyObject *__pyx_n_s_log_errors;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_loguru;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_name_patch;
  PyObject *__pyx_n_s_names;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_node;
  PyObject *__pyx_n_s_opt;
  PyObject *__pyx_n_s_parameters;
  PyObject *__pyx_n_s_parent;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_pragmas;
  PyObject *__pyx_n_s_program;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_Add;
  PyObject *__pyx_n_s_pyx_unpickle_And;
  PyObject *__pyx_n_s_pyx_unpickle_Append;
  PyObject *__pyx_n_s_pyx_unpickle_Attributes;
  PyObject *__pyx_n_s_pyx_unpickle_BinaryOperation;
  PyObject *__pyx_n_s_pyx_unpickle_Binding;
  PyObject *__pyx_n_s_pyx_unpickle_Call;
  PyObject *__pyx_n_s_pyx_unpickle_Ceil;
  PyObject *__pyx_n_s_pyx_unpickle_Comparison;
  PyObject *__pyx_n_s_pyx_unpickle_Divide;
  PyObject *__pyx_n_s_pyx_unpickle_EqualTo;
  PyObject *__pyx_n_s_pyx_unpickle_Export;
  PyObject *__pyx_n_s_pyx_unpickle_Expression;
  PyObject *__pyx_n_s_pyx_unpickle_Floor;
  PyObject *__pyx_n_s_pyx_unpickle_FloorDivide;
  PyObject *__pyx_n_s_pyx_unpickle_For;
  PyObject *__pyx_n_s_pyx_unpickle_Fract;
  PyObject *__pyx_n_s_pyx_unpickle_Function;
  PyObject *__pyx_n_s_pyx_unpickle_GreaterThan;
  PyObject *__pyx_n_s_pyx_unpickle_GreaterThanOrEqua;
  PyObject *__pyx_n_s_pyx_unpickle_IfCondition;
  PyObject *__pyx_n_s_pyx_unpickle_IfElse;
  PyObject *__pyx_n_s_pyx_unpickle_Import;
  PyObject *__pyx_n_s_pyx_unpickle_LessThan;
  PyObject *__pyx_n_s_pyx_unpickle_LessThanOrEqualTo;
  PyObject *__pyx_n_s_pyx_unpickle_Let;
  PyObject *__pyx_n_s_pyx_unpickle_Literal;
  PyObject *__pyx_n_s_pyx_unpickle_Lookup;
  PyObject *__pyx_n_s_pyx_unpickle_MathsBinaryOperat;
  PyObject *__pyx_n_s_pyx_unpickle_Modulo;
  PyObject *__pyx_n_s_pyx_unpickle_Multiply;
  PyObject *__pyx_n_s_pyx_unpickle_Name;
  PyObject *__pyx_n_s_pyx_unpickle_Negative;
  PyObject *__pyx_n_s_pyx_unpickle_NodeModifier;
  PyObject *__pyx_n_s_pyx_unpickle_Not;
  PyObject *__pyx_n_s_pyx_unpickle_NotEqualTo;
  PyObject *__pyx_n_s_pyx_unpickle_Or;
  PyObject *__pyx_n_s_pyx_unpickle_PolyBinding;
  PyObject *__pyx_n_s_pyx_unpickle_Positive;
  PyObject *__pyx_n_s_pyx_unpickle_Power;
  PyObject *__pyx_n_s_pyx_unpickle_Range;
  PyObject *__pyx_n_s_pyx_unpickle_Sequence;
  PyObject *__pyx_n_s_pyx_unpickle_Slice;
  PyObject *__pyx_n_s_pyx_unpickle_Subtract;
  PyObject *__pyx_n_s_pyx_unpickle_Tag;
  PyObject *__pyx_n_s_pyx_unpickle_Top;
  PyObject *__pyx_n_s_pyx_unpickle_UnaryOperation;
  PyObject *__pyx_n_s_pyx_unpickle_Xor;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_read_flitter_top;
  PyObject *__pyx_n_s_recursive;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_return_context;
  PyObject *__pyx_n_s_reversed;
  PyObject *__pyx_n_s_right;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set_pragma;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_simplify;
  PyObject *__pyx_n_s_source;
  PyObject *__pyx_kp_s_src_flitter_language_tree_pyx;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_static;
  PyObject *__pyx_n_s_static_exports;
  PyObject *__pyx_n_s_step;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_tag;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_tests;
  PyObject *__pyx_n_s_then;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_5413075;
  PyObject *__pyx_int_16586848;
  PyObject *__pyx_int_18525531;
  PyObject *__pyx_int_19056541;
  PyObject *__pyx_int_19136085;
  PyObject *__pyx_int_19657503;
  PyObject *__pyx_int_20102196;
  PyObject *__pyx_int_21458427;
  PyObject *__pyx_int_25936640;
  PyObject *__pyx_int_29776540;
  PyObject *__pyx_int_33963030;
  PyObject *__pyx_int_39246950;
  PyObject *__pyx_int_42909475;
  PyObject *__pyx_int_46599698;
  PyObject *__pyx_int_48890310;
  PyObject *__pyx_int_54761546;
  PyObject *__pyx_int_57427254;
  PyObject *__pyx_int_58806657;
  PyObject *__pyx_int_60181930;
  PyObject *__pyx_int_60997603;
  PyObject *__pyx_int_61976268;
  PyObject *__pyx_int_63365304;
  PyObject *__pyx_int_64598595;
  PyObject *__pyx_int_65788850;
  PyObject *__pyx_int_81599198;
  PyObject *__pyx_int_87851177;
  PyObject *__pyx_int_88468051;
  PyObject *__pyx_int_89160345;
  PyObject *__pyx_int_89314571;
  PyObject *__pyx_int_101425629;
  PyObject *__pyx_int_103159781;
  PyObject *__pyx_int_112105877;
  PyObject *__pyx_int_114850062;
  PyObject *__pyx_int_126472276;
  PyObject *__pyx_int_132590791;
  PyObject *__pyx_int_134496108;
  PyObject *__pyx_int_136983863;
  PyObject *__pyx_int_147834117;
  PyObject *__pyx_int_147971132;
  PyObject *__pyx_int_149349987;
  PyObject *__pyx_int_158473136;
  PyObject *__pyx_int_161766002;
  PyObject *__pyx_int_168508340;
  PyObject *__pyx_int_174256674;
  PyObject *__pyx_int_184977713;
  PyObject *__pyx_int_193537176;
  PyObject *__pyx_int_201276235;
  PyObject *__pyx_int_209144725;
  PyObject *__pyx_int_209608191;
  PyObject *__pyx_int_211607581;
  PyObject *__pyx_int_211749586;
  PyObject *__pyx_int_215229444;
  PyObject *__pyx_int_217770541;
  PyObject *__pyx_int_221199149;
  PyObject *__pyx_int_222419149;
  PyObject *__pyx_int_222754444;
  PyObject *__pyx_int_228825662;
  PyObject *__pyx_int_231565212;
  PyObject *__pyx_int_237118084;
  PyObject *__pyx_int_238750788;
  PyObject *__pyx_int_241444487;
  PyObject *__pyx_int_251934659;
  PyObject *__pyx_int_253263515;
  PyObject *__pyx_int_254981756;
  PyObject *__pyx_int_256597255;
  PyObject *__pyx_int_256930855;
  PyObject *__pyx_int_257251258;
  PyObject *__pyx_int_257836063;
  PyObject *__pyx_int_261002861;
  PyObject *__pyx_int_262293510;
  PyObject *__pyx_int_264124616;
  PyObject *__pyx_int_264265876;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__89;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__91;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__95;
  PyObject *__pyx_codeobj__96;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__98;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__106;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__108;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__111;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__117;
  PyObject *__pyx_codeobj__118;
  PyObject *__pyx_codeobj__119;
  PyObject *__pyx_codeobj__120;
  PyObject *__pyx_codeobj__121;
  PyObject *__pyx_codeobj__122;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__124;
  PyObject *__pyx_codeobj__125;
  PyObject *__pyx_codeobj__126;
  PyObject *__pyx_codeobj__127;
  PyObject *__pyx_codeobj__128;
  PyObject *__pyx_codeobj__129;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__131;
  PyObject *__pyx_codeobj__132;
  PyObject *__pyx_codeobj__133;
  PyObject *__pyx_codeobj__134;
  PyObject *__pyx_codeobj__135;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__138;
  PyObject *__pyx_codeobj__139;
  PyObject *__pyx_codeobj__140;
  PyObject *__pyx_codeobj__141;
  PyObject *__pyx_codeobj__142;
  PyObject *__pyx_codeobj__143;
  PyObject *__pyx_codeobj__144;
  PyObject *__pyx_codeobj__145;
  PyObject *__pyx_codeobj__146;
  PyObject *__pyx_codeobj__147;
  PyObject *__pyx_codeobj__148;
  PyObject *__pyx_codeobj__149;
  PyObject *__pyx_codeobj__150;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__152;
  PyObject *__pyx_codeobj__153;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__155;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__157;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__159;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__161;
  PyObject *__pyx_codeobj__162;
  PyObject *__pyx_codeobj__163;
  PyObject *__pyx_codeobj__164;
  PyObject *__pyx_codeobj__165;
  PyObject *__pyx_codeobj__166;
  PyObject *__pyx_codeobj__167;
  PyObject *__pyx_codeobj__168;
  PyObject *__pyx_codeobj__169;
  PyObject *__pyx_codeobj__170;
  PyObject *__pyx_codeobj__171;
  PyObject *__pyx_codeobj__172;
  PyObject *__pyx_codeobj__173;
  PyObject *__pyx_codeobj__174;
  PyObject *__pyx_codeobj__175;
  PyObject *__pyx_codeobj__176;
  PyObject *__pyx_codeobj__177;
  PyObject *__pyx_codeobj__178;
  PyObject *__pyx_codeobj__179;
  PyObject *__pyx_codeobj__180;
  PyObject *__pyx_codeobj__181;
  PyObject *__pyx_codeobj__182;
  PyObject *__pyx_codeobj__183;
  PyObject *__pyx_codeobj__184;
  PyObject *__pyx_codeobj__185;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_VectorStack);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_Instruction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_2vm_Program);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Expression);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Expression);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Top);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Top);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Export);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Export);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Import);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Import);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Sequence);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Sequence);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Literal);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Literal);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Name);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Name);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Lookup);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Lookup);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Range);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Range);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_UnaryOperation);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_UnaryOperation);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Negative);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Negative);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Positive);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Positive);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Not);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Not);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Ceil);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Ceil);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Floor);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Floor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Fract);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Fract);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_BinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_BinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_MathsBinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Add);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Add);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Subtract);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Subtract);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Multiply);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Multiply);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Divide);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Divide);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_FloorDivide);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_FloorDivide);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Modulo);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Modulo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Power);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Power);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Comparison);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Comparison);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_EqualTo);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_EqualTo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_NotEqualTo);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_NotEqualTo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_LessThan);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_LessThan);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_GreaterThan);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_GreaterThan);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_And);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_And);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Or);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Or);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Xor);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Xor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Slice);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Slice);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Call);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Call);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_NodeModifier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_NodeModifier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Tag);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Tag);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Attributes);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Attributes);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Append);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Append);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Binding);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Binding);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_PolyBinding);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_PolyBinding);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Let);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Let);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_For);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_For);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_IfCondition);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_IfCondition);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_IfElse);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_IfElse);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree_Function);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree_Function);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_n_s_Add);
  Py_CLEAR(clear_module_state->__pyx_n_s_Add___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Add___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_And);
  Py_CLEAR(clear_module_state->__pyx_n_s_And___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_And___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Append);
  Py_CLEAR(clear_module_state->__pyx_n_s_Append___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Append___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_Attributes___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Attributes___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_n_s_BinaryOperation___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BinaryOperation___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Binding);
  Py_CLEAR(clear_module_state->__pyx_n_s_Binding_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Binding___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Binding___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Call);
  Py_CLEAR(clear_module_state->__pyx_n_s_Call_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Call___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Call___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ceil___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ceil___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Circular_import_of);
  Py_CLEAR(clear_module_state->__pyx_n_s_Comparison);
  Py_CLEAR(clear_module_state->__pyx_n_s_Comparison___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Comparison___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Compiler_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_Divide);
  Py_CLEAR(clear_module_state->__pyx_n_s_Divide___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Divide___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_EqualTo___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EqualTo___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_calling);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Export);
  Py_CLEAR(clear_module_state->__pyx_n_s_Export_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Export___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Export___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Expression);
  Py_CLEAR(clear_module_state->__pyx_n_s_Expression___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Expression___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Expression_compile);
  Py_CLEAR(clear_module_state->__pyx_n_s_Expression_simplify);
  Py_CLEAR(clear_module_state->__pyx_n_u_False);
  Py_CLEAR(clear_module_state->__pyx_n_s_Floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloorDivide);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloorDivide___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloorDivide___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Floor___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Floor___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_For);
  Py_CLEAR(clear_module_state->__pyx_n_s_For_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_For___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_For___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Fract);
  Py_CLEAR(clear_module_state->__pyx_n_s_Fract___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Fract___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Function);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Function___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThan);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThanOrEqualTo___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThanOrEqualTo___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThan___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_GreaterThan___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IfCondition);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfCondition_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfCondition___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfCondition___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IfElse);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfElse_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfElse___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IfElse___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Import);
  Py_CLEAR(clear_module_state->__pyx_n_s_Import_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Import___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Import___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Import__simplify_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThan);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThanOrEqualTo___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThanOrEqualTo___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThan___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LessThan___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Let);
  Py_CLEAR(clear_module_state->__pyx_n_s_Let_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Let___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Let___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Literal_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Literal___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Literal___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Lookup);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lookup_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lookup___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lookup___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MathsBinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_n_s_MathsBinaryOperation___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_MathsBinaryOperation___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Modulo);
  Py_CLEAR(clear_module_state->__pyx_n_s_Modulo___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Modulo___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Multiply);
  Py_CLEAR(clear_module_state->__pyx_n_s_Multiply___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Multiply___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Name);
  Py_CLEAR(clear_module_state->__pyx_n_s_Name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Name___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Name___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Negative);
  Py_CLEAR(clear_module_state->__pyx_n_s_Negative___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Negative___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NodeModifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_NodeModifier___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NodeModifier___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_Not);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotEqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotEqualTo___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotEqualTo___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplementedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Not___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Not___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Or);
  Py_CLEAR(clear_module_state->__pyx_n_s_Or___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Or___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PolyBinding);
  Py_CLEAR(clear_module_state->__pyx_n_s_PolyBinding_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PolyBinding___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PolyBinding___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Positive);
  Py_CLEAR(clear_module_state->__pyx_n_s_Positive___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Positive___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Power);
  Py_CLEAR(clear_module_state->__pyx_n_s_Power___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Power___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Range);
  Py_CLEAR(clear_module_state->__pyx_n_s_Range_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Range___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Range___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SharedCache);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Simplifier_error);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Slice);
  Py_CLEAR(clear_module_state->__pyx_n_s_Slice_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Slice___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Slice___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Subtract);
  Py_CLEAR(clear_module_state->__pyx_n_s_Subtract___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Subtract___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tag_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tag___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tag___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Top);
  Py_CLEAR(clear_module_state->__pyx_n_s_Top_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Top___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Top___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_u_True);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unable_to_simplify_program);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnaryOperation);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnaryOperation___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnaryOperation___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unbound_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_Xor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Xor___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Xor___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s__186);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_n_s__32);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_bindings);
  Py_CLEAR(clear_module_state->__pyx_n_s_body);
  Py_CLEAR(clear_module_state->__pyx_n_s_cache);
  Py_CLEAR(clear_module_state->__pyx_n_s_captures);
  Py_CLEAR(clear_module_state->__pyx_n_u_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_children);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_compile);
  Py_CLEAR(clear_module_state->__pyx_n_s_condition);
  Py_CLEAR(clear_module_state->__pyx_n_s_context);
  Py_CLEAR(clear_module_state->__pyx_n_u_context_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_context_vars);
  Py_CLEAR(clear_module_state->__pyx_n_s_dependencies);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_difference_update);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_dynamic);
  Py_CLEAR(clear_module_state->__pyx_n_s_else);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc);
  Py_CLEAR(clear_module_state->__pyx_n_s_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_expr);
  Py_CLEAR(clear_module_state->__pyx_n_s_expressions);
  Py_CLEAR(clear_module_state->__pyx_n_s_filename);
  Py_CLEAR(clear_module_state->__pyx_n_s_flitter_language_tree);
  Py_CLEAR(clear_module_state->__pyx_n_u_floor);
  Py_CLEAR(clear_module_state->__pyx_n_u_fract);
  Py_CLEAR(clear_module_state->__pyx_n_s_function);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_with_root);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_initial_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_initial_lnames);
  Py_CLEAR(clear_module_state->__pyx_n_s_inlineable);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_keyword_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_left);
  Py_CLEAR(clear_module_state->__pyx_n_s_log_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_loguru);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_patch);
  Py_CLEAR(clear_module_state->__pyx_n_s_names);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_node);
  Py_CLEAR(clear_module_state->__pyx_n_s_opt);
  Py_CLEAR(clear_module_state->__pyx_n_s_parameters);
  Py_CLEAR(clear_module_state->__pyx_n_s_parent);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_pragmas);
  Py_CLEAR(clear_module_state->__pyx_n_s_program);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Add);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_And);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Append);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_BinaryOperation);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Binding);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Call);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Comparison);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Divide);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_EqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Export);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Expression);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FloorDivide);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_For);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Fract);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Function);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GreaterThan);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_GreaterThanOrEqua);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IfCondition);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IfElse);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Import);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LessThan);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LessThanOrEqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Let);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Literal);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Lookup);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_MathsBinaryOperat);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Modulo);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Multiply);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Name);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Negative);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_NodeModifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Not);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_NotEqualTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Or);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PolyBinding);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Positive);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Power);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Range);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Slice);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Subtract);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Top);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UnaryOperation);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Xor);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_flitter_top);
  Py_CLEAR(clear_module_state->__pyx_n_s_recursive);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_context);
  Py_CLEAR(clear_module_state->__pyx_n_s_reversed);
  Py_CLEAR(clear_module_state->__pyx_n_s_right);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_pragma);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_simplify);
  Py_CLEAR(clear_module_state->__pyx_n_s_source);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_flitter_language_tree_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_static);
  Py_CLEAR(clear_module_state->__pyx_n_s_static_exports);
  Py_CLEAR(clear_module_state->__pyx_n_s_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests);
  Py_CLEAR(clear_module_state->__pyx_n_s_then);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_5413075);
  Py_CLEAR(clear_module_state->__pyx_int_16586848);
  Py_CLEAR(clear_module_state->__pyx_int_18525531);
  Py_CLEAR(clear_module_state->__pyx_int_19056541);
  Py_CLEAR(clear_module_state->__pyx_int_19136085);
  Py_CLEAR(clear_module_state->__pyx_int_19657503);
  Py_CLEAR(clear_module_state->__pyx_int_20102196);
  Py_CLEAR(clear_module_state->__pyx_int_21458427);
  Py_CLEAR(clear_module_state->__pyx_int_25936640);
  Py_CLEAR(clear_module_state->__pyx_int_29776540);
  Py_CLEAR(clear_module_state->__pyx_int_33963030);
  Py_CLEAR(clear_module_state->__pyx_int_39246950);
  Py_CLEAR(clear_module_state->__pyx_int_42909475);
  Py_CLEAR(clear_module_state->__pyx_int_46599698);
  Py_CLEAR(clear_module_state->__pyx_int_48890310);
  Py_CLEAR(clear_module_state->__pyx_int_54761546);
  Py_CLEAR(clear_module_state->__pyx_int_57427254);
  Py_CLEAR(clear_module_state->__pyx_int_58806657);
  Py_CLEAR(clear_module_state->__pyx_int_60181930);
  Py_CLEAR(clear_module_state->__pyx_int_60997603);
  Py_CLEAR(clear_module_state->__pyx_int_61976268);
  Py_CLEAR(clear_module_state->__pyx_int_63365304);
  Py_CLEAR(clear_module_state->__pyx_int_64598595);
  Py_CLEAR(clear_module_state->__pyx_int_65788850);
  Py_CLEAR(clear_module_state->__pyx_int_81599198);
  Py_CLEAR(clear_module_state->__pyx_int_87851177);
  Py_CLEAR(clear_module_state->__pyx_int_88468051);
  Py_CLEAR(clear_module_state->__pyx_int_89160345);
  Py_CLEAR(clear_module_state->__pyx_int_89314571);
  Py_CLEAR(clear_module_state->__pyx_int_101425629);
  Py_CLEAR(clear_module_state->__pyx_int_103159781);
  Py_CLEAR(clear_module_state->__pyx_int_112105877);
  Py_CLEAR(clear_module_state->__pyx_int_114850062);
  Py_CLEAR(clear_module_state->__pyx_int_126472276);
  Py_CLEAR(clear_module_state->__pyx_int_132590791);
  Py_CLEAR(clear_module_state->__pyx_int_134496108);
  Py_CLEAR(clear_module_state->__pyx_int_136983863);
  Py_CLEAR(clear_module_state->__pyx_int_147834117);
  Py_CLEAR(clear_module_state->__pyx_int_147971132);
  Py_CLEAR(clear_module_state->__pyx_int_149349987);
  Py_CLEAR(clear_module_state->__pyx_int_158473136);
  Py_CLEAR(clear_module_state->__pyx_int_161766002);
  Py_CLEAR(clear_module_state->__pyx_int_168508340);
  Py_CLEAR(clear_module_state->__pyx_int_174256674);
  Py_CLEAR(clear_module_state->__pyx_int_184977713);
  Py_CLEAR(clear_module_state->__pyx_int_193537176);
  Py_CLEAR(clear_module_state->__pyx_int_201276235);
  Py_CLEAR(clear_module_state->__pyx_int_209144725);
  Py_CLEAR(clear_module_state->__pyx_int_209608191);
  Py_CLEAR(clear_module_state->__pyx_int_211607581);
  Py_CLEAR(clear_module_state->__pyx_int_211749586);
  Py_CLEAR(clear_module_state->__pyx_int_215229444);
  Py_CLEAR(clear_module_state->__pyx_int_217770541);
  Py_CLEAR(clear_module_state->__pyx_int_221199149);
  Py_CLEAR(clear_module_state->__pyx_int_222419149);
  Py_CLEAR(clear_module_state->__pyx_int_222754444);
  Py_CLEAR(clear_module_state->__pyx_int_228825662);
  Py_CLEAR(clear_module_state->__pyx_int_231565212);
  Py_CLEAR(clear_module_state->__pyx_int_237118084);
  Py_CLEAR(clear_module_state->__pyx_int_238750788);
  Py_CLEAR(clear_module_state->__pyx_int_241444487);
  Py_CLEAR(clear_module_state->__pyx_int_251934659);
  Py_CLEAR(clear_module_state->__pyx_int_253263515);
  Py_CLEAR(clear_module_state->__pyx_int_254981756);
  Py_CLEAR(clear_module_state->__pyx_int_256597255);
  Py_CLEAR(clear_module_state->__pyx_int_256930855);
  Py_CLEAR(clear_module_state->__pyx_int_257251258);
  Py_CLEAR(clear_module_state->__pyx_int_257836063);
  Py_CLEAR(clear_module_state->__pyx_int_261002861);
  Py_CLEAR(clear_module_state->__pyx_int_262293510);
  Py_CLEAR(clear_module_state->__pyx_int_264124616);
  Py_CLEAR(clear_module_state->__pyx_int_264265876);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__89);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__91);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  Py_CLEAR(clear_module_state->__pyx_codeobj__96);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__106);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__108);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__111);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  Py_CLEAR(clear_module_state->__pyx_codeobj__118);
  Py_CLEAR(clear_module_state->__pyx_codeobj__119);
  Py_CLEAR(clear_module_state->__pyx_codeobj__120);
  Py_CLEAR(clear_module_state->__pyx_codeobj__121);
  Py_CLEAR(clear_module_state->__pyx_codeobj__122);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__124);
  Py_CLEAR(clear_module_state->__pyx_codeobj__125);
  Py_CLEAR(clear_module_state->__pyx_codeobj__126);
  Py_CLEAR(clear_module_state->__pyx_codeobj__127);
  Py_CLEAR(clear_module_state->__pyx_codeobj__128);
  Py_CLEAR(clear_module_state->__pyx_codeobj__129);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__131);
  Py_CLEAR(clear_module_state->__pyx_codeobj__132);
  Py_CLEAR(clear_module_state->__pyx_codeobj__133);
  Py_CLEAR(clear_module_state->__pyx_codeobj__134);
  Py_CLEAR(clear_module_state->__pyx_codeobj__135);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__138);
  Py_CLEAR(clear_module_state->__pyx_codeobj__139);
  Py_CLEAR(clear_module_state->__pyx_codeobj__140);
  Py_CLEAR(clear_module_state->__pyx_codeobj__141);
  Py_CLEAR(clear_module_state->__pyx_codeobj__142);
  Py_CLEAR(clear_module_state->__pyx_codeobj__143);
  Py_CLEAR(clear_module_state->__pyx_codeobj__144);
  Py_CLEAR(clear_module_state->__pyx_codeobj__145);
  Py_CLEAR(clear_module_state->__pyx_codeobj__146);
  Py_CLEAR(clear_module_state->__pyx_codeobj__147);
  Py_CLEAR(clear_module_state->__pyx_codeobj__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__149);
  Py_CLEAR(clear_module_state->__pyx_codeobj__150);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__152);
  Py_CLEAR(clear_module_state->__pyx_codeobj__153);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__155);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__157);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__159);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__161);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__163);
  Py_CLEAR(clear_module_state->__pyx_codeobj__164);
  Py_CLEAR(clear_module_state->__pyx_codeobj__165);
  Py_CLEAR(clear_module_state->__pyx_codeobj__166);
  Py_CLEAR(clear_module_state->__pyx_codeobj__167);
  Py_CLEAR(clear_module_state->__pyx_codeobj__168);
  Py_CLEAR(clear_module_state->__pyx_codeobj__169);
  Py_CLEAR(clear_module_state->__pyx_codeobj__170);
  Py_CLEAR(clear_module_state->__pyx_codeobj__171);
  Py_CLEAR(clear_module_state->__pyx_codeobj__172);
  Py_CLEAR(clear_module_state->__pyx_codeobj__173);
  Py_CLEAR(clear_module_state->__pyx_codeobj__174);
  Py_CLEAR(clear_module_state->__pyx_codeobj__175);
  Py_CLEAR(clear_module_state->__pyx_codeobj__176);
  Py_CLEAR(clear_module_state->__pyx_codeobj__177);
  Py_CLEAR(clear_module_state->__pyx_codeobj__178);
  Py_CLEAR(clear_module_state->__pyx_codeobj__179);
  Py_CLEAR(clear_module_state->__pyx_codeobj__180);
  Py_CLEAR(clear_module_state->__pyx_codeobj__181);
  Py_CLEAR(clear_module_state->__pyx_codeobj__182);
  Py_CLEAR(clear_module_state->__pyx_codeobj__183);
  Py_CLEAR(clear_module_state->__pyx_codeobj__184);
  Py_CLEAR(clear_module_state->__pyx_codeobj__185);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_VectorStack);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_Instruction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionVector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionNode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionStr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionTuple);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionInt);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionLabel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionJump);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionFunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_2vm_Program);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Expression);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Expression);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Top);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Top);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Export);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Export);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Import);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Import);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Sequence);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Sequence);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Literal);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Literal);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Name);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Name);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Lookup);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Lookup);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Range);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Range);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_UnaryOperation);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_UnaryOperation);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Negative);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Negative);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Positive);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Positive);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Not);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Not);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Ceil);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Ceil);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Floor);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Floor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Fract);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Fract);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_BinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_BinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_MathsBinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Add);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Add);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Subtract);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Subtract);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Multiply);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Multiply);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Divide);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Divide);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_FloorDivide);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_FloorDivide);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Modulo);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Modulo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Power);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Power);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Comparison);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Comparison);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_EqualTo);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_EqualTo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_NotEqualTo);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_NotEqualTo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_LessThan);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_LessThan);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_GreaterThan);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_GreaterThan);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_And);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_And);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Or);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Or);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Xor);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Xor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Slice);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Slice);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Call);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Call);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_NodeModifier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_NodeModifier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Tag);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Tag);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Attributes);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Attributes);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Append);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Append);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Binding);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Binding);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_PolyBinding);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_PolyBinding);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Let);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Let);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_For);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_For);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_IfCondition);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_IfCondition);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_IfElse);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_IfElse);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree_Function);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree_Function);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_n_s_Add);
  Py_VISIT(traverse_module_state->__pyx_n_s_Add___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Add___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_And);
  Py_VISIT(traverse_module_state->__pyx_n_s_And___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_And___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Append);
  Py_VISIT(traverse_module_state->__pyx_n_s_Append___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Append___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_Attributes___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Attributes___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_n_s_BinaryOperation___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BinaryOperation___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Binding);
  Py_VISIT(traverse_module_state->__pyx_n_s_Binding_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Binding___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Binding___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Call);
  Py_VISIT(traverse_module_state->__pyx_n_s_Call_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Call___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Call___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ceil___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ceil___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Circular_import_of);
  Py_VISIT(traverse_module_state->__pyx_n_s_Comparison);
  Py_VISIT(traverse_module_state->__pyx_n_s_Comparison___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Comparison___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Compiler_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_Divide);
  Py_VISIT(traverse_module_state->__pyx_n_s_Divide___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Divide___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_EqualTo___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EqualTo___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_calling);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Export);
  Py_VISIT(traverse_module_state->__pyx_n_s_Export_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Export___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Export___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Expression);
  Py_VISIT(traverse_module_state->__pyx_n_s_Expression___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Expression___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Expression_compile);
  Py_VISIT(traverse_module_state->__pyx_n_s_Expression_simplify);
  Py_VISIT(traverse_module_state->__pyx_n_u_False);
  Py_VISIT(traverse_module_state->__pyx_n_s_Floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloorDivide);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloorDivide___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloorDivide___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Floor___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Floor___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_For);
  Py_VISIT(traverse_module_state->__pyx_n_s_For_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_For___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_For___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Fract);
  Py_VISIT(traverse_module_state->__pyx_n_s_Fract___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Fract___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Function);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Function___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThan);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThanOrEqualTo___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThanOrEqualTo___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThan___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_GreaterThan___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IfCondition);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfCondition_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfCondition___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfCondition___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IfElse);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfElse_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfElse___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IfElse___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Import);
  Py_VISIT(traverse_module_state->__pyx_n_s_Import_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Import___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Import___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Import__simplify_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThan);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThanOrEqualTo___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThanOrEqualTo___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThan___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LessThan___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Let);
  Py_VISIT(traverse_module_state->__pyx_n_s_Let_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Let___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Let___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Literal_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Literal___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Literal___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Lookup);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lookup_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lookup___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lookup___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MathsBinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_n_s_MathsBinaryOperation___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_MathsBinaryOperation___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Modulo);
  Py_VISIT(traverse_module_state->__pyx_n_s_Modulo___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Modulo___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Multiply);
  Py_VISIT(traverse_module_state->__pyx_n_s_Multiply___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Multiply___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Name);
  Py_VISIT(traverse_module_state->__pyx_n_s_Name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Name___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Name___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Negative);
  Py_VISIT(traverse_module_state->__pyx_n_s_Negative___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Negative___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NodeModifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_NodeModifier___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NodeModifier___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_Not);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotEqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotEqualTo___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotEqualTo___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplementedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Not___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Not___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Or);
  Py_VISIT(traverse_module_state->__pyx_n_s_Or___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Or___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PolyBinding);
  Py_VISIT(traverse_module_state->__pyx_n_s_PolyBinding_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PolyBinding___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PolyBinding___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Positive);
  Py_VISIT(traverse_module_state->__pyx_n_s_Positive___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Positive___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Power);
  Py_VISIT(traverse_module_state->__pyx_n_s_Power___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Power___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Range);
  Py_VISIT(traverse_module_state->__pyx_n_s_Range_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Range___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Range___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SharedCache);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Simplifier_error);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Slice);
  Py_VISIT(traverse_module_state->__pyx_n_s_Slice_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Slice___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Slice___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Subtract);
  Py_VISIT(traverse_module_state->__pyx_n_s_Subtract___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Subtract___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tag_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tag___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tag___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Top);
  Py_VISIT(traverse_module_state->__pyx_n_s_Top_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Top___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Top___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_u_True);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unable_to_simplify_program);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnaryOperation);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnaryOperation___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnaryOperation___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unbound_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_Xor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Xor___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Xor___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s__186);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_n_s__32);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_bindings);
  Py_VISIT(traverse_module_state->__pyx_n_s_body);
  Py_VISIT(traverse_module_state->__pyx_n_s_cache);
  Py_VISIT(traverse_module_state->__pyx_n_s_captures);
  Py_VISIT(traverse_module_state->__pyx_n_u_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_children);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_compile);
  Py_VISIT(traverse_module_state->__pyx_n_s_condition);
  Py_VISIT(traverse_module_state->__pyx_n_s_context);
  Py_VISIT(traverse_module_state->__pyx_n_u_context_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_context_vars);
  Py_VISIT(traverse_module_state->__pyx_n_s_dependencies);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_difference_update);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_dynamic);
  Py_VISIT(traverse_module_state->__pyx_n_s_else);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc);
  Py_VISIT(traverse_module_state->__pyx_n_s_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_expr);
  Py_VISIT(traverse_module_state->__pyx_n_s_expressions);
  Py_VISIT(traverse_module_state->__pyx_n_s_filename);
  Py_VISIT(traverse_module_state->__pyx_n_s_flitter_language_tree);
  Py_VISIT(traverse_module_state->__pyx_n_u_floor);
  Py_VISIT(traverse_module_state->__pyx_n_u_fract);
  Py_VISIT(traverse_module_state->__pyx_n_s_function);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_with_root);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_initial_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_initial_lnames);
  Py_VISIT(traverse_module_state->__pyx_n_s_inlineable);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_keyword_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_left);
  Py_VISIT(traverse_module_state->__pyx_n_s_log_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_loguru);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_patch);
  Py_VISIT(traverse_module_state->__pyx_n_s_names);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_node);
  Py_VISIT(traverse_module_state->__pyx_n_s_opt);
  Py_VISIT(traverse_module_state->__pyx_n_s_parameters);
  Py_VISIT(traverse_module_state->__pyx_n_s_parent);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_pragmas);
  Py_VISIT(traverse_module_state->__pyx_n_s_program);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Add);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_And);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Append);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_BinaryOperation);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Binding);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Call);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Comparison);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Divide);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_EqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Export);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Expression);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FloorDivide);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_For);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Fract);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Function);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GreaterThan);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_GreaterThanOrEqua);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IfCondition);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IfElse);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Import);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LessThan);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LessThanOrEqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Let);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Literal);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Lookup);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_MathsBinaryOperat);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Modulo);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Multiply);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Name);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Negative);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_NodeModifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Not);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_NotEqualTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Or);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PolyBinding);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Positive);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Power);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Range);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Slice);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Subtract);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Top);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UnaryOperation);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Xor);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_flitter_top);
  Py_VISIT(traverse_module_state->__pyx_n_s_recursive);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_context);
  Py_VISIT(traverse_module_state->__pyx_n_s_reversed);
  Py_VISIT(traverse_module_state->__pyx_n_s_right);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_pragma);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_simplify);
  Py_VISIT(traverse_module_state->__pyx_n_s_source);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_flitter_language_tree_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_static);
  Py_VISIT(traverse_module_state->__pyx_n_s_static_exports);
  Py_VISIT(traverse_module_state->__pyx_n_s_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests);
  Py_VISIT(traverse_module_state->__pyx_n_s_then);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_5413075);
  Py_VISIT(traverse_module_state->__pyx_int_16586848);
  Py_VISIT(traverse_module_state->__pyx_int_18525531);
  Py_VISIT(traverse_module_state->__pyx_int_19056541);
  Py_VISIT(traverse_module_state->__pyx_int_19136085);
  Py_VISIT(traverse_module_state->__pyx_int_19657503);
  Py_VISIT(traverse_module_state->__pyx_int_20102196);
  Py_VISIT(traverse_module_state->__pyx_int_21458427);
  Py_VISIT(traverse_module_state->__pyx_int_25936640);
  Py_VISIT(traverse_module_state->__pyx_int_29776540);
  Py_VISIT(traverse_module_state->__pyx_int_33963030);
  Py_VISIT(traverse_module_state->__pyx_int_39246950);
  Py_VISIT(traverse_module_state->__pyx_int_42909475);
  Py_VISIT(traverse_module_state->__pyx_int_46599698);
  Py_VISIT(traverse_module_state->__pyx_int_48890310);
  Py_VISIT(traverse_module_state->__pyx_int_54761546);
  Py_VISIT(traverse_module_state->__pyx_int_57427254);
  Py_VISIT(traverse_module_state->__pyx_int_58806657);
  Py_VISIT(traverse_module_state->__pyx_int_60181930);
  Py_VISIT(traverse_module_state->__pyx_int_60997603);
  Py_VISIT(traverse_module_state->__pyx_int_61976268);
  Py_VISIT(traverse_module_state->__pyx_int_63365304);
  Py_VISIT(traverse_module_state->__pyx_int_64598595);
  Py_VISIT(traverse_module_state->__pyx_int_65788850);
  Py_VISIT(traverse_module_state->__pyx_int_81599198);
  Py_VISIT(traverse_module_state->__pyx_int_87851177);
  Py_VISIT(traverse_module_state->__pyx_int_88468051);
  Py_VISIT(traverse_module_state->__pyx_int_89160345);
  Py_VISIT(traverse_module_state->__pyx_int_89314571);
  Py_VISIT(traverse_module_state->__pyx_int_101425629);
  Py_VISIT(traverse_module_state->__pyx_int_103159781);
  Py_VISIT(traverse_module_state->__pyx_int_112105877);
  Py_VISIT(traverse_module_state->__pyx_int_114850062);
  Py_VISIT(traverse_module_state->__pyx_int_126472276);
  Py_VISIT(traverse_module_state->__pyx_int_132590791);
  Py_VISIT(traverse_module_state->__pyx_int_134496108);
  Py_VISIT(traverse_module_state->__pyx_int_136983863);
  Py_VISIT(traverse_module_state->__pyx_int_147834117);
  Py_VISIT(traverse_module_state->__pyx_int_147971132);
  Py_VISIT(traverse_module_state->__pyx_int_149349987);
  Py_VISIT(traverse_module_state->__pyx_int_158473136);
  Py_VISIT(traverse_module_state->__pyx_int_161766002);
  Py_VISIT(traverse_module_state->__pyx_int_168508340);
  Py_VISIT(traverse_module_state->__pyx_int_174256674);
  Py_VISIT(traverse_module_state->__pyx_int_184977713);
  Py_VISIT(traverse_module_state->__pyx_int_193537176);
  Py_VISIT(traverse_module_state->__pyx_int_201276235);
  Py_VISIT(traverse_module_state->__pyx_int_209144725);
  Py_VISIT(traverse_module_state->__pyx_int_209608191);
  Py_VISIT(traverse_module_state->__pyx_int_211607581);
  Py_VISIT(traverse_module_state->__pyx_int_211749586);
  Py_VISIT(traverse_module_state->__pyx_int_215229444);
  Py_VISIT(traverse_module_state->__pyx_int_217770541);
  Py_VISIT(traverse_module_state->__pyx_int_221199149);
  Py_VISIT(traverse_module_state->__pyx_int_222419149);
  Py_VISIT(traverse_module_state->__pyx_int_222754444);
  Py_VISIT(traverse_module_state->__pyx_int_228825662);
  Py_VISIT(traverse_module_state->__pyx_int_231565212);
  Py_VISIT(traverse_module_state->__pyx_int_237118084);
  Py_VISIT(traverse_module_state->__pyx_int_238750788);
  Py_VISIT(traverse_module_state->__pyx_int_241444487);
  Py_VISIT(traverse_module_state->__pyx_int_251934659);
  Py_VISIT(traverse_module_state->__pyx_int_253263515);
  Py_VISIT(traverse_module_state->__pyx_int_254981756);
  Py_VISIT(traverse_module_state->__pyx_int_256597255);
  Py_VISIT(traverse_module_state->__pyx_int_256930855);
  Py_VISIT(traverse_module_state->__pyx_int_257251258);
  Py_VISIT(traverse_module_state->__pyx_int_257836063);
  Py_VISIT(traverse_module_state->__pyx_int_261002861);
  Py_VISIT(traverse_module_state->__pyx_int_262293510);
  Py_VISIT(traverse_module_state->__pyx_int_264124616);
  Py_VISIT(traverse_module_state->__pyx_int_264265876);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__89);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__91);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  Py_VISIT(traverse_module_state->__pyx_codeobj__96);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__106);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__108);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__111);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  Py_VISIT(traverse_module_state->__pyx_codeobj__118);
  Py_VISIT(traverse_module_state->__pyx_codeobj__119);
  Py_VISIT(traverse_module_state->__pyx_codeobj__120);
  Py_VISIT(traverse_module_state->__pyx_codeobj__121);
  Py_VISIT(traverse_module_state->__pyx_codeobj__122);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__124);
  Py_VISIT(traverse_module_state->__pyx_codeobj__125);
  Py_VISIT(traverse_module_state->__pyx_codeobj__126);
  Py_VISIT(traverse_module_state->__pyx_codeobj__127);
  Py_VISIT(traverse_module_state->__pyx_codeobj__128);
  Py_VISIT(traverse_module_state->__pyx_codeobj__129);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__131);
  Py_VISIT(traverse_module_state->__pyx_codeobj__132);
  Py_VISIT(traverse_module_state->__pyx_codeobj__133);
  Py_VISIT(traverse_module_state->__pyx_codeobj__134);
  Py_VISIT(traverse_module_state->__pyx_codeobj__135);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__138);
  Py_VISIT(traverse_module_state->__pyx_codeobj__139);
  Py_VISIT(traverse_module_state->__pyx_codeobj__140);
  Py_VISIT(traverse_module_state->__pyx_codeobj__141);
  Py_VISIT(traverse_module_state->__pyx_codeobj__142);
  Py_VISIT(traverse_module_state->__pyx_codeobj__143);
  Py_VISIT(traverse_module_state->__pyx_codeobj__144);
  Py_VISIT(traverse_module_state->__pyx_codeobj__145);
  Py_VISIT(traverse_module_state->__pyx_codeobj__146);
  Py_VISIT(traverse_module_state->__pyx_codeobj__147);
  Py_VISIT(traverse_module_state->__pyx_codeobj__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__149);
  Py_VISIT(traverse_module_state->__pyx_codeobj__150);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__152);
  Py_VISIT(traverse_module_state->__pyx_codeobj__153);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__155);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__157);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__159);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__161);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__163);
  Py_VISIT(traverse_module_state->__pyx_codeobj__164);
  Py_VISIT(traverse_module_state->__pyx_codeobj__165);
  Py_VISIT(traverse_module_state->__pyx_codeobj__166);
  Py_VISIT(traverse_module_state->__pyx_codeobj__167);
  Py_VISIT(traverse_module_state->__pyx_codeobj__168);
  Py_VISIT(traverse_module_state->__pyx_codeobj__169);
  Py_VISIT(traverse_module_state->__pyx_codeobj__170);
  Py_VISIT(traverse_module_state->__pyx_codeobj__171);
  Py_VISIT(traverse_module_state->__pyx_codeobj__172);
  Py_VISIT(traverse_module_state->__pyx_codeobj__173);
  Py_VISIT(traverse_module_state->__pyx_codeobj__174);
  Py_VISIT(traverse_module_state->__pyx_codeobj__175);
  Py_VISIT(traverse_module_state->__pyx_codeobj__176);
  Py_VISIT(traverse_module_state->__pyx_codeobj__177);
  Py_VISIT(traverse_module_state->__pyx_codeobj__178);
  Py_VISIT(traverse_module_state->__pyx_codeobj__179);
  Py_VISIT(traverse_module_state->__pyx_codeobj__180);
  Py_VISIT(traverse_module_state->__pyx_codeobj__181);
  Py_VISIT(traverse_module_state->__pyx_codeobj__182);
  Py_VISIT(traverse_module_state->__pyx_codeobj__183);
  Py_VISIT(traverse_module_state->__pyx_codeobj__184);
  Py_VISIT(traverse_module_state->__pyx_codeobj__185);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7flitter_5model_Vector __pyx_mstate_global->__pyx_ptype_7flitter_5model_Vector
#define __pyx_ptype_7flitter_5model_Matrix33 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix33
#define __pyx_ptype_7flitter_5model_Matrix44 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix44
#define __pyx_ptype_7flitter_5model_Quaternion __pyx_mstate_global->__pyx_ptype_7flitter_5model_Quaternion
#define __pyx_ptype_7flitter_5model_Node __pyx_mstate_global->__pyx_ptype_7flitter_5model_Node
#define __pyx_ptype_7flitter_5model_StateDict __pyx_mstate_global->__pyx_ptype_7flitter_5model_StateDict
#define __pyx_ptype_7flitter_5model_Context __pyx_mstate_global->__pyx_ptype_7flitter_5model_Context
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7flitter_8language_2vm_VectorStack __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_VectorStack
#define __pyx_ptype_7flitter_8language_2vm_Instruction __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_Instruction
#define __pyx_ptype_7flitter_8language_2vm_InstructionVector __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionVector
#define __pyx_ptype_7flitter_8language_2vm_InstructionNode __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionNode
#define __pyx_ptype_7flitter_8language_2vm_InstructionStr __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionStr
#define __pyx_ptype_7flitter_8language_2vm_InstructionTuple __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionTuple
#define __pyx_ptype_7flitter_8language_2vm_InstructionInt __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionInt
#define __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple
#define __pyx_ptype_7flitter_8language_2vm_InstructionLabel __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionLabel
#define __pyx_ptype_7flitter_8language_2vm_InstructionJump __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionJump
#define __pyx_ptype_7flitter_8language_2vm_InstructionFunc __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionFunc
#define __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt
#define __pyx_ptype_7flitter_8language_2vm_Program __pyx_mstate_global->__pyx_ptype_7flitter_8language_2vm_Program
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_7flitter_8language_4tree_Expression __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Expression
#define __pyx_type_7flitter_8language_4tree_Top __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Top
#define __pyx_type_7flitter_8language_4tree_Export __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Export
#define __pyx_type_7flitter_8language_4tree_Import __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Import
#define __pyx_type_7flitter_8language_4tree_Sequence __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Sequence
#define __pyx_type_7flitter_8language_4tree_Literal __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Literal
#define __pyx_type_7flitter_8language_4tree_Name __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Name
#define __pyx_type_7flitter_8language_4tree_Lookup __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Lookup
#define __pyx_type_7flitter_8language_4tree_Range __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Range
#define __pyx_type_7flitter_8language_4tree_UnaryOperation __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_UnaryOperation
#define __pyx_type_7flitter_8language_4tree_Negative __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Negative
#define __pyx_type_7flitter_8language_4tree_Positive __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Positive
#define __pyx_type_7flitter_8language_4tree_Not __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Not
#define __pyx_type_7flitter_8language_4tree_Ceil __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Ceil
#define __pyx_type_7flitter_8language_4tree_Floor __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Floor
#define __pyx_type_7flitter_8language_4tree_Fract __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Fract
#define __pyx_type_7flitter_8language_4tree_BinaryOperation __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_BinaryOperation
#define __pyx_type_7flitter_8language_4tree_MathsBinaryOperation __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_MathsBinaryOperation
#define __pyx_type_7flitter_8language_4tree_Add __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Add
#define __pyx_type_7flitter_8language_4tree_Subtract __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Subtract
#define __pyx_type_7flitter_8language_4tree_Multiply __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Multiply
#define __pyx_type_7flitter_8language_4tree_Divide __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Divide
#define __pyx_type_7flitter_8language_4tree_FloorDivide __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_FloorDivide
#define __pyx_type_7flitter_8language_4tree_Modulo __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Modulo
#define __pyx_type_7flitter_8language_4tree_Power __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Power
#define __pyx_type_7flitter_8language_4tree_Comparison __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Comparison
#define __pyx_type_7flitter_8language_4tree_EqualTo __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_EqualTo
#define __pyx_type_7flitter_8language_4tree_NotEqualTo __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_NotEqualTo
#define __pyx_type_7flitter_8language_4tree_LessThan __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_LessThan
#define __pyx_type_7flitter_8language_4tree_GreaterThan __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_GreaterThan
#define __pyx_type_7flitter_8language_4tree_LessThanOrEqualTo __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo
#define __pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo
#define __pyx_type_7flitter_8language_4tree_And __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_And
#define __pyx_type_7flitter_8language_4tree_Or __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Or
#define __pyx_type_7flitter_8language_4tree_Xor __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Xor
#define __pyx_type_7flitter_8language_4tree_Slice __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Slice
#define __pyx_type_7flitter_8language_4tree_Call __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Call
#define __pyx_type_7flitter_8language_4tree_NodeModifier __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_NodeModifier
#define __pyx_type_7flitter_8language_4tree_Tag __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Tag
#define __pyx_type_7flitter_8language_4tree_Attributes __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Attributes
#define __pyx_type_7flitter_8language_4tree_Append __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Append
#define __pyx_type_7flitter_8language_4tree_Binding __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Binding
#define __pyx_type_7flitter_8language_4tree_PolyBinding __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_PolyBinding
#define __pyx_type_7flitter_8language_4tree_Let __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Let
#define __pyx_type_7flitter_8language_4tree_For __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_For
#define __pyx_type_7flitter_8language_4tree_IfCondition __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_IfCondition
#define __pyx_type_7flitter_8language_4tree_IfElse __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_IfElse
#define __pyx_type_7flitter_8language_4tree_Function __pyx_mstate_global->__pyx_type_7flitter_8language_4tree_Function
#define __pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr
#endif
#define __pyx_ptype_7flitter_8language_4tree_Expression __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Expression
#define __pyx_ptype_7flitter_8language_4tree_Top __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Top
#define __pyx_ptype_7flitter_8language_4tree_Export __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Export
#define __pyx_ptype_7flitter_8language_4tree_Import __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Import
#define __pyx_ptype_7flitter_8language_4tree_Sequence __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Sequence
#define __pyx_ptype_7flitter_8language_4tree_Literal __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Literal
#define __pyx_ptype_7flitter_8language_4tree_Name __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Name
#define __pyx_ptype_7flitter_8language_4tree_Lookup __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Lookup
#define __pyx_ptype_7flitter_8language_4tree_Range __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Range
#define __pyx_ptype_7flitter_8language_4tree_UnaryOperation __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_UnaryOperation
#define __pyx_ptype_7flitter_8language_4tree_Negative __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Negative
#define __pyx_ptype_7flitter_8language_4tree_Positive __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Positive
#define __pyx_ptype_7flitter_8language_4tree_Not __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Not
#define __pyx_ptype_7flitter_8language_4tree_Ceil __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Ceil
#define __pyx_ptype_7flitter_8language_4tree_Floor __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Floor
#define __pyx_ptype_7flitter_8language_4tree_Fract __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Fract
#define __pyx_ptype_7flitter_8language_4tree_BinaryOperation __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_BinaryOperation
#define __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation
#define __pyx_ptype_7flitter_8language_4tree_Add __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Add
#define __pyx_ptype_7flitter_8language_4tree_Subtract __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Subtract
#define __pyx_ptype_7flitter_8language_4tree_Multiply __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Multiply
#define __pyx_ptype_7flitter_8language_4tree_Divide __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Divide
#define __pyx_ptype_7flitter_8language_4tree_FloorDivide __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_FloorDivide
#define __pyx_ptype_7flitter_8language_4tree_Modulo __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Modulo
#define __pyx_ptype_7flitter_8language_4tree_Power __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Power
#define __pyx_ptype_7flitter_8language_4tree_Comparison __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Comparison
#define __pyx_ptype_7flitter_8language_4tree_EqualTo __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_EqualTo
#define __pyx_ptype_7flitter_8language_4tree_NotEqualTo __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_NotEqualTo
#define __pyx_ptype_7flitter_8language_4tree_LessThan __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_LessThan
#define __pyx_ptype_7flitter_8language_4tree_GreaterThan __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_GreaterThan
#define __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo
#define __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo
#define __pyx_ptype_7flitter_8language_4tree_And __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_And
#define __pyx_ptype_7flitter_8language_4tree_Or __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Or
#define __pyx_ptype_7flitter_8language_4tree_Xor __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Xor
#define __pyx_ptype_7flitter_8language_4tree_Slice __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Slice
#define __pyx_ptype_7flitter_8language_4tree_Call __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Call
#define __pyx_ptype_7flitter_8language_4tree_NodeModifier __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_NodeModifier
#define __pyx_ptype_7flitter_8language_4tree_Tag __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Tag
#define __pyx_ptype_7flitter_8language_4tree_Attributes __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Attributes
#define __pyx_ptype_7flitter_8language_4tree_Append __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Append
#define __pyx_ptype_7flitter_8language_4tree_Binding __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Binding
#define __pyx_ptype_7flitter_8language_4tree_PolyBinding __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_PolyBinding
#define __pyx_ptype_7flitter_8language_4tree_Let __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Let
#define __pyx_ptype_7flitter_8language_4tree_For __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_For
#define __pyx_ptype_7flitter_8language_4tree_IfCondition __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_IfCondition
#define __pyx_ptype_7flitter_8language_4tree_IfElse __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_IfElse
#define __pyx_ptype_7flitter_8language_4tree_Function __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree_Function
#define __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_n_s_Add __pyx_mstate_global->__pyx_n_s_Add
#define __pyx_n_s_Add___reduce_cython __pyx_mstate_global->__pyx_n_s_Add___reduce_cython
#define __pyx_n_s_Add___setstate_cython __pyx_mstate_global->__pyx_n_s_Add___setstate_cython
#define __pyx_n_s_And __pyx_mstate_global->__pyx_n_s_And
#define __pyx_n_s_And___reduce_cython __pyx_mstate_global->__pyx_n_s_And___reduce_cython
#define __pyx_n_s_And___setstate_cython __pyx_mstate_global->__pyx_n_s_And___setstate_cython
#define __pyx_n_s_Append __pyx_mstate_global->__pyx_n_s_Append
#define __pyx_n_s_Append___reduce_cython __pyx_mstate_global->__pyx_n_s_Append___reduce_cython
#define __pyx_n_s_Append___setstate_cython __pyx_mstate_global->__pyx_n_s_Append___setstate_cython
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_Attributes __pyx_mstate_global->__pyx_n_s_Attributes
#define __pyx_n_s_Attributes___reduce_cython __pyx_mstate_global->__pyx_n_s_Attributes___reduce_cython
#define __pyx_n_s_Attributes___setstate_cython __pyx_mstate_global->__pyx_n_s_Attributes___setstate_cython
#define __pyx_n_s_BinaryOperation __pyx_mstate_global->__pyx_n_s_BinaryOperation
#define __pyx_n_s_BinaryOperation___reduce_cython __pyx_mstate_global->__pyx_n_s_BinaryOperation___reduce_cython
#define __pyx_n_s_BinaryOperation___setstate_cytho __pyx_mstate_global->__pyx_n_s_BinaryOperation___setstate_cytho
#define __pyx_kp_u_Binding __pyx_mstate_global->__pyx_kp_u_Binding
#define __pyx_n_s_Binding_2 __pyx_mstate_global->__pyx_n_s_Binding_2
#define __pyx_n_s_Binding___reduce_cython __pyx_mstate_global->__pyx_n_s_Binding___reduce_cython
#define __pyx_n_s_Binding___setstate_cython __pyx_mstate_global->__pyx_n_s_Binding___setstate_cython
#define __pyx_kp_u_Call __pyx_mstate_global->__pyx_kp_u_Call
#define __pyx_n_s_Call_2 __pyx_mstate_global->__pyx_n_s_Call_2
#define __pyx_n_s_Call___reduce_cython __pyx_mstate_global->__pyx_n_s_Call___reduce_cython
#define __pyx_n_s_Call___setstate_cython __pyx_mstate_global->__pyx_n_s_Call___setstate_cython
#define __pyx_n_s_Ceil __pyx_mstate_global->__pyx_n_s_Ceil
#define __pyx_n_s_Ceil___reduce_cython __pyx_mstate_global->__pyx_n_s_Ceil___reduce_cython
#define __pyx_n_s_Ceil___setstate_cython __pyx_mstate_global->__pyx_n_s_Ceil___setstate_cython
#define __pyx_kp_u_Circular_import_of __pyx_mstate_global->__pyx_kp_u_Circular_import_of
#define __pyx_n_s_Comparison __pyx_mstate_global->__pyx_n_s_Comparison
#define __pyx_n_s_Comparison___reduce_cython __pyx_mstate_global->__pyx_n_s_Comparison___reduce_cython
#define __pyx_n_s_Comparison___setstate_cython __pyx_mstate_global->__pyx_n_s_Comparison___setstate_cython
#define __pyx_kp_u_Compiler_error __pyx_mstate_global->__pyx_kp_u_Compiler_error
#define __pyx_n_s_Divide __pyx_mstate_global->__pyx_n_s_Divide
#define __pyx_n_s_Divide___reduce_cython __pyx_mstate_global->__pyx_n_s_Divide___reduce_cython
#define __pyx_n_s_Divide___setstate_cython __pyx_mstate_global->__pyx_n_s_Divide___setstate_cython
#define __pyx_n_s_EqualTo __pyx_mstate_global->__pyx_n_s_EqualTo
#define __pyx_n_s_EqualTo___reduce_cython __pyx_mstate_global->__pyx_n_s_EqualTo___reduce_cython
#define __pyx_n_s_EqualTo___setstate_cython __pyx_mstate_global->__pyx_n_s_EqualTo___setstate_cython
#define __pyx_kp_u_Error_calling __pyx_mstate_global->__pyx_kp_u_Error_calling
#define __pyx_kp_u_Export __pyx_mstate_global->__pyx_kp_u_Export
#define __pyx_n_s_Export_2 __pyx_mstate_global->__pyx_n_s_Export_2
#define __pyx_n_s_Export___reduce_cython __pyx_mstate_global->__pyx_n_s_Export___reduce_cython
#define __pyx_n_s_Export___setstate_cython __pyx_mstate_global->__pyx_n_s_Export___setstate_cython
#define __pyx_n_s_Expression __pyx_mstate_global->__pyx_n_s_Expression
#define __pyx_n_s_Expression___reduce_cython __pyx_mstate_global->__pyx_n_s_Expression___reduce_cython
#define __pyx_n_s_Expression___setstate_cython __pyx_mstate_global->__pyx_n_s_Expression___setstate_cython
#define __pyx_n_s_Expression_compile __pyx_mstate_global->__pyx_n_s_Expression_compile
#define __pyx_n_s_Expression_simplify __pyx_mstate_global->__pyx_n_s_Expression_simplify
#define __pyx_n_u_False __pyx_mstate_global->__pyx_n_u_False
#define __pyx_n_s_Floor __pyx_mstate_global->__pyx_n_s_Floor
#define __pyx_n_s_FloorDivide __pyx_mstate_global->__pyx_n_s_FloorDivide
#define __pyx_n_s_FloorDivide___reduce_cython __pyx_mstate_global->__pyx_n_s_FloorDivide___reduce_cython
#define __pyx_n_s_FloorDivide___setstate_cython __pyx_mstate_global->__pyx_n_s_FloorDivide___setstate_cython
#define __pyx_n_s_Floor___reduce_cython __pyx_mstate_global->__pyx_n_s_Floor___reduce_cython
#define __pyx_n_s_Floor___setstate_cython __pyx_mstate_global->__pyx_n_s_Floor___setstate_cython
#define __pyx_kp_u_For __pyx_mstate_global->__pyx_kp_u_For
#define __pyx_n_s_For_2 __pyx_mstate_global->__pyx_n_s_For_2
#define __pyx_n_s_For___reduce_cython __pyx_mstate_global->__pyx_n_s_For___reduce_cython
#define __pyx_n_s_For___setstate_cython __pyx_mstate_global->__pyx_n_s_For___setstate_cython
#define __pyx_n_s_Fract __pyx_mstate_global->__pyx_n_s_Fract
#define __pyx_n_s_Fract___reduce_cython __pyx_mstate_global->__pyx_n_s_Fract___reduce_cython
#define __pyx_n_s_Fract___setstate_cython __pyx_mstate_global->__pyx_n_s_Fract___setstate_cython
#define __pyx_kp_u_Function __pyx_mstate_global->__pyx_kp_u_Function
#define __pyx_n_s_Function_2 __pyx_mstate_global->__pyx_n_s_Function_2
#define __pyx_n_s_Function___reduce_cython __pyx_mstate_global->__pyx_n_s_Function___reduce_cython
#define __pyx_n_s_Function___setstate_cython __pyx_mstate_global->__pyx_n_s_Function___setstate_cython
#define __pyx_n_s_GreaterThan __pyx_mstate_global->__pyx_n_s_GreaterThan
#define __pyx_n_s_GreaterThanOrEqualTo __pyx_mstate_global->__pyx_n_s_GreaterThanOrEqualTo
#define __pyx_n_s_GreaterThanOrEqualTo___reduce_cy __pyx_mstate_global->__pyx_n_s_GreaterThanOrEqualTo___reduce_cy
#define __pyx_n_s_GreaterThanOrEqualTo___setstate __pyx_mstate_global->__pyx_n_s_GreaterThanOrEqualTo___setstate
#define __pyx_n_s_GreaterThan___reduce_cython __pyx_mstate_global->__pyx_n_s_GreaterThan___reduce_cython
#define __pyx_n_s_GreaterThan___setstate_cython __pyx_mstate_global->__pyx_n_s_GreaterThan___setstate_cython
#define __pyx_kp_u_IfCondition __pyx_mstate_global->__pyx_kp_u_IfCondition
#define __pyx_n_s_IfCondition_2 __pyx_mstate_global->__pyx_n_s_IfCondition_2
#define __pyx_n_s_IfCondition___reduce_cython __pyx_mstate_global->__pyx_n_s_IfCondition___reduce_cython
#define __pyx_n_s_IfCondition___setstate_cython __pyx_mstate_global->__pyx_n_s_IfCondition___setstate_cython
#define __pyx_kp_u_IfElse __pyx_mstate_global->__pyx_kp_u_IfElse
#define __pyx_n_s_IfElse_2 __pyx_mstate_global->__pyx_n_s_IfElse_2
#define __pyx_n_s_IfElse___reduce_cython __pyx_mstate_global->__pyx_n_s_IfElse___reduce_cython
#define __pyx_n_s_IfElse___setstate_cython __pyx_mstate_global->__pyx_n_s_IfElse___setstate_cython
#define __pyx_kp_u_Import __pyx_mstate_global->__pyx_kp_u_Import
#define __pyx_n_s_Import_2 __pyx_mstate_global->__pyx_n_s_Import_2
#define __pyx_n_s_Import___reduce_cython __pyx_mstate_global->__pyx_n_s_Import___reduce_cython
#define __pyx_n_s_Import___setstate_cython __pyx_mstate_global->__pyx_n_s_Import___setstate_cython
#define __pyx_n_s_Import__simplify_locals_genexpr __pyx_mstate_global->__pyx_n_s_Import__simplify_locals_genexpr
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9
#define __pyx_n_s_LessThan __pyx_mstate_global->__pyx_n_s_LessThan
#define __pyx_n_s_LessThanOrEqualTo __pyx_mstate_global->__pyx_n_s_LessThanOrEqualTo
#define __pyx_n_s_LessThanOrEqualTo___reduce_cytho __pyx_mstate_global->__pyx_n_s_LessThanOrEqualTo___reduce_cytho
#define __pyx_n_s_LessThanOrEqualTo___setstate_cyt __pyx_mstate_global->__pyx_n_s_LessThanOrEqualTo___setstate_cyt
#define __pyx_n_s_LessThan___reduce_cython __pyx_mstate_global->__pyx_n_s_LessThan___reduce_cython
#define __pyx_n_s_LessThan___setstate_cython __pyx_mstate_global->__pyx_n_s_LessThan___setstate_cython
#define __pyx_kp_u_Let __pyx_mstate_global->__pyx_kp_u_Let
#define __pyx_n_s_Let_2 __pyx_mstate_global->__pyx_n_s_Let_2
#define __pyx_n_s_Let___reduce_cython __pyx_mstate_global->__pyx_n_s_Let___reduce_cython
#define __pyx_n_s_Let___setstate_cython __pyx_mstate_global->__pyx_n_s_Let___setstate_cython
#define __pyx_kp_u_Literal __pyx_mstate_global->__pyx_kp_u_Literal
#define __pyx_n_s_Literal_2 __pyx_mstate_global->__pyx_n_s_Literal_2
#define __pyx_n_s_Literal___reduce_cython __pyx_mstate_global->__pyx_n_s_Literal___reduce_cython
#define __pyx_n_s_Literal___setstate_cython __pyx_mstate_global->__pyx_n_s_Literal___setstate_cython
#define __pyx_kp_u_Lookup __pyx_mstate_global->__pyx_kp_u_Lookup
#define __pyx_n_s_Lookup_2 __pyx_mstate_global->__pyx_n_s_Lookup_2
#define __pyx_n_s_Lookup___reduce_cython __pyx_mstate_global->__pyx_n_s_Lookup___reduce_cython
#define __pyx_n_s_Lookup___setstate_cython __pyx_mstate_global->__pyx_n_s_Lookup___setstate_cython
#define __pyx_n_s_MathsBinaryOperation __pyx_mstate_global->__pyx_n_s_MathsBinaryOperation
#define __pyx_n_s_MathsBinaryOperation___reduce_cy __pyx_mstate_global->__pyx_n_s_MathsBinaryOperation___reduce_cy
#define __pyx_n_s_MathsBinaryOperation___setstate __pyx_mstate_global->__pyx_n_s_MathsBinaryOperation___setstate
#define __pyx_n_s_Modulo __pyx_mstate_global->__pyx_n_s_Modulo
#define __pyx_n_s_Modulo___reduce_cython __pyx_mstate_global->__pyx_n_s_Modulo___reduce_cython
#define __pyx_n_s_Modulo___setstate_cython __pyx_mstate_global->__pyx_n_s_Modulo___setstate_cython
#define __pyx_n_s_Multiply __pyx_mstate_global->__pyx_n_s_Multiply
#define __pyx_n_s_Multiply___reduce_cython __pyx_mstate_global->__pyx_n_s_Multiply___reduce_cython
#define __pyx_n_s_Multiply___setstate_cython __pyx_mstate_global->__pyx_n_s_Multiply___setstate_cython
#define __pyx_kp_u_Name __pyx_mstate_global->__pyx_kp_u_Name
#define __pyx_n_s_Name_2 __pyx_mstate_global->__pyx_n_s_Name_2
#define __pyx_n_s_Name___reduce_cython __pyx_mstate_global->__pyx_n_s_Name___reduce_cython
#define __pyx_n_s_Name___setstate_cython __pyx_mstate_global->__pyx_n_s_Name___setstate_cython
#define __pyx_n_s_Negative __pyx_mstate_global->__pyx_n_s_Negative
#define __pyx_n_s_Negative___reduce_cython __pyx_mstate_global->__pyx_n_s_Negative___reduce_cython
#define __pyx_n_s_Negative___setstate_cython __pyx_mstate_global->__pyx_n_s_Negative___setstate_cython
#define __pyx_n_s_NodeModifier __pyx_mstate_global->__pyx_n_s_NodeModifier
#define __pyx_n_s_NodeModifier___reduce_cython __pyx_mstate_global->__pyx_n_s_NodeModifier___reduce_cython
#define __pyx_n_s_NodeModifier___setstate_cython __pyx_mstate_global->__pyx_n_s_NodeModifier___setstate_cython
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_n_s_Not __pyx_mstate_global->__pyx_n_s_Not
#define __pyx_n_s_NotEqualTo __pyx_mstate_global->__pyx_n_s_NotEqualTo
#define __pyx_n_s_NotEqualTo___reduce_cython __pyx_mstate_global->__pyx_n_s_NotEqualTo___reduce_cython
#define __pyx_n_s_NotEqualTo___setstate_cython __pyx_mstate_global->__pyx_n_s_NotEqualTo___setstate_cython
#define __pyx_n_s_NotImplementedError __pyx_mstate_global->__pyx_n_s_NotImplementedError
#define __pyx_n_s_Not___reduce_cython __pyx_mstate_global->__pyx_n_s_Not___reduce_cython
#define __pyx_n_s_Not___setstate_cython __pyx_mstate_global->__pyx_n_s_Not___setstate_cython
#define __pyx_n_s_Or __pyx_mstate_global->__pyx_n_s_Or
#define __pyx_n_s_Or___reduce_cython __pyx_mstate_global->__pyx_n_s_Or___reduce_cython
#define __pyx_n_s_Or___setstate_cython __pyx_mstate_global->__pyx_n_s_Or___setstate_cython
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_u_PolyBinding __pyx_mstate_global->__pyx_kp_u_PolyBinding
#define __pyx_n_s_PolyBinding_2 __pyx_mstate_global->__pyx_n_s_PolyBinding_2
#define __pyx_n_s_PolyBinding___reduce_cython __pyx_mstate_global->__pyx_n_s_PolyBinding___reduce_cython
#define __pyx_n_s_PolyBinding___setstate_cython __pyx_mstate_global->__pyx_n_s_PolyBinding___setstate_cython
#define __pyx_n_s_Positive __pyx_mstate_global->__pyx_n_s_Positive
#define __pyx_n_s_Positive___reduce_cython __pyx_mstate_global->__pyx_n_s_Positive___reduce_cython
#define __pyx_n_s_Positive___setstate_cython __pyx_mstate_global->__pyx_n_s_Positive___setstate_cython
#define __pyx_n_s_Power __pyx_mstate_global->__pyx_n_s_Power
#define __pyx_n_s_Power___reduce_cython __pyx_mstate_global->__pyx_n_s_Power___reduce_cython
#define __pyx_n_s_Power___setstate_cython __pyx_mstate_global->__pyx_n_s_Power___setstate_cython
#define __pyx_kp_u_Range __pyx_mstate_global->__pyx_kp_u_Range
#define __pyx_n_s_Range_2 __pyx_mstate_global->__pyx_n_s_Range_2
#define __pyx_n_s_Range___reduce_cython __pyx_mstate_global->__pyx_n_s_Range___reduce_cython
#define __pyx_n_s_Range___setstate_cython __pyx_mstate_global->__pyx_n_s_Range___setstate_cython
#define __pyx_kp_u_Sequence __pyx_mstate_global->__pyx_kp_u_Sequence
#define __pyx_n_s_Sequence_2 __pyx_mstate_global->__pyx_n_s_Sequence_2
#define __pyx_n_s_Sequence___reduce_cython __pyx_mstate_global->__pyx_n_s_Sequence___reduce_cython
#define __pyx_n_s_Sequence___setstate_cython __pyx_mstate_global->__pyx_n_s_Sequence___setstate_cython
#define __pyx_n_s_SharedCache __pyx_mstate_global->__pyx_n_s_SharedCache
#define __pyx_kp_u_Simplifier_error __pyx_mstate_global->__pyx_kp_u_Simplifier_error
#define __pyx_kp_u_Slice __pyx_mstate_global->__pyx_kp_u_Slice
#define __pyx_n_s_Slice_2 __pyx_mstate_global->__pyx_n_s_Slice_2
#define __pyx_n_s_Slice___reduce_cython __pyx_mstate_global->__pyx_n_s_Slice___reduce_cython
#define __pyx_n_s_Slice___setstate_cython __pyx_mstate_global->__pyx_n_s_Slice___setstate_cython
#define __pyx_n_s_Subtract __pyx_mstate_global->__pyx_n_s_Subtract
#define __pyx_n_s_Subtract___reduce_cython __pyx_mstate_global->__pyx_n_s_Subtract___reduce_cython
#define __pyx_n_s_Subtract___setstate_cython __pyx_mstate_global->__pyx_n_s_Subtract___setstate_cython
#define __pyx_kp_u_Tag __pyx_mstate_global->__pyx_kp_u_Tag
#define __pyx_n_s_Tag_2 __pyx_mstate_global->__pyx_n_s_Tag_2
#define __pyx_n_s_Tag___reduce_cython __pyx_mstate_global->__pyx_n_s_Tag___reduce_cython
#define __pyx_n_s_Tag___setstate_cython __pyx_mstate_global->__pyx_n_s_Tag___setstate_cython
#define __pyx_kp_u_Top __pyx_mstate_global->__pyx_kp_u_Top
#define __pyx_n_s_Top_2 __pyx_mstate_global->__pyx_n_s_Top_2
#define __pyx_n_s_Top___reduce_cython __pyx_mstate_global->__pyx_n_s_Top___reduce_cython
#define __pyx_n_s_Top___setstate_cython __pyx_mstate_global->__pyx_n_s_Top___setstate_cython
#define __pyx_n_u_True __pyx_mstate_global->__pyx_n_u_True
#define __pyx_kp_u_Unable_to_simplify_program __pyx_mstate_global->__pyx_kp_u_Unable_to_simplify_program
#define __pyx_n_s_UnaryOperation __pyx_mstate_global->__pyx_n_s_UnaryOperation
#define __pyx_n_s_UnaryOperation___reduce_cython __pyx_mstate_global->__pyx_n_s_UnaryOperation___reduce_cython
#define __pyx_n_s_UnaryOperation___setstate_cython __pyx_mstate_global->__pyx_n_s_UnaryOperation___setstate_cython
#define __pyx_kp_u_Unbound_name __pyx_mstate_global->__pyx_kp_u_Unbound_name
#define __pyx_n_s_Xor __pyx_mstate_global->__pyx_n_s_Xor
#define __pyx_n_s_Xor___reduce_cython __pyx_mstate_global->__pyx_n_s_Xor___reduce_cython
#define __pyx_n_s_Xor___setstate_cython __pyx_mstate_global->__pyx_n_s_Xor___setstate_cython
#define __pyx_n_s__186 __pyx_mstate_global->__pyx_n_s__186
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_n_s__32 __pyx_mstate_global->__pyx_n_s__32
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_bindings __pyx_mstate_global->__pyx_n_s_bindings
#define __pyx_n_s_body __pyx_mstate_global->__pyx_n_s_body
#define __pyx_n_s_cache __pyx_mstate_global->__pyx_n_s_cache
#define __pyx_n_s_captures __pyx_mstate_global->__pyx_n_s_captures
#define __pyx_n_u_ceil __pyx_mstate_global->__pyx_n_u_ceil
#define __pyx_n_s_children __pyx_mstate_global->__pyx_n_s_children
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_compile __pyx_mstate_global->__pyx_n_s_compile
#define __pyx_n_s_condition __pyx_mstate_global->__pyx_n_s_condition
#define __pyx_n_s_context __pyx_mstate_global->__pyx_n_s_context
#define __pyx_n_u_context_func __pyx_mstate_global->__pyx_n_u_context_func
#define __pyx_n_s_context_vars __pyx_mstate_global->__pyx_n_s_context_vars
#define __pyx_n_s_dependencies __pyx_mstate_global->__pyx_n_s_dependencies
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_s_difference_update __pyx_mstate_global->__pyx_n_s_difference_update
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_dynamic __pyx_mstate_global->__pyx_n_s_dynamic
#define __pyx_n_s_else __pyx_mstate_global->__pyx_n_s_else
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_s_exc __pyx_mstate_global->__pyx_n_s_exc
#define __pyx_n_s_exception __pyx_mstate_global->__pyx_n_s_exception
#define __pyx_n_s_expr __pyx_mstate_global->__pyx_n_s_expr
#define __pyx_n_s_expressions __pyx_mstate_global->__pyx_n_s_expressions
#define __pyx_n_s_filename __pyx_mstate_global->__pyx_n_s_filename
#define __pyx_n_s_flitter_language_tree __pyx_mstate_global->__pyx_n_s_flitter_language_tree
#define __pyx_n_u_floor __pyx_mstate_global->__pyx_n_u_floor
#define __pyx_n_u_fract __pyx_mstate_global->__pyx_n_u_fract
#define __pyx_n_s_function __pyx_mstate_global->__pyx_n_s_function
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_with_root __pyx_mstate_global->__pyx_n_s_get_with_root
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_initial_errors __pyx_mstate_global->__pyx_n_s_initial_errors
#define __pyx_n_s_initial_lnames __pyx_mstate_global->__pyx_n_s_initial_lnames
#define __pyx_n_s_inlineable __pyx_mstate_global->__pyx_n_s_inlineable
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_keyword_args __pyx_mstate_global->__pyx_n_s_keyword_args
#define __pyx_n_s_left __pyx_mstate_global->__pyx_n_s_left
#define __pyx_n_s_log_errors __pyx_mstate_global->__pyx_n_s_log_errors
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_loguru __pyx_mstate_global->__pyx_n_s_loguru
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_name_patch __pyx_mstate_global->__pyx_n_s_name_patch
#define __pyx_n_s_names __pyx_mstate_global->__pyx_n_s_names
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_node __pyx_mstate_global->__pyx_n_s_node
#define __pyx_n_s_opt __pyx_mstate_global->__pyx_n_s_opt
#define __pyx_n_s_parameters __pyx_mstate_global->__pyx_n_s_parameters
#define __pyx_n_s_parent __pyx_mstate_global->__pyx_n_s_parent
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_pragmas __pyx_mstate_global->__pyx_n_s_pragmas
#define __pyx_n_s_program __pyx_mstate_global->__pyx_n_s_program
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_Add __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Add
#define __pyx_n_s_pyx_unpickle_And __pyx_mstate_global->__pyx_n_s_pyx_unpickle_And
#define __pyx_n_s_pyx_unpickle_Append __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Append
#define __pyx_n_s_pyx_unpickle_Attributes __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Attributes
#define __pyx_n_s_pyx_unpickle_BinaryOperation __pyx_mstate_global->__pyx_n_s_pyx_unpickle_BinaryOperation
#define __pyx_n_s_pyx_unpickle_Binding __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Binding
#define __pyx_n_s_pyx_unpickle_Call __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Call
#define __pyx_n_s_pyx_unpickle_Ceil __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Ceil
#define __pyx_n_s_pyx_unpickle_Comparison __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Comparison
#define __pyx_n_s_pyx_unpickle_Divide __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Divide
#define __pyx_n_s_pyx_unpickle_EqualTo __pyx_mstate_global->__pyx_n_s_pyx_unpickle_EqualTo
#define __pyx_n_s_pyx_unpickle_Export __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Export
#define __pyx_n_s_pyx_unpickle_Expression __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Expression
#define __pyx_n_s_pyx_unpickle_Floor __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Floor
#define __pyx_n_s_pyx_unpickle_FloorDivide __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FloorDivide
#define __pyx_n_s_pyx_unpickle_For __pyx_mstate_global->__pyx_n_s_pyx_unpickle_For
#define __pyx_n_s_pyx_unpickle_Fract __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Fract
#define __pyx_n_s_pyx_unpickle_Function __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Function
#define __pyx_n_s_pyx_unpickle_GreaterThan __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GreaterThan
#define __pyx_n_s_pyx_unpickle_GreaterThanOrEqua __pyx_mstate_global->__pyx_n_s_pyx_unpickle_GreaterThanOrEqua
#define __pyx_n_s_pyx_unpickle_IfCondition __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IfCondition
#define __pyx_n_s_pyx_unpickle_IfElse __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IfElse
#define __pyx_n_s_pyx_unpickle_Import __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Import
#define __pyx_n_s_pyx_unpickle_LessThan __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LessThan
#define __pyx_n_s_pyx_unpickle_LessThanOrEqualTo __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LessThanOrEqualTo
#define __pyx_n_s_pyx_unpickle_Let __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Let
#define __pyx_n_s_pyx_unpickle_Literal __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Literal
#define __pyx_n_s_pyx_unpickle_Lookup __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Lookup
#define __pyx_n_s_pyx_unpickle_MathsBinaryOperat __pyx_mstate_global->__pyx_n_s_pyx_unpickle_MathsBinaryOperat
#define __pyx_n_s_pyx_unpickle_Modulo __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Modulo
#define __pyx_n_s_pyx_unpickle_Multiply __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Multiply
#define __pyx_n_s_pyx_unpickle_Name __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Name
#define __pyx_n_s_pyx_unpickle_Negative __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Negative
#define __pyx_n_s_pyx_unpickle_NodeModifier __pyx_mstate_global->__pyx_n_s_pyx_unpickle_NodeModifier
#define __pyx_n_s_pyx_unpickle_Not __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Not
#define __pyx_n_s_pyx_unpickle_NotEqualTo __pyx_mstate_global->__pyx_n_s_pyx_unpickle_NotEqualTo
#define __pyx_n_s_pyx_unpickle_Or __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Or
#define __pyx_n_s_pyx_unpickle_PolyBinding __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PolyBinding
#define __pyx_n_s_pyx_unpickle_Positive __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Positive
#define __pyx_n_s_pyx_unpickle_Power __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Power
#define __pyx_n_s_pyx_unpickle_Range __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Range
#define __pyx_n_s_pyx_unpickle_Sequence __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Sequence
#define __pyx_n_s_pyx_unpickle_Slice __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Slice
#define __pyx_n_s_pyx_unpickle_Subtract __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Subtract
#define __pyx_n_s_pyx_unpickle_Tag __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Tag
#define __pyx_n_s_pyx_unpickle_Top __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Top
#define __pyx_n_s_pyx_unpickle_UnaryOperation __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UnaryOperation
#define __pyx_n_s_pyx_unpickle_Xor __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Xor
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_read_flitter_top __pyx_mstate_global->__pyx_n_s_read_flitter_top
#define __pyx_n_s_recursive __pyx_mstate_global->__pyx_n_s_recursive
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_return_context __pyx_mstate_global->__pyx_n_s_return_context
#define __pyx_n_s_reversed __pyx_mstate_global->__pyx_n_s_reversed
#define __pyx_n_s_right __pyx_mstate_global->__pyx_n_s_right
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set_pragma __pyx_mstate_global->__pyx_n_s_set_pragma
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_simplify __pyx_mstate_global->__pyx_n_s_simplify
#define __pyx_n_s_source __pyx_mstate_global->__pyx_n_s_source
#define __pyx_kp_s_src_flitter_language_tree_pyx __pyx_mstate_global->__pyx_kp_s_src_flitter_language_tree_pyx
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_static __pyx_mstate_global->__pyx_n_s_static
#define __pyx_n_s_static_exports __pyx_mstate_global->__pyx_n_s_static_exports
#define __pyx_n_s_step __pyx_mstate_global->__pyx_n_s_step
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_tag __pyx_mstate_global->__pyx_n_s_tag
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_tests __pyx_mstate_global->__pyx_n_s_tests
#define __pyx_n_s_then __pyx_mstate_global->__pyx_n_s_then
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_5413075 __pyx_mstate_global->__pyx_int_5413075
#define __pyx_int_16586848 __pyx_mstate_global->__pyx_int_16586848
#define __pyx_int_18525531 __pyx_mstate_global->__pyx_int_18525531
#define __pyx_int_19056541 __pyx_mstate_global->__pyx_int_19056541
#define __pyx_int_19136085 __pyx_mstate_global->__pyx_int_19136085
#define __pyx_int_19657503 __pyx_mstate_global->__pyx_int_19657503
#define __pyx_int_20102196 __pyx_mstate_global->__pyx_int_20102196
#define __pyx_int_21458427 __pyx_mstate_global->__pyx_int_21458427
#define __pyx_int_25936640 __pyx_mstate_global->__pyx_int_25936640
#define __pyx_int_29776540 __pyx_mstate_global->__pyx_int_29776540
#define __pyx_int_33963030 __pyx_mstate_global->__pyx_int_33963030
#define __pyx_int_39246950 __pyx_mstate_global->__pyx_int_39246950
#define __pyx_int_42909475 __pyx_mstate_global->__pyx_int_42909475
#define __pyx_int_46599698 __pyx_mstate_global->__pyx_int_46599698
#define __pyx_int_48890310 __pyx_mstate_global->__pyx_int_48890310
#define __pyx_int_54761546 __pyx_mstate_global->__pyx_int_54761546
#define __pyx_int_57427254 __pyx_mstate_global->__pyx_int_57427254
#define __pyx_int_58806657 __pyx_mstate_global->__pyx_int_58806657
#define __pyx_int_60181930 __pyx_mstate_global->__pyx_int_60181930
#define __pyx_int_60997603 __pyx_mstate_global->__pyx_int_60997603
#define __pyx_int_61976268 __pyx_mstate_global->__pyx_int_61976268
#define __pyx_int_63365304 __pyx_mstate_global->__pyx_int_63365304
#define __pyx_int_64598595 __pyx_mstate_global->__pyx_int_64598595
#define __pyx_int_65788850 __pyx_mstate_global->__pyx_int_65788850
#define __pyx_int_81599198 __pyx_mstate_global->__pyx_int_81599198
#define __pyx_int_87851177 __pyx_mstate_global->__pyx_int_87851177
#define __pyx_int_88468051 __pyx_mstate_global->__pyx_int_88468051
#define __pyx_int_89160345 __pyx_mstate_global->__pyx_int_89160345
#define __pyx_int_89314571 __pyx_mstate_global->__pyx_int_89314571
#define __pyx_int_101425629 __pyx_mstate_global->__pyx_int_101425629
#define __pyx_int_103159781 __pyx_mstate_global->__pyx_int_103159781
#define __pyx_int_112105877 __pyx_mstate_global->__pyx_int_112105877
#define __pyx_int_114850062 __pyx_mstate_global->__pyx_int_114850062
#define __pyx_int_126472276 __pyx_mstate_global->__pyx_int_126472276
#define __pyx_int_132590791 __pyx_mstate_global->__pyx_int_132590791
#define __pyx_int_134496108 __pyx_mstate_global->__pyx_int_134496108
#define __pyx_int_136983863 __pyx_mstate_global->__pyx_int_136983863
#define __pyx_int_147834117 __pyx_mstate_global->__pyx_int_147834117
#define __pyx_int_147971132 __pyx_mstate_global->__pyx_int_147971132
#define __pyx_int_149349987 __pyx_mstate_global->__pyx_int_149349987
#define __pyx_int_158473136 __pyx_mstate_global->__pyx_int_158473136
#define __pyx_int_161766002 __pyx_mstate_global->__pyx_int_161766002
#define __pyx_int_168508340 __pyx_mstate_global->__pyx_int_168508340
#define __pyx_int_174256674 __pyx_mstate_global->__pyx_int_174256674
#define __pyx_int_184977713 __pyx_mstate_global->__pyx_int_184977713
#define __pyx_int_193537176 __pyx_mstate_global->__pyx_int_193537176
#define __pyx_int_201276235 __pyx_mstate_global->__pyx_int_201276235
#define __pyx_int_209144725 __pyx_mstate_global->__pyx_int_209144725
#define __pyx_int_209608191 __pyx_mstate_global->__pyx_int_209608191
#define __pyx_int_211607581 __pyx_mstate_global->__pyx_int_211607581
#define __pyx_int_211749586 __pyx_mstate_global->__pyx_int_211749586
#define __pyx_int_215229444 __pyx_mstate_global->__pyx_int_215229444
#define __pyx_int_217770541 __pyx_mstate_global->__pyx_int_217770541
#define __pyx_int_221199149 __pyx_mstate_global->__pyx_int_221199149
#define __pyx_int_222419149 __pyx_mstate_global->__pyx_int_222419149
#define __pyx_int_222754444 __pyx_mstate_global->__pyx_int_222754444
#define __pyx_int_228825662 __pyx_mstate_global->__pyx_int_228825662
#define __pyx_int_231565212 __pyx_mstate_global->__pyx_int_231565212
#define __pyx_int_237118084 __pyx_mstate_global->__pyx_int_237118084
#define __pyx_int_238750788 __pyx_mstate_global->__pyx_int_238750788
#define __pyx_int_241444487 __pyx_mstate_global->__pyx_int_241444487
#define __pyx_int_251934659 __pyx_mstate_global->__pyx_int_251934659
#define __pyx_int_253263515 __pyx_mstate_global->__pyx_int_253263515
#define __pyx_int_254981756 __pyx_mstate_global->__pyx_int_254981756
#define __pyx_int_256597255 __pyx_mstate_global->__pyx_int_256597255
#define __pyx_int_256930855 __pyx_mstate_global->__pyx_int_256930855
#define __pyx_int_257251258 __pyx_mstate_global->__pyx_int_257251258
#define __pyx_int_257836063 __pyx_mstate_global->__pyx_int_257836063
#define __pyx_int_261002861 __pyx_mstate_global->__pyx_int_261002861
#define __pyx_int_262293510 __pyx_mstate_global->__pyx_int_262293510
#define __pyx_int_264124616 __pyx_mstate_global->__pyx_int_264124616
#define __pyx_int_264265876 __pyx_mstate_global->__pyx_int_264265876
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__89 __pyx_mstate_global->__pyx_codeobj__89
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__91 __pyx_mstate_global->__pyx_codeobj__91
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
#define __pyx_codeobj__96 __pyx_mstate_global->__pyx_codeobj__96
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__98 __pyx_mstate_global->__pyx_codeobj__98
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__106 __pyx_mstate_global->__pyx_codeobj__106
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__108 __pyx_mstate_global->__pyx_codeobj__108
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__111 __pyx_mstate_global->__pyx_codeobj__111
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#define __pyx_codeobj__118 __pyx_mstate_global->__pyx_codeobj__118
#define __pyx_codeobj__119 __pyx_mstate_global->__pyx_codeobj__119
#define __pyx_codeobj__120 __pyx_mstate_global->__pyx_codeobj__120
#define __pyx_codeobj__121 __pyx_mstate_global->__pyx_codeobj__121
#define __pyx_codeobj__122 __pyx_mstate_global->__pyx_codeobj__122
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__124 __pyx_mstate_global->__pyx_codeobj__124
#define __pyx_codeobj__125 __pyx_mstate_global->__pyx_codeobj__125
#define __pyx_codeobj__126 __pyx_mstate_global->__pyx_codeobj__126
#define __pyx_codeobj__127 __pyx_mstate_global->__pyx_codeobj__127
#define __pyx_codeobj__128 __pyx_mstate_global->__pyx_codeobj__128
#define __pyx_codeobj__129 __pyx_mstate_global->__pyx_codeobj__129
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__131 __pyx_mstate_global->__pyx_codeobj__131
#define __pyx_codeobj__132 __pyx_mstate_global->__pyx_codeobj__132
#define __pyx_codeobj__133 __pyx_mstate_global->__pyx_codeobj__133
#define __pyx_codeobj__134 __pyx_mstate_global->__pyx_codeobj__134
#define __pyx_codeobj__135 __pyx_mstate_global->__pyx_codeobj__135
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__138 __pyx_mstate_global->__pyx_codeobj__138
#define __pyx_codeobj__139 __pyx_mstate_global->__pyx_codeobj__139
#define __pyx_codeobj__140 __pyx_mstate_global->__pyx_codeobj__140
#define __pyx_codeobj__141 __pyx_mstate_global->__pyx_codeobj__141
#define __pyx_codeobj__142 __pyx_mstate_global->__pyx_codeobj__142
#define __pyx_codeobj__143 __pyx_mstate_global->__pyx_codeobj__143
#define __pyx_codeobj__144 __pyx_mstate_global->__pyx_codeobj__144
#define __pyx_codeobj__145 __pyx_mstate_global->__pyx_codeobj__145
#define __pyx_codeobj__146 __pyx_mstate_global->__pyx_codeobj__146
#define __pyx_codeobj__147 __pyx_mstate_global->__pyx_codeobj__147
#define __pyx_codeobj__148 __pyx_mstate_global->__pyx_codeobj__148
#define __pyx_codeobj__149 __pyx_mstate_global->__pyx_codeobj__149
#define __pyx_codeobj__150 __pyx_mstate_global->__pyx_codeobj__150
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__152 __pyx_mstate_global->__pyx_codeobj__152
#define __pyx_codeobj__153 __pyx_mstate_global->__pyx_codeobj__153
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__155 __pyx_mstate_global->__pyx_codeobj__155
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__157 __pyx_mstate_global->__pyx_codeobj__157
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__159 __pyx_mstate_global->__pyx_codeobj__159
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__161 __pyx_mstate_global->__pyx_codeobj__161
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
#define __pyx_codeobj__163 __pyx_mstate_global->__pyx_codeobj__163
#define __pyx_codeobj__164 __pyx_mstate_global->__pyx_codeobj__164
#define __pyx_codeobj__165 __pyx_mstate_global->__pyx_codeobj__165
#define __pyx_codeobj__166 __pyx_mstate_global->__pyx_codeobj__166
#define __pyx_codeobj__167 __pyx_mstate_global->__pyx_codeobj__167
#define __pyx_codeobj__168 __pyx_mstate_global->__pyx_codeobj__168
#define __pyx_codeobj__169 __pyx_mstate_global->__pyx_codeobj__169
#define __pyx_codeobj__170 __pyx_mstate_global->__pyx_codeobj__170
#define __pyx_codeobj__171 __pyx_mstate_global->__pyx_codeobj__171
#define __pyx_codeobj__172 __pyx_mstate_global->__pyx_codeobj__172
#define __pyx_codeobj__173 __pyx_mstate_global->__pyx_codeobj__173
#define __pyx_codeobj__174 __pyx_mstate_global->__pyx_codeobj__174
#define __pyx_codeobj__175 __pyx_mstate_global->__pyx_codeobj__175
#define __pyx_codeobj__176 __pyx_mstate_global->__pyx_codeobj__176
#define __pyx_codeobj__177 __pyx_mstate_global->__pyx_codeobj__177
#define __pyx_codeobj__178 __pyx_mstate_global->__pyx_codeobj__178
#define __pyx_codeobj__179 __pyx_mstate_global->__pyx_codeobj__179
#define __pyx_codeobj__180 __pyx_mstate_global->__pyx_codeobj__180
#define __pyx_codeobj__181 __pyx_mstate_global->__pyx_codeobj__181
#define __pyx_codeobj__182 __pyx_mstate_global->__pyx_codeobj__182
#define __pyx_codeobj__183 __pyx_mstate_global->__pyx_codeobj__183
#define __pyx_codeobj__184 __pyx_mstate_global->__pyx_codeobj__184
#define __pyx_codeobj__185 __pyx_mstate_global->__pyx_codeobj__185
/* #### Code section: module_code ### */

/* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_UPDATE(uint64_t __pyx_v__hash, uint64_t __pyx_v_y) {
  uint64_t __pyx_r;

  /* "model.pxd":13
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y             # <<<<<<<<<<<<<<
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 */
  __pyx_v__hash = (__pyx_v__hash ^ __pyx_v_y);

  /* "model.pxd":14
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 */
  __pyx_v__hash = (__pyx_v__hash + ((uint64_t)0x9e3779b97f4a7c15));

  /* "model.pxd":15
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 30));

  /* "model.pxd":16
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0xbf58476d1ce4e5b9));

  /* "model.pxd":17
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 27));

  /* "model.pxd":18
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 31
 *     return _hash
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0x94d049bb133111eb));

  /* "model.pxd":19
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31             # <<<<<<<<<<<<<<
 *     return _hash
 * 
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 31));

  /* "model.pxd":20
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 *     return _hash             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v__hash;
  goto __pyx_L0;

  /* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_STRING(PyObject *__pyx_v_value) {
  void *__pyx_v_data;
  uint64_t __pyx_v_i;
  uint64_t __pyx_v_n;
  uint64_t __pyx_v_kind;
  Py_UCS4 __pyx_v_c;
  uint64_t __pyx_v_y;
  uint64_t __pyx_r;
  uint64_t __pyx_t_1;
  uint64_t __pyx_t_2;
  uint64_t __pyx_t_3;

  /* "model.pxd":25
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)             # <<<<<<<<<<<<<<
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 */
  __pyx_v_data = PyUnicode_DATA(__pyx_v_value);

  /* "model.pxd":26
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)             # <<<<<<<<<<<<<<
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 */
  __pyx_v_n = PyUnicode_GET_LENGTH(__pyx_v_value);
  __pyx_v_kind = PyUnicode_KIND(__pyx_v_value);

  /* "model.pxd":28
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 */
  __pyx_v_y = ((uint64_t)0xcbf29ce484222325);

  /* "model.pxd":29
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "model.pxd":30
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)             # <<<<<<<<<<<<<<
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y
 */
    __pyx_v_c = PyUnicode_READ(__pyx_v_kind, __pyx_v_data, __pyx_v_i);

    /* "model.pxd":31
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)             # <<<<<<<<<<<<<<
 *     return y
 * 
 */
    __pyx_v_y = ((__pyx_v_y ^ ((uint64_t)__pyx_v_c)) * ((uint64_t)0x100000001b3));
  }

  /* "model.pxd":32
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_y;
  goto __pyx_L0;

  /* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":25
 * 
 * 
 * cdef bint sequence_pack(list expressions):             # <<<<<<<<<<<<<<
 *     cdef Expression expr
 *     cdef bint touched = False
 */

static int __pyx_f_7flitter_8language_4tree_sequence_pack(PyObject *__pyx_v_expressions) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  int __pyx_v_touched;
  PyObject *__pyx_v_vectors = 0;
  PyObject *__pyx_v_todo = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sequence_pack", 1);

  /* "flitter/language/tree.pyx":27
 * cdef bint sequence_pack(list expressions):
 *     cdef Expression expr
 *     cdef bint touched = False             # <<<<<<<<<<<<<<
 *     cdef list vectors, todo=[]
 *     while expressions:
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":28
 *     cdef Expression expr
 *     cdef bint touched = False
 *     cdef list vectors, todo=[]             # <<<<<<<<<<<<<<
 *     while expressions:
 *         todo.append(expressions.pop())
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_todo = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":29
 *     cdef bint touched = False
 *     cdef list vectors, todo=[]
 *     while expressions:             # <<<<<<<<<<<<<<
 *         todo.append(expressions.pop())
 *     while todo:
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_expressions != Py_None)&&(PyList_GET_SIZE(__pyx_v_expressions) != 0);
    if (!__pyx_t_2) break;

    /* "flitter/language/tree.pyx":30
 *     cdef list vectors, todo=[]
 *     while expressions:
 *         todo.append(expressions.pop())             # <<<<<<<<<<<<<<
 *     while todo:
 *         expr = <Expression>todo.pop()
 */
    if (unlikely(__pyx_v_expressions == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(0, 30, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_expressions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_todo, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "flitter/language/tree.pyx":31
 *     while expressions:
 *         todo.append(expressions.pop())
 *     while todo:             # <<<<<<<<<<<<<<
 *         expr = <Expression>todo.pop()
 *         if todo and isinstance(expr, Literal):
 */
  while (1) {
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_todo) != 0);
    if (!__pyx_t_2) break;

    /* "flitter/language/tree.pyx":32
 *         todo.append(expressions.pop())
 *     while todo:
 *         expr = <Expression>todo.pop()             # <<<<<<<<<<<<<<
 *         if todo and isinstance(expr, Literal):
 *             vectors = [(<Literal>expr).value]
 */
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_todo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":33
 *     while todo:
 *         expr = <Expression>todo.pop()
 *         if todo and isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             vectors = [(<Literal>expr).value]
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 */
    __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_todo) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
    __pyx_t_2 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":34
 *         expr = <Expression>todo.pop()
 *         if todo and isinstance(expr, Literal):
 *             vectors = [(<Literal>expr).value]             # <<<<<<<<<<<<<<
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 *                 vectors.append((<Literal>todo.pop()).value)
 */
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value);
      __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value))) __PYX_ERR(0, 34, __pyx_L1_error);
      __Pyx_XDECREF_SET(__pyx_v_vectors, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":35
 *         if todo and isinstance(expr, Literal):
 *             vectors = [(<Literal>expr).value]
 *             while todo and isinstance(todo[len(todo)-1], Literal):             # <<<<<<<<<<<<<<
 *                 vectors.append((<Literal>todo.pop()).value)
 *             if len(vectors) > 1:
 */
      while (1) {
        __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_todo) != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_2 = __pyx_t_5;
          goto __pyx_L12_bool_binop_done;
        }
        __pyx_t_6 = __Pyx_PyList_GET_SIZE(__pyx_v_todo); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 35, __pyx_L1_error)
        __pyx_t_7 = (__pyx_t_6 - 1);
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_todo, __pyx_t_7, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_TypeCheck(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Literal); 
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_2 = __pyx_t_5;
        __pyx_L12_bool_binop_done:;
        if (!__pyx_t_2) break;

        /* "flitter/language/tree.pyx":36
 *             vectors = [(<Literal>expr).value]
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 *                 vectors.append((<Literal>todo.pop()).value)             # <<<<<<<<<<<<<<
 *             if len(vectors) > 1:
 *                 expr = Literal(Vector._compose(vectors))
 */
        __pyx_t_4 = __Pyx_PyList_Pop(__pyx_v_todo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_t_4)->value);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_vectors, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 36, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "flitter/language/tree.pyx":37
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 *                 vectors.append((<Literal>todo.pop()).value)
 *             if len(vectors) > 1:             # <<<<<<<<<<<<<<
 *                 expr = Literal(Vector._compose(vectors))
 *                 touched = True
 */
      __pyx_t_7 = __Pyx_PyList_GET_SIZE(__pyx_v_vectors); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 37, __pyx_L1_error)
      __pyx_t_2 = (__pyx_t_7 > 1);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":38
 *                 vectors.append((<Literal>todo.pop()).value)
 *             if len(vectors) > 1:
 *                 expr = Literal(Vector._compose(vectors))             # <<<<<<<<<<<<<<
 *                 touched = True
 *         elif isinstance(expr, Sequence):
 */
        __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_compose(__pyx_v_vectors)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "flitter/language/tree.pyx":39
 *             if len(vectors) > 1:
 *                 expr = Literal(Vector._compose(vectors))
 *                 touched = True             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, Sequence):
 *             for expr in reversed((<Sequence>expr).expressions):
 */
        __pyx_v_touched = 1;

        /* "flitter/language/tree.pyx":37
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 *                 vectors.append((<Literal>todo.pop()).value)
 *             if len(vectors) > 1:             # <<<<<<<<<<<<<<
 *                 expr = Literal(Vector._compose(vectors))
 *                 touched = True
 */
      }

      /* "flitter/language/tree.pyx":33
 *     while todo:
 *         expr = <Expression>todo.pop()
 *         if todo and isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             vectors = [(<Literal>expr).value]
 *             while todo and isinstance(todo[len(todo)-1], Literal):
 */
      goto __pyx_L7;
    }

    /* "flitter/language/tree.pyx":40
 *                 expr = Literal(Vector._compose(vectors))
 *                 touched = True
 *         elif isinstance(expr, Sequence):             # <<<<<<<<<<<<<<
 *             for expr in reversed((<Sequence>expr).expressions):
 *                 todo.append(expr)
 */
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Sequence); 
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":41
 *                 touched = True
 *         elif isinstance(expr, Sequence):
 *             for expr in reversed((<Sequence>expr).expressions):             # <<<<<<<<<<<<<<
 *                 todo.append(expr)
 *             touched = True
 */
      if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_expr)->expressions == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 41, __pyx_L1_error)
      }
      __pyx_t_4 = ((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_expr)->expressions; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 41, __pyx_L1_error)
      #endif
      --__pyx_t_7;
      for (;;) {
        if (__pyx_t_7 < 0) break;
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 41, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7--; if (unlikely((0 < 0))) __PYX_ERR(0, 41, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7--; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 41, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":42
 *         elif isinstance(expr, Sequence):
 *             for expr in reversed((<Sequence>expr).expressions):
 *                 todo.append(expr)             # <<<<<<<<<<<<<<
 *             touched = True
 *             continue
 */
        __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_todo, ((PyObject *)__pyx_v_expr)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 42, __pyx_L1_error)

        /* "flitter/language/tree.pyx":41
 *                 touched = True
 *         elif isinstance(expr, Sequence):
 *             for expr in reversed((<Sequence>expr).expressions):             # <<<<<<<<<<<<<<
 *                 todo.append(expr)
 *             touched = True
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":43
 *             for expr in reversed((<Sequence>expr).expressions):
 *                 todo.append(expr)
 *             touched = True             # <<<<<<<<<<<<<<
 *             continue
 *         expressions.append(expr)
 */
      __pyx_v_touched = 1;

      /* "flitter/language/tree.pyx":44
 *                 todo.append(expr)
 *             touched = True
 *             continue             # <<<<<<<<<<<<<<
 *         expressions.append(expr)
 *     return touched
 */
      goto __pyx_L5_continue;

      /* "flitter/language/tree.pyx":40
 *                 expr = Literal(Vector._compose(vectors))
 *                 touched = True
 *         elif isinstance(expr, Sequence):             # <<<<<<<<<<<<<<
 *             for expr in reversed((<Sequence>expr).expressions):
 *                 todo.append(expr)
 */
    }
    __pyx_L7:;

    /* "flitter/language/tree.pyx":45
 *             touched = True
 *             continue
 *         expressions.append(expr)             # <<<<<<<<<<<<<<
 *     return touched
 * 
 */
    if (unlikely(__pyx_v_expressions == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 45, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_expressions, ((PyObject *)__pyx_v_expr)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 45, __pyx_L1_error)
    __pyx_L5_continue:;
  }

  /* "flitter/language/tree.pyx":46
 *             continue
 *         expressions.append(expr)
 *     return touched             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_touched;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":25
 * 
 * 
 * cdef bint sequence_pack(list expressions):             # <<<<<<<<<<<<<<
 *     cdef Expression expr
 *     cdef bint touched = False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.sequence_pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_vectors);
  __Pyx_XDECREF(__pyx_v_todo);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":50
 * 
 * cdef class Expression:
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):             # <<<<<<<<<<<<<<
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_1compile(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Expression_1compile = {"compile", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_1compile, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_1compile(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_initial_lnames = 0;
  PyObject *__pyx_v_initial_errors = 0;
  int __pyx_v_log_errors;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compile (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_initial_lnames,&__pyx_n_s_initial_errors,&__pyx_n_s_log_errors,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)__pyx_empty_tuple));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_initial_lnames);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_initial_errors);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_log_errors);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "compile") < 0)) __PYX_ERR(0, 50, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_initial_lnames = ((PyObject*)values[0]);
    __pyx_v_initial_errors = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_log_errors = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_log_errors == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)
    } else {
      __pyx_v_log_errors = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compile", 0, 0, 3, __pyx_nargs); __PYX_ERR(0, 50, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Expression.compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_initial_lnames), (&PyTuple_Type), 1, "initial_lnames", 1))) __PYX_ERR(0, 50, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_initial_errors), (&PySet_Type), 1, "initial_errors", 1))) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Expression_compile(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self), __pyx_v_initial_lnames, __pyx_v_initial_errors, __pyx_v_log_errors);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_compile(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, PyObject *__pyx_v_initial_lnames, PyObject *__pyx_v_initial_errors, int __pyx_v_log_errors) {
  struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program = 0;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compile", 1);

  /* "flitter/language/tree.pyx":51
 * cdef class Expression:
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):
 *         cdef Program program = Program.__new__(Program, initial_lnames)             # <<<<<<<<<<<<<<
 *         if initial_errors:
 *             program.compiler_errors.update(initial_errors)
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_initial_lnames);
  __Pyx_GIVEREF(__pyx_v_initial_lnames);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_initial_lnames)) __PYX_ERR(0, 51, __pyx_L1_error);
  __pyx_t_2 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_8language_2vm_Program), ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_8language_2vm_Program)))) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_program = ((struct __pyx_obj_7flitter_8language_2vm_Program *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":52
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:             # <<<<<<<<<<<<<<
 *             program.compiler_errors.update(initial_errors)
 *         self._compile(program, list(initial_lnames))
 */
  __pyx_t_3 = (__pyx_v_initial_errors != Py_None)&&(PySet_GET_SIZE(__pyx_v_initial_errors) != 0);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":53
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:
 *             program.compiler_errors.update(initial_errors)             # <<<<<<<<<<<<<<
 *         self._compile(program, list(initial_lnames))
 *         program.link()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_program->compiler_errors, __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_initial_errors};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":52
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:             # <<<<<<<<<<<<<<
 *             program.compiler_errors.update(initial_errors)
 *         self._compile(program, list(initial_lnames))
 */
  }

  /* "flitter/language/tree.pyx":54
 *         if initial_errors:
 *             program.compiler_errors.update(initial_errors)
 *         self._compile(program, list(initial_lnames))             # <<<<<<<<<<<<<<
 *         program.link()
 *         if log_errors:
 */
  __pyx_t_2 = PySequence_List(__pyx_v_initial_lnames); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_vtab)->_compile(__pyx_v_self, __pyx_v_program, ((PyObject*)__pyx_t_2)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":55
 *             program.compiler_errors.update(initial_errors)
 *         self._compile(program, list(initial_lnames))
 *         program.link()             # <<<<<<<<<<<<<<
 *         if log_errors:
 *             for error in program.compiler_errors:
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->link(__pyx_v_program, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":56
 *         self._compile(program, list(initial_lnames))
 *         program.link()
 *         if log_errors:             # <<<<<<<<<<<<<<
 *             for error in program.compiler_errors:
 *                 logger.warning("Compiler error: {}", error)
 */
  if (__pyx_v_log_errors) {

    /* "flitter/language/tree.pyx":57
 *         program.link()
 *         if log_errors:
 *             for error in program.compiler_errors:             # <<<<<<<<<<<<<<
 *                 logger.warning("Compiler error: {}", error)
 *         return program
 */
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_set_iterator(__pyx_v_program->compiler_errors, 1, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_set_iter_next(__pyx_t_2, __pyx_t_7, &__pyx_t_6, &__pyx_t_1, __pyx_t_8);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_error, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":58
 *         if log_errors:
 *             for error in program.compiler_errors:
 *                 logger.warning("Compiler error: {}", error)             # <<<<<<<<<<<<<<
 *         return program
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warning); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_Compiler_error, __pyx_v_error};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":56
 *         self._compile(program, list(initial_lnames))
 *         program.link()
 *         if log_errors:             # <<<<<<<<<<<<<<
 *             for error in program.compiler_errors:
 *                 logger.warning("Compiler error: {}", error)
 */
  }

  /* "flitter/language/tree.pyx":59
 *             for error in program.compiler_errors:
 *                 logger.warning("Compiler error: {}", error)
 *         return program             # <<<<<<<<<<<<<<
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_program);
  __pyx_r = ((PyObject *)__pyx_v_program);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":50
 * 
 * cdef class Expression:
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):             # <<<<<<<<<<<<<<
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("flitter.language.tree.Expression.compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_program);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":61
 *         return program
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):             # <<<<<<<<<<<<<<
 *         cdef dict context_vars = {}
 *         cdef str key
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_3simplify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Expression_3simplify = {"simplify", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_3simplify, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_3simplify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state = 0;
  PyObject *__pyx_v_static = 0;
  PyObject *__pyx_v_dynamic = 0;
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_parent = 0;
  PyObject *__pyx_v_path = 0;
  int __pyx_v_return_context;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("simplify (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,&__pyx_n_s_static,&__pyx_n_s_dynamic,&__pyx_n_s_parent,&__pyx_n_s_path,&__pyx_n_s_return_context,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL((PyObject *)((struct __pyx_obj_7flitter_5model_StateDict *)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[3] = __Pyx_Arg_NewRef_FASTCALL((PyObject *)((struct __pyx_obj_7flitter_5model_Context *)Py_None));
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_static);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dynamic);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parent);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_return_context);
          if (value) { values[5] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "simplify") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_state = ((struct __pyx_obj_7flitter_5model_StateDict *)values[0]);
    __pyx_v_static = ((PyObject*)values[1]);
    __pyx_v_dynamic = values[2];
    __pyx_v_parent = ((struct __pyx_obj_7flitter_5model_Context *)values[3]);
    __pyx_v_path = values[4];
    if (values[5]) {
      __pyx_v_return_context = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_return_context == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    } else {
      __pyx_v_return_context = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("simplify", 0, 0, 6, __pyx_nargs); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Expression.simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_state), __pyx_ptype_7flitter_5model_StateDict, 1, "state", 0))) __PYX_ERR(0, 61, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_static), (&PyDict_Type), 1, "static", 1))) __PYX_ERR(0, 61, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent), __pyx_ptype_7flitter_5model_Context, 1, "parent", 0))) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Expression_2simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self), __pyx_v_state, __pyx_v_static, __pyx_v_dynamic, __pyx_v_parent, __pyx_v_path, __pyx_v_return_context);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_2simplify(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state, PyObject *__pyx_v_static, PyObject *__pyx_v_dynamic, struct __pyx_obj_7flitter_5model_Context *__pyx_v_parent, PyObject *__pyx_v_path, int __pyx_v_return_context) {
  PyObject *__pyx_v_context_vars = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_context = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  unsigned int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("simplify", 1);

  /* "flitter/language/tree.pyx":62
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):
 *         cdef dict context_vars = {}             # <<<<<<<<<<<<<<
 *         cdef str key
 *         if static is not None:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_context_vars = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":64
 *         cdef dict context_vars = {}
 *         cdef str key
 *         if static is not None:             # <<<<<<<<<<<<<<
 *             for key, value in static.items():
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 */
  __pyx_t_2 = (__pyx_v_static != ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":65
 *         cdef str key
 *         if static is not None:
 *             for key, value in static.items():             # <<<<<<<<<<<<<<
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 *         if dynamic is not None:
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_static == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 65, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_static, 1, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_6))) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":66
 *         if static is not None:
 *             for key, value in static.items():
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)             # <<<<<<<<<<<<<<
 *         if dynamic is not None:
 *             for key in dynamic:
 */
      __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7flitter_8language_4tree_Expression); 
      if (__pyx_t_2) {
        __Pyx_INCREF(__pyx_v_value);
        __pyx_t_7 = __pyx_v_value;
      } else {
        __pyx_t_6 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_coerce(__pyx_v_value)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __pyx_t_6;
        __pyx_t_6 = 0;
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context_vars, __pyx_v_key, __pyx_t_7) < 0))) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":64
 *         cdef dict context_vars = {}
 *         cdef str key
 *         if static is not None:             # <<<<<<<<<<<<<<
 *             for key, value in static.items():
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 */
  }

  /* "flitter/language/tree.pyx":67
 *             for key, value in static.items():
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 *         if dynamic is not None:             # <<<<<<<<<<<<<<
 *             for key in dynamic:
 *                 context_vars[key] = None
 */
  __pyx_t_2 = (__pyx_v_dynamic != Py_None);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":68
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 *         if dynamic is not None:
 *             for key in dynamic:             # <<<<<<<<<<<<<<
 *                 context_vars[key] = None
 *         cdef Context context = Context(state=state, names=context_vars)
 */
    if (likely(PyList_CheckExact(__pyx_v_dynamic)) || PyTuple_CheckExact(__pyx_v_dynamic)) {
      __pyx_t_1 = __pyx_v_dynamic; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_dynamic); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 68, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 68, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 68, __pyx_L1_error)
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 68, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 68, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 68, __pyx_L1_error)
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 68, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 68, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 68, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":69
 *         if dynamic is not None:
 *             for key in dynamic:
 *                 context_vars[key] = None             # <<<<<<<<<<<<<<
 *         cdef Context context = Context(state=state, names=context_vars)
 *         context.path = path
 */
      if (unlikely((PyDict_SetItem(__pyx_v_context_vars, __pyx_v_key, Py_None) < 0))) __PYX_ERR(0, 69, __pyx_L1_error)

      /* "flitter/language/tree.pyx":68
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 *         if dynamic is not None:
 *             for key in dynamic:             # <<<<<<<<<<<<<<
 *                 context_vars[key] = None
 *         cdef Context context = Context(state=state, names=context_vars)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":67
 *             for key, value in static.items():
 *                 context_vars[key] = value if isinstance(value, Expression) else Vector._coerce(value)
 *         if dynamic is not None:             # <<<<<<<<<<<<<<
 *             for key in dynamic:
 *                 context_vars[key] = None
 */
  }

  /* "flitter/language/tree.pyx":70
 *             for key in dynamic:
 *                 context_vars[key] = None
 *         cdef Context context = Context(state=state, names=context_vars)             # <<<<<<<<<<<<<<
 *         context.path = path
 *         context.parent = parent
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_state, ((PyObject *)__pyx_v_state)) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_names, __pyx_v_context_vars) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_5model_Context), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_context = ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "flitter/language/tree.pyx":71
 *                 context_vars[key] = None
 *         cdef Context context = Context(state=state, names=context_vars)
 *         context.path = path             # <<<<<<<<<<<<<<
 *         context.parent = parent
 *         cdef Expression expr = self
 */
  __Pyx_INCREF(__pyx_v_path);
  __Pyx_GIVEREF(__pyx_v_path);
  __Pyx_GOTREF(__pyx_v_context->path);
  __Pyx_DECREF(__pyx_v_context->path);
  __pyx_v_context->path = __pyx_v_path;

  /* "flitter/language/tree.pyx":72
 *         cdef Context context = Context(state=state, names=context_vars)
 *         context.path = path
 *         context.parent = parent             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self
 *         try:
 */
  __Pyx_INCREF((PyObject *)__pyx_v_parent);
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent);
  __Pyx_GOTREF((PyObject *)__pyx_v_context->parent);
  __Pyx_DECREF((PyObject *)__pyx_v_context->parent);
  __pyx_v_context->parent = __pyx_v_parent;

  /* "flitter/language/tree.pyx":73
 *         context.path = path
 *         context.parent = parent
 *         cdef Expression expr = self             # <<<<<<<<<<<<<<
 *         try:
 *             expr = expr._simplify(context)
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_v_expr = __pyx_v_self;

  /* "flitter/language/tree.pyx":74
 *         context.parent = parent
 *         cdef Expression expr = self
 *         try:             # <<<<<<<<<<<<<<
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    /*try:*/ {

      /* "flitter/language/tree.pyx":75
 *         cdef Expression expr = self
 *         try:
 *             expr = expr._simplify(context)             # <<<<<<<<<<<<<<
 *         except Exception as exc:                                             # pragma: no cover
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 */
      __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 75, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":74
 *         context.parent = parent
 *         cdef Expression expr = self
 *         try:             # <<<<<<<<<<<<<<
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L15_try_end;
    __pyx_L10_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":76
 *         try:
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover             # <<<<<<<<<<<<<<
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 *         if return_context:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("flitter.language.tree.Expression.simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 76, __pyx_L12_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      /*try:*/ {

        /* "flitter/language/tree.pyx":77
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover             # <<<<<<<<<<<<<<
 *         if return_context:
 *             return expr, context
 */
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_logger); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 77, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_opt); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 77, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 77, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_exception, __pyx_v_exc) < 0) __PYX_ERR(0, 77, __pyx_L21_error)
        __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_empty_tuple, __pyx_t_14); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 77, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_warning); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 77, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = NULL;
        __pyx_t_17 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
            __pyx_t_17 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_kp_u_Unable_to_simplify_program};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 77, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }

      /* "flitter/language/tree.pyx":76
 *         try:
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover             # <<<<<<<<<<<<<<
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 *         if return_context:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L22;
        }
        __pyx_L21_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_24);
          __pyx_t_5 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
          }
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_18;
          goto __pyx_L12_except_error;
        }
        __pyx_L22:;
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L11_exception_handled;
    }
    goto __pyx_L12_except_error;

    /* "flitter/language/tree.pyx":74
 *         context.parent = parent
 *         cdef Expression expr = self
 *         try:             # <<<<<<<<<<<<<<
 *             expr = expr._simplify(context)
 *         except Exception as exc:                                             # pragma: no cover
 */
    __pyx_L12_except_error:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    goto __pyx_L1_error;
    __pyx_L11_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    __pyx_L15_try_end:;
  }

  /* "flitter/language/tree.pyx":78
 *         except Exception as exc:                                             # pragma: no cover
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 *         if return_context:             # <<<<<<<<<<<<<<
 *             return expr, context
 *         else:
 */
  if (__pyx_v_return_context) {

    /* "flitter/language/tree.pyx":79
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 *         if return_context:
 *             return expr, context             # <<<<<<<<<<<<<<
 *         else:
 *             for error in context.errors:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF((PyObject *)__pyx_v_expr);
    __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 79, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_context);
    __Pyx_GIVEREF((PyObject *)__pyx_v_context);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_context))) __PYX_ERR(0, 79, __pyx_L1_error);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":78
 *         except Exception as exc:                                             # pragma: no cover
 *             logger.opt(exception=exc).warning("Unable to simplify program")  # pragma: no cover
 *         if return_context:             # <<<<<<<<<<<<<<
 *             return expr, context
 *         else:
 */
  }

  /* "flitter/language/tree.pyx":81
 *             return expr, context
 *         else:
 *             for error in context.errors:             # <<<<<<<<<<<<<<
 *                 logger.warning("Simplifier error: {}", error)
 *         return expr
 */
  /*else*/ {
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_set_iterator(__pyx_v_context->errors, 1, (&__pyx_t_3), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_6);
    __pyx_t_6 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_5 = __Pyx_set_iter_next(__pyx_t_6, __pyx_t_3, &__pyx_t_4, &__pyx_t_1, __pyx_t_8);
      if (unlikely(__pyx_t_5 == 0)) break;
      if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_error, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":82
 *         else:
 *             for error in context.errors:
 *                 logger.warning("Simplifier error: {}", error)             # <<<<<<<<<<<<<<
 *         return expr
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_logger); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warning); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_kp_u_Simplifier_error, __pyx_v_error};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_17, 2+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "flitter/language/tree.pyx":83
 *             for error in context.errors:
 *                 logger.warning("Simplifier error: {}", error)
 *         return expr             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __pyx_r = ((PyObject *)__pyx_v_expr);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":61
 *         return program
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):             # <<<<<<<<<<<<<<
 *         cdef dict context_vars = {}
 *         cdef str key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("flitter.language.tree.Expression.simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context_vars);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_context);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":85
 *         return expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_10Expression__compile(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, CYTHON_UNUSED PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":86
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 86, __pyx_L1_error)

  /* "flitter/language/tree.pyx":85
 *         return expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Expression._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":88
 *         raise NotImplementedError()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_10Expression__simplify(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":89
 * 
 *     cdef Expression _simplify(self, Context context):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 89, __pyx_L1_error)

  /* "flitter/language/tree.pyx":88
 *         raise NotImplementedError()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Expression._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Expression_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Expression_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Expression); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Expression.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Expression_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Expression_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Expression.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Expression_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Expression_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Expression__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Expression.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":97
 *     cdef readonly set dependencies
 * 
 *     def __init__(self, tuple pragmas, Expression body, set dependencies=None):             # <<<<<<<<<<<<<<
 *         self.pragmas = pragmas
 *         self.body = body
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_3Top_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_3Top_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pragmas = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  PyObject *__pyx_v_dependencies = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pragmas,&__pyx_n_s_body,&__pyx_n_s_dependencies,0};
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pragmas)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_body)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 97, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dependencies);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 97, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_pragmas = ((PyObject*)values[0]);
    __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
    __pyx_v_dependencies = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 97, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Top.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pragmas), (&PyTuple_Type), 1, "pragmas", 1))) __PYX_ERR(0, 97, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_body), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "body", 0))) __PYX_ERR(0, 97, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dependencies), (&PySet_Type), 1, "dependencies", 1))) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top___init__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self), __pyx_v_pragmas, __pyx_v_body, __pyx_v_dependencies);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_3Top___init__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, PyObject *__pyx_v_pragmas, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body, PyObject *__pyx_v_dependencies) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":98
 * 
 *     def __init__(self, tuple pragmas, Expression body, set dependencies=None):
 *         self.pragmas = pragmas             # <<<<<<<<<<<<<<
 *         self.body = body
 *         self.dependencies = dependencies if dependencies is not None else set()
 */
  __Pyx_INCREF(__pyx_v_pragmas);
  __Pyx_GIVEREF(__pyx_v_pragmas);
  __Pyx_GOTREF(__pyx_v_self->pragmas);
  __Pyx_DECREF(__pyx_v_self->pragmas);
  __pyx_v_self->pragmas = __pyx_v_pragmas;

  /* "flitter/language/tree.pyx":99
 *     def __init__(self, tuple pragmas, Expression body, set dependencies=None):
 *         self.pragmas = pragmas
 *         self.body = body             # <<<<<<<<<<<<<<
 *         self.dependencies = dependencies if dependencies is not None else set()
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->body);
  __Pyx_DECREF((PyObject *)__pyx_v_self->body);
  __pyx_v_self->body = __pyx_v_body;

  /* "flitter/language/tree.pyx":100
 *         self.pragmas = pragmas
 *         self.body = body
 *         self.dependencies = dependencies if dependencies is not None else set()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __pyx_t_2 = (__pyx_v_dependencies != ((PyObject*)Py_None));
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_v_dependencies);
    __pyx_t_1 = __pyx_v_dependencies;
  } else {
    __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->dependencies);
  __Pyx_DECREF(__pyx_v_self->dependencies);
  __pyx_v_self->dependencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":97
 *     cdef readonly set dependencies
 * 
 *     def __init__(self, tuple pragmas, Expression body, set dependencies=None):             # <<<<<<<<<<<<<<
 *         self.pragmas = pragmas
 *         self.body = body
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Top.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":102
 *         self.dependencies = dependencies if dependencies is not None else set()
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Binding binding
 *         for binding in self.pragmas:
 */

static void __pyx_f_7flitter_8language_4tree_3Top__compile(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_binding = 0;
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instr = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_append __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":104
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef Binding binding
 *         for binding in self.pragmas:             # <<<<<<<<<<<<<<
 *             program.set_pragma(binding.name, (<Literal>binding.expr).value)
 *         self.body._compile(program, lnames)
 */
  if (unlikely(__pyx_v_self->pragmas == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 104, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->pragmas; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 104, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 104, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":105
 *         cdef Binding binding
 *         for binding in self.pragmas:
 *             program.set_pragma(binding.name, (<Literal>binding.expr).value)             # <<<<<<<<<<<<<<
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_program), __pyx_n_s_set_pragma); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_binding->name, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_binding->expr)->value)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":104
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef Binding binding
 *         for binding in self.pragmas:             # <<<<<<<<<<<<<<
 *             program.set_pragma(binding.name, (<Literal>binding.expr).value)
 *         self.body._compile(program, lnames)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":106
 *         for binding in self.pragmas:
 *             program.set_pragma(binding.name, (<Literal>binding.expr).value)
 *         self.body._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_compile(__pyx_v_self->body, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "flitter/language/tree.pyx":107
 *             program.set_pragma(binding.name, (<Literal>binding.expr).value)
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()             # <<<<<<<<<<<<<<
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":108
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 */
  __pyx_t_7 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_Compose);
  if (__pyx_t_7) {

    /* "flitter/language/tree.pyx":109
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_instr, ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":110
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)             # <<<<<<<<<<<<<<
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 *             program.pop_instruction()
 */
    __pyx_t_8.__pyx_n = 1;
    __pyx_t_8.count = ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instr)->value;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->append(__pyx_v_program, 0, &__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":108
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 */
    goto __pyx_L6;
  }

  /* "flitter/language/tree.pyx":111
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *         else:
 */
  __pyx_t_9 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_Literal);
  if (__pyx_t_9) {
  } else {
    __pyx_t_7 = __pyx_t_9;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_9 = (((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instr)->value->length == 0);
  __pyx_t_7 = __pyx_t_9;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_7) {

    /* "flitter/language/tree.pyx":112
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *         else:
 *             program.append()
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":111
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *         else:
 */
    goto __pyx_L6;
  }

  /* "flitter/language/tree.pyx":114
 *             program.pop_instruction()
 *         else:
 *             program.append()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->append(__pyx_v_program, 0, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L6:;

  /* "flitter/language/tree.pyx":102
 *         self.dependencies = dependencies if dependencies is not None else set()
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Binding binding
 *         for binding in self.pragmas:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Top._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":116
 *             program.append()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression body = self.body._simplify(context)
 *         if body is self.body and context.dependencies == self.dependencies:
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Top__simplify(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":117
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression body = self.body._simplify(context)             # <<<<<<<<<<<<<<
 *         if body is self.body and context.dependencies == self.dependencies:
 *             return self
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_simplify(__pyx_v_self->body, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":118
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression body = self.body._simplify(context)
 *         if body is self.body and context.dependencies == self.dependencies:             # <<<<<<<<<<<<<<
 *             return self
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)
 */
  __pyx_t_3 = (__pyx_v_body == __pyx_v_self->body);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_context->dependencies, __pyx_v_self->dependencies, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":119
 *         cdef Expression body = self.body._simplify(context)
 *         if body is self.body and context.dependencies == self.dependencies:
 *             return self             # <<<<<<<<<<<<<<
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":118
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression body = self.body._simplify(context)
 *         if body is self.body and context.dependencies == self.dependencies:             # <<<<<<<<<<<<<<
 *             return self
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)
 */
  }

  /* "flitter/language/tree.pyx":120
 *         if body is self.body and context.dependencies == self.dependencies:
 *             return self
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyNumber_Xor(__pyx_v_self->dependencies, __pyx_v_context->dependencies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->pragmas);
  __Pyx_GIVEREF(__pyx_v_self->pragmas);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->pragmas)) __PYX_ERR(0, 120, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_body))) __PYX_ERR(0, 120, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Top), __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":116
 *             program.append()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression body = self.body._simplify(context)
 *         if body is self.body and context.dependencies == self.dependencies:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Top._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_body);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":122
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Top({self.pragmas!r}, {self.body!r}, {self.dependencies!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":123
 * 
 *     def __repr__(self):
 *         return f'Top({self.pragmas!r}, {self.body!r}, {self.dependencies!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Top);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_Top);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Top);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->pragmas), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->body)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->dependencies), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":122
 *         return Top(self.pragmas, body, self.dependencies ^ context.dependencies)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Top({self.pragmas!r}, {self.body!r}, {self.dependencies!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Top.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":93
 * 
 * cdef class Top(Expression):
 *     cdef readonly tuple pragmas             # <<<<<<<<<<<<<<
 *     cdef readonly Expression body
 *     cdef readonly set dependencies
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_7pragmas_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_7pragmas_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_7pragmas___get__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_7pragmas___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->pragmas);
  __pyx_r = __pyx_v_self->pragmas;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":94
 * cdef class Top(Expression):
 *     cdef readonly tuple pragmas
 *     cdef readonly Expression body             # <<<<<<<<<<<<<<
 *     cdef readonly set dependencies
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_4body_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_4body___get__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __pyx_r = ((PyObject *)__pyx_v_self->body);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":95
 *     cdef readonly tuple pragmas
 *     cdef readonly Expression body
 *     cdef readonly set dependencies             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple pragmas, Expression body, set dependencies=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_12dependencies_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_12dependencies_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_12dependencies___get__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_12dependencies___get__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->dependencies);
  __pyx_r = __pyx_v_self->dependencies;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Top_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Top_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.body, self.dependencies, self.pragmas)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->body))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->dependencies);
  __Pyx_GIVEREF(__pyx_v_self->dependencies);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->dependencies)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->pragmas);
  __Pyx_GIVEREF(__pyx_v_self->pragmas);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->pragmas)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.body, self.dependencies, self.pragmas)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.body, self.dependencies, self.pragmas)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.body is not None or self.dependencies is not None or self.pragmas is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.body, self.dependencies, self.pragmas)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.body is not None or self.dependencies is not None or self.pragmas is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->body) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->dependencies != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->pragmas != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.dependencies is not None or self.pragmas is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.body is not None or self.dependencies is not None or self.pragmas is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Top); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_193537176);
    __Pyx_GIVEREF(__pyx_int_193537176);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_193537176)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.dependencies is not None or self.pragmas is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, None), state
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Top); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_193537176);
    __Pyx_GIVEREF(__pyx_int_193537176);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_193537176)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Top.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Top_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Top_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Top_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Top.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Top_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Top_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Top__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Top.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":129
 *     cdef readonly dict static_exports
 * 
 *     def __init__(self, dict static_exports=None):             # <<<<<<<<<<<<<<
 *         self.static_exports = static_exports
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_6Export_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_6Export_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_static_exports = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_static_exports,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_static_exports);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 129, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_static_exports = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 129, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Export.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_static_exports), (&PyDict_Type), 1, "static_exports", 1))) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Export___init__(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v_self), __pyx_v_static_exports);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_6Export___init__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, PyObject *__pyx_v_static_exports) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":130
 * 
 *     def __init__(self, dict static_exports=None):
 *         self.static_exports = static_exports             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_static_exports);
  __Pyx_GIVEREF(__pyx_v_static_exports);
  __Pyx_GOTREF(__pyx_v_self->static_exports);
  __Pyx_DECREF(__pyx_v_self->static_exports);
  __pyx_v_self->static_exports = __pyx_v_static_exports;

  /* "flitter/language/tree.pyx":129
 *     cdef readonly dict static_exports
 * 
 *     def __init__(self, dict static_exports=None):             # <<<<<<<<<<<<<<
 *         self.static_exports = static_exports
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":132
 *         self.static_exports = static_exports
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef Vector value
 */

static void __pyx_f_7flitter_8language_4tree_6Export__compile(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  PyObject *__pyx_v_name = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  PyObject *__pyx_v_m = NULL;
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  int64_t __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":135
 *         cdef str name
 *         cdef Vector value
 *         if self.static_exports:             # <<<<<<<<<<<<<<
 *             for name, value in self.static_exports.items():
 *                 program.literal(value)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->static_exports); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 135, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":136
 *         cdef Vector value
 *         if self.static_exports:
 *             for name, value in self.static_exports.items():             # <<<<<<<<<<<<<<
 *                 program.literal(value)
 *                 program.export(name)
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_self->static_exports == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 136, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_self->static_exports, 1, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_6))) __PYX_ERR(0, 136, __pyx_L1_error)
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 136, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":137
 *         if self.static_exports:
 *             for name, value in self.static_exports.items():
 *                 program.literal(value)             # <<<<<<<<<<<<<<
 *                 program.export(name)
 *         m = len(program.initial_lnames)
 */
      __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, ((PyObject *)__pyx_v_value), 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":138
 *             for name, value in self.static_exports.items():
 *                 program.literal(value)
 *                 program.export(name)             # <<<<<<<<<<<<<<
 *         m = len(program.initial_lnames)
 *         cdef int64_t i, n=len(lnames)
 */
      __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->export(__pyx_v_program, __pyx_v_name, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":135
 *         cdef str name
 *         cdef Vector value
 *         if self.static_exports:             # <<<<<<<<<<<<<<
 *             for name, value in self.static_exports.items():
 *                 program.literal(value)
 */
  }

  /* "flitter/language/tree.pyx":139
 *                 program.literal(value)
 *                 program.export(name)
 *         m = len(program.initial_lnames)             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=len(lnames)
 *         if n > m:
 */
  __pyx_t_2 = __pyx_v_program->initial_lnames;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 139, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_m = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":140
 *                 program.export(name)
 *         m = len(program.initial_lnames)
 *         cdef int64_t i, n=len(lnames)             # <<<<<<<<<<<<<<
 *         if n > m:
 *             for i in range(n-m):
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 140, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_v_n = __pyx_t_4;

  /* "flitter/language/tree.pyx":141
 *         m = len(program.initial_lnames)
 *         cdef int64_t i, n=len(lnames)
 *         if n > m:             # <<<<<<<<<<<<<<
 *             for i in range(n-m):
 *                 name = lnames[n-1-i]
 */
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_v_m, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":142
 *         cdef int64_t i, n=len(lnames)
 *         if n > m:
 *             for i in range(n-m):             # <<<<<<<<<<<<<<
 *                 name = lnames[n-1-i]
 *                 program.local_load(i)
 */
    __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_n); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_7, __pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_9 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "flitter/language/tree.pyx":143
 *         if n > m:
 *             for i in range(n-m):
 *                 name = lnames[n-1-i]             # <<<<<<<<<<<<<<
 *                 program.local_load(i)
 *                 program.export(name)
 */
      if (unlikely(__pyx_v_lnames == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 143, __pyx_L1_error)
      }
      __pyx_t_12 = ((__pyx_v_n - 1) - __pyx_v_i);
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_lnames, __pyx_t_12, int64_t, 1, __Pyx_PyInt_From_int64_t, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/language/tree.pyx":144
 *             for i in range(n-m):
 *                 name = lnames[n-1-i]
 *                 program.local_load(i)             # <<<<<<<<<<<<<<
 *                 program.export(name)
 *         program.literal(null_)
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_load(__pyx_v_program, __pyx_v_i, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "flitter/language/tree.pyx":145
 *                 name = lnames[n-1-i]
 *                 program.local_load(i)
 *                 program.export(name)             # <<<<<<<<<<<<<<
 *         program.literal(null_)
 * 
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->export(__pyx_v_program, __pyx_v_name, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "flitter/language/tree.pyx":141
 *         m = len(program.initial_lnames)
 *         cdef int64_t i, n=len(lnames)
 *         if n > m:             # <<<<<<<<<<<<<<
 *             for i in range(n-m):
 *                 name = lnames[n-1-i]
 */
  }

  /* "flitter/language/tree.pyx":146
 *                 program.local_load(i)
 *                 program.export(name)
 *         program.literal(null_)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_7flitter_5model_null_);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_2, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "flitter/language/tree.pyx":132
 *         self.static_exports = static_exports
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef Vector value
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.Export._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":148
 *         program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef dict static_exports = dict(self.static_exports) if self.static_exports else {}
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Export__simplify(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_static_exports = 0;
  int __pyx_v_touched;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":150
 *     cdef Expression _simplify(self, Context context):
 *         cdef str name
 *         cdef dict static_exports = dict(self.static_exports) if self.static_exports else {}             # <<<<<<<<<<<<<<
 *         cdef bint touched = False
 *         for name, value in context.names.items():
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->static_exports); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 150, __pyx_L1_error)
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_self->static_exports == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
      __PYX_ERR(0, 150, __pyx_L1_error)
    }
    __pyx_t_3 = PyDict_Copy(__pyx_v_self->static_exports); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_static_exports = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":151
 *         cdef str name
 *         cdef dict static_exports = dict(self.static_exports) if self.static_exports else {}
 *         cdef bint touched = False             # <<<<<<<<<<<<<<
 *         for name, value in context.names.items():
 *             if value is not None:
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":152
 *         cdef dict static_exports = dict(self.static_exports) if self.static_exports else {}
 *         cdef bint touched = False
 *         for name, value in context.names.items():             # <<<<<<<<<<<<<<
 *             if value is not None:
 *                 context.exports[name] = value
 */
  __pyx_t_4 = 0;
  if (unlikely(__pyx_v_context->names == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 152, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_context->names, 1, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_5, &__pyx_t_4, &__pyx_t_3, &__pyx_t_7, NULL, __pyx_t_6);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_7);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":153
 *         cdef bint touched = False
 *         for name, value in context.names.items():
 *             if value is not None:             # <<<<<<<<<<<<<<
 *                 context.exports[name] = value
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):
 */
    __pyx_t_2 = (__pyx_v_value != Py_None);
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":154
 *         for name, value in context.names.items():
 *             if value is not None:
 *                 context.exports[name] = value             # <<<<<<<<<<<<<<
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):
 *                 static_exports[name] = value
 */
      if (unlikely(__pyx_v_context->exports == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 154, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->exports, __pyx_v_name, __pyx_v_value) < 0))) __PYX_ERR(0, 154, __pyx_L1_error)

      /* "flitter/language/tree.pyx":153
 *         cdef bint touched = False
 *         for name, value in context.names.items():
 *             if value is not None:             # <<<<<<<<<<<<<<
 *                 context.exports[name] = value
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):
 */
    }

    /* "flitter/language/tree.pyx":155
 *             if value is not None:
 *                 context.exports[name] = value
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):             # <<<<<<<<<<<<<<
 *                 static_exports[name] = value
 *                 touched = True
 */
    __pyx_t_9 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7flitter_5model_Vector); 
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    if (unlikely(__pyx_v_static_exports == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 155, __pyx_L1_error)
    }
    __pyx_t_9 = (__Pyx_PyDict_ContainsTF(__pyx_v_name, __pyx_v_static_exports, Py_NE)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 155, __pyx_L1_error)
    if (!__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    if (unlikely(__pyx_v_static_exports == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 155, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_static_exports, __pyx_v_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":156
 *                 context.exports[name] = value
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):
 *                 static_exports[name] = value             # <<<<<<<<<<<<<<
 *                 touched = True
 *         if touched:
 */
      if (unlikely(__pyx_v_static_exports == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 156, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_static_exports, __pyx_v_name, __pyx_v_value) < 0))) __PYX_ERR(0, 156, __pyx_L1_error)

      /* "flitter/language/tree.pyx":157
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):
 *                 static_exports[name] = value
 *                 touched = True             # <<<<<<<<<<<<<<
 *         if touched:
 *             return Export(static_exports)
 */
      __pyx_v_touched = 1;

      /* "flitter/language/tree.pyx":155
 *             if value is not None:
 *                 context.exports[name] = value
 *             if isinstance(value, Vector) and (name not in static_exports or value != static_exports[name]):             # <<<<<<<<<<<<<<
 *                 static_exports[name] = value
 *                 touched = True
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":158
 *                 static_exports[name] = value
 *                 touched = True
 *         if touched:             # <<<<<<<<<<<<<<
 *             return Export(static_exports)
 *         return self
 */
  if (__pyx_v_touched) {

    /* "flitter/language/tree.pyx":159
 *                 touched = True
 *         if touched:
 *             return Export(static_exports)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Export), __pyx_v_static_exports); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":158
 *                 static_exports[name] = value
 *                 touched = True
 *         if touched:             # <<<<<<<<<<<<<<
 *             return Export(static_exports)
 *         return self
 */
  }

  /* "flitter/language/tree.pyx":160
 *         if touched:
 *             return Export(static_exports)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":148
 *         program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef dict static_exports = dict(self.static_exports) if self.static_exports else {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.Export._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_static_exports);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":162
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Export({self.static_exports!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Export_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":163
 * 
 *     def __repr__(self):
 *         return f'Export({self.static_exports!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Export);
  __pyx_t_2 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_Export);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Export);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->static_exports), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":162
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Export({self.static_exports!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Export.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":127
 * 
 * cdef class Export(Expression):
 *     cdef readonly dict static_exports             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, dict static_exports=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_14static_exports_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_14static_exports_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Export_14static_exports___get__(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_14static_exports___get__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->static_exports);
  __pyx_r = __pyx_v_self->static_exports;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Export_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Export_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Export_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.static_exports,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->static_exports);
  __Pyx_GIVEREF(__pyx_v_self->static_exports);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->static_exports)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.static_exports,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.static_exports,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.static_exports is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.static_exports,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.static_exports is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->static_exports != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.static_exports is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.static_exports is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Export); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19136085);
    __Pyx_GIVEREF(__pyx_int_19136085);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_19136085)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.static_exports is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, None), state
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Export); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19136085);
    __Pyx_GIVEREF(__pyx_int_19136085);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_19136085)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Export.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Export_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Export_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Export_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Export.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Export_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Export_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Export__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Export.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":171
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, tuple names, Expression filename, Expression expr):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.filename = filename
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_6Import_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_6Import_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_names = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_filename = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_names,&__pyx_n_s_filename,&__pyx_n_s_expr,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_filename)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expr)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 171, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 171, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_names = ((PyObject*)values[0]);
    __pyx_v_filename = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 171, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Import.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 171, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filename), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "filename", 0))) __PYX_ERR(0, 171, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "expr", 0))) __PYX_ERR(0, 171, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import___init__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self), __pyx_v_names, __pyx_v_filename, __pyx_v_expr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_6Import___init__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_filename, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":172
 * 
 *     def __init__(self, tuple names, Expression filename, Expression expr):
 *         self.names = names             # <<<<<<<<<<<<<<
 *         self.filename = filename
 *         self.expr = expr
 */
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = __pyx_v_names;

  /* "flitter/language/tree.pyx":173
 *     def __init__(self, tuple names, Expression filename, Expression expr):
 *         self.names = names
 *         self.filename = filename             # <<<<<<<<<<<<<<
 *         self.expr = expr
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_filename);
  __Pyx_GIVEREF((PyObject *)__pyx_v_filename);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->filename);
  __Pyx_DECREF((PyObject *)__pyx_v_self->filename);
  __pyx_v_self->filename = __pyx_v_filename;

  /* "flitter/language/tree.pyx":174
 *         self.names = names
 *         self.filename = filename
 *         self.expr = expr             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->expr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->expr);
  __pyx_v_self->expr = __pyx_v_expr;

  /* "flitter/language/tree.pyx":171
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, tuple names, Expression filename, Expression expr):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.filename = filename
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":176
 *         self.expr = expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.filename._compile(program, lnames)
 *         program.import_(self.names)
 */

static void __pyx_f_7flitter_8language_4tree_6Import__compile(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instr = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":177
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.filename._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.import_(self.names)
 *         self.expr._compile(program, lnames + list(self.names))
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->filename->__pyx_vtab)->_compile(__pyx_v_self->filename, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)

  /* "flitter/language/tree.pyx":178
 *     cdef void _compile(self, Program program, list lnames):
 *         self.filename._compile(program, lnames)
 *         program.import_(self.names)             # <<<<<<<<<<<<<<
 *         self.expr._compile(program, lnames + list(self.names))
 *         cdef Instruction instr = program.last_instruction()
 */
  __pyx_t_1 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->import_(__pyx_v_program, ((PyObject*)__pyx_t_1), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":179
 *         self.filename._compile(program, lnames)
 *         program.import_(self.names)
 *         self.expr._compile(program, lnames + list(self.names))             # <<<<<<<<<<<<<<
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:
 */
  __pyx_t_2 = PySequence_List(__pyx_v_self->names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_lnames, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->expr->__pyx_vtab)->_compile(__pyx_v_self->expr, __pyx_v_program, ((PyObject*)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":180
 *         program.import_(self.names)
 *         self.expr._compile(program, lnames + list(self.names))
 *         cdef Instruction instr = program.last_instruction()             # <<<<<<<<<<<<<<
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":181
 *         self.expr._compile(program, lnames + list(self.names))
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(self.names))
 */
  __pyx_t_3 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_LocalDrop);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":182
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.local_drop((<InstructionInt>instr).value + len(self.names))
 *         else:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":183
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(self.names))             # <<<<<<<<<<<<<<
 *         else:
 *             program.local_drop(len(self.names))
 */
    __pyx_t_1 = __pyx_v_self->names;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 183, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_drop(__pyx_v_program, (((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instr)->value + __pyx_t_4), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":181
 *         self.expr._compile(program, lnames + list(self.names))
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(self.names))
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":185
 *             program.local_drop((<InstructionInt>instr).value + len(self.names))
 *         else:
 *             program.local_drop(len(self.names))             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  /*else*/ {
    __pyx_t_1 = __pyx_v_self->names;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 185, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_drop(__pyx_v_program, __pyx_t_4, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":176
 *         self.expr = expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.filename._compile(program, lnames)
 *         program.import_(self.names)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Import._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_RefNannyFinishContext();
}
static PyObject *__pyx_gb_7flitter_8language_4tree_6Import_9_simplify_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/language/tree.pyx":221
 *         cdef Expression expr = self.expr
 *         if let_names:
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)             # <<<<<<<<<<<<<<
 *         expr = expr._simplify(context)
 *         context.names = saved
 */

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_9_simplify_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)__pyx_tp_new_7flitter_8language_4tree___pyx_scope_struct__genexpr(__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 221, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_8language_4tree_6Import_9_simplify_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Import__simplify_locals_genexpr, __pyx_n_s_flitter_language_tree); if (unlikely(!gen)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.tree.Import._simplify.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_8language_4tree_6Import_9_simplify_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 221, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 221, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_genexpr_arg_0, 1, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_name);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_name, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_value);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_name);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_name);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_cur_scope->__pyx_v_name)) __PYX_ERR(0, 221, __pyx_L1_error);
    __pyx_t_8 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_value, __pyx_ptype_7flitter_8language_4tree_Function); 
    if (__pyx_t_8) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __pyx_t_5 = __pyx_cur_scope->__pyx_v_value;
    } else {
      __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_cur_scope->__pyx_v_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = __pyx_t_9;
      __pyx_t_9 = 0;
    }
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 221, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":187
 *             program.local_drop(len(self.names))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef Expression filename = self.filename._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Import__simplify(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_name = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_filename = 0;
  struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v_top = 0;
  struct __pyx_obj_7flitter_5model_Context *__pyx_v_import_context = 0;
  PyObject *__pyx_v_import_static_names = 0;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_let_names = 0;
  PyObject *__pyx_v_saved = 0;
  PyObject *__pyx_v_remaining = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  PyObject *__pyx_7genexpr__pyx_v_name = NULL;
  PyObject *__pyx_gb_7flitter_8language_4tree_6Import_9_simplify_2generator = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":189
 *     cdef Expression _simplify(self, Context context):
 *         cdef str name
 *         cdef Expression filename = self.filename._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Top top
 *         cdef Context import_context
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->filename->__pyx_vtab)->_simplify(__pyx_v_self->filename, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_filename = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":192
 *         cdef Top top
 *         cdef Context import_context
 *         cdef dict import_static_names = None             # <<<<<<<<<<<<<<
 *         if isinstance(filename, Literal) and context.path is not None:
 *             name = (<Literal>filename).value.as_string()
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_import_static_names = ((PyObject*)Py_None);

  /* "flitter/language/tree.pyx":193
 *         cdef Context import_context
 *         cdef dict import_static_names = None
 *         if isinstance(filename, Literal) and context.path is not None:             # <<<<<<<<<<<<<<
 *             name = (<Literal>filename).value.as_string()
 *             path = SharedCache.get_with_root(name, context.path)
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_filename), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_context->path != Py_None);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":194
 *         cdef dict import_static_names = None
 *         if isinstance(filename, Literal) and context.path is not None:
 *             name = (<Literal>filename).value.as_string()             # <<<<<<<<<<<<<<
 *             path = SharedCache.get_with_root(name, context.path)
 *             import_context = context.parent
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_filename)->value->__pyx_vtab)->as_string(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_filename)->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_name = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":195
 *         if isinstance(filename, Literal) and context.path is not None:
 *             name = (<Literal>filename).value.as_string()
 *             path = SharedCache.get_with_root(name, context.path)             # <<<<<<<<<<<<<<
 *             import_context = context.parent
 *             while import_context is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SharedCache); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get_with_root); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_name, __pyx_v_context->path};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_v_path = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":196
 *             name = (<Literal>filename).value.as_string()
 *             path = SharedCache.get_with_root(name, context.path)
 *             import_context = context.parent             # <<<<<<<<<<<<<<
 *             while import_context is not None:
 *                 if import_context.path is path:
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_context->parent);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_import_context = ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":197
 *             path = SharedCache.get_with_root(name, context.path)
 *             import_context = context.parent
 *             while import_context is not None:             # <<<<<<<<<<<<<<
 *                 if import_context.path is path:
 *                     context.errors.add(f"Circular import of '{name}'")
 */
    while (1) {
      __pyx_t_2 = (((PyObject *)__pyx_v_import_context) != Py_None);
      if (!__pyx_t_2) break;

      /* "flitter/language/tree.pyx":198
 *             import_context = context.parent
 *             while import_context is not None:
 *                 if import_context.path is path:             # <<<<<<<<<<<<<<
 *                     context.errors.add(f"Circular import of '{name}'")
 *                     import_static_names = {name: null_ for name in self.names}
 */
      __pyx_t_2 = (__pyx_v_import_context->path == __pyx_v_path);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":199
 *             while import_context is not None:
 *                 if import_context.path is path:
 *                     context.errors.add(f"Circular import of '{name}'")             # <<<<<<<<<<<<<<
 *                     import_static_names = {name: null_ for name in self.names}
 *                     break
 */
        if (unlikely(__pyx_v_context->errors == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
          __PYX_ERR(0, 199, __pyx_L1_error)
        }
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_8 = 127;
        __Pyx_INCREF(__pyx_kp_u_Circular_import_of);
        __pyx_t_7 += 20;
        __Pyx_GIVEREF(__pyx_kp_u_Circular_import_of);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Circular_import_of);
        __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_8;
        __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_INCREF(__pyx_kp_u__3);
        __pyx_t_7 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__3);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__3);
        __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = PySet_Add(__pyx_v_context->errors, __pyx_t_5); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "flitter/language/tree.pyx":200
 *                 if import_context.path is path:
 *                     context.errors.add(f"Circular import of '{name}'")
 *                     import_static_names = {name: null_ for name in self.names}             # <<<<<<<<<<<<<<
 *                     break
 *                 import_context = import_context.parent
 */
        { /* enter inner scope */
          __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 200, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely(__pyx_v_self->names == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 200, __pyx_L11_error)
          }
          __pyx_t_1 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_1);
          __pyx_t_7 = 0;
          for (;;) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 200, __pyx_L11_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 200, __pyx_L11_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 200, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
            if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(0, 200, __pyx_L11_error)
            __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_name, ((PyObject*)__pyx_t_4));
            __pyx_t_4 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_5, (PyObject*)__pyx_7genexpr__pyx_v_name, (PyObject*)__pyx_v_7flitter_5model_null_))) __PYX_ERR(0, 200, __pyx_L11_error)
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_7genexpr__pyx_v_name); __pyx_7genexpr__pyx_v_name = 0;
          goto __pyx_L15_exit_scope;
          __pyx_L11_error:;
          __Pyx_XDECREF(__pyx_7genexpr__pyx_v_name); __pyx_7genexpr__pyx_v_name = 0;
          goto __pyx_L1_error;
          __pyx_L15_exit_scope:;
        } /* exit inner scope */
        __Pyx_DECREF_SET(__pyx_v_import_static_names, ((PyObject*)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "flitter/language/tree.pyx":201
 *                     context.errors.add(f"Circular import of '{name}'")
 *                     import_static_names = {name: null_ for name in self.names}
 *                     break             # <<<<<<<<<<<<<<
 *                 import_context = import_context.parent
 *             else:
 */
        goto __pyx_L7_break;

        /* "flitter/language/tree.pyx":198
 *             import_context = context.parent
 *             while import_context is not None:
 *                 if import_context.path is path:             # <<<<<<<<<<<<<<
 *                     context.errors.add(f"Circular import of '{name}'")
 *                     import_static_names = {name: null_ for name in self.names}
 */
      }

      /* "flitter/language/tree.pyx":202
 *                     import_static_names = {name: null_ for name in self.names}
 *                     break
 *                 import_context = import_context.parent             # <<<<<<<<<<<<<<
 *             else:
 *                 if (top := path.read_flitter_top()) is not None:
 */
      __pyx_t_5 = ((PyObject *)__pyx_v_import_context->parent);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_import_context, ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_5));
      __pyx_t_5 = 0;
    }

    /* "flitter/language/tree.pyx":204
 *                 import_context = import_context.parent
 *             else:
 *                 if (top := path.read_flitter_top()) is not None:             # <<<<<<<<<<<<<<
 *                     top, import_context = top.simplify(path=path, parent=context, return_context=True)
 *                     import_static_names = import_context.exports
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_read_flitter_top); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_8language_4tree_Top))))) __PYX_ERR(0, 204, __pyx_L1_error)
      __pyx_v_top = ((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_2 = (__pyx_t_5 != Py_None);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":205
 *             else:
 *                 if (top := path.read_flitter_top()) is not None:
 *                     top, import_context = top.simplify(path=path, parent=context, return_context=True)             # <<<<<<<<<<<<<<
 *                     import_static_names = import_context.exports
 *                     context.errors.update(import_context.errors)
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_top), __pyx_n_s_simplify); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_path, __pyx_v_path) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_parent, ((PyObject *)__pyx_v_context)) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_context, Py_True) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
          PyObject* sequence = __pyx_t_4;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 205, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 205, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
          index = 0; __pyx_t_1 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_1)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_5 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_5)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L18_unpacking_done;
          __pyx_L17_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 205, __pyx_L1_error)
          __pyx_L18_unpacking_done:;
        }
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Top))))) __PYX_ERR(0, 205, __pyx_L1_error)
        if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_5model_Context))))) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_top, ((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_t_1));
        __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_import_context, ((struct __pyx_obj_7flitter_5model_Context *)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "flitter/language/tree.pyx":206
 *                 if (top := path.read_flitter_top()) is not None:
 *                     top, import_context = top.simplify(path=path, parent=context, return_context=True)
 *                     import_static_names = import_context.exports             # <<<<<<<<<<<<<<
 *                     context.errors.update(import_context.errors)
 *                     context.dependencies.update(import_context.dependencies)
 */
        __pyx_t_4 = __pyx_v_import_context->exports;
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_import_static_names, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "flitter/language/tree.pyx":207
 *                     top, import_context = top.simplify(path=path, parent=context, return_context=True)
 *                     import_static_names = import_context.exports
 *                     context.errors.update(import_context.errors)             # <<<<<<<<<<<<<<
 *                     context.dependencies.update(import_context.dependencies)
 *                     context.dependencies.add(path)
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->errors, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_import_context->errors};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 207, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "flitter/language/tree.pyx":208
 *                     import_static_names = import_context.exports
 *                     context.errors.update(import_context.errors)
 *                     context.dependencies.update(import_context.dependencies)             # <<<<<<<<<<<<<<
 *                     context.dependencies.add(path)
 *         cdef dict let_names = {}
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->dependencies, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_import_context->dependencies};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "flitter/language/tree.pyx":209
 *                     context.errors.update(import_context.errors)
 *                     context.dependencies.update(import_context.dependencies)
 *                     context.dependencies.add(path)             # <<<<<<<<<<<<<<
 *         cdef dict let_names = {}
 *         cdef dict saved = dict(context.names)
 */
        if (unlikely(__pyx_v_context->dependencies == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
          __PYX_ERR(0, 209, __pyx_L1_error)
        }
        __pyx_t_9 = PySet_Add(__pyx_v_context->dependencies, __pyx_v_path); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 209, __pyx_L1_error)

        /* "flitter/language/tree.pyx":204
 *                 import_context = import_context.parent
 *             else:
 *                 if (top := path.read_flitter_top()) is not None:             # <<<<<<<<<<<<<<
 *                     top, import_context = top.simplify(path=path, parent=context, return_context=True)
 *                     import_static_names = import_context.exports
 */
      }
    }
    __pyx_L7_break:;

    /* "flitter/language/tree.pyx":193
 *         cdef Context import_context
 *         cdef dict import_static_names = None
 *         if isinstance(filename, Literal) and context.path is not None:             # <<<<<<<<<<<<<<
 *             name = (<Literal>filename).value.as_string()
 *             path = SharedCache.get_with_root(name, context.path)
 */
  }

  /* "flitter/language/tree.pyx":210
 *                     context.dependencies.update(import_context.dependencies)
 *                     context.dependencies.add(path)
 *         cdef dict let_names = {}             # <<<<<<<<<<<<<<
 *         cdef dict saved = dict(context.names)
 *         cdef list remaining = []
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_let_names = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":211
 *                     context.dependencies.add(path)
 *         cdef dict let_names = {}
 *         cdef dict saved = dict(context.names)             # <<<<<<<<<<<<<<
 *         cdef list remaining = []
 *         for name in self.names:
 */
  if (unlikely(__pyx_v_context->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(0, 211, __pyx_L1_error)
  }
  __pyx_t_4 = PyDict_Copy(__pyx_v_context->names); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_saved = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":212
 *         cdef dict let_names = {}
 *         cdef dict saved = dict(context.names)
 *         cdef list remaining = []             # <<<<<<<<<<<<<<
 *         for name in self.names:
 *             if import_static_names is not None and name in import_static_names:
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_remaining = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":213
 *         cdef dict saved = dict(context.names)
 *         cdef list remaining = []
 *         for name in self.names:             # <<<<<<<<<<<<<<
 *             if import_static_names is not None and name in import_static_names:
 *                 let_names[name] = import_static_names[name]
 */
  if (unlikely(__pyx_v_self->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 213, __pyx_L1_error)
  }
  __pyx_t_4 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_7 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 213, __pyx_L1_error)
      #endif
      if (__pyx_t_7 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 213, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "flitter/language/tree.pyx":214
 *         cdef list remaining = []
 *         for name in self.names:
 *             if import_static_names is not None and name in import_static_names:             # <<<<<<<<<<<<<<
 *                 let_names[name] = import_static_names[name]
 *             else:
 */
    __pyx_t_3 = (__pyx_v_import_static_names != ((PyObject*)Py_None));
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L22_bool_binop_done;
    }
    if (unlikely(__pyx_v_import_static_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 214, __pyx_L1_error)
    }
    __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_name, __pyx_v_import_static_names, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 214, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_3;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":215
 *         for name in self.names:
 *             if import_static_names is not None and name in import_static_names:
 *                 let_names[name] = import_static_names[name]             # <<<<<<<<<<<<<<
 *             else:
 *                 context.names[name] = None
 */
      if (unlikely(__pyx_v_import_static_names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 215, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_import_static_names, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((PyDict_SetItem(__pyx_v_let_names, __pyx_v_name, __pyx_t_5) < 0))) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":214
 *         cdef list remaining = []
 *         for name in self.names:
 *             if import_static_names is not None and name in import_static_names:             # <<<<<<<<<<<<<<
 *                 let_names[name] = import_static_names[name]
 *             else:
 */
      goto __pyx_L21;
    }

    /* "flitter/language/tree.pyx":217
 *                 let_names[name] = import_static_names[name]
 *             else:
 *                 context.names[name] = None             # <<<<<<<<<<<<<<
 *                 remaining.append(name)
 *         cdef Expression expr = self.expr
 */
    /*else*/ {
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 217, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, Py_None) < 0))) __PYX_ERR(0, 217, __pyx_L1_error)

      /* "flitter/language/tree.pyx":218
 *             else:
 *                 context.names[name] = None
 *                 remaining.append(name)             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr
 *         if let_names:
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_v_name); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 218, __pyx_L1_error)
    }
    __pyx_L21:;

    /* "flitter/language/tree.pyx":213
 *         cdef dict saved = dict(context.names)
 *         cdef list remaining = []
 *         for name in self.names:             # <<<<<<<<<<<<<<
 *             if import_static_names is not None and name in import_static_names:
 *                 let_names[name] = import_static_names[name]
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":219
 *                 context.names[name] = None
 *                 remaining.append(name)
 *         cdef Expression expr = self.expr             # <<<<<<<<<<<<<<
 *         if let_names:
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->expr);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":220
 *                 remaining.append(name)
 *         cdef Expression expr = self.expr
 *         if let_names:             # <<<<<<<<<<<<<<
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)
 *         expr = expr._simplify(context)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_let_names); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 220, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":221
 *         cdef Expression expr = self.expr
 *         if let_names:
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)             # <<<<<<<<<<<<<<
 *         expr = expr._simplify(context)
 *         context.names = saved
 */
    __pyx_t_4 = __pyx_pf_7flitter_8language_4tree_6Import_9_simplify_genexpr(NULL, __pyx_v_let_names); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_expr);
    __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 221, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "flitter/language/tree.pyx":220
 *                 remaining.append(name)
 *         cdef Expression expr = self.expr
 *         if let_names:             # <<<<<<<<<<<<<<
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)
 *         expr = expr._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":222
 *         if let_names:
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)
 *         expr = expr._simplify(context)             # <<<<<<<<<<<<<<
 *         context.names = saved
 *         if not remaining:
 */
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5));
  __pyx_t_5 = 0;

  /* "flitter/language/tree.pyx":223
 *             expr = Let(tuple(PolyBinding((name,), value if isinstance(value, Function) else Literal(value)) for name, value in let_names.items()), expr)
 *         expr = expr._simplify(context)
 *         context.names = saved             # <<<<<<<<<<<<<<
 *         if not remaining:
 *             return expr
 */
  __Pyx_INCREF(__pyx_v_saved);
  __Pyx_GIVEREF(__pyx_v_saved);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = __pyx_v_saved;

  /* "flitter/language/tree.pyx":224
 *         expr = expr._simplify(context)
 *         context.names = saved
 *         if not remaining:             # <<<<<<<<<<<<<<
 *             return expr
 *         if filename is self.filename and expr is self.expr:
 */
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_remaining) != 0);
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":225
 *         context.names = saved
 *         if not remaining:
 *             return expr             # <<<<<<<<<<<<<<
 *         if filename is self.filename and expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_expr);
    __pyx_r = __pyx_v_expr;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":224
 *         expr = expr._simplify(context)
 *         context.names = saved
 *         if not remaining:             # <<<<<<<<<<<<<<
 *             return expr
 *         if filename is self.filename and expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":226
 *         if not remaining:
 *             return expr
 *         if filename is self.filename and expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Import(tuple(remaining), filename, expr)
 */
  __pyx_t_2 = (__pyx_v_filename == __pyx_v_self->filename);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L28_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->expr);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L28_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":227
 *             return expr
 *         if filename is self.filename and expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Import(tuple(remaining), filename, expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":226
 *         if not remaining:
 *             return expr
 *         if filename is self.filename and expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Import(tuple(remaining), filename, expr)
 */
  }

  /* "flitter/language/tree.pyx":228
 *         if filename is self.filename and expr is self.expr:
 *             return self
 *         return Import(tuple(remaining), filename, expr)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_5 = PyList_AsTuple(__pyx_v_remaining); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(0, 228, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_filename);
  __Pyx_GIVEREF((PyObject *)__pyx_v_filename);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_filename))) __PYX_ERR(0, 228, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 228, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Import), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":187
 *             program.local_drop(len(self.names))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef Expression filename = self.filename._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("flitter.language.tree.Import._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF((PyObject *)__pyx_v_filename);
  __Pyx_XDECREF((PyObject *)__pyx_v_top);
  __Pyx_XDECREF((PyObject *)__pyx_v_import_context);
  __Pyx_XDECREF(__pyx_v_import_static_names);
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_let_names);
  __Pyx_XDECREF(__pyx_v_saved);
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_name);
  __Pyx_XDECREF(__pyx_gb_7flitter_8language_4tree_6Import_9_simplify_2generator);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":230
 *         return Import(tuple(remaining), filename, expr)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Import({self.names!r}, {self.filename!r}, {self.expr!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":231
 * 
 *     def __repr__(self):
 *         return f'Import({self.names!r}, {self.filename!r}, {self.expr!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Import);
  __pyx_t_2 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_Import);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Import);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->names), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->filename)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->expr)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":230
 *         return Import(tuple(remaining), filename, expr)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Import({self.names!r}, {self.filename!r}, {self.expr!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Import.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":167
 * 
 * cdef class Import(Expression):
 *     cdef readonly tuple names             # <<<<<<<<<<<<<<
 *     cdef readonly Expression filename
 *     cdef readonly Expression expr
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_5names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_5names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_5names___get__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_5names___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->names);
  __pyx_r = __pyx_v_self->names;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":168
 * cdef class Import(Expression):
 *     cdef readonly tuple names
 *     cdef readonly Expression filename             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_8filename_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_8filename_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_8filename___get__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_8filename___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->filename);
  __pyx_r = ((PyObject *)__pyx_v_self->filename);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":169
 *     cdef readonly tuple names
 *     cdef readonly Expression filename
 *     cdef readonly Expression expr             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple names, Expression filename, Expression expr):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_4expr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_4expr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_4expr___get__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __pyx_r = ((PyObject *)__pyx_v_self->expr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Import_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Import_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr, self.filename, self.names)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->filename);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->filename);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->filename))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->names);
  __Pyx_GIVEREF(__pyx_v_self->names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->names)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr, self.filename, self.names)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr, self.filename, self.names)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None or self.filename is not None or self.names is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr, self.filename, self.names)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None or self.filename is not None or self.names is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->expr) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->filename) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->names != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.filename is not None or self.names is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None or self.filename is not None or self.names is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Import); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_65788850);
    __Pyx_GIVEREF(__pyx_int_65788850);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_65788850)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.filename is not None or self.names is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, None), state
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Import); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_65788850);
    __Pyx_GIVEREF(__pyx_int_65788850);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_65788850)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Import.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Import_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Import_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Import_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Import.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Import_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Import_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Import__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Import.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":237
 *     cdef readonly tuple expressions
 * 
 *     def __init__(self, tuple expressions):             # <<<<<<<<<<<<<<
 *         self.expressions = expressions
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_8Sequence_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_8Sequence_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_expressions = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_expressions,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expressions)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 237, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_expressions = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Sequence.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expressions), (&PyTuple_Type), 1, "expressions", 1))) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Sequence___init__(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_self), __pyx_v_expressions);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_8Sequence___init__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, PyObject *__pyx_v_expressions) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":238
 * 
 *     def __init__(self, tuple expressions):
 *         self.expressions = expressions             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_expressions);
  __Pyx_GIVEREF(__pyx_v_expressions);
  __Pyx_GOTREF(__pyx_v_self->expressions);
  __Pyx_DECREF(__pyx_v_self->expressions);
  __pyx_v_self->expressions = __pyx_v_expressions;

  /* "flitter/language/tree.pyx":237
 *     cdef readonly tuple expressions
 * 
 *     def __init__(self, tuple expressions):             # <<<<<<<<<<<<<<
 *         self.expressions = expressions
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":240
 *         self.expressions = expressions
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Expression expr
 *         cdef InstructionInt instr = None
 */

static void __pyx_f_7flitter_8language_4tree_8Sequence__compile(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_2vm_InstructionInt *__pyx_v_instr = 0;
  int64_t __pyx_v_n;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":242
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef Expression expr
 *         cdef InstructionInt instr = None             # <<<<<<<<<<<<<<
 *         cdef int64_t n=len(self.expressions)
 *         if n:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)Py_None);

  /* "flitter/language/tree.pyx":243
 *         cdef Expression expr
 *         cdef InstructionInt instr = None
 *         cdef int64_t n=len(self.expressions)             # <<<<<<<<<<<<<<
 *         if n:
 *             for expr in self.expressions:
 */
  __pyx_t_1 = __pyx_v_self->expressions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 243, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n = __pyx_t_2;

  /* "flitter/language/tree.pyx":244
 *         cdef InstructionInt instr = None
 *         cdef int64_t n=len(self.expressions)
 *         if n:             # <<<<<<<<<<<<<<
 *             for expr in self.expressions:
 *                 expr._compile(program, lnames)
 */
  __pyx_t_3 = (__pyx_v_n != 0);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":245
 *         cdef int64_t n=len(self.expressions)
 *         if n:
 *             for expr in self.expressions:             # <<<<<<<<<<<<<<
 *                 expr._compile(program, lnames)
 *                 if program.last_instruction().code == OpCode.Compose:
 */
    if (unlikely(__pyx_v_self->expressions == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 245, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->expressions; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 245, __pyx_L1_error)
        #endif
        if (__pyx_t_2 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 245, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":246
 *         if n:
 *             for expr in self.expressions:
 *                 expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *                 if program.last_instruction().code == OpCode.Compose:
 *                     instr = program.pop_instruction()
 */
      ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_compile(__pyx_v_expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)

      /* "flitter/language/tree.pyx":247
 *             for expr in self.expressions:
 *                 expr._compile(program, lnames)
 *                 if program.last_instruction().code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *                     instr = program.pop_instruction()
 *                     n += instr.value - 1
 */
      __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = (((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_4)->code == __pyx_e_7flitter_8language_2vm_Compose);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_3) {

        /* "flitter/language/tree.pyx":248
 *                 expr._compile(program, lnames)
 *                 if program.last_instruction().code == OpCode.Compose:
 *                     instr = program.pop_instruction()             # <<<<<<<<<<<<<<
 *                     n += instr.value - 1
 *             if n > 1:
 */
        __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_2vm_InstructionInt))))) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_instr, ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "flitter/language/tree.pyx":249
 *                 if program.last_instruction().code == OpCode.Compose:
 *                     instr = program.pop_instruction()
 *                     n += instr.value - 1             # <<<<<<<<<<<<<<
 *             if n > 1:
 *                 program.compose(n)
 */
        __pyx_v_n = (__pyx_v_n + (__pyx_v_instr->value - 1));

        /* "flitter/language/tree.pyx":247
 *             for expr in self.expressions:
 *                 expr._compile(program, lnames)
 *                 if program.last_instruction().code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *                     instr = program.pop_instruction()
 *                     n += instr.value - 1
 */
      }

      /* "flitter/language/tree.pyx":245
 *         cdef int64_t n=len(self.expressions)
 *         if n:
 *             for expr in self.expressions:             # <<<<<<<<<<<<<<
 *                 expr._compile(program, lnames)
 *                 if program.last_instruction().code == OpCode.Compose:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":250
 *                     instr = program.pop_instruction()
 *                     n += instr.value - 1
 *             if n > 1:             # <<<<<<<<<<<<<<
 *                 program.compose(n)
 *         else:
 */
    __pyx_t_3 = (__pyx_v_n > 1);
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":251
 *                     n += instr.value - 1
 *             if n > 1:
 *                 program.compose(n)             # <<<<<<<<<<<<<<
 *         else:
 *             program.literal(null_)
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->compose(__pyx_v_program, __pyx_v_n, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":250
 *                     instr = program.pop_instruction()
 *                     n += instr.value - 1
 *             if n > 1:             # <<<<<<<<<<<<<<
 *                 program.compose(n)
 *         else:
 */
    }

    /* "flitter/language/tree.pyx":244
 *         cdef InstructionInt instr = None
 *         cdef int64_t n=len(self.expressions)
 *         if n:             # <<<<<<<<<<<<<<
 *             for expr in self.expressions:
 *                 expr._compile(program, lnames)
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":253
 *                 program.compose(n)
 *         else:
 *             program.literal(null_)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_null_);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_1, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":240
 *         self.expressions = expressions
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Expression expr
 *         cdef InstructionInt instr = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Sequence._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":255
 *             program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list expressions = []
 *         cdef Expression expr, sexpr
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Sequence__simplify(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_expressions = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_sexpr = 0;
  int __pyx_v_touched;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":256
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef list expressions = []             # <<<<<<<<<<<<<<
 *         cdef Expression expr, sexpr
 *         cdef bint touched = False
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expressions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":258
 *         cdef list expressions = []
 *         cdef Expression expr, sexpr
 *         cdef bint touched = False             # <<<<<<<<<<<<<<
 *         for expr in self.expressions:
 *             sexpr = expr._simplify(context)
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":259
 *         cdef Expression expr, sexpr
 *         cdef bint touched = False
 *         for expr in self.expressions:             # <<<<<<<<<<<<<<
 *             sexpr = expr._simplify(context)
 *             expressions.append(sexpr)
 */
  if (unlikely(__pyx_v_self->expressions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 259, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->expressions; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 259, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 259, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":260
 *         cdef bint touched = False
 *         for expr in self.expressions:
 *             sexpr = expr._simplify(context)             # <<<<<<<<<<<<<<
 *             expressions.append(sexpr)
 *             touched |= sexpr is not expr
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_sexpr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":261
 *         for expr in self.expressions:
 *             sexpr = expr._simplify(context)
 *             expressions.append(sexpr)             # <<<<<<<<<<<<<<
 *             touched |= sexpr is not expr
 *         touched |= sequence_pack(expressions)
 */
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_expressions, ((PyObject *)__pyx_v_sexpr)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 261, __pyx_L1_error)

    /* "flitter/language/tree.pyx":262
 *             sexpr = expr._simplify(context)
 *             expressions.append(sexpr)
 *             touched |= sexpr is not expr             # <<<<<<<<<<<<<<
 *         touched |= sequence_pack(expressions)
 *         if not expressions:
 */
    __pyx_t_5 = (__pyx_v_sexpr != __pyx_v_expr);
    __pyx_v_touched = (__pyx_v_touched | __pyx_t_5);

    /* "flitter/language/tree.pyx":259
 *         cdef Expression expr, sexpr
 *         cdef bint touched = False
 *         for expr in self.expressions:             # <<<<<<<<<<<<<<
 *             sexpr = expr._simplify(context)
 *             expressions.append(sexpr)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":263
 *             expressions.append(sexpr)
 *             touched |= sexpr is not expr
 *         touched |= sequence_pack(expressions)             # <<<<<<<<<<<<<<
 *         if not expressions:
 *             return NoOp
 */
  __pyx_t_5 = __pyx_f_7flitter_8language_4tree_sequence_pack(__pyx_v_expressions); if (unlikely(__pyx_t_5 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 263, __pyx_L1_error)
  __pyx_v_touched = (__pyx_v_touched | __pyx_t_5);

  /* "flitter/language/tree.pyx":264
 *             touched |= sexpr is not expr
 *         touched |= sequence_pack(expressions)
 *         if not expressions:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         if len(expressions) == 1:
 */
  __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_expressions) != 0);
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":265
 *         touched |= sequence_pack(expressions)
 *         if not expressions:
 *             return NoOp             # <<<<<<<<<<<<<<
 *         if len(expressions) == 1:
 *             return expressions[0]
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_7flitter_8language_4tree_NoOp);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":264
 *             touched |= sexpr is not expr
 *         touched |= sequence_pack(expressions)
 *         if not expressions:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         if len(expressions) == 1:
 */
  }

  /* "flitter/language/tree.pyx":266
 *         if not expressions:
 *             return NoOp
 *         if len(expressions) == 1:             # <<<<<<<<<<<<<<
 *             return expressions[0]
 *         if not touched:
 */
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_expressions); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 266, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_2 == 1);
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":267
 *             return NoOp
 *         if len(expressions) == 1:
 *             return expressions[0]             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_expressions, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 267, __pyx_L1_error)
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":266
 *         if not expressions:
 *             return NoOp
 *         if len(expressions) == 1:             # <<<<<<<<<<<<<<
 *             return expressions[0]
 *         if not touched:
 */
  }

  /* "flitter/language/tree.pyx":268
 *         if len(expressions) == 1:
 *             return expressions[0]
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Sequence(tuple(expressions))
 */
  __pyx_t_6 = (!__pyx_v_touched);
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":269
 *             return expressions[0]
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         return Sequence(tuple(expressions))
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":268
 *         if len(expressions) == 1:
 *             return expressions[0]
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Sequence(tuple(expressions))
 */
  }

  /* "flitter/language/tree.pyx":270
 *         if not touched:
 *             return self
 *         return Sequence(tuple(expressions))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyList_AsTuple(__pyx_v_expressions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":255
 *             program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list expressions = []
 *         cdef Expression expr, sexpr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Sequence._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expressions);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_sexpr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":272
 *         return Sequence(tuple(expressions))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Sequence({self.expressions!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Sequence_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":273
 * 
 *     def __repr__(self):
 *         return f'Sequence({self.expressions!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Sequence);
  __pyx_t_2 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_Sequence);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Sequence);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->expressions), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":272
 *         return Sequence(tuple(expressions))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Sequence({self.expressions!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Sequence.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":235
 * 
 * cdef class Sequence(Expression):
 *     cdef readonly tuple expressions             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple expressions):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_11expressions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_11expressions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Sequence_11expressions___get__(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_11expressions___get__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->expressions);
  __pyx_r = __pyx_v_self->expressions;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Sequence_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Sequence_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Sequence_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expressions,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->expressions);
  __Pyx_GIVEREF(__pyx_v_self->expressions);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->expressions)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expressions,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expressions,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expressions is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expressions,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expressions is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->expressions != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expressions is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expressions is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Sequence); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_25936640);
    __Pyx_GIVEREF(__pyx_int_25936640);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_25936640)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expressions is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, None), state
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Sequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_25936640);
    __Pyx_GIVEREF(__pyx_int_25936640);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_25936640)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Sequence.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Sequence_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Sequence_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Sequence_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Sequence.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Sequence_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Sequence_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Sequence__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Sequence.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":279
 *     cdef readonly Vector value
 * 
 *     def __init__(self, value):             # <<<<<<<<<<<<<<
 *         self.value = Vector._coerce(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_7Literal_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_7Literal_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 279, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 279, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 279, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Literal.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Literal___init__(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_7Literal___init__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":280
 * 
 *     def __init__(self, value):
 *         self.value = Vector._coerce(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_coerce(__pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->value);
  __Pyx_DECREF((PyObject *)__pyx_v_self->value);
  __pyx_v_self->value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":279
 *     cdef readonly Vector value
 * 
 *     def __init__(self, value):             # <<<<<<<<<<<<<<
 *         self.value = Vector._coerce(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Literal.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":282
 *         self.value = Vector._coerce(value)
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         program.literal(self.value.intern())
 * 
 */

static void __pyx_f_7flitter_8language_4tree_7Literal__compile(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, CYTHON_UNUSED PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":283
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         program.literal(self.value.intern())             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_self->value->__pyx_vtab)->intern(__pyx_v_self->value, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_1, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":282
 *         self.value = Vector._coerce(value)
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         program.literal(self.value.intern())
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Literal._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":285
 *         program.literal(self.value.intern())
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_7Literal__simplify(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":286
 * 
 *     cdef Expression _simplify(self, Context context):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":285
 *         program.literal(self.value.intern())
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":288
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Literal({self.value!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Literal_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":289
 * 
 *     def __repr__(self):
 *         return f'Literal({self.value!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Literal);
  __pyx_t_2 += 8;
  __Pyx_GIVEREF(__pyx_kp_u_Literal);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Literal);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":288
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Literal({self.value!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Literal.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":277
 * 
 * cdef class Literal(Expression):
 *     cdef readonly Vector value             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Literal_5value___get__(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_5value___get__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __pyx_r = ((PyObject *)__pyx_v_self->value);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7Literal_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Literal_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Literal_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.value,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->value))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->value) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Literal); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215229444);
    __Pyx_GIVEREF(__pyx_int_215229444);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_215229444)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, None), state
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Literal); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215229444);
    __Pyx_GIVEREF(__pyx_int_215229444);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_215229444)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Literal.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7Literal_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Literal_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7Literal_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Literal.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Literal_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Literal_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Literal__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Literal.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":295
 *     cdef readonly str name
 * 
 *     def __init__(self, str name):             # <<<<<<<<<<<<<<
 *         self.name = name
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_4Name_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_4Name_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 295, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_name = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 295, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Name.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Name___init__(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_4Name___init__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":296
 * 
 *     def __init__(self, str name):
 *         self.name = name             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "flitter/language/tree.pyx":295
 *     cdef readonly str name
 * 
 *     def __init__(self, str name):             # <<<<<<<<<<<<<<
 *         self.name = name
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":298
 *         self.name = name
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=len(lnames)-1
 *         for i in range(len(lnames)):
 */

static void __pyx_f_7flitter_8language_4tree_4Name__compile(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":299
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef int64_t i, n=len(lnames)-1             # <<<<<<<<<<<<<<
 *         for i in range(len(lnames)):
 *             if self.name == <str>lnames[n-i]:
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 299, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 299, __pyx_L1_error)
  __pyx_v_n = (__pyx_t_1 - 1);

  /* "flitter/language/tree.pyx":300
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef int64_t i, n=len(lnames)-1
 *         for i in range(len(lnames)):             # <<<<<<<<<<<<<<
 *             if self.name == <str>lnames[n-i]:
 *                 program.local_load(i)
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 300, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 300, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/language/tree.pyx":301
 *         cdef int64_t i, n=len(lnames)-1
 *         for i in range(len(lnames)):
 *             if self.name == <str>lnames[n-i]:             # <<<<<<<<<<<<<<
 *                 program.local_load(i)
 *                 break
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 301, __pyx_L1_error)
    }
    __pyx_t_4 = (__pyx_v_n - __pyx_v_i);
    __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_lnames, __pyx_t_4, int64_t, 1, __Pyx_PyInt_From_int64_t, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_self->name, __pyx_t_5, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {

      /* "flitter/language/tree.pyx":302
 *         for i in range(len(lnames)):
 *             if self.name == <str>lnames[n-i]:
 *                 program.local_load(i)             # <<<<<<<<<<<<<<
 *                 break
 *         else:
 */
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_load(__pyx_v_program, __pyx_v_i, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":303
 *             if self.name == <str>lnames[n-i]:
 *                 program.local_load(i)
 *                 break             # <<<<<<<<<<<<<<
 *         else:
 *             if self.name in static_builtins:
 */
      goto __pyx_L4_break;

      /* "flitter/language/tree.pyx":301
 *         cdef int64_t i, n=len(lnames)-1
 *         for i in range(len(lnames)):
 *             if self.name == <str>lnames[n-i]:             # <<<<<<<<<<<<<<
 *                 program.local_load(i)
 *                 break
 */
    }
  }
  /*else*/ {

    /* "flitter/language/tree.pyx":305
 *                 break
 *         else:
 *             if self.name in static_builtins:             # <<<<<<<<<<<<<<
 *                 program.literal(static_builtins[self.name])
 *             elif self.name in dynamic_builtins:
 */
    if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 305, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PyDict_ContainsTF(__pyx_v_self->name, __pyx_v_7flitter_8language_2vm_static_builtins, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 305, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "flitter/language/tree.pyx":306
 *         else:
 *             if self.name in static_builtins:
 *                 program.literal(static_builtins[self.name])             # <<<<<<<<<<<<<<
 *             elif self.name in dynamic_builtins:
 *                 program.literal(dynamic_builtins[self.name])
 */
      if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 306, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_static_builtins, __pyx_v_self->name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_5, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":305
 *                 break
 *         else:
 *             if self.name in static_builtins:             # <<<<<<<<<<<<<<
 *                 program.literal(static_builtins[self.name])
 *             elif self.name in dynamic_builtins:
 */
      goto __pyx_L6;
    }

    /* "flitter/language/tree.pyx":307
 *             if self.name in static_builtins:
 *                 program.literal(static_builtins[self.name])
 *             elif self.name in dynamic_builtins:             # <<<<<<<<<<<<<<
 *                 program.literal(dynamic_builtins[self.name])
 *             else:
 */
    if (unlikely(__pyx_v_7flitter_8language_2vm_dynamic_builtins == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 307, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PyDict_ContainsTF(__pyx_v_self->name, __pyx_v_7flitter_8language_2vm_dynamic_builtins, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 307, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "flitter/language/tree.pyx":308
 *                 program.literal(static_builtins[self.name])
 *             elif self.name in dynamic_builtins:
 *                 program.literal(dynamic_builtins[self.name])             # <<<<<<<<<<<<<<
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{self.name}'")
 */
      if (unlikely(__pyx_v_7flitter_8language_2vm_dynamic_builtins == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 308, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_dynamic_builtins, __pyx_v_self->name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_7, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":307
 *             if self.name in static_builtins:
 *                 program.literal(static_builtins[self.name])
 *             elif self.name in dynamic_builtins:             # <<<<<<<<<<<<<<
 *                 program.literal(dynamic_builtins[self.name])
 *             else:
 */
      goto __pyx_L6;
    }

    /* "flitter/language/tree.pyx":310
 *                 program.literal(dynamic_builtins[self.name])
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{self.name}'")             # <<<<<<<<<<<<<<
 *                 program.literal(null_)
 * 
 */
    /*else*/ {
      if (unlikely(__pyx_v_program->compiler_errors == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 310, __pyx_L1_error)
      }
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = 0;
      __pyx_t_9 = 127;
      __Pyx_INCREF(__pyx_kp_u_Unbound_name);
      __pyx_t_8 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_Unbound_name);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Unbound_name);
      __pyx_t_7 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_8 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__3);
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = PySet_Add(__pyx_v_program->compiler_errors, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":311
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{self.name}'")
 *                 program.literal(null_)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
      __pyx_t_7 = ((PyObject *)__pyx_v_7flitter_5model_null_);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_7, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_L6:;
  }
  __pyx_L4_break:;

  /* "flitter/language/tree.pyx":298
 *         self.name = name
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=len(lnames)-1
 *         for i in range(len(lnames)):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.Name._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":313
 *                 program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         if self.name in context.names:
 *             value = context.names[self.name]
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Name__simplify(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_value = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":314
 * 
 *     cdef Expression _simplify(self, Context context):
 *         if self.name in context.names:             # <<<<<<<<<<<<<<
 *             value = context.names[self.name]
 *             if value is None or isinstance(value, Function):
 */
  if (unlikely(__pyx_v_context->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 314, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_self->name, __pyx_v_context->names, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 314, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":315
 *     cdef Expression _simplify(self, Context context):
 *         if self.name in context.names:
 *             value = context.names[self.name]             # <<<<<<<<<<<<<<
 *             if value is None or isinstance(value, Function):
 *                 return self
 */
    if (unlikely(__pyx_v_context->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 315, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_context->names, __pyx_v_self->name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_value = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":316
 *         if self.name in context.names:
 *             value = context.names[self.name]
 *             if value is None or isinstance(value, Function):             # <<<<<<<<<<<<<<
 *                 return self
 *             elif isinstance(value, Name):
 */
    __pyx_t_3 = (__pyx_v_value == Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7flitter_8language_4tree_Function); 
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "flitter/language/tree.pyx":317
 *             value = context.names[self.name]
 *             if value is None or isinstance(value, Function):
 *                 return self             # <<<<<<<<<<<<<<
 *             elif isinstance(value, Name):
 *                 return (<Name>value)._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_self);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":316
 *         if self.name in context.names:
 *             value = context.names[self.name]
 *             if value is None or isinstance(value, Function):             # <<<<<<<<<<<<<<
 *                 return self
 *             elif isinstance(value, Name):
 */
    }

    /* "flitter/language/tree.pyx":318
 *             if value is None or isinstance(value, Function):
 *                 return self
 *             elif isinstance(value, Name):             # <<<<<<<<<<<<<<
 *                 return (<Name>value)._simplify(context)
 *             elif type(value) is Vector:
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7flitter_8language_4tree_Name); 
    if (__pyx_t_1) {

      /* "flitter/language/tree.pyx":319
 *                 return self
 *             elif isinstance(value, Name):
 *                 return (<Name>value)._simplify(context)             # <<<<<<<<<<<<<<
 *             elif type(value) is Vector:
 *                 return Literal((<Vector>value).copy())
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Name *)((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_value)->__pyx_base.__pyx_vtab)->__pyx_base._simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_value)), __pyx_v_context)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":318
 *             if value is None or isinstance(value, Function):
 *                 return self
 *             elif isinstance(value, Name):             # <<<<<<<<<<<<<<
 *                 return (<Name>value)._simplify(context)
 *             elif type(value) is Vector:
 */
    }

    /* "flitter/language/tree.pyx":320
 *             elif isinstance(value, Name):
 *                 return (<Name>value)._simplify(context)
 *             elif type(value) is Vector:             # <<<<<<<<<<<<<<
 *                 return Literal((<Vector>value).copy())
 *             else:
 */
    __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)__pyx_ptype_7flitter_5model_Vector));
    if (__pyx_t_1) {

      /* "flitter/language/tree.pyx":321
 *                 return (<Name>value)._simplify(context)
 *             elif type(value) is Vector:
 *                 return Literal((<Vector>value).copy())             # <<<<<<<<<<<<<<
 *             else:
 *                 return Literal(value)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_value)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_v_value), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":320
 *             elif isinstance(value, Name):
 *                 return (<Name>value)._simplify(context)
 *             elif type(value) is Vector:             # <<<<<<<<<<<<<<
 *                 return Literal((<Vector>value).copy())
 *             else:
 */
    }

    /* "flitter/language/tree.pyx":323
 *                 return Literal((<Vector>value).copy())
 *             else:
 *                 return Literal(value)             # <<<<<<<<<<<<<<
 *         elif (value := static_builtins.get(self.name)) is not None:
 *             return Literal(value)
 */
    /*else*/ {
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;
    }

    /* "flitter/language/tree.pyx":314
 * 
 *     cdef Expression _simplify(self, Context context):
 *         if self.name in context.names:             # <<<<<<<<<<<<<<
 *             value = context.names[self.name]
 *             if value is None or isinstance(value, Function):
 */
  }

  /* "flitter/language/tree.pyx":324
 *             else:
 *                 return Literal(value)
 *         elif (value := static_builtins.get(self.name)) is not None:             # <<<<<<<<<<<<<<
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:
 */
  if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(0, 324, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyDict_GetItemDefault(__pyx_v_7flitter_8language_2vm_static_builtins, __pyx_v_self->name, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_value = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = (__pyx_t_4 != Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":325
 *                 return Literal(value)
 *         elif (value := static_builtins.get(self.name)) is not None:
 *             return Literal(value)             # <<<<<<<<<<<<<<
 *         elif self.name not in dynamic_builtins:
 *             if context.captures is not None:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":324
 *             else:
 *                 return Literal(value)
 *         elif (value := static_builtins.get(self.name)) is not None:             # <<<<<<<<<<<<<<
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:
 */
  }

  /* "flitter/language/tree.pyx":326
 *         elif (value := static_builtins.get(self.name)) is not None:
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:             # <<<<<<<<<<<<<<
 *             if context.captures is not None:
 *                 context.captures.add(self.name)
 */
  if (unlikely(__pyx_v_7flitter_8language_2vm_dynamic_builtins == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 326, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_self->name, __pyx_v_7flitter_8language_2vm_dynamic_builtins, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":327
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:
 *             if context.captures is not None:             # <<<<<<<<<<<<<<
 *                 context.captures.add(self.name)
 *             else:
 */
    __pyx_t_1 = (__pyx_v_context->captures != ((PyObject*)Py_None));
    if (__pyx_t_1) {

      /* "flitter/language/tree.pyx":328
 *         elif self.name not in dynamic_builtins:
 *             if context.captures is not None:
 *                 context.captures.add(self.name)             # <<<<<<<<<<<<<<
 *             else:
 *                 context.errors.add(f"Unbound name '{self.name}'")
 */
      if (unlikely(__pyx_v_context->captures == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 328, __pyx_L1_error)
      }
      __pyx_t_4 = __pyx_v_self->name;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = PySet_Add(__pyx_v_context->captures, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 328, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":327
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:
 *             if context.captures is not None:             # <<<<<<<<<<<<<<
 *                 context.captures.add(self.name)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "flitter/language/tree.pyx":330
 *                 context.captures.add(self.name)
 *             else:
 *                 context.errors.add(f"Unbound name '{self.name}'")             # <<<<<<<<<<<<<<
 *                 return NoOp
 *         return self
 */
    /*else*/ {
      if (unlikely(__pyx_v_context->errors == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 330, __pyx_L1_error)
      }
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = 0;
      __pyx_t_7 = 127;
      __Pyx_INCREF(__pyx_kp_u_Unbound_name);
      __pyx_t_6 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_Unbound_name);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Unbound_name);
      __pyx_t_2 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
      __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_6 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u__3);
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_5 = PySet_Add(__pyx_v_context->errors, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "flitter/language/tree.pyx":331
 *             else:
 *                 context.errors.add(f"Unbound name '{self.name}'")
 *                 return NoOp             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_7flitter_8language_4tree_NoOp);
      goto __pyx_L0;
    }
    __pyx_L7:;

    /* "flitter/language/tree.pyx":326
 *         elif (value := static_builtins.get(self.name)) is not None:
 *             return Literal(value)
 *         elif self.name not in dynamic_builtins:             # <<<<<<<<<<<<<<
 *             if context.captures is not None:
 *                 context.captures.add(self.name)
 */
  }

  /* "flitter/language/tree.pyx":332
 *                 context.errors.add(f"Unbound name '{self.name}'")
 *                 return NoOp
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":313
 *                 program.literal(null_)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         if self.name in context.names:
 *             value = context.names[self.name]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Name._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":334
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Name({self.name!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Name_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":335
 * 
 *     def __repr__(self):
 *         return f'Name({self.name!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Name);
  __pyx_t_2 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_Name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Name);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->name), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":334
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Name({self.name!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Name.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":293
 * 
 * cdef class Name(Expression):
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Name_4name___get__(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Name_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Name_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Name_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Name); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, None), state
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Name.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Name_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Name_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Name_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Name.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Name_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Name_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Name__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Name.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":341
 *     cdef readonly Expression key
 * 
 *     def __init__(self, Expression key):             # <<<<<<<<<<<<<<
 *         self.key = key
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_6Lookup_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_6Lookup_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_key = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 341, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_key = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 341, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Lookup.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "key", 0))) __PYX_ERR(0, 341, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Lookup___init__(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v_self), __pyx_v_key);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_6Lookup___init__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":342
 * 
 *     def __init__(self, Expression key):
 *         self.key = key             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_key);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->key);
  __Pyx_DECREF((PyObject *)__pyx_v_self->key);
  __pyx_v_self->key = __pyx_v_key;

  /* "flitter/language/tree.pyx":341
 *     cdef readonly Expression key
 * 
 *     def __init__(self, Expression key):             # <<<<<<<<<<<<<<
 *         self.key = key
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":344
 *         self.key = key
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         if isinstance(self.key, Literal):
 *             program.lookup_literal((<Literal>self.key).value.intern())
 */

static void __pyx_f_7flitter_8language_4tree_6Lookup__compile(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":345
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         if isinstance(self.key, Literal):             # <<<<<<<<<<<<<<
 *             program.lookup_literal((<Literal>self.key).value.intern())
 *         else:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->key);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":346
 *     cdef void _compile(self, Program program, list lnames):
 *         if isinstance(self.key, Literal):
 *             program.lookup_literal((<Literal>self.key).value.intern())             # <<<<<<<<<<<<<<
 *         else:
 *             self.key._compile(program, lnames)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->key)->value->__pyx_vtab)->intern(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->key)->value, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->lookup_literal(__pyx_v_program, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":345
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         if isinstance(self.key, Literal):             # <<<<<<<<<<<<<<
 *             program.lookup_literal((<Literal>self.key).value.intern())
 *         else:
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":348
 *             program.lookup_literal((<Literal>self.key).value.intern())
 *         else:
 *             self.key._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.lookup()
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->key->__pyx_vtab)->_compile(__pyx_v_self->key, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 348, __pyx_L1_error)

    /* "flitter/language/tree.pyx":349
 *         else:
 *             self.key._compile(program, lnames)
 *             program.lookup()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->lookup(__pyx_v_program, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":344
 *         self.key = key
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         if isinstance(self.key, Literal):
 *             program.lookup_literal((<Literal>self.key).value.intern())
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Lookup._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":351
 *             program.lookup()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression key = self.key._simplify(context)
 *         cdef Vector value
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Lookup__simplify(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_key = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":352
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression key = self.key._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Vector value
 *         if isinstance(key, Literal):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->key->__pyx_vtab)->_simplify(__pyx_v_self->key, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":354
 *         cdef Expression key = self.key._simplify(context)
 *         cdef Vector value
 *         if isinstance(key, Literal):             # <<<<<<<<<<<<<<
 *             if context.state is not None and context.state.contains((<Literal>key).value):
 *                 value = context.state.get_item((<Literal>key).value)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_key), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":355
 *         cdef Vector value
 *         if isinstance(key, Literal):
 *             if context.state is not None and context.state.contains((<Literal>key).value):             # <<<<<<<<<<<<<<
 *                 value = context.state.get_item((<Literal>key).value)
 *                 return Literal(value)
 */
    __pyx_t_3 = (((PyObject *)__pyx_v_context->state) != Py_None);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_key)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_v_context->state->__pyx_vtab)->contains(__pyx_v_context->state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":356
 *         if isinstance(key, Literal):
 *             if context.state is not None and context.state.contains((<Literal>key).value):
 *                 value = context.state.get_item((<Literal>key).value)             # <<<<<<<<<<<<<<
 *                 return Literal(value)
 *         if key is self.key:
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_key)->value);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_v_context->state->__pyx_vtab)->get_item(__pyx_v_context->state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":357
 *             if context.state is not None and context.state.contains((<Literal>key).value):
 *                 value = context.state.get_item((<Literal>key).value)
 *                 return Literal(value)             # <<<<<<<<<<<<<<
 *         if key is self.key:
 *             return self
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_value)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":355
 *         cdef Vector value
 *         if isinstance(key, Literal):
 *             if context.state is not None and context.state.contains((<Literal>key).value):             # <<<<<<<<<<<<<<
 *                 value = context.state.get_item((<Literal>key).value)
 *                 return Literal(value)
 */
    }

    /* "flitter/language/tree.pyx":354
 *         cdef Expression key = self.key._simplify(context)
 *         cdef Vector value
 *         if isinstance(key, Literal):             # <<<<<<<<<<<<<<
 *             if context.state is not None and context.state.contains((<Literal>key).value):
 *                 value = context.state.get_item((<Literal>key).value)
 */
  }

  /* "flitter/language/tree.pyx":358
 *                 value = context.state.get_item((<Literal>key).value)
 *                 return Literal(value)
 *         if key is self.key:             # <<<<<<<<<<<<<<
 *             return self
 *         return Lookup(key)
 */
  __pyx_t_2 = (__pyx_v_key == __pyx_v_self->key);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":359
 *                 return Literal(value)
 *         if key is self.key:
 *             return self             # <<<<<<<<<<<<<<
 *         return Lookup(key)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":358
 *                 value = context.state.get_item((<Literal>key).value)
 *                 return Literal(value)
 *         if key is self.key:             # <<<<<<<<<<<<<<
 *             return self
 *         return Lookup(key)
 */
  }

  /* "flitter/language/tree.pyx":360
 *         if key is self.key:
 *             return self
 *         return Lookup(key)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Lookup), ((PyObject *)__pyx_v_key)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":351
 *             program.lookup()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression key = self.key._simplify(context)
 *         cdef Vector value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Lookup._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":362
 *         return Lookup(key)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Lookup({self.key!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Lookup_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":363
 * 
 *     def __repr__(self):
 *         return f'Lookup({self.key!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Lookup);
  __pyx_t_2 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_Lookup);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Lookup);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->key)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":362
 *         return Lookup(key)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Lookup({self.key!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Lookup.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":339
 * 
 * cdef class Lookup(Expression):
 *     cdef readonly Expression key             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression key):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Lookup_3key___get__(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_3key___get__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->key);
  __pyx_r = ((PyObject *)__pyx_v_self->key);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Lookup_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Lookup_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Lookup_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.key,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->key))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.key,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.key,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.key is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.key,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.key is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->key) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.key is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.key is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Lookup); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_46599698);
    __Pyx_GIVEREF(__pyx_int_46599698);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_46599698)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.key is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, None), state
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Lookup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_46599698);
    __Pyx_GIVEREF(__pyx_int_46599698);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_46599698)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Lookup.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Lookup_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Lookup_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Lookup_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Lookup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Lookup_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Lookup_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Lookup__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Lookup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":371
 *     cdef readonly Expression step
 * 
 *     def __init__(self, Expression start, Expression stop, Expression step):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.stop = stop
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_5Range_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_5Range_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_start = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_stop = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_step = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_stop,&__pyx_n_s_step,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_start)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 371, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_step)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 371, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 371, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_start = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_stop = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
    __pyx_v_step = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 371, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Range.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_start), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "start", 0))) __PYX_ERR(0, 371, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_stop), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "stop", 0))) __PYX_ERR(0, 371, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_step), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "step", 0))) __PYX_ERR(0, 371, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range___init__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self), __pyx_v_start, __pyx_v_stop, __pyx_v_step);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_5Range___init__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_start, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_stop, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_step) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":372
 * 
 *     def __init__(self, Expression start, Expression stop, Expression step):
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.stop = stop
 *         self.step = step
 */
  __Pyx_INCREF((PyObject *)__pyx_v_start);
  __Pyx_GIVEREF((PyObject *)__pyx_v_start);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->start);
  __Pyx_DECREF((PyObject *)__pyx_v_self->start);
  __pyx_v_self->start = __pyx_v_start;

  /* "flitter/language/tree.pyx":373
 *     def __init__(self, Expression start, Expression stop, Expression step):
 *         self.start = start
 *         self.stop = stop             # <<<<<<<<<<<<<<
 *         self.step = step
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_stop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_stop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->stop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->stop);
  __pyx_v_self->stop = __pyx_v_stop;

  /* "flitter/language/tree.pyx":374
 *         self.start = start
 *         self.stop = stop
 *         self.step = step             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_step);
  __Pyx_GIVEREF((PyObject *)__pyx_v_step);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->step);
  __Pyx_DECREF((PyObject *)__pyx_v_self->step);
  __pyx_v_self->step = __pyx_v_step;

  /* "flitter/language/tree.pyx":371
 *     cdef readonly Expression step
 * 
 *     def __init__(self, Expression start, Expression stop, Expression step):             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.stop = stop
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":376
 *         self.step = step
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.start._compile(program, lnames)
 *         self.stop._compile(program, lnames)
 */

static void __pyx_f_7flitter_8language_4tree_5Range__compile(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":377
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.start._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self.stop._compile(program, lnames)
 *         self.step._compile(program, lnames)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->start->__pyx_vtab)->_compile(__pyx_v_self->start, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 377, __pyx_L1_error)

  /* "flitter/language/tree.pyx":378
 *     cdef void _compile(self, Program program, list lnames):
 *         self.start._compile(program, lnames)
 *         self.stop._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self.step._compile(program, lnames)
 *         program.range()
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->stop->__pyx_vtab)->_compile(__pyx_v_self->stop, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 378, __pyx_L1_error)

  /* "flitter/language/tree.pyx":379
 *         self.start._compile(program, lnames)
 *         self.stop._compile(program, lnames)
 *         self.step._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.range()
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->step->__pyx_vtab)->_compile(__pyx_v_self->step, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)

  /* "flitter/language/tree.pyx":380
 *         self.stop._compile(program, lnames)
 *         self.step._compile(program, lnames)
 *         program.range()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->range(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":376
 *         self.step = step
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.start._compile(program, lnames)
 *         self.stop._compile(program, lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Range._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":382
 *         program.range()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression start = self.start._simplify(context)
 *         cdef Expression stop = self.stop._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Range__simplify(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_start = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_stop = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_step = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_result = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":383
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression start = self.start._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression stop = self.stop._simplify(context)
 *         cdef Expression step = self.step._simplify(context)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->start->__pyx_vtab)->_simplify(__pyx_v_self->start, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_start = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":384
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression start = self.start._simplify(context)
 *         cdef Expression stop = self.stop._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression step = self.step._simplify(context)
 *         cdef Vector result
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->stop->__pyx_vtab)->_simplify(__pyx_v_self->stop, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_stop = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":385
 *         cdef Expression start = self.start._simplify(context)
 *         cdef Expression stop = self.stop._simplify(context)
 *         cdef Expression step = self.step._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Vector result
 *         if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->step->__pyx_vtab)->_simplify(__pyx_v_self->step, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_step = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":387
 *         cdef Expression step = self.step._simplify(context)
 *         cdef Vector result
 *         if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):             # <<<<<<<<<<<<<<
 *             result = Vector.__new__(Vector)
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_start), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_stop), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_step), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":388
 *         cdef Vector result
 *         if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):
 *             result = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 *             return Literal(result)
 */
    __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(0, 388, __pyx_L1_error)
    __pyx_v_result = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":389
 *         if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):
 *             result = Vector.__new__(Vector)
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)             # <<<<<<<<<<<<<<
 *             return Literal(result)
 *         if start is self.start and stop is self.stop and step is self.step:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_start)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_stop)->value);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_step)->value);
    __Pyx_INCREF(__pyx_t_5);
    ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_result->__pyx_vtab)->fill_range(__pyx_v_result, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "flitter/language/tree.pyx":390
 *             result = Vector.__new__(Vector)
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 *             return Literal(result)             # <<<<<<<<<<<<<<
 *         if start is self.start and stop is self.stop and step is self.step:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_result)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":387
 *         cdef Expression step = self.step._simplify(context)
 *         cdef Vector result
 *         if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):             # <<<<<<<<<<<<<<
 *             result = Vector.__new__(Vector)
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 */
  }

  /* "flitter/language/tree.pyx":391
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 *             return Literal(result)
 *         if start is self.start and stop is self.stop and step is self.step:             # <<<<<<<<<<<<<<
 *             return self
 *         return Range(start, stop, step)
 */
  __pyx_t_3 = (__pyx_v_start == __pyx_v_self->start);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_stop == __pyx_v_self->stop);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_step == __pyx_v_self->step);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":392
 *             return Literal(result)
 *         if start is self.start and stop is self.stop and step is self.step:
 *             return self             # <<<<<<<<<<<<<<
 *         return Range(start, stop, step)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":391
 *             result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
 *             return Literal(result)
 *         if start is self.start and stop is self.stop and step is self.step:             # <<<<<<<<<<<<<<
 *             return self
 *         return Range(start, stop, step)
 */
  }

  /* "flitter/language/tree.pyx":393
 *         if start is self.start and stop is self.stop and step is self.step:
 *             return self
 *         return Range(start, stop, step)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_v_start);
  __Pyx_GIVEREF((PyObject *)__pyx_v_start);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_start))) __PYX_ERR(0, 393, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_stop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_stop);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_stop))) __PYX_ERR(0, 393, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_step);
  __Pyx_GIVEREF((PyObject *)__pyx_v_step);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, ((PyObject *)__pyx_v_step))) __PYX_ERR(0, 393, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Range), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":382
 *         program.range()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression start = self.start._simplify(context)
 *         cdef Expression stop = self.stop._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Range._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_start);
  __Pyx_XDECREF((PyObject *)__pyx_v_stop);
  __Pyx_XDECREF((PyObject *)__pyx_v_step);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":395
 *         return Range(start, stop, step)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Range({self.start!r}, {self.stop!r}, {self.step!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":396
 * 
 *     def __repr__(self):
 *         return f'Range({self.start!r}, {self.stop!r}, {self.step!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Range);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_Range);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Range);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->start)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->stop)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->step)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":395
 *         return Range(start, stop, step)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Range({self.start!r}, {self.stop!r}, {self.step!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Range.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":367
 * 
 * cdef class Range(Expression):
 *     cdef readonly Expression start             # <<<<<<<<<<<<<<
 *     cdef readonly Expression stop
 *     cdef readonly Expression step
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_5start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_5start_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_5start___get__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_5start___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->start);
  __pyx_r = ((PyObject *)__pyx_v_self->start);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":368
 * cdef class Range(Expression):
 *     cdef readonly Expression start
 *     cdef readonly Expression stop             # <<<<<<<<<<<<<<
 *     cdef readonly Expression step
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_4stop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_4stop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_4stop___get__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4stop___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->stop);
  __pyx_r = ((PyObject *)__pyx_v_self->stop);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":369
 *     cdef readonly Expression start
 *     cdef readonly Expression stop
 *     cdef readonly Expression step             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression start, Expression stop, Expression step):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_4step_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_4step_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_4step___get__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4step___get__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->step);
  __pyx_r = ((PyObject *)__pyx_v_self->step);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Range_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Range_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.start, self.step, self.stop)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->start);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->start);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->start))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->step);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->step);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->step))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->stop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->stop);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->stop))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.start, self.step, self.stop)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.start, self.step, self.stop)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.start is not None or self.step is not None or self.stop is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.start, self.step, self.stop)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.start is not None or self.step is not None or self.stop is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->start) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->step) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->stop) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.start is not None or self.step is not None or self.stop is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.start is not None or self.step is not None or self.stop is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Range); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_264124616);
    __Pyx_GIVEREF(__pyx_int_264124616);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_264124616)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.start is not None or self.step is not None or self.stop is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, None), state
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Range); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_264124616);
    __Pyx_GIVEREF(__pyx_int_264124616);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_264124616)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Range.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Range_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Range_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Range_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Range.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Range_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Range_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Range__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Range.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":402
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, Expression expr):             # <<<<<<<<<<<<<<
 *         self.expr = expr
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_expr,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expr)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 402, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 402, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "expr", 0))) __PYX_ERR(0, 402, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14UnaryOperation___init__(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self), __pyx_v_expr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_14UnaryOperation___init__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":403
 * 
 *     def __init__(self, Expression expr):
 *         self.expr = expr             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->expr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->expr);
  __pyx_v_self->expr = __pyx_v_expr;

  /* "flitter/language/tree.pyx":402
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, Expression expr):             # <<<<<<<<<<<<<<
 *         self.expr = expr
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":405
 *         self.expr = expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.expr._compile(program, lnames)
 *         self._compile_op(program)
 */

static void __pyx_f_7flitter_8language_4tree_14UnaryOperation__compile(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/language/tree.pyx":406
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self._compile_op(program)
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->expr->__pyx_vtab)->_compile(__pyx_v_self->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 406, __pyx_L1_error)

  /* "flitter/language/tree.pyx":407
 *     cdef void _compile(self, Program program, list lnames):
 *         self.expr._compile(program, lnames)
 *         self._compile_op(program)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->_compile_op(__pyx_v_self, __pyx_v_program); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "flitter/language/tree.pyx":405
 *         self.expr = expr
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.expr._compile(program, lnames)
 *         self._compile_op(program)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "flitter/language/tree.pyx":409
 *         self._compile_op(program)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_14UnaryOperation__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":410
 * 
 *     cdef void _compile_op(self, Program program):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 410, __pyx_L1_error)

  /* "flitter/language/tree.pyx":409
 *         self._compile_op(program)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":412
 *         raise NotImplementedError()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.expr!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14UnaryOperation_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_2__repr__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":413
 * 
 *     def __repr__(self):
 *         return f'{self.__class__.__name__}({self.expr!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->expr)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":412
 *         raise NotImplementedError()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.expr!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":400
 * 
 * cdef class UnaryOperation(Expression):
 *     cdef readonly Expression expr             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression expr):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_4expr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_4expr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14UnaryOperation_4expr___get__(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __pyx_r = ((PyObject *)__pyx_v_self->expr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14UnaryOperation_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_UnaryOperation); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_UnaryOperation); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14UnaryOperation_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_14UnaryOperation_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_UnaryOperation__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.UnaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":417
 * 
 * cdef class Negative(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.neg()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_8Negative__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":418
 * cdef class Negative(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.neg()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->neg(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":417
 * 
 * cdef class Negative(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.neg()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Negative._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":420
 *         program.neg()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Negative__simplify(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_maths = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":421
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.neg())
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":422
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.neg())
 *         if isinstance(expr, Negative):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":423
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.neg())             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Negative):
 *             expr = Positive((<Negative>expr).expr)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->neg(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":422
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.neg())
 *         if isinstance(expr, Negative):
 */
  }

  /* "flitter/language/tree.pyx":424
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.neg())
 *         if isinstance(expr, Negative):             # <<<<<<<<<<<<<<
 *             expr = Positive((<Negative>expr).expr)
 *             return expr._simplify(context)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":425
 *             return Literal((<Literal>expr).value.neg())
 *         if isinstance(expr, Negative):
 *             expr = Positive((<Negative>expr).expr)             # <<<<<<<<<<<<<<
 *             return expr._simplify(context)
 *         cdef MathsBinaryOperation maths
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_expr)->__pyx_base.expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":426
 *         if isinstance(expr, Negative):
 *             expr = Positive((<Negative>expr).expr)
 *             return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef MathsBinaryOperation maths
 *         if isinstance(expr, (Multiply, Divide)):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":424
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.neg())
 *         if isinstance(expr, Negative):             # <<<<<<<<<<<<<<
 *             expr = Positive((<Negative>expr).expr)
 *             return expr._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":428
 *             return expr._simplify(context)
 *         cdef MathsBinaryOperation maths
 *         if isinstance(expr, (Multiply, Divide)):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal):
 */
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Multiply); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Divide); 
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":429
 *         cdef MathsBinaryOperation maths
 *         if isinstance(expr, (Multiply, Divide)):
 *             maths = expr             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal):
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 */
    if (!(likely(((((PyObject *)__pyx_v_expr)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 429, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_expr);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":430
 *         if isinstance(expr, (Multiply, Divide)):
 *             maths = expr
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":431
 *             maths = expr
 *             if isinstance(maths.left, Literal):
 *                 expr = type(expr)(Negative(maths.left), maths.right)             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *             if isinstance(maths.right, Literal):
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_maths->__pyx_base.left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))));
      __pyx_t_5 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))); __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_1, ((PyObject *)__pyx_v_maths->__pyx_base.right)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 431, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "flitter/language/tree.pyx":432
 *             if isinstance(maths.left, Literal):
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *             if isinstance(maths.right, Literal):
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":430
 *         if isinstance(expr, (Multiply, Divide)):
 *             maths = expr
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 */
    }

    /* "flitter/language/tree.pyx":433
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 *             if isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 *                 return expr._simplify(context)
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_maths->__pyx_base.right);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":434
 *                 return expr._simplify(context)
 *             if isinstance(maths.right, Literal):
 *                 expr = type(expr)(maths.left, Negative(maths.right))             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Add):
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_maths->__pyx_base.right)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))));
      __pyx_t_1 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))); __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, ((PyObject *)__pyx_v_maths->__pyx_base.left), __pyx_t_5};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "flitter/language/tree.pyx":435
 *             if isinstance(maths.right, Literal):
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, Add):
 *             maths = expr
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":433
 *                 expr = type(expr)(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 *             if isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 *                 return expr._simplify(context)
 */
    }

    /* "flitter/language/tree.pyx":428
 *             return expr._simplify(context)
 *         cdef MathsBinaryOperation maths
 *         if isinstance(expr, (Multiply, Divide)):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal):
 */
    goto __pyx_L5;
  }

  /* "flitter/language/tree.pyx":436
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Add):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Add); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":437
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Add):
 *             maths = expr             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 */
    if (!(likely(((((PyObject *)__pyx_v_expr)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 437, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_expr);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":438
 *         elif isinstance(expr, Add):
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 *                 return expr._simplify(context)
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_3 = ((PyObject *)__pyx_v_maths->__pyx_base.right);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_4;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":439
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), Negative(maths.right))             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Subtract):
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_maths->__pyx_base.left)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_maths->__pyx_base.right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error);
      __pyx_t_3 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":440
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, Subtract):
 *             maths = expr
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":438
 *         elif isinstance(expr, Add):
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 *                 return expr._simplify(context)
 */
    }

    /* "flitter/language/tree.pyx":436
 *                 expr = type(expr)(maths.left, Negative(maths.right))
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Add):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
    goto __pyx_L5;
  }

  /* "flitter/language/tree.pyx":441
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Subtract):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Subtract); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":442
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Subtract):
 *             maths = expr             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), maths.right)
 */
    if (!(likely(((((PyObject *)__pyx_v_expr)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 442, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_v_expr);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":443
 *         elif isinstance(expr, Subtract):
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = Add(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_4;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":444
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), maths.right)             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         if expr is self.expr:
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_maths->__pyx_base.left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 444, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_maths->__pyx_base.right))) __PYX_ERR(0, 444, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":445
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 expr = Add(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":443
 *         elif isinstance(expr, Subtract):
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 expr = Add(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 */
    }

    /* "flitter/language/tree.pyx":441
 *                 expr = Add(Negative(maths.left), Negative(maths.right))
 *                 return expr._simplify(context)
 *         elif isinstance(expr, Subtract):             # <<<<<<<<<<<<<<
 *             maths = expr
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
  }
  __pyx_L5:;

  /* "flitter/language/tree.pyx":446
 *                 expr = Add(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Negative(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":447
 *                 return expr._simplify(context)
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Negative(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":446
 *                 expr = Add(Negative(maths.left), maths.right)
 *                 return expr._simplify(context)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Negative(expr)
 */
  }

  /* "flitter/language/tree.pyx":448
 *         if expr is self.expr:
 *             return self
 *         return Negative(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":420
 *         program.neg()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.tree.Negative._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_maths);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Negative_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Negative_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Negative_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Negative_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Negative___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Negative___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Negative); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Negative.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Negative_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Negative_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Negative_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Negative_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Negative.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Negative_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Negative_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Negative__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Negative.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":452
 * 
 * cdef class Positive(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.pos()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_8Positive__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":453
 * cdef class Positive(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.pos()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pos(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":452
 * 
 * cdef class Positive(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.pos()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Positive._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":455
 *         program.pos()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Positive__simplify(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":456
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.pos())
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":457
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.pos())
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":458
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.pos())             # <<<<<<<<<<<<<<
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 *             return expr._simplify(context)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->pos(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":457
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.pos())
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 */
  }

  /* "flitter/language/tree.pyx":459
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.pos())
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):             # <<<<<<<<<<<<<<
 *             return expr._simplify(context)
 *         if expr is self.expr:
 */
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Positive); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); 
  __pyx_t_2 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":460
 *             return Literal((<Literal>expr).value.pos())
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 *             return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":459
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.pos())
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):             # <<<<<<<<<<<<<<
 *             return expr._simplify(context)
 *         if expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":461
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 *             return expr._simplify(context)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Positive(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":462
 *             return expr._simplify(context)
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Positive(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":461
 *         if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
 *             return expr._simplify(context)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Positive(expr)
 */
  }

  /* "flitter/language/tree.pyx":463
 *         if expr is self.expr:
 *             return self
 *         return Positive(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":455
 *         program.pos()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Positive._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Positive_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Positive_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Positive_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Positive_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Positive___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Positive *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Positive___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Positive); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Positive.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Positive_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Positive_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Positive_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Positive_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Positive.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Positive_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Positive *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Positive_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Positive__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Positive.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":467
 * 
 * cdef class Not(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.not_()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_3Not__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":468
 * cdef class Not(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.not_()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->not_(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":467
 * 
 * cdef class Not(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.not_()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Not._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":470
 *         program.not_()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Not__simplify(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":471
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":472
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 *         if expr is self.expr:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":473
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_false_);
      __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_false_);
    } else {
      __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_true_);
      __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
    }
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":472
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 *         if expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":474
 *         if isinstance(expr, Literal):
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Not(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":475
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Not(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":474
 *         if isinstance(expr, Literal):
 *             return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Not(expr)
 */
  }

  /* "flitter/language/tree.pyx":476
 *         if expr is self.expr:
 *             return self
 *         return Not(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Not), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":470
 *         program.not_()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Not._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Not_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Not_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Not_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Not_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Not___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Not *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Not___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Not); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Not); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Not.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Not_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Not_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Not_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Not_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Not.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Not_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Not *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Not_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Not__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Not.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":480
 * 
 * cdef class Ceil(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ceil()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_4Ceil__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":481
 * cdef class Ceil(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.ceil()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->ceil(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":480
 * 
 * cdef class Ceil(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ceil()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Ceil._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":483
 *         program.ceil()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Ceil__simplify(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":484
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.ceil())
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":485
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.ceil())
 *         if expr is self.expr:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":486
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.ceil())             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->ceil(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":485
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.ceil())
 *         if expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":487
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.ceil())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Ceil(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":488
 *             return Literal((<Literal>expr).value.ceil())
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Ceil(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":487
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.ceil())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Ceil(expr)
 */
  }

  /* "flitter/language/tree.pyx":489
 *         if expr is self.expr:
 *             return self
 *         return Ceil(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Ceil), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":483
 *         program.ceil()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Ceil._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Ceil_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Ceil_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Ceil_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Ceil_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Ceil___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Ceil *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Ceil___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Ceil); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Ceil); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Ceil.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Ceil_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Ceil_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Ceil_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Ceil_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Ceil.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Ceil_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Ceil *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Ceil_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Ceil__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Ceil.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":493
 * 
 * cdef class Floor(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.floor()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_5Floor__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":494
 * cdef class Floor(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.floor()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->floor(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":493
 * 
 * cdef class Floor(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.floor()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Floor._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":496
 *         program.floor()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Floor__simplify(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":497
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.floor())
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":498
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.floor())
 *         if expr is self.expr:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":499
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.floor())             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->floor(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":498
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.floor())
 *         if expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":500
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.floor())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Floor(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":501
 *             return Literal((<Literal>expr).value.floor())
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Floor(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":500
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.floor())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Floor(expr)
 */
  }

  /* "flitter/language/tree.pyx":502
 *         if expr is self.expr:
 *             return self
 *         return Floor(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":496
 *         program.floor()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Floor._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Floor_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Floor_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Floor_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Floor_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Floor___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Floor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Floor___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Floor); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Floor); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Floor.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Floor_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Floor_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Floor_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Floor_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Floor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Floor_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Floor *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Floor_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Floor__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Floor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":506
 * 
 * cdef class Fract(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.fract()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_5Fract__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":507
 * cdef class Fract(UnaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.fract()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->fract(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":506
 * 
 * cdef class Fract(UnaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.fract()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Fract._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":509
 *         program.fract()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Fract__simplify(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":510
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.fract())
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.expr->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":511
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.fract())
 *         if expr is self.expr:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":512
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.fract())             # <<<<<<<<<<<<<<
 *         if expr is self.expr:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value->__pyx_vtab)->fract(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":511
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *             return Literal((<Literal>expr).value.fract())
 *         if expr is self.expr:
 */
  }

  /* "flitter/language/tree.pyx":513
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.fract())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Fract(expr)
 */
  __pyx_t_2 = (__pyx_v_expr == __pyx_v_self->__pyx_base.expr);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":514
 *             return Literal((<Literal>expr).value.fract())
 *         if expr is self.expr:
 *             return self             # <<<<<<<<<<<<<<
 *         return Fract(expr)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":513
 *         if isinstance(expr, Literal):
 *             return Literal((<Literal>expr).value.fract())
 *         if expr is self.expr:             # <<<<<<<<<<<<<<
 *             return self
 *         return Fract(expr)
 */
  }

  /* "flitter/language/tree.pyx":515
 *         if expr is self.expr:
 *             return self
 *         return Fract(expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract), ((PyObject *)__pyx_v_expr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":509
 *         program.fract()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         if isinstance(expr, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Fract._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Fract_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Fract_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Fract_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Fract_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Fract___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Fract *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Fract___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.expr) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Fract); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, None), state
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Fract); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_101425629);
    __Pyx_GIVEREF(__pyx_int_101425629);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_101425629)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Fract.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Fract_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Fract_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Fract_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Fract_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Fract.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Fract_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Fract *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Fract_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Fract__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Fract.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":522
 *     cdef readonly Expression right
 * 
 *     def __init__(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         self.left = left
 *         self.right = right
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_left)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 522, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_right)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 522, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 522, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 522, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 522, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "left", 0))) __PYX_ERR(0, 522, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "right", 0))) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation___init__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self), __pyx_v_left, __pyx_v_right);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_15BinaryOperation___init__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":523
 * 
 *     def __init__(self, Expression left, Expression right):
 *         self.left = left             # <<<<<<<<<<<<<<
 *         self.right = right
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->left);
  __Pyx_DECREF((PyObject *)__pyx_v_self->left);
  __pyx_v_self->left = __pyx_v_left;

  /* "flitter/language/tree.pyx":524
 *     def __init__(self, Expression left, Expression right):
 *         self.left = left
 *         self.right = right             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_right);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->right);
  __Pyx_DECREF((PyObject *)__pyx_v_self->right);
  __pyx_v_self->right = __pyx_v_right;

  /* "flitter/language/tree.pyx":522
 *     cdef readonly Expression right
 * 
 *     def __init__(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         self.left = left
 *         self.right = right
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":526
 *         self.right = right
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.left._compile(program, lnames)
 *         self.right._compile(program, lnames)
 */

static void __pyx_f_7flitter_8language_4tree_15BinaryOperation__compile(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/language/tree.pyx":527
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.left._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self.right._compile(program, lnames)
 *         self._compile_op(program)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->left->__pyx_vtab)->_compile(__pyx_v_self->left, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L1_error)

  /* "flitter/language/tree.pyx":528
 *     cdef void _compile(self, Program program, list lnames):
 *         self.left._compile(program, lnames)
 *         self.right._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self._compile_op(program)
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->right->__pyx_vtab)->_compile(__pyx_v_self->right, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)

  /* "flitter/language/tree.pyx":529
 *         self.left._compile(program, lnames)
 *         self.right._compile(program, lnames)
 *         self._compile_op(program)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->_compile_op(__pyx_v_self, __pyx_v_program); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 529, __pyx_L1_error)

  /* "flitter/language/tree.pyx":526
 *         self.right = right
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.left._compile(program, lnames)
 *         self.right._compile(program, lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "flitter/language/tree.pyx":531
 *         self._compile_op(program)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_15BinaryOperation__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":532
 * 
 *     cdef void _compile_op(self, Program program):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 532, __pyx_L1_error)

  /* "flitter/language/tree.pyx":531
 *         self._compile_op(program)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":534
 *         raise NotImplementedError()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation__simplify(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right = 0;
  int __pyx_v_literal_left;
  int __pyx_v_literal_right;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":535
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->left->__pyx_vtab)->_simplify(__pyx_v_self->left, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":536
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->right->__pyx_vtab)->_simplify(__pyx_v_self->right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":537
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)             # <<<<<<<<<<<<<<
 *         cdef bint literal_right = isinstance(right, Literal)
 *         cdef Expression expr
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_v_literal_left = __pyx_t_2;

  /* "flitter/language/tree.pyx":538
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)             # <<<<<<<<<<<<<<
 *         cdef Expression expr
 *         if literal_left and literal_right:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_v_literal_right = __pyx_t_2;

  /* "flitter/language/tree.pyx":540
 *         cdef bint literal_right = isinstance(right, Literal)
 *         cdef Expression expr
 *         if literal_left and literal_right:             # <<<<<<<<<<<<<<
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:
 */
  if (__pyx_v_literal_left) {
  } else {
    __pyx_t_2 = __pyx_v_literal_left;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_literal_right;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":541
 *         cdef Expression expr
 *         if literal_left and literal_right:
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))             # <<<<<<<<<<<<<<
 *         elif literal_left:
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_right)->value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->op(__pyx_v_self, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":540
 *         cdef bint literal_right = isinstance(right, Literal)
 *         cdef Expression expr
 *         if literal_left and literal_right:             # <<<<<<<<<<<<<<
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:
 */
  }

  /* "flitter/language/tree.pyx":542
 *         if literal_left and literal_right:
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:             # <<<<<<<<<<<<<<
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 *                 return expr._simplify(context)
 */
  if (__pyx_v_literal_left) {

    /* "flitter/language/tree.pyx":543
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         elif literal_right:
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->constant_left(__pyx_v_self, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3), __pyx_v_right)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = (((PyObject *)__pyx_t_4) != Py_None);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":544
 *         elif literal_left:
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         elif literal_right:
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":543
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         elif literal_right:
 */
    }

    /* "flitter/language/tree.pyx":542
 *         if literal_left and literal_right:
 *             return Literal(self.op((<Literal>left).value, (<Literal>right).value))
 *         elif literal_left:             # <<<<<<<<<<<<<<
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 *                 return expr._simplify(context)
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":545
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 *                 return expr._simplify(context)
 *         elif literal_right:             # <<<<<<<<<<<<<<
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 *                 return expr._simplify(context)
 */
  if (__pyx_v_literal_right) {

    /* "flitter/language/tree.pyx":546
 *                 return expr._simplify(context)
 *         elif literal_right:
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         if (expr := self.additional_rules(left, right)) is not None:
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_right)->value);
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->constant_right(__pyx_v_self, __pyx_v_left, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = (((PyObject *)__pyx_t_3) != Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":547
 *         elif literal_right:
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 *                 return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if (expr := self.additional_rules(left, right)) is not None:
 *             return expr._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":546
 *                 return expr._simplify(context)
 *         elif literal_right:
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:             # <<<<<<<<<<<<<<
 *                 return expr._simplify(context)
 *         if (expr := self.additional_rules(left, right)) is not None:
 */
    }

    /* "flitter/language/tree.pyx":545
 *             if (expr := self.constant_left((<Literal>left).value, right)) is not None:
 *                 return expr._simplify(context)
 *         elif literal_right:             # <<<<<<<<<<<<<<
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 *                 return expr._simplify(context)
 */
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":548
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 *                 return expr._simplify(context)
 *         if (expr := self.additional_rules(left, right)) is not None:             # <<<<<<<<<<<<<<
 *             return expr._simplify(context)
 *         if left is self.left and right is self.right:
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self->__pyx_base.__pyx_vtab)->additional_rules(__pyx_v_self, __pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = (((PyObject *)__pyx_t_3) != Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":549
 *                 return expr._simplify(context)
 *         if (expr := self.additional_rules(left, right)) is not None:
 *             return expr._simplify(context)             # <<<<<<<<<<<<<<
 *         if left is self.left and right is self.right:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_simplify(__pyx_v_expr, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":548
 *             if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
 *                 return expr._simplify(context)
 *         if (expr := self.additional_rules(left, right)) is not None:             # <<<<<<<<<<<<<<
 *             return expr._simplify(context)
 *         if left is self.left and right is self.right:
 */
  }

  /* "flitter/language/tree.pyx":550
 *         if (expr := self.additional_rules(left, right)) is not None:
 *             return expr._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return type(self)(left, right)
 */
  __pyx_t_5 = (__pyx_v_left == __pyx_v_self->left);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_right == __pyx_v_self->right);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":551
 *             return expr._simplify(context)
 *         if left is self.left and right is self.right:
 *             return self             # <<<<<<<<<<<<<<
 *         return type(self)(left, right)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":550
 *         if (expr := self.additional_rules(left, right)) is not None:
 *             return expr._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return type(self)(left, right)
 */
  }

  /* "flitter/language/tree.pyx":552
 *         if left is self.left and right is self.right:
 *             return self
 *         return type(self)(left, right)             # <<<<<<<<<<<<<<
 * 
 *     cdef Vector op(self, Vector left, Vector right):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __pyx_t_4 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))); __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_1, ((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_right)};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 552, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":534
 *         raise NotImplementedError()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":554
 *         return type(self)(left, right)
 * 
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_15BinaryOperation_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":555
 * 
 *     cdef Vector op(self, Vector left, Vector right):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 555, __pyx_L1_error)

  /* "flitter/language/tree.pyx":554
 *         return type(self)(left, right)
 * 
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":557
 *         raise NotImplementedError()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("constant_left", 1);

  /* "flitter/language/tree.pyx":558
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":557
 *         raise NotImplementedError()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":560
 *         return None
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":561
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":560
 *         return None
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":563
 *         return None
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_15BinaryOperation_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("additional_rules", 1);

  /* "flitter/language/tree.pyx":564
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":563
 *         return None
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":566
 *         return None
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.left!r}, {self.right!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_2__repr__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":567
 * 
 *     def __repr__(self):
 *         return f'{self.__class__.__name__}({self.left!r}, {self.right!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->left)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->right)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":566
 *         return None
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.left!r}, {self.right!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":519
 * 
 * cdef class BinaryOperation(Expression):
 *     cdef readonly Expression left             # <<<<<<<<<<<<<<
 *     cdef readonly Expression right
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_4left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_4left_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation_4left___get__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_4left___get__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->left);
  __pyx_r = ((PyObject *)__pyx_v_self->left);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":520
 * cdef class BinaryOperation(Expression):
 *     cdef readonly Expression left
 *     cdef readonly Expression right             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression left, Expression right):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5right_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation_5right___get__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_5right___get__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->right);
  __pyx_r = ((PyObject *)__pyx_v_self->right);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_BinaryOperation); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_BinaryOperation); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_15BinaryOperation_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_15BinaryOperation_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_BinaryOperation__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.BinaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":571
 * 
 * cdef class MathsBinaryOperation(BinaryOperation):
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_20MathsBinaryOperation__simplify(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":572
 * cdef class MathsBinaryOperation(BinaryOperation):
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr=BinaryOperation._simplify(self, context)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, MathsBinaryOperation):
 *             if isinstance(expr.left, Positive):
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7flitter_8language_4tree_15BinaryOperation__simplify(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self), __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":573
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *             if isinstance(expr.left, Positive):
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":574
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):
 *             if isinstance(expr.left, Positive):             # <<<<<<<<<<<<<<
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 *             elif isinstance(expr.right, Positive):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Positive); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":575
 *         if isinstance(expr, MathsBinaryOperation):
 *             if isinstance(expr.left, Positive):
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)             # <<<<<<<<<<<<<<
 *             elif isinstance(expr.right, Positive):
 *                 return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))));
      __pyx_t_5 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))); __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_4, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1)->__pyx_vtab)->_simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1), __pyx_v_context)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":574
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):
 *             if isinstance(expr.left, Positive):             # <<<<<<<<<<<<<<
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 *             elif isinstance(expr.right, Positive):
 */
    }

    /* "flitter/language/tree.pyx":576
 *             if isinstance(expr.left, Positive):
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 *             elif isinstance(expr.right, Positive):             # <<<<<<<<<<<<<<
 *                 return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)
 *         return expr
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_5, __pyx_ptype_7flitter_8language_4tree_Positive); 
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":577
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 *             elif isinstance(expr.right, Positive):
 *                 return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)             # <<<<<<<<<<<<<<
 *         return expr
 * 
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_expr), __pyx_n_s_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))));
      __pyx_t_3 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_expr))); __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_1, __pyx_t_4};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 577, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5)->__pyx_vtab)->_simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5), __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":576
 *             if isinstance(expr.left, Positive):
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 *             elif isinstance(expr.right, Positive):             # <<<<<<<<<<<<<<
 *                 return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)
 *         return expr
 */
    }

    /* "flitter/language/tree.pyx":573
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):             # <<<<<<<<<<<<<<
 *             if isinstance(expr.left, Positive):
 *                 return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":578
 *             elif isinstance(expr.right, Positive):
 *                 return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)
 *         return expr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __pyx_r = __pyx_v_expr;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":571
 * 
 * cdef class MathsBinaryOperation(BinaryOperation):
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr=BinaryOperation._simplify(self, context)
 *         if isinstance(expr, MathsBinaryOperation):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.tree.MathsBinaryOperation._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_MathsBinaryOperat); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_MathsBinaryOperat); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.MathsBinaryOperation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.MathsBinaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_20MathsBinaryOperation_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_MathsBinaryOperation__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.MathsBinaryOperation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":582
 * 
 * cdef class Add(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.add(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_3Add_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":583
 * cdef class Add(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.add(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->add(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":582
 * 
 * cdef class Add(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.add(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Add.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":585
 *         return left.add(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         if program.last_instruction().code == OpCode.Mul:
 *             program.pop_instruction()
 */

static void __pyx_f_7flitter_8language_4tree_3Add__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":586
 * 
 *     cdef void _compile_op(self, Program program):
 *         if program.last_instruction().code == OpCode.Mul:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.mul_add()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1)->code == __pyx_e_7flitter_8language_2vm_Mul);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":587
 *     cdef void _compile_op(self, Program program):
 *         if program.last_instruction().code == OpCode.Mul:
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.mul_add()
 *         else:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":588
 *         if program.last_instruction().code == OpCode.Mul:
 *             program.pop_instruction()
 *             program.mul_add()             # <<<<<<<<<<<<<<
 *         else:
 *             program.add()
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->mul_add(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":586
 * 
 *     cdef void _compile_op(self, Program program):
 *         if program.last_instruction().code == OpCode.Mul:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.mul_add()
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":590
 *             program.mul_add()
 *         else:
 *             program.add()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->add(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":585
 *         return left.add(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         if program.last_instruction().code == OpCode.Mul:
 *             program.pop_instruction()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Add._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":592
 *             program.add()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(false_):
 *             return Positive(right)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_left", 1);

  /* "flitter/language/tree.pyx":593
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):             # <<<<<<<<<<<<<<
 *             return Positive(right)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_false_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->eq(__pyx_v_left, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":594
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):
 *             return Positive(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_right)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":593
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):             # <<<<<<<<<<<<<<
 *             return Positive(right)
 * 
 */
  }

  /* "flitter/language/tree.pyx":592
 *             program.add()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(false_):
 *             return Positive(right)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Add.constant_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":596
 *             return Positive(right)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return self.constant_left(right, left)
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_constant_right(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":597
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         return self.constant_left(right, left)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Add *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.constant_left(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self), __pyx_v_right, __pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":596
 *             return Positive(right)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return self.constant_left(right, left)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Add.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":599
 *         return self.constant_left(right, left)
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):
 *             return Add(right, left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Add_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("additional_rules", 1);

  /* "flitter/language/tree.pyx":600
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):             # <<<<<<<<<<<<<<
 *             return Add(right, left)
 *         if isinstance(right, Negative):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Multiply); 
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Multiply); 
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":601
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):
 *             return Add(right, left)             # <<<<<<<<<<<<<<
 *         if isinstance(right, Negative):
 *             return Subtract(left, (<Negative>right).expr)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF((PyObject *)__pyx_v_right);
    __Pyx_GIVEREF((PyObject *)__pyx_v_right);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_right))) __PYX_ERR(0, 601, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __Pyx_GIVEREF((PyObject *)__pyx_v_left);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 601, __pyx_L1_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":600
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):             # <<<<<<<<<<<<<<
 *             return Add(right, left)
 *         if isinstance(right, Negative):
 */
  }

  /* "flitter/language/tree.pyx":602
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):
 *             return Add(right, left)
 *         if isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Subtract(left, (<Negative>right).expr)
 *         if isinstance(left, Negative):
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":603
 *             return Add(right, left)
 *         if isinstance(right, Negative):
 *             return Subtract(left, (<Negative>right).expr)             # <<<<<<<<<<<<<<
 *         if isinstance(left, Negative):
 *             return Subtract(right, (<Negative>left).expr)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __Pyx_GIVEREF((PyObject *)__pyx_v_left);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 603, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr))) __PYX_ERR(0, 603, __pyx_L1_error);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Subtract), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":602
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):
 *             return Add(right, left)
 *         if isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Subtract(left, (<Negative>right).expr)
 *         if isinstance(left, Negative):
 */
  }

  /* "flitter/language/tree.pyx":604
 *         if isinstance(right, Negative):
 *             return Subtract(left, (<Negative>right).expr)
 *         if isinstance(left, Negative):             # <<<<<<<<<<<<<<
 *             return Subtract(right, (<Negative>left).expr)
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":605
 *             return Subtract(left, (<Negative>right).expr)
 *         if isinstance(left, Negative):
 *             return Subtract(right, (<Negative>left).expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF((PyObject *)__pyx_v_right);
    __Pyx_GIVEREF((PyObject *)__pyx_v_right);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_right))) __PYX_ERR(0, 605, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_left)->__pyx_base.expr);
    __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_left)->__pyx_base.expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_left)->__pyx_base.expr))) __PYX_ERR(0, 605, __pyx_L1_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Subtract), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":604
 *         if isinstance(right, Negative):
 *             return Subtract(left, (<Negative>right).expr)
 *         if isinstance(left, Negative):             # <<<<<<<<<<<<<<
 *             return Subtract(right, (<Negative>left).expr)
 * 
 */
  }

  /* "flitter/language/tree.pyx":599
 *         return self.constant_left(right, left)
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         if not isinstance(right, Multiply) and isinstance(left, Multiply):
 *             return Add(right, left)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Add.additional_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Add_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Add_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Add_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Add_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Add___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Add *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Add___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Add); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Add); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Add.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Add_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Add_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Add_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Add_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Add.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Add_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Add *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Add_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Add__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Add.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":609
 * 
 * cdef class Subtract(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.sub(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8Subtract_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":610
 * cdef class Subtract(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.sub(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->sub(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":609
 * 
 * cdef class Subtract(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.sub(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":612
 *         return left.sub(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.sub()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_8Subtract__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":613
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.sub()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->sub(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":612
 *         return left.sub(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.sub()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Subtract._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":615
 *         program.sub()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(false_):
 *             return Negative(right)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_left", 1);

  /* "flitter/language/tree.pyx":616
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):             # <<<<<<<<<<<<<<
 *             return Negative(right)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_false_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->eq(__pyx_v_left, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":617
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):
 *             return Negative(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_right)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":616
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(false_):             # <<<<<<<<<<<<<<
 *             return Negative(right)
 * 
 */
  }

  /* "flitter/language/tree.pyx":615
 *         program.sub()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(false_):
 *             return Negative(right)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.constant_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":619
 *             return Negative(right)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(false_):
 *             return Positive(left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":620
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(false_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 *         return Add(left, Literal(right.neg()))
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_false_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->eq(__pyx_v_right, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":621
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(false_):
 *             return Positive(left)             # <<<<<<<<<<<<<<
 *         return Add(left, Literal(right.neg()))
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":620
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(false_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 *         return Add(left, Literal(right.neg()))
 */
  }

  /* "flitter/language/tree.pyx":622
 *         if right.eq(false_):
 *             return Positive(left)
 *         return Add(left, Literal(right.neg()))             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->neg(__pyx_v_right)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 622, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":619
 *             return Negative(right)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(false_):
 *             return Positive(left)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":624
 *         return Add(left, Literal(right.neg()))
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         if isinstance(right, Negative):
 *             return Add(left, (<Negative>right).expr)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Subtract_additional_rules(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("additional_rules", 1);

  /* "flitter/language/tree.pyx":625
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Add(left, (<Negative>right).expr)
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (__pyx_t_1) {

    /* "flitter/language/tree.pyx":626
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if isinstance(right, Negative):
 *             return Add(left, (<Negative>right).expr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __Pyx_GIVEREF((PyObject *)__pyx_v_left);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 626, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr))) __PYX_ERR(0, 626, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":625
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):
 *         if isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Add(left, (<Negative>right).expr)
 * 
 */
  }

  /* "flitter/language/tree.pyx":624
 *         return Add(left, Literal(right.neg()))
 * 
 *     cdef Expression additional_rules(self, Expression left, Expression right):             # <<<<<<<<<<<<<<
 *         if isinstance(right, Negative):
 *             return Add(left, (<Negative>right).expr)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.additional_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Subtract_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Subtract_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Subtract_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Subtract_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Subtract___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Subtract *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Subtract___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Subtract); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Subtract); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Subtract_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Subtract_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Subtract_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Subtract_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Subtract.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Subtract_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Subtract *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Subtract_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Subtract__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Subtract.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":630
 * 
 * cdef class Multiply(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.mul(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8Multiply_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":631
 * cdef class Multiply(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mul(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->mul(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":630
 * 
 * cdef class Multiply(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.mul(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Multiply.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":633
 *         return left.mul(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.mul()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_8Multiply__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":634
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.mul()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->mul(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":633
 *         return left.mul(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.mul()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Multiply._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":636
 *         program.mul()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(true_):
 *             return Positive(right)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Multiply_constant_left(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v_maths = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_left", 1);

  /* "flitter/language/tree.pyx":637
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(right)
 *         if left.eq(minusone_):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->eq(__pyx_v_left, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":638
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(true_):
 *             return Positive(right)             # <<<<<<<<<<<<<<
 *         if left.eq(minusone_):
 *             return Negative(right)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_right)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":637
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):
 *         if left.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(right)
 *         if left.eq(minusone_):
 */
  }

  /* "flitter/language/tree.pyx":639
 *         if left.eq(true_):
 *             return Positive(right)
 *         if left.eq(minusone_):             # <<<<<<<<<<<<<<
 *             return Negative(right)
 *         cdef MathsBinaryOperation maths
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_7flitter_5model_minusone_);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->eq(__pyx_v_left, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":640
 *             return Positive(right)
 *         if left.eq(minusone_):
 *             return Negative(right)             # <<<<<<<<<<<<<<
 *         cdef MathsBinaryOperation maths
 *         if isinstance(right, Add) or isinstance(right, Subtract):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), ((PyObject *)__pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":639
 *         if left.eq(true_):
 *             return Positive(right)
 *         if left.eq(minusone_):             # <<<<<<<<<<<<<<
 *             return Negative(right)
 *         cdef MathsBinaryOperation maths
 */
  }

  /* "flitter/language/tree.pyx":642
 *             return Negative(right)
 *         cdef MathsBinaryOperation maths
 *         if isinstance(right, Add) or isinstance(right, Subtract):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Add); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Subtract); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":643
 *         cdef MathsBinaryOperation maths
 *         if isinstance(right, Add) or isinstance(right, Subtract):
 *             maths = right             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 */
    if (!(likely(((((PyObject *)__pyx_v_right)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 643, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_v_right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":644
 *         if isinstance(right, Add) or isinstance(right, Subtract):
 *             maths = right
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 *         elif isinstance(right, Multiply):
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_4;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":645
 *             maths = right
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))             # <<<<<<<<<<<<<<
 *         elif isinstance(right, Multiply):
 *             maths = right
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_maths->__pyx_base.left))) __PYX_ERR(0, 645, __pyx_L1_error);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.right))) __PYX_ERR(0, 645, __pyx_L1_error);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_maths))));
      __pyx_t_6 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_maths))); __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_2, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":644
 *         if isinstance(right, Add) or isinstance(right, Subtract):
 *             maths = right
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 *         elif isinstance(right, Multiply):
 */
    }

    /* "flitter/language/tree.pyx":642
 *             return Negative(right)
 *         cdef MathsBinaryOperation maths
 *         if isinstance(right, Add) or isinstance(right, Subtract):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 */
    goto __pyx_L5;
  }

  /* "flitter/language/tree.pyx":646
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 *         elif isinstance(right, Multiply):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal):
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Multiply); 
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":647
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 *         elif isinstance(right, Multiply):
 *             maths = right             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 */
    if (!(likely(((((PyObject *)__pyx_v_right)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 647, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_v_right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":648
 *         elif isinstance(right, Multiply):
 *             maths = right
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 *             if isinstance(maths.right, Literal):
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":649
 *             maths = right
 *             if isinstance(maths.left, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)             # <<<<<<<<<<<<<<
 *             if isinstance(maths.right, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.left))) __PYX_ERR(0, 649, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.right))) __PYX_ERR(0, 649, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":648
 *         elif isinstance(right, Multiply):
 *             maths = right
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 *             if isinstance(maths.right, Literal):
 */
    }

    /* "flitter/language/tree.pyx":650
 *             if isinstance(maths.left, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 *             if isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 *         elif isinstance(right, Divide):
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":651
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 *             if isinstance(maths.right, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)             # <<<<<<<<<<<<<<
 *         elif isinstance(right, Divide):
 *             maths = right
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.right))) __PYX_ERR(0, 651, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.left))) __PYX_ERR(0, 651, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":650
 *             if isinstance(maths.left, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.left), maths.right)
 *             if isinstance(maths.right, Literal):             # <<<<<<<<<<<<<<
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 *         elif isinstance(right, Divide):
 */
    }

    /* "flitter/language/tree.pyx":646
 *             if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
 *                 return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
 *         elif isinstance(right, Multiply):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal):
 */
    goto __pyx_L5;
  }

  /* "flitter/language/tree.pyx":652
 *             if isinstance(maths.right, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 *         elif isinstance(right, Divide):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal):
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Divide); 
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":653
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 *         elif isinstance(right, Divide):
 *             maths = right             # <<<<<<<<<<<<<<
 *             if isinstance(maths.left, Literal):
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 */
    if (!(likely(((((PyObject *)__pyx_v_right)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation))))) __PYX_ERR(0, 653, __pyx_L1_error)
    __pyx_t_1 = ((PyObject *)__pyx_v_right);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_maths = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":654
 *         elif isinstance(right, Divide):
 *             maths = right
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 *         elif isinstance(right, Negative):
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_maths->__pyx_base.left);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":655
 *             maths = right
 *             if isinstance(maths.left, Literal):
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)             # <<<<<<<<<<<<<<
 *         elif isinstance(right, Negative):
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.left);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.left))) __PYX_ERR(0, 655, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      __Pyx_GIVEREF((PyObject *)__pyx_v_maths->__pyx_base.right);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_maths->__pyx_base.right))) __PYX_ERR(0, 655, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Divide), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":654
 *         elif isinstance(right, Divide):
 *             maths = right
 *             if isinstance(maths.left, Literal):             # <<<<<<<<<<<<<<
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 *         elif isinstance(right, Negative):
 */
    }

    /* "flitter/language/tree.pyx":652
 *             if isinstance(maths.right, Literal):
 *                 return Multiply(Multiply(Literal(left), maths.right), maths.left)
 *         elif isinstance(right, Divide):             # <<<<<<<<<<<<<<
 *             maths = right
 *             if isinstance(maths.left, Literal):
 */
    goto __pyx_L5;
  }

  /* "flitter/language/tree.pyx":656
 *             if isinstance(maths.left, Literal):
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 *         elif isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)
 * 
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Negative); 
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":657
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 *         elif isinstance(right, Negative):
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->neg(__pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6)) __PYX_ERR(0, 657, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v_right)->__pyx_base.expr))) __PYX_ERR(0, 657, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":656
 *             if isinstance(maths.left, Literal):
 *                 return Divide(Multiply(Literal(left), maths.left), maths.right)
 *         elif isinstance(right, Negative):             # <<<<<<<<<<<<<<
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)
 * 
 */
  }
  __pyx_L5:;

  /* "flitter/language/tree.pyx":636
 *         program.mul()
 * 
 *     cdef Expression constant_left(self, Vector left, Expression right):             # <<<<<<<<<<<<<<
 *         if left.eq(true_):
 *             return Positive(right)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.Multiply.constant_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_maths);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":659
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return self.constant_left(right, left)
 * 
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Multiply_constant_right(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":660
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         return self.constant_left(right, left)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Multiply *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.constant_left(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v_self), __pyx_v_right, __pyx_v_left)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":659
 *             return Multiply(Literal(left.neg()), (<Negative>right).expr)
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         return self.constant_left(right, left)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Multiply.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Multiply_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Multiply_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Multiply_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Multiply_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Multiply___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Multiply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Multiply___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Multiply); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Multiply); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Multiply.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Multiply_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Multiply_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Multiply_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Multiply_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Multiply.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Multiply_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Multiply *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Multiply_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Multiply__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Multiply.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":664
 * 
 * cdef class Divide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.truediv(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_6Divide_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":665
 * cdef class Divide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.truediv(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->truediv(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":664
 * 
 * cdef class Divide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.truediv(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Divide.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":667
 *         return left.truediv(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.truediv()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_6Divide__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":668
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.truediv()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->truediv(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":667
 *         return left.truediv(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.truediv()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Divide._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":670
 *         program.truediv()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Positive(left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Divide_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":671
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 *         return Multiply(Literal(true_.truediv(right)), left)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->eq(__pyx_v_right, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":672
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):
 *             return Positive(left)             # <<<<<<<<<<<<<<
 *         return Multiply(Literal(true_.truediv(right)), left)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":671
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 *         return Multiply(Literal(true_.truediv(right)), left)
 */
  }

  /* "flitter/language/tree.pyx":673
 *         if right.eq(true_):
 *             return Positive(left)
 *         return Multiply(Literal(true_.truediv(right)), left)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_7flitter_5model_true_->__pyx_vtab)->truediv(__pyx_v_7flitter_5model_true_, __pyx_v_right)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 673, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":670
 *         program.truediv()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Positive(left)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Divide.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Divide_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Divide_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Divide_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Divide_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Divide___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Divide *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Divide___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Divide); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Divide); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Divide.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Divide_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Divide_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Divide_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Divide_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Divide.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Divide_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Divide *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Divide_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Divide__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Divide.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":677
 * 
 * cdef class FloorDivide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.floordiv(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_11FloorDivide_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":678
 * cdef class FloorDivide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.floordiv(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->floordiv(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":677
 * 
 * cdef class FloorDivide(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.floordiv(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":680
 *         return left.floordiv(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.floordiv()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_11FloorDivide__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":681
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.floordiv()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->floordiv(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":680
 *         return left.floordiv(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.floordiv()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":683
 *         program.floordiv()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Floor(left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_11FloorDivide_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":684
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Floor(left)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->eq(__pyx_v_right, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":685
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):
 *             return Floor(left)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":684
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Floor(left)
 * 
 */
  }

  /* "flitter/language/tree.pyx":683
 *         program.floordiv()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Floor(left)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11FloorDivide___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_FloorDivide *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11FloorDivide___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_FloorDivide); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_FloorDivide); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11FloorDivide_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_FloorDivide *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11FloorDivide_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_FloorDivide__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.FloorDivide.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":689
 * 
 * cdef class Modulo(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.mod(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_6Modulo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":690
 * cdef class Modulo(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.mod(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->mod(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":689
 * 
 * cdef class Modulo(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.mod(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Modulo.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":692
 *         return left.mod(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.mod()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_6Modulo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":693
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.mod()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->mod(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":692
 *         return left.mod(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.mod()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Modulo._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":695
 *         program.mod()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Fract(left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Modulo_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":696
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Fract(left)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->eq(__pyx_v_right, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":697
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):
 *             return Fract(left)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":696
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Fract(left)
 * 
 */
  }

  /* "flitter/language/tree.pyx":695
 *         program.mod()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Fract(left)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Modulo.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Modulo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Modulo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Modulo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Modulo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Modulo___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Modulo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Modulo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Modulo); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Modulo); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Modulo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Modulo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Modulo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Modulo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Modulo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Modulo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Modulo_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Modulo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Modulo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Modulo__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Modulo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":701
 * 
 * cdef class Power(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.pow(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_5Power_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":702
 * cdef class Power(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.pow(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->pow(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":701
 * 
 * cdef class Power(MathsBinaryOperation):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.pow(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Power.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":704
 *         return left.pow(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):
 */

static void __pyx_f_7flitter_8language_4tree_5Power__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instr = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":705
 * 
 *     cdef void _compile_op(self, Program program):
 *         cdef Instruction instr = program.last_instruction()             # <<<<<<<<<<<<<<
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):
 *             program.pop_instruction()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":706
 *     cdef void _compile_op(self, Program program):
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.dup()
 */
  __pyx_t_3 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_Literal);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instr)->value->__pyx_vtab)->eq(((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instr)->value, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":707
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.dup()
 *             program.mul()
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":708
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):
 *             program.pop_instruction()
 *             program.dup()             # <<<<<<<<<<<<<<
 *             program.mul()
 *         else:
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->dup(__pyx_v_program, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":709
 *             program.pop_instruction()
 *             program.dup()
 *             program.mul()             # <<<<<<<<<<<<<<
 *         else:
 *             program.pow()
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->mul(__pyx_v_program, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":706
 *     cdef void _compile_op(self, Program program):
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.dup()
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":711
 *             program.mul()
 *         else:
 *             program.pow()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 */
  /*else*/ {
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pow(__pyx_v_program, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":704
 *         return left.pow(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Power._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":713
 *             program.pow()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Positive(left)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Power_constant_right(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("constant_right", 1);

  /* "flitter/language/tree.pyx":714
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_true_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_right->__pyx_vtab)->eq(__pyx_v_right, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":715
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):
 *             return Positive(left)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), ((PyObject *)__pyx_v_left)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":714
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):
 *         if right.eq(true_):             # <<<<<<<<<<<<<<
 *             return Positive(left)
 * 
 */
  }

  /* "flitter/language/tree.pyx":713
 *             program.pow()
 * 
 *     cdef Expression constant_right(self, Expression left, Vector right):             # <<<<<<<<<<<<<<
 *         if right.eq(true_):
 *             return Positive(left)
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Power.constant_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Power_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Power_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Power_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Power_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Power___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Power *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Power___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Power); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Power); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Power.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Power_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Power_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Power_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Power_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Power.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Power_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Power *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Power_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Power__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Power.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Comparison_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Comparison_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Comparison_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Comparison_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Comparison___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Comparison *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Comparison___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Comparison *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Comparison); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Comparison); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Comparison.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Comparison_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Comparison_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Comparison_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Comparison_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Comparison.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Comparison_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Comparison *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Comparison_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Comparison *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Comparison__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Comparison.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":723
 * 
 * cdef class EqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.eq(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_7EqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":724
 * cdef class EqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.eq(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->eq(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":723
 * 
 * cdef class EqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.eq(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.EqualTo.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":726
 *         return left.eq(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.eq()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_7EqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":727
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.eq()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->eq(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":726
 *         return left.eq(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.eq()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.EqualTo._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7EqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7EqualTo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7EqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7EqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7EqualTo___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_EqualTo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7EqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_EqualTo); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_EqualTo); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.EqualTo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7EqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7EqualTo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7EqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7EqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.EqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7EqualTo_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_EqualTo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7EqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_EqualTo__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.EqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":731
 * 
 * cdef class NotEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.ne(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_10NotEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":732
 * cdef class NotEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ne(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->ne(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":731
 * 
 * cdef class NotEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.ne(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.NotEqualTo.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":734
 *         return left.ne(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ne()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_10NotEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":735
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.ne()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->ne(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":734
 *         return left.ne(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ne()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.NotEqualTo._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10NotEqualTo___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10NotEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_NotEqualTo); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_NotEqualTo); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.NotEqualTo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.NotEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10NotEqualTo_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10NotEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_NotEqualTo__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.NotEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":739
 * 
 * cdef class LessThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.lt(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_8LessThan_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":740
 * cdef class LessThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.lt(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->lt(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":739
 * 
 * cdef class LessThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.lt(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThan.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":742
 *         return left.lt(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.lt()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_8LessThan__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":743
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.lt()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->lt(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":742
 *         return left.lt(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.lt()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThan._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8LessThan_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8LessThan_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8LessThan_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8LessThan_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8LessThan___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_LessThan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8LessThan___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_LessThan); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_LessThan); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.LessThan.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8LessThan_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8LessThan_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8LessThan_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8LessThan_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.LessThan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8LessThan_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_LessThan *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8LessThan_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThan__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":747
 * 
 * cdef class GreaterThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.gt(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_11GreaterThan_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":748
 * cdef class GreaterThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.gt(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->gt(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":747
 * 
 * cdef class GreaterThan(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.gt(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThan.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":750
 *         return left.gt(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.gt()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_11GreaterThan__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":751
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.gt()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->gt(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":750
 *         return left.gt(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.gt()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThan._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11GreaterThan___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_GreaterThan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11GreaterThan___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_GreaterThan); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_GreaterThan); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThan.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.GreaterThan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11GreaterThan_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_GreaterThan *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11GreaterThan_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThan__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":755
 * 
 * cdef class LessThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.le(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":756
 * cdef class LessThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.le(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->le(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":755
 * 
 * cdef class LessThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.le(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThanOrEqualTo.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":758
 *         return left.le(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.le()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":759
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.le()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->le(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":758
 *         return left.le(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.le()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThanOrEqualTo._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_LessThanOrEqualTo); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_LessThanOrEqualTo); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.LessThanOrEqualTo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.LessThanOrEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_17LessThanOrEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThanOrEqualTo__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.LessThanOrEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":763
 * 
 * cdef class GreaterThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.ge(right)
 * 
 */

static struct __pyx_obj_7flitter_5model_Vector *__pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_left, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_right) {
  struct __pyx_obj_7flitter_5model_Vector *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op", 1);

  /* "flitter/language/tree.pyx":764
 * cdef class GreaterThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):
 *         return left.ge(right)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile_op(self, Program program):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_left->__pyx_vtab)->ge(__pyx_v_left, __pyx_v_right)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":763
 * 
 * cdef class GreaterThanOrEqualTo(Comparison):
 *     cdef Vector op(self, Vector left, Vector right):             # <<<<<<<<<<<<<<
 *         return left.ge(right)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThanOrEqualTo.op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":766
 *         return left.ge(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ge()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":767
 * 
 *     cdef void _compile_op(self, Program program):
 *         program.ge()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->ge(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":766
 *         return left.ge(right)
 * 
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.ge()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThanOrEqualTo._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_GreaterThanOrEqua); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_GreaterThanOrEqua); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThanOrEqualTo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.GreaterThanOrEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_20GreaterThanOrEqualTo_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThanOrEqualTo__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.GreaterThanOrEqualTo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":771
 * 
 * cdef class And(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 */

static void __pyx_f_7flitter_8language_4tree_3And__compile(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  int64_t __pyx_v_end_label;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":772
 * cdef class And(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()             # <<<<<<<<<<<<<<
 *         self.left._compile(program, lnames)
 *         program.dup()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 772, __pyx_L1_error)
  __pyx_v_end_label = __pyx_t_1;

  /* "flitter/language/tree.pyx":773
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.dup()
 *         program.branch_false(end_label)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.left->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.left, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 773, __pyx_L1_error)

  /* "flitter/language/tree.pyx":774
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 *         program.dup()             # <<<<<<<<<<<<<<
 *         program.branch_false(end_label)
 *         program.drop()
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->dup(__pyx_v_program, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":775
 *         self.left._compile(program, lnames)
 *         program.dup()
 *         program.branch_false(end_label)             # <<<<<<<<<<<<<<
 *         program.drop()
 *         self.right._compile(program, lnames)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->branch_false(__pyx_v_program, __pyx_v_end_label, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":776
 *         program.dup()
 *         program.branch_false(end_label)
 *         program.drop()             # <<<<<<<<<<<<<<
 *         self.right._compile(program, lnames)
 *         program.label(end_label)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->drop(__pyx_v_program, 0, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":777
 *         program.branch_false(end_label)
 *         program.drop()
 *         self.right._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.label(end_label)
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.right, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L1_error)

  /* "flitter/language/tree.pyx":778
 *         program.drop()
 *         self.right._compile(program, lnames)
 *         program.label(end_label)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_end_label, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":771
 * 
 * cdef class And(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.And._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":780
 *         program.label(end_label)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3And__simplify(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":781
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(left, Literal):
 *             if (<Literal>left).value.as_bool():
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.left->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.left, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":782
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):             # <<<<<<<<<<<<<<
 *             if (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":783
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 *             if (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *                 return self.right._simplify(context)
 *             return left
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":784
 *         if isinstance(left, Literal):
 *             if (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)             # <<<<<<<<<<<<<<
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":783
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 *             if (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *                 return self.right._simplify(context)
 *             return left
 */
    }

    /* "flitter/language/tree.pyx":785
 *             if (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 *             return left             # <<<<<<<<<<<<<<
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __pyx_r = __pyx_v_left;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":782
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):             # <<<<<<<<<<<<<<
 *             if (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":786
 *                 return self.right._simplify(context)
 *             return left
 *         cdef Expression right = self.right._simplify(context)             # <<<<<<<<<<<<<<
 *         if left is self.left and right is self.right:
 *             return self
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":787
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return And(left, right)
 */
  __pyx_t_3 = (__pyx_v_left == __pyx_v_self->__pyx_base.left);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_right == __pyx_v_self->__pyx_base.right);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":788
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:
 *             return self             # <<<<<<<<<<<<<<
 *         return And(left, right)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":787
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return And(left, right)
 */
  }

  /* "flitter/language/tree.pyx":789
 *         if left is self.left and right is self.right:
 *             return self
 *         return And(left, right)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 789, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_right))) __PYX_ERR(0, 789, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_And), __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":780
 *         program.label(end_label)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.And._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3And_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3And_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3And_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3And_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3And___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_And *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3And___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_And); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_And__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_And); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.And.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_And__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3And_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3And_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3And_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3And_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.And.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3And_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_And *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3And_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_And__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_And__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_And__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.And.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":793
 * 
 * cdef class Or(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 */

static void __pyx_f_7flitter_8language_4tree_2Or__compile(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  int64_t __pyx_v_end_label;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":794
 * cdef class Or(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()             # <<<<<<<<<<<<<<
 *         self.left._compile(program, lnames)
 *         program.dup()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L1_error)
  __pyx_v_end_label = __pyx_t_1;

  /* "flitter/language/tree.pyx":795
 *     cdef void _compile(self, Program program, list lnames):
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.dup()
 *         program.branch_true(end_label)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.left->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.left, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "flitter/language/tree.pyx":796
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 *         program.dup()             # <<<<<<<<<<<<<<
 *         program.branch_true(end_label)
 *         program.drop()
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->dup(__pyx_v_program, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":797
 *         self.left._compile(program, lnames)
 *         program.dup()
 *         program.branch_true(end_label)             # <<<<<<<<<<<<<<
 *         program.drop()
 *         self.right._compile(program, lnames)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->branch_true(__pyx_v_program, __pyx_v_end_label, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":798
 *         program.dup()
 *         program.branch_true(end_label)
 *         program.drop()             # <<<<<<<<<<<<<<
 *         self.right._compile(program, lnames)
 *         program.label(end_label)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->drop(__pyx_v_program, 0, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":799
 *         program.branch_true(end_label)
 *         program.drop()
 *         self.right._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.label(end_label)
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.right, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L1_error)

  /* "flitter/language/tree.pyx":800
 *         program.drop()
 *         self.right._compile(program, lnames)
 *         program.label(end_label)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_end_label, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":793
 * 
 * cdef class Or(BinaryOperation):
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         end_label = program.new_label()
 *         self.left._compile(program, lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Or._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":802
 *         program.label(end_label)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_2Or__simplify(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":803
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)             # <<<<<<<<<<<<<<
 *         if isinstance(left, Literal):
 *             if not (<Literal>left).value.as_bool():
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.left->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.left, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":804
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):             # <<<<<<<<<<<<<<
 *             if not (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":805
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 *             if not (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *                 return self.right._simplify(context)
 *             return left
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 805, __pyx_L1_error)
    __pyx_t_3 = (!__pyx_t_2);
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":806
 *         if isinstance(left, Literal):
 *             if not (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)             # <<<<<<<<<<<<<<
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":805
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 *             if not (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *                 return self.right._simplify(context)
 *             return left
 */
    }

    /* "flitter/language/tree.pyx":807
 *             if not (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 *             return left             # <<<<<<<<<<<<<<
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __pyx_r = __pyx_v_left;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":804
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):             # <<<<<<<<<<<<<<
 *             if not (<Literal>left).value.as_bool():
 *                 return self.right._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":808
 *                 return self.right._simplify(context)
 *             return left
 *         cdef Expression right = self.right._simplify(context)             # <<<<<<<<<<<<<<
 *         if left is self.left and right is self.right:
 *             return self
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":809
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return Or(left, right)
 */
  __pyx_t_2 = (__pyx_v_left == __pyx_v_self->__pyx_base.left);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_right == __pyx_v_self->__pyx_base.right);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":810
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:
 *             return self             # <<<<<<<<<<<<<<
 *         return Or(left, right)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":809
 *             return left
 *         cdef Expression right = self.right._simplify(context)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return Or(left, right)
 */
  }

  /* "flitter/language/tree.pyx":811
 *         if left is self.left and right is self.right:
 *             return self
 *         return Or(left, right)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 811, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_right))) __PYX_ERR(0, 811, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Or), __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":802
 *         program.label(end_label)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         if isinstance(left, Literal):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Or._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_2Or_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_2Or_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_2Or_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_2Or_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_2Or___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Or *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_2Or___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Or); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Or); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Or.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_2Or_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_2Or_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_2Or_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_2Or_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Or.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_2Or_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Or *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_2Or_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Or__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Or.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":815
 * 
 * cdef class Xor(BinaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.xor()
 * 
 */

static void __pyx_f_7flitter_8language_4tree_3Xor__compile_op(CYTHON_UNUSED struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile_op", 1);

  /* "flitter/language/tree.pyx":816
 * cdef class Xor(BinaryOperation):
 *     cdef void _compile_op(self, Program program):
 *         program.xor()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->xor(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":815
 * 
 * cdef class Xor(BinaryOperation):
 *     cdef void _compile_op(self, Program program):             # <<<<<<<<<<<<<<
 *         program.xor()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Xor._compile_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":818
 *         program.xor()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Xor__simplify(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_left = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_right = 0;
  int __pyx_v_literal_left;
  int __pyx_v_literal_right;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":819
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.left->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.left, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":820
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.right->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.right, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":821
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)             # <<<<<<<<<<<<<<
 *         cdef bint literal_right = isinstance(right, Literal)
 *         if literal_left and not (<Literal>left).value.as_bool():
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_left), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_v_literal_left = __pyx_t_2;

  /* "flitter/language/tree.pyx":822
 *         cdef Expression right = self.right._simplify(context)
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)             # <<<<<<<<<<<<<<
 *         if literal_left and not (<Literal>left).value.as_bool():
 *             return right
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_right), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_v_literal_right = __pyx_t_2;

  /* "flitter/language/tree.pyx":823
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)
 *         if literal_left and not (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *             return right
 *         if literal_right and not (<Literal>right).value.as_bool():
 */
  if (__pyx_v_literal_left) {
  } else {
    __pyx_t_2 = __pyx_v_literal_left;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_left)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":824
 *         cdef bint literal_right = isinstance(right, Literal)
 *         if literal_left and not (<Literal>left).value.as_bool():
 *             return right             # <<<<<<<<<<<<<<
 *         if literal_right and not (<Literal>right).value.as_bool():
 *             return left
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_right);
    __pyx_r = __pyx_v_right;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":823
 *         cdef bint literal_left = isinstance(left, Literal)
 *         cdef bint literal_right = isinstance(right, Literal)
 *         if literal_left and not (<Literal>left).value.as_bool():             # <<<<<<<<<<<<<<
 *             return right
 *         if literal_right and not (<Literal>right).value.as_bool():
 */
  }

  /* "flitter/language/tree.pyx":825
 *         if literal_left and not (<Literal>left).value.as_bool():
 *             return right
 *         if literal_right and not (<Literal>right).value.as_bool():             # <<<<<<<<<<<<<<
 *             return left
 *         if literal_left and literal_right:
 */
  if (__pyx_v_literal_right) {
  } else {
    __pyx_t_2 = __pyx_v_literal_right;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_right)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_right)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 825, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_4);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":826
 *             return right
 *         if literal_right and not (<Literal>right).value.as_bool():
 *             return left             # <<<<<<<<<<<<<<
 *         if literal_left and literal_right:
 *             return Literal(false_)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_left);
    __pyx_r = __pyx_v_left;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":825
 *         if literal_left and not (<Literal>left).value.as_bool():
 *             return right
 *         if literal_right and not (<Literal>right).value.as_bool():             # <<<<<<<<<<<<<<
 *             return left
 *         if literal_left and literal_right:
 */
  }

  /* "flitter/language/tree.pyx":827
 *         if literal_right and not (<Literal>right).value.as_bool():
 *             return left
 *         if literal_left and literal_right:             # <<<<<<<<<<<<<<
 *             return Literal(false_)
 *         if left is self.left and right is self.right:
 */
  if (__pyx_v_literal_left) {
  } else {
    __pyx_t_2 = __pyx_v_literal_left;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_literal_right;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":828
 *             return left
 *         if literal_left and literal_right:
 *             return Literal(false_)             # <<<<<<<<<<<<<<
 *         if left is self.left and right is self.right:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_7flitter_5model_false_)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":827
 *         if literal_right and not (<Literal>right).value.as_bool():
 *             return left
 *         if literal_left and literal_right:             # <<<<<<<<<<<<<<
 *             return Literal(false_)
 *         if left is self.left and right is self.right:
 */
  }

  /* "flitter/language/tree.pyx":829
 *         if literal_left and literal_right:
 *             return Literal(false_)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return Xor(left, right)
 */
  __pyx_t_3 = (__pyx_v_left == __pyx_v_self->__pyx_base.left);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_right == __pyx_v_self->__pyx_base.right);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":830
 *             return Literal(false_)
 *         if left is self.left and right is self.right:
 *             return self             # <<<<<<<<<<<<<<
 *         return Xor(left, right)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":829
 *         if literal_left and literal_right:
 *             return Literal(false_)
 *         if left is self.left and right is self.right:             # <<<<<<<<<<<<<<
 *             return self
 *         return Xor(left, right)
 */
  }

  /* "flitter/language/tree.pyx":831
 *         if left is self.left and right is self.right:
 *             return self
 *         return Xor(left, right)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_left))) __PYX_ERR(0, 831, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_right))) __PYX_ERR(0, 831, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Xor), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":818
 *         program.xor()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression left = self.left._simplify(context)
 *         cdef Expression right = self.right._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Xor._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Xor_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Xor_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Xor_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Xor_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Xor___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Xor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Xor___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.left, self.right)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.left);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.left);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.left))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.right);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.right);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.right))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.left, self.right)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left is not None or self.right is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.left) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.right) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Xor); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left is not None or self.right is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, None), state
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Xor); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_20102196);
    __Pyx_GIVEREF(__pyx_int_20102196);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_20102196)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Xor.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Xor_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Xor_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Xor_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Xor_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Xor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Xor_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Xor *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Xor_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Xor__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Xor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":838
 *     cdef readonly Expression index
 * 
 *     def __init__(self, Expression expr, Expression index):             # <<<<<<<<<<<<<<
 *         self.expr = expr
 *         self.index = index
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_5Slice_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_5Slice_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_index = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_expr,&__pyx_n_s_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expr)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 838, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 838, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_index = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 838, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Slice.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "expr", 0))) __PYX_ERR(0, 838, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "index", 0))) __PYX_ERR(0, 838, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice___init__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self), __pyx_v_expr, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_5Slice___init__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_index) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":839
 * 
 *     def __init__(self, Expression expr, Expression index):
 *         self.expr = expr             # <<<<<<<<<<<<<<
 *         self.index = index
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->expr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->expr);
  __pyx_v_self->expr = __pyx_v_expr;

  /* "flitter/language/tree.pyx":840
 *     def __init__(self, Expression expr, Expression index):
 *         self.expr = expr
 *         self.index = index             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_index);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->index);
  __Pyx_DECREF((PyObject *)__pyx_v_self->index);
  __pyx_v_self->index = __pyx_v_index;

  /* "flitter/language/tree.pyx":838
 *     cdef readonly Expression index
 * 
 *     def __init__(self, Expression expr, Expression index):             # <<<<<<<<<<<<<<
 *         self.expr = expr
 *         self.index = index
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":842
 *         self.index = index
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.expr._compile(program, lnames)
 *         if isinstance(self.index, Literal):
 */

static void __pyx_f_7flitter_8language_4tree_5Slice__compile(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":843
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         if isinstance(self.index, Literal):
 *             program.slice_literal((<Literal>self.index).value)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->expr->__pyx_vtab)->_compile(__pyx_v_self->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L1_error)

  /* "flitter/language/tree.pyx":844
 *     cdef void _compile(self, Program program, list lnames):
 *         self.expr._compile(program, lnames)
 *         if isinstance(self.index, Literal):             # <<<<<<<<<<<<<<
 *             program.slice_literal((<Literal>self.index).value)
 *         else:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->index);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":845
 *         self.expr._compile(program, lnames)
 *         if isinstance(self.index, Literal):
 *             program.slice_literal((<Literal>self.index).value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.index._compile(program, lnames)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->index)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->slice_literal(__pyx_v_program, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":844
 *     cdef void _compile(self, Program program, list lnames):
 *         self.expr._compile(program, lnames)
 *         if isinstance(self.index, Literal):             # <<<<<<<<<<<<<<
 *             program.slice_literal((<Literal>self.index).value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":847
 *             program.slice_literal((<Literal>self.index).value)
 *         else:
 *             self.index._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.slice()
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->index->__pyx_vtab)->_compile(__pyx_v_self->index, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L1_error)

    /* "flitter/language/tree.pyx":848
 *         else:
 *             self.index._compile(program, lnames)
 *             program.slice()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->slice(__pyx_v_program, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":842
 *         self.index = index
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.expr._compile(program, lnames)
 *         if isinstance(self.index, Literal):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Slice._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":850
 *             program.slice()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         cdef Expression index = self.index._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_5Slice__simplify(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_index = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_expr_value = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_index_value = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":851
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression index = self.index._simplify(context)
 *         cdef Vector expr_value
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->expr->__pyx_vtab)->_simplify(__pyx_v_self->expr, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":852
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression expr = self.expr._simplify(context)
 *         cdef Expression index = self.index._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Vector expr_value
 *         cdef Vector index_value
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->index->__pyx_vtab)->_simplify(__pyx_v_self->index, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_index = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":855
 *         cdef Vector expr_value
 *         cdef Vector index_value
 *         if isinstance(expr, Literal) and isinstance(index, Literal):             # <<<<<<<<<<<<<<
 *             expr_value = (<Literal>expr).value
 *             index_value = (<Literal>index).value
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_index), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":856
 *         cdef Vector index_value
 *         if isinstance(expr, Literal) and isinstance(index, Literal):
 *             expr_value = (<Literal>expr).value             # <<<<<<<<<<<<<<
 *             index_value = (<Literal>index).value
 *             return Literal(expr_value.slice(index_value))
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_expr_value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":857
 *         if isinstance(expr, Literal) and isinstance(index, Literal):
 *             expr_value = (<Literal>expr).value
 *             index_value = (<Literal>index).value             # <<<<<<<<<<<<<<
 *             return Literal(expr_value.slice(index_value))
 *         if expr is self.expr and index is self.index:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_index)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_index_value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":858
 *             expr_value = (<Literal>expr).value
 *             index_value = (<Literal>index).value
 *             return Literal(expr_value.slice(index_value))             # <<<<<<<<<<<<<<
 *         if expr is self.expr and index is self.index:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_expr_value->__pyx_vtab)->slice(__pyx_v_expr_value, __pyx_v_index_value, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":855
 *         cdef Vector expr_value
 *         cdef Vector index_value
 *         if isinstance(expr, Literal) and isinstance(index, Literal):             # <<<<<<<<<<<<<<
 *             expr_value = (<Literal>expr).value
 *             index_value = (<Literal>index).value
 */
  }

  /* "flitter/language/tree.pyx":859
 *             index_value = (<Literal>index).value
 *             return Literal(expr_value.slice(index_value))
 *         if expr is self.expr and index is self.index:             # <<<<<<<<<<<<<<
 *             return self
 *         return Slice(expr, index)
 */
  __pyx_t_3 = (__pyx_v_expr == __pyx_v_self->expr);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_index == __pyx_v_self->index);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":860
 *             return Literal(expr_value.slice(index_value))
 *         if expr is self.expr and index is self.index:
 *             return self             # <<<<<<<<<<<<<<
 *         return Slice(expr, index)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":859
 *             index_value = (<Literal>index).value
 *             return Literal(expr_value.slice(index_value))
 *         if expr is self.expr and index is self.index:             # <<<<<<<<<<<<<<
 *             return self
 *         return Slice(expr, index)
 */
  }

  /* "flitter/language/tree.pyx":861
 *         if expr is self.expr and index is self.index:
 *             return self
 *         return Slice(expr, index)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 861, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_index))) __PYX_ERR(0, 861, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Slice), __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":850
 *             program.slice()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression expr = self.expr._simplify(context)
 *         cdef Expression index = self.index._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Slice._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_index);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_index_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":863
 *         return Slice(expr, index)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Slice({self.expr!r}, {self.index!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":864
 * 
 *     def __repr__(self):
 *         return f'Slice({self.expr!r}, {self.index!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Slice);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_Slice);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Slice);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->expr)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->index)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":863
 *         return Slice(expr, index)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Slice({self.expr!r}, {self.index!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Slice.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":835
 * 
 * cdef class Slice(Expression):
 *     cdef readonly Expression expr             # <<<<<<<<<<<<<<
 *     cdef readonly Expression index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_4expr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_4expr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice_4expr___get__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __pyx_r = ((PyObject *)__pyx_v_self->expr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":836
 * cdef class Slice(Expression):
 *     cdef readonly Expression expr
 *     cdef readonly Expression index             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression expr, Expression index):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_5index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice_5index___get__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_5index___get__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->index);
  __pyx_r = ((PyObject *)__pyx_v_self->index);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Slice_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Slice_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr, self.index)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->index))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr, self.index)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr, self.index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None or self.index is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr, self.index)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None or self.index is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->expr) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->index) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.index is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None or self.index is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Slice); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_61976268);
    __Pyx_GIVEREF(__pyx_int_61976268);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_61976268)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.index is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, None), state
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Slice); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_61976268);
    __Pyx_GIVEREF(__pyx_int_61976268);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_61976268)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Slice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5Slice_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Slice_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5Slice_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Slice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_5Slice_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_5Slice_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Slice__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Slice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":872
 *     cdef readonly tuple keyword_args
 * 
 *     def __init__(self, Expression function, tuple args, tuple keyword_args=None):             # <<<<<<<<<<<<<<
 *         self.function = function
 *         self.args = args
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_4Call_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_4Call_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_function = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_keyword_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_function,&__pyx_n_s_args,&__pyx_n_s_keyword_args,0};
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_function)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_keyword_args);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 872, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 872, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_function = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_args = ((PyObject*)values[1]);
    __pyx_v_keyword_args = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 872, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Call.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_function), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "function", 0))) __PYX_ERR(0, 872, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyTuple_Type), 1, "args", 1))) __PYX_ERR(0, 872, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_keyword_args), (&PyTuple_Type), 1, "keyword_args", 1))) __PYX_ERR(0, 872, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call___init__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self), __pyx_v_function, __pyx_v_args, __pyx_v_keyword_args);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_4Call___init__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_function, PyObject *__pyx_v_args, PyObject *__pyx_v_keyword_args) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":873
 * 
 *     def __init__(self, Expression function, tuple args, tuple keyword_args=None):
 *         self.function = function             # <<<<<<<<<<<<<<
 *         self.args = args
 *         self.keyword_args = keyword_args
 */
  __Pyx_INCREF((PyObject *)__pyx_v_function);
  __Pyx_GIVEREF((PyObject *)__pyx_v_function);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->function);
  __Pyx_DECREF((PyObject *)__pyx_v_self->function);
  __pyx_v_self->function = __pyx_v_function;

  /* "flitter/language/tree.pyx":874
 *     def __init__(self, Expression function, tuple args, tuple keyword_args=None):
 *         self.function = function
 *         self.args = args             # <<<<<<<<<<<<<<
 *         self.keyword_args = keyword_args
 * 
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = __pyx_v_args;

  /* "flitter/language/tree.pyx":875
 *         self.function = function
 *         self.args = args
 *         self.keyword_args = keyword_args             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_keyword_args);
  __Pyx_GIVEREF(__pyx_v_keyword_args);
  __Pyx_GOTREF(__pyx_v_self->keyword_args);
  __Pyx_DECREF(__pyx_v_self->keyword_args);
  __pyx_v_self->keyword_args = __pyx_v_keyword_args;

  /* "flitter/language/tree.pyx":872
 *     cdef readonly tuple keyword_args
 * 
 *     def __init__(self, Expression function, tuple args, tuple keyword_args=None):             # <<<<<<<<<<<<<<
 *         self.function = function
 *         self.args = args
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":877
 *         self.keyword_args = keyword_args
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Expression expr
 *         cdef list names = []
 */

static void __pyx_f_7flitter_8language_4tree_4Call__compile(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  PyObject *__pyx_v_names = 0;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_keyword_arg = 0;
  PyObject *__pyx_v_function = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_call __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":879
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef Expression expr
 *         cdef list names = []             # <<<<<<<<<<<<<<
 *         if self.args:
 *             for expr in self.args:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":880
 *         cdef Expression expr
 *         cdef list names = []
 *         if self.args:             # <<<<<<<<<<<<<<
 *             for expr in self.args:
 *                 expr._compile(program, lnames)
 */
  __pyx_t_2 = (__pyx_v_self->args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->args) != 0);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":881
 *         cdef list names = []
 *         if self.args:
 *             for expr in self.args:             # <<<<<<<<<<<<<<
 *                 expr._compile(program, lnames)
 *         cdef Binding keyword_arg
 */
    if (unlikely(__pyx_v_self->args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 881, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->args; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 881, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 881, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":882
 *         if self.args:
 *             for expr in self.args:
 *                 expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef Binding keyword_arg
 *         if self.keyword_args:
 */
      ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_expr->__pyx_vtab)->_compile(__pyx_v_expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L1_error)

      /* "flitter/language/tree.pyx":881
 *         cdef list names = []
 *         if self.args:
 *             for expr in self.args:             # <<<<<<<<<<<<<<
 *                 expr._compile(program, lnames)
 *         cdef Binding keyword_arg
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":880
 *         cdef Expression expr
 *         cdef list names = []
 *         if self.args:             # <<<<<<<<<<<<<<
 *             for expr in self.args:
 *                 expr._compile(program, lnames)
 */
  }

  /* "flitter/language/tree.pyx":884
 *                 expr._compile(program, lnames)
 *         cdef Binding keyword_arg
 *         if self.keyword_args:             # <<<<<<<<<<<<<<
 *             for keyword_arg in self.keyword_args:
 *                 names.append(keyword_arg.name)
 */
  __pyx_t_2 = (__pyx_v_self->keyword_args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->keyword_args) != 0);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":885
 *         cdef Binding keyword_arg
 *         if self.keyword_args:
 *             for keyword_arg in self.keyword_args:             # <<<<<<<<<<<<<<
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 */
    if (unlikely(__pyx_v_self->keyword_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 885, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->keyword_args; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 885, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 885, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 885, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_keyword_arg, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":886
 *         if self.keyword_args:
 *             for keyword_arg in self.keyword_args:
 *                 names.append(keyword_arg.name)             # <<<<<<<<<<<<<<
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \
 */
      __pyx_t_4 = __pyx_v_keyword_arg->name;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_names, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":887
 *             for keyword_arg in self.keyword_args:
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         if not names and isinstance(self.function, Literal) \
 *                 and (<Literal>self.function).value.length == 1 \
 */
      ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_keyword_arg->expr->__pyx_vtab)->_compile(__pyx_v_keyword_arg->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L1_error)

      /* "flitter/language/tree.pyx":885
 *         cdef Binding keyword_arg
 *         if self.keyword_args:
 *             for keyword_arg in self.keyword_args:             # <<<<<<<<<<<<<<
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":884
 *                 expr._compile(program, lnames)
 *         cdef Binding keyword_arg
 *         if self.keyword_args:             # <<<<<<<<<<<<<<
 *             for keyword_arg in self.keyword_args:
 *                 names.append(keyword_arg.name)
 */
  }

  /* "flitter/language/tree.pyx":888
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \
 */
  __pyx_t_6 = (PyList_GET_SIZE(__pyx_v_names) != 0);
  __pyx_t_7 = (!__pyx_t_6);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L12_bool_binop_done;
  }

  /* "flitter/language/tree.pyx":889
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \
 *                 and (<Literal>self.function).value.length == 1 \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.objects is not None \
 *                 and not isinstance(function := (<Literal>self.function).value.objects[0], Function):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->function);
  __Pyx_INCREF(__pyx_t_1);

  /* "flitter/language/tree.pyx":888
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \
 */
  __pyx_t_7 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L12_bool_binop_done;
  }

  /* "flitter/language/tree.pyx":889
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \
 *                 and (<Literal>self.function).value.length == 1 \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.objects is not None \
 *                 and not isinstance(function := (<Literal>self.function).value.objects[0], Function):
 */
  __pyx_t_7 = (((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->function)->value->length == 1);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L12_bool_binop_done;
  }

  /* "flitter/language/tree.pyx":890
 *         if not names and isinstance(self.function, Literal) \
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \             # <<<<<<<<<<<<<<
 *                 and not isinstance(function := (<Literal>self.function).value.objects[0], Function):
 *             program.call_fast(function, len(self.args) if self.args else 0)
 */
  __pyx_t_7 = (((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->function)->value->objects != ((PyObject*)Py_None));
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L12_bool_binop_done;
  }

  /* "flitter/language/tree.pyx":891
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \
 *                 and not isinstance(function := (<Literal>self.function).value.objects[0], Function):             # <<<<<<<<<<<<<<
 *             program.call_fast(function, len(self.args) if self.args else 0)
 *         else:
 */
  if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->function)->value->objects == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 891, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_self->function)->value->objects, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_function = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_TypeCheck(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Function); 
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = (!__pyx_t_7);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L12_bool_binop_done:;

  /* "flitter/language/tree.pyx":888
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \
 */
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":892
 *                 and (<Literal>self.function).value.objects is not None \
 *                 and not isinstance(function := (<Literal>self.function).value.objects[0], Function):
 *             program.call_fast(function, len(self.args) if self.args else 0)             # <<<<<<<<<<<<<<
 *         else:
 *             self.function._compile(program, lnames)
 */
    __pyx_t_2 = (__pyx_v_self->args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->args) != 0);
    if (__pyx_t_2) {
      __pyx_t_4 = __pyx_v_self->args;
      __Pyx_INCREF(__pyx_t_4);
      if (unlikely(__pyx_t_4 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 892, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PyTuple_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_8;
    } else {
      __pyx_t_3 = 0;
    }
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->call_fast(__pyx_v_program, __pyx_v_function, __pyx_t_3, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":888
 *                 names.append(keyword_arg.name)
 *                 keyword_arg.expr._compile(program, lnames)
 *         if not names and isinstance(self.function, Literal) \             # <<<<<<<<<<<<<<
 *                 and (<Literal>self.function).value.length == 1 \
 *                 and (<Literal>self.function).value.objects is not None \
 */
    goto __pyx_L11;
  }

  /* "flitter/language/tree.pyx":894
 *             program.call_fast(function, len(self.args) if self.args else 0)
 *         else:
 *             self.function._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.call(len(self.args) if self.args else 0, tuple(names) if names else None)
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->function->__pyx_vtab)->_compile(__pyx_v_self->function, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L1_error)

    /* "flitter/language/tree.pyx":895
 *         else:
 *             self.function._compile(program, lnames)
 *             program.call(len(self.args) if self.args else 0, tuple(names) if names else None)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
    __pyx_t_2 = (__pyx_v_self->args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->args) != 0);
    if (__pyx_t_2) {
      __pyx_t_4 = __pyx_v_self->args;
      __Pyx_INCREF(__pyx_t_4);
      if (unlikely(__pyx_t_4 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 895, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PyTuple_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_8;
    } else {
      __pyx_t_3 = 0;
    }
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_names) != 0);
    if (__pyx_t_2) {
      __pyx_t_1 = PyList_AsTuple(__pyx_v_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1;
      __pyx_t_1 = 0;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_4 = Py_None;
    }
    if (!(likely(PyTuple_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_4))) __PYX_ERR(0, 895, __pyx_L1_error)
    __pyx_t_9.__pyx_n = 1;
    __pyx_t_9.names = ((PyObject*)__pyx_t_4);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->call(__pyx_v_program, __pyx_t_3, 0, &__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L11:;

  /* "flitter/language/tree.pyx":877
 *         self.keyword_args = keyword_args
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef Expression expr
 *         cdef list names = []
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Call._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF((PyObject *)__pyx_v_keyword_arg);
  __Pyx_XDECREF(__pyx_v_function);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":897
 *             program.call(len(self.args) if self.args else 0, tuple(names) if names else None)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression function = self.function._simplify(context)
 *         cdef bint touched = function is not self.function
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_4Call__simplify(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_function = 0;
  int __pyx_v_touched;
  struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_func_expr = 0;
  PyObject *__pyx_v_value = NULL;
  int __pyx_v_literal_func;
  int __pyx_v_literal_args;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_arg = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_sarg = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_keyword_args = 0;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_binding = 0;
  PyObject *__pyx_v_bindings = 0;
  PyObject *__pyx_v_kwargs = 0;
  int64_t __pyx_v_i;
  PyObject *__pyx_v_vector_args = 0;
  PyObject *__pyx_v_results = 0;
  PyObject *__pyx_v_func = NULL;
  PyObject *__pyx_v_exc = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_8genexpr2__pyx_v_binding = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_8genexpr3__pyx_v_literal_arg = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_8genexpr4__pyx_v_binding = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int64_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_UCS4 __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  char const *__pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":898
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression function = self.function._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef bint touched = function is not self.function
 *         cdef Function func_expr = None
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->function->__pyx_vtab)->_simplify(__pyx_v_self->function, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_function = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":899
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression function = self.function._simplify(context)
 *         cdef bint touched = function is not self.function             # <<<<<<<<<<<<<<
 *         cdef Function func_expr = None
 *         if isinstance(function, Function):
 */
  __pyx_t_2 = (__pyx_v_function != __pyx_v_self->function);
  __pyx_v_touched = __pyx_t_2;

  /* "flitter/language/tree.pyx":900
 *         cdef Expression function = self.function._simplify(context)
 *         cdef bint touched = function is not self.function
 *         cdef Function func_expr = None             # <<<<<<<<<<<<<<
 *         if isinstance(function, Function):
 *             func_expr = <Function>function
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_func_expr = ((struct __pyx_obj_7flitter_8language_4tree_Function *)Py_None);

  /* "flitter/language/tree.pyx":901
 *         cdef bint touched = function is not self.function
 *         cdef Function func_expr = None
 *         if isinstance(function, Function):             # <<<<<<<<<<<<<<
 *             func_expr = <Function>function
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_function), __pyx_ptype_7flitter_8language_4tree_Function); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":902
 *         cdef Function func_expr = None
 *         if isinstance(function, Function):
 *             func_expr = <Function>function             # <<<<<<<<<<<<<<
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 *             value = context.names[(<Name>function).name]
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_function);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_func_expr, ((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":901
 *         cdef bint touched = function is not self.function
 *         cdef Function func_expr = None
 *         if isinstance(function, Function):             # <<<<<<<<<<<<<<
 *             func_expr = <Function>function
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":903
 *         if isinstance(function, Function):
 *             func_expr = <Function>function
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:             # <<<<<<<<<<<<<<
 *             value = context.names[(<Name>function).name]
 *             if isinstance(value, Function):
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_function), __pyx_ptype_7flitter_8language_4tree_Name); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_context->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 903, __pyx_L1_error)
  }
  __pyx_t_3 = (__Pyx_PyDict_ContainsTF(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_function)->name, __pyx_v_context->names, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 903, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":904
 *             func_expr = <Function>function
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 *             value = context.names[(<Name>function).name]             # <<<<<<<<<<<<<<
 *             if isinstance(value, Function):
 *                 func_expr = <Function>value
 */
    if (unlikely(__pyx_v_context->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 904, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_context->names, ((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_function)->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":905
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 *             value = context.names[(<Name>function).name]
 *             if isinstance(value, Function):             # <<<<<<<<<<<<<<
 *                 func_expr = <Function>value
 *         cdef bint literal_func = isinstance(function, Literal)
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7flitter_8language_4tree_Function); 
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":906
 *             value = context.names[(<Name>function).name]
 *             if isinstance(value, Function):
 *                 func_expr = <Function>value             # <<<<<<<<<<<<<<
 *         cdef bint literal_func = isinstance(function, Literal)
 *         if literal_func and not (<Literal>function).value.objects:
 */
      __pyx_t_1 = __pyx_v_value;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_func_expr, ((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":905
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:
 *             value = context.names[(<Name>function).name]
 *             if isinstance(value, Function):             # <<<<<<<<<<<<<<
 *                 func_expr = <Function>value
 *         cdef bint literal_func = isinstance(function, Literal)
 */
    }

    /* "flitter/language/tree.pyx":903
 *         if isinstance(function, Function):
 *             func_expr = <Function>function
 *         elif isinstance(function, Name) and (<Name>function).name in context.names:             # <<<<<<<<<<<<<<
 *             value = context.names[(<Name>function).name]
 *             if isinstance(value, Function):
 */
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":907
 *             if isinstance(value, Function):
 *                 func_expr = <Function>value
 *         cdef bint literal_func = isinstance(function, Literal)             # <<<<<<<<<<<<<<
 *         if literal_func and not (<Literal>function).value.objects:
 *             return NoOp
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_function), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_v_literal_func = __pyx_t_2;

  /* "flitter/language/tree.pyx":908
 *                 func_expr = <Function>value
 *         cdef bint literal_func = isinstance(function, Literal)
 *         if literal_func and not (<Literal>function).value.objects:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         cdef bint literal_args = True
 */
  if (__pyx_v_literal_func) {
  } else {
    __pyx_t_2 = __pyx_v_literal_func;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_3 = (((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value->objects != Py_None)&&(PyTuple_GET_SIZE(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value->objects) != 0);
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":909
 *         cdef bint literal_func = isinstance(function, Literal)
 *         if literal_func and not (<Literal>function).value.objects:
 *             return NoOp             # <<<<<<<<<<<<<<
 *         cdef bint literal_args = True
 *         cdef Expression arg, sarg, expr
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_7flitter_8language_4tree_NoOp);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":908
 *                 func_expr = <Function>value
 *         cdef bint literal_func = isinstance(function, Literal)
 *         if literal_func and not (<Literal>function).value.objects:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         cdef bint literal_args = True
 */
  }

  /* "flitter/language/tree.pyx":910
 *         if literal_func and not (<Literal>function).value.objects:
 *             return NoOp
 *         cdef bint literal_args = True             # <<<<<<<<<<<<<<
 *         cdef Expression arg, sarg, expr
 *         cdef list args = []
 */
  __pyx_v_literal_args = 1;

  /* "flitter/language/tree.pyx":912
 *         cdef bint literal_args = True
 *         cdef Expression arg, sarg, expr
 *         cdef list args = []             # <<<<<<<<<<<<<<
 *         if self.args:
 *             for arg in self.args:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":913
 *         cdef Expression arg, sarg, expr
 *         cdef list args = []
 *         if self.args:             # <<<<<<<<<<<<<<
 *             for arg in self.args:
 *                 sarg = arg._simplify(context)
 */
  __pyx_t_2 = (__pyx_v_self->args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->args) != 0);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":914
 *         cdef list args = []
 *         if self.args:
 *             for arg in self.args:             # <<<<<<<<<<<<<<
 *                 sarg = arg._simplify(context)
 *                 touched |= sarg is not arg
 */
    if (unlikely(__pyx_v_self->args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 914, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->args; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 914, __pyx_L1_error)
        #endif
        if (__pyx_t_5 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 914, __pyx_L1_error)
      #else
      __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_arg, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":915
 *         if self.args:
 *             for arg in self.args:
 *                 sarg = arg._simplify(context)             # <<<<<<<<<<<<<<
 *                 touched |= sarg is not arg
 *                 args.append(sarg)
 */
      __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_arg->__pyx_vtab)->_simplify(__pyx_v_arg, __pyx_v_context)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 915, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_sarg, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":916
 *             for arg in self.args:
 *                 sarg = arg._simplify(context)
 *                 touched |= sarg is not arg             # <<<<<<<<<<<<<<
 *                 args.append(sarg)
 *                 if not isinstance(sarg, Literal):
 */
      __pyx_t_2 = (__pyx_v_sarg != __pyx_v_arg);
      __pyx_v_touched = (__pyx_v_touched | __pyx_t_2);

      /* "flitter/language/tree.pyx":917
 *                 sarg = arg._simplify(context)
 *                 touched |= sarg is not arg
 *                 args.append(sarg)             # <<<<<<<<<<<<<<
 *                 if not isinstance(sarg, Literal):
 *                     literal_args = False
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_args, ((PyObject *)__pyx_v_sarg)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 917, __pyx_L1_error)

      /* "flitter/language/tree.pyx":918
 *                 touched |= sarg is not arg
 *                 args.append(sarg)
 *                 if not isinstance(sarg, Literal):             # <<<<<<<<<<<<<<
 *                     literal_args = False
 *         cdef list keyword_args = []
 */
      __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_sarg), __pyx_ptype_7flitter_8language_4tree_Literal); 
      __pyx_t_4 = (!__pyx_t_2);
      if (__pyx_t_4) {

        /* "flitter/language/tree.pyx":919
 *                 args.append(sarg)
 *                 if not isinstance(sarg, Literal):
 *                     literal_args = False             # <<<<<<<<<<<<<<
 *         cdef list keyword_args = []
 *         cdef Binding binding
 */
        __pyx_v_literal_args = 0;

        /* "flitter/language/tree.pyx":918
 *                 touched |= sarg is not arg
 *                 args.append(sarg)
 *                 if not isinstance(sarg, Literal):             # <<<<<<<<<<<<<<
 *                     literal_args = False
 *         cdef list keyword_args = []
 */
      }

      /* "flitter/language/tree.pyx":914
 *         cdef list args = []
 *         if self.args:
 *             for arg in self.args:             # <<<<<<<<<<<<<<
 *                 sarg = arg._simplify(context)
 *                 touched |= sarg is not arg
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":913
 *         cdef Expression arg, sarg, expr
 *         cdef list args = []
 *         if self.args:             # <<<<<<<<<<<<<<
 *             for arg in self.args:
 *                 sarg = arg._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":920
 *                 if not isinstance(sarg, Literal):
 *                     literal_args = False
 *         cdef list keyword_args = []             # <<<<<<<<<<<<<<
 *         cdef Binding binding
 *         if self.keyword_args:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_keyword_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":922
 *         cdef list keyword_args = []
 *         cdef Binding binding
 *         if self.keyword_args:             # <<<<<<<<<<<<<<
 *             for binding in self.keyword_args:
 *                 arg = binding.expr._simplify(context)
 */
  __pyx_t_4 = (__pyx_v_self->keyword_args != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_self->keyword_args) != 0);
  if (__pyx_t_4) {

    /* "flitter/language/tree.pyx":923
 *         cdef Binding binding
 *         if self.keyword_args:
 *             for binding in self.keyword_args:             # <<<<<<<<<<<<<<
 *                 arg = binding.expr._simplify(context)
 *                 touched |= arg is not binding.expr
 */
    if (unlikely(__pyx_v_self->keyword_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 923, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->keyword_args; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 923, __pyx_L1_error)
        #endif
        if (__pyx_t_5 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 923, __pyx_L1_error)
      #else
      __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 923, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":924
 *         if self.keyword_args:
 *             for binding in self.keyword_args:
 *                 arg = binding.expr._simplify(context)             # <<<<<<<<<<<<<<
 *                 touched |= arg is not binding.expr
 *                 keyword_args.append(Binding(binding.name, arg))
 */
      __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_binding->expr->__pyx_vtab)->_simplify(__pyx_v_binding->expr, __pyx_v_context)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_arg, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":925
 *             for binding in self.keyword_args:
 *                 arg = binding.expr._simplify(context)
 *                 touched |= arg is not binding.expr             # <<<<<<<<<<<<<<
 *                 keyword_args.append(Binding(binding.name, arg))
 *                 if not isinstance(arg, Literal):
 */
      __pyx_t_4 = (__pyx_v_arg != __pyx_v_binding->expr);
      __pyx_v_touched = (__pyx_v_touched | __pyx_t_4);

      /* "flitter/language/tree.pyx":926
 *                 arg = binding.expr._simplify(context)
 *                 touched |= arg is not binding.expr
 *                 keyword_args.append(Binding(binding.name, arg))             # <<<<<<<<<<<<<<
 *                 if not isinstance(arg, Literal):
 *                     literal_args = False
 */
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_binding->name);
      __Pyx_GIVEREF(__pyx_v_binding->name);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_binding->name)) __PYX_ERR(0, 926, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_arg);
      __Pyx_GIVEREF((PyObject *)__pyx_v_arg);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_arg))) __PYX_ERR(0, 926, __pyx_L1_error);
      __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding), __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_keyword_args, __pyx_t_8); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "flitter/language/tree.pyx":927
 *                 touched |= arg is not binding.expr
 *                 keyword_args.append(Binding(binding.name, arg))
 *                 if not isinstance(arg, Literal):             # <<<<<<<<<<<<<<
 *                     literal_args = False
 *         cdef list bindings
 */
      __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_arg), __pyx_ptype_7flitter_8language_4tree_Literal); 
      __pyx_t_2 = (!__pyx_t_4);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":928
 *                 keyword_args.append(Binding(binding.name, arg))
 *                 if not isinstance(arg, Literal):
 *                     literal_args = False             # <<<<<<<<<<<<<<
 *         cdef list bindings
 *         cdef dict kwargs
 */
        __pyx_v_literal_args = 0;

        /* "flitter/language/tree.pyx":927
 *                 touched |= arg is not binding.expr
 *                 keyword_args.append(Binding(binding.name, arg))
 *                 if not isinstance(arg, Literal):             # <<<<<<<<<<<<<<
 *                     literal_args = False
 *         cdef list bindings
 */
      }

      /* "flitter/language/tree.pyx":923
 *         cdef Binding binding
 *         if self.keyword_args:
 *             for binding in self.keyword_args:             # <<<<<<<<<<<<<<
 *                 arg = binding.expr._simplify(context)
 *                 touched |= arg is not binding.expr
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":922
 *         cdef list keyword_args = []
 *         cdef Binding binding
 *         if self.keyword_args:             # <<<<<<<<<<<<<<
 *             for binding in self.keyword_args:
 *                 arg = binding.expr._simplify(context)
 */
  }

  /* "flitter/language/tree.pyx":932
 *         cdef dict kwargs
 *         cdef int64_t i
 *         if func_expr is not None and not func_expr.captures and (not func_expr.recursive or literal_args):             # <<<<<<<<<<<<<<
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}
 *             bindings = []
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_func_expr) != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L21_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_func_expr->captures != Py_None)&&(PyTuple_GET_SIZE(__pyx_v_func_expr->captures) != 0);
  __pyx_t_3 = (!__pyx_t_4);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L21_bool_binop_done;
  }
  __pyx_t_3 = (!__pyx_v_func_expr->recursive);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L21_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_literal_args;
  __pyx_L21_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":933
 *         cdef int64_t i
 *         if func_expr is not None and not func_expr.captures and (not func_expr.recursive or literal_args):
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}             # <<<<<<<<<<<<<<
 *             bindings = []
 *             for i, binding in enumerate(func_expr.parameters):
 */
    { /* enter inner scope */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __pyx_v_keyword_args; __Pyx_INCREF(__pyx_t_8);
      __pyx_t_5 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 933, __pyx_L27_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 933, __pyx_L27_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 933, __pyx_L27_error)
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_6));
        __pyx_t_6 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_binding->name, (PyObject*)__pyx_8genexpr2__pyx_v_binding->expr))) __PYX_ERR(0, 933, __pyx_L27_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr2__pyx_v_binding); __pyx_8genexpr2__pyx_v_binding = 0;
      goto __pyx_L31_exit_scope;
      __pyx_L27_error:;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr2__pyx_v_binding); __pyx_8genexpr2__pyx_v_binding = 0;
      goto __pyx_L1_error;
      __pyx_L31_exit_scope:;
    } /* exit inner scope */
    __pyx_v_kwargs = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":934
 *         if func_expr is not None and not func_expr.captures and (not func_expr.recursive or literal_args):
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}
 *             bindings = []             # <<<<<<<<<<<<<<
 *             for i, binding in enumerate(func_expr.parameters):
 *                 if i < len(args):
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_bindings = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":935
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}
 *             bindings = []
 *             for i, binding in enumerate(func_expr.parameters):             # <<<<<<<<<<<<<<
 *                 if i < len(args):
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 */
    __pyx_t_9 = 0;
    __pyx_t_1 = __pyx_v_func_expr->parameters; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 935, __pyx_L1_error)
        #endif
        if (__pyx_t_5 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_8); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 935, __pyx_L1_error)
      #else
      __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_8));
      __pyx_t_8 = 0;
      __pyx_v_i = __pyx_t_9;
      __pyx_t_9 = (__pyx_t_9 + 1);

      /* "flitter/language/tree.pyx":936
 *             bindings = []
 *             for i, binding in enumerate(func_expr.parameters):
 *                 if i < len(args):             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 *                 elif binding.name in kwargs:
 */
      __pyx_t_10 = __Pyx_PyList_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 936, __pyx_L1_error)
      __pyx_t_2 = (__pyx_v_i < __pyx_t_10);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":937
 *             for i, binding in enumerate(func_expr.parameters):
 *                 if i < len(args):
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))             # <<<<<<<<<<<<<<
 *                 elif binding.name in kwargs:
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 */
        __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_binding->name);
        __Pyx_GIVEREF(__pyx_v_binding->name);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_binding->name)) __PYX_ERR(0, 937, __pyx_L1_error);
        __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_args, __pyx_v_i, int64_t, 1, __Pyx_PyInt_From_int64_t, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_8);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8)) __PYX_ERR(0, 937, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6));
        __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_6));
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error);
        __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_bindings, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 937, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/language/tree.pyx":936
 *             bindings = []
 *             for i, binding in enumerate(func_expr.parameters):
 *                 if i < len(args):             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 *                 elif binding.name in kwargs:
 */
        goto __pyx_L34;
      }

      /* "flitter/language/tree.pyx":938
 *                 if i < len(args):
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 *                 elif binding.name in kwargs:             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 *                 elif binding.expr is not None:
 */
      __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_binding->name, __pyx_v_kwargs, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 938, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":939
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 *                 elif binding.name in kwargs:
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))             # <<<<<<<<<<<<<<
 *                 elif binding.expr is not None:
 *                     bindings.append(PolyBinding((binding.name,), binding.expr))
 */
        __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_v_binding->name);
        __Pyx_GIVEREF(__pyx_v_binding->name);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_binding->name)) __PYX_ERR(0, 939, __pyx_L1_error);
        __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_kwargs, __pyx_v_binding->name); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6)) __PYX_ERR(0, 939, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_11));
        __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_11));
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_11)) __PYX_ERR(0, 939, __pyx_L1_error);
        __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_8, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_bindings, __pyx_t_11); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "flitter/language/tree.pyx":938
 *                 if i < len(args):
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 *                 elif binding.name in kwargs:             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 *                 elif binding.expr is not None:
 */
        goto __pyx_L34;
      }

      /* "flitter/language/tree.pyx":940
 *                 elif binding.name in kwargs:
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 *                 elif binding.expr is not None:             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), binding.expr))
 *                 else:
 */
      __pyx_t_2 = (((PyObject *)__pyx_v_binding->expr) != Py_None);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":941
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 *                 elif binding.expr is not None:
 *                     bindings.append(PolyBinding((binding.name,), binding.expr))             # <<<<<<<<<<<<<<
 *                 else:
 *                     bindings.append(PolyBinding((binding.name,), Literal(null_)))
 */
        __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 941, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_v_binding->name);
        __Pyx_GIVEREF(__pyx_v_binding->name);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_binding->name)) __PYX_ERR(0, 941, __pyx_L1_error);
        __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 941, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_11);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_11)) __PYX_ERR(0, 941, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_binding->expr);
        __Pyx_GIVEREF((PyObject *)__pyx_v_binding->expr);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_binding->expr))) __PYX_ERR(0, 941, __pyx_L1_error);
        __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_8, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 941, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_bindings, __pyx_t_11); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 941, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "flitter/language/tree.pyx":940
 *                 elif binding.name in kwargs:
 *                     bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
 *                 elif binding.expr is not None:             # <<<<<<<<<<<<<<
 *                     bindings.append(PolyBinding((binding.name,), binding.expr))
 *                 else:
 */
        goto __pyx_L34;
      }

      /* "flitter/language/tree.pyx":943
 *                     bindings.append(PolyBinding((binding.name,), binding.expr))
 *                 else:
 *                     bindings.append(PolyBinding((binding.name,), Literal(null_)))             # <<<<<<<<<<<<<<
 *             expr = Let(tuple(bindings), func_expr.body)._simplify(context)
 *             return expr
 */
      /*else*/ {
        __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_v_binding->name);
        __Pyx_GIVEREF(__pyx_v_binding->name);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_binding->name)) __PYX_ERR(0, 943, __pyx_L1_error);
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_7flitter_5model_null_)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_11);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11)) __PYX_ERR(0, 943, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_8);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_8)) __PYX_ERR(0, 943, __pyx_L1_error);
        __pyx_t_11 = 0;
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_bindings, __pyx_t_8); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 943, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_L34:;

      /* "flitter/language/tree.pyx":935
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}
 *             bindings = []
 *             for i, binding in enumerate(func_expr.parameters):             # <<<<<<<<<<<<<<
 *                 if i < len(args):
 *                     bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":944
 *                 else:
 *                     bindings.append(PolyBinding((binding.name,), Literal(null_)))
 *             expr = Let(tuple(bindings), func_expr.body)._simplify(context)             # <<<<<<<<<<<<<<
 *             return expr
 *         cdef list vector_args, results
 */
    __pyx_t_1 = PyList_AsTuple(__pyx_v_bindings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_func_expr->body);
    __Pyx_GIVEREF((PyObject *)__pyx_v_func_expr->body);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_func_expr->body))) __PYX_ERR(0, 944, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Let *)((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_t_1)->__pyx_base.__pyx_vtab)->__pyx_base._simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1), __pyx_v_context)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "flitter/language/tree.pyx":945
 *                     bindings.append(PolyBinding((binding.name,), Literal(null_)))
 *             expr = Let(tuple(bindings), func_expr.body)._simplify(context)
 *             return expr             # <<<<<<<<<<<<<<
 *         cdef list vector_args, results
 *         cdef Literal literal_arg
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_expr);
    __pyx_r = __pyx_v_expr;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":932
 *         cdef dict kwargs
 *         cdef int64_t i
 *         if func_expr is not None and not func_expr.captures and (not func_expr.recursive or literal_args):             # <<<<<<<<<<<<<<
 *             kwargs = {binding.name: binding.expr for binding in keyword_args}
 *             bindings = []
 */
  }

  /* "flitter/language/tree.pyx":948
 *         cdef list vector_args, results
 *         cdef Literal literal_arg
 *         if literal_func and literal_args:             # <<<<<<<<<<<<<<
 *             vector_args = [literal_arg.value for literal_arg in args]
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 */
  if (__pyx_v_literal_func) {
  } else {
    __pyx_t_2 = __pyx_v_literal_func;
    goto __pyx_L37_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_literal_args;
  __pyx_L37_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":949
 *         cdef Literal literal_arg
 *         if literal_func and literal_args:
 *             vector_args = [literal_arg.value for literal_arg in args]             # <<<<<<<<<<<<<<
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 *             results = []
 */
    { /* enter inner scope */
      __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 949, __pyx_L41_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __pyx_v_args; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 949, __pyx_L41_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 949, __pyx_L41_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 949, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_Literal))))) __PYX_ERR(0, 949, __pyx_L41_error)
        __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_literal_arg, ((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_t_6));
        __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_8genexpr3__pyx_v_literal_arg->value))) __PYX_ERR(0, 949, __pyx_L41_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr3__pyx_v_literal_arg); __pyx_8genexpr3__pyx_v_literal_arg = 0;
      goto __pyx_L45_exit_scope;
      __pyx_L41_error:;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr3__pyx_v_literal_arg); __pyx_8genexpr3__pyx_v_literal_arg = 0;
      goto __pyx_L1_error;
      __pyx_L45_exit_scope:;
    } /* exit inner scope */
    __pyx_v_vector_args = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "flitter/language/tree.pyx":950
 *         if literal_func and literal_args:
 *             vector_args = [literal_arg.value for literal_arg in args]
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}             # <<<<<<<<<<<<<<
 *             results = []
 *             for func in (<Literal>function).value.objects:
 */
    { /* enter inner scope */
      __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 950, __pyx_L48_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __pyx_v_keyword_args; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 950, __pyx_L48_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 950, __pyx_L48_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 950, __pyx_L48_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 950, __pyx_L48_error)
        __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_6));
        __pyx_t_6 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_8, (PyObject*)__pyx_8genexpr4__pyx_v_binding->name, (PyObject*)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_8genexpr4__pyx_v_binding->expr)->value))) __PYX_ERR(0, 950, __pyx_L48_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr4__pyx_v_binding); __pyx_8genexpr4__pyx_v_binding = 0;
      goto __pyx_L52_exit_scope;
      __pyx_L48_error:;
      __Pyx_XDECREF((PyObject *)__pyx_8genexpr4__pyx_v_binding); __pyx_8genexpr4__pyx_v_binding = 0;
      goto __pyx_L1_error;
      __pyx_L52_exit_scope:;
    } /* exit inner scope */
    __pyx_v_kwargs = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "flitter/language/tree.pyx":951
 *             vector_args = [literal_arg.value for literal_arg in args]
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 *             results = []             # <<<<<<<<<<<<<<
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):
 */
    __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_v_results = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "flitter/language/tree.pyx":952
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 *             results = []
 *             for func in (<Literal>function).value.objects:             # <<<<<<<<<<<<<<
 *                 if callable(func):
 *                     try:
 */
    if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value->objects == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 952, __pyx_L1_error)
    }
    __pyx_t_8 = ((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value->objects; __Pyx_INCREF(__pyx_t_8);
    __pyx_t_5 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 952, __pyx_L1_error)
        #endif
        if (__pyx_t_5 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 952, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_func, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":953
 *             results = []
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):             # <<<<<<<<<<<<<<
 *                     try:
 *                         assert not hasattr(func, 'context_func')
 */
      __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_v_func); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 953, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":954
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):
 *                     try:             # <<<<<<<<<<<<<<
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          /*try:*/ {

            /* "flitter/language/tree.pyx":955
 *                 if callable(func):
 *                     try:
 *                         assert not hasattr(func, 'context_func')             # <<<<<<<<<<<<<<
 *                         results.append(func(*vector_args, **kwargs))
 *                     except Exception as exc:
 */
            #ifndef CYTHON_WITHOUT_ASSERTIONS
            if (unlikely(__pyx_assertions_enabled())) {
              __pyx_t_2 = __Pyx_HasAttr(__pyx_v_func, __pyx_n_u_context_func); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 955, __pyx_L56_error)
              __pyx_t_3 = (!__pyx_t_2);
              if (unlikely(!__pyx_t_3)) {
                __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
                __PYX_ERR(0, 955, __pyx_L56_error)
              }
            }
            #else
            if ((1)); else __PYX_ERR(0, 955, __pyx_L56_error)
            #endif

            /* "flitter/language/tree.pyx":956
 *                     try:
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))             # <<<<<<<<<<<<<<
 *                     except Exception as exc:
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 */
            __pyx_t_1 = PySequence_Tuple(__pyx_v_vector_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L56_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_6 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 956, __pyx_L56_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_v_func, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 956, __pyx_L56_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_11); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 956, __pyx_L56_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "flitter/language/tree.pyx":954
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):
 *                     try:             # <<<<<<<<<<<<<<
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))
 */
          }
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          goto __pyx_L63_try_end;
          __pyx_L56_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "flitter/language/tree.pyx":957
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))
 *                     except Exception as exc:             # <<<<<<<<<<<<<<
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 *             return Literal(Vector._compose(results))
 */
          __pyx_t_15 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
          if (__pyx_t_15) {
            __Pyx_AddTraceback("flitter.language.tree.Call._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_6, &__pyx_t_1) < 0) __PYX_ERR(0, 957, __pyx_L58_except_error)
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_6);
            __Pyx_XGOTREF(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_6);
            __pyx_v_exc = __pyx_t_6;
            /*try:*/ {

              /* "flitter/language/tree.pyx":958
 *                         results.append(func(*vector_args, **kwargs))
 *                     except Exception as exc:
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")             # <<<<<<<<<<<<<<
 *             return Literal(Vector._compose(results))
 *         if isinstance(function, Literal) and len(args) == 1:
 */
              if (unlikely(__pyx_v_context->errors == Py_None)) {
                PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
                __PYX_ERR(0, 958, __pyx_L69_error)
              }
              __pyx_t_16 = PyTuple_New(4); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_10 = 0;
              __pyx_t_17 = 127;
              __Pyx_INCREF(__pyx_kp_u_Error_calling);
              __pyx_t_10 += 14;
              __Pyx_GIVEREF(__pyx_kp_u_Error_calling);
              PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_kp_u_Error_calling);
              __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_name_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_19 = __Pyx_PyObject_FormatSimple(__pyx_t_18, __pyx_empty_unicode); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) : __pyx_t_17;
              __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_19);
              __Pyx_GIVEREF(__pyx_t_19);
              PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_19);
              __pyx_t_19 = 0;
              __Pyx_INCREF(__pyx_kp_u__6);
              __pyx_t_10 += 2;
              __Pyx_GIVEREF(__pyx_kp_u__6);
              PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_kp_u__6);
              __pyx_t_19 = __Pyx_PyObject_Unicode(__pyx_v_exc); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_19) : __pyx_t_17;
              __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_19);
              __Pyx_GIVEREF(__pyx_t_19);
              PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_19);
              __pyx_t_19 = 0;
              __pyx_t_19 = __Pyx_PyUnicode_Join(__pyx_t_16, 4, __pyx_t_10, __pyx_t_17); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_7 = PySet_Add(__pyx_v_context->errors, __pyx_t_19); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 958, __pyx_L69_error)
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            }

            /* "flitter/language/tree.pyx":957
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))
 *                     except Exception as exc:             # <<<<<<<<<<<<<<
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 *             return Literal(Vector._compose(results))
 */
            /*finally:*/ {
              /*normal exit:*/{
                __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                goto __pyx_L70;
              }
              __pyx_L69_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24) < 0)) __Pyx_ErrFetch(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
                __Pyx_XGOTREF(__pyx_t_22);
                __Pyx_XGOTREF(__pyx_t_23);
                __Pyx_XGOTREF(__pyx_t_24);
                __Pyx_XGOTREF(__pyx_t_25);
                __Pyx_XGOTREF(__pyx_t_26);
                __Pyx_XGOTREF(__pyx_t_27);
                __pyx_t_15 = __pyx_lineno; __pyx_t_20 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_25);
                  __Pyx_XGIVEREF(__pyx_t_26);
                  __Pyx_XGIVEREF(__pyx_t_27);
                  __Pyx_ExceptionReset(__pyx_t_25, __pyx_t_26, __pyx_t_27);
                }
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_XGIVEREF(__pyx_t_24);
                __Pyx_ErrRestore(__pyx_t_22, __pyx_t_23, __pyx_t_24);
                __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                __pyx_lineno = __pyx_t_15; __pyx_clineno = __pyx_t_20; __pyx_filename = __pyx_t_21;
                goto __pyx_L58_except_error;
              }
              __pyx_L70:;
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L57_exception_handled;
          }
          goto __pyx_L58_except_error;

          /* "flitter/language/tree.pyx":954
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):
 *                     try:             # <<<<<<<<<<<<<<
 *                         assert not hasattr(func, 'context_func')
 *                         results.append(func(*vector_args, **kwargs))
 */
          __pyx_L58_except_error:;
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          goto __pyx_L1_error;
          __pyx_L57_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_L63_try_end:;
        }

        /* "flitter/language/tree.pyx":953
 *             results = []
 *             for func in (<Literal>function).value.objects:
 *                 if callable(func):             # <<<<<<<<<<<<<<
 *                     try:
 *                         assert not hasattr(func, 'context_func')
 */
      }

      /* "flitter/language/tree.pyx":952
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 *             results = []
 *             for func in (<Literal>function).value.objects:             # <<<<<<<<<<<<<<
 *                 if callable(func):
 *                     try:
 */
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "flitter/language/tree.pyx":959
 *                     except Exception as exc:
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 *             return Literal(Vector._compose(results))             # <<<<<<<<<<<<<<
 *         if isinstance(function, Literal) and len(args) == 1:
 *             if (<Literal>function).value == static_builtins['ceil']:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_8 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_compose(__pyx_v_results)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 959, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 959, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":948
 *         cdef list vector_args, results
 *         cdef Literal literal_arg
 *         if literal_func and literal_args:             # <<<<<<<<<<<<<<
 *             vector_args = [literal_arg.value for literal_arg in args]
 *             kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
 */
  }

  /* "flitter/language/tree.pyx":960
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 *             return Literal(Vector._compose(results))
 *         if isinstance(function, Literal) and len(args) == 1:             # <<<<<<<<<<<<<<
 *             if (<Literal>function).value == static_builtins['ceil']:
 *                 return Ceil(args[0])
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_function), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L77_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyList_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 960, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_5 == 1);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L77_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":961
 *             return Literal(Vector._compose(results))
 *         if isinstance(function, Literal) and len(args) == 1:
 *             if (<Literal>function).value == static_builtins['ceil']:             # <<<<<<<<<<<<<<
 *                 return Ceil(args[0])
 *             elif (<Literal>function).value == static_builtins['floor']:
 */
    if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 961, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_static_builtins, __pyx_n_u_ceil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_RichCompare(((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value), __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 961, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 961, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":962
 *         if isinstance(function, Literal) and len(args) == 1:
 *             if (<Literal>function).value == static_builtins['ceil']:
 *                 return Ceil(args[0])             # <<<<<<<<<<<<<<
 *             elif (<Literal>function).value == static_builtins['floor']:
 *                 return Floor(args[0])
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Ceil), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":961
 *             return Literal(Vector._compose(results))
 *         if isinstance(function, Literal) and len(args) == 1:
 *             if (<Literal>function).value == static_builtins['ceil']:             # <<<<<<<<<<<<<<
 *                 return Ceil(args[0])
 *             elif (<Literal>function).value == static_builtins['floor']:
 */
    }

    /* "flitter/language/tree.pyx":963
 *             if (<Literal>function).value == static_builtins['ceil']:
 *                 return Ceil(args[0])
 *             elif (<Literal>function).value == static_builtins['floor']:             # <<<<<<<<<<<<<<
 *                 return Floor(args[0])
 *             if (<Literal>function).value == static_builtins['fract']:
 */
    if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 963, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_static_builtins, __pyx_n_u_floor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_RichCompare(((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value), __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":964
 *                 return Ceil(args[0])
 *             elif (<Literal>function).value == static_builtins['floor']:
 *                 return Floor(args[0])             # <<<<<<<<<<<<<<
 *             if (<Literal>function).value == static_builtins['fract']:
 *                 return Fract(args[0])
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":963
 *             if (<Literal>function).value == static_builtins['ceil']:
 *                 return Ceil(args[0])
 *             elif (<Literal>function).value == static_builtins['floor']:             # <<<<<<<<<<<<<<
 *                 return Floor(args[0])
 *             if (<Literal>function).value == static_builtins['fract']:
 */
    }

    /* "flitter/language/tree.pyx":965
 *             elif (<Literal>function).value == static_builtins['floor']:
 *                 return Floor(args[0])
 *             if (<Literal>function).value == static_builtins['fract']:             # <<<<<<<<<<<<<<
 *                 return Fract(args[0])
 *         if not touched:
 */
    if (unlikely(__pyx_v_7flitter_8language_2vm_static_builtins == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 965, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_7flitter_8language_2vm_static_builtins, __pyx_n_u_fract); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_RichCompare(((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_function)->value), __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":966
 *                 return Floor(args[0])
 *             if (<Literal>function).value == static_builtins['fract']:
 *                 return Fract(args[0])             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":965
 *             elif (<Literal>function).value == static_builtins['floor']:
 *                 return Floor(args[0])
 *             if (<Literal>function).value == static_builtins['fract']:             # <<<<<<<<<<<<<<
 *                 return Fract(args[0])
 *         if not touched:
 */
    }

    /* "flitter/language/tree.pyx":960
 *                         context.errors.add(f"Error calling {func.__name__}: {str(exc)}")
 *             return Literal(Vector._compose(results))
 *         if isinstance(function, Literal) and len(args) == 1:             # <<<<<<<<<<<<<<
 *             if (<Literal>function).value == static_builtins['ceil']:
 *                 return Ceil(args[0])
 */
  }

  /* "flitter/language/tree.pyx":967
 *             if (<Literal>function).value == static_builtins['fract']:
 *                 return Fract(args[0])
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)
 */
  __pyx_t_3 = (!__pyx_v_touched);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":968
 *                 return Fract(args[0])
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":967
 *             if (<Literal>function).value == static_builtins['fract']:
 *                 return Fract(args[0])
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)
 */
  }

  /* "flitter/language/tree.pyx":969
 *         if not touched:
 *             return self
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyList_AsTuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_keyword_args) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = PyList_AsTuple(__pyx_v_keyword_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_8 = Py_None;
  }
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF((PyObject *)__pyx_v_function);
  __Pyx_GIVEREF((PyObject *)__pyx_v_function);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_function))) __PYX_ERR(0, 969, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_8)) __PYX_ERR(0, 969, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Call), __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_8);
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":897
 *             program.call(len(self.args) if self.args else 0, tuple(names) if names else None)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression function = self.function._simplify(context)
 *         cdef bint touched = function is not self.function
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("flitter.language.tree.Call._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_function);
  __Pyx_XDECREF((PyObject *)__pyx_v_func_expr);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_arg);
  __Pyx_XDECREF((PyObject *)__pyx_v_sarg);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_keyword_args);
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF(__pyx_v_bindings);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XDECREF(__pyx_v_vector_args);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_func);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF((PyObject *)__pyx_8genexpr2__pyx_v_binding);
  __Pyx_XDECREF((PyObject *)__pyx_8genexpr3__pyx_v_literal_arg);
  __Pyx_XDECREF((PyObject *)__pyx_8genexpr4__pyx_v_binding);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":971
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Call({self.function!r}, {self.args!r}, {self.keyword_args!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":972
 * 
 *     def __repr__(self):
 *         return f'Call({self.function!r}, {self.args!r}, {self.keyword_args!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Call);
  __pyx_t_2 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_Call);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Call);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->function)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->args), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->keyword_args), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":971
 *         return Call(function, tuple(args), tuple(keyword_args) if keyword_args else None)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Call({self.function!r}, {self.args!r}, {self.keyword_args!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Call.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":868
 * 
 * cdef class Call(Expression):
 *     cdef readonly Expression function             # <<<<<<<<<<<<<<
 *     cdef readonly tuple args
 *     cdef readonly tuple keyword_args
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_8function_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_8function_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_8function___get__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_8function___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->function);
  __pyx_r = ((PyObject *)__pyx_v_self->function);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":869
 * cdef class Call(Expression):
 *     cdef readonly Expression function
 *     cdef readonly tuple args             # <<<<<<<<<<<<<<
 *     cdef readonly tuple keyword_args
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_4args_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_4args_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_4args___get__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_4args___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->args);
  __pyx_r = __pyx_v_self->args;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":870
 *     cdef readonly Expression function
 *     cdef readonly tuple args
 *     cdef readonly tuple keyword_args             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression function, tuple args, tuple keyword_args=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_12keyword_args_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_12keyword_args_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_12keyword_args___get__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_12keyword_args___get__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->keyword_args);
  __pyx_r = __pyx_v_self->keyword_args;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Call_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Call_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.args, self.function, self.keyword_args)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->args)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->function);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->function);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->function))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->keyword_args);
  __Pyx_GIVEREF(__pyx_v_self->keyword_args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->keyword_args)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.args, self.function, self.keyword_args)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.args, self.function, self.keyword_args)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.args is not None or self.function is not None or self.keyword_args is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.args, self.function, self.keyword_args)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.args is not None or self.function is not None or self.keyword_args is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->args != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->function) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->keyword_args != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.function is not None or self.keyword_args is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.args is not None or self.function is not None or self.keyword_args is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Call); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_126472276);
    __Pyx_GIVEREF(__pyx_int_126472276);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_126472276)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.function is not None or self.keyword_args is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, None), state
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Call); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_126472276);
    __Pyx_GIVEREF(__pyx_int_126472276);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_126472276)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Call.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_4Call_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Call_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_4Call_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Call.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4Call_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4Call_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Call__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Call.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":976
 * 
 * cdef class NodeModifier(Expression):
 *     cdef readonly Expression node             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_4node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_4node_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_12NodeModifier_4node___get__(((struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier_4node___get__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node);
  __pyx_r = ((PyObject *)__pyx_v_self->node);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_12NodeModifier___reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier___reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.node,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->node))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.node,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.node,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.node,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, None), state
 */
  /*else*/ {
    __pyx_t_2 = (((PyObject *)__pyx_v_self->node) != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_NodeModifier); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_88468051);
    __Pyx_GIVEREF(__pyx_int_88468051);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_88468051)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, None), state
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_NodeModifier); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_88468051);
    __Pyx_GIVEREF(__pyx_int_88468051);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_88468051)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.NodeModifier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.NodeModifier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_12NodeModifier_2__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_12NodeModifier_2__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_NodeModifier__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.NodeModifier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":982
 *     cdef readonly str tag
 * 
 *     def __init__(self, Expression node, str tag):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.tag = tag
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_3Tag_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_3Tag_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  PyObject *__pyx_v_tag = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 982, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 982, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 982, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Tag.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "node", 0))) __PYX_ERR(0, 982, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 1, "tag", 1))) __PYX_ERR(0, 982, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Tag___init__(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v_self), __pyx_v_node, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_3Tag___init__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, PyObject *__pyx_v_tag) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":983
 * 
 *     def __init__(self, Expression node, str tag):
 *         self.node = node             # <<<<<<<<<<<<<<
 *         self.tag = tag
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __pyx_v_self->__pyx_base.node = __pyx_v_node;

  /* "flitter/language/tree.pyx":984
 *     def __init__(self, Expression node, str tag):
 *         self.node = node
 *         self.tag = tag             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_tag);
  __Pyx_GIVEREF(__pyx_v_tag);
  __Pyx_GOTREF(__pyx_v_self->tag);
  __Pyx_DECREF(__pyx_v_self->tag);
  __pyx_v_self->tag = __pyx_v_tag;

  /* "flitter/language/tree.pyx":982
 *     cdef readonly str tag
 * 
 *     def __init__(self, Expression node, str tag):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.tag = tag
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":986
 *         self.tag = tag
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         program.tag(self.tag)
 */

static void __pyx_f_7flitter_8language_4tree_3Tag__compile(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":987
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.node._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.tag(self.tag)
 * 
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.node->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.node, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L1_error)

  /* "flitter/language/tree.pyx":988
 *     cdef void _compile(self, Program program, list lnames):
 *         self.node._compile(program, lnames)
 *         program.tag(self.tag)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = __pyx_v_self->tag;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->tag(__pyx_v_program, ((PyObject*)__pyx_t_1), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":986
 *         self.tag = tag
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         program.tag(self.tag)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.language.tree.Tag._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":990
 *         program.tag(self.tag)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self.node._simplify(context)
 *         cdef Vector nodes
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Tag__simplify(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes = 0;
  PyObject *__pyx_v_objects = 0;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_n = 0;
  int64_t __pyx_v_i;
  PyObject *__pyx_v_obj = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":991
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression node = self.node._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Vector nodes
 *         cdef list objects
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.node->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.node, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":996
 *         cdef Node n
 *         cdef int64_t i
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":997
 *         cdef int64_t i
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value             # <<<<<<<<<<<<<<
 *             if nodes.objects is None:
 *                 return node
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_node)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_nodes = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":998
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:             # <<<<<<<<<<<<<<
 *                 return node
 *             objects = []
 */
    __pyx_t_2 = (__pyx_v_nodes->objects == ((PyObject*)Py_None));
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":999
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 *                 return node             # <<<<<<<<<<<<<<
 *             objects = []
 *             for i in range(nodes.length):
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_node);
      __pyx_r = __pyx_v_node;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":998
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:             # <<<<<<<<<<<<<<
 *                 return node
 *             objects = []
 */
    }

    /* "flitter/language/tree.pyx":1000
 *             if nodes.objects is None:
 *                 return node
 *             objects = []             # <<<<<<<<<<<<<<
 *             for i in range(nodes.length):
 *                 obj = nodes.objects[i]
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_objects = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1001
 *                 return node
 *             objects = []
 *             for i in range(nodes.length):             # <<<<<<<<<<<<<<
 *                 obj = nodes.objects[i]
 *                 if isinstance(obj, Node):
 */
    __pyx_t_3 = __pyx_v_nodes->length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/language/tree.pyx":1002
 *             objects = []
 *             for i in range(nodes.length):
 *                 obj = nodes.objects[i]             # <<<<<<<<<<<<<<
 *                 if isinstance(obj, Node):
 *                     n = (<Node>obj).copy()
 */
      if (unlikely(__pyx_v_nodes->objects == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1002, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_nodes->objects, __pyx_v_i, int64_t, 1, __Pyx_PyInt_From_int64_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1003
 *             for i in range(nodes.length):
 *                 obj = nodes.objects[i]
 *                 if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                     n = (<Node>obj).copy()
 *                     n.add_tag(self.tag)
 */
      __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7flitter_5model_Node); 
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":1004
 *                 obj = nodes.objects[i]
 *                 if isinstance(obj, Node):
 *                     n = (<Node>obj).copy()             # <<<<<<<<<<<<<<
 *                     n.add_tag(self.tag)
 *                     objects.append(n)
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_n, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1005
 *                 if isinstance(obj, Node):
 *                     n = (<Node>obj).copy()
 *                     n.add_tag(self.tag)             # <<<<<<<<<<<<<<
 *                     objects.append(n)
 *                 else:
 */
        __pyx_t_1 = __pyx_v_self->tag;
        __Pyx_INCREF(__pyx_t_1);
        ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_n->__pyx_vtab)->add_tag(__pyx_v_n, ((PyObject*)__pyx_t_1), 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1006
 *                     n = (<Node>obj).copy()
 *                     n.add_tag(self.tag)
 *                     objects.append(n)             # <<<<<<<<<<<<<<
 *                 else:
 *                     objects.append(obj)
 */
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_objects, ((PyObject *)__pyx_v_n)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)

        /* "flitter/language/tree.pyx":1003
 *             for i in range(nodes.length):
 *                 obj = nodes.objects[i]
 *                 if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                     n = (<Node>obj).copy()
 *                     n.add_tag(self.tag)
 */
        goto __pyx_L7;
      }

      /* "flitter/language/tree.pyx":1008
 *                     objects.append(n)
 *                 else:
 *                     objects.append(obj)             # <<<<<<<<<<<<<<
 *             return Literal(objects)
 *         if node is self.node:
 */
      /*else*/ {
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_objects, __pyx_v_obj); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1008, __pyx_L1_error)
      }
      __pyx_L7:;
    }

    /* "flitter/language/tree.pyx":1009
 *                 else:
 *                     objects.append(obj)
 *             return Literal(objects)             # <<<<<<<<<<<<<<
 *         if node is self.node:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_v_objects); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":996
 *         cdef Node n
 *         cdef int64_t i
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 */
  }

  /* "flitter/language/tree.pyx":1010
 *                     objects.append(obj)
 *             return Literal(objects)
 *         if node is self.node:             # <<<<<<<<<<<<<<
 *             return self
 *         return Tag(node, self.tag)
 */
  __pyx_t_2 = (__pyx_v_node == __pyx_v_self->__pyx_base.node);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1011
 *             return Literal(objects)
 *         if node is self.node:
 *             return self             # <<<<<<<<<<<<<<
 *         return Tag(node, self.tag)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1010
 *                     objects.append(obj)
 *             return Literal(objects)
 *         if node is self.node:             # <<<<<<<<<<<<<<
 *             return self
 *         return Tag(node, self.tag)
 */
  }

  /* "flitter/language/tree.pyx":1012
 *         if node is self.node:
 *             return self
 *         return Tag(node, self.tag)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_node))) __PYX_ERR(0, 1012, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->tag);
  __Pyx_GIVEREF(__pyx_v_self->tag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->tag)) __PYX_ERR(0, 1012, __pyx_L1_error);
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Tag), __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7);
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":990
 *         program.tag(self.tag)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self.node._simplify(context)
 *         cdef Vector nodes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.Tag._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes);
  __Pyx_XDECREF(__pyx_v_objects);
  __Pyx_XDECREF((PyObject *)__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1014
 *         return Tag(node, self.tag)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Tag({self.node!r}, {self.tag!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Tag_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1015
 * 
 *     def __repr__(self):
 *         return f'Tag({self.node!r}, {self.tag!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Tag);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_Tag);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Tag);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->__pyx_base.node)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->tag), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1014
 *         return Tag(node, self.tag)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Tag({self.node!r}, {self.tag!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Tag.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":980
 * 
 * cdef class Tag(NodeModifier):
 *     cdef readonly str tag             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression node, str tag):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_3tag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_3tag_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Tag_3tag___get__(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_3tag___get__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->tag);
  __pyx_r = __pyx_v_self->tag;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Tag_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Tag_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Tag_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.node, self.tag)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.node))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->tag);
  __Pyx_GIVEREF(__pyx_v_self->tag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->tag)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.node, self.tag)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.node, self.tag)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.node is not None or self.tag is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.node, self.tag)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.node is not None or self.tag is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.node) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->tag != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.node is not None or self.tag is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.node is not None or self.tag is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Tag); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_158473136);
    __Pyx_GIVEREF(__pyx_int_158473136);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_158473136)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.node is not None or self.tag is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, None), state
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Tag); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_158473136);
    __Pyx_GIVEREF(__pyx_int_158473136);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_158473136)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Tag.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Tag_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Tag_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Tag_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Tag.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Tag_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Tag_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Tag__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Tag.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1021
 *     cdef readonly tuple bindings
 * 
 *     def __init__(self, Expression node, tuple bindings):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.bindings = bindings
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_10Attributes_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_10Attributes_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  PyObject *__pyx_v_bindings = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_bindings,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bindings)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1021, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1021, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_bindings = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1021, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Attributes.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "node", 0))) __PYX_ERR(0, 1021, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bindings), (&PyTuple_Type), 1, "bindings", 1))) __PYX_ERR(0, 1021, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Attributes___init__(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_self), __pyx_v_node, __pyx_v_bindings);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_10Attributes___init__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, PyObject *__pyx_v_bindings) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1022
 * 
 *     def __init__(self, Expression node, tuple bindings):
 *         self.node = node             # <<<<<<<<<<<<<<
 *         self.bindings = bindings
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __pyx_v_self->__pyx_base.node = __pyx_v_node;

  /* "flitter/language/tree.pyx":1023
 *     def __init__(self, Expression node, tuple bindings):
 *         self.node = node
 *         self.bindings = bindings             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF(__pyx_v_bindings);
  __Pyx_GIVEREF(__pyx_v_bindings);
  __Pyx_GOTREF(__pyx_v_self->bindings);
  __Pyx_DECREF(__pyx_v_self->bindings);
  __pyx_v_self->bindings = __pyx_v_bindings;

  /* "flitter/language/tree.pyx":1021
 *     cdef readonly tuple bindings
 * 
 *     def __init__(self, Expression node, tuple bindings):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.bindings = bindings
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1025
 *         self.bindings = bindings
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         cdef Binding binding
 */

static void __pyx_f_7flitter_8language_4tree_10Attributes__compile(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_binding = 0;
  PyObject *__pyx_v_names = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1026
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.node._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef Binding binding
 *         cdef list names = []
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.node->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.node, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1028
 *         self.node._compile(program, lnames)
 *         cdef Binding binding
 *         cdef list names = []             # <<<<<<<<<<<<<<
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1029
 *         cdef Binding binding
 *         cdef list names = []
 *         for binding in self.bindings:             # <<<<<<<<<<<<<<
 *             binding.expr._compile(program, lnames)
 *             names.append(binding.name)
 */
  if (unlikely(__pyx_v_self->bindings == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1029, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->bindings; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1029, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1029, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1030
 *         cdef list names = []
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             names.append(binding.name)
 *         program.attributes(tuple(names))
 */
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_binding->expr->__pyx_vtab)->_compile(__pyx_v_binding->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1030, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1031
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)
 *             names.append(binding.name)             # <<<<<<<<<<<<<<
 *         program.attributes(tuple(names))
 * 
 */
    __pyx_t_3 = __pyx_v_binding->name;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_names, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1031, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1029
 *         cdef Binding binding
 *         cdef list names = []
 *         for binding in self.bindings:             # <<<<<<<<<<<<<<
 *             binding.expr._compile(program, lnames)
 *             names.append(binding.name)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1032
 *             binding.expr._compile(program, lnames)
 *             names.append(binding.name)
 *         program.attributes(tuple(names))             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = PyList_AsTuple(__pyx_v_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->attributes(__pyx_v_program, ((PyObject*)__pyx_t_1), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":1025
 *         self.bindings = bindings
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         cdef Binding binding
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Attributes._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1034
 *         program.attributes(tuple(names))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self
 *         cdef list bindings = []
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_10Attributes__simplify(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  PyObject *__pyx_v_bindings = 0;
  struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_attrs = 0;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_binding = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  int __pyx_v_touched;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  PyObject *__pyx_v_objects = 0;
  PyObject *__pyx_v_obj = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1035
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression node = self             # <<<<<<<<<<<<<<
 *         cdef list bindings = []
 *         cdef Attributes attrs
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);

  /* "flitter/language/tree.pyx":1036
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression node = self
 *         cdef list bindings = []             # <<<<<<<<<<<<<<
 *         cdef Attributes attrs
 *         cdef Binding binding
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bindings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1040
 *         cdef Binding binding
 *         cdef Expression expr
 *         cdef bint touched = False             # <<<<<<<<<<<<<<
 *         while isinstance(node, Attributes):
 *             attrs = <Attributes>node
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":1041
 *         cdef Expression expr
 *         cdef bint touched = False
 *         while isinstance(node, Attributes):             # <<<<<<<<<<<<<<
 *             attrs = <Attributes>node
 *             for binding in reversed(attrs.bindings):
 */
  while (1) {
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Attributes); 
    if (!__pyx_t_2) break;

    /* "flitter/language/tree.pyx":1042
 *         cdef bint touched = False
 *         while isinstance(node, Attributes):
 *             attrs = <Attributes>node             # <<<<<<<<<<<<<<
 *             for binding in reversed(attrs.bindings):
 *                 expr = binding.expr._simplify(context)
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_attrs, ((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1043
 *         while isinstance(node, Attributes):
 *             attrs = <Attributes>node
 *             for binding in reversed(attrs.bindings):             # <<<<<<<<<<<<<<
 *                 expr = binding.expr._simplify(context)
 *                 bindings.append(Binding(binding.name, expr))
 */
    if (unlikely(__pyx_v_attrs->bindings == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1043, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_attrs->bindings; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1043, __pyx_L1_error)
    #endif
    --__pyx_t_3;
    for (;;) {
      if (__pyx_t_3 < 0) break;
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1043, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3--; if (unlikely((0 < 0))) __PYX_ERR(0, 1043, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3--; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":1044
 *             attrs = <Attributes>node
 *             for binding in reversed(attrs.bindings):
 *                 expr = binding.expr._simplify(context)             # <<<<<<<<<<<<<<
 *                 bindings.append(Binding(binding.name, expr))
 *                 touched |= expr is not binding.expr
 */
      __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_binding->expr->__pyx_vtab)->_simplify(__pyx_v_binding->expr, __pyx_v_context)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "flitter/language/tree.pyx":1045
 *             for binding in reversed(attrs.bindings):
 *                 expr = binding.expr._simplify(context)
 *                 bindings.append(Binding(binding.name, expr))             # <<<<<<<<<<<<<<
 *                 touched |= expr is not binding.expr
 *             node = attrs.node
 */
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_binding->name);
      __Pyx_GIVEREF(__pyx_v_binding->name);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_binding->name)) __PYX_ERR(0, 1045, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_expr);
      __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 1045, __pyx_L1_error);
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_bindings, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":1046
 *                 expr = binding.expr._simplify(context)
 *                 bindings.append(Binding(binding.name, expr))
 *                 touched |= expr is not binding.expr             # <<<<<<<<<<<<<<
 *             node = attrs.node
 *         node = node._simplify(context)
 */
      __pyx_t_2 = (__pyx_v_expr != __pyx_v_binding->expr);
      __pyx_v_touched = (__pyx_v_touched | __pyx_t_2);

      /* "flitter/language/tree.pyx":1043
 *         while isinstance(node, Attributes):
 *             attrs = <Attributes>node
 *             for binding in reversed(attrs.bindings):             # <<<<<<<<<<<<<<
 *                 expr = binding.expr._simplify(context)
 *                 bindings.append(Binding(binding.name, expr))
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1047
 *                 bindings.append(Binding(binding.name, expr))
 *                 touched |= expr is not binding.expr
 *             node = attrs.node             # <<<<<<<<<<<<<<
 *         node = node._simplify(context)
 *         touched |= node is not self.node
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_attrs->__pyx_base.node);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
    __pyx_t_1 = 0;
  }

  /* "flitter/language/tree.pyx":1048
 *                 touched |= expr is not binding.expr
 *             node = attrs.node
 *         node = node._simplify(context)             # <<<<<<<<<<<<<<
 *         touched |= node is not self.node
 *         cdef Vector nodes, value
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_node->__pyx_vtab)->_simplify(__pyx_v_node, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1049
 *             node = attrs.node
 *         node = node._simplify(context)
 *         touched |= node is not self.node             # <<<<<<<<<<<<<<
 *         cdef Vector nodes, value
 *         cdef list objects
 */
  __pyx_t_2 = (__pyx_v_node != __pyx_v_self->__pyx_base.node);
  __pyx_v_touched = (__pyx_v_touched | __pyx_t_2);

  /* "flitter/language/tree.pyx":1052
 *         cdef Vector nodes, value
 *         cdef list objects
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1053
 *         cdef list objects
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value             # <<<<<<<<<<<<<<
 *             if nodes.objects is None:
 *                 return node
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_node)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_nodes = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1054
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:             # <<<<<<<<<<<<<<
 *                 return node
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 */
    __pyx_t_2 = (__pyx_v_nodes->objects == ((PyObject*)Py_None));
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1055
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 *                 return node             # <<<<<<<<<<<<<<
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                 objects = []
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_node);
      __pyx_r = __pyx_v_node;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":1054
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:             # <<<<<<<<<<<<<<
 *                 return node
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 */
    }

    /* "flitter/language/tree.pyx":1056
 *             if nodes.objects is None:
 *                 return node
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):             # <<<<<<<<<<<<<<
 *                 objects = []
 *                 for obj in nodes.objects:
 */
    __pyx_t_7 = (PyList_GET_SIZE(__pyx_v_bindings) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_2 = __pyx_t_7;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_bindings, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_1)->expr);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_TypeCheck(__pyx_t_5, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __pyx_t_7;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1057
 *                 return node
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                 objects = []             # <<<<<<<<<<<<<<
 *                 for obj in nodes.objects:
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_objects = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":1058
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                 objects = []
 *                 for obj in nodes.objects:             # <<<<<<<<<<<<<<
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 */
      if (unlikely(__pyx_v_nodes->objects == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 1058, __pyx_L1_error)
      }
      __pyx_t_5 = __pyx_v_nodes->objects; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1058, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1058, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1059
 *                 objects = []
 *                 for obj in nodes.objects:
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)             # <<<<<<<<<<<<<<
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                     binding = bindings.pop()
 */
        __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7flitter_5model_Node); 
        if (__pyx_t_2) {
          __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj), 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1059, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __pyx_t_4;
          __pyx_t_4 = 0;
        } else {
          __Pyx_INCREF(__pyx_v_obj);
          __pyx_t_1 = __pyx_v_obj;
        }
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_objects, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1059, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1058
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                 objects = []
 *                 for obj in nodes.objects:             # <<<<<<<<<<<<<<
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":1060
 *                 for obj in nodes.objects:
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):             # <<<<<<<<<<<<<<
 *                     binding = bindings.pop()
 *                     value = (<Literal>binding.expr).value
 */
      while (1) {
        __pyx_t_7 = (PyList_GET_SIZE(__pyx_v_bindings) != 0);
        if (__pyx_t_7) {
        } else {
          __pyx_t_2 = __pyx_t_7;
          goto __pyx_L18_bool_binop_done;
        }
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_bindings, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1060, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_5)->expr);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_7 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __pyx_t_7;
        __pyx_L18_bool_binop_done:;
        if (!__pyx_t_2) break;

        /* "flitter/language/tree.pyx":1061
 *                     objects.append((<Node>obj).copy() if isinstance(obj, Node) else obj)
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                     binding = bindings.pop()             # <<<<<<<<<<<<<<
 *                     value = (<Literal>binding.expr).value
 *                     for obj in objects:
 */
        __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_bindings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1061, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1062
 *                 while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
 *                     binding = bindings.pop()
 *                     value = (<Literal>binding.expr).value             # <<<<<<<<<<<<<<
 *                     for obj in objects:
 *                         if isinstance(obj, Node):
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_binding->expr)->value);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1063
 *                     binding = bindings.pop()
 *                     value = (<Literal>binding.expr).value
 *                     for obj in objects:             # <<<<<<<<<<<<<<
 *                         if isinstance(obj, Node):
 *                             (<Node>obj).set_attribute(binding.name, value)
 */
        __pyx_t_1 = __pyx_v_objects; __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = 0;
        for (;;) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1063, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1063, __pyx_L1_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1063, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "flitter/language/tree.pyx":1064
 *                     value = (<Literal>binding.expr).value
 *                     for obj in objects:
 *                         if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                             (<Node>obj).set_attribute(binding.name, value)
 *                 node = Literal(objects)
 */
          __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7flitter_5model_Node); 
          if (__pyx_t_2) {

            /* "flitter/language/tree.pyx":1065
 *                     for obj in objects:
 *                         if isinstance(obj, Node):
 *                             (<Node>obj).set_attribute(binding.name, value)             # <<<<<<<<<<<<<<
 *                 node = Literal(objects)
 *                 touched = True
 */
            __pyx_t_5 = __pyx_v_binding->name;
            __Pyx_INCREF(__pyx_t_5);
            ((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj)->__pyx_vtab)->set_attribute(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj), ((PyObject*)__pyx_t_5), __pyx_v_value, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1065, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "flitter/language/tree.pyx":1064
 *                     value = (<Literal>binding.expr).value
 *                     for obj in objects:
 *                         if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                             (<Node>obj).set_attribute(binding.name, value)
 *                 node = Literal(objects)
 */
          }

          /* "flitter/language/tree.pyx":1063
 *                     binding = bindings.pop()
 *                     value = (<Literal>binding.expr).value
 *                     for obj in objects:             # <<<<<<<<<<<<<<
 *                         if isinstance(obj, Node):
 *                             (<Node>obj).set_attribute(binding.name, value)
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "flitter/language/tree.pyx":1066
 *                         if isinstance(obj, Node):
 *                             (<Node>obj).set_attribute(binding.name, value)
 *                 node = Literal(objects)             # <<<<<<<<<<<<<<
 *                 touched = True
 *         if not touched:
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_v_objects); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1067
 *                             (<Node>obj).set_attribute(binding.name, value)
 *                 node = Literal(objects)
 *                 touched = True             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
      __pyx_v_touched = 1;

      /* "flitter/language/tree.pyx":1056
 *             if nodes.objects is None:
 *                 return node
 *             if bindings and isinstance((<Binding>bindings[-1]).expr, Literal):             # <<<<<<<<<<<<<<
 *                 objects = []
 *                 for obj in nodes.objects:
 */
    }

    /* "flitter/language/tree.pyx":1052
 *         cdef Vector nodes, value
 *         cdef list objects
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if nodes.objects is None:
 */
  }

  /* "flitter/language/tree.pyx":1068
 *                 node = Literal(objects)
 *                 touched = True
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if bindings:
 */
  __pyx_t_2 = (!__pyx_v_touched);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1069
 *                 touched = True
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         if bindings:
 *             bindings.reverse()
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1068
 *                 node = Literal(objects)
 *                 touched = True
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if bindings:
 */
  }

  /* "flitter/language/tree.pyx":1070
 *         if not touched:
 *             return self
 *         if bindings:             # <<<<<<<<<<<<<<
 *             bindings.reverse()
 *             node = Attributes(node, tuple(bindings))
 */
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_bindings) != 0);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1071
 *             return self
 *         if bindings:
 *             bindings.reverse()             # <<<<<<<<<<<<<<
 *             node = Attributes(node, tuple(bindings))
 *         return node
 */
    __pyx_t_6 = PyList_Reverse(__pyx_v_bindings); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1071, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1072
 *         if bindings:
 *             bindings.reverse()
 *             node = Attributes(node, tuple(bindings))             # <<<<<<<<<<<<<<
 *         return node
 * 
 */
    __pyx_t_1 = PyList_AsTuple(__pyx_v_bindings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF((PyObject *)__pyx_v_node);
    __Pyx_GIVEREF((PyObject *)__pyx_v_node);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_node))) __PYX_ERR(0, 1072, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Attributes), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1070
 *         if not touched:
 *             return self
 *         if bindings:             # <<<<<<<<<<<<<<
 *             bindings.reverse()
 *             node = Attributes(node, tuple(bindings))
 */
  }

  /* "flitter/language/tree.pyx":1073
 *             bindings.reverse()
 *             node = Attributes(node, tuple(bindings))
 *         return node             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __pyx_r = __pyx_v_node;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1034
 *         program.attributes(tuple(names))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self
 *         cdef list bindings = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Attributes._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_bindings);
  __Pyx_XDECREF((PyObject *)__pyx_v_attrs);
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_objects);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1075
 *         return node
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.node!r}, {self.bindings!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Attributes_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1076
 * 
 *     def __repr__(self):
 *         return f'{self.__class__.__name__}({self.node!r}, {self.bindings!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->__pyx_base.node)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->bindings), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1075
 *         return node
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.node!r}, {self.bindings!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Attributes.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1019
 * 
 * cdef class Attributes(NodeModifier):
 *     cdef readonly tuple bindings             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression node, tuple bindings):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_8bindings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_8bindings_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Attributes_8bindings___get__(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_8bindings___get__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->bindings);
  __pyx_r = __pyx_v_self->bindings;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Attributes_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Attributes_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Attributes_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.bindings, self.node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->bindings);
  __Pyx_GIVEREF(__pyx_v_self->bindings);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->bindings)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.node))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.bindings, self.node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.bindings, self.node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.bindings is not None or self.node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.bindings, self.node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.bindings is not None or self.node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->bindings != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.node) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bindings is not None or self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.bindings is not None or self.node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Attributes); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_89314571);
    __Pyx_GIVEREF(__pyx_int_89314571);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_89314571)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bindings is not None or self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, None), state
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Attributes); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_89314571);
    __Pyx_GIVEREF(__pyx_int_89314571);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_89314571)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Attributes.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_10Attributes_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Attributes_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_10Attributes_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Attributes.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10Attributes_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10Attributes_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Attributes__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Attributes.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1082
 *     cdef readonly Expression children
 * 
 *     def __init__(self, Expression node, Expression children):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.children = children
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_6Append_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_6Append_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_children = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_children,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1082, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_children)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1082, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1082, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_children = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1082, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Append.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "node", 0))) __PYX_ERR(0, 1082, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_children), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "children", 0))) __PYX_ERR(0, 1082, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Append___init__(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v_self), __pyx_v_node, __pyx_v_children);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_6Append___init__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_children) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1083
 * 
 *     def __init__(self, Expression node, Expression children):
 *         self.node = node             # <<<<<<<<<<<<<<
 *         self.children = children
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __pyx_v_self->__pyx_base.node = __pyx_v_node;

  /* "flitter/language/tree.pyx":1084
 *     def __init__(self, Expression node, Expression children):
 *         self.node = node
 *         self.children = children             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_children);
  __Pyx_GIVEREF((PyObject *)__pyx_v_children);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->children);
  __Pyx_DECREF((PyObject *)__pyx_v_self->children);
  __pyx_v_self->children = __pyx_v_children;

  /* "flitter/language/tree.pyx":1082
 *     cdef readonly Expression children
 * 
 *     def __init__(self, Expression node, Expression children):             # <<<<<<<<<<<<<<
 *         self.node = node
 *         self.children = children
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1086
 *         self.children = children
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         self.children._compile(program, lnames)
 */

static void __pyx_f_7flitter_8language_4tree_6Append__compile(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instr = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_append __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1087
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.node._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         self.children._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.node->__pyx_vtab)->_compile(__pyx_v_self->__pyx_base.node, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1087, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1088
 *     cdef void _compile(self, Program program, list lnames):
 *         self.node._compile(program, lnames)
 *         self.children._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->children->__pyx_vtab)->_compile(__pyx_v_self->children, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1088, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1089
 *         self.node._compile(program, lnames)
 *         self.children._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()             # <<<<<<<<<<<<<<
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1090
 *         self.children._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 */
  __pyx_t_2 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_Compose);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1091
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_instr, ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1092
 *         if instr.code == OpCode.Compose:
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)             # <<<<<<<<<<<<<<
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 *             program.pop_instruction()
 */
    __pyx_t_3.__pyx_n = 1;
    __pyx_t_3.count = ((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instr)->value;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->append(__pyx_v_program, 0, &__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1090
 *         self.children._compile(program, lnames)
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":1093
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *         else:
 */
  __pyx_t_4 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_Literal);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (((struct __pyx_obj_7flitter_8language_2vm_InstructionVector *)__pyx_v_instr)->value->length == 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1094
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *         else:
 *             program.append()
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1093
 *             instr = program.pop_instruction()
 *             program.append((<InstructionInt>instr).value)
 *         elif instr.code == OpCode.Literal and (<InstructionVector>instr).value.length == 0:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "flitter/language/tree.pyx":1096
 *             program.pop_instruction()
 *         else:
 *             program.append()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->append(__pyx_v_program, 0, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "flitter/language/tree.pyx":1086
 *         self.children = children
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.node._compile(program, lnames)
 *         self.children._compile(program, lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Append._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1098
 *             program.append()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self.node._simplify(context)
 *         cdef Expression children = self.children._simplify(context)
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6Append__simplify(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_node = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_children = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_nodes = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_children_vector = 0;
  int64_t __pyx_v_i;
  PyObject *__pyx_v_objects = 0;
  PyObject *__pyx_v_node_objects = 0;
  PyObject *__pyx_v_obj = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1099
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression node = self.node._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Expression children = self.children._simplify(context)
 *         cdef Vector nodes, children_vector
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->__pyx_base.node->__pyx_vtab)->_simplify(__pyx_v_self->__pyx_base.node, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1100
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression node = self.node._simplify(context)
 *         cdef Expression children = self.children._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef Vector nodes, children_vector
 *         cdef int64_t i
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->children->__pyx_vtab)->_simplify(__pyx_v_self->children, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_children = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1105
 *         cdef list objects
 *         cdef tuple node_objects
 *         if (isinstance(node, Literal) and (<Literal>node).value.objects is None) or \             # <<<<<<<<<<<<<<
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (!__pyx_t_3) {
    goto __pyx_L5_next_or;
  } else {
  }
  __pyx_t_3 = (((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_node)->value->objects == ((PyObject*)Py_None));
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L5_next_or:;

  /* "flitter/language/tree.pyx":1106
 *         cdef tuple node_objects
 *         if (isinstance(node, Literal) and (<Literal>node).value.objects is None) or \
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):             # <<<<<<<<<<<<<<
 *             return node
 *         if isinstance(node, Literal):
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_children), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_children)->value->objects == ((PyObject*)Py_None));
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "flitter/language/tree.pyx":1105
 *         cdef list objects
 *         cdef tuple node_objects
 *         if (isinstance(node, Literal) and (<Literal>node).value.objects is None) or \             # <<<<<<<<<<<<<<
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node
 */
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1107
 *         if (isinstance(node, Literal) and (<Literal>node).value.objects is None) or \
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node             # <<<<<<<<<<<<<<
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_node);
    __pyx_r = __pyx_v_node;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1105
 *         cdef list objects
 *         cdef tuple node_objects
 *         if (isinstance(node, Literal) and (<Literal>node).value.objects is None) or \             # <<<<<<<<<<<<<<
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node
 */
  }

  /* "flitter/language/tree.pyx":1108
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if isinstance(children, Literal):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1109
 *             return node
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value             # <<<<<<<<<<<<<<
 *             if isinstance(children, Literal):
 *                 children_vector = (<Literal>children).value
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_node)->value);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_nodes = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1110
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if isinstance(children, Literal):             # <<<<<<<<<<<<<<
 *                 children_vector = (<Literal>children).value
 *                 node_objects = nodes.objects
 */
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_children), __pyx_ptype_7flitter_8language_4tree_Literal); 
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1111
 *             nodes = (<Literal>node).value
 *             if isinstance(children, Literal):
 *                 children_vector = (<Literal>children).value             # <<<<<<<<<<<<<<
 *                 node_objects = nodes.objects
 *                 objects = []
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_children)->value);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_children_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1112
 *             if isinstance(children, Literal):
 *                 children_vector = (<Literal>children).value
 *                 node_objects = nodes.objects             # <<<<<<<<<<<<<<
 *                 objects = []
 *                 for i in range(nodes.length):
 */
      __pyx_t_1 = __pyx_v_nodes->objects;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_node_objects = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1113
 *                 children_vector = (<Literal>children).value
 *                 node_objects = nodes.objects
 *                 objects = []             # <<<<<<<<<<<<<<
 *                 for i in range(nodes.length):
 *                     obj = node_objects[i]
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_objects = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1114
 *                 node_objects = nodes.objects
 *                 objects = []
 *                 for i in range(nodes.length):             # <<<<<<<<<<<<<<
 *                     obj = node_objects[i]
 *                     if isinstance(obj, Node):
 */
      __pyx_t_4 = __pyx_v_nodes->length;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "flitter/language/tree.pyx":1115
 *                 objects = []
 *                 for i in range(nodes.length):
 *                     obj = node_objects[i]             # <<<<<<<<<<<<<<
 *                     if isinstance(obj, Node):
 *                         obj = (<Node>obj).copy()
 */
        if (unlikely(__pyx_v_node_objects == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1115, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_node_objects, __pyx_v_i, int64_t, 1, __Pyx_PyInt_From_int64_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1115, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "flitter/language/tree.pyx":1116
 *                 for i in range(nodes.length):
 *                     obj = node_objects[i]
 *                     if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                         obj = (<Node>obj).copy()
 *                         (<Node>obj).append_vector(children_vector)
 */
        __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7flitter_5model_Node); 
        if (__pyx_t_2) {

          /* "flitter/language/tree.pyx":1117
 *                     obj = node_objects[i]
 *                     if isinstance(obj, Node):
 *                         obj = (<Node>obj).copy()             # <<<<<<<<<<<<<<
 *                         (<Node>obj).append_vector(children_vector)
 *                     objects.append(obj)
 */
          __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "flitter/language/tree.pyx":1118
 *                     if isinstance(obj, Node):
 *                         obj = (<Node>obj).copy()
 *                         (<Node>obj).append_vector(children_vector)             # <<<<<<<<<<<<<<
 *                     objects.append(obj)
 *                 return Literal(objects)
 */
          ((struct __pyx_vtabstruct_7flitter_5model_Node *)((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj)->__pyx_vtab)->append_vector(((struct __pyx_obj_7flitter_5model_Node *)__pyx_v_obj), __pyx_v_children_vector, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1118, __pyx_L1_error)

          /* "flitter/language/tree.pyx":1116
 *                 for i in range(nodes.length):
 *                     obj = node_objects[i]
 *                     if isinstance(obj, Node):             # <<<<<<<<<<<<<<
 *                         obj = (<Node>obj).copy()
 *                         (<Node>obj).append_vector(children_vector)
 */
        }

        /* "flitter/language/tree.pyx":1119
 *                         obj = (<Node>obj).copy()
 *                         (<Node>obj).append_vector(children_vector)
 *                     objects.append(obj)             # <<<<<<<<<<<<<<
 *                 return Literal(objects)
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):
 */
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_objects, __pyx_v_obj); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1119, __pyx_L1_error)
      }

      /* "flitter/language/tree.pyx":1120
 *                         (<Node>obj).append_vector(children_vector)
 *                     objects.append(obj)
 *                 return Literal(objects)             # <<<<<<<<<<<<<<
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):
 *                 node = Append(node, (<Sequence>children).expressions[0])
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_v_objects); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":1110
 *         if isinstance(node, Literal):
 *             nodes = (<Literal>node).value
 *             if isinstance(children, Literal):             # <<<<<<<<<<<<<<
 *                 children_vector = (<Literal>children).value
 *                 node_objects = nodes.objects
 */
    }

    /* "flitter/language/tree.pyx":1121
 *                     objects.append(obj)
 *                 return Literal(objects)
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):             # <<<<<<<<<<<<<<
 *                 node = Append(node, (<Sequence>children).expressions[0])
 *                 children = Sequence((<Sequence>children).expressions[1:])
 */
    __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_children), __pyx_ptype_7flitter_8language_4tree_Sequence); 
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L13_bool_binop_done;
    }
    if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1121, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_L13_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1122
 *                 return Literal(objects)
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):
 *                 node = Append(node, (<Sequence>children).expressions[0])             # <<<<<<<<<<<<<<
 *                 children = Sequence((<Sequence>children).expressions[1:])
 *                 return Append(node, children)._simplify(context)
 */
      if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1122, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_Tuple(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF((PyObject *)__pyx_v_node);
      __Pyx_GIVEREF((PyObject *)__pyx_v_node);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_v_node))) __PYX_ERR(0, 1122, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1123
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):
 *                 node = Append(node, (<Sequence>children).expressions[0])
 *                 children = Sequence((<Sequence>children).expressions[1:])             # <<<<<<<<<<<<<<
 *                 return Append(node, children)._simplify(context)
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):
 */
      if (unlikely(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1123, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyTuple_GetSlice(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v_children)->expressions, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence), __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_children, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "flitter/language/tree.pyx":1124
 *                 node = Append(node, (<Sequence>children).expressions[0])
 *                 children = Sequence((<Sequence>children).expressions[1:])
 *                 return Append(node, children)._simplify(context)             # <<<<<<<<<<<<<<
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF((PyObject *)__pyx_v_node);
      __Pyx_GIVEREF((PyObject *)__pyx_v_node);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_v_node))) __PYX_ERR(0, 1124, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_children);
      __Pyx_GIVEREF((PyObject *)__pyx_v_children);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_children))) __PYX_ERR(0, 1124, __pyx_L1_error);
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Append *)((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_t_1)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1), __pyx_v_context)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_8);
      __pyx_t_8 = 0;
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":1121
 *                     objects.append(obj)
 *                 return Literal(objects)
 *             elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):             # <<<<<<<<<<<<<<
 *                 node = Append(node, (<Sequence>children).expressions[0])
 *                 children = Sequence((<Sequence>children).expressions[1:])
 */
    }

    /* "flitter/language/tree.pyx":1108
 *                 (isinstance(children, Literal) and (<Literal>children).value.objects is None):
 *             return node
 *         if isinstance(node, Literal):             # <<<<<<<<<<<<<<
 *             nodes = (<Literal>node).value
 *             if isinstance(children, Literal):
 */
    goto __pyx_L8;
  }

  /* "flitter/language/tree.pyx":1125
 *                 children = Sequence((<Sequence>children).expressions[1:])
 *                 return Append(node, children)._simplify(context)
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):             # <<<<<<<<<<<<<<
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 *         if node is self.node and children is self.children:
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_8language_4tree_Attributes); 
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_8 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->__pyx_base.node);
  __Pyx_INCREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_t_8, __pyx_ptype_7flitter_8language_4tree_Literal); 
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1126
 *                 return Append(node, children)._simplify(context)
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)             # <<<<<<<<<<<<<<
 *         if node is self.node and children is self.children:
 *             return self
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->__pyx_base.node);
    __Pyx_GIVEREF((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->__pyx_base.node);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->__pyx_base.node))) __PYX_ERR(0, 1126, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_children);
    __Pyx_GIVEREF((PyObject *)__pyx_v_children);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_children))) __PYX_ERR(0, 1126, __pyx_L1_error);
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error);
    __Pyx_INCREF(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->bindings);
    __Pyx_GIVEREF(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->bindings);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v_node)->bindings)) __PYX_ERR(0, 1126, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Attributes), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Attributes *)((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_t_1)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._simplify(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1), __pyx_v_context)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_8);
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1125
 *                 children = Sequence((<Sequence>children).expressions[1:])
 *                 return Append(node, children)._simplify(context)
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):             # <<<<<<<<<<<<<<
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 *         if node is self.node and children is self.children:
 */
  }
  __pyx_L8:;

  /* "flitter/language/tree.pyx":1127
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 *         if node is self.node and children is self.children:             # <<<<<<<<<<<<<<
 *             return self
 *         return Append(node, children)
 */
  __pyx_t_3 = (__pyx_v_node == __pyx_v_self->__pyx_base.node);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L18_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_children == __pyx_v_self->children);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L18_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1128
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 *         if node is self.node and children is self.children:
 *             return self             # <<<<<<<<<<<<<<
 *         return Append(node, children)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1127
 *         elif isinstance(node, Attributes) and isinstance((<Attributes>node).node, Literal):
 *             return Attributes(Append((<Attributes>node).node, children), (<Attributes>node).bindings)._simplify(context)
 *         if node is self.node and children is self.children:             # <<<<<<<<<<<<<<
 *             return self
 *         return Append(node, children)
 */
  }

  /* "flitter/language/tree.pyx":1129
 *         if node is self.node and children is self.children:
 *             return self
 *         return Append(node, children)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF((PyObject *)__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_v_node))) __PYX_ERR(0, 1129, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_children);
  __Pyx_GIVEREF((PyObject *)__pyx_v_children);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_children))) __PYX_ERR(0, 1129, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append), __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1098
 *             program.append()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression node = self.node._simplify(context)
 *         cdef Expression children = self.children._simplify(context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.language.tree.Append._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF((PyObject *)__pyx_v_children);
  __Pyx_XDECREF((PyObject *)__pyx_v_nodes);
  __Pyx_XDECREF((PyObject *)__pyx_v_children_vector);
  __Pyx_XDECREF(__pyx_v_objects);
  __Pyx_XDECREF(__pyx_v_node_objects);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1131
 *         return Append(node, children)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.node!r}, {self.children!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Append_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1132
 * 
 *     def __repr__(self):
 *         return f'{self.__class__.__name__}({self.node!r}, {self.children!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->__pyx_base.node)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->children)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1131
 *         return Append(node, children)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'{self.__class__.__name__}({self.node!r}, {self.children!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Append.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1080
 * 
 * cdef class Append(NodeModifier):
 *     cdef readonly Expression children             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression node, Expression children):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_8children_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_8children_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Append_8children___get__(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_8children___get__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->children);
  __pyx_r = ((PyObject *)__pyx_v_self->children);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Append_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Append_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Append_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.children, self.node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->children);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->children);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->children))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.node);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.node);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.node))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.children, self.node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.children, self.node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.children is not None or self.node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.children, self.node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.children is not None or self.node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->children) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.node) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.children is not None or self.node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Append); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_217770541);
    __Pyx_GIVEREF(__pyx_int_217770541);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_217770541)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, None), state
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Append); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_217770541);
    __Pyx_GIVEREF(__pyx_int_217770541);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_217770541)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Append.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6Append_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Append_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6Append_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Append.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6Append_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6Append_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Append__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Append.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1139
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, str name, Expression expr):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.expr = expr
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_7Binding_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_7Binding_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_expr,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expr)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1139, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1139, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Binding.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1139, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "expr", 0))) __PYX_ERR(0, 1139, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding___init__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self), __pyx_v_name, __pyx_v_expr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_7Binding___init__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1140
 * 
 *     def __init__(self, str name, Expression expr):
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.expr = expr
 * 
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "flitter/language/tree.pyx":1141
 *     def __init__(self, str name, Expression expr):
 *         self.name = name
 *         self.expr = expr             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->expr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->expr);
  __pyx_v_self->expr = __pyx_v_expr;

  /* "flitter/language/tree.pyx":1139
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, str name, Expression expr):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.expr = expr
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1143
 *         self.expr = expr
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Binding({self.name!r}, {self.expr!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1144
 * 
 *     def __repr__(self):
 *         return f'Binding({self.name!r}, {self.expr!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Binding);
  __pyx_t_2 += 8;
  __Pyx_GIVEREF(__pyx_kp_u_Binding);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Binding);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->name), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->expr)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1143
 *         self.expr = expr
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Binding({self.name!r}, {self.expr!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Binding.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1136
 * 
 * cdef class Binding:
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding_4name___get__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1137
 * cdef class Binding:
 *     cdef readonly str name
 *     cdef readonly Expression expr             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str name, Expression expr):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_4expr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_4expr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding_4expr___get__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __pyx_r = ((PyObject *)__pyx_v_self->expr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7Binding_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Binding_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr, self.name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->name)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr, self.name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None or self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None or self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->expr) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None or self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Binding); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_60997603);
    __Pyx_GIVEREF(__pyx_int_60997603);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_60997603)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, None), state
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Binding); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_60997603);
    __Pyx_GIVEREF(__pyx_int_60997603);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_60997603)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Binding.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7Binding_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Binding_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7Binding_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Binding.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_7Binding_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_7Binding_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Binding__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Binding.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1151
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, tuple names, Expression expr):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.expr = expr
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_11PolyBinding_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_11PolyBinding_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_names = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_names,&__pyx_n_s_expr,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1151, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expr)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1151, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1151, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1151, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_names = ((PyObject*)values[0]);
    __pyx_v_expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1151, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.PolyBinding.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 1151, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "expr", 0))) __PYX_ERR(0, 1151, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding___init__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self), __pyx_v_names, __pyx_v_expr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_11PolyBinding___init__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1152
 * 
 *     def __init__(self, tuple names, Expression expr):
 *         self.names = names             # <<<<<<<<<<<<<<
 *         self.expr = expr
 * 
 */
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = __pyx_v_names;

  /* "flitter/language/tree.pyx":1153
 *     def __init__(self, tuple names, Expression expr):
 *         self.names = names
 *         self.expr = expr             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->expr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->expr);
  __pyx_v_self->expr = __pyx_v_expr;

  /* "flitter/language/tree.pyx":1151
 *     cdef readonly Expression expr
 * 
 *     def __init__(self, tuple names, Expression expr):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.expr = expr
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1155
 *         self.expr = expr
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'PolyBinding({self.names!r}, {self.expr!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_2__repr__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1156
 * 
 *     def __repr__(self):
 *         return f'PolyBinding({self.names!r}, {self.expr!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_PolyBinding);
  __pyx_t_2 += 12;
  __Pyx_GIVEREF(__pyx_kp_u_PolyBinding);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_PolyBinding);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->names), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->expr)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1155
 *         self.expr = expr
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'PolyBinding({self.names!r}, {self.expr!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.PolyBinding.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1148
 * 
 * cdef class PolyBinding:
 *     cdef readonly tuple names             # <<<<<<<<<<<<<<
 *     cdef readonly Expression expr
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_5names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_5names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding_5names___get__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_5names___get__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->names);
  __pyx_r = __pyx_v_self->names;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1149
 * cdef class PolyBinding:
 *     cdef readonly tuple names
 *     cdef readonly Expression expr             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple names, Expression expr):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_4expr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_4expr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding_4expr___get__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_4expr___get__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __pyx_r = ((PyObject *)__pyx_v_self->expr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.expr, self.names)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->expr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->expr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->expr))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->names);
  __Pyx_GIVEREF(__pyx_v_self->names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->names)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.expr, self.names)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.expr, self.names)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.expr is not None or self.names is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.expr, self.names)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.expr is not None or self.names is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->expr) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->names != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.names is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.expr is not None or self.names is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PolyBinding); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19657503);
    __Pyx_GIVEREF(__pyx_int_19657503);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_19657503)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.expr is not None or self.names is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, None), state
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_PolyBinding); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_19657503);
    __Pyx_GIVEREF(__pyx_int_19657503);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_19657503)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.PolyBinding.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.PolyBinding.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11PolyBinding_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11PolyBinding_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_PolyBinding__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.PolyBinding.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1163
 *     cdef readonly Expression body
 * 
 *     def __init__(self, tuple bindings, Expression body):             # <<<<<<<<<<<<<<
 *         self.bindings = bindings
 *         self.body = body
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_3Let_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_3Let_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bindings = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bindings,&__pyx_n_s_body,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bindings)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1163, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_body)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1163, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1163, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1163, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_bindings = ((PyObject*)values[0]);
    __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1163, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Let.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bindings), (&PyTuple_Type), 1, "bindings", 1))) __PYX_ERR(0, 1163, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_body), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "body", 0))) __PYX_ERR(0, 1163, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let___init__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self), __pyx_v_bindings, __pyx_v_body);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_3Let___init__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, PyObject *__pyx_v_bindings, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1164
 * 
 *     def __init__(self, tuple bindings, Expression body):
 *         self.bindings = bindings             # <<<<<<<<<<<<<<
 *         self.body = body
 * 
 */
  __Pyx_INCREF(__pyx_v_bindings);
  __Pyx_GIVEREF(__pyx_v_bindings);
  __Pyx_GOTREF(__pyx_v_self->bindings);
  __Pyx_DECREF(__pyx_v_self->bindings);
  __pyx_v_self->bindings = __pyx_v_bindings;

  /* "flitter/language/tree.pyx":1165
 *     def __init__(self, tuple bindings, Expression body):
 *         self.bindings = bindings
 *         self.body = body             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->body);
  __Pyx_DECREF((PyObject *)__pyx_v_self->body);
  __pyx_v_self->body = __pyx_v_body;

  /* "flitter/language/tree.pyx":1163
 *     cdef readonly Expression body
 * 
 *     def __init__(self, tuple bindings, Expression body):             # <<<<<<<<<<<<<<
 *         self.bindings = bindings
 *         self.body = body
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1167
 *         self.body = body
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef PolyBinding binding
 *         cdef int64_t n=len(lnames)
 */

static void __pyx_f_7flitter_8language_4tree_3Let__compile(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_binding = 0;
  int64_t __pyx_v_n;
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_instr = 0;
  struct __pyx_obj_7flitter_8language_2vm_Instruction *__pyx_v_compose = 0;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1169
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef PolyBinding binding
 *         cdef int64_t n=len(lnames)             # <<<<<<<<<<<<<<
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1169, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1169, __pyx_L1_error)
  __pyx_v_n = __pyx_t_1;

  /* "flitter/language/tree.pyx":1170
 *         cdef PolyBinding binding
 *         cdef int64_t n=len(lnames)
 *         for binding in self.bindings:             # <<<<<<<<<<<<<<
 *             binding.expr._compile(program, lnames)
 *             program.local_push(len(binding.names))
 */
  if (unlikely(__pyx_v_self->bindings == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1170, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->bindings; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1170, __pyx_L1_error)
      #endif
      if (__pyx_t_1 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(0, 1170, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_PolyBinding))))) __PYX_ERR(0, 1170, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1171
 *         cdef int64_t n=len(lnames)
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.local_push(len(binding.names))
 *             lnames.extend(binding.names)
 */
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_binding->expr->__pyx_vtab)->_compile(__pyx_v_binding->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1171, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1172
 *         for binding in self.bindings:
 *             binding.expr._compile(program, lnames)
 *             program.local_push(len(binding.names))             # <<<<<<<<<<<<<<
 *             lnames.extend(binding.names)
 *         self.body._compile(program, lnames)
 */
    __pyx_t_3 = __pyx_v_binding->names;
    __Pyx_INCREF(__pyx_t_3);
    if (unlikely(__pyx_t_3 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1172, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1172, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_push(__pyx_v_program, __pyx_t_4, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1173
 *             binding.expr._compile(program, lnames)
 *             program.local_push(len(binding.names))
 *             lnames.extend(binding.names)             # <<<<<<<<<<<<<<
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr, compose=None
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 1173, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_v_binding->names;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyList_Extend(__pyx_v_lnames, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1170
 *         cdef PolyBinding binding
 *         cdef int64_t n=len(lnames)
 *         for binding in self.bindings:             # <<<<<<<<<<<<<<
 *             binding.expr._compile(program, lnames)
 *             program.local_push(len(binding.names))
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":1174
 *             program.local_push(len(binding.names))
 *             lnames.extend(binding.names)
 *         self.body._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef Instruction instr, compose=None
 *         if program.last_instruction().code == OpCode.Compose:
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_compile(__pyx_v_self->body, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1174, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1175
 *             lnames.extend(binding.names)
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr, compose=None             # <<<<<<<<<<<<<<
 *         if program.last_instruction().code == OpCode.Compose:
 *             compose = program.pop_instruction()
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_compose = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)Py_None);

  /* "flitter/language/tree.pyx":1176
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr, compose=None
 *         if program.last_instruction().code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             compose = program.pop_instruction()
 *         instr = program.last_instruction()
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = (((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_2)->code == __pyx_e_7flitter_8language_2vm_Compose);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":1177
 *         cdef Instruction instr, compose=None
 *         if program.last_instruction().code == OpCode.Compose:
 *             compose = program.pop_instruction()             # <<<<<<<<<<<<<<
 *         instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_compose, ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":1176
 *         self.body._compile(program, lnames)
 *         cdef Instruction instr, compose=None
 *         if program.last_instruction().code == OpCode.Compose:             # <<<<<<<<<<<<<<
 *             compose = program.pop_instruction()
 *         instr = program.last_instruction()
 */
  }

  /* "flitter/language/tree.pyx":1178
 *         if program.last_instruction().code == OpCode.Compose:
 *             compose = program.pop_instruction()
 *         instr = program.last_instruction()             # <<<<<<<<<<<<<<
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->last_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_instr = ((struct __pyx_obj_7flitter_8language_2vm_Instruction *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":1179
 *             compose = program.pop_instruction()
 *         instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(lnames) - n)
 */
  __pyx_t_6 = (__pyx_v_instr->code == __pyx_e_7flitter_8language_2vm_LocalDrop);
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":1180
 *         instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()             # <<<<<<<<<<<<<<
 *             program.local_drop((<InstructionInt>instr).value + len(lnames) - n)
 *         else:
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->pop_instruction(__pyx_v_program)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":1181
 *         if instr.code == OpCode.LocalDrop:
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(lnames) - n)             # <<<<<<<<<<<<<<
 *         else:
 *             program.local_drop(len(lnames) - n)
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1181, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1181, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_drop(__pyx_v_program, ((((struct __pyx_obj_7flitter_8language_2vm_InstructionInt *)__pyx_v_instr)->value + __pyx_t_1) - __pyx_v_n), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":1179
 *             compose = program.pop_instruction()
 *         instr = program.last_instruction()
 *         if instr.code == OpCode.LocalDrop:             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.local_drop((<InstructionInt>instr).value + len(lnames) - n)
 */
    goto __pyx_L7;
  }

  /* "flitter/language/tree.pyx":1183
 *             program.local_drop((<InstructionInt>instr).value + len(lnames) - n)
 *         else:
 *             program.local_drop(len(lnames) - n)             # <<<<<<<<<<<<<<
 *         if compose is not None:
 *             program.push_instruction(compose)
 */
  /*else*/ {
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1183, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1183, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_drop(__pyx_v_program, (__pyx_t_1 - __pyx_v_n), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L7:;

  /* "flitter/language/tree.pyx":1184
 *         else:
 *             program.local_drop(len(lnames) - n)
 *         if compose is not None:             # <<<<<<<<<<<<<<
 *             program.push_instruction(compose)
 *         while len(lnames) > n:
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_compose) != Py_None);
  if (__pyx_t_6) {

    /* "flitter/language/tree.pyx":1185
 *             program.local_drop(len(lnames) - n)
 *         if compose is not None:
 *             program.push_instruction(compose)             # <<<<<<<<<<<<<<
 *         while len(lnames) > n:
 *             lnames.pop()
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->push_instruction(__pyx_v_program, __pyx_v_compose)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flitter/language/tree.pyx":1184
 *         else:
 *             program.local_drop(len(lnames) - n)
 *         if compose is not None:             # <<<<<<<<<<<<<<
 *             program.push_instruction(compose)
 *         while len(lnames) > n:
 */
  }

  /* "flitter/language/tree.pyx":1186
 *         if compose is not None:
 *             program.push_instruction(compose)
 *         while len(lnames) > n:             # <<<<<<<<<<<<<<
 *             lnames.pop()
 * 
 */
  while (1) {
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1186, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1186, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_1 > __pyx_v_n);
    if (!__pyx_t_6) break;

    /* "flitter/language/tree.pyx":1187
 *             program.push_instruction(compose)
 *         while len(lnames) > n:
 *             lnames.pop()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(0, 1187, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyList_Pop(__pyx_v_lnames); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "flitter/language/tree.pyx":1167
 *         self.body = body
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef PolyBinding binding
 *         cdef int64_t n=len(lnames)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Let._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF((PyObject *)__pyx_v_instr);
  __Pyx_XDECREF((PyObject *)__pyx_v_compose);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1189
 *             lnames.pop()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef dict saved = context.names
 *         context.names = saved.copy()
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3Let__simplify(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_saved = 0;
  PyObject *__pyx_v_bindings = 0;
  PyObject *__pyx_v_remaining = 0;
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v_binding = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_expr = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_value = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_existing_name = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_j;
  int64_t __pyx_v_n;
  PyObject *__pyx_v_touched = 0;
  PyObject *__pyx_v_shadowed = 0;
  PyObject *__pyx_v_discarded = 0;
  PyObject *__pyx_v_existing_value = NULL;
  int __pyx_v_resimplify;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_sbody = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int64_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1190
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef dict saved = context.names             # <<<<<<<<<<<<<<
 *         context.names = saved.copy()
 *         cdef list bindings = list(self.bindings)
 */
  __pyx_t_1 = __pyx_v_context->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_saved = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1191
 *     cdef Expression _simplify(self, Context context):
 *         cdef dict saved = context.names
 *         context.names = saved.copy()             # <<<<<<<<<<<<<<
 *         cdef list bindings = list(self.bindings)
 *         cdef list remaining = []
 */
  if (unlikely(__pyx_v_saved == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(0, 1191, __pyx_L1_error)
  }
  __pyx_t_1 = PyDict_Copy(__pyx_v_saved); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1192
 *         cdef dict saved = context.names
 *         context.names = saved.copy()
 *         cdef list bindings = list(self.bindings)             # <<<<<<<<<<<<<<
 *         cdef list remaining = []
 *         cdef PolyBinding binding
 */
  __pyx_t_1 = PySequence_List(__pyx_v_self->bindings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bindings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1193
 *         context.names = saved.copy()
 *         cdef list bindings = list(self.bindings)
 *         cdef list remaining = []             # <<<<<<<<<<<<<<
 *         cdef PolyBinding binding
 *         cdef Expression expr, body=self.body
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_remaining = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1195
 *         cdef list remaining = []
 *         cdef PolyBinding binding
 *         cdef Expression expr, body=self.body             # <<<<<<<<<<<<<<
 *         cdef Vector value
 *         cdef str name, existing_name
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->body);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1199
 *         cdef str name, existing_name
 *         cdef int64_t i, j, n
 *         cdef touched = False             # <<<<<<<<<<<<<<
 *         cdef set shadowed=set(), discarded=set()
 *         while isinstance(body, Let):
 */
  __Pyx_INCREF(Py_False);
  __pyx_v_touched = Py_False;

  /* "flitter/language/tree.pyx":1200
 *         cdef int64_t i, j, n
 *         cdef touched = False
 *         cdef set shadowed=set(), discarded=set()             # <<<<<<<<<<<<<<
 *         while isinstance(body, Let):
 *             bindings.extend((<Let>body).bindings)
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shadowed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_discarded = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1201
 *         cdef touched = False
 *         cdef set shadowed=set(), discarded=set()
 *         while isinstance(body, Let):             # <<<<<<<<<<<<<<
 *             bindings.extend((<Let>body).bindings)
 *             body = (<Let>body).body
 */
  while (1) {
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_body), __pyx_ptype_7flitter_8language_4tree_Let); 
    if (!__pyx_t_2) break;

    /* "flitter/language/tree.pyx":1202
 *         cdef set shadowed=set(), discarded=set()
 *         while isinstance(body, Let):
 *             bindings.extend((<Let>body).bindings)             # <<<<<<<<<<<<<<
 *             body = (<Let>body).body
 *             touched = True
 */
    __pyx_t_1 = ((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_body)->bindings;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_bindings, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1203
 *         while isinstance(body, Let):
 *             bindings.extend((<Let>body).bindings)
 *             body = (<Let>body).body             # <<<<<<<<<<<<<<
 *             touched = True
 *         for i, binding in enumerate(bindings):
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_body)->body);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_body, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1204
 *             bindings.extend((<Let>body).bindings)
 *             body = (<Let>body).body
 *             touched = True             # <<<<<<<<<<<<<<
 *         for i, binding in enumerate(bindings):
 *             for existing_name in list(context.names):
 */
    __Pyx_INCREF(Py_True);
    __Pyx_DECREF_SET(__pyx_v_touched, Py_True);
  }

  /* "flitter/language/tree.pyx":1205
 *             body = (<Let>body).body
 *             touched = True
 *         for i, binding in enumerate(bindings):             # <<<<<<<<<<<<<<
 *             for existing_name in list(context.names):
 *                 existing_value = context.names[existing_name]
 */
  __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_v_bindings; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1205, __pyx_L1_error)
      #endif
      if (__pyx_t_5 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 1205, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_8language_4tree_PolyBinding))))) __PYX_ERR(0, 1205, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_binding, ((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_t_6));
    __pyx_t_6 = 0;
    __pyx_v_i = __pyx_t_4;
    __pyx_t_4 = (__pyx_t_4 + 1);

    /* "flitter/language/tree.pyx":1206
 *             touched = True
 *         for i, binding in enumerate(bindings):
 *             for existing_name in list(context.names):             # <<<<<<<<<<<<<<
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):
 */
    __pyx_t_6 = PySequence_List(__pyx_v_context->names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __pyx_t_6; __Pyx_INCREF(__pyx_t_7);
    __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
      #else
      __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_6))) __PYX_ERR(0, 1206, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_existing_name, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":1207
 *         for i, binding in enumerate(bindings):
 *             for existing_name in list(context.names):
 *                 existing_value = context.names[existing_name]             # <<<<<<<<<<<<<<
 *                 if existing_value is not None and isinstance(existing_value, Name):
 *                     for name in binding.names:
 */
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1207, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_context->names, __pyx_v_existing_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_existing_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":1208
 *             for existing_name in list(context.names):
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):             # <<<<<<<<<<<<<<
 *                     for name in binding.names:
 *                         if name == (<Name>existing_value).name:
 */
      __pyx_t_9 = (__pyx_v_existing_value != Py_None);
      if (__pyx_t_9) {
      } else {
        __pyx_t_2 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_9 = __Pyx_TypeCheck(__pyx_v_existing_value, __pyx_ptype_7flitter_8language_4tree_Name); 
      __pyx_t_2 = __pyx_t_9;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":1209
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):
 *                     for name in binding.names:             # <<<<<<<<<<<<<<
 *                         if name == (<Name>existing_value).name:
 *                             context.names[existing_name] = None
 */
        if (unlikely(__pyx_v_binding->names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 1209, __pyx_L1_error)
        }
        __pyx_t_6 = __pyx_v_binding->names; __Pyx_INCREF(__pyx_t_6);
        __pyx_t_10 = 0;
        for (;;) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1209, __pyx_L1_error)
            #endif
            if (__pyx_t_10 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_11); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 1209, __pyx_L1_error)
          #else
          __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
          if (!(likely(PyUnicode_CheckExact(__pyx_t_11))||((__pyx_t_11) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_11))) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_11));
          __pyx_t_11 = 0;

          /* "flitter/language/tree.pyx":1210
 *                 if existing_value is not None and isinstance(existing_value, Name):
 *                     for name in binding.names:
 *                         if name == (<Name>existing_value).name:             # <<<<<<<<<<<<<<
 *                             context.names[existing_name] = None
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))
 */
          __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_name, ((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_existing_value)->name, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1210, __pyx_L1_error)
          if (__pyx_t_2) {

            /* "flitter/language/tree.pyx":1211
 *                     for name in binding.names:
 *                         if name == (<Name>existing_value).name:
 *                             context.names[existing_name] = None             # <<<<<<<<<<<<<<
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))
 *                             shadowed.add(name)
 */
            if (unlikely(__pyx_v_context->names == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 1211, __pyx_L1_error)
            }
            if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_existing_name, Py_None) < 0))) __PYX_ERR(0, 1211, __pyx_L1_error)

            /* "flitter/language/tree.pyx":1212
 *                         if name == (<Name>existing_value).name:
 *                             context.names[existing_name] = None
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))             # <<<<<<<<<<<<<<
 *                             shadowed.add(name)
 *                             touched = True
 */
            __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_INCREF(__pyx_v_existing_name);
            __Pyx_GIVEREF(__pyx_v_existing_name);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_existing_name)) __PYX_ERR(0, 1212, __pyx_L1_error);
            __pyx_t_12 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Name), __pyx_v_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_11);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11)) __PYX_ERR(0, 1212, __pyx_L1_error);
            __Pyx_GIVEREF(__pyx_t_12);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_12)) __PYX_ERR(0, 1212, __pyx_L1_error);
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_12 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_13, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_t_12); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

            /* "flitter/language/tree.pyx":1213
 *                             context.names[existing_name] = None
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))
 *                             shadowed.add(name)             # <<<<<<<<<<<<<<
 *                             touched = True
 *             n = len(binding.names)
 */
            __pyx_t_3 = PySet_Add(__pyx_v_shadowed, __pyx_v_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1213, __pyx_L1_error)

            /* "flitter/language/tree.pyx":1214
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))
 *                             shadowed.add(name)
 *                             touched = True             # <<<<<<<<<<<<<<
 *             n = len(binding.names)
 *             expr = binding.expr._simplify(context)
 */
            __Pyx_INCREF(Py_True);
            __Pyx_DECREF_SET(__pyx_v_touched, Py_True);

            /* "flitter/language/tree.pyx":1210
 *                 if existing_value is not None and isinstance(existing_value, Name):
 *                     for name in binding.names:
 *                         if name == (<Name>existing_value).name:             # <<<<<<<<<<<<<<
 *                             context.names[existing_name] = None
 *                             remaining.append(PolyBinding((existing_name,), Name(name)))
 */
          }

          /* "flitter/language/tree.pyx":1209
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):
 *                     for name in binding.names:             # <<<<<<<<<<<<<<
 *                         if name == (<Name>existing_value).name:
 *                             context.names[existing_name] = None
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/language/tree.pyx":1208
 *             for existing_name in list(context.names):
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):             # <<<<<<<<<<<<<<
 *                     for name in binding.names:
 *                         if name == (<Name>existing_value).name:
 */
      }

      /* "flitter/language/tree.pyx":1206
 *             touched = True
 *         for i, binding in enumerate(bindings):
 *             for existing_name in list(context.names):             # <<<<<<<<<<<<<<
 *                 existing_value = context.names[existing_name]
 *                 if existing_value is not None and isinstance(existing_value, Name):
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1215
 *                             shadowed.add(name)
 *                             touched = True
 *             n = len(binding.names)             # <<<<<<<<<<<<<<
 *             expr = binding.expr._simplify(context)
 *             touched |= expr is not binding.expr
 */
    __pyx_t_7 = __pyx_v_binding->names;
    __Pyx_INCREF(__pyx_t_7);
    if (unlikely(__pyx_t_7 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1215, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyTuple_GET_SIZE(__pyx_t_7); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_n = __pyx_t_8;

    /* "flitter/language/tree.pyx":1216
 *                             touched = True
 *             n = len(binding.names)
 *             expr = binding.expr._simplify(context)             # <<<<<<<<<<<<<<
 *             touched |= expr is not binding.expr
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:
 */
    __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_binding->expr->__pyx_vtab)->_simplify(__pyx_v_binding->expr, __pyx_v_context)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_expr, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1217
 *             n = len(binding.names)
 *             expr = binding.expr._simplify(context)
 *             touched |= expr is not binding.expr             # <<<<<<<<<<<<<<
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:
 *                 context.names[binding.names[0]] = expr
 */
    __pyx_t_2 = (__pyx_v_expr != __pyx_v_binding->expr);
    __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PyNumber_InPlaceOr(__pyx_v_touched, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_touched, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "flitter/language/tree.pyx":1218
 *             expr = binding.expr._simplify(context)
 *             touched |= expr is not binding.expr
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:             # <<<<<<<<<<<<<<
 *                 context.names[binding.names[0]] = expr
 *                 remaining.append(PolyBinding(binding.names, expr))
 */
    __pyx_t_9 = (__pyx_v_n == 1);
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_9 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Function); 
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_2 = ((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_expr)->inlineable;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1219
 *             touched |= expr is not binding.expr
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:
 *                 context.names[binding.names[0]] = expr             # <<<<<<<<<<<<<<
 *                 remaining.append(PolyBinding(binding.names, expr))
 *                 continue
 */
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1219, __pyx_L1_error)
      }
      if (unlikely(__pyx_v_binding->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1219, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v_binding->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_t_6, ((PyObject *)__pyx_v_expr)) < 0))) __PYX_ERR(0, 1219, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":1220
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:
 *                 context.names[binding.names[0]] = expr
 *                 remaining.append(PolyBinding(binding.names, expr))             # <<<<<<<<<<<<<<
 *                 continue
 *             if isinstance(expr, Literal):
 */
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_binding->names);
      __Pyx_GIVEREF(__pyx_v_binding->names);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_binding->names)) __PYX_ERR(0, 1220, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_expr);
      __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 1220, __pyx_L1_error);
      __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_t_7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1220, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":1221
 *                 context.names[binding.names[0]] = expr
 *                 remaining.append(PolyBinding(binding.names, expr))
 *                 continue             # <<<<<<<<<<<<<<
 *             if isinstance(expr, Literal):
 *                 value = (<Literal>expr).value
 */
      goto __pyx_L5_continue;

      /* "flitter/language/tree.pyx":1218
 *             expr = binding.expr._simplify(context)
 *             touched |= expr is not binding.expr
 *             if n == 1 and isinstance(expr, Function) and (<Function>expr).inlineable:             # <<<<<<<<<<<<<<
 *                 context.names[binding.names[0]] = expr
 *                 remaining.append(PolyBinding(binding.names, expr))
 */
    }

    /* "flitter/language/tree.pyx":1222
 *                 remaining.append(PolyBinding(binding.names, expr))
 *                 continue
 *             if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *                 value = (<Literal>expr).value
 *                 if n == 1:
 */
    __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Literal); 
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1223
 *                 continue
 *             if isinstance(expr, Literal):
 *                 value = (<Literal>expr).value             # <<<<<<<<<<<<<<
 *                 if n == 1:
 *                     name = binding.names[0]
 */
      __pyx_t_7 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_expr)->value);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":1224
 *             if isinstance(expr, Literal):
 *                 value = (<Literal>expr).value
 *                 if n == 1:             # <<<<<<<<<<<<<<
 *                     name = binding.names[0]
 *                     context.names[name] = value
 */
      __pyx_t_2 = (__pyx_v_n == 1);
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":1225
 *                 value = (<Literal>expr).value
 *                 if n == 1:
 *                     name = binding.names[0]             # <<<<<<<<<<<<<<
 *                     context.names[name] = value
 *                     discarded.add(name)
 */
        if (unlikely(__pyx_v_binding->names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1225, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v_binding->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1225, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 1225, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "flitter/language/tree.pyx":1226
 *                 if n == 1:
 *                     name = binding.names[0]
 *                     context.names[name] = value             # <<<<<<<<<<<<<<
 *                     discarded.add(name)
 *                 else:
 */
        if (unlikely(__pyx_v_context->names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1226, __pyx_L1_error)
        }
        if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, ((PyObject *)__pyx_v_value)) < 0))) __PYX_ERR(0, 1226, __pyx_L1_error)

        /* "flitter/language/tree.pyx":1227
 *                     name = binding.names[0]
 *                     context.names[name] = value
 *                     discarded.add(name)             # <<<<<<<<<<<<<<
 *                 else:
 *                     for j, name in enumerate(binding.names):
 */
        __pyx_t_3 = PySet_Add(__pyx_v_discarded, __pyx_v_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1227, __pyx_L1_error)

        /* "flitter/language/tree.pyx":1224
 *             if isinstance(expr, Literal):
 *                 value = (<Literal>expr).value
 *                 if n == 1:             # <<<<<<<<<<<<<<
 *                     name = binding.names[0]
 *                     context.names[name] = value
 */
        goto __pyx_L22;
      }

      /* "flitter/language/tree.pyx":1229
 *                     discarded.add(name)
 *                 else:
 *                     for j, name in enumerate(binding.names):             # <<<<<<<<<<<<<<
 *                         context.names[name] = value.item(j)
 *                         discarded.add(name)
 */
      /*else*/ {
        __pyx_t_14 = 0;
        __pyx_t_7 = __pyx_v_binding->names; __Pyx_INCREF(__pyx_t_7);
        __pyx_t_8 = 0;
        for (;;) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1229, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1229, __pyx_L1_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1229, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_6))) __PYX_ERR(0, 1229, __pyx_L1_error)
          __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_6));
          __pyx_t_6 = 0;
          __pyx_v_j = __pyx_t_14;
          __pyx_t_14 = (__pyx_t_14 + 1);

          /* "flitter/language/tree.pyx":1230
 *                 else:
 *                     for j, name in enumerate(binding.names):
 *                         context.names[name] = value.item(j)             # <<<<<<<<<<<<<<
 *                         discarded.add(name)
 *                 touched = True
 */
          __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_value->__pyx_vtab)->item(__pyx_v_value, __pyx_v_j, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1230, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (unlikely(__pyx_v_context->names == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 1230, __pyx_L1_error)
          }
          if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, __pyx_t_6) < 0))) __PYX_ERR(0, 1230, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "flitter/language/tree.pyx":1231
 *                     for j, name in enumerate(binding.names):
 *                         context.names[name] = value.item(j)
 *                         discarded.add(name)             # <<<<<<<<<<<<<<
 *                 touched = True
 *                 continue
 */
          __pyx_t_3 = PySet_Add(__pyx_v_discarded, __pyx_v_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1231, __pyx_L1_error)

          /* "flitter/language/tree.pyx":1229
 *                     discarded.add(name)
 *                 else:
 *                     for j, name in enumerate(binding.names):             # <<<<<<<<<<<<<<
 *                         context.names[name] = value.item(j)
 *                         discarded.add(name)
 */
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_L22:;

      /* "flitter/language/tree.pyx":1232
 *                         context.names[name] = value.item(j)
 *                         discarded.add(name)
 *                 touched = True             # <<<<<<<<<<<<<<
 *                 continue
 *             if n == 1 and isinstance(expr, Name):
 */
      __Pyx_INCREF(Py_True);
      __Pyx_DECREF_SET(__pyx_v_touched, Py_True);

      /* "flitter/language/tree.pyx":1233
 *                         discarded.add(name)
 *                 touched = True
 *                 continue             # <<<<<<<<<<<<<<
 *             if n == 1 and isinstance(expr, Name):
 *                 name = binding.names[0]
 */
      goto __pyx_L5_continue;

      /* "flitter/language/tree.pyx":1222
 *                 remaining.append(PolyBinding(binding.names, expr))
 *                 continue
 *             if isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *                 value = (<Literal>expr).value
 *                 if n == 1:
 */
    }

    /* "flitter/language/tree.pyx":1234
 *                 touched = True
 *                 continue
 *             if n == 1 and isinstance(expr, Name):             # <<<<<<<<<<<<<<
 *                 name = binding.names[0]
 *                 if (<Name>expr).name == name:
 */
    __pyx_t_9 = (__pyx_v_n == 1);
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L27_bool_binop_done;
    }
    __pyx_t_9 = __Pyx_TypeCheck(((PyObject *)__pyx_v_expr), __pyx_ptype_7flitter_8language_4tree_Name); 
    __pyx_t_2 = __pyx_t_9;
    __pyx_L27_bool_binop_done:;
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1235
 *                 continue
 *             if n == 1 and isinstance(expr, Name):
 *                 name = binding.names[0]             # <<<<<<<<<<<<<<
 *                 if (<Name>expr).name == name:
 *                     touched = True
 */
      if (unlikely(__pyx_v_binding->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1235, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v_binding->names, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 1235, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":1236
 *             if n == 1 and isinstance(expr, Name):
 *                 name = binding.names[0]
 *                 if (<Name>expr).name == name:             # <<<<<<<<<<<<<<
 *                     touched = True
 *                     continue
 */
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_expr)->name, __pyx_v_name, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1236, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":1237
 *                 name = binding.names[0]
 *                 if (<Name>expr).name == name:
 *                     touched = True             # <<<<<<<<<<<<<<
 *                     continue
 *                 for j in range(i+1, len(bindings)):
 */
        __Pyx_INCREF(Py_True);
        __Pyx_DECREF_SET(__pyx_v_touched, Py_True);

        /* "flitter/language/tree.pyx":1238
 *                 if (<Name>expr).name == name:
 *                     touched = True
 *                     continue             # <<<<<<<<<<<<<<
 *                 for j in range(i+1, len(bindings)):
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:
 */
        goto __pyx_L5_continue;

        /* "flitter/language/tree.pyx":1236
 *             if n == 1 and isinstance(expr, Name):
 *                 name = binding.names[0]
 *                 if (<Name>expr).name == name:             # <<<<<<<<<<<<<<
 *                     touched = True
 *                     continue
 */
      }

      /* "flitter/language/tree.pyx":1239
 *                     touched = True
 *                     continue
 *                 for j in range(i+1, len(bindings)):             # <<<<<<<<<<<<<<
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:
 *                         context.names[name] = None
 */
      __pyx_t_8 = __Pyx_PyList_GET_SIZE(__pyx_v_bindings); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1239, __pyx_L1_error)
      __pyx_t_10 = __pyx_t_8;
      for (__pyx_t_14 = (__pyx_v_i + 1); __pyx_t_14 < __pyx_t_10; __pyx_t_14+=1) {
        __pyx_v_j = __pyx_t_14;

        /* "flitter/language/tree.pyx":1240
 *                     continue
 *                 for j in range(i+1, len(bindings)):
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:             # <<<<<<<<<<<<<<
 *                         context.names[name] = None
 *                         shadowed.add((<Name>expr).name)
 */
        __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_bindings, __pyx_v_j, int64_t, 1, __Pyx_PyInt_From_int64_t, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = (__Pyx_PySequence_ContainsTF(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_expr)->name, __pyx_t_6, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1240, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_2) {

          /* "flitter/language/tree.pyx":1241
 *                 for j in range(i+1, len(bindings)):
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:
 *                         context.names[name] = None             # <<<<<<<<<<<<<<
 *                         shadowed.add((<Name>expr).name)
 *                         break
 */
          if (unlikely(__pyx_v_context->names == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 1241, __pyx_L1_error)
          }
          if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, Py_None) < 0))) __PYX_ERR(0, 1241, __pyx_L1_error)

          /* "flitter/language/tree.pyx":1242
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:
 *                         context.names[name] = None
 *                         shadowed.add((<Name>expr).name)             # <<<<<<<<<<<<<<
 *                         break
 *                 else:
 */
          __pyx_t_6 = ((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v_expr)->name;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_3 = PySet_Add(__pyx_v_shadowed, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1242, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "flitter/language/tree.pyx":1243
 *                         context.names[name] = None
 *                         shadowed.add((<Name>expr).name)
 *                         break             # <<<<<<<<<<<<<<
 *                 else:
 *                     context.names[name] = expr
 */
          goto __pyx_L31_break;

          /* "flitter/language/tree.pyx":1240
 *                     continue
 *                 for j in range(i+1, len(bindings)):
 *                     if (<Name>expr).name in (<Binding>bindings[j]).names:             # <<<<<<<<<<<<<<
 *                         context.names[name] = None
 *                         shadowed.add((<Name>expr).name)
 */
        }
      }
      /*else*/ {

        /* "flitter/language/tree.pyx":1245
 *                         break
 *                 else:
 *                     context.names[name] = expr             # <<<<<<<<<<<<<<
 *                     discarded.add(name)
 *                     touched = True
 */
        if (unlikely(__pyx_v_context->names == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1245, __pyx_L1_error)
        }
        if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, ((PyObject *)__pyx_v_expr)) < 0))) __PYX_ERR(0, 1245, __pyx_L1_error)

        /* "flitter/language/tree.pyx":1246
 *                 else:
 *                     context.names[name] = expr
 *                     discarded.add(name)             # <<<<<<<<<<<<<<
 *                     touched = True
 *                     continue
 */
        __pyx_t_3 = PySet_Add(__pyx_v_discarded, __pyx_v_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1246, __pyx_L1_error)

        /* "flitter/language/tree.pyx":1247
 *                     context.names[name] = expr
 *                     discarded.add(name)
 *                     touched = True             # <<<<<<<<<<<<<<
 *                     continue
 *             for name in binding.names:
 */
        __Pyx_INCREF(Py_True);
        __Pyx_DECREF_SET(__pyx_v_touched, Py_True);

        /* "flitter/language/tree.pyx":1248
 *                     discarded.add(name)
 *                     touched = True
 *                     continue             # <<<<<<<<<<<<<<
 *             for name in binding.names:
 *                 context.names[name] = None
 */
        goto __pyx_L5_continue;
      }
      __pyx_L31_break:;

      /* "flitter/language/tree.pyx":1234
 *                 touched = True
 *                 continue
 *             if n == 1 and isinstance(expr, Name):             # <<<<<<<<<<<<<<
 *                 name = binding.names[0]
 *                 if (<Name>expr).name == name:
 */
    }

    /* "flitter/language/tree.pyx":1249
 *                     touched = True
 *                     continue
 *             for name in binding.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = None
 *             remaining.append(PolyBinding(binding.names, expr))
 */
    if (unlikely(__pyx_v_binding->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1249, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_binding->names; __Pyx_INCREF(__pyx_t_6);
    __pyx_t_8 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1249, __pyx_L1_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1249, __pyx_L1_error)
      #else
      __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1249, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 1249, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "flitter/language/tree.pyx":1250
 *                     continue
 *             for name in binding.names:
 *                 context.names[name] = None             # <<<<<<<<<<<<<<
 *             remaining.append(PolyBinding(binding.names, expr))
 *         cdef bint resimplify = shadowed and shadowed & discarded
 */
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1250, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, Py_None) < 0))) __PYX_ERR(0, 1250, __pyx_L1_error)

      /* "flitter/language/tree.pyx":1249
 *                     touched = True
 *                     continue
 *             for name in binding.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = None
 *             remaining.append(PolyBinding(binding.names, expr))
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flitter/language/tree.pyx":1251
 *             for name in binding.names:
 *                 context.names[name] = None
 *             remaining.append(PolyBinding(binding.names, expr))             # <<<<<<<<<<<<<<
 *         cdef bint resimplify = shadowed and shadowed & discarded
 *         cdef Expression sbody = body._simplify(context)
 */
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_binding->names);
    __Pyx_GIVEREF(__pyx_v_binding->names);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_binding->names)) __PYX_ERR(0, 1251, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_expr);
    __Pyx_GIVEREF((PyObject *)__pyx_v_expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_expr))) __PYX_ERR(0, 1251, __pyx_L1_error);
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_t_7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1205
 *             body = (<Let>body).body
 *             touched = True
 *         for i, binding in enumerate(bindings):             # <<<<<<<<<<<<<<
 *             for existing_name in list(context.names):
 *                 existing_value = context.names[existing_name]
 */
    __pyx_L5_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1252
 *                 context.names[name] = None
 *             remaining.append(PolyBinding(binding.names, expr))
 *         cdef bint resimplify = shadowed and shadowed & discarded             # <<<<<<<<<<<<<<
 *         cdef Expression sbody = body._simplify(context)
 *         touched |= sbody is not body
 */
  __pyx_t_9 = (PySet_GET_SIZE(__pyx_v_shadowed) != 0);
  if (__pyx_t_9) {
  } else {
    __pyx_t_2 = __pyx_t_9;
    goto __pyx_L37_bool_binop_done;
  }
  __pyx_t_1 = PyNumber_And(__pyx_v_shadowed, __pyx_v_discarded); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = (PySet_GET_SIZE(__pyx_t_1) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_9;
  __pyx_L37_bool_binop_done:;
  __pyx_v_resimplify = __pyx_t_2;

  /* "flitter/language/tree.pyx":1253
 *             remaining.append(PolyBinding(binding.names, expr))
 *         cdef bint resimplify = shadowed and shadowed & discarded
 *         cdef Expression sbody = body._simplify(context)             # <<<<<<<<<<<<<<
 *         touched |= sbody is not body
 *         context.names = saved
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_body->__pyx_vtab)->_simplify(__pyx_v_body, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sbody = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1254
 *         cdef bint resimplify = shadowed and shadowed & discarded
 *         cdef Expression sbody = body._simplify(context)
 *         touched |= sbody is not body             # <<<<<<<<<<<<<<
 *         context.names = saved
 *         if isinstance(sbody, Literal):
 */
  __pyx_t_2 = (__pyx_v_sbody != __pyx_v_body);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyNumber_InPlaceOr(__pyx_v_touched, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_touched, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "flitter/language/tree.pyx":1255
 *         cdef Expression sbody = body._simplify(context)
 *         touched |= sbody is not body
 *         context.names = saved             # <<<<<<<<<<<<<<
 *         if isinstance(sbody, Literal):
 *             return sbody
 */
  __Pyx_INCREF(__pyx_v_saved);
  __Pyx_GIVEREF(__pyx_v_saved);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = __pyx_v_saved;

  /* "flitter/language/tree.pyx":1256
 *         touched |= sbody is not body
 *         context.names = saved
 *         if isinstance(sbody, Literal):             # <<<<<<<<<<<<<<
 *             return sbody
 *         if isinstance(sbody, Let):
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_sbody), __pyx_ptype_7flitter_8language_4tree_Literal); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1257
 *         context.names = saved
 *         if isinstance(sbody, Literal):
 *             return sbody             # <<<<<<<<<<<<<<
 *         if isinstance(sbody, Let):
 *             remaining.extend((<Let>sbody).bindings)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_sbody);
    __pyx_r = __pyx_v_sbody;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1256
 *         touched |= sbody is not body
 *         context.names = saved
 *         if isinstance(sbody, Literal):             # <<<<<<<<<<<<<<
 *             return sbody
 *         if isinstance(sbody, Let):
 */
  }

  /* "flitter/language/tree.pyx":1258
 *         if isinstance(sbody, Literal):
 *             return sbody
 *         if isinstance(sbody, Let):             # <<<<<<<<<<<<<<
 *             remaining.extend((<Let>sbody).bindings)
 *             sbody = (<Let>sbody).body
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_sbody), __pyx_ptype_7flitter_8language_4tree_Let); 
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1259
 *             return sbody
 *         if isinstance(sbody, Let):
 *             remaining.extend((<Let>sbody).bindings)             # <<<<<<<<<<<<<<
 *             sbody = (<Let>sbody).body
 *             resimplify = True
 */
    __pyx_t_7 = ((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_sbody)->bindings;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_remaining, __pyx_t_7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1259, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1260
 *         if isinstance(sbody, Let):
 *             remaining.extend((<Let>sbody).bindings)
 *             sbody = (<Let>sbody).body             # <<<<<<<<<<<<<<
 *             resimplify = True
 *             touched = True
 */
    __pyx_t_7 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_sbody)->body);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_sbody, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1261
 *             remaining.extend((<Let>sbody).bindings)
 *             sbody = (<Let>sbody).body
 *             resimplify = True             # <<<<<<<<<<<<<<
 *             touched = True
 *         if not touched:
 */
    __pyx_v_resimplify = 1;

    /* "flitter/language/tree.pyx":1262
 *             sbody = (<Let>sbody).body
 *             resimplify = True
 *             touched = True             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
    __Pyx_INCREF(Py_True);
    __Pyx_DECREF_SET(__pyx_v_touched, Py_True);

    /* "flitter/language/tree.pyx":1258
 *         if isinstance(sbody, Literal):
 *             return sbody
 *         if isinstance(sbody, Let):             # <<<<<<<<<<<<<<
 *             remaining.extend((<Let>sbody).bindings)
 *             sbody = (<Let>sbody).body
 */
  }

  /* "flitter/language/tree.pyx":1263
 *             resimplify = True
 *             touched = True
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if remaining:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_touched); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1263, __pyx_L1_error)
  __pyx_t_9 = (!__pyx_t_2);
  if (__pyx_t_9) {

    /* "flitter/language/tree.pyx":1264
 *             touched = True
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         if remaining:
 *             sbody = Let(tuple(remaining), sbody)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1263
 *             resimplify = True
 *             touched = True
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if remaining:
 */
  }

  /* "flitter/language/tree.pyx":1265
 *         if not touched:
 *             return self
 *         if remaining:             # <<<<<<<<<<<<<<
 *             sbody = Let(tuple(remaining), sbody)
 *         if resimplify:
 */
  __pyx_t_9 = (PyList_GET_SIZE(__pyx_v_remaining) != 0);
  if (__pyx_t_9) {

    /* "flitter/language/tree.pyx":1266
 *             return self
 *         if remaining:
 *             sbody = Let(tuple(remaining), sbody)             # <<<<<<<<<<<<<<
 *         if resimplify:
 *             return sbody._simplify(context)
 */
    __pyx_t_7 = PyList_AsTuple(__pyx_v_remaining); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_sbody);
    __Pyx_GIVEREF((PyObject *)__pyx_v_sbody);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_sbody))) __PYX_ERR(0, 1266, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let), __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_sbody, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "flitter/language/tree.pyx":1265
 *         if not touched:
 *             return self
 *         if remaining:             # <<<<<<<<<<<<<<
 *             sbody = Let(tuple(remaining), sbody)
 *         if resimplify:
 */
  }

  /* "flitter/language/tree.pyx":1267
 *         if remaining:
 *             sbody = Let(tuple(remaining), sbody)
 *         if resimplify:             # <<<<<<<<<<<<<<
 *             return sbody._simplify(context)
 *         return sbody
 */
  if (__pyx_v_resimplify) {

    /* "flitter/language/tree.pyx":1268
 *             sbody = Let(tuple(remaining), sbody)
 *         if resimplify:
 *             return sbody._simplify(context)             # <<<<<<<<<<<<<<
 *         return sbody
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_sbody->__pyx_vtab)->_simplify(__pyx_v_sbody, __pyx_v_context)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_7);
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1267
 *         if remaining:
 *             sbody = Let(tuple(remaining), sbody)
 *         if resimplify:             # <<<<<<<<<<<<<<
 *             return sbody._simplify(context)
 *         return sbody
 */
  }

  /* "flitter/language/tree.pyx":1269
 *         if resimplify:
 *             return sbody._simplify(context)
 *         return sbody             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_sbody);
  __pyx_r = __pyx_v_sbody;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1189
 *             lnames.pop()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef dict saved = context.names
 *         context.names = saved.copy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("flitter.language.tree.Let._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_saved);
  __Pyx_XDECREF(__pyx_v_bindings);
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF((PyObject *)__pyx_v_binding);
  __Pyx_XDECREF((PyObject *)__pyx_v_expr);
  __Pyx_XDECREF((PyObject *)__pyx_v_body);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_existing_name);
  __Pyx_XDECREF(__pyx_v_touched);
  __Pyx_XDECREF(__pyx_v_shadowed);
  __Pyx_XDECREF(__pyx_v_discarded);
  __Pyx_XDECREF(__pyx_v_existing_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_sbody);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1271
 *         return sbody
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Let({self.bindings!r}, {self.body!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1272
 * 
 *     def __repr__(self):
 *         return f'Let({self.bindings!r}, {self.body!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Let);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_Let);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Let);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->bindings), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->body)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1271
 *         return sbody
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Let({self.bindings!r}, {self.body!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Let.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1160
 * 
 * cdef class Let(Expression):
 *     cdef readonly tuple bindings             # <<<<<<<<<<<<<<
 *     cdef readonly Expression body
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_8bindings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_8bindings_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let_8bindings___get__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_8bindings___get__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->bindings);
  __pyx_r = __pyx_v_self->bindings;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1161
 * cdef class Let(Expression):
 *     cdef readonly tuple bindings
 *     cdef readonly Expression body             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple bindings, Expression body):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_4body_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let_4body___get__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __pyx_r = ((PyObject *)__pyx_v_self->body);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Let_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Let_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.bindings, self.body)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->bindings);
  __Pyx_GIVEREF(__pyx_v_self->bindings);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->bindings)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->body))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.bindings, self.body)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.bindings, self.body)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.bindings is not None or self.body is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.bindings, self.body)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.bindings is not None or self.body is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->bindings != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->body) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bindings is not None or self.body is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.bindings is not None or self.body is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Let); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_237118084);
    __Pyx_GIVEREF(__pyx_int_237118084);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_237118084)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bindings is not None or self.body is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, None), state
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Let); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_237118084);
    __Pyx_GIVEREF(__pyx_int_237118084);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_237118084)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Let.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3Let_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Let_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3Let_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Let.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3Let_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3Let_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Let__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Let.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1280
 *     cdef readonly Expression body
 * 
 *     def __init__(self, tuple names, Expression source, Expression body):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.source = source
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_3For_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_3For_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_names = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_source = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_names,&__pyx_n_s_source,&__pyx_n_s_body,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_names)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1280, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_body)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1280, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1280, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_names = ((PyObject*)values[0]);
    __pyx_v_source = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
    __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1280, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.For.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_names), (&PyTuple_Type), 1, "names", 1))) __PYX_ERR(0, 1280, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_source), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "source", 0))) __PYX_ERR(0, 1280, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_body), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "body", 0))) __PYX_ERR(0, 1280, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For___init__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self), __pyx_v_names, __pyx_v_source, __pyx_v_body);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_3For___init__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, PyObject *__pyx_v_names, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_source, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1281
 * 
 *     def __init__(self, tuple names, Expression source, Expression body):
 *         self.names = names             # <<<<<<<<<<<<<<
 *         self.source = source
 *         self.body = body
 */
  __Pyx_INCREF(__pyx_v_names);
  __Pyx_GIVEREF(__pyx_v_names);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = __pyx_v_names;

  /* "flitter/language/tree.pyx":1282
 *     def __init__(self, tuple names, Expression source, Expression body):
 *         self.names = names
 *         self.source = source             # <<<<<<<<<<<<<<
 *         self.body = body
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_source);
  __Pyx_GIVEREF((PyObject *)__pyx_v_source);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->source);
  __Pyx_DECREF((PyObject *)__pyx_v_self->source);
  __pyx_v_self->source = __pyx_v_source;

  /* "flitter/language/tree.pyx":1283
 *         self.names = names
 *         self.source = source
 *         self.body = body             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->body);
  __Pyx_DECREF((PyObject *)__pyx_v_self->body);
  __pyx_v_self->body = __pyx_v_body;

  /* "flitter/language/tree.pyx":1280
 *     cdef readonly Expression body
 * 
 *     def __init__(self, tuple names, Expression source, Expression body):             # <<<<<<<<<<<<<<
 *         self.names = names
 *         self.source = source
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1285
 *         self.body = body
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.source._compile(program, lnames)
 *         cdef int64_t i, n=len(self.names)
 */

static void __pyx_f_7flitter_8language_4tree_3For__compile(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  CYTHON_UNUSED int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  int64_t __pyx_v_START;
  int64_t __pyx_v_END;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int64_t __pyx_t_3;
  int __pyx_t_4;
  int64_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1286
 * 
 *     cdef void _compile(self, Program program, list lnames):
 *         self.source._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=len(self.names)
 *         program.begin_for(n)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->source->__pyx_vtab)->_compile(__pyx_v_self->source, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1286, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1287
 *     cdef void _compile(self, Program program, list lnames):
 *         self.source._compile(program, lnames)
 *         cdef int64_t i, n=len(self.names)             # <<<<<<<<<<<<<<
 *         program.begin_for(n)
 *         START = program.new_label()
 */
  __pyx_t_1 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1287, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n = __pyx_t_2;

  /* "flitter/language/tree.pyx":1288
 *         self.source._compile(program, lnames)
 *         cdef int64_t i, n=len(self.names)
 *         program.begin_for(n)             # <<<<<<<<<<<<<<
 *         START = program.new_label()
 *         END = program.new_label()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->begin_for(__pyx_v_program, __pyx_v_n, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1289
 *         cdef int64_t i, n=len(self.names)
 *         program.begin_for(n)
 *         START = program.new_label()             # <<<<<<<<<<<<<<
 *         END = program.new_label()
 *         program.label(START)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_v_START = __pyx_t_3;

  /* "flitter/language/tree.pyx":1290
 *         program.begin_for(n)
 *         START = program.new_label()
 *         END = program.new_label()             # <<<<<<<<<<<<<<
 *         program.label(START)
 *         program.next(END)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L1_error)
  __pyx_v_END = __pyx_t_3;

  /* "flitter/language/tree.pyx":1291
 *         START = program.new_label()
 *         END = program.new_label()
 *         program.label(START)             # <<<<<<<<<<<<<<
 *         program.next(END)
 *         lnames.extend(self.names)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_START, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1292
 *         END = program.new_label()
 *         program.label(START)
 *         program.next(END)             # <<<<<<<<<<<<<<
 *         lnames.extend(self.names)
 *         self.body._compile(program, lnames)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->next(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1293
 *         program.label(START)
 *         program.next(END)
 *         lnames.extend(self.names)             # <<<<<<<<<<<<<<
 *         self.body._compile(program, lnames)
 *         program.jump(START)
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
    __PYX_ERR(0, 1293, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyList_Extend(__pyx_v_lnames, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1294
 *         program.next(END)
 *         lnames.extend(self.names)
 *         self.body._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         program.jump(START)
 *         program.label(END)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_compile(__pyx_v_self->body, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1295
 *         lnames.extend(self.names)
 *         self.body._compile(program, lnames)
 *         program.jump(START)             # <<<<<<<<<<<<<<
 *         program.label(END)
 *         program.end_for()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->jump(__pyx_v_program, __pyx_v_START, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1296
 *         self.body._compile(program, lnames)
 *         program.jump(START)
 *         program.label(END)             # <<<<<<<<<<<<<<
 *         program.end_for()
 *         for i in range(n):
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1297
 *         program.jump(START)
 *         program.label(END)
 *         program.end_for()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             lnames.pop()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->end_for(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1298
 *         program.label(END)
 *         program.end_for()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             lnames.pop()
 * 
 */
  __pyx_t_3 = __pyx_v_n;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "flitter/language/tree.pyx":1299
 *         program.end_for()
 *         for i in range(n):
 *             lnames.pop()             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(0, 1299, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_lnames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "flitter/language/tree.pyx":1285
 *         self.body = body
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         self.source._compile(program, lnames)
 *         cdef int64_t i, n=len(self.names)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.For._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1301
 *             lnames.pop()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression body, source=self.source._simplify(context)
 *         cdef list remaining = []
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_3For__simplify(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_source = 0;
  PyObject *__pyx_v_remaining = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_values = 0;
  PyObject *__pyx_v_saved = 0;
  PyObject *__pyx_v_name = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int64_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1302
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression body, source=self.source._simplify(context)             # <<<<<<<<<<<<<<
 *         cdef list remaining = []
 *         cdef Vector values
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->source->__pyx_vtab)->_simplify(__pyx_v_self->source, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_source = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1303
 *     cdef Expression _simplify(self, Context context):
 *         cdef Expression body, source=self.source._simplify(context)
 *         cdef list remaining = []             # <<<<<<<<<<<<<<
 *         cdef Vector values
 *         cdef dict saved = context.names
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_remaining = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1305
 *         cdef list remaining = []
 *         cdef Vector values
 *         cdef dict saved = context.names             # <<<<<<<<<<<<<<
 *         context.names = saved.copy()
 *         cdef str name
 */
  __pyx_t_1 = __pyx_v_context->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_saved = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1306
 *         cdef Vector values
 *         cdef dict saved = context.names
 *         context.names = saved.copy()             # <<<<<<<<<<<<<<
 *         cdef str name
 *         if not isinstance(source, Literal):
 */
  if (unlikely(__pyx_v_saved == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(0, 1306, __pyx_L1_error)
  }
  __pyx_t_1 = PyDict_Copy(__pyx_v_saved); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1308
 *         context.names = saved.copy()
 *         cdef str name
 *         if not isinstance(source, Literal):             # <<<<<<<<<<<<<<
 *             for name in self.names:
 *                 context.names[name] = None
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_source), __pyx_ptype_7flitter_8language_4tree_Literal); 
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "flitter/language/tree.pyx":1309
 *         cdef str name
 *         if not isinstance(source, Literal):
 *             for name in self.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = None
 *             body = self.body._simplify(context)
 */
    if (unlikely(__pyx_v_self->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1309, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1309, __pyx_L1_error)
        #endif
        if (__pyx_t_4 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1309, __pyx_L1_error)
      #else
      __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 1309, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "flitter/language/tree.pyx":1310
 *         if not isinstance(source, Literal):
 *             for name in self.names:
 *                 context.names[name] = None             # <<<<<<<<<<<<<<
 *             body = self.body._simplify(context)
 *             context.names = saved
 */
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1310, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, Py_None) < 0))) __PYX_ERR(0, 1310, __pyx_L1_error)

      /* "flitter/language/tree.pyx":1309
 *         cdef str name
 *         if not isinstance(source, Literal):
 *             for name in self.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = None
 *             body = self.body._simplify(context)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1311
 *             for name in self.names:
 *                 context.names[name] = None
 *             body = self.body._simplify(context)             # <<<<<<<<<<<<<<
 *             context.names = saved
 *             if source is self.source and body is self.body:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_simplify(__pyx_v_self->body, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1312
 *                 context.names[name] = None
 *             body = self.body._simplify(context)
 *             context.names = saved             # <<<<<<<<<<<<<<
 *             if source is self.source and body is self.body:
 *                 return self
 */
    __Pyx_INCREF(__pyx_v_saved);
    __Pyx_GIVEREF(__pyx_v_saved);
    __Pyx_GOTREF(__pyx_v_context->names);
    __Pyx_DECREF(__pyx_v_context->names);
    __pyx_v_context->names = __pyx_v_saved;

    /* "flitter/language/tree.pyx":1313
 *             body = self.body._simplify(context)
 *             context.names = saved
 *             if source is self.source and body is self.body:             # <<<<<<<<<<<<<<
 *                 return self
 *             return For(self.names, source, body)
 */
    __pyx_t_2 = (__pyx_v_source == __pyx_v_self->source);
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_body == __pyx_v_self->body);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_3) {

      /* "flitter/language/tree.pyx":1314
 *             context.names = saved
 *             if source is self.source and body is self.body:
 *                 return self             # <<<<<<<<<<<<<<
 *             return For(self.names, source, body)
 *         values = (<Literal>source).value
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_self);
      __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
      goto __pyx_L0;

      /* "flitter/language/tree.pyx":1313
 *             body = self.body._simplify(context)
 *             context.names = saved
 *             if source is self.source and body is self.body:             # <<<<<<<<<<<<<<
 *                 return self
 *             return For(self.names, source, body)
 */
    }

    /* "flitter/language/tree.pyx":1315
 *             if source is self.source and body is self.body:
 *                 return self
 *             return For(self.names, source, body)             # <<<<<<<<<<<<<<
 *         values = (<Literal>source).value
 *         cdef int64_t i=0, n=values.length
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->names);
    __Pyx_GIVEREF(__pyx_v_self->names);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->names)) __PYX_ERR(0, 1315, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_source);
    __Pyx_GIVEREF((PyObject *)__pyx_v_source);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_source))) __PYX_ERR(0, 1315, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_body);
    __Pyx_GIVEREF((PyObject *)__pyx_v_body);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_body))) __PYX_ERR(0, 1315, __pyx_L1_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_For), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1308
 *         context.names = saved.copy()
 *         cdef str name
 *         if not isinstance(source, Literal):             # <<<<<<<<<<<<<<
 *             for name in self.names:
 *                 context.names[name] = None
 */
  }

  /* "flitter/language/tree.pyx":1316
 *                 return self
 *             return For(self.names, source, body)
 *         values = (<Literal>source).value             # <<<<<<<<<<<<<<
 *         cdef int64_t i=0, n=values.length
 *         while i < n:
 */
  __pyx_t_5 = ((PyObject *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_source)->value);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_values = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/language/tree.pyx":1317
 *             return For(self.names, source, body)
 *         values = (<Literal>source).value
 *         cdef int64_t i=0, n=values.length             # <<<<<<<<<<<<<<
 *         while i < n:
 *             for name in self.names:
 */
  __pyx_v_i = 0;
  __pyx_t_6 = __pyx_v_values->length;
  __pyx_v_n = __pyx_t_6;

  /* "flitter/language/tree.pyx":1318
 *         values = (<Literal>source).value
 *         cdef int64_t i=0, n=values.length
 *         while i < n:             # <<<<<<<<<<<<<<
 *             for name in self.names:
 *                 context.names[name] = values.item(i) if i < n else null_
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_i < __pyx_v_n);
    if (!__pyx_t_3) break;

    /* "flitter/language/tree.pyx":1319
 *         cdef int64_t i=0, n=values.length
 *         while i < n:
 *             for name in self.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = values.item(i) if i < n else null_
 *                 i += 1
 */
    if (unlikely(__pyx_v_self->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1319, __pyx_L1_error)
    }
    __pyx_t_5 = __pyx_v_self->names; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1319, __pyx_L1_error)
        #endif
        if (__pyx_t_4 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1319, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1320
 *         while i < n:
 *             for name in self.names:
 *                 context.names[name] = values.item(i) if i < n else null_             # <<<<<<<<<<<<<<
 *                 i += 1
 *             remaining.append(self.body._simplify(context))
 */
      __pyx_t_3 = (__pyx_v_i < __pyx_v_n);
      if (__pyx_t_3) {
        __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_values->__pyx_vtab)->item(__pyx_v_values, __pyx_v_i, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
        __pyx_t_7 = 0;
      } else {
        __Pyx_INCREF((PyObject *)__pyx_v_7flitter_5model_null_);
        __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_5model_null_);
      }
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1320, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_name, __pyx_t_1) < 0))) __PYX_ERR(0, 1320, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flitter/language/tree.pyx":1321
 *             for name in self.names:
 *                 context.names[name] = values.item(i) if i < n else null_
 *                 i += 1             # <<<<<<<<<<<<<<
 *             remaining.append(self.body._simplify(context))
 *         context.names = saved
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "flitter/language/tree.pyx":1319
 *         cdef int64_t i=0, n=values.length
 *         while i < n:
 *             for name in self.names:             # <<<<<<<<<<<<<<
 *                 context.names[name] = values.item(i) if i < n else null_
 *                 i += 1
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "flitter/language/tree.pyx":1322
 *                 context.names[name] = values.item(i) if i < n else null_
 *                 i += 1
 *             remaining.append(self.body._simplify(context))             # <<<<<<<<<<<<<<
 *         context.names = saved
 *         sequence_pack(remaining)
 */
    __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_simplify(__pyx_v_self->body, __pyx_v_context)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "flitter/language/tree.pyx":1323
 *                 i += 1
 *             remaining.append(self.body._simplify(context))
 *         context.names = saved             # <<<<<<<<<<<<<<
 *         sequence_pack(remaining)
 *         if not remaining:
 */
  __Pyx_INCREF(__pyx_v_saved);
  __Pyx_GIVEREF(__pyx_v_saved);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = __pyx_v_saved;

  /* "flitter/language/tree.pyx":1324
 *             remaining.append(self.body._simplify(context))
 *         context.names = saved
 *         sequence_pack(remaining)             # <<<<<<<<<<<<<<
 *         if not remaining:
 *             return NoOp
 */
  __pyx_t_3 = __pyx_f_7flitter_8language_4tree_sequence_pack(__pyx_v_remaining); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1324, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1325
 *         context.names = saved
 *         sequence_pack(remaining)
 *         if not remaining:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         if len(remaining) == 1:
 */
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_remaining) != 0);
  __pyx_t_2 = (!__pyx_t_3);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1326
 *         sequence_pack(remaining)
 *         if not remaining:
 *             return NoOp             # <<<<<<<<<<<<<<
 *         if len(remaining) == 1:
 *             return remaining[0]
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_7flitter_8language_4tree_NoOp);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1325
 *         context.names = saved
 *         sequence_pack(remaining)
 *         if not remaining:             # <<<<<<<<<<<<<<
 *             return NoOp
 *         if len(remaining) == 1:
 */
  }

  /* "flitter/language/tree.pyx":1327
 *         if not remaining:
 *             return NoOp
 *         if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *             return remaining[0]
 *         return Sequence(tuple(remaining))
 */
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_remaining); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1327, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 == 1);
  if (__pyx_t_2) {

    /* "flitter/language/tree.pyx":1328
 *             return NoOp
 *         if len(remaining) == 1:
 *             return remaining[0]             # <<<<<<<<<<<<<<
 *         return Sequence(tuple(remaining))
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_remaining, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 1328, __pyx_L1_error)
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1327
 *         if not remaining:
 *             return NoOp
 *         if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *             return remaining[0]
 *         return Sequence(tuple(remaining))
 */
  }

  /* "flitter/language/tree.pyx":1329
 *         if len(remaining) == 1:
 *             return remaining[0]
 *         return Sequence(tuple(remaining))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_5 = PyList_AsTuple(__pyx_v_remaining); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence), __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1301
 *             lnames.pop()
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef Expression body, source=self.source._simplify(context)
 *         cdef list remaining = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.For._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_body);
  __Pyx_XDECREF((PyObject *)__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF((PyObject *)__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_saved);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1331
 *         return Sequence(tuple(remaining))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'For({self.names!r}, {self.source!r}, {self.body!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_2__repr__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1332
 * 
 *     def __repr__(self):
 *         return f'For({self.names!r}, {self.source!r}, {self.body!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_For);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_For);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_For);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->names), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->source)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->body)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1331
 *         return Sequence(tuple(remaining))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'For({self.names!r}, {self.source!r}, {self.body!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.For.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1276
 * 
 * cdef class For(Expression):
 *     cdef readonly tuple names             # <<<<<<<<<<<<<<
 *     cdef readonly Expression source
 *     cdef readonly Expression body
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_5names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_5names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_5names___get__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_5names___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->names);
  __pyx_r = __pyx_v_self->names;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1277
 * cdef class For(Expression):
 *     cdef readonly tuple names
 *     cdef readonly Expression source             # <<<<<<<<<<<<<<
 *     cdef readonly Expression body
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_6source_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_6source_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_6source___get__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_6source___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->source);
  __pyx_r = ((PyObject *)__pyx_v_self->source);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1278
 *     cdef readonly tuple names
 *     cdef readonly Expression source
 *     cdef readonly Expression body             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple names, Expression source, Expression body):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_4body_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_4body___get__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_4body___get__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __pyx_r = ((PyObject *)__pyx_v_self->body);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3For_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3For_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.body, self.names, self.source)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->body))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->names);
  __Pyx_GIVEREF(__pyx_v_self->names);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->names)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->source);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->source);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->source))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.body, self.names, self.source)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.body, self.names, self.source)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.body is not None or self.names is not None or self.source is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.body, self.names, self.source)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.body is not None or self.names is not None or self.source is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->body) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->names != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->source) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.names is not None or self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.body is not None or self.names is not None or self.source is not None
 *     if use_setstate:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_For); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_64598595);
    __Pyx_GIVEREF(__pyx_int_64598595);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_64598595)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.names is not None or self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, None), state
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_For__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_For); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_64598595);
    __Pyx_GIVEREF(__pyx_int_64598595);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_64598595)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.For.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_For__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3For_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3For_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3For_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.For.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_3For_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_3For_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_For__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_For__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_For__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.For.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1339
 *     cdef readonly Expression then
 * 
 *     def __init__(self, Expression condition, Expression then):             # <<<<<<<<<<<<<<
 *         self.condition = condition
 *         self.then = then
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_11IfCondition_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_11IfCondition_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_condition = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_then = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_condition,&__pyx_n_s_then,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_condition)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_then)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1339, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1339, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_condition = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[0]);
    __pyx_v_then = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1339, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.IfCondition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_condition), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "condition", 0))) __PYX_ERR(0, 1339, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_then), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "then", 0))) __PYX_ERR(0, 1339, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition___init__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self), __pyx_v_condition, __pyx_v_then);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_11IfCondition___init__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_condition, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_then) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1340
 * 
 *     def __init__(self, Expression condition, Expression then):
 *         self.condition = condition             # <<<<<<<<<<<<<<
 *         self.then = then
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_condition);
  __Pyx_GIVEREF((PyObject *)__pyx_v_condition);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->condition);
  __Pyx_DECREF((PyObject *)__pyx_v_self->condition);
  __pyx_v_self->condition = __pyx_v_condition;

  /* "flitter/language/tree.pyx":1341
 *     def __init__(self, Expression condition, Expression then):
 *         self.condition = condition
 *         self.then = then             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_then);
  __Pyx_GIVEREF((PyObject *)__pyx_v_then);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->then);
  __Pyx_DECREF((PyObject *)__pyx_v_self->then);
  __pyx_v_self->then = __pyx_v_then;

  /* "flitter/language/tree.pyx":1339
 *     cdef readonly Expression then
 * 
 *     def __init__(self, Expression condition, Expression then):             # <<<<<<<<<<<<<<
 *         self.condition = condition
 *         self.then = then
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1343
 *         self.then = then
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'IfCondition({self.condition!r}, {self.then!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_2__repr__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1344
 * 
 *     def __repr__(self):
 *         return f'IfCondition({self.condition!r}, {self.then!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_IfCondition);
  __pyx_t_2 += 12;
  __Pyx_GIVEREF(__pyx_kp_u_IfCondition);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_IfCondition);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->condition)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->then)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1343
 *         self.then = then
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'IfCondition({self.condition!r}, {self.then!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.IfCondition.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1336
 * 
 * cdef class IfCondition:
 *     cdef readonly Expression condition             # <<<<<<<<<<<<<<
 *     cdef readonly Expression then
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_9condition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_9condition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition_9condition___get__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_9condition___get__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->condition);
  __pyx_r = ((PyObject *)__pyx_v_self->condition);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1337
 * cdef class IfCondition:
 *     cdef readonly Expression condition
 *     cdef readonly Expression then             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Expression condition, Expression then):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_4then_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_4then_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition_4then___get__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_4then___get__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->then);
  __pyx_r = ((PyObject *)__pyx_v_self->then);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11IfCondition_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11IfCondition_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.condition, self.then)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->condition);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->condition);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->condition))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->then);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->then);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->then))) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.condition, self.then)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.condition, self.then)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.condition is not None or self.then is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.condition, self.then)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.condition is not None or self.then is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->condition) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->then) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.condition is not None or self.then is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.condition is not None or self.then is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IfCondition); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_161766002);
    __Pyx_GIVEREF(__pyx_int_161766002);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_161766002)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.condition is not None or self.then is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, None), state
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_IfCondition); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_161766002);
    __Pyx_GIVEREF(__pyx_int_161766002);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_161766002)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.IfCondition.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11IfCondition_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11IfCondition_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11IfCondition_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.IfCondition.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_11IfCondition_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_11IfCondition_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_IfCondition__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.IfCondition.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1351
 *     cdef readonly Expression else_
 * 
 *     def __init__(self, tuple tests, Expression else_):             # <<<<<<<<<<<<<<
 *         self.tests = tests
 *         self.else_ = else_
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_6IfElse_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_6IfElse_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_tests = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_else_ = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tests,&__pyx_n_s_else,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tests)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1351, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_else)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1351, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1351, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1351, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_tests = ((PyObject*)values[0]);
    __pyx_v_else_ = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1351, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.IfElse.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tests), (&PyTuple_Type), 1, "tests", 1))) __PYX_ERR(0, 1351, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_else_), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "else_", 0))) __PYX_ERR(0, 1351, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse___init__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self), __pyx_v_tests, __pyx_v_else_);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_6IfElse___init__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, PyObject *__pyx_v_tests, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_else_) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1352
 * 
 *     def __init__(self, tuple tests, Expression else_):
 *         self.tests = tests             # <<<<<<<<<<<<<<
 *         self.else_ = else_
 * 
 */
  __Pyx_INCREF(__pyx_v_tests);
  __Pyx_GIVEREF(__pyx_v_tests);
  __Pyx_GOTREF(__pyx_v_self->tests);
  __Pyx_DECREF(__pyx_v_self->tests);
  __pyx_v_self->tests = __pyx_v_tests;

  /* "flitter/language/tree.pyx":1353
 *     def __init__(self, tuple tests, Expression else_):
 *         self.tests = tests
 *         self.else_ = else_             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_else_);
  __Pyx_GIVEREF((PyObject *)__pyx_v_else_);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->else_);
  __Pyx_DECREF((PyObject *)__pyx_v_self->else_);
  __pyx_v_self->else_ = __pyx_v_else_;

  /* "flitter/language/tree.pyx":1351
 *     cdef readonly Expression else_
 * 
 *     def __init__(self, tuple tests, Expression else_):             # <<<<<<<<<<<<<<
 *         self.tests = tests
 *         self.else_ = else_
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1355
 *         self.else_ = else_
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef IfCondition test
 *         END = program.new_label()
 */

static void __pyx_f_7flitter_8language_4tree_6IfElse__compile(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_test = 0;
  int64_t __pyx_v_END;
  int64_t __pyx_v_NEXT;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1357
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef IfCondition test
 *         END = program.new_label()             # <<<<<<<<<<<<<<
 *         for test in self.tests:
 *             NEXT = program.new_label()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1357, __pyx_L1_error)
  __pyx_v_END = __pyx_t_1;

  /* "flitter/language/tree.pyx":1358
 *         cdef IfCondition test
 *         END = program.new_label()
 *         for test in self.tests:             # <<<<<<<<<<<<<<
 *             NEXT = program.new_label()
 *             test.condition._compile(program, lnames)
 */
  if (unlikely(__pyx_v_self->tests == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1358, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->tests; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1358, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1358, __pyx_L1_error)
    #else
    __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_7flitter_8language_4tree_IfCondition))))) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_test, ((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":1359
 *         END = program.new_label()
 *         for test in self.tests:
 *             NEXT = program.new_label()             # <<<<<<<<<<<<<<
 *             test.condition._compile(program, lnames)
 *             program.branch_false(NEXT)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1359, __pyx_L1_error)
    __pyx_v_NEXT = __pyx_t_1;

    /* "flitter/language/tree.pyx":1360
 *         for test in self.tests:
 *             NEXT = program.new_label()
 *             test.condition._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.branch_false(NEXT)
 *             test.then._compile(program, lnames)
 */
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_test->condition->__pyx_vtab)->_compile(__pyx_v_test->condition, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1361
 *             NEXT = program.new_label()
 *             test.condition._compile(program, lnames)
 *             program.branch_false(NEXT)             # <<<<<<<<<<<<<<
 *             test.then._compile(program, lnames)
 *             program.jump(END)
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->branch_false(__pyx_v_program, __pyx_v_NEXT, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":1362
 *             test.condition._compile(program, lnames)
 *             program.branch_false(NEXT)
 *             test.then._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             program.jump(END)
 *             program.label(NEXT)
 */
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_test->then->__pyx_vtab)->_compile(__pyx_v_test->then, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1363
 *             program.branch_false(NEXT)
 *             test.then._compile(program, lnames)
 *             program.jump(END)             # <<<<<<<<<<<<<<
 *             program.label(NEXT)
 *         if self.else_ is not None:
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->jump(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":1364
 *             test.then._compile(program, lnames)
 *             program.jump(END)
 *             program.label(NEXT)             # <<<<<<<<<<<<<<
 *         if self.else_ is not None:
 *             self.else_._compile(program, lnames)
 */
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_NEXT, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flitter/language/tree.pyx":1358
 *         cdef IfCondition test
 *         END = program.new_label()
 *         for test in self.tests:             # <<<<<<<<<<<<<<
 *             NEXT = program.new_label()
 *             test.condition._compile(program, lnames)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":1365
 *             program.jump(END)
 *             program.label(NEXT)
 *         if self.else_ is not None:             # <<<<<<<<<<<<<<
 *             self.else_._compile(program, lnames)
 *         else:
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->else_) != Py_None);
  if (__pyx_t_5) {

    /* "flitter/language/tree.pyx":1366
 *             program.label(NEXT)
 *         if self.else_ is not None:
 *             self.else_._compile(program, lnames)             # <<<<<<<<<<<<<<
 *         else:
 *             program.literal(null_)
 */
    ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->else_->__pyx_vtab)->_compile(__pyx_v_self->else_, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1365
 *             program.jump(END)
 *             program.label(NEXT)
 *         if self.else_ is not None:             # <<<<<<<<<<<<<<
 *             self.else_._compile(program, lnames)
 *         else:
 */
    goto __pyx_L6;
  }

  /* "flitter/language/tree.pyx":1368
 *             self.else_._compile(program, lnames)
 *         else:
 *             program.literal(null_)             # <<<<<<<<<<<<<<
 *         program.label(END)
 * 
 */
  /*else*/ {
    __pyx_t_2 = ((PyObject *)__pyx_v_7flitter_5model_null_);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_2, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L6:;

  /* "flitter/language/tree.pyx":1369
 *         else:
 *             program.literal(null_)
 *         program.label(END)             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":1355
 *         self.else_ = else_
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef IfCondition test
 *         END = program.new_label()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.IfElse._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_test);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1371
 *         program.label(END)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list remaining = []
 *         cdef IfCondition test
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_6IfElse__simplify(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_remaining = 0;
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v_test = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_condition = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_then = 0;
  int __pyx_v_touched;
  PyObject *__pyx_v_else_ = NULL;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1372
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef list remaining = []             # <<<<<<<<<<<<<<
 *         cdef IfCondition test
 *         cdef Expression condition, then
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_remaining = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1375
 *         cdef IfCondition test
 *         cdef Expression condition, then
 *         cdef bint touched = False             # <<<<<<<<<<<<<<
 *         for test in self.tests:
 *             condition = test.condition._simplify(context)
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":1376
 *         cdef Expression condition, then
 *         cdef bint touched = False
 *         for test in self.tests:             # <<<<<<<<<<<<<<
 *             condition = test.condition._simplify(context)
 *             touched |= condition is not test.condition
 */
  if (unlikely(__pyx_v_self->tests == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1376, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->tests; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1376, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1376, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_IfCondition))))) __PYX_ERR(0, 1376, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_test, ((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1377
 *         cdef bint touched = False
 *         for test in self.tests:
 *             condition = test.condition._simplify(context)             # <<<<<<<<<<<<<<
 *             touched |= condition is not test.condition
 *             if isinstance(condition, Literal):
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_test->condition->__pyx_vtab)->_simplify(__pyx_v_test->condition, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_condition, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1378
 *         for test in self.tests:
 *             condition = test.condition._simplify(context)
 *             touched |= condition is not test.condition             # <<<<<<<<<<<<<<
 *             if isinstance(condition, Literal):
 *                 if (<Literal>condition).value.as_bool():
 */
    __pyx_t_4 = (__pyx_v_condition != __pyx_v_test->condition);
    __pyx_v_touched = (__pyx_v_touched | __pyx_t_4);

    /* "flitter/language/tree.pyx":1379
 *             condition = test.condition._simplify(context)
 *             touched |= condition is not test.condition
 *             if isinstance(condition, Literal):             # <<<<<<<<<<<<<<
 *                 if (<Literal>condition).value.as_bool():
 *                     then = test.then._simplify(context)
 */
    __pyx_t_4 = __Pyx_TypeCheck(((PyObject *)__pyx_v_condition), __pyx_ptype_7flitter_8language_4tree_Literal); 
    if (__pyx_t_4) {

      /* "flitter/language/tree.pyx":1380
 *             touched |= condition is not test.condition
 *             if isinstance(condition, Literal):
 *                 if (<Literal>condition).value.as_bool():             # <<<<<<<<<<<<<<
 *                     then = test.then._simplify(context)
 *                     if not remaining:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_condition)->value->__pyx_vtab)->as_bool(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v_condition)->value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L1_error)
      if (__pyx_t_4) {

        /* "flitter/language/tree.pyx":1381
 *             if isinstance(condition, Literal):
 *                 if (<Literal>condition).value.as_bool():
 *                     then = test.then._simplify(context)             # <<<<<<<<<<<<<<
 *                     if not remaining:
 *                         return then
 */
        __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_test->then->__pyx_vtab)->_simplify(__pyx_v_test->then, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_then, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "flitter/language/tree.pyx":1382
 *                 if (<Literal>condition).value.as_bool():
 *                     then = test.then._simplify(context)
 *                     if not remaining:             # <<<<<<<<<<<<<<
 *                         return then
 *                     else:
 */
        __pyx_t_4 = (PyList_GET_SIZE(__pyx_v_remaining) != 0);
        __pyx_t_5 = (!__pyx_t_4);
        if (__pyx_t_5) {

          /* "flitter/language/tree.pyx":1383
 *                     then = test.then._simplify(context)
 *                     if not remaining:
 *                         return then             # <<<<<<<<<<<<<<
 *                     else:
 *                         return IfElse(tuple(remaining), then)
 */
          __Pyx_XDECREF((PyObject *)__pyx_r);
          __Pyx_INCREF((PyObject *)__pyx_v_then);
          __pyx_r = __pyx_v_then;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L0;

          /* "flitter/language/tree.pyx":1382
 *                 if (<Literal>condition).value.as_bool():
 *                     then = test.then._simplify(context)
 *                     if not remaining:             # <<<<<<<<<<<<<<
 *                         return then
 *                     else:
 */
        }

        /* "flitter/language/tree.pyx":1385
 *                         return then
 *                     else:
 *                         return IfElse(tuple(remaining), then)             # <<<<<<<<<<<<<<
 *                 touched = True
 *             else:
 */
        /*else*/ {
          __Pyx_XDECREF((PyObject *)__pyx_r);
          __pyx_t_3 = PyList_AsTuple(__pyx_v_remaining); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_3);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L1_error);
          __Pyx_INCREF((PyObject *)__pyx_v_then);
          __Pyx_GIVEREF((PyObject *)__pyx_v_then);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_then))) __PYX_ERR(0, 1385, __pyx_L1_error);
          __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfElse), __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3);
          __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L0;
        }

        /* "flitter/language/tree.pyx":1380
 *             touched |= condition is not test.condition
 *             if isinstance(condition, Literal):
 *                 if (<Literal>condition).value.as_bool():             # <<<<<<<<<<<<<<
 *                     then = test.then._simplify(context)
 *                     if not remaining:
 */
      }

      /* "flitter/language/tree.pyx":1386
 *                     else:
 *                         return IfElse(tuple(remaining), then)
 *                 touched = True             # <<<<<<<<<<<<<<
 *             else:
 *                 then = test.then._simplify(context)
 */
      __pyx_v_touched = 1;

      /* "flitter/language/tree.pyx":1379
 *             condition = test.condition._simplify(context)
 *             touched |= condition is not test.condition
 *             if isinstance(condition, Literal):             # <<<<<<<<<<<<<<
 *                 if (<Literal>condition).value.as_bool():
 *                     then = test.then._simplify(context)
 */
      goto __pyx_L5;
    }

    /* "flitter/language/tree.pyx":1388
 *                 touched = True
 *             else:
 *                 then = test.then._simplify(context)             # <<<<<<<<<<<<<<
 *                 remaining.append(IfCondition(condition, then))
 *                 touched |= then is not test.then
 */
    /*else*/ {
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_test->then->__pyx_vtab)->_simplify(__pyx_v_test->then, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_then, ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "flitter/language/tree.pyx":1389
 *             else:
 *                 then = test.then._simplify(context)
 *                 remaining.append(IfCondition(condition, then))             # <<<<<<<<<<<<<<
 *                 touched |= then is not test.then
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF((PyObject *)__pyx_v_condition);
      __Pyx_GIVEREF((PyObject *)__pyx_v_condition);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_condition))) __PYX_ERR(0, 1389, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_then);
      __Pyx_GIVEREF((PyObject *)__pyx_v_then);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_then))) __PYX_ERR(0, 1389, __pyx_L1_error);
      __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfCondition), __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_remaining, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1389, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "flitter/language/tree.pyx":1390
 *                 then = test.then._simplify(context)
 *                 remaining.append(IfCondition(condition, then))
 *                 touched |= then is not test.then             # <<<<<<<<<<<<<<
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None
 *         touched |= else_ is not self.else_
 */
      __pyx_t_5 = (__pyx_v_then != __pyx_v_test->then);
      __pyx_v_touched = (__pyx_v_touched | __pyx_t_5);
    }
    __pyx_L5:;

    /* "flitter/language/tree.pyx":1376
 *         cdef Expression condition, then
 *         cdef bint touched = False
 *         for test in self.tests:             # <<<<<<<<<<<<<<
 *             condition = test.condition._simplify(context)
 *             touched |= condition is not test.condition
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1391
 *                 remaining.append(IfCondition(condition, then))
 *                 touched |= then is not test.then
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None             # <<<<<<<<<<<<<<
 *         touched |= else_ is not self.else_
 *         if not touched:
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->else_) != Py_None);
  if (__pyx_t_5) {
    __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->else_->__pyx_vtab)->_simplify(__pyx_v_self->else_, __pyx_v_context)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_v_else_ = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1392
 *                 touched |= then is not test.then
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None
 *         touched |= else_ is not self.else_             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
  __pyx_t_5 = (__pyx_v_else_ != ((PyObject *)__pyx_v_self->else_));
  __pyx_v_touched = (__pyx_v_touched | __pyx_t_5);

  /* "flitter/language/tree.pyx":1393
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None
 *         touched |= else_ is not self.else_
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if remaining:
 */
  __pyx_t_5 = (!__pyx_v_touched);
  if (__pyx_t_5) {

    /* "flitter/language/tree.pyx":1394
 *         touched |= else_ is not self.else_
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         if remaining:
 *             return IfElse(tuple(remaining), else_)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1393
 *         else_ = self.else_._simplify(context) if self.else_ is not None else None
 *         touched |= else_ is not self.else_
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         if remaining:
 */
  }

  /* "flitter/language/tree.pyx":1395
 *         if not touched:
 *             return self
 *         if remaining:             # <<<<<<<<<<<<<<
 *             return IfElse(tuple(remaining), else_)
 *         return NoOp if else_ is None else else_
 */
  __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_remaining) != 0);
  if (__pyx_t_5) {

    /* "flitter/language/tree.pyx":1396
 *             return self
 *         if remaining:
 *             return IfElse(tuple(remaining), else_)             # <<<<<<<<<<<<<<
 *         return NoOp if else_ is None else else_
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = PyList_AsTuple(__pyx_v_remaining); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_else_);
    __Pyx_GIVEREF(__pyx_v_else_);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_else_)) __PYX_ERR(0, 1396, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfElse), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1395
 *         if not touched:
 *             return self
 *         if remaining:             # <<<<<<<<<<<<<<
 *             return IfElse(tuple(remaining), else_)
 *         return NoOp if else_ is None else else_
 */
  }

  /* "flitter/language/tree.pyx":1397
 *         if remaining:
 *             return IfElse(tuple(remaining), else_)
 *         return NoOp if else_ is None else else_             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_5 = (__pyx_v_else_ == Py_None);
  if (__pyx_t_5) {
    __Pyx_INCREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
    __pyx_t_1 = ((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
  } else {
    if (!(likely(((__pyx_v_else_) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_else_, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_else_);
    __pyx_t_1 = __pyx_v_else_;
  }
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1371
 *         program.label(END)
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list remaining = []
 *         cdef IfCondition test
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.language.tree.IfElse._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF((PyObject *)__pyx_v_test);
  __Pyx_XDECREF((PyObject *)__pyx_v_condition);
  __Pyx_XDECREF((PyObject *)__pyx_v_then);
  __Pyx_XDECREF(__pyx_v_else_);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1399
 *         return NoOp if else_ is None else else_
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'IfElse({self.tests!r}, {self.else_!r})'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_2__repr__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1400
 * 
 *     def __repr__(self):
 *         return f'IfElse({self.tests!r}, {self.else_!r})'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_IfElse);
  __pyx_t_2 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_IfElse);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_IfElse);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->tests), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->else_)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1399
 *         return NoOp if else_ is None else else_
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'IfElse({self.tests!r}, {self.else_!r})'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.IfElse.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1348
 * 
 * cdef class IfElse(Expression):
 *     cdef readonly tuple tests             # <<<<<<<<<<<<<<
 *     cdef readonly Expression else_
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5tests_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5tests_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse_5tests___get__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_5tests___get__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->tests);
  __pyx_r = __pyx_v_self->tests;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1349
 * cdef class IfElse(Expression):
 *     cdef readonly tuple tests
 *     cdef readonly Expression else_             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, tuple tests, Expression else_):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5else__1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5else__1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse_5else____get__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_5else____get__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->else_);
  __pyx_r = ((PyObject *)__pyx_v_self->else_);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6IfElse_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6IfElse_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.else_, self.tests)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->else_);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->else_);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->else_))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->tests);
  __Pyx_GIVEREF(__pyx_v_self->tests);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->tests)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.else_, self.tests)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.else_, self.tests)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.else_ is not None or self.tests is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.else_, self.tests)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.else_ is not None or self.tests is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->else_) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->tests != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.else_ is not None or self.tests is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.else_ is not None or self.tests is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IfElse); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_231565212);
    __Pyx_GIVEREF(__pyx_int_231565212);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_231565212)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.else_ is not None or self.tests is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, None), state
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_IfElse); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_231565212);
    __Pyx_GIVEREF(__pyx_int_231565212);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_231565212)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.IfElse.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_6IfElse_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6IfElse_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_6IfElse_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.IfElse.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6IfElse_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6IfElse_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_IfElse__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.IfElse.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1411
 *     cdef readonly bint recursive
 * 
 *     def __init__(self, str name, tuple parameters, Expression body, tuple captures=None, bint inlineable=False, bint recursive=False):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.parameters = parameters
 */

/* Python wrapper */
static int __pyx_pw_7flitter_8language_4tree_8Function_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_8language_4tree_8Function_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parameters = 0;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  PyObject *__pyx_v_captures = 0;
  int __pyx_v_inlineable;
  int __pyx_v_recursive;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_parameters,&__pyx_n_s_body,&__pyx_n_s_captures,&__pyx_n_s_inlineable,&__pyx_n_s_recursive,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parameters)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, 1); __PYX_ERR(0, 1411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_body)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, 2); __PYX_ERR(0, 1411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_captures);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_inlineable);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_recursive);
          if (value) { values[5] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 1411, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_parameters = ((PyObject*)values[1]);
    __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)values[2]);
    __pyx_v_captures = ((PyObject*)values[3]);
    if (values[4]) {
      __pyx_v_inlineable = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_inlineable == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
    } else {
      __pyx_v_inlineable = ((int)0);
    }
    if (values[5]) {
      __pyx_v_recursive = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_recursive == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L3_error)
    } else {
      __pyx_v_recursive = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 6, __pyx_nargs); __PYX_ERR(0, 1411, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Function.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parameters), (&PyTuple_Type), 1, "parameters", 1))) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_body), __pyx_ptype_7flitter_8language_4tree_Expression, 1, "body", 0))) __PYX_ERR(0, 1411, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_captures), (&PyTuple_Type), 1, "captures", 1))) __PYX_ERR(0, 1411, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function___init__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self), __pyx_v_name, __pyx_v_parameters, __pyx_v_body, __pyx_v_captures, __pyx_v_inlineable, __pyx_v_recursive);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_8language_4tree_8Function___init__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_parameters, struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body, PyObject *__pyx_v_captures, int __pyx_v_inlineable, int __pyx_v_recursive) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/language/tree.pyx":1412
 * 
 *     def __init__(self, str name, tuple parameters, Expression body, tuple captures=None, bint inlineable=False, bint recursive=False):
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.parameters = parameters
 *         self.body = body
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "flitter/language/tree.pyx":1413
 *     def __init__(self, str name, tuple parameters, Expression body, tuple captures=None, bint inlineable=False, bint recursive=False):
 *         self.name = name
 *         self.parameters = parameters             # <<<<<<<<<<<<<<
 *         self.body = body
 *         self.captures = captures
 */
  __Pyx_INCREF(__pyx_v_parameters);
  __Pyx_GIVEREF(__pyx_v_parameters);
  __Pyx_GOTREF(__pyx_v_self->parameters);
  __Pyx_DECREF(__pyx_v_self->parameters);
  __pyx_v_self->parameters = __pyx_v_parameters;

  /* "flitter/language/tree.pyx":1414
 *         self.name = name
 *         self.parameters = parameters
 *         self.body = body             # <<<<<<<<<<<<<<
 *         self.captures = captures
 *         self.inlineable = inlineable
 */
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->body);
  __Pyx_DECREF((PyObject *)__pyx_v_self->body);
  __pyx_v_self->body = __pyx_v_body;

  /* "flitter/language/tree.pyx":1415
 *         self.parameters = parameters
 *         self.body = body
 *         self.captures = captures             # <<<<<<<<<<<<<<
 *         self.inlineable = inlineable
 *         self.recursive = recursive
 */
  __Pyx_INCREF(__pyx_v_captures);
  __Pyx_GIVEREF(__pyx_v_captures);
  __Pyx_GOTREF(__pyx_v_self->captures);
  __Pyx_DECREF(__pyx_v_self->captures);
  __pyx_v_self->captures = __pyx_v_captures;

  /* "flitter/language/tree.pyx":1416
 *         self.body = body
 *         self.captures = captures
 *         self.inlineable = inlineable             # <<<<<<<<<<<<<<
 *         self.recursive = recursive
 * 
 */
  __pyx_v_self->inlineable = __pyx_v_inlineable;

  /* "flitter/language/tree.pyx":1417
 *         self.captures = captures
 *         self.inlineable = inlineable
 *         self.recursive = recursive             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compile(self, Program program, list lnames):
 */
  __pyx_v_self->recursive = __pyx_v_recursive;

  /* "flitter/language/tree.pyx":1411
 *     cdef readonly bint recursive
 * 
 *     def __init__(self, str name, tuple parameters, Expression body, tuple captures=None, bint inlineable=False, bint recursive=False):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.parameters = parameters
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1419
 *         self.recursive = recursive
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef list function_lnames = []
 */

static void __pyx_f_7flitter_8language_4tree_8Function__compile(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, struct __pyx_obj_7flitter_8language_2vm_Program *__pyx_v_program, PyObject *__pyx_v_lnames) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_function_lnames = 0;
  PyObject *__pyx_v_captures = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  PyObject *__pyx_v_parameters = 0;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_parameter = 0;
  int64_t __pyx_v_START;
  int64_t __pyx_v_END;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  struct __pyx_opt_args_7flitter_8language_2vm_7Program_func __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compile", 1);

  /* "flitter/language/tree.pyx":1421
 *     cdef void _compile(self, Program program, list lnames):
 *         cdef str name
 *         cdef list function_lnames = []             # <<<<<<<<<<<<<<
 *         cdef tuple captures = self.captures if self.captures is not None else tuple(lnames)
 *         cdef int64_t i, n=len(lnames)-1
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_function_lnames = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1422
 *         cdef str name
 *         cdef list function_lnames = []
 *         cdef tuple captures = self.captures if self.captures is not None else tuple(lnames)             # <<<<<<<<<<<<<<
 *         cdef int64_t i, n=len(lnames)-1
 *         for name in captures:
 */
  __pyx_t_2 = (__pyx_v_self->captures != ((PyObject*)Py_None));
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_v_self->captures);
    __pyx_t_1 = __pyx_v_self->captures;
  } else {
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1422, __pyx_L1_error)
    }
    __pyx_t_3 = PyList_AsTuple(__pyx_v_lnames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_captures = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1423
 *         cdef list function_lnames = []
 *         cdef tuple captures = self.captures if self.captures is not None else tuple(lnames)
 *         cdef int64_t i, n=len(lnames)-1             # <<<<<<<<<<<<<<
 *         for name in captures:
 *             for i in range(len(lnames)):
 */
  if (unlikely(__pyx_v_lnames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1423, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1423, __pyx_L1_error)
  __pyx_v_n = (__pyx_t_4 - 1);

  /* "flitter/language/tree.pyx":1424
 *         cdef tuple captures = self.captures if self.captures is not None else tuple(lnames)
 *         cdef int64_t i, n=len(lnames)-1
 *         for name in captures:             # <<<<<<<<<<<<<<
 *             for i in range(len(lnames)):
 *                 if name == <str>lnames[n-i]:
 */
  if (unlikely(__pyx_v_captures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1424, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_captures; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1424, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1424, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 1424, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1425
 *         cdef int64_t i, n=len(lnames)-1
 *         for name in captures:
 *             for i in range(len(lnames)):             # <<<<<<<<<<<<<<
 *                 if name == <str>lnames[n-i]:
 *                     program.local_load(i)
 */
    if (unlikely(__pyx_v_lnames == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1425, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_GET_SIZE(__pyx_v_lnames); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1425, __pyx_L1_error)
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "flitter/language/tree.pyx":1426
 *         for name in captures:
 *             for i in range(len(lnames)):
 *                 if name == <str>lnames[n-i]:             # <<<<<<<<<<<<<<
 *                     program.local_load(i)
 *                     break
 */
      if (unlikely(__pyx_v_lnames == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1426, __pyx_L1_error)
      }
      __pyx_t_8 = (__pyx_v_n - __pyx_v_i);
      __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_lnames, __pyx_t_8, int64_t, 1, __Pyx_PyInt_From_int64_t, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_t_3, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1426, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_2) {

        /* "flitter/language/tree.pyx":1427
 *             for i in range(len(lnames)):
 *                 if name == <str>lnames[n-i]:
 *                     program.local_load(i)             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
        __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->local_load(__pyx_v_program, __pyx_v_i, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "flitter/language/tree.pyx":1428
 *                 if name == <str>lnames[n-i]:
 *                     program.local_load(i)
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{name}'")
 */
        goto __pyx_L6_break;

        /* "flitter/language/tree.pyx":1426
 *         for name in captures:
 *             for i in range(len(lnames)):
 *                 if name == <str>lnames[n-i]:             # <<<<<<<<<<<<<<
 *                     program.local_load(i)
 *                     break
 */
      }
    }
    /*else*/ {

      /* "flitter/language/tree.pyx":1430
 *                     break
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{name}'")             # <<<<<<<<<<<<<<
 *                 program.literal(null_)
 *             function_lnames.append(name)
 */
      if (unlikely(__pyx_v_program->compiler_errors == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 1430, __pyx_L1_error)
      }
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Unbound_name);
      __pyx_t_9 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_Unbound_name);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Unbound_name);
      __pyx_t_11 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_9 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__3);
      __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = PySet_Add(__pyx_v_program->compiler_errors, __pyx_t_11); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "flitter/language/tree.pyx":1431
 *             else:
 *                 program.compiler_errors.add(f"Unbound name '{name}'")
 *                 program.literal(null_)             # <<<<<<<<<<<<<<
 *             function_lnames.append(name)
 *         function_lnames.append(self.name)
 */
      __pyx_t_11 = ((PyObject *)__pyx_v_7flitter_5model_null_);
      __Pyx_INCREF(__pyx_t_11);
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_11, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L6_break:;

    /* "flitter/language/tree.pyx":1432
 *                 program.compiler_errors.add(f"Unbound name '{name}'")
 *                 program.literal(null_)
 *             function_lnames.append(name)             # <<<<<<<<<<<<<<
 *         function_lnames.append(self.name)
 *         cdef list parameters = []
 */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_function_lnames, __pyx_v_name); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1432, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1424
 *         cdef tuple captures = self.captures if self.captures is not None else tuple(lnames)
 *         cdef int64_t i, n=len(lnames)-1
 *         for name in captures:             # <<<<<<<<<<<<<<
 *             for i in range(len(lnames)):
 *                 if name == <str>lnames[n-i]:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1433
 *                 program.literal(null_)
 *             function_lnames.append(name)
 *         function_lnames.append(self.name)             # <<<<<<<<<<<<<<
 *         cdef list parameters = []
 *         cdef Binding parameter
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_function_lnames, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1434
 *             function_lnames.append(name)
 *         function_lnames.append(self.name)
 *         cdef list parameters = []             # <<<<<<<<<<<<<<
 *         cdef Binding parameter
 *         for parameter in self.parameters:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parameters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1436
 *         cdef list parameters = []
 *         cdef Binding parameter
 *         for parameter in self.parameters:             # <<<<<<<<<<<<<<
 *             parameters.append(parameter.name)
 *             if parameter.expr is None:
 */
  if (unlikely(__pyx_v_self->parameters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1436, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->parameters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1436, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1436, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1436, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_parameter, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1437
 *         cdef Binding parameter
 *         for parameter in self.parameters:
 *             parameters.append(parameter.name)             # <<<<<<<<<<<<<<
 *             if parameter.expr is None:
 *                 program.literal(null_)
 */
    __pyx_t_3 = __pyx_v_parameter->name;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1437, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1438
 *         for parameter in self.parameters:
 *             parameters.append(parameter.name)
 *             if parameter.expr is None:             # <<<<<<<<<<<<<<
 *                 program.literal(null_)
 *             else:
 */
    __pyx_t_2 = (((PyObject *)__pyx_v_parameter->expr) == Py_None);
    if (__pyx_t_2) {

      /* "flitter/language/tree.pyx":1439
 *             parameters.append(parameter.name)
 *             if parameter.expr is None:
 *                 program.literal(null_)             # <<<<<<<<<<<<<<
 *             else:
 *                 parameter.expr._compile(program, lnames)
 */
      __pyx_t_3 = ((PyObject *)__pyx_v_7flitter_5model_null_);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_11 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->literal(__pyx_v_program, __pyx_t_3, 0)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "flitter/language/tree.pyx":1438
 *         for parameter in self.parameters:
 *             parameters.append(parameter.name)
 *             if parameter.expr is None:             # <<<<<<<<<<<<<<
 *                 program.literal(null_)
 *             else:
 */
      goto __pyx_L11;
    }

    /* "flitter/language/tree.pyx":1441
 *                 program.literal(null_)
 *             else:
 *                 parameter.expr._compile(program, lnames)             # <<<<<<<<<<<<<<
 *             function_lnames.append(parameter.name)
 *         START = program.new_label()
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_parameter->expr->__pyx_vtab)->_compile(__pyx_v_parameter->expr, __pyx_v_program, __pyx_v_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L1_error)
    }
    __pyx_L11:;

    /* "flitter/language/tree.pyx":1442
 *             else:
 *                 parameter.expr._compile(program, lnames)
 *             function_lnames.append(parameter.name)             # <<<<<<<<<<<<<<
 *         START = program.new_label()
 *         END = program.new_label()
 */
    __pyx_t_11 = __pyx_v_parameter->name;
    __Pyx_INCREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_function_lnames, __pyx_t_11); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1442, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "flitter/language/tree.pyx":1436
 *         cdef list parameters = []
 *         cdef Binding parameter
 *         for parameter in self.parameters:             # <<<<<<<<<<<<<<
 *             parameters.append(parameter.name)
 *             if parameter.expr is None:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1443
 *                 parameter.expr._compile(program, lnames)
 *             function_lnames.append(parameter.name)
 *         START = program.new_label()             # <<<<<<<<<<<<<<
 *         END = program.new_label()
 *         program.jump(END)
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L1_error)
  __pyx_v_START = __pyx_t_7;

  /* "flitter/language/tree.pyx":1444
 *             function_lnames.append(parameter.name)
 *         START = program.new_label()
 *         END = program.new_label()             # <<<<<<<<<<<<<<
 *         program.jump(END)
 *         program.label(START)
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->new_label(__pyx_v_program, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1444, __pyx_L1_error)
  __pyx_v_END = __pyx_t_7;

  /* "flitter/language/tree.pyx":1445
 *         START = program.new_label()
 *         END = program.new_label()
 *         program.jump(END)             # <<<<<<<<<<<<<<
 *         program.label(START)
 *         self.body._compile(program, function_lnames)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->jump(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1446
 *         END = program.new_label()
 *         program.jump(END)
 *         program.label(START)             # <<<<<<<<<<<<<<
 *         self.body._compile(program, function_lnames)
 *         program.exit()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_START, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1447
 *         program.jump(END)
 *         program.label(START)
 *         self.body._compile(program, function_lnames)             # <<<<<<<<<<<<<<
 *         program.exit()
 *         program.label(END)
 */
  ((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_compile(__pyx_v_self->body, __pyx_v_program, __pyx_v_function_lnames); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1447, __pyx_L1_error)

  /* "flitter/language/tree.pyx":1448
 *         program.label(START)
 *         self.body._compile(program, function_lnames)
 *         program.exit()             # <<<<<<<<<<<<<<
 *         program.label(END)
 *         program.func(START, self.name, tuple(parameters), len(captures))
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->exit(__pyx_v_program, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1449
 *         self.body._compile(program, function_lnames)
 *         program.exit()
 *         program.label(END)             # <<<<<<<<<<<<<<
 *         program.func(START, self.name, tuple(parameters), len(captures))
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->label(__pyx_v_program, __pyx_v_END, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1450
 *         program.exit()
 *         program.label(END)
 *         program.func(START, self.name, tuple(parameters), len(captures))             # <<<<<<<<<<<<<<
 * 
 *     cdef Expression _simplify(self, Context context):
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_11 = PyList_AsTuple(__pyx_v_parameters); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (unlikely(__pyx_v_captures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1450, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v_captures); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1450, __pyx_L1_error)
  __pyx_t_13.__pyx_n = 1;
  __pyx_t_13.ncaptures = __pyx_t_4;
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_2vm_Program *)__pyx_v_program->__pyx_vtab)->func(__pyx_v_program, __pyx_v_START, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_11), 0, &__pyx_t_13)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":1419
 *         self.recursive = recursive
 * 
 *     cdef void _compile(self, Program program, list lnames):             # <<<<<<<<<<<<<<
 *         cdef str name
 *         cdef list function_lnames = []
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("flitter.language.tree.Function._compile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_function_lnames);
  __Pyx_XDECREF(__pyx_v_captures);
  __Pyx_XDECREF(__pyx_v_parameters);
  __Pyx_XDECREF((PyObject *)__pyx_v_parameter);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/language/tree.pyx":1452
 *         program.func(START, self.name, tuple(parameters), len(captures))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list parameters = []
 *         cdef Binding parameter
 */

static struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_f_7flitter_8language_4tree_8Function__simplify(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, struct __pyx_obj_7flitter_5model_Context *__pyx_v_context) {
  PyObject *__pyx_v_parameters = 0;
  struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v_parameter = 0;
  int __pyx_v_literal;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v_body = 0;
  int __pyx_v_touched;
  PyObject *__pyx_v_expr = NULL;
  PyObject *__pyx_v_saved = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_captures = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_recursive = 0;
  PyObject *__pyx_v_captures_t = 0;
  int __pyx_v_inlineable;
  struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  unsigned int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simplify", 1);

  /* "flitter/language/tree.pyx":1453
 * 
 *     cdef Expression _simplify(self, Context context):
 *         cdef list parameters = []             # <<<<<<<<<<<<<<
 *         cdef Binding parameter
 *         cdef bint literal = True
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parameters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1455
 *         cdef list parameters = []
 *         cdef Binding parameter
 *         cdef bint literal = True             # <<<<<<<<<<<<<<
 *         cdef Expression body
 *         cdef bint touched = False
 */
  __pyx_v_literal = 1;

  /* "flitter/language/tree.pyx":1457
 *         cdef bint literal = True
 *         cdef Expression body
 *         cdef bint touched = False             # <<<<<<<<<<<<<<
 *         for parameter in self.parameters:
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 */
  __pyx_v_touched = 0;

  /* "flitter/language/tree.pyx":1458
 *         cdef Expression body
 *         cdef bint touched = False
 *         for parameter in self.parameters:             # <<<<<<<<<<<<<<
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 *             if expr is not None and not isinstance(expr, Literal):
 */
  if (unlikely(__pyx_v_self->parameters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1458, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->parameters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1458, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1458, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1458, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_parameter, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1459
 *         cdef bint touched = False
 *         for parameter in self.parameters:
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None             # <<<<<<<<<<<<<<
 *             if expr is not None and not isinstance(expr, Literal):
 *                 literal = False
 */
    __pyx_t_4 = (((PyObject *)__pyx_v_parameter->expr) != Py_None);
    if (__pyx_t_4) {
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_parameter->expr->__pyx_vtab)->_simplify(__pyx_v_parameter->expr, __pyx_v_context)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_5 = 0;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_3 = Py_None;
    }
    __Pyx_XDECREF_SET(__pyx_v_expr, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1460
 *         for parameter in self.parameters:
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 *             if expr is not None and not isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *                 literal = False
 *             parameters.append(Binding(parameter.name, expr))
 */
    __pyx_t_6 = (__pyx_v_expr != Py_None);
    if (__pyx_t_6) {
    } else {
      __pyx_t_4 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_expr, __pyx_ptype_7flitter_8language_4tree_Literal); 
    __pyx_t_7 = (!__pyx_t_6);
    __pyx_t_4 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_4) {

      /* "flitter/language/tree.pyx":1461
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 *             if expr is not None and not isinstance(expr, Literal):
 *                 literal = False             # <<<<<<<<<<<<<<
 *             parameters.append(Binding(parameter.name, expr))
 *             touched |= expr is not parameter.expr
 */
      __pyx_v_literal = 0;

      /* "flitter/language/tree.pyx":1460
 *         for parameter in self.parameters:
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 *             if expr is not None and not isinstance(expr, Literal):             # <<<<<<<<<<<<<<
 *                 literal = False
 *             parameters.append(Binding(parameter.name, expr))
 */
    }

    /* "flitter/language/tree.pyx":1462
 *             if expr is not None and not isinstance(expr, Literal):
 *                 literal = False
 *             parameters.append(Binding(parameter.name, expr))             # <<<<<<<<<<<<<<
 *             touched |= expr is not parameter.expr
 *         cdef dict saved = context.names
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_parameter->name);
    __Pyx_GIVEREF(__pyx_v_parameter->name);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_parameter->name)) __PYX_ERR(0, 1462, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_expr);
    __Pyx_GIVEREF(__pyx_v_expr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_expr)) __PYX_ERR(0, 1462, __pyx_L1_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding), __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_parameters, __pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "flitter/language/tree.pyx":1463
 *                 literal = False
 *             parameters.append(Binding(parameter.name, expr))
 *             touched |= expr is not parameter.expr             # <<<<<<<<<<<<<<
 *         cdef dict saved = context.names
 *         cdef str key
 */
    __pyx_t_4 = (__pyx_v_expr != ((PyObject *)__pyx_v_parameter->expr));
    __pyx_v_touched = (__pyx_v_touched | __pyx_t_4);

    /* "flitter/language/tree.pyx":1458
 *         cdef Expression body
 *         cdef bint touched = False
 *         for parameter in self.parameters:             # <<<<<<<<<<<<<<
 *             expr = parameter.expr._simplify(context) if parameter.expr is not None else None
 *             if expr is not None and not isinstance(expr, Literal):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1464
 *             parameters.append(Binding(parameter.name, expr))
 *             touched |= expr is not parameter.expr
 *         cdef dict saved = context.names             # <<<<<<<<<<<<<<
 *         cdef str key
 *         cdef set captures = context.captures
 */
  __pyx_t_1 = __pyx_v_context->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_saved = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1466
 *         cdef dict saved = context.names
 *         cdef str key
 *         cdef set captures = context.captures             # <<<<<<<<<<<<<<
 *         context.captures = set()
 *         context.names = {}
 */
  __pyx_t_1 = __pyx_v_context->captures;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_captures = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1467
 *         cdef str key
 *         cdef set captures = context.captures
 *         context.captures = set()             # <<<<<<<<<<<<<<
 *         context.names = {}
 *         for key, value in saved.items():
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_context->captures);
  __Pyx_DECREF(__pyx_v_context->captures);
  __pyx_v_context->captures = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1468
 *         cdef set captures = context.captures
 *         context.captures = set()
 *         context.names = {}             # <<<<<<<<<<<<<<
 *         for key, value in saved.items():
 *             if value is not None and key != self.name:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1469
 *         context.captures = set()
 *         context.names = {}
 *         for key, value in saved.items():             # <<<<<<<<<<<<<<
 *             if value is not None and key != self.name:
 *                 context.names[key] = value
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_saved == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1469, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_saved, 1, __pyx_n_s_items, (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_9, &__pyx_t_2, &__pyx_t_5, &__pyx_t_3, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1470
 *         context.names = {}
 *         for key, value in saved.items():
 *             if value is not None and key != self.name:             # <<<<<<<<<<<<<<
 *                 context.names[key] = value
 *         for parameter in parameters:
 */
    __pyx_t_7 = (__pyx_v_value != Py_None);
    if (__pyx_t_7) {
    } else {
      __pyx_t_4 = __pyx_t_7;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_key, __pyx_v_self->name, Py_NE)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1470, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_7;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_4) {

      /* "flitter/language/tree.pyx":1471
 *         for key, value in saved.items():
 *             if value is not None and key != self.name:
 *                 context.names[key] = value             # <<<<<<<<<<<<<<
 *         for parameter in parameters:
 *             context.names[parameter.name] = None
 */
      if (unlikely(__pyx_v_context->names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1471, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_key, __pyx_v_value) < 0))) __PYX_ERR(0, 1471, __pyx_L1_error)

      /* "flitter/language/tree.pyx":1470
 *         context.names = {}
 *         for key, value in saved.items():
 *             if value is not None and key != self.name:             # <<<<<<<<<<<<<<
 *                 context.names[key] = value
 *         for parameter in parameters:
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1472
 *             if value is not None and key != self.name:
 *                 context.names[key] = value
 *         for parameter in parameters:             # <<<<<<<<<<<<<<
 *             context.names[parameter.name] = None
 *         body = self.body._simplify(context)
 */
  __pyx_t_1 = __pyx_v_parameters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_9 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1472, __pyx_L1_error)
      #endif
      if (__pyx_t_9 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 1472, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7flitter_8language_4tree_Binding))))) __PYX_ERR(0, 1472, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_parameter, ((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1473
 *                 context.names[key] = value
 *         for parameter in parameters:
 *             context.names[parameter.name] = None             # <<<<<<<<<<<<<<
 *         body = self.body._simplify(context)
 *         touched |= body is not self.body
 */
    if (unlikely(__pyx_v_context->names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1473, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_context->names, __pyx_v_parameter->name, Py_None) < 0))) __PYX_ERR(0, 1473, __pyx_L1_error)

    /* "flitter/language/tree.pyx":1472
 *             if value is not None and key != self.name:
 *                 context.names[key] = value
 *         for parameter in parameters:             # <<<<<<<<<<<<<<
 *             context.names[parameter.name] = None
 *         body = self.body._simplify(context)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1474
 *         for parameter in parameters:
 *             context.names[parameter.name] = None
 *         body = self.body._simplify(context)             # <<<<<<<<<<<<<<
 *         touched |= body is not self.body
 *         cdef recursive = self.name in context.captures
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_8language_4tree_Expression *)__pyx_v_self->body->__pyx_vtab)->_simplify(__pyx_v_self->body, __pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1475
 *             context.names[parameter.name] = None
 *         body = self.body._simplify(context)
 *         touched |= body is not self.body             # <<<<<<<<<<<<<<
 *         cdef recursive = self.name in context.captures
 *         if recursive:
 */
  __pyx_t_4 = (__pyx_v_body != __pyx_v_self->body);
  __pyx_v_touched = (__pyx_v_touched | __pyx_t_4);

  /* "flitter/language/tree.pyx":1476
 *         body = self.body._simplify(context)
 *         touched |= body is not self.body
 *         cdef recursive = self.name in context.captures             # <<<<<<<<<<<<<<
 *         if recursive:
 *             context.captures.discard(self.name)
 */
  if (unlikely(__pyx_v_context->captures == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1476, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PySet_ContainsTF(__pyx_v_self->name, __pyx_v_context->captures, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1476, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_recursive = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flitter/language/tree.pyx":1477
 *         touched |= body is not self.body
 *         cdef recursive = self.name in context.captures
 *         if recursive:             # <<<<<<<<<<<<<<
 *             context.captures.discard(self.name)
 *         touched |= recursive != self.recursive
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_recursive); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1477, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "flitter/language/tree.pyx":1478
 *         cdef recursive = self.name in context.captures
 *         if recursive:
 *             context.captures.discard(self.name)             # <<<<<<<<<<<<<<
 *         touched |= recursive != self.recursive
 *         cdef tuple captures_t = tuple(context.captures)
 */
    if (unlikely(__pyx_v_context->captures == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
      __PYX_ERR(0, 1478, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->name;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PySet_Discard(__pyx_v_context->captures, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1478, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/language/tree.pyx":1477
 *         touched |= body is not self.body
 *         cdef recursive = self.name in context.captures
 *         if recursive:             # <<<<<<<<<<<<<<
 *             context.captures.discard(self.name)
 *         touched |= recursive != self.recursive
 */
  }

  /* "flitter/language/tree.pyx":1479
 *         if recursive:
 *             context.captures.discard(self.name)
 *         touched |= recursive != self.recursive             # <<<<<<<<<<<<<<
 *         cdef tuple captures_t = tuple(context.captures)
 *         cdef bint inlineable = literal and not captures_t
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_touched); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->recursive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_recursive, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceOr(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_touched = __pyx_t_4;

  /* "flitter/language/tree.pyx":1480
 *             context.captures.discard(self.name)
 *         touched |= recursive != self.recursive
 *         cdef tuple captures_t = tuple(context.captures)             # <<<<<<<<<<<<<<
 *         cdef bint inlineable = literal and not captures_t
 *         touched |= inlineable != self.inlineable
 */
  __pyx_t_3 = PySequence_Tuple(__pyx_v_context->captures); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_captures_t = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":1481
 *         touched |= recursive != self.recursive
 *         cdef tuple captures_t = tuple(context.captures)
 *         cdef bint inlineable = literal and not captures_t             # <<<<<<<<<<<<<<
 *         touched |= inlineable != self.inlineable
 *         context.names = dict(saved)
 */
  if (__pyx_v_literal) {
  } else {
    __pyx_t_4 = __pyx_v_literal;
    goto __pyx_L18_bool_binop_done;
  }
  __pyx_t_7 = (PyTuple_GET_SIZE(__pyx_v_captures_t) != 0);
  __pyx_t_6 = (!__pyx_t_7);
  __pyx_t_4 = __pyx_t_6;
  __pyx_L18_bool_binop_done:;
  __pyx_v_inlineable = __pyx_t_4;

  /* "flitter/language/tree.pyx":1482
 *         cdef tuple captures_t = tuple(context.captures)
 *         cdef bint inlineable = literal and not captures_t
 *         touched |= inlineable != self.inlineable             # <<<<<<<<<<<<<<
 *         context.names = dict(saved)
 *         if captures is not None:
 */
  __pyx_v_touched = (__pyx_v_touched | (__pyx_v_inlineable != __pyx_v_self->inlineable));

  /* "flitter/language/tree.pyx":1483
 *         cdef bint inlineable = literal and not captures_t
 *         touched |= inlineable != self.inlineable
 *         context.names = dict(saved)             # <<<<<<<<<<<<<<
 *         if captures is not None:
 *             context.captures.difference_update(saved)
 */
  if (unlikely(__pyx_v_saved == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(0, 1483, __pyx_L1_error)
  }
  __pyx_t_3 = PyDict_Copy(__pyx_v_saved); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_context->names);
  __Pyx_DECREF(__pyx_v_context->names);
  __pyx_v_context->names = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":1484
 *         touched |= inlineable != self.inlineable
 *         context.names = dict(saved)
 *         if captures is not None:             # <<<<<<<<<<<<<<
 *             context.captures.difference_update(saved)
 *             captures.update(context.captures)
 */
  __pyx_t_4 = (__pyx_v_captures != ((PyObject*)Py_None));
  if (__pyx_t_4) {

    /* "flitter/language/tree.pyx":1485
 *         context.names = dict(saved)
 *         if captures is not None:
 *             context.captures.difference_update(saved)             # <<<<<<<<<<<<<<
 *             captures.update(context.captures)
 *         context.captures = captures
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_context->captures, __pyx_n_s_difference_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = NULL;
    __pyx_t_12 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_12 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_saved};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1486
 *         if captures is not None:
 *             context.captures.difference_update(saved)
 *             captures.update(context.captures)             # <<<<<<<<<<<<<<
 *         context.captures = captures
 *         touched |= captures_t != self.captures
 */
    __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_update, __pyx_v_captures, __pyx_v_context->captures); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flitter/language/tree.pyx":1484
 *         touched |= inlineable != self.inlineable
 *         context.names = dict(saved)
 *         if captures is not None:             # <<<<<<<<<<<<<<
 *             context.captures.difference_update(saved)
 *             captures.update(context.captures)
 */
  }

  /* "flitter/language/tree.pyx":1487
 *             context.captures.difference_update(saved)
 *             captures.update(context.captures)
 *         context.captures = captures             # <<<<<<<<<<<<<<
 *         touched |= captures_t != self.captures
 *         if not touched:
 */
  __Pyx_INCREF(__pyx_v_captures);
  __Pyx_GIVEREF(__pyx_v_captures);
  __Pyx_GOTREF(__pyx_v_context->captures);
  __Pyx_DECREF(__pyx_v_context->captures);
  __pyx_v_context->captures = __pyx_v_captures;

  /* "flitter/language/tree.pyx":1488
 *             captures.update(context.captures)
 *         context.captures = captures
 *         touched |= captures_t != self.captures             # <<<<<<<<<<<<<<
 *         if not touched:
 *             return self
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_touched); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_captures_t, __pyx_v_self->captures, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __pyx_t_1 = PyNumber_InPlaceOr(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_touched = __pyx_t_4;

  /* "flitter/language/tree.pyx":1489
 *         context.captures = captures
 *         touched |= captures_t != self.captures
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)
 */
  __pyx_t_4 = (!__pyx_v_touched);
  if (__pyx_t_4) {

    /* "flitter/language/tree.pyx":1490
 *         touched |= captures_t != self.captures
 *         if not touched:
 *             return self             # <<<<<<<<<<<<<<
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v_self);
    goto __pyx_L0;

    /* "flitter/language/tree.pyx":1489
 *         context.captures = captures
 *         touched |= captures_t != self.captures
 *         if not touched:             # <<<<<<<<<<<<<<
 *             return self
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)
 */
  }

  /* "flitter/language/tree.pyx":1491
 *         if not touched:
 *             return self
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyList_AsTuple(__pyx_v_parameters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_inlineable); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->name)) __PYX_ERR(0, 1491, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_body))) __PYX_ERR(0, 1491, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_captures_t);
  __Pyx_GIVEREF(__pyx_v_captures_t);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_captures_t)) __PYX_ERR(0, 1491, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_5)) __PYX_ERR(0, 1491, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_recursive);
  __Pyx_GIVEREF(__pyx_v_recursive);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_recursive)) __PYX_ERR(0, 1491, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Function), __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1452
 *         program.func(START, self.name, tuple(parameters), len(captures))
 * 
 *     cdef Expression _simplify(self, Context context):             # <<<<<<<<<<<<<<
 *         cdef list parameters = []
 *         cdef Binding parameter
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.language.tree.Function._simplify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parameters);
  __Pyx_XDECREF((PyObject *)__pyx_v_parameter);
  __Pyx_XDECREF((PyObject *)__pyx_v_body);
  __Pyx_XDECREF(__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_saved);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_captures);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_recursive);
  __Pyx_XDECREF(__pyx_v_captures_t);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1493
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Function({self.name!r}, {self.parameters!r}, {self.body!r}, {self.captures!r}, {self.inlineable!r}, {self.recursive!r})'
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_2__repr__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_2__repr__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "flitter/language/tree.pyx":1494
 * 
 *     def __repr__(self):
 *         return f'Function({self.name!r}, {self.parameters!r}, {self.body!r}, {self.captures!r}, {self.inlineable!r}, {self.recursive!r})'             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Function);
  __pyx_t_2 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_Function);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Function);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->name), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->parameters), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_v_self->body)), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->captures), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->inlineable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->recursive); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 13, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flitter/language/tree.pyx":1493
 *         return Function(self.name, tuple(parameters), body, captures_t, inlineable, recursive)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'Function({self.name!r}, {self.parameters!r}, {self.body!r}, {self.captures!r}, {self.inlineable!r}, {self.recursive!r})'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.Function.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1404
 * 
 * cdef class Function(Expression):
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 *     cdef readonly tuple parameters
 *     cdef readonly Expression body
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_4name___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4name___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1405
 * cdef class Function(Expression):
 *     cdef readonly str name
 *     cdef readonly tuple parameters             # <<<<<<<<<<<<<<
 *     cdef readonly Expression body
 *     cdef readonly tuple captures
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_10parameters_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_10parameters_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_10parameters___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_10parameters___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __pyx_r = __pyx_v_self->parameters;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1406
 *     cdef readonly str name
 *     cdef readonly tuple parameters
 *     cdef readonly Expression body             # <<<<<<<<<<<<<<
 *     cdef readonly tuple captures
 *     cdef readonly bint inlineable
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_4body_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_4body___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4body___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __pyx_r = ((PyObject *)__pyx_v_self->body);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1407
 *     cdef readonly tuple parameters
 *     cdef readonly Expression body
 *     cdef readonly tuple captures             # <<<<<<<<<<<<<<
 *     cdef readonly bint inlineable
 *     cdef readonly bint recursive
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_8captures_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_8captures_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_8captures___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_8captures___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->captures);
  __pyx_r = __pyx_v_self->captures;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1408
 *     cdef readonly Expression body
 *     cdef readonly tuple captures
 *     cdef readonly bint inlineable             # <<<<<<<<<<<<<<
 *     cdef readonly bint recursive
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_10inlineable_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_10inlineable_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_10inlineable___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_10inlineable___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->inlineable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Function.inlineable.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/language/tree.pyx":1409
 *     cdef readonly tuple captures
 *     cdef readonly bint inlineable
 *     cdef readonly bint recursive             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str name, tuple parameters, Expression body, tuple captures=None, bint inlineable=False, bint recursive=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_9recursive_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_9recursive_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_9recursive___get__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_9recursive___get__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->recursive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Function.recursive.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Function_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Function_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_4__reduce_cython__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_4__reduce_cython__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.body, self.captures, self.inlineable, self.name, self.parameters, self.recursive)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->inlineable); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->recursive); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self->body);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->body);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->body))) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->captures);
  __Pyx_GIVEREF(__pyx_v_self->captures);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->captures)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->name)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->parameters);
  __Pyx_GIVEREF(__pyx_v_self->parameters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->parameters)) __PYX_ERR(2, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.body, self.captures, self.inlineable, self.name, self.parameters, self.recursive)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.body, self.captures, self.inlineable, self.name, self.parameters, self.recursive)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(2, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.body is not None or self.captures is not None or self.name is not None or self.parameters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.body, self.captures, self.inlineable, self.name, self.parameters, self.recursive)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.body is not None or self.captures is not None or self.name is not None or self.parameters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, None), state
 */
  /*else*/ {
    __pyx_t_5 = (((PyObject *)__pyx_v_self->body) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->captures != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->name != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->parameters != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.captures is not None or self.name is not None or self.parameters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.body is not None or self.captures is not None or self.name is not None or self.parameters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Function); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_256597255);
    __Pyx_GIVEREF(__pyx_int_256597255);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_256597255)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(2, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.body is not None or self.captures is not None or self.name is not None or self.parameters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, None), state
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Function); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_256597255);
    __Pyx_GIVEREF(__pyx_int_256597255);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_256597255)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flitter.language.tree.Function.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_8Function_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Function_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_8Function_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(2, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.Function.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8Function_6__setstate_cython__(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8Function_6__setstate_cython__(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Function__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.language.tree.Function.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Expression(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_1__pyx_unpickle_Expression(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_1__pyx_unpickle_Expression = {"__pyx_unpickle_Expression", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_1__pyx_unpickle_Expression, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_1__pyx_unpickle_Expression(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Expression (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Expression", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Expression", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Expression") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Expression", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree___pyx_unpickle_Expression(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree___pyx_unpickle_Expression(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Expression", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__7, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Expression.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Expression.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Expression.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Expression.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Expression.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Expression__set_state(((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Expression.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Expression(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Expression__set_state(struct __pyx_obj_7flitter_8language_4tree_Expression *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Expression__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Expression__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Top(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_3__pyx_unpickle_Top(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_3__pyx_unpickle_Top = {"__pyx_unpickle_Top", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3__pyx_unpickle_Top, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_3__pyx_unpickle_Top(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Top (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Top", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Top", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Top") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Top", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_2__pyx_unpickle_Top(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_2__pyx_unpickle_Top(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Top", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb892498, 0x4dd1ade, 0x1c65a9c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb892498, 0x4dd1ade, 0x1c65a9c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 *     __pyx_result = Top.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb892498, 0x4dd1ade, 0x1c65a9c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Top.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb892498, 0x4dd1ade, 0x1c65a9c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 *     __pyx_result = Top.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Top), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 *     __pyx_result = Top.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Top.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Top__set_state(((struct __pyx_obj_7flitter_8language_4tree_Top *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb892498, 0x4dd1ade, 0x1c65a9c) = (body, dependencies, pragmas))" % __pyx_checksum
 *     __pyx_result = Top.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Top(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Top__set_state(struct __pyx_obj_7flitter_8language_4tree_Top *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Top__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->body);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->body);
  __pyx_v___pyx_result->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->dependencies);
  __Pyx_DECREF(__pyx_v___pyx_result->dependencies);
  __pyx_v___pyx_result->dependencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->pragmas);
  __Pyx_DECREF(__pyx_v___pyx_result->pragmas);
  __pyx_v___pyx_result->pragmas = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Top__set_state(<Top> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Top__set_state(Top __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.dependencies = __pyx_state[1]; __pyx_result.pragmas = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Top__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Export(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_5__pyx_unpickle_Export(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_5__pyx_unpickle_Export = {"__pyx_unpickle_Export", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5__pyx_unpickle_Export, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_5__pyx_unpickle_Export(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Export (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Export", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Export", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Export") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Export", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Export", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_4__pyx_unpickle_Export(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_4__pyx_unpickle_Export(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Export", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x123fe55, 0xf507427, 0x2ea01c6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__10, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x123fe55, 0xf507427, 0x2ea01c6):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 *     __pyx_result = Export.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x123fe55, 0xf507427, 0x2ea01c6):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Export.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x123fe55, 0xf507427, 0x2ea01c6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 *     __pyx_result = Export.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Export), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 *     __pyx_result = Export.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Export.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Export__set_state(((struct __pyx_obj_7flitter_8language_4tree_Export *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x123fe55, 0xf507427, 0x2ea01c6) = (static_exports))" % __pyx_checksum
 *     __pyx_result = Export.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):
 *     __pyx_result.static_exports = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Export(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Export", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Export__set_state(struct __pyx_obj_7flitter_8language_4tree_Export *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Export__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):
 *     __pyx_result.static_exports = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->static_exports);
  __Pyx_DECREF(__pyx_v___pyx_result->static_exports);
  __pyx_v___pyx_result->static_exports = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Export__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Import(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_7__pyx_unpickle_Import(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_7__pyx_unpickle_Import = {"__pyx_unpickle_Import", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7__pyx_unpickle_Import, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_7__pyx_unpickle_Import(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Import (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Import", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Import", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Import") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Import", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Import", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_6__pyx_unpickle_Import(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_6__pyx_unpickle_Import(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Import", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3ebdbb2, 0x256dc66, 0x8e6e663):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__11, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3ebdbb2, 0x256dc66, 0x8e6e663):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 *     __pyx_result = Import.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3ebdbb2, 0x256dc66, 0x8e6e663):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Import.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3ebdbb2, 0x256dc66, 0x8e6e663):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 *     __pyx_result = Import.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Import), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 *     __pyx_result = Import.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Import.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Import__set_state(((struct __pyx_obj_7flitter_8language_4tree_Import *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3ebdbb2, 0x256dc66, 0x8e6e663) = (expr, filename, names))" % __pyx_checksum
 *     __pyx_result = Import.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Import(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Import", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Import__set_state(struct __pyx_obj_7flitter_8language_4tree_Import *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Import__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->expr);
  __pyx_v___pyx_result->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->filename);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->filename);
  __pyx_v___pyx_result->filename = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->names);
  __Pyx_DECREF(__pyx_v___pyx_result->names);
  __pyx_v___pyx_result->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Import__set_state(<Import> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Import__set_state(Import __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.filename = __pyx_state[1]; __pyx_result.names = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Import__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Sequence(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_9__pyx_unpickle_Sequence(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_9__pyx_unpickle_Sequence = {"__pyx_unpickle_Sequence", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_9__pyx_unpickle_Sequence, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_9__pyx_unpickle_Sequence(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Sequence (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Sequence", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Sequence", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Sequence") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Sequence", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_8__pyx_unpickle_Sequence(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_8__pyx_unpickle_Sequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Sequence", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x18bc300, 0xfa24806, 0x1476dfb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__12, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x18bc300, 0xfa24806, 0x1476dfb):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 *     __pyx_result = Sequence.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x18bc300, 0xfa24806, 0x1476dfb):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Sequence.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x18bc300, 0xfa24806, 0x1476dfb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 *     __pyx_result = Sequence.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 *     __pyx_result = Sequence.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Sequence.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Sequence__set_state(((struct __pyx_obj_7flitter_8language_4tree_Sequence *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x18bc300, 0xfa24806, 0x1476dfb) = (expressions))" % __pyx_checksum
 *     __pyx_result = Sequence.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):
 *     __pyx_result.expressions = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Sequence(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Sequence__set_state(struct __pyx_obj_7flitter_8language_4tree_Sequence *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Sequence__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):
 *     __pyx_result.expressions = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->expressions);
  __Pyx_DECREF(__pyx_v___pyx_result->expressions);
  __pyx_v___pyx_result->expressions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Sequence__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Literal(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_11__pyx_unpickle_Literal(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_11__pyx_unpickle_Literal = {"__pyx_unpickle_Literal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11__pyx_unpickle_Literal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_11__pyx_unpickle_Literal(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Literal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Literal", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Literal", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Literal") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Literal", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_10__pyx_unpickle_Literal(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_10__pyx_unpickle_Literal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Literal", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__13, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Literal.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Literal.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Literal.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Literal.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Literal.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Literal__set_state(((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Literal.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Literal(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Literal__set_state(struct __pyx_obj_7flitter_8language_4tree_Literal *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Literal__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->value);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Literal__set_state(<Literal> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Literal__set_state(Literal __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Literal__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Name(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_13__pyx_unpickle_Name(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_13__pyx_unpickle_Name = {"__pyx_unpickle_Name", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_13__pyx_unpickle_Name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_13__pyx_unpickle_Name(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Name", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Name", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Name") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Name", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_12__pyx_unpickle_Name(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_12__pyx_unpickle_Name(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Name", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__14, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Name.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Name.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Name.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Name), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Name.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Name.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Name__set_state(((struct __pyx_obj_7flitter_8language_4tree_Name *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Name.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Name(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Name__set_state(struct __pyx_obj_7flitter_8language_4tree_Name *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Name__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Name__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Lookup(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_15__pyx_unpickle_Lookup(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_15__pyx_unpickle_Lookup = {"__pyx_unpickle_Lookup", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_15__pyx_unpickle_Lookup, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_15__pyx_unpickle_Lookup(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Lookup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Lookup", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Lookup", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Lookup") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Lookup", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_14__pyx_unpickle_Lookup(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_14__pyx_unpickle_Lookup(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Lookup", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2c70e12, 0xa62f222, 0x3c6e0b8):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__15, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2c70e12, 0xa62f222, 0x3c6e0b8):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 *     __pyx_result = Lookup.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x2c70e12, 0xa62f222, 0x3c6e0b8):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Lookup.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2c70e12, 0xa62f222, 0x3c6e0b8):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 *     __pyx_result = Lookup.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Lookup), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 *     __pyx_result = Lookup.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Lookup.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Lookup__set_state(((struct __pyx_obj_7flitter_8language_4tree_Lookup *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2c70e12, 0xa62f222, 0x3c6e0b8) = (key))" % __pyx_checksum
 *     __pyx_result = Lookup.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):
 *     __pyx_result.key = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Lookup(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.key = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Lookup__set_state(struct __pyx_obj_7flitter_8language_4tree_Lookup *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Lookup__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):
 *     __pyx_result.key = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->key);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):
 *     __pyx_result.key = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.key = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):
 *     __pyx_result.key = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Lookup__set_state(<Lookup> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Lookup__set_state(Lookup __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.key = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Lookup__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Range(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_17__pyx_unpickle_Range(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_17__pyx_unpickle_Range = {"__pyx_unpickle_Range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_17__pyx_unpickle_Range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_17__pyx_unpickle_Range(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Range (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Range", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Range", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Range") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Range", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_16__pyx_unpickle_Range(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_16__pyx_unpickle_Range(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Range", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfbe38c8, 0x8cfc505, 0x53c80a9):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__16, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfbe38c8, 0x8cfc505, 0x53c80a9):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 *     __pyx_result = Range.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xfbe38c8, 0x8cfc505, 0x53c80a9):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Range.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfbe38c8, 0x8cfc505, 0x53c80a9):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 *     __pyx_result = Range.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Range), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 *     __pyx_result = Range.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Range.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Range__set_state(((struct __pyx_obj_7flitter_8language_4tree_Range *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfbe38c8, 0x8cfc505, 0x53c80a9) = (start, step, stop))" % __pyx_checksum
 *     __pyx_result = Range.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Range(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Range__set_state(struct __pyx_obj_7flitter_8language_4tree_Range *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Range__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->start);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->start);
  __pyx_v___pyx_result->start = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->step);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->step);
  __pyx_v___pyx_result->step = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->stop);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->stop);
  __pyx_v___pyx_result->stop = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Range__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UnaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_19__pyx_unpickle_UnaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_19__pyx_unpickle_UnaryOperation = {"__pyx_unpickle_UnaryOperation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_19__pyx_unpickle_UnaryOperation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_19__pyx_unpickle_UnaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnaryOperation (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnaryOperation", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnaryOperation", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UnaryOperation") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnaryOperation", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_UnaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_18__pyx_unpickle_UnaryOperation(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_18__pyx_unpickle_UnaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnaryOperation", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_UnaryOperation__set_state(((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = UnaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UnaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_UnaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_UnaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnaryOperation__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->expr);
  __pyx_v___pyx_result->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UnaryOperation__set_state(<UnaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UnaryOperation__set_state(UnaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_UnaryOperation__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Negative(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_21__pyx_unpickle_Negative(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_21__pyx_unpickle_Negative = {"__pyx_unpickle_Negative", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_21__pyx_unpickle_Negative, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_21__pyx_unpickle_Negative(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Negative (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Negative", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Negative", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Negative") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Negative", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Negative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_20__pyx_unpickle_Negative(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_20__pyx_unpickle_Negative(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Negative", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Negative.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Negative.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Negative.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Negative.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Negative.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Negative__set_state(((struct __pyx_obj_7flitter_8language_4tree_Negative *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Negative.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Negative(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Negative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Negative__set_state(struct __pyx_obj_7flitter_8language_4tree_Negative *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Negative__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Negative__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Positive(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_23__pyx_unpickle_Positive(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_23__pyx_unpickle_Positive = {"__pyx_unpickle_Positive", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_23__pyx_unpickle_Positive, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_23__pyx_unpickle_Positive(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Positive (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Positive", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Positive", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Positive") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Positive", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Positive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_22__pyx_unpickle_Positive(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_22__pyx_unpickle_Positive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Positive", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Positive.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Positive.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Positive.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Positive.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Positive.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Positive__set_state(((struct __pyx_obj_7flitter_8language_4tree_Positive *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Positive.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Positive(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Positive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Positive__set_state(struct __pyx_obj_7flitter_8language_4tree_Positive *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Positive__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Positive__set_state(<Positive> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Positive__set_state(Positive __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Positive__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Not(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_25__pyx_unpickle_Not(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_25__pyx_unpickle_Not = {"__pyx_unpickle_Not", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_25__pyx_unpickle_Not, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_25__pyx_unpickle_Not(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Not (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Not", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Not", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Not") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Not", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Not", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_24__pyx_unpickle_Not(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_24__pyx_unpickle_Not(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Not", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Not.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Not.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Not.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Not), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Not.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Not.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Not__set_state(((struct __pyx_obj_7flitter_8language_4tree_Not *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Not.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Not(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Not", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Not__set_state(struct __pyx_obj_7flitter_8language_4tree_Not *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Not__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Not__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Ceil(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_27__pyx_unpickle_Ceil(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_27__pyx_unpickle_Ceil = {"__pyx_unpickle_Ceil", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_27__pyx_unpickle_Ceil, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_27__pyx_unpickle_Ceil(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Ceil (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Ceil", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Ceil", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Ceil") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Ceil", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Ceil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_26__pyx_unpickle_Ceil(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_26__pyx_unpickle_Ceil(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Ceil", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Ceil.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Ceil.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Ceil.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Ceil), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Ceil.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Ceil.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Ceil__set_state(((struct __pyx_obj_7flitter_8language_4tree_Ceil *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Ceil.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Ceil(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Ceil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Ceil__set_state(struct __pyx_obj_7flitter_8language_4tree_Ceil *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Ceil__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Ceil__set_state(<Ceil> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Ceil__set_state(Ceil __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Ceil__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Floor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_29__pyx_unpickle_Floor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_29__pyx_unpickle_Floor = {"__pyx_unpickle_Floor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_29__pyx_unpickle_Floor, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_29__pyx_unpickle_Floor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Floor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Floor", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Floor", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Floor") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Floor", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Floor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_28__pyx_unpickle_Floor(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_28__pyx_unpickle_Floor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Floor", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Floor.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Floor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Floor.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Floor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Floor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Floor__set_state(((struct __pyx_obj_7flitter_8language_4tree_Floor *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Floor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Floor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Floor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Floor__set_state(struct __pyx_obj_7flitter_8language_4tree_Floor *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Floor__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Floor__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Fract(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_31__pyx_unpickle_Fract(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_31__pyx_unpickle_Fract = {"__pyx_unpickle_Fract", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_31__pyx_unpickle_Fract, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_31__pyx_unpickle_Fract(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Fract (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Fract", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Fract", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Fract") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Fract", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Fract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_30__pyx_unpickle_Fract(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_30__pyx_unpickle_Fract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Fract", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__17, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Fract.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Fract.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x60ba1dd, 0x0fd1860, 0x8d1dc3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Fract.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Fract.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Fract.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Fract__set_state(((struct __pyx_obj_7flitter_8language_4tree_Fract *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x60ba1dd, 0x0fd1860, 0x8d1dc3c) = (expr))" % __pyx_checksum
 *     __pyx_result = Fract.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Fract(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Fract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Fract__set_state(struct __pyx_obj_7flitter_8language_4tree_Fract *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Fract__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.expr);
  __pyx_v___pyx_result->__pyx_base.expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Fract__set_state(<Fract> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Fract__set_state(Fract __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Fract__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BinaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_33__pyx_unpickle_BinaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_33__pyx_unpickle_BinaryOperation = {"__pyx_unpickle_BinaryOperation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_33__pyx_unpickle_BinaryOperation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_33__pyx_unpickle_BinaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryOperation (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryOperation", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryOperation", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_BinaryOperation") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryOperation", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_BinaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_32__pyx_unpickle_BinaryOperation(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_32__pyx_unpickle_BinaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryOperation", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_BinaryOperation__set_state(((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = BinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BinaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_BinaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_BinaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryOperation__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->left);
  __pyx_v___pyx_result->left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->right);
  __pyx_v___pyx_result->right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_BinaryOperation__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MathsBinaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_35__pyx_unpickle_MathsBinaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_35__pyx_unpickle_MathsBinaryOperation = {"__pyx_unpickle_MathsBinaryOperation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_35__pyx_unpickle_MathsBinaryOperation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_35__pyx_unpickle_MathsBinaryOperation(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MathsBinaryOperation (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MathsBinaryOperation", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MathsBinaryOperation", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_MathsBinaryOperation") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MathsBinaryOperation", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_MathsBinaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_34__pyx_unpickle_MathsBinaryOperation(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_34__pyx_unpickle_MathsBinaryOperation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MathsBinaryOperation", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_MathsBinaryOperation__set_state(((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = MathsBinaryOperation.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MathsBinaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_MathsBinaryOperation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_MathsBinaryOperation__set_state(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MathsBinaryOperation__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MathsBinaryOperation__set_state(<MathsBinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MathsBinaryOperation__set_state(MathsBinaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_MathsBinaryOperation__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Add(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_37__pyx_unpickle_Add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_37__pyx_unpickle_Add = {"__pyx_unpickle_Add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_37__pyx_unpickle_Add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_37__pyx_unpickle_Add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Add", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Add", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Add") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Add", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_36__pyx_unpickle_Add(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_36__pyx_unpickle_Add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Add", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Add.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Add.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Add.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Add.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Add.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Add__set_state(((struct __pyx_obj_7flitter_8language_4tree_Add *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Add.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Add(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Add__set_state(struct __pyx_obj_7flitter_8language_4tree_Add *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Add__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Add__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Subtract(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_39__pyx_unpickle_Subtract(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_39__pyx_unpickle_Subtract = {"__pyx_unpickle_Subtract", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_39__pyx_unpickle_Subtract, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_39__pyx_unpickle_Subtract(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Subtract (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Subtract", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Subtract", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Subtract") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Subtract", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Subtract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_38__pyx_unpickle_Subtract(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_38__pyx_unpickle_Subtract(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Subtract", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Subtract.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Subtract.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Subtract.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Subtract), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Subtract.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Subtract.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Subtract__set_state(((struct __pyx_obj_7flitter_8language_4tree_Subtract *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Subtract.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Subtract(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Subtract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Subtract__set_state(struct __pyx_obj_7flitter_8language_4tree_Subtract *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Subtract__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Subtract__set_state(<Subtract> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Subtract__set_state(Subtract __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Subtract__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Multiply(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_41__pyx_unpickle_Multiply(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_41__pyx_unpickle_Multiply = {"__pyx_unpickle_Multiply", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_41__pyx_unpickle_Multiply, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_41__pyx_unpickle_Multiply(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Multiply (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Multiply", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Multiply", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Multiply") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Multiply", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Multiply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_40__pyx_unpickle_Multiply(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_40__pyx_unpickle_Multiply(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Multiply", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Multiply.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Multiply.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Multiply.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Multiply.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Multiply.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Multiply__set_state(((struct __pyx_obj_7flitter_8language_4tree_Multiply *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Multiply.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Multiply(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Multiply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Multiply__set_state(struct __pyx_obj_7flitter_8language_4tree_Multiply *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Multiply__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Multiply__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Divide(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_43__pyx_unpickle_Divide(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_43__pyx_unpickle_Divide = {"__pyx_unpickle_Divide", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_43__pyx_unpickle_Divide, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_43__pyx_unpickle_Divide(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Divide (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Divide", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Divide", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Divide") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Divide", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Divide", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_42__pyx_unpickle_Divide(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_42__pyx_unpickle_Divide(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Divide", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Divide.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Divide.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Divide.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Divide), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Divide.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Divide.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Divide__set_state(((struct __pyx_obj_7flitter_8language_4tree_Divide *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Divide.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Divide(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Divide", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Divide__set_state(struct __pyx_obj_7flitter_8language_4tree_Divide *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Divide__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Divide__set_state(<Divide> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Divide__set_state(Divide __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Divide__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FloorDivide(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_45__pyx_unpickle_FloorDivide(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_45__pyx_unpickle_FloorDivide = {"__pyx_unpickle_FloorDivide", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_45__pyx_unpickle_FloorDivide, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_45__pyx_unpickle_FloorDivide(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloorDivide (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloorDivide", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloorDivide", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FloorDivide") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloorDivide", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_FloorDivide", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_44__pyx_unpickle_FloorDivide(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_44__pyx_unpickle_FloorDivide(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloorDivide", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = FloorDivide.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FloorDivide.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = FloorDivide.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_FloorDivide), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = FloorDivide.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FloorDivide.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_FloorDivide__set_state(((struct __pyx_obj_7flitter_8language_4tree_FloorDivide *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = FloorDivide.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FloorDivide(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_FloorDivide", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_FloorDivide__set_state(struct __pyx_obj_7flitter_8language_4tree_FloorDivide *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloorDivide__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_FloorDivide__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Modulo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_47__pyx_unpickle_Modulo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_47__pyx_unpickle_Modulo = {"__pyx_unpickle_Modulo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_47__pyx_unpickle_Modulo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_47__pyx_unpickle_Modulo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Modulo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Modulo", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Modulo", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Modulo") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Modulo", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Modulo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_46__pyx_unpickle_Modulo(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_46__pyx_unpickle_Modulo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Modulo", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Modulo.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Modulo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Modulo.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Modulo), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Modulo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Modulo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Modulo__set_state(((struct __pyx_obj_7flitter_8language_4tree_Modulo *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Modulo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Modulo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Modulo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Modulo__set_state(struct __pyx_obj_7flitter_8language_4tree_Modulo *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Modulo__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Modulo__set_state(<Modulo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Modulo__set_state(Modulo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Modulo__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Power(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_49__pyx_unpickle_Power(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_49__pyx_unpickle_Power = {"__pyx_unpickle_Power", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_49__pyx_unpickle_Power, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_49__pyx_unpickle_Power(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Power (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Power", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Power", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Power") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Power", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Power", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_48__pyx_unpickle_Power(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_48__pyx_unpickle_Power(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Power", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Power.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Power.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Power.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Power), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Power.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Power.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Power__set_state(((struct __pyx_obj_7flitter_8language_4tree_Power *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Power.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Power(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Power", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Power__set_state(struct __pyx_obj_7flitter_8language_4tree_Power *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Power__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Power__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Comparison(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_51__pyx_unpickle_Comparison(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_51__pyx_unpickle_Comparison = {"__pyx_unpickle_Comparison", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_51__pyx_unpickle_Comparison, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_51__pyx_unpickle_Comparison(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Comparison (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Comparison", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Comparison", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Comparison") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Comparison", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Comparison", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_50__pyx_unpickle_Comparison(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_50__pyx_unpickle_Comparison(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Comparison", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Comparison.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Comparison.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Comparison.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Comparison.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Comparison.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Comparison__set_state(((struct __pyx_obj_7flitter_8language_4tree_Comparison *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Comparison.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Comparison(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Comparison", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Comparison__set_state(struct __pyx_obj_7flitter_8language_4tree_Comparison *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Comparison__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Comparison__set_state(<Comparison> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Comparison__set_state(Comparison __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Comparison__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_EqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_53__pyx_unpickle_EqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_53__pyx_unpickle_EqualTo = {"__pyx_unpickle_EqualTo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_53__pyx_unpickle_EqualTo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_53__pyx_unpickle_EqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_EqualTo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EqualTo", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EqualTo", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_EqualTo") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EqualTo", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_EqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_52__pyx_unpickle_EqualTo(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_52__pyx_unpickle_EqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EqualTo", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = EqualTo.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = EqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = EqualTo.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_EqualTo), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = EqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = EqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_EqualTo__set_state(((struct __pyx_obj_7flitter_8language_4tree_EqualTo *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = EqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_EqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_EqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_EqualTo *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EqualTo__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_EqualTo__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NotEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_55__pyx_unpickle_NotEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_55__pyx_unpickle_NotEqualTo = {"__pyx_unpickle_NotEqualTo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_55__pyx_unpickle_NotEqualTo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_55__pyx_unpickle_NotEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NotEqualTo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NotEqualTo", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NotEqualTo", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_NotEqualTo") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NotEqualTo", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NotEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_54__pyx_unpickle_NotEqualTo(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_54__pyx_unpickle_NotEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NotEqualTo", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_NotEqualTo), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_NotEqualTo__set_state(((struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = NotEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NotEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NotEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_NotEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NotEqualTo__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NotEqualTo__set_state(<NotEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NotEqualTo__set_state(NotEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NotEqualTo__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LessThan(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_57__pyx_unpickle_LessThan(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_57__pyx_unpickle_LessThan = {"__pyx_unpickle_LessThan", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_57__pyx_unpickle_LessThan, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_57__pyx_unpickle_LessThan(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThan", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThan", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LessThan") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThan", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_56__pyx_unpickle_LessThan(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_56__pyx_unpickle_LessThan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThan", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThan.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LessThan.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThan.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThan), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThan.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LessThan.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThan__set_state(((struct __pyx_obj_7flitter_8language_4tree_LessThan *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThan.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LessThan(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThan__set_state(struct __pyx_obj_7flitter_8language_4tree_LessThan *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThan__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThan__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GreaterThan(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_59__pyx_unpickle_GreaterThan(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_59__pyx_unpickle_GreaterThan = {"__pyx_unpickle_GreaterThan", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_59__pyx_unpickle_GreaterThan, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_59__pyx_unpickle_GreaterThan(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThan", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThan", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GreaterThan") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThan", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_58__pyx_unpickle_GreaterThan(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_58__pyx_unpickle_GreaterThan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThan", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThan.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GreaterThan.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThan.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThan), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThan.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GreaterThan.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThan__set_state(((struct __pyx_obj_7flitter_8language_4tree_GreaterThan *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThan.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GreaterThan(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThan__set_state(struct __pyx_obj_7flitter_8language_4tree_GreaterThan *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThan__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GreaterThan__set_state(<GreaterThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThan__set_state(GreaterThan __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThan__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LessThanOrEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_61__pyx_unpickle_LessThanOrEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_61__pyx_unpickle_LessThanOrEqualTo = {"__pyx_unpickle_LessThanOrEqualTo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_61__pyx_unpickle_LessThanOrEqualTo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_61__pyx_unpickle_LessThanOrEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThanOrEqualTo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThanOrEqualTo", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThanOrEqualTo", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LessThanOrEqualTo") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LessThanOrEqualTo", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThanOrEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_60__pyx_unpickle_LessThanOrEqualTo(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_60__pyx_unpickle_LessThanOrEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThanOrEqualTo", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThanOrEqualTo__set_state(((struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = LessThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LessThanOrEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThanOrEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_LessThanOrEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LessThanOrEqualTo__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_LessThanOrEqualTo__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GreaterThanOrEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_63__pyx_unpickle_GreaterThanOrEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_63__pyx_unpickle_GreaterThanOrEqualTo = {"__pyx_unpickle_GreaterThanOrEqualTo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_63__pyx_unpickle_GreaterThanOrEqualTo, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_63__pyx_unpickle_GreaterThanOrEqualTo(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThanOrEqualTo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThanOrEqualTo", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThanOrEqualTo", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_GreaterThanOrEqualTo") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GreaterThanOrEqualTo", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThanOrEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_62__pyx_unpickle_GreaterThanOrEqualTo(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_62__pyx_unpickle_GreaterThanOrEqualTo(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThanOrEqualTo", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThanOrEqualTo__set_state(((struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = GreaterThanOrEqualTo.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GreaterThanOrEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThanOrEqualTo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_GreaterThanOrEqualTo__set_state(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GreaterThanOrEqualTo__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GreaterThanOrEqualTo__set_state(<GreaterThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GreaterThanOrEqualTo__set_state(GreaterThanOrEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_GreaterThanOrEqualTo__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_And(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_65__pyx_unpickle_And(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_65__pyx_unpickle_And = {"__pyx_unpickle_And", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_65__pyx_unpickle_And, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_65__pyx_unpickle_And(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_And (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_And", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_And", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_And") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_And", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_And", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_64__pyx_unpickle_And(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_64__pyx_unpickle_And(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_And", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = And.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = And.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = And.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_And), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = And.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = And.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_And__set_state(((struct __pyx_obj_7flitter_8language_4tree_And *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = And.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_And(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_And", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_And__set_state(struct __pyx_obj_7flitter_8language_4tree_And *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_And__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_And__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Or(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_67__pyx_unpickle_Or(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_67__pyx_unpickle_Or = {"__pyx_unpickle_Or", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_67__pyx_unpickle_Or, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_67__pyx_unpickle_Or(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Or (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Or", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Or", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Or") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Or", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Or", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_66__pyx_unpickle_Or(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_66__pyx_unpickle_Or(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Or", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Or.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Or.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Or.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Or), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Or.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Or.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Or__set_state(((struct __pyx_obj_7flitter_8language_4tree_Or *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Or.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Or(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Or", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Or__set_state(struct __pyx_obj_7flitter_8language_4tree_Or *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Or__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Or__set_state(<Or> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Or__set_state(Or __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Or__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Xor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_69__pyx_unpickle_Xor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_69__pyx_unpickle_Xor = {"__pyx_unpickle_Xor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_69__pyx_unpickle_Xor, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_69__pyx_unpickle_Xor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Xor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Xor", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Xor", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Xor") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Xor", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Xor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_68__pyx_unpickle_Xor(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_68__pyx_unpickle_Xor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Xor", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__18, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Xor.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Xor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x132bc34, 0xf5e441f, 0x7e72cc7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Xor.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Xor), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Xor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Xor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Xor__set_state(((struct __pyx_obj_7flitter_8language_4tree_Xor *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x132bc34, 0xf5e441f, 0x7e72cc7) = (left, right))" % __pyx_checksum
 *     __pyx_result = Xor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Xor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Xor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Xor__set_state(struct __pyx_obj_7flitter_8language_4tree_Xor *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Xor__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.left);
  __pyx_v___pyx_result->__pyx_base.left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.right);
  __pyx_v___pyx_result->__pyx_base.right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Xor__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Slice(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_71__pyx_unpickle_Slice(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_71__pyx_unpickle_Slice = {"__pyx_unpickle_Slice", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_71__pyx_unpickle_Slice, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_71__pyx_unpickle_Slice(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Slice (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Slice", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Slice", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Slice") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Slice", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_70__pyx_unpickle_Slice(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_70__pyx_unpickle_Slice(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Slice", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3b1aecc, 0x62617e5, 0xd2f3b2d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__19, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3b1aecc, 0x62617e5, 0xd2f3b2d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 *     __pyx_result = Slice.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3b1aecc, 0x62617e5, 0xd2f3b2d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Slice.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3b1aecc, 0x62617e5, 0xd2f3b2d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 *     __pyx_result = Slice.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Slice), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 *     __pyx_result = Slice.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Slice.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Slice__set_state(((struct __pyx_obj_7flitter_8language_4tree_Slice *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3b1aecc, 0x62617e5, 0xd2f3b2d) = (expr, index))" % __pyx_checksum
 *     __pyx_result = Slice.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Slice(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Slice__set_state(struct __pyx_obj_7flitter_8language_4tree_Slice *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Slice__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->expr);
  __pyx_v___pyx_result->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->index);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->index);
  __pyx_v___pyx_result->index = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Slice__set_state(<Slice> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Slice__set_state(Slice __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.index = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Slice__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Call(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_73__pyx_unpickle_Call(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_73__pyx_unpickle_Call = {"__pyx_unpickle_Call", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_73__pyx_unpickle_Call, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_73__pyx_unpickle_Call(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Call (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Call", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Call", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Call") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Call", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_72__pyx_unpickle_Call(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_72__pyx_unpickle_Call(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Call", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x789d054, 0xe642687, 0xa0b3bb4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__20, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x789d054, 0xe642687, 0xa0b3bb4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 *     __pyx_result = Call.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x789d054, 0xe642687, 0xa0b3bb4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Call.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x789d054, 0xe642687, 0xa0b3bb4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 *     __pyx_result = Call.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Call), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 *     __pyx_result = Call.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Call.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Call__set_state(((struct __pyx_obj_7flitter_8language_4tree_Call *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x789d054, 0xe642687, 0xa0b3bb4) = (args, function, keyword_args))" % __pyx_checksum
 *     __pyx_result = Call.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Call(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Call__set_state(struct __pyx_obj_7flitter_8language_4tree_Call *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Call__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->function);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->function);
  __pyx_v___pyx_result->function = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->keyword_args);
  __Pyx_DECREF(__pyx_v___pyx_result->keyword_args);
  __pyx_v___pyx_result->keyword_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Call__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NodeModifier(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_75__pyx_unpickle_NodeModifier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_75__pyx_unpickle_NodeModifier = {"__pyx_unpickle_NodeModifier", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_75__pyx_unpickle_NodeModifier, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_75__pyx_unpickle_NodeModifier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NodeModifier (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NodeModifier", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NodeModifier", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_NodeModifier") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NodeModifier", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NodeModifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_74__pyx_unpickle_NodeModifier(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_74__pyx_unpickle_NodeModifier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NodeModifier", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x545ea53, 0xf8e966d, 0x36c4536):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__21, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x545ea53, 0xf8e966d, 0x36c4536):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 *     __pyx_result = NodeModifier.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x545ea53, 0xf8e966d, 0x36c4536):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = NodeModifier.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x545ea53, 0xf8e966d, 0x36c4536):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 *     __pyx_result = NodeModifier.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 *     __pyx_result = NodeModifier.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = NodeModifier.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_NodeModifier__set_state(((struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x545ea53, 0xf8e966d, 0x36c4536) = (node))" % __pyx_checksum
 *     __pyx_result = NodeModifier.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NodeModifier(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NodeModifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.node = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_NodeModifier__set_state(struct __pyx_obj_7flitter_8language_4tree_NodeModifier *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NodeModifier__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->node);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->node);
  __pyx_v___pyx_result->node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.node = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NodeModifier__set_state(<NodeModifier> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NodeModifier__set_state(NodeModifier __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.node = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_NodeModifier__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Tag(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_77__pyx_unpickle_Tag(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_77__pyx_unpickle_Tag = {"__pyx_unpickle_Tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_77__pyx_unpickle_Tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_77__pyx_unpickle_Tag(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Tag", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Tag", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Tag") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Tag", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_76__pyx_unpickle_Tag(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_76__pyx_unpickle_Tag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Tag", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9721bb0, 0xc9ce01d, 0xfc06094):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__22, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9721bb0, 0xc9ce01d, 0xfc06094):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 *     __pyx_result = Tag.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9721bb0, 0xc9ce01d, 0xfc06094):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Tag.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9721bb0, 0xc9ce01d, 0xfc06094):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 *     __pyx_result = Tag.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Tag), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 *     __pyx_result = Tag.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Tag.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Tag__set_state(((struct __pyx_obj_7flitter_8language_4tree_Tag *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9721bb0, 0xc9ce01d, 0xfc06094) = (node, tag))" % __pyx_checksum
 *     __pyx_result = Tag.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Tag(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Tag__set_state(struct __pyx_obj_7flitter_8language_4tree_Tag *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Tag__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __pyx_v___pyx_result->__pyx_base.node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->tag);
  __Pyx_DECREF(__pyx_v___pyx_result->tag);
  __pyx_v___pyx_result->tag = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Tag__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Attributes(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_79__pyx_unpickle_Attributes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_79__pyx_unpickle_Attributes = {"__pyx_unpickle_Attributes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_79__pyx_unpickle_Attributes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_79__pyx_unpickle_Attributes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Attributes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Attributes", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Attributes", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Attributes") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Attributes", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_78__pyx_unpickle_Attributes(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_78__pyx_unpickle_Attributes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Attributes", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x552d50b, 0x28ebf23, 0xf187e9b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__23, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x552d50b, 0x28ebf23, 0xf187e9b):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 *     __pyx_result = Attributes.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x552d50b, 0x28ebf23, 0xf187e9b):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Attributes.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x552d50b, 0x28ebf23, 0xf187e9b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 *     __pyx_result = Attributes.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Attributes), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 *     __pyx_result = Attributes.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Attributes.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Attributes__set_state(((struct __pyx_obj_7flitter_8language_4tree_Attributes *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x552d50b, 0x28ebf23, 0xf187e9b) = (bindings, node))" % __pyx_checksum
 *     __pyx_result = Attributes.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Attributes(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Attributes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Attributes__set_state(struct __pyx_obj_7flitter_8language_4tree_Attributes *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Attributes__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bindings);
  __Pyx_DECREF(__pyx_v___pyx_result->bindings);
  __pyx_v___pyx_result->bindings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __pyx_v___pyx_result->__pyx_base.node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Attributes__set_state(<Attributes> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Attributes__set_state(Attributes __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Attributes__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Append(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_81__pyx_unpickle_Append(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_81__pyx_unpickle_Append = {"__pyx_unpickle_Append", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_81__pyx_unpickle_Append, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_81__pyx_unpickle_Append(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Append (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Append", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Append", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Append") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Append", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_80__pyx_unpickle_Append(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_80__pyx_unpickle_Append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Append", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcfaea2d, 0x5507a99, 0x6d8790e):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__24, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcfaea2d, 0x5507a99, 0x6d8790e):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 *     __pyx_result = Append.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xcfaea2d, 0x5507a99, 0x6d8790e):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Append.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcfaea2d, 0x5507a99, 0x6d8790e):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 *     __pyx_result = Append.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 *     __pyx_result = Append.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Append.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Append__set_state(((struct __pyx_obj_7flitter_8language_4tree_Append *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcfaea2d, 0x5507a99, 0x6d8790e) = (children, node))" % __pyx_checksum
 *     __pyx_result = Append.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Append(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Append__set_state(struct __pyx_obj_7flitter_8language_4tree_Append *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Append__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->children);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->children);
  __pyx_v___pyx_result->children = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.node);
  __pyx_v___pyx_result->__pyx_base.node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Append__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Binding(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_83__pyx_unpickle_Binding(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_83__pyx_unpickle_Binding = {"__pyx_unpickle_Binding", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_83__pyx_unpickle_Binding, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_83__pyx_unpickle_Binding(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Binding (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Binding", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Binding", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Binding") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Binding", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Binding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_82__pyx_unpickle_Binding(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_82__pyx_unpickle_Binding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Binding", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a2bfe3, 0x122c79d, 0xf5557ba):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__25, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a2bfe3, 0x122c79d, 0xf5557ba):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 *     __pyx_result = Binding.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3a2bfe3, 0x122c79d, 0xf5557ba):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Binding.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a2bfe3, 0x122c79d, 0xf5557ba):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 *     __pyx_result = Binding.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 *     __pyx_result = Binding.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Binding.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Binding__set_state(((struct __pyx_obj_7flitter_8language_4tree_Binding *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a2bfe3, 0x122c79d, 0xf5557ba) = (expr, name))" % __pyx_checksum
 *     __pyx_result = Binding.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Binding(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Binding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Binding__set_state(struct __pyx_obj_7flitter_8language_4tree_Binding *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Binding__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->expr);
  __pyx_v___pyx_result->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Binding__set_state(<Binding> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Binding__set_state(Binding __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.name = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Binding__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PolyBinding(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_85__pyx_unpickle_PolyBinding(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_85__pyx_unpickle_PolyBinding = {"__pyx_unpickle_PolyBinding", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_85__pyx_unpickle_PolyBinding, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_85__pyx_unpickle_PolyBinding(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PolyBinding (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PolyBinding", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PolyBinding", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PolyBinding") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PolyBinding", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_PolyBinding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_84__pyx_unpickle_PolyBinding(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_84__pyx_unpickle_PolyBinding(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PolyBinding", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x12bf31f, 0xbff3b4b, 0x3964daa):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__26, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x12bf31f, 0xbff3b4b, 0x3964daa):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 *     __pyx_result = PolyBinding.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x12bf31f, 0xbff3b4b, 0x3964daa):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PolyBinding.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x12bf31f, 0xbff3b4b, 0x3964daa):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 *     __pyx_result = PolyBinding.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 *     __pyx_result = PolyBinding.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PolyBinding.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_PolyBinding__set_state(((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x12bf31f, 0xbff3b4b, 0x3964daa) = (expr, names))" % __pyx_checksum
 *     __pyx_result = PolyBinding.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PolyBinding(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_PolyBinding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_PolyBinding__set_state(struct __pyx_obj_7flitter_8language_4tree_PolyBinding *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PolyBinding__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->expr);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->expr);
  __pyx_v___pyx_result->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->names);
  __Pyx_DECREF(__pyx_v___pyx_result->names);
  __pyx_v___pyx_result->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_PolyBinding__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Let(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_87__pyx_unpickle_Let(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_87__pyx_unpickle_Let = {"__pyx_unpickle_Let", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_87__pyx_unpickle_Let, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_87__pyx_unpickle_Let(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Let (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Let", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Let", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Let") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Let", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Let", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_86__pyx_unpickle_Let(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_86__pyx_unpickle_Let(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Let", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe222284, 0xc774b95, 0x343984a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__27, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe222284, 0xc774b95, 0x343984a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 *     __pyx_result = Let.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe222284, 0xc774b95, 0x343984a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Let.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe222284, 0xc774b95, 0x343984a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 *     __pyx_result = Let.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 *     __pyx_result = Let.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Let.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Let__set_state(((struct __pyx_obj_7flitter_8language_4tree_Let *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe222284, 0xc774b95, 0x343984a) = (bindings, body))" % __pyx_checksum
 *     __pyx_result = Let.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Let(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Let", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Let__set_state(struct __pyx_obj_7flitter_8language_4tree_Let *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Let__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bindings);
  __Pyx_DECREF(__pyx_v___pyx_result->bindings);
  __pyx_v___pyx_result->bindings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->body);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->body);
  __pyx_v___pyx_result->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Let__set_state(<Let> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Let__set_state(Let __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.bindings = __pyx_state[0]; __pyx_result.body = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Let__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_For(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_89__pyx_unpickle_For(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_89__pyx_unpickle_For = {"__pyx_unpickle_For", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_89__pyx_unpickle_For, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_89__pyx_unpickle_For(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_For (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_For", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_For", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_For") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_For", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_For", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_88__pyx_unpickle_For(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_88__pyx_unpickle_For(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_For", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3d9b243, 0x3815181, 0xd46f68c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__28, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3d9b243, 0x3815181, 0xd46f68c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 *     __pyx_result = For.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3d9b243, 0x3815181, 0xd46f68c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = For.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3d9b243, 0x3815181, 0xd46f68c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 *     __pyx_result = For.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_For), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 *     __pyx_result = For.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = For.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_For__set_state(((struct __pyx_obj_7flitter_8language_4tree_For *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3d9b243, 0x3815181, 0xd46f68c) = (body, names, source))" % __pyx_checksum
 *     __pyx_result = For.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_For(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_For", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_For__set_state(struct __pyx_obj_7flitter_8language_4tree_For *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_For__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->body);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->body);
  __pyx_v___pyx_result->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->names);
  __Pyx_DECREF(__pyx_v___pyx_result->names);
  __pyx_v___pyx_result->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->source);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->source);
  __pyx_v___pyx_result->source = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_For__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IfCondition(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_91__pyx_unpickle_IfCondition(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_91__pyx_unpickle_IfCondition = {"__pyx_unpickle_IfCondition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_91__pyx_unpickle_IfCondition, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_91__pyx_unpickle_IfCondition(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfCondition (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfCondition", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfCondition", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IfCondition") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfCondition", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfCondition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_90__pyx_unpickle_IfCondition(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_90__pyx_unpickle_IfCondition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfCondition", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9a45a72, 0xf0437c3, 0xc7e5dff):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__29, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9a45a72, 0xf0437c3, 0xc7e5dff):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 *     __pyx_result = IfCondition.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9a45a72, 0xf0437c3, 0xc7e5dff):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IfCondition.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9a45a72, 0xf0437c3, 0xc7e5dff):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 *     __pyx_result = IfCondition.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfCondition), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 *     __pyx_result = IfCondition.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IfCondition.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_IfCondition__set_state(((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9a45a72, 0xf0437c3, 0xc7e5dff) = (condition, then))" % __pyx_checksum
 *     __pyx_result = IfCondition.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IfCondition(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfCondition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_IfCondition__set_state(struct __pyx_obj_7flitter_8language_4tree_IfCondition *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfCondition__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->condition);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->condition);
  __pyx_v___pyx_result->condition = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->then);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->then);
  __pyx_v___pyx_result->then = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IfCondition__set_state(<IfCondition> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IfCondition__set_state(IfCondition __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.condition = __pyx_state[0]; __pyx_result.then = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfCondition__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IfElse(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_93__pyx_unpickle_IfElse(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_93__pyx_unpickle_IfElse = {"__pyx_unpickle_IfElse", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_93__pyx_unpickle_IfElse, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_93__pyx_unpickle_IfElse(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfElse (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfElse", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfElse", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IfElse") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IfElse", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfElse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_92__pyx_unpickle_IfElse(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_92__pyx_unpickle_IfElse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfElse", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xdcd679c, 0x11aad5b, 0x05298d3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__30, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xdcd679c, 0x11aad5b, 0x05298d3):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 *     __pyx_result = IfElse.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xdcd679c, 0x11aad5b, 0x05298d3):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IfElse.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xdcd679c, 0x11aad5b, 0x05298d3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 *     __pyx_result = IfElse.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfElse), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 *     __pyx_result = IfElse.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IfElse.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_IfElse__set_state(((struct __pyx_obj_7flitter_8language_4tree_IfElse *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xdcd679c, 0x11aad5b, 0x05298d3) = (else_, tests))" % __pyx_checksum
 *     __pyx_result = IfElse.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IfElse(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfElse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_IfElse__set_state(struct __pyx_obj_7flitter_8language_4tree_IfElse *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IfElse__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->else_);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->else_);
  __pyx_v___pyx_result->else_ = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->tests);
  __Pyx_DECREF(__pyx_v___pyx_result->tests);
  __pyx_v___pyx_result->tests = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_IfElse__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Function(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_8language_4tree_95__pyx_unpickle_Function(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_8language_4tree_95__pyx_unpickle_Function = {"__pyx_unpickle_Function", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_95__pyx_unpickle_Function, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_8language_4tree_95__pyx_unpickle_Function(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Function") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Function", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_8language_4tree_94__pyx_unpickle_Function(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_8language_4tree_94__pyx_unpickle_Function(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf4b5d07, 0x8043f6c, 0xc9f0ad2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__31, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf4b5d07, 0x8043f6c, 0xc9f0ad2):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(2, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xf4b5d07, 0x8043f6c, 0xc9f0ad2):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf4b5d07, 0x8043f6c, 0xc9f0ad2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Function), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_8language_4tree___pyx_unpickle_Function__set_state(((struct __pyx_obj_7flitter_8language_4tree_Function *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf4b5d07, 0x8043f6c, 0xc9f0ad2) = (body, captures, inlineable, name, parameters, recursive))" % __pyx_checksum
 *     __pyx_result = Function.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Function(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_8language_4tree___pyx_unpickle_Function__set_state(struct __pyx_obj_7flitter_8language_4tree_Function *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Function__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_8language_4tree_Expression))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->body);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->body);
  __pyx_v___pyx_result->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->captures);
  __Pyx_DECREF(__pyx_v___pyx_result->captures);
  __pyx_v___pyx_result->captures = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->inlineable = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->parameters);
  __Pyx_DECREF(__pyx_v___pyx_result->parameters);
  __pyx_v___pyx_result->parameters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->recursive = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Function__set_state(<Function> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Function__set_state(Function __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.captures = __pyx_state[1]; __pyx_result.inlineable = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.parameters = __pyx_state[4]; __pyx_result.recursive = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.language.tree.__pyx_unpickle_Function__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_7flitter_8language_4tree_Expression __pyx_vtable_7flitter_8language_4tree_Expression;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Expression(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Expression *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_8language_4tree_Expression;
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Expression(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Expression) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Expression[] = {
  {"compile", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_1compile, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"simplify", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_3simplify, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Expression_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Expression_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Expression},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Expression},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Expression},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Expression_spec = {
  "flitter.language.tree.Expression",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Expression),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_4tree_Expression_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Expression = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Expression", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Expression), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Expression, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Expression, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Top __pyx_vtable_7flitter_8language_4tree_Top;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Top(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Top *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Top *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Top;
  p->pragmas = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->dependencies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Top(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Top *p = (struct __pyx_obj_7flitter_8language_4tree_Top *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Top) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->pragmas);
  Py_CLEAR(p->body);
  Py_CLEAR(p->dependencies);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Top(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Top *p = (struct __pyx_obj_7flitter_8language_4tree_Top *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Top)); if (e) return e;
  if (p->pragmas) {
    e = (*v)(p->pragmas, a); if (e) return e;
  }
  if (p->body) {
    e = (*v)(((PyObject *)p->body), a); if (e) return e;
  }
  if (p->dependencies) {
    e = (*v)(p->dependencies, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Top(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Top *p = (struct __pyx_obj_7flitter_8language_4tree_Top *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Top);
  tmp = ((PyObject*)p->pragmas);
  p->pragmas = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->body);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dependencies);
  p->dependencies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Top_pragmas(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Top_7pragmas_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Top_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Top_4body_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Top_dependencies(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Top_12dependencies_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Top_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_3Top_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Top[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Top_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Top_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Top_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Top[] = {
  {(char *)"pragmas", __pyx_getprop_7flitter_8language_4tree_3Top_pragmas, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_7flitter_8language_4tree_3Top_body, 0, (char *)0, 0},
  {(char *)"dependencies", __pyx_getprop_7flitter_8language_4tree_3Top_dependencies, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Top_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Top},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_3Top_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Top},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Top},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Top},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Top},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_3Top_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Top},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Top_spec = {
  "flitter.language.tree.Top",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Top),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Top_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Top = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Top", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Top), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Top, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Top_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Top, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Top, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Top, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Top, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Top_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Top, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Export __pyx_vtable_7flitter_8language_4tree_Export;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Export(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Export *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Export *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Export;
  p->static_exports = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Export(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Export *p = (struct __pyx_obj_7flitter_8language_4tree_Export *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Export) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->static_exports);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Export(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Export *p = (struct __pyx_obj_7flitter_8language_4tree_Export *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Export)); if (e) return e;
  if (p->static_exports) {
    e = (*v)(p->static_exports, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Export(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Export *p = (struct __pyx_obj_7flitter_8language_4tree_Export *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Export);
  tmp = ((PyObject*)p->static_exports);
  p->static_exports = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Export_static_exports(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Export_14static_exports_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Export_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_6Export_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Export[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Export_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Export_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Export_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Export[] = {
  {(char *)"static_exports", __pyx_getprop_7flitter_8language_4tree_6Export_static_exports, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Export_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Export},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_6Export_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Export},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Export},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Export},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Export},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_6Export_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Export},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Export_spec = {
  "flitter.language.tree.Export",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Export),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Export_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Export = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Export", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Export), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Export, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Export_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Export, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Export, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Export, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Export, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Export_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Export, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Import __pyx_vtable_7flitter_8language_4tree_Import;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Import(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Import *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Import *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Import;
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->filename = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Import(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Import *p = (struct __pyx_obj_7flitter_8language_4tree_Import *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Import) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->names);
  Py_CLEAR(p->filename);
  Py_CLEAR(p->expr);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Import(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Import *p = (struct __pyx_obj_7flitter_8language_4tree_Import *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Import)); if (e) return e;
  if (p->names) {
    e = (*v)(p->names, a); if (e) return e;
  }
  if (p->filename) {
    e = (*v)(((PyObject *)p->filename), a); if (e) return e;
  }
  if (p->expr) {
    e = (*v)(((PyObject *)p->expr), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Import(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Import *p = (struct __pyx_obj_7flitter_8language_4tree_Import *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Import);
  tmp = ((PyObject*)p->names);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->filename);
  p->filename = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->expr);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Import_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Import_5names_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Import_filename(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Import_8filename_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Import_expr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Import_4expr_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Import_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_6Import_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Import[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Import_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Import_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Import_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Import[] = {
  {(char *)"names", __pyx_getprop_7flitter_8language_4tree_6Import_names, 0, (char *)0, 0},
  {(char *)"filename", __pyx_getprop_7flitter_8language_4tree_6Import_filename, 0, (char *)0, 0},
  {(char *)"expr", __pyx_getprop_7flitter_8language_4tree_6Import_expr, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Import_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Import},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_6Import_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Import},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Import},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Import},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Import},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_6Import_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Import},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Import_spec = {
  "flitter.language.tree.Import",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Import),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Import_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Import = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Import", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Import), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Import, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Import_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Import, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Import, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Import, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Import, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Import_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Import, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Sequence __pyx_vtable_7flitter_8language_4tree_Sequence;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Sequence(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Sequence *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Sequence *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Sequence;
  p->expressions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Sequence(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Sequence *p = (struct __pyx_obj_7flitter_8language_4tree_Sequence *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Sequence) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->expressions);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Sequence(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Sequence *p = (struct __pyx_obj_7flitter_8language_4tree_Sequence *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Sequence)); if (e) return e;
  if (p->expressions) {
    e = (*v)(p->expressions, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Sequence(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Sequence *p = (struct __pyx_obj_7flitter_8language_4tree_Sequence *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Sequence);
  tmp = ((PyObject*)p->expressions);
  p->expressions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Sequence_expressions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Sequence_11expressions_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Sequence[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Sequence_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Sequence_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Sequence[] = {
  {(char *)"expressions", __pyx_getprop_7flitter_8language_4tree_8Sequence_expressions, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Sequence_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Sequence},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Sequence},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Sequence},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Sequence},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Sequence},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_8Sequence_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Sequence},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Sequence_spec = {
  "flitter.language.tree.Sequence",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Sequence),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Sequence_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Sequence = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Sequence", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Sequence), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Sequence, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_8Sequence_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Sequence, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Sequence, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Sequence, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Sequence, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_8Sequence_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Sequence, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Literal __pyx_vtable_7flitter_8language_4tree_Literal;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Literal(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Literal *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Literal *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Literal;
  p->value = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Literal(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Literal *p = (struct __pyx_obj_7flitter_8language_4tree_Literal *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Literal) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->value);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Literal(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Literal *p = (struct __pyx_obj_7flitter_8language_4tree_Literal *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Literal)); if (e) return e;
  if (p->value) {
    e = (*v)(((PyObject *)p->value), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Literal(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Literal *p = (struct __pyx_obj_7flitter_8language_4tree_Literal *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Literal);
  tmp = ((PyObject*)p->value);
  p->value = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_7Literal_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_7Literal_5value_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_7Literal_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_7Literal_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Literal[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_7Literal_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Literal_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Literal_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Literal[] = {
  {(char *)"value", __pyx_getprop_7flitter_8language_4tree_7Literal_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Literal_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Literal},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_7Literal_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Literal},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Literal},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Literal},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Literal},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_7Literal_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Literal},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Literal_spec = {
  "flitter.language.tree.Literal",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Literal),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Literal_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Literal = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Literal", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Literal), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Literal, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_7Literal_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Literal, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Literal, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Literal, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Literal, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_7Literal_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Literal, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Name __pyx_vtable_7flitter_8language_4tree_Name;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Name(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Name *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Name *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Name;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Name(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Name *p = (struct __pyx_obj_7flitter_8language_4tree_Name *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Name) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->name);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_4Name_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_4Name_4name_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_4Name_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_4Name_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Name[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_4Name_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Name_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Name_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Name[] = {
  {(char *)"name", __pyx_getprop_7flitter_8language_4tree_4Name_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Name_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Name},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_4Name_3__repr__},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Name},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Name},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_4Name_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Name},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Name_spec = {
  "flitter.language.tree.Name",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Name),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_8language_4tree_Name_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Name = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Name", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Name), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Name, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_4Name_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Name, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Name, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_4Name_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Name, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Lookup __pyx_vtable_7flitter_8language_4tree_Lookup;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Lookup(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Lookup *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Lookup *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Lookup;
  p->key = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Lookup(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Lookup *p = (struct __pyx_obj_7flitter_8language_4tree_Lookup *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Lookup) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Lookup(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Lookup *p = (struct __pyx_obj_7flitter_8language_4tree_Lookup *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Lookup)); if (e) return e;
  if (p->key) {
    e = (*v)(((PyObject *)p->key), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Lookup(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Lookup *p = (struct __pyx_obj_7flitter_8language_4tree_Lookup *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Lookup);
  tmp = ((PyObject*)p->key);
  p->key = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Lookup_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Lookup_3key_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Lookup[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Lookup_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Lookup_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Lookup[] = {
  {(char *)"key", __pyx_getprop_7flitter_8language_4tree_6Lookup_key, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Lookup_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Lookup},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Lookup},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Lookup},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Lookup},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Lookup},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_6Lookup_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Lookup},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Lookup_spec = {
  "flitter.language.tree.Lookup",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Lookup),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Lookup_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Lookup = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Lookup", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Lookup), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Lookup, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Lookup_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Lookup, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Lookup, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Lookup, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Lookup, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Lookup_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Lookup, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Range __pyx_vtable_7flitter_8language_4tree_Range;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Range(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Range *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Range *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Range;
  p->start = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->stop = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->step = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Range(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Range *p = (struct __pyx_obj_7flitter_8language_4tree_Range *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Range) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->start);
  Py_CLEAR(p->stop);
  Py_CLEAR(p->step);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Range(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Range *p = (struct __pyx_obj_7flitter_8language_4tree_Range *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Range)); if (e) return e;
  if (p->start) {
    e = (*v)(((PyObject *)p->start), a); if (e) return e;
  }
  if (p->stop) {
    e = (*v)(((PyObject *)p->stop), a); if (e) return e;
  }
  if (p->step) {
    e = (*v)(((PyObject *)p->step), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Range(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Range *p = (struct __pyx_obj_7flitter_8language_4tree_Range *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Range);
  tmp = ((PyObject*)p->start);
  p->start = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->stop);
  p->stop = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->step);
  p->step = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_5Range_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_5Range_5start_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_5Range_stop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_5Range_4stop_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_5Range_step(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_5Range_4step_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_5Range_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_5Range_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Range[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_5Range_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Range_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Range_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Range[] = {
  {(char *)"start", __pyx_getprop_7flitter_8language_4tree_5Range_start, 0, (char *)0, 0},
  {(char *)"stop", __pyx_getprop_7flitter_8language_4tree_5Range_stop, 0, (char *)0, 0},
  {(char *)"step", __pyx_getprop_7flitter_8language_4tree_5Range_step, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Range_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Range},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_5Range_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Range},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Range},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Range},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Range},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_5Range_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Range},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Range_spec = {
  "flitter.language.tree.Range",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Range),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Range_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Range = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Range", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Range), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Range, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_5Range_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Range, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Range, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Range, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Range, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_5Range_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Range, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_UnaryOperation __pyx_vtable_7flitter_8language_4tree_UnaryOperation;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_UnaryOperation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->expr);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation)); if (e) return e;
  if (p->expr) {
    e = (*v)(((PyObject *)p->expr), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation);
  tmp = ((PyObject*)p->expr);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_14UnaryOperation_expr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_14UnaryOperation_4expr_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_UnaryOperation[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_UnaryOperation[] = {
  {(char *)"expr", __pyx_getprop_7flitter_8language_4tree_14UnaryOperation_expr, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_UnaryOperation_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_UnaryOperation},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_UnaryOperation_spec = {
  "flitter.language.tree.UnaryOperation",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_UnaryOperation_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_UnaryOperation = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""UnaryOperation", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_UnaryOperation, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_UnaryOperation, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_UnaryOperation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Negative __pyx_vtable_7flitter_8language_4tree_Negative;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Negative(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Negative *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Negative *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Negative;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Negative[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Negative_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Negative_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Negative_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Negative},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Negative},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Negative_spec = {
  "flitter.language.tree.Negative",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Negative),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Negative_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Negative = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Negative", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Negative), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Negative, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Negative, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Positive __pyx_vtable_7flitter_8language_4tree_Positive;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Positive(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Positive *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Positive *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Positive;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Positive[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Positive_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Positive_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Positive_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Positive},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Positive},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Positive_spec = {
  "flitter.language.tree.Positive",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Positive),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Positive_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Positive = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Positive", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Positive), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Positive, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Positive, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Not __pyx_vtable_7flitter_8language_4tree_Not;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Not(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Not *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Not *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Not;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Not[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Not_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Not_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Not_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Not},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Not},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Not_spec = {
  "flitter.language.tree.Not",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Not),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Not_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Not = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Not", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Not), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Not, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Not, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Ceil __pyx_vtable_7flitter_8language_4tree_Ceil;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Ceil(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Ceil *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Ceil *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Ceil;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Ceil[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Ceil_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Ceil_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Ceil_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Ceil},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Ceil},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Ceil_spec = {
  "flitter.language.tree.Ceil",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Ceil),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Ceil_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Ceil = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Ceil", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Ceil), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Ceil, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Ceil, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Floor __pyx_vtable_7flitter_8language_4tree_Floor;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Floor(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Floor *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Floor *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Floor;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Floor[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Floor_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Floor_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Floor_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Floor},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Floor},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Floor_spec = {
  "flitter.language.tree.Floor",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Floor),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Floor_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Floor = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Floor", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Floor), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Floor, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Floor, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Fract __pyx_vtable_7flitter_8language_4tree_Fract;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Fract(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Fract *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_UnaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Fract *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Fract;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Fract[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Fract_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Fract_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Fract_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_UnaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Fract},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Fract},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Fract_spec = {
  "flitter.language.tree.Fract",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Fract),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Fract_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Fract = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Fract", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Fract), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_UnaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_UnaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_UnaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Fract, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_14UnaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Fract, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_BinaryOperation __pyx_vtable_7flitter_8language_4tree_BinaryOperation;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_BinaryOperation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  p->left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left);
  Py_CLEAR(p->right);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation)); if (e) return e;
  if (p->left) {
    e = (*v)(((PyObject *)p->left), a); if (e) return e;
  }
  if (p->right) {
    e = (*v)(((PyObject *)p->right), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *p = (struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation);
  tmp = ((PyObject*)p->left);
  p->left = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right);
  p->right = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_15BinaryOperation_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_15BinaryOperation_4left_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_15BinaryOperation_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_15BinaryOperation_5right_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_BinaryOperation[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_BinaryOperation[] = {
  {(char *)"left", __pyx_getprop_7flitter_8language_4tree_15BinaryOperation_left, 0, (char *)0, 0},
  {(char *)"right", __pyx_getprop_7flitter_8language_4tree_15BinaryOperation_right, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_BinaryOperation_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_BinaryOperation},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_BinaryOperation_spec = {
  "flitter.language.tree.BinaryOperation",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_BinaryOperation_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_BinaryOperation = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""BinaryOperation", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_BinaryOperation, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_BinaryOperation, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_BinaryOperation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_MathsBinaryOperation __pyx_vtable_7flitter_8language_4tree_MathsBinaryOperation;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_BinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_MathsBinaryOperation[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_MathsBinaryOperation_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_MathsBinaryOperation},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_MathsBinaryOperation_spec = {
  "flitter.language.tree.MathsBinaryOperation",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_MathsBinaryOperation_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_MathsBinaryOperation = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""MathsBinaryOperation", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_MathsBinaryOperation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_MathsBinaryOperation, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Add __pyx_vtable_7flitter_8language_4tree_Add;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Add(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Add *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Add *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Add;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Add[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Add_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Add_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Add_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Add},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Add},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Add_spec = {
  "flitter.language.tree.Add",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Add),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Add_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Add = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Add", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Add), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Add, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Add, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Subtract __pyx_vtable_7flitter_8language_4tree_Subtract;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Subtract(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Subtract *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Subtract *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Subtract;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Subtract[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Subtract_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Subtract_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Subtract_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Subtract},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Subtract},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Subtract_spec = {
  "flitter.language.tree.Subtract",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Subtract),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Subtract_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Subtract = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Subtract", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Subtract), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Subtract, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Subtract, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Multiply __pyx_vtable_7flitter_8language_4tree_Multiply;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Multiply(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Multiply *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Multiply *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Multiply;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Multiply[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Multiply_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Multiply_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Multiply_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Multiply},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Multiply},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Multiply_spec = {
  "flitter.language.tree.Multiply",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Multiply),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Multiply_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Multiply = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Multiply", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Multiply), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Multiply, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Multiply, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Divide __pyx_vtable_7flitter_8language_4tree_Divide;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Divide(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Divide *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Divide *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Divide;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Divide[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Divide_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Divide_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Divide_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Divide},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Divide},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Divide_spec = {
  "flitter.language.tree.Divide",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Divide),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Divide_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Divide = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Divide", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Divide), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Divide, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Divide, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_FloorDivide __pyx_vtable_7flitter_8language_4tree_FloorDivide;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_FloorDivide(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_FloorDivide *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_FloorDivide *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_FloorDivide;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_FloorDivide[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_FloorDivide_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_FloorDivide},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_FloorDivide},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_FloorDivide_spec = {
  "flitter.language.tree.FloorDivide",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_FloorDivide),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_FloorDivide_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_FloorDivide = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""FloorDivide", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_FloorDivide), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_FloorDivide, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_FloorDivide, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Modulo __pyx_vtable_7flitter_8language_4tree_Modulo;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Modulo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Modulo *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Modulo *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Modulo;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Modulo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Modulo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Modulo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Modulo_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Modulo},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Modulo},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Modulo_spec = {
  "flitter.language.tree.Modulo",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Modulo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Modulo_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Modulo = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Modulo", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Modulo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Modulo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Modulo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Power __pyx_vtable_7flitter_8language_4tree_Power;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Power(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Power *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_MathsBinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Power *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Power;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Power[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Power_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Power_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Power_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Power},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Power},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Power_spec = {
  "flitter.language.tree.Power",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Power),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Power_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Power = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Power", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Power), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Power, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Power, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Comparison __pyx_vtable_7flitter_8language_4tree_Comparison;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Comparison(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Comparison *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_BinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Comparison *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Comparison[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Comparison_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Comparison_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Comparison_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Comparison},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Comparison},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Comparison_spec = {
  "flitter.language.tree.Comparison",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Comparison),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Comparison_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Comparison = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Comparison", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Comparison), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Comparison, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Comparison, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_EqualTo __pyx_vtable_7flitter_8language_4tree_EqualTo;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_EqualTo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_EqualTo *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_EqualTo *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_EqualTo;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_EqualTo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7EqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7EqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_EqualTo_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_EqualTo},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_EqualTo},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_EqualTo_spec = {
  "flitter.language.tree.EqualTo",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_EqualTo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_EqualTo_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_EqualTo = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""EqualTo", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_EqualTo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_EqualTo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_EqualTo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_NotEqualTo __pyx_vtable_7flitter_8language_4tree_NotEqualTo;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_NotEqualTo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_NotEqualTo *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_NotEqualTo;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_NotEqualTo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_NotEqualTo_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_NotEqualTo},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_NotEqualTo},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_NotEqualTo_spec = {
  "flitter.language.tree.NotEqualTo",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_NotEqualTo_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_NotEqualTo = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""NotEqualTo", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_NotEqualTo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_NotEqualTo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_NotEqualTo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_LessThan __pyx_vtable_7flitter_8language_4tree_LessThan;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_LessThan(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_LessThan *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_LessThan *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_LessThan;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_LessThan[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8LessThan_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8LessThan_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_LessThan_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_LessThan},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_LessThan},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_LessThan_spec = {
  "flitter.language.tree.LessThan",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_LessThan),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_LessThan_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_LessThan = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""LessThan", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_LessThan), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_LessThan, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_LessThan, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThan __pyx_vtable_7flitter_8language_4tree_GreaterThan;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_GreaterThan(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_GreaterThan *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_GreaterThan *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_GreaterThan;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_GreaterThan[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_GreaterThan_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_GreaterThan},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_GreaterThan},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_GreaterThan_spec = {
  "flitter.language.tree.GreaterThan",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_GreaterThan),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_GreaterThan_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_GreaterThan = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""GreaterThan", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_GreaterThan), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_GreaterThan, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_GreaterThan, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_LessThanOrEqualTo __pyx_vtable_7flitter_8language_4tree_LessThanOrEqualTo;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_LessThanOrEqualTo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_LessThanOrEqualTo;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_LessThanOrEqualTo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_LessThanOrEqualTo_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_LessThanOrEqualTo},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_LessThanOrEqualTo},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_LessThanOrEqualTo_spec = {
  "flitter.language.tree.LessThanOrEqualTo",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_LessThanOrEqualTo_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_LessThanOrEqualTo = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""LessThanOrEqualTo", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_LessThanOrEqualTo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_LessThanOrEqualTo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_LessThanOrEqualTo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_GreaterThanOrEqualTo __pyx_vtable_7flitter_8language_4tree_GreaterThanOrEqualTo;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_GreaterThanOrEqualTo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Comparison(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_GreaterThanOrEqualTo;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_GreaterThanOrEqualTo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_GreaterThanOrEqualTo},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_GreaterThanOrEqualTo},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo_spec = {
  "flitter.language.tree.GreaterThanOrEqualTo",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""GreaterThanOrEqualTo", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_GreaterThanOrEqualTo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_GreaterThanOrEqualTo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_GreaterThanOrEqualTo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_And __pyx_vtable_7flitter_8language_4tree_And;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_And(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_And *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_BinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_And *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_And;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_And[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3And_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3And_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_And_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_And},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_And},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_And_spec = {
  "flitter.language.tree.And",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_And),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_And_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_And = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""And", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_And), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_And, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_And, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Or __pyx_vtable_7flitter_8language_4tree_Or;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Or(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Or *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_BinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Or *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Or;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Or[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_2Or_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_2Or_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Or_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Or},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Or},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Or_spec = {
  "flitter.language.tree.Or",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Or),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Or_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Or = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Or", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Or), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Or, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Or, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Xor __pyx_vtable_7flitter_8language_4tree_Xor;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Xor(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Xor *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_BinaryOperation(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Xor *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Xor;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Xor[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Xor_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Xor_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Xor_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_BinaryOperation},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Xor},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Xor},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Xor_spec = {
  "flitter.language.tree.Xor",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Xor),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Xor_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Xor = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Xor", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Xor), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_BinaryOperation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_BinaryOperation, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_BinaryOperation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Xor, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7flitter_8language_4tree_15BinaryOperation_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Xor, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Slice __pyx_vtable_7flitter_8language_4tree_Slice;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Slice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Slice *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Slice *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Slice;
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->index = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Slice(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Slice *p = (struct __pyx_obj_7flitter_8language_4tree_Slice *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Slice) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->expr);
  Py_CLEAR(p->index);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Slice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Slice *p = (struct __pyx_obj_7flitter_8language_4tree_Slice *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Slice)); if (e) return e;
  if (p->expr) {
    e = (*v)(((PyObject *)p->expr), a); if (e) return e;
  }
  if (p->index) {
    e = (*v)(((PyObject *)p->index), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Slice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Slice *p = (struct __pyx_obj_7flitter_8language_4tree_Slice *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Slice);
  tmp = ((PyObject*)p->expr);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->index);
  p->index = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_5Slice_expr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_5Slice_4expr_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_5Slice_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_5Slice_5index_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_5Slice_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_5Slice_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Slice[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_5Slice_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Slice_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_5Slice_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Slice[] = {
  {(char *)"expr", __pyx_getprop_7flitter_8language_4tree_5Slice_expr, 0, (char *)0, 0},
  {(char *)"index", __pyx_getprop_7flitter_8language_4tree_5Slice_index, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Slice_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Slice},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_5Slice_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Slice},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Slice},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Slice},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Slice},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_5Slice_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Slice},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Slice_spec = {
  "flitter.language.tree.Slice",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Slice),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Slice_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Slice = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Slice", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Slice), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Slice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_5Slice_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Slice, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Slice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Slice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Slice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_5Slice_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Slice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Call __pyx_vtable_7flitter_8language_4tree_Call;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Call(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Call *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Call *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Call;
  p->function = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->keyword_args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Call(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Call *p = (struct __pyx_obj_7flitter_8language_4tree_Call *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Call) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->function);
  Py_CLEAR(p->args);
  Py_CLEAR(p->keyword_args);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Call(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Call *p = (struct __pyx_obj_7flitter_8language_4tree_Call *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Call)); if (e) return e;
  if (p->function) {
    e = (*v)(((PyObject *)p->function), a); if (e) return e;
  }
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  if (p->keyword_args) {
    e = (*v)(p->keyword_args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Call(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Call *p = (struct __pyx_obj_7flitter_8language_4tree_Call *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Call);
  tmp = ((PyObject*)p->function);
  p->function = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->args);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->keyword_args);
  p->keyword_args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_4Call_function(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_4Call_8function_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_4Call_args(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_4Call_4args_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_4Call_keyword_args(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_4Call_12keyword_args_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_4Call_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_4Call_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Call[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_4Call_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Call_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_4Call_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Call[] = {
  {(char *)"function", __pyx_getprop_7flitter_8language_4tree_4Call_function, 0, (char *)0, 0},
  {(char *)"args", __pyx_getprop_7flitter_8language_4tree_4Call_args, 0, (char *)0, 0},
  {(char *)"keyword_args", __pyx_getprop_7flitter_8language_4tree_4Call_keyword_args, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Call_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Call},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_4Call_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Call},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Call},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Call},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Call},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_4Call_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Call},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Call_spec = {
  "flitter.language.tree.Call",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Call),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Call_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Call = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Call", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Call), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Call, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_4Call_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Call, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Call, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Call, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Call, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_4Call_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Call, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_NodeModifier __pyx_vtable_7flitter_8language_4tree_NodeModifier;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_NodeModifier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_NodeModifier;
  p->node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier *p = (struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->node);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier *p = (struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier)); if (e) return e;
  if (p->node) {
    e = (*v)(((PyObject *)p->node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_NodeModifier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_NodeModifier *p = (struct __pyx_obj_7flitter_8language_4tree_NodeModifier *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_NodeModifier);
  tmp = ((PyObject*)p->node);
  p->node = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_12NodeModifier_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_12NodeModifier_4node_1__get__(o);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_NodeModifier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_NodeModifier[] = {
  {(char *)"node", __pyx_getprop_7flitter_8language_4tree_12NodeModifier_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_NodeModifier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_NodeModifier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_NodeModifier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_NodeModifier},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_NodeModifier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_NodeModifier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_NodeModifier_spec = {
  "flitter.language.tree.NodeModifier",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_NodeModifier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_NodeModifier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_NodeModifier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""NodeModifier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_NodeModifier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_NodeModifier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_NodeModifier, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_NodeModifier, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_NodeModifier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Tag __pyx_vtable_7flitter_8language_4tree_Tag;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Tag(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Tag *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_NodeModifier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Tag *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Tag;
  p->tag = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Tag(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Tag *p = (struct __pyx_obj_7flitter_8language_4tree_Tag *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Tag) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->tag);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Tag_tag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Tag_3tag_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Tag_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_3Tag_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Tag[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Tag_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Tag_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Tag_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Tag[] = {
  {(char *)"tag", __pyx_getprop_7flitter_8language_4tree_3Tag_tag, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Tag_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Tag},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_3Tag_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_NodeModifier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_NodeModifier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Tag},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Tag},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_3Tag_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Tag},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Tag_spec = {
  "flitter.language.tree.Tag",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Tag),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Tag_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Tag = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Tag", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Tag), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Tag, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Tag_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_NodeModifier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Tag, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Tag, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Tag_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Tag, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Attributes __pyx_vtable_7flitter_8language_4tree_Attributes;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Attributes(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Attributes *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_NodeModifier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Attributes *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Attributes;
  p->bindings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Attributes(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Attributes *p = (struct __pyx_obj_7flitter_8language_4tree_Attributes *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Attributes) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->bindings);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Attributes(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Attributes *p = (struct __pyx_obj_7flitter_8language_4tree_Attributes *)o;
  e = __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier(o, v, a); if (e) return e;
  if (p->bindings) {
    e = (*v)(p->bindings, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Attributes(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Attributes *p = (struct __pyx_obj_7flitter_8language_4tree_Attributes *)o;
  __pyx_tp_clear_7flitter_8language_4tree_NodeModifier(o);
  tmp = ((PyObject*)p->bindings);
  p->bindings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_10Attributes_bindings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_10Attributes_8bindings_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Attributes[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Attributes_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_10Attributes_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Attributes[] = {
  {(char *)"bindings", __pyx_getprop_7flitter_8language_4tree_10Attributes_bindings, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Attributes_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Attributes},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Attributes},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Attributes},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Attributes},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Attributes},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_10Attributes_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Attributes},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Attributes_spec = {
  "flitter.language.tree.Attributes",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Attributes),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Attributes_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Attributes = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Attributes", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Attributes), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Attributes, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_10Attributes_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Attributes, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Attributes, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Attributes, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Attributes, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_10Attributes_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Attributes, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Append __pyx_vtable_7flitter_8language_4tree_Append;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Append(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Append *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_NodeModifier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Append *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Append;
  p->children = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Append(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Append *p = (struct __pyx_obj_7flitter_8language_4tree_Append *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Append) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->children);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7flitter_8language_4tree_NodeModifier(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Append(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Append *p = (struct __pyx_obj_7flitter_8language_4tree_Append *)o;
  e = __pyx_tp_traverse_7flitter_8language_4tree_NodeModifier(o, v, a); if (e) return e;
  if (p->children) {
    e = (*v)(((PyObject *)p->children), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Append(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Append *p = (struct __pyx_obj_7flitter_8language_4tree_Append *)o;
  __pyx_tp_clear_7flitter_8language_4tree_NodeModifier(o);
  tmp = ((PyObject*)p->children);
  p->children = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6Append_children(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6Append_8children_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Append_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_6Append_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Append[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6Append_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Append_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6Append_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Append[] = {
  {(char *)"children", __pyx_getprop_7flitter_8language_4tree_6Append_children, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Append_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Append},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_6Append_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Append},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Append},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Append},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Append},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_6Append_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Append},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Append_spec = {
  "flitter.language.tree.Append",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Append),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Append_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Append = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Append", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Append), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Append, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Append_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Append, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Append, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Append, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Append, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6Append_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Append, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Binding(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Binding *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_4tree_Binding *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Binding(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Binding *p = (struct __pyx_obj_7flitter_8language_4tree_Binding *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Binding) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->expr);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Binding(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Binding *p = (struct __pyx_obj_7flitter_8language_4tree_Binding *)o;
  if (p->expr) {
    e = (*v)(((PyObject *)p->expr), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Binding(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Binding *p = (struct __pyx_obj_7flitter_8language_4tree_Binding *)o;
  tmp = ((PyObject*)p->expr);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_7Binding_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_7Binding_4name_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_7Binding_expr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_7Binding_4expr_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_7Binding_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_7Binding_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Binding[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_7Binding_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Binding_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_7Binding_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Binding[] = {
  {(char *)"name", __pyx_getprop_7flitter_8language_4tree_7Binding_name, 0, (char *)0, 0},
  {(char *)"expr", __pyx_getprop_7flitter_8language_4tree_7Binding_expr, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Binding_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Binding},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_7Binding_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Binding},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Binding},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Binding},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Binding},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_7Binding_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Binding},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Binding_spec = {
  "flitter.language.tree.Binding",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Binding),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Binding_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Binding = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Binding", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Binding), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Binding, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_7Binding_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Binding, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Binding, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Binding, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Binding, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_7Binding_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Binding, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_4tree_PolyBinding(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)o);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_PolyBinding(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *p = (struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_PolyBinding) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->names);
  Py_CLEAR(p->expr);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_4tree_PolyBinding(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *p = (struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)o;
  if (p->names) {
    e = (*v)(p->names, a); if (e) return e;
  }
  if (p->expr) {
    e = (*v)(((PyObject *)p->expr), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_PolyBinding(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_PolyBinding *p = (struct __pyx_obj_7flitter_8language_4tree_PolyBinding *)o;
  tmp = ((PyObject*)p->names);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->expr);
  p->expr = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_11PolyBinding_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_11PolyBinding_5names_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_11PolyBinding_expr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_11PolyBinding_4expr_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_PolyBinding[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_PolyBinding[] = {
  {(char *)"names", __pyx_getprop_7flitter_8language_4tree_11PolyBinding_names, 0, (char *)0, 0},
  {(char *)"expr", __pyx_getprop_7flitter_8language_4tree_11PolyBinding_expr, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_PolyBinding_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_PolyBinding},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_PolyBinding},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_PolyBinding},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_PolyBinding},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_PolyBinding},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_11PolyBinding_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_PolyBinding},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_PolyBinding_spec = {
  "flitter.language.tree.PolyBinding",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_PolyBinding),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_PolyBinding_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_PolyBinding = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""PolyBinding", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_PolyBinding), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_PolyBinding, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_11PolyBinding_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_PolyBinding, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_PolyBinding, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_PolyBinding, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_PolyBinding, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_11PolyBinding_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_PolyBinding, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Let __pyx_vtable_7flitter_8language_4tree_Let;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Let(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Let *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Let *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Let;
  p->bindings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Let(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Let *p = (struct __pyx_obj_7flitter_8language_4tree_Let *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Let) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->bindings);
  Py_CLEAR(p->body);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Let(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Let *p = (struct __pyx_obj_7flitter_8language_4tree_Let *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Let)); if (e) return e;
  if (p->bindings) {
    e = (*v)(p->bindings, a); if (e) return e;
  }
  if (p->body) {
    e = (*v)(((PyObject *)p->body), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Let(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Let *p = (struct __pyx_obj_7flitter_8language_4tree_Let *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Let);
  tmp = ((PyObject*)p->bindings);
  p->bindings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->body);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Let_bindings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Let_8bindings_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3Let_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3Let_4body_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Let_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_3Let_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Let[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3Let_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Let_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3Let_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Let[] = {
  {(char *)"bindings", __pyx_getprop_7flitter_8language_4tree_3Let_bindings, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_7flitter_8language_4tree_3Let_body, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Let_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Let},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_3Let_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Let},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Let},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Let},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Let},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_3Let_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Let},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Let_spec = {
  "flitter.language.tree.Let",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Let),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Let_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Let = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Let", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Let), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Let, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Let_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Let, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Let, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Let, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Let, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3Let_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Let, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_For __pyx_vtable_7flitter_8language_4tree_For;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_For(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_For *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_For *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_For;
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->source = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_For(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_For *p = (struct __pyx_obj_7flitter_8language_4tree_For *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_For) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->names);
  Py_CLEAR(p->source);
  Py_CLEAR(p->body);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_For(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_For *p = (struct __pyx_obj_7flitter_8language_4tree_For *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_For)); if (e) return e;
  if (p->names) {
    e = (*v)(p->names, a); if (e) return e;
  }
  if (p->source) {
    e = (*v)(((PyObject *)p->source), a); if (e) return e;
  }
  if (p->body) {
    e = (*v)(((PyObject *)p->body), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_For(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_For *p = (struct __pyx_obj_7flitter_8language_4tree_For *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_For);
  tmp = ((PyObject*)p->names);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->source);
  p->source = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->body);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3For_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3For_5names_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3For_source(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3For_6source_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_3For_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_3For_4body_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3For_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_3For_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_For[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_3For_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3For_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_3For_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_For[] = {
  {(char *)"names", __pyx_getprop_7flitter_8language_4tree_3For_names, 0, (char *)0, 0},
  {(char *)"source", __pyx_getprop_7flitter_8language_4tree_3For_source, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_7flitter_8language_4tree_3For_body, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_For_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_For},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_3For_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_For},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_For},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_For},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_For},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_3For_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_For},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_For_spec = {
  "flitter.language.tree.For",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_For),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_For_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_For = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""For", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_For), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_For, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3For_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_For, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_For, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_For, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_For, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_3For_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_For, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7flitter_8language_4tree_IfCondition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_8language_4tree_IfCondition *)o);
  p->condition = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->then = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_IfCondition(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *p = (struct __pyx_obj_7flitter_8language_4tree_IfCondition *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_IfCondition) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->condition);
  Py_CLEAR(p->then);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_8language_4tree_IfCondition(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *p = (struct __pyx_obj_7flitter_8language_4tree_IfCondition *)o;
  if (p->condition) {
    e = (*v)(((PyObject *)p->condition), a); if (e) return e;
  }
  if (p->then) {
    e = (*v)(((PyObject *)p->then), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_IfCondition(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_IfCondition *p = (struct __pyx_obj_7flitter_8language_4tree_IfCondition *)o;
  tmp = ((PyObject*)p->condition);
  p->condition = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->then);
  p->then = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_11IfCondition_condition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_11IfCondition_9condition_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_11IfCondition_then(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_11IfCondition_4then_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_IfCondition[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11IfCondition_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_11IfCondition_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_IfCondition[] = {
  {(char *)"condition", __pyx_getprop_7flitter_8language_4tree_11IfCondition_condition, 0, (char *)0, 0},
  {(char *)"then", __pyx_getprop_7flitter_8language_4tree_11IfCondition_then, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_IfCondition_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_IfCondition},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_IfCondition},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_IfCondition},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_IfCondition},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_IfCondition},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_11IfCondition_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_IfCondition},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_IfCondition_spec = {
  "flitter.language.tree.IfCondition",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_IfCondition),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_IfCondition_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_IfCondition = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""IfCondition", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_IfCondition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_IfCondition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_11IfCondition_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_IfCondition, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_IfCondition, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_IfCondition, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_IfCondition, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_11IfCondition_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_IfCondition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_IfElse __pyx_vtable_7flitter_8language_4tree_IfElse;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_IfElse(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_IfElse *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_IfElse *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_IfElse;
  p->tests = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->else_ = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_IfElse(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_IfElse *p = (struct __pyx_obj_7flitter_8language_4tree_IfElse *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_IfElse) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->tests);
  Py_CLEAR(p->else_);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_IfElse(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_IfElse *p = (struct __pyx_obj_7flitter_8language_4tree_IfElse *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_IfElse)); if (e) return e;
  if (p->tests) {
    e = (*v)(p->tests, a); if (e) return e;
  }
  if (p->else_) {
    e = (*v)(((PyObject *)p->else_), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_IfElse(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_IfElse *p = (struct __pyx_obj_7flitter_8language_4tree_IfElse *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_IfElse);
  tmp = ((PyObject*)p->tests);
  p->tests = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->else_);
  p->else_ = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6IfElse_tests(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6IfElse_5tests_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_6IfElse_else_(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_6IfElse_5else__1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_IfElse[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6IfElse_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_6IfElse_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_IfElse[] = {
  {(char *)"tests", __pyx_getprop_7flitter_8language_4tree_6IfElse_tests, 0, (char *)0, 0},
  {(char *)"else_", __pyx_getprop_7flitter_8language_4tree_6IfElse_else_, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_IfElse_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_IfElse},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_IfElse},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_IfElse},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_IfElse},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_IfElse},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_6IfElse_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_IfElse},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_IfElse_spec = {
  "flitter.language.tree.IfElse",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_IfElse),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_IfElse_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_IfElse = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""IfElse", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_IfElse), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_IfElse, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6IfElse_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_IfElse, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_IfElse, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_IfElse, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_IfElse, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_6IfElse_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_IfElse, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_8language_4tree_Function __pyx_vtable_7flitter_8language_4tree_Function;

static PyObject *__pyx_tp_new_7flitter_8language_4tree_Function(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_8language_4tree_Function *p;
  PyObject *o = __pyx_tp_new_7flitter_8language_4tree_Expression(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_8language_4tree_Function *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_8language_4tree_Expression*)__pyx_vtabptr_7flitter_8language_4tree_Function;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  p->captures = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree_Function(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree_Function *p = (struct __pyx_obj_7flitter_8language_4tree_Function *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree_Function) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->parameters);
  Py_CLEAR(p->body);
  Py_CLEAR(p->captures);
  __pyx_tp_dealloc_7flitter_8language_4tree_Expression(o);
}

static int __pyx_tp_traverse_7flitter_8language_4tree_Function(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree_Function *p = (struct __pyx_obj_7flitter_8language_4tree_Function *)o;
  e = ((likely(__pyx_ptype_7flitter_8language_4tree_Expression)) ? ((__pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse) ? __pyx_ptype_7flitter_8language_4tree_Expression->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_8language_4tree_Function)); if (e) return e;
  if (p->parameters) {
    e = (*v)(p->parameters, a); if (e) return e;
  }
  if (p->body) {
    e = (*v)(((PyObject *)p->body), a); if (e) return e;
  }
  if (p->captures) {
    e = (*v)(p->captures, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_8language_4tree_Function(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_8language_4tree_Function *p = (struct __pyx_obj_7flitter_8language_4tree_Function *)o;
  if (likely(__pyx_ptype_7flitter_8language_4tree_Expression)) { if (__pyx_ptype_7flitter_8language_4tree_Expression->tp_clear) __pyx_ptype_7flitter_8language_4tree_Expression->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_8language_4tree_Function);
  tmp = ((PyObject*)p->parameters);
  p->parameters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->body);
  p->body = ((struct __pyx_obj_7flitter_8language_4tree_Expression *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->captures);
  p->captures = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_4name_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_parameters(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_10parameters_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_4body_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_captures(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_8captures_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_inlineable(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_10inlineable_1__get__(o);
}

static PyObject *__pyx_getprop_7flitter_8language_4tree_8Function_recursive(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7flitter_8language_4tree_8Function_9recursive_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_8Function_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7flitter_8language_4tree_8Function_3__repr__(self);
}

static PyMethodDef __pyx_methods_7flitter_8language_4tree_Function[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7flitter_8language_4tree_8Function_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Function_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_8language_4tree_8Function_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7flitter_8language_4tree_Function[] = {
  {(char *)"name", __pyx_getprop_7flitter_8language_4tree_8Function_name, 0, (char *)0, 0},
  {(char *)"parameters", __pyx_getprop_7flitter_8language_4tree_8Function_parameters, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_7flitter_8language_4tree_8Function_body, 0, (char *)0, 0},
  {(char *)"captures", __pyx_getprop_7flitter_8language_4tree_8Function_captures, 0, (char *)0, 0},
  {(char *)"inlineable", __pyx_getprop_7flitter_8language_4tree_8Function_inlineable, 0, (char *)0, 0},
  {(char *)"recursive", __pyx_getprop_7flitter_8language_4tree_8Function_recursive, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree_Function_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree_Function},
  {Py_tp_repr, (void *)__pyx_pw_7flitter_8language_4tree_8Function_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree_Function},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_8language_4tree_Function},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_8language_4tree_Function},
  {Py_tp_getset, (void *)__pyx_getsets_7flitter_8language_4tree_Function},
  {Py_tp_init, (void *)__pyx_pw_7flitter_8language_4tree_8Function_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree_Function},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree_Function_spec = {
  "flitter.language.tree.Function",
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Function),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_8language_4tree_Function_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree_Function = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""Function", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree_Function), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree_Function, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7flitter_8language_4tree_8Function_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree_Function, /*tp_traverse*/
  __pyx_tp_clear_7flitter_8language_4tree_Function, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_8language_4tree_Function, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7flitter_8language_4tree_Function, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_8language_4tree_8Function_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree_Function, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *__pyx_freelist_7flitter_8language_4tree___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_7flitter_8language_4tree___pyx_scope_struct__genexpr = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_8language_4tree___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_8language_4tree___pyx_scope_struct__genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_7flitter_8language_4tree___pyx_scope_struct__genexpr[--__pyx_freecount_7flitter_8language_4tree___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_8language_4tree___pyx_scope_struct__genexpr(PyObject *o) {
  struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *p = (struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_8language_4tree___pyx_scope_struct__genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_name);
  Py_CLEAR(p->__pyx_v_value);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_8language_4tree___pyx_scope_struct__genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_7flitter_8language_4tree___pyx_scope_struct__genexpr[__pyx_freecount_7flitter_8language_4tree___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_8language_4tree___pyx_scope_struct__genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *p = (struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_name) {
    e = (*v)(p->__pyx_v_name, a); if (e) return e;
  }
  if (p->__pyx_v_value) {
    e = (*v)(p->__pyx_v_value, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_8language_4tree___pyx_scope_struct__genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_8language_4tree___pyx_scope_struct__genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_8language_4tree___pyx_scope_struct__genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr_spec = {
  "flitter.language.tree.__pyx_scope_struct__genexpr",
  sizeof(struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.language.tree.""__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_8language_4tree___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_8language_4tree___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_8language_4tree___pyx_scope_struct__genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_8language_4tree___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_n_s_Add, __pyx_k_Add, sizeof(__pyx_k_Add), 0, 0, 1, 1},
    {&__pyx_n_s_Add___reduce_cython, __pyx_k_Add___reduce_cython, sizeof(__pyx_k_Add___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Add___setstate_cython, __pyx_k_Add___setstate_cython, sizeof(__pyx_k_Add___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_And, __pyx_k_And, sizeof(__pyx_k_And), 0, 0, 1, 1},
    {&__pyx_n_s_And___reduce_cython, __pyx_k_And___reduce_cython, sizeof(__pyx_k_And___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_And___setstate_cython, __pyx_k_And___setstate_cython, sizeof(__pyx_k_And___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Append, __pyx_k_Append, sizeof(__pyx_k_Append), 0, 0, 1, 1},
    {&__pyx_n_s_Append___reduce_cython, __pyx_k_Append___reduce_cython, sizeof(__pyx_k_Append___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Append___setstate_cython, __pyx_k_Append___setstate_cython, sizeof(__pyx_k_Append___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_Attributes, __pyx_k_Attributes, sizeof(__pyx_k_Attributes), 0, 0, 1, 1},
    {&__pyx_n_s_Attributes___reduce_cython, __pyx_k_Attributes___reduce_cython, sizeof(__pyx_k_Attributes___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Attributes___setstate_cython, __pyx_k_Attributes___setstate_cython, sizeof(__pyx_k_Attributes___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BinaryOperation, __pyx_k_BinaryOperation, sizeof(__pyx_k_BinaryOperation), 0, 0, 1, 1},
    {&__pyx_n_s_BinaryOperation___reduce_cython, __pyx_k_BinaryOperation___reduce_cython, sizeof(__pyx_k_BinaryOperation___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BinaryOperation___setstate_cytho, __pyx_k_BinaryOperation___setstate_cytho, sizeof(__pyx_k_BinaryOperation___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_kp_u_Binding, __pyx_k_Binding, sizeof(__pyx_k_Binding), 0, 1, 0, 0},
    {&__pyx_n_s_Binding_2, __pyx_k_Binding_2, sizeof(__pyx_k_Binding_2), 0, 0, 1, 1},
    {&__pyx_n_s_Binding___reduce_cython, __pyx_k_Binding___reduce_cython, sizeof(__pyx_k_Binding___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Binding___setstate_cython, __pyx_k_Binding___setstate_cython, sizeof(__pyx_k_Binding___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Call, __pyx_k_Call, sizeof(__pyx_k_Call), 0, 1, 0, 0},
    {&__pyx_n_s_Call_2, __pyx_k_Call_2, sizeof(__pyx_k_Call_2), 0, 0, 1, 1},
    {&__pyx_n_s_Call___reduce_cython, __pyx_k_Call___reduce_cython, sizeof(__pyx_k_Call___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Call___setstate_cython, __pyx_k_Call___setstate_cython, sizeof(__pyx_k_Call___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Ceil, __pyx_k_Ceil, sizeof(__pyx_k_Ceil), 0, 0, 1, 1},
    {&__pyx_n_s_Ceil___reduce_cython, __pyx_k_Ceil___reduce_cython, sizeof(__pyx_k_Ceil___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Ceil___setstate_cython, __pyx_k_Ceil___setstate_cython, sizeof(__pyx_k_Ceil___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Circular_import_of, __pyx_k_Circular_import_of, sizeof(__pyx_k_Circular_import_of), 0, 1, 0, 0},
    {&__pyx_n_s_Comparison, __pyx_k_Comparison, sizeof(__pyx_k_Comparison), 0, 0, 1, 1},
    {&__pyx_n_s_Comparison___reduce_cython, __pyx_k_Comparison___reduce_cython, sizeof(__pyx_k_Comparison___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Comparison___setstate_cython, __pyx_k_Comparison___setstate_cython, sizeof(__pyx_k_Comparison___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Compiler_error, __pyx_k_Compiler_error, sizeof(__pyx_k_Compiler_error), 0, 1, 0, 0},
    {&__pyx_n_s_Divide, __pyx_k_Divide, sizeof(__pyx_k_Divide), 0, 0, 1, 1},
    {&__pyx_n_s_Divide___reduce_cython, __pyx_k_Divide___reduce_cython, sizeof(__pyx_k_Divide___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Divide___setstate_cython, __pyx_k_Divide___setstate_cython, sizeof(__pyx_k_Divide___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EqualTo, __pyx_k_EqualTo, sizeof(__pyx_k_EqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_EqualTo___reduce_cython, __pyx_k_EqualTo___reduce_cython, sizeof(__pyx_k_EqualTo___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EqualTo___setstate_cython, __pyx_k_EqualTo___setstate_cython, sizeof(__pyx_k_EqualTo___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Error_calling, __pyx_k_Error_calling, sizeof(__pyx_k_Error_calling), 0, 1, 0, 0},
    {&__pyx_kp_u_Export, __pyx_k_Export, sizeof(__pyx_k_Export), 0, 1, 0, 0},
    {&__pyx_n_s_Export_2, __pyx_k_Export_2, sizeof(__pyx_k_Export_2), 0, 0, 1, 1},
    {&__pyx_n_s_Export___reduce_cython, __pyx_k_Export___reduce_cython, sizeof(__pyx_k_Export___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Export___setstate_cython, __pyx_k_Export___setstate_cython, sizeof(__pyx_k_Export___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Expression, __pyx_k_Expression, sizeof(__pyx_k_Expression), 0, 0, 1, 1},
    {&__pyx_n_s_Expression___reduce_cython, __pyx_k_Expression___reduce_cython, sizeof(__pyx_k_Expression___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Expression___setstate_cython, __pyx_k_Expression___setstate_cython, sizeof(__pyx_k_Expression___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Expression_compile, __pyx_k_Expression_compile, sizeof(__pyx_k_Expression_compile), 0, 0, 1, 1},
    {&__pyx_n_s_Expression_simplify, __pyx_k_Expression_simplify, sizeof(__pyx_k_Expression_simplify), 0, 0, 1, 1},
    {&__pyx_n_u_False, __pyx_k_False, sizeof(__pyx_k_False), 0, 1, 0, 1},
    {&__pyx_n_s_Floor, __pyx_k_Floor, sizeof(__pyx_k_Floor), 0, 0, 1, 1},
    {&__pyx_n_s_FloorDivide, __pyx_k_FloorDivide, sizeof(__pyx_k_FloorDivide), 0, 0, 1, 1},
    {&__pyx_n_s_FloorDivide___reduce_cython, __pyx_k_FloorDivide___reduce_cython, sizeof(__pyx_k_FloorDivide___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FloorDivide___setstate_cython, __pyx_k_FloorDivide___setstate_cython, sizeof(__pyx_k_FloorDivide___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Floor___reduce_cython, __pyx_k_Floor___reduce_cython, sizeof(__pyx_k_Floor___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Floor___setstate_cython, __pyx_k_Floor___setstate_cython, sizeof(__pyx_k_Floor___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_For, __pyx_k_For, sizeof(__pyx_k_For), 0, 1, 0, 0},
    {&__pyx_n_s_For_2, __pyx_k_For_2, sizeof(__pyx_k_For_2), 0, 0, 1, 1},
    {&__pyx_n_s_For___reduce_cython, __pyx_k_For___reduce_cython, sizeof(__pyx_k_For___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_For___setstate_cython, __pyx_k_For___setstate_cython, sizeof(__pyx_k_For___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Fract, __pyx_k_Fract, sizeof(__pyx_k_Fract), 0, 0, 1, 1},
    {&__pyx_n_s_Fract___reduce_cython, __pyx_k_Fract___reduce_cython, sizeof(__pyx_k_Fract___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Fract___setstate_cython, __pyx_k_Fract___setstate_cython, sizeof(__pyx_k_Fract___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Function, __pyx_k_Function, sizeof(__pyx_k_Function), 0, 1, 0, 0},
    {&__pyx_n_s_Function_2, __pyx_k_Function_2, sizeof(__pyx_k_Function_2), 0, 0, 1, 1},
    {&__pyx_n_s_Function___reduce_cython, __pyx_k_Function___reduce_cython, sizeof(__pyx_k_Function___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Function___setstate_cython, __pyx_k_Function___setstate_cython, sizeof(__pyx_k_Function___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThan, __pyx_k_GreaterThan, sizeof(__pyx_k_GreaterThan), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThanOrEqualTo, __pyx_k_GreaterThanOrEqualTo, sizeof(__pyx_k_GreaterThanOrEqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThanOrEqualTo___reduce_cy, __pyx_k_GreaterThanOrEqualTo___reduce_cy, sizeof(__pyx_k_GreaterThanOrEqualTo___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThanOrEqualTo___setstate, __pyx_k_GreaterThanOrEqualTo___setstate, sizeof(__pyx_k_GreaterThanOrEqualTo___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThan___reduce_cython, __pyx_k_GreaterThan___reduce_cython, sizeof(__pyx_k_GreaterThan___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_GreaterThan___setstate_cython, __pyx_k_GreaterThan___setstate_cython, sizeof(__pyx_k_GreaterThan___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_IfCondition, __pyx_k_IfCondition, sizeof(__pyx_k_IfCondition), 0, 1, 0, 0},
    {&__pyx_n_s_IfCondition_2, __pyx_k_IfCondition_2, sizeof(__pyx_k_IfCondition_2), 0, 0, 1, 1},
    {&__pyx_n_s_IfCondition___reduce_cython, __pyx_k_IfCondition___reduce_cython, sizeof(__pyx_k_IfCondition___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IfCondition___setstate_cython, __pyx_k_IfCondition___setstate_cython, sizeof(__pyx_k_IfCondition___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_IfElse, __pyx_k_IfElse, sizeof(__pyx_k_IfElse), 0, 1, 0, 0},
    {&__pyx_n_s_IfElse_2, __pyx_k_IfElse_2, sizeof(__pyx_k_IfElse_2), 0, 0, 1, 1},
    {&__pyx_n_s_IfElse___reduce_cython, __pyx_k_IfElse___reduce_cython, sizeof(__pyx_k_IfElse___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IfElse___setstate_cython, __pyx_k_IfElse___setstate_cython, sizeof(__pyx_k_IfElse___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Import, __pyx_k_Import, sizeof(__pyx_k_Import), 0, 1, 0, 0},
    {&__pyx_n_s_Import_2, __pyx_k_Import_2, sizeof(__pyx_k_Import_2), 0, 0, 1, 1},
    {&__pyx_n_s_Import___reduce_cython, __pyx_k_Import___reduce_cython, sizeof(__pyx_k_Import___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Import___setstate_cython, __pyx_k_Import___setstate_cython, sizeof(__pyx_k_Import___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Import__simplify_locals_genexpr, __pyx_k_Import__simplify_locals_genexpr, sizeof(__pyx_k_Import__simplify_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_k_Incompatible_checksums_0x_x_vs_0_10, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_10), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_k_Incompatible_checksums_0x_x_vs_0_11, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_11), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_k_Incompatible_checksums_0x_x_vs_0_12, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_12), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_k_Incompatible_checksums_0x_x_vs_0_13, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_13), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_k_Incompatible_checksums_0x_x_vs_0_14, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_14), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_k_Incompatible_checksums_0x_x_vs_0_15, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_15), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_k_Incompatible_checksums_0x_x_vs_0_16, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_16), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17, __pyx_k_Incompatible_checksums_0x_x_vs_0_17, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_17), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18, __pyx_k_Incompatible_checksums_0x_x_vs_0_18, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_18), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19, __pyx_k_Incompatible_checksums_0x_x_vs_0_19, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_19), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20, __pyx_k_Incompatible_checksums_0x_x_vs_0_20, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_20), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21, __pyx_k_Incompatible_checksums_0x_x_vs_0_21, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_21), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22, __pyx_k_Incompatible_checksums_0x_x_vs_0_22, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_22), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_k_Incompatible_checksums_0x_x_vs_0_23, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_23), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24, __pyx_k_Incompatible_checksums_0x_x_vs_0_24, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_24), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_k_Incompatible_checksums_0x_x_vs_0_8, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_8), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_k_Incompatible_checksums_0x_x_vs_0_9, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_9), 0, 0, 1, 0},
    {&__pyx_n_s_LessThan, __pyx_k_LessThan, sizeof(__pyx_k_LessThan), 0, 0, 1, 1},
    {&__pyx_n_s_LessThanOrEqualTo, __pyx_k_LessThanOrEqualTo, sizeof(__pyx_k_LessThanOrEqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_LessThanOrEqualTo___reduce_cytho, __pyx_k_LessThanOrEqualTo___reduce_cytho, sizeof(__pyx_k_LessThanOrEqualTo___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_LessThanOrEqualTo___setstate_cyt, __pyx_k_LessThanOrEqualTo___setstate_cyt, sizeof(__pyx_k_LessThanOrEqualTo___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_LessThan___reduce_cython, __pyx_k_LessThan___reduce_cython, sizeof(__pyx_k_LessThan___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LessThan___setstate_cython, __pyx_k_LessThan___setstate_cython, sizeof(__pyx_k_LessThan___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Let, __pyx_k_Let, sizeof(__pyx_k_Let), 0, 1, 0, 0},
    {&__pyx_n_s_Let_2, __pyx_k_Let_2, sizeof(__pyx_k_Let_2), 0, 0, 1, 1},
    {&__pyx_n_s_Let___reduce_cython, __pyx_k_Let___reduce_cython, sizeof(__pyx_k_Let___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Let___setstate_cython, __pyx_k_Let___setstate_cython, sizeof(__pyx_k_Let___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Literal, __pyx_k_Literal, sizeof(__pyx_k_Literal), 0, 1, 0, 0},
    {&__pyx_n_s_Literal_2, __pyx_k_Literal_2, sizeof(__pyx_k_Literal_2), 0, 0, 1, 1},
    {&__pyx_n_s_Literal___reduce_cython, __pyx_k_Literal___reduce_cython, sizeof(__pyx_k_Literal___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Literal___setstate_cython, __pyx_k_Literal___setstate_cython, sizeof(__pyx_k_Literal___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Lookup, __pyx_k_Lookup, sizeof(__pyx_k_Lookup), 0, 1, 0, 0},
    {&__pyx_n_s_Lookup_2, __pyx_k_Lookup_2, sizeof(__pyx_k_Lookup_2), 0, 0, 1, 1},
    {&__pyx_n_s_Lookup___reduce_cython, __pyx_k_Lookup___reduce_cython, sizeof(__pyx_k_Lookup___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Lookup___setstate_cython, __pyx_k_Lookup___setstate_cython, sizeof(__pyx_k_Lookup___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MathsBinaryOperation, __pyx_k_MathsBinaryOperation, sizeof(__pyx_k_MathsBinaryOperation), 0, 0, 1, 1},
    {&__pyx_n_s_MathsBinaryOperation___reduce_cy, __pyx_k_MathsBinaryOperation___reduce_cy, sizeof(__pyx_k_MathsBinaryOperation___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_MathsBinaryOperation___setstate, __pyx_k_MathsBinaryOperation___setstate, sizeof(__pyx_k_MathsBinaryOperation___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_Modulo, __pyx_k_Modulo, sizeof(__pyx_k_Modulo), 0, 0, 1, 1},
    {&__pyx_n_s_Modulo___reduce_cython, __pyx_k_Modulo___reduce_cython, sizeof(__pyx_k_Modulo___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Modulo___setstate_cython, __pyx_k_Modulo___setstate_cython, sizeof(__pyx_k_Modulo___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Multiply, __pyx_k_Multiply, sizeof(__pyx_k_Multiply), 0, 0, 1, 1},
    {&__pyx_n_s_Multiply___reduce_cython, __pyx_k_Multiply___reduce_cython, sizeof(__pyx_k_Multiply___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Multiply___setstate_cython, __pyx_k_Multiply___setstate_cython, sizeof(__pyx_k_Multiply___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Name, __pyx_k_Name, sizeof(__pyx_k_Name), 0, 1, 0, 0},
    {&__pyx_n_s_Name_2, __pyx_k_Name_2, sizeof(__pyx_k_Name_2), 0, 0, 1, 1},
    {&__pyx_n_s_Name___reduce_cython, __pyx_k_Name___reduce_cython, sizeof(__pyx_k_Name___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Name___setstate_cython, __pyx_k_Name___setstate_cython, sizeof(__pyx_k_Name___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Negative, __pyx_k_Negative, sizeof(__pyx_k_Negative), 0, 0, 1, 1},
    {&__pyx_n_s_Negative___reduce_cython, __pyx_k_Negative___reduce_cython, sizeof(__pyx_k_Negative___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Negative___setstate_cython, __pyx_k_Negative___setstate_cython, sizeof(__pyx_k_Negative___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NodeModifier, __pyx_k_NodeModifier, sizeof(__pyx_k_NodeModifier), 0, 0, 1, 1},
    {&__pyx_n_s_NodeModifier___reduce_cython, __pyx_k_NodeModifier___reduce_cython, sizeof(__pyx_k_NodeModifier___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NodeModifier___setstate_cython, __pyx_k_NodeModifier___setstate_cython, sizeof(__pyx_k_NodeModifier___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_n_s_Not, __pyx_k_Not, sizeof(__pyx_k_Not), 0, 0, 1, 1},
    {&__pyx_n_s_NotEqualTo, __pyx_k_NotEqualTo, sizeof(__pyx_k_NotEqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_NotEqualTo___reduce_cython, __pyx_k_NotEqualTo___reduce_cython, sizeof(__pyx_k_NotEqualTo___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NotEqualTo___setstate_cython, __pyx_k_NotEqualTo___setstate_cython, sizeof(__pyx_k_NotEqualTo___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
    {&__pyx_n_s_Not___reduce_cython, __pyx_k_Not___reduce_cython, sizeof(__pyx_k_Not___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Not___setstate_cython, __pyx_k_Not___setstate_cython, sizeof(__pyx_k_Not___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Or, __pyx_k_Or, sizeof(__pyx_k_Or), 0, 0, 1, 1},
    {&__pyx_n_s_Or___reduce_cython, __pyx_k_Or___reduce_cython, sizeof(__pyx_k_Or___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Or___setstate_cython, __pyx_k_Or___setstate_cython, sizeof(__pyx_k_Or___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_u_PolyBinding, __pyx_k_PolyBinding, sizeof(__pyx_k_PolyBinding), 0, 1, 0, 0},
    {&__pyx_n_s_PolyBinding_2, __pyx_k_PolyBinding_2, sizeof(__pyx_k_PolyBinding_2), 0, 0, 1, 1},
    {&__pyx_n_s_PolyBinding___reduce_cython, __pyx_k_PolyBinding___reduce_cython, sizeof(__pyx_k_PolyBinding___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PolyBinding___setstate_cython, __pyx_k_PolyBinding___setstate_cython, sizeof(__pyx_k_PolyBinding___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Positive, __pyx_k_Positive, sizeof(__pyx_k_Positive), 0, 0, 1, 1},
    {&__pyx_n_s_Positive___reduce_cython, __pyx_k_Positive___reduce_cython, sizeof(__pyx_k_Positive___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Positive___setstate_cython, __pyx_k_Positive___setstate_cython, sizeof(__pyx_k_Positive___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Power, __pyx_k_Power, sizeof(__pyx_k_Power), 0, 0, 1, 1},
    {&__pyx_n_s_Power___reduce_cython, __pyx_k_Power___reduce_cython, sizeof(__pyx_k_Power___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Power___setstate_cython, __pyx_k_Power___setstate_cython, sizeof(__pyx_k_Power___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Range, __pyx_k_Range, sizeof(__pyx_k_Range), 0, 1, 0, 0},
    {&__pyx_n_s_Range_2, __pyx_k_Range_2, sizeof(__pyx_k_Range_2), 0, 0, 1, 1},
    {&__pyx_n_s_Range___reduce_cython, __pyx_k_Range___reduce_cython, sizeof(__pyx_k_Range___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Range___setstate_cython, __pyx_k_Range___setstate_cython, sizeof(__pyx_k_Range___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 1, 0, 0},
    {&__pyx_n_s_Sequence_2, __pyx_k_Sequence_2, sizeof(__pyx_k_Sequence_2), 0, 0, 1, 1},
    {&__pyx_n_s_Sequence___reduce_cython, __pyx_k_Sequence___reduce_cython, sizeof(__pyx_k_Sequence___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Sequence___setstate_cython, __pyx_k_Sequence___setstate_cython, sizeof(__pyx_k_Sequence___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SharedCache, __pyx_k_SharedCache, sizeof(__pyx_k_SharedCache), 0, 0, 1, 1},
    {&__pyx_kp_u_Simplifier_error, __pyx_k_Simplifier_error, sizeof(__pyx_k_Simplifier_error), 0, 1, 0, 0},
    {&__pyx_kp_u_Slice, __pyx_k_Slice, sizeof(__pyx_k_Slice), 0, 1, 0, 0},
    {&__pyx_n_s_Slice_2, __pyx_k_Slice_2, sizeof(__pyx_k_Slice_2), 0, 0, 1, 1},
    {&__pyx_n_s_Slice___reduce_cython, __pyx_k_Slice___reduce_cython, sizeof(__pyx_k_Slice___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Slice___setstate_cython, __pyx_k_Slice___setstate_cython, sizeof(__pyx_k_Slice___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Subtract, __pyx_k_Subtract, sizeof(__pyx_k_Subtract), 0, 0, 1, 1},
    {&__pyx_n_s_Subtract___reduce_cython, __pyx_k_Subtract___reduce_cython, sizeof(__pyx_k_Subtract___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Subtract___setstate_cython, __pyx_k_Subtract___setstate_cython, sizeof(__pyx_k_Subtract___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Tag, __pyx_k_Tag, sizeof(__pyx_k_Tag), 0, 1, 0, 0},
    {&__pyx_n_s_Tag_2, __pyx_k_Tag_2, sizeof(__pyx_k_Tag_2), 0, 0, 1, 1},
    {&__pyx_n_s_Tag___reduce_cython, __pyx_k_Tag___reduce_cython, sizeof(__pyx_k_Tag___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Tag___setstate_cython, __pyx_k_Tag___setstate_cython, sizeof(__pyx_k_Tag___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Top, __pyx_k_Top, sizeof(__pyx_k_Top), 0, 1, 0, 0},
    {&__pyx_n_s_Top_2, __pyx_k_Top_2, sizeof(__pyx_k_Top_2), 0, 0, 1, 1},
    {&__pyx_n_s_Top___reduce_cython, __pyx_k_Top___reduce_cython, sizeof(__pyx_k_Top___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Top___setstate_cython, __pyx_k_Top___setstate_cython, sizeof(__pyx_k_Top___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_u_True, __pyx_k_True, sizeof(__pyx_k_True), 0, 1, 0, 1},
    {&__pyx_kp_u_Unable_to_simplify_program, __pyx_k_Unable_to_simplify_program, sizeof(__pyx_k_Unable_to_simplify_program), 0, 1, 0, 0},
    {&__pyx_n_s_UnaryOperation, __pyx_k_UnaryOperation, sizeof(__pyx_k_UnaryOperation), 0, 0, 1, 1},
    {&__pyx_n_s_UnaryOperation___reduce_cython, __pyx_k_UnaryOperation___reduce_cython, sizeof(__pyx_k_UnaryOperation___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnaryOperation___setstate_cython, __pyx_k_UnaryOperation___setstate_cython, sizeof(__pyx_k_UnaryOperation___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Unbound_name, __pyx_k_Unbound_name, sizeof(__pyx_k_Unbound_name), 0, 1, 0, 0},
    {&__pyx_n_s_Xor, __pyx_k_Xor, sizeof(__pyx_k_Xor), 0, 0, 1, 1},
    {&__pyx_n_s_Xor___reduce_cython, __pyx_k_Xor___reduce_cython, sizeof(__pyx_k_Xor___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Xor___setstate_cython, __pyx_k_Xor___setstate_cython, sizeof(__pyx_k_Xor___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s__186, __pyx_k__186, sizeof(__pyx_k__186), 0, 0, 1, 1},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_n_s__32, __pyx_k__32, sizeof(__pyx_k__32), 0, 0, 1, 1},
    {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_bindings, __pyx_k_bindings, sizeof(__pyx_k_bindings), 0, 0, 1, 1},
    {&__pyx_n_s_body, __pyx_k_body, sizeof(__pyx_k_body), 0, 0, 1, 1},
    {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
    {&__pyx_n_s_captures, __pyx_k_captures, sizeof(__pyx_k_captures), 0, 0, 1, 1},
    {&__pyx_n_u_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 1, 0, 1},
    {&__pyx_n_s_children, __pyx_k_children, sizeof(__pyx_k_children), 0, 0, 1, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
    {&__pyx_n_s_condition, __pyx_k_condition, sizeof(__pyx_k_condition), 0, 0, 1, 1},
    {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
    {&__pyx_n_u_context_func, __pyx_k_context_func, sizeof(__pyx_k_context_func), 0, 1, 0, 1},
    {&__pyx_n_s_context_vars, __pyx_k_context_vars, sizeof(__pyx_k_context_vars), 0, 0, 1, 1},
    {&__pyx_n_s_dependencies, __pyx_k_dependencies, sizeof(__pyx_k_dependencies), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_n_s_difference_update, __pyx_k_difference_update, sizeof(__pyx_k_difference_update), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_dynamic, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
    {&__pyx_n_s_else, __pyx_k_else, sizeof(__pyx_k_else), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_s_exc, __pyx_k_exc, sizeof(__pyx_k_exc), 0, 0, 1, 1},
    {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
    {&__pyx_n_s_expr, __pyx_k_expr, sizeof(__pyx_k_expr), 0, 0, 1, 1},
    {&__pyx_n_s_expressions, __pyx_k_expressions, sizeof(__pyx_k_expressions), 0, 0, 1, 1},
    {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
    {&__pyx_n_s_flitter_language_tree, __pyx_k_flitter_language_tree, sizeof(__pyx_k_flitter_language_tree), 0, 0, 1, 1},
    {&__pyx_n_u_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 1, 0, 1},
    {&__pyx_n_u_fract, __pyx_k_fract, sizeof(__pyx_k_fract), 0, 1, 0, 1},
    {&__pyx_n_s_function, __pyx_k_function, sizeof(__pyx_k_function), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_with_root, __pyx_k_get_with_root, sizeof(__pyx_k_get_with_root), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_initial_errors, __pyx_k_initial_errors, sizeof(__pyx_k_initial_errors), 0, 0, 1, 1},
    {&__pyx_n_s_initial_lnames, __pyx_k_initial_lnames, sizeof(__pyx_k_initial_lnames), 0, 0, 1, 1},
    {&__pyx_n_s_inlineable, __pyx_k_inlineable, sizeof(__pyx_k_inlineable), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_keyword_args, __pyx_k_keyword_args, sizeof(__pyx_k_keyword_args), 0, 0, 1, 1},
    {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
    {&__pyx_n_s_log_errors, __pyx_k_log_errors, sizeof(__pyx_k_log_errors), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_loguru, __pyx_k_loguru, sizeof(__pyx_k_loguru), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_name_patch, __pyx_k_name_patch, sizeof(__pyx_k_name_patch), 0, 0, 1, 1},
    {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
    {&__pyx_n_s_opt, __pyx_k_opt, sizeof(__pyx_k_opt), 0, 0, 1, 1},
    {&__pyx_n_s_parameters, __pyx_k_parameters, sizeof(__pyx_k_parameters), 0, 0, 1, 1},
    {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_pragmas, __pyx_k_pragmas, sizeof(__pyx_k_pragmas), 0, 0, 1, 1},
    {&__pyx_n_s_program, __pyx_k_program, sizeof(__pyx_k_program), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Add, __pyx_k_pyx_unpickle_Add, sizeof(__pyx_k_pyx_unpickle_Add), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_And, __pyx_k_pyx_unpickle_And, sizeof(__pyx_k_pyx_unpickle_And), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Append, __pyx_k_pyx_unpickle_Append, sizeof(__pyx_k_pyx_unpickle_Append), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Attributes, __pyx_k_pyx_unpickle_Attributes, sizeof(__pyx_k_pyx_unpickle_Attributes), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_BinaryOperation, __pyx_k_pyx_unpickle_BinaryOperation, sizeof(__pyx_k_pyx_unpickle_BinaryOperation), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Binding, __pyx_k_pyx_unpickle_Binding, sizeof(__pyx_k_pyx_unpickle_Binding), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Call, __pyx_k_pyx_unpickle_Call, sizeof(__pyx_k_pyx_unpickle_Call), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Ceil, __pyx_k_pyx_unpickle_Ceil, sizeof(__pyx_k_pyx_unpickle_Ceil), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Comparison, __pyx_k_pyx_unpickle_Comparison, sizeof(__pyx_k_pyx_unpickle_Comparison), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Divide, __pyx_k_pyx_unpickle_Divide, sizeof(__pyx_k_pyx_unpickle_Divide), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_EqualTo, __pyx_k_pyx_unpickle_EqualTo, sizeof(__pyx_k_pyx_unpickle_EqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Export, __pyx_k_pyx_unpickle_Export, sizeof(__pyx_k_pyx_unpickle_Export), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Expression, __pyx_k_pyx_unpickle_Expression, sizeof(__pyx_k_pyx_unpickle_Expression), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Floor, __pyx_k_pyx_unpickle_Floor, sizeof(__pyx_k_pyx_unpickle_Floor), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_FloorDivide, __pyx_k_pyx_unpickle_FloorDivide, sizeof(__pyx_k_pyx_unpickle_FloorDivide), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_For, __pyx_k_pyx_unpickle_For, sizeof(__pyx_k_pyx_unpickle_For), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Fract, __pyx_k_pyx_unpickle_Fract, sizeof(__pyx_k_pyx_unpickle_Fract), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Function, __pyx_k_pyx_unpickle_Function, sizeof(__pyx_k_pyx_unpickle_Function), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GreaterThan, __pyx_k_pyx_unpickle_GreaterThan, sizeof(__pyx_k_pyx_unpickle_GreaterThan), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_GreaterThanOrEqua, __pyx_k_pyx_unpickle_GreaterThanOrEqua, sizeof(__pyx_k_pyx_unpickle_GreaterThanOrEqua), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IfCondition, __pyx_k_pyx_unpickle_IfCondition, sizeof(__pyx_k_pyx_unpickle_IfCondition), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IfElse, __pyx_k_pyx_unpickle_IfElse, sizeof(__pyx_k_pyx_unpickle_IfElse), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Import, __pyx_k_pyx_unpickle_Import, sizeof(__pyx_k_pyx_unpickle_Import), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LessThan, __pyx_k_pyx_unpickle_LessThan, sizeof(__pyx_k_pyx_unpickle_LessThan), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LessThanOrEqualTo, __pyx_k_pyx_unpickle_LessThanOrEqualTo, sizeof(__pyx_k_pyx_unpickle_LessThanOrEqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Let, __pyx_k_pyx_unpickle_Let, sizeof(__pyx_k_pyx_unpickle_Let), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Literal, __pyx_k_pyx_unpickle_Literal, sizeof(__pyx_k_pyx_unpickle_Literal), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Lookup, __pyx_k_pyx_unpickle_Lookup, sizeof(__pyx_k_pyx_unpickle_Lookup), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_MathsBinaryOperat, __pyx_k_pyx_unpickle_MathsBinaryOperat, sizeof(__pyx_k_pyx_unpickle_MathsBinaryOperat), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Modulo, __pyx_k_pyx_unpickle_Modulo, sizeof(__pyx_k_pyx_unpickle_Modulo), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Multiply, __pyx_k_pyx_unpickle_Multiply, sizeof(__pyx_k_pyx_unpickle_Multiply), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Name, __pyx_k_pyx_unpickle_Name, sizeof(__pyx_k_pyx_unpickle_Name), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Negative, __pyx_k_pyx_unpickle_Negative, sizeof(__pyx_k_pyx_unpickle_Negative), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_NodeModifier, __pyx_k_pyx_unpickle_NodeModifier, sizeof(__pyx_k_pyx_unpickle_NodeModifier), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Not, __pyx_k_pyx_unpickle_Not, sizeof(__pyx_k_pyx_unpickle_Not), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_NotEqualTo, __pyx_k_pyx_unpickle_NotEqualTo, sizeof(__pyx_k_pyx_unpickle_NotEqualTo), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Or, __pyx_k_pyx_unpickle_Or, sizeof(__pyx_k_pyx_unpickle_Or), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PolyBinding, __pyx_k_pyx_unpickle_PolyBinding, sizeof(__pyx_k_pyx_unpickle_PolyBinding), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Positive, __pyx_k_pyx_unpickle_Positive, sizeof(__pyx_k_pyx_unpickle_Positive), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Power, __pyx_k_pyx_unpickle_Power, sizeof(__pyx_k_pyx_unpickle_Power), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Range, __pyx_k_pyx_unpickle_Range, sizeof(__pyx_k_pyx_unpickle_Range), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Sequence, __pyx_k_pyx_unpickle_Sequence, sizeof(__pyx_k_pyx_unpickle_Sequence), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Slice, __pyx_k_pyx_unpickle_Slice, sizeof(__pyx_k_pyx_unpickle_Slice), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Subtract, __pyx_k_pyx_unpickle_Subtract, sizeof(__pyx_k_pyx_unpickle_Subtract), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Tag, __pyx_k_pyx_unpickle_Tag, sizeof(__pyx_k_pyx_unpickle_Tag), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Top, __pyx_k_pyx_unpickle_Top, sizeof(__pyx_k_pyx_unpickle_Top), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UnaryOperation, __pyx_k_pyx_unpickle_UnaryOperation, sizeof(__pyx_k_pyx_unpickle_UnaryOperation), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Xor, __pyx_k_pyx_unpickle_Xor, sizeof(__pyx_k_pyx_unpickle_Xor), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_read_flitter_top, __pyx_k_read_flitter_top, sizeof(__pyx_k_read_flitter_top), 0, 0, 1, 1},
    {&__pyx_n_s_recursive, __pyx_k_recursive, sizeof(__pyx_k_recursive), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_return_context, __pyx_k_return_context, sizeof(__pyx_k_return_context), 0, 0, 1, 1},
    {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
    {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set_pragma, __pyx_k_set_pragma, sizeof(__pyx_k_set_pragma), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_simplify, __pyx_k_simplify, sizeof(__pyx_k_simplify), 0, 0, 1, 1},
    {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
    {&__pyx_kp_s_src_flitter_language_tree_pyx, __pyx_k_src_flitter_language_tree_pyx, sizeof(__pyx_k_src_flitter_language_tree_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_static, __pyx_k_static, sizeof(__pyx_k_static), 0, 0, 1, 1},
    {&__pyx_n_s_static_exports, __pyx_k_static_exports, sizeof(__pyx_k_static_exports), 0, 0, 1, 1},
    {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_tag, __pyx_k_tag, sizeof(__pyx_k_tag), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_tests, __pyx_k_tests, sizeof(__pyx_k_tests), 0, 0, 1, 1},
    {&__pyx_n_s_then, __pyx_k_then, sizeof(__pyx_k_then), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_reversed = __Pyx_GetBuiltinName(__pyx_n_s_reversed); if (!__pyx_builtin_reversed) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 935, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 955, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "flitter/language/tree.pyx":706
 *     cdef void _compile_op(self, Program program):
 *         cdef Instruction instr = program.last_instruction()
 *         if instr.code == OpCode.Literal and (<InstructionVector>instr).value.eq(Vector(2)):             # <<<<<<<<<<<<<<
 *             program.pop_instruction()
 *             program.dup()
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_tuple__7 = PyTuple_Pack(3, __pyx_int_238750788, __pyx_int_228825662, __pyx_int_222419149); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_193537176, __pyx_int_81599198, __pyx_int_29776540); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_int_19136085, __pyx_int_256930855, __pyx_int_48890310); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_65788850, __pyx_int_39246950, __pyx_int_149349987); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__12 = PyTuple_Pack(3, __pyx_int_25936640, __pyx_int_262293510, __pyx_int_21458427); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_int_215229444, __pyx_int_254981756, __pyx_int_33963030); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_136983863, __pyx_int_112105877, __pyx_int_184977713); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_int_46599698, __pyx_int_174256674, __pyx_int_63365304); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_tuple__16 = PyTuple_Pack(3, __pyx_int_264124616, __pyx_int_147834117, __pyx_int_87851177); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_int_101425629, __pyx_int_16586848, __pyx_int_147971132); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_int_20102196, __pyx_int_257836063, __pyx_int_132590791); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_tuple__19 = PyTuple_Pack(3, __pyx_int_61976268, __pyx_int_103159781, __pyx_int_221199149); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_tuple__20 = PyTuple_Pack(3, __pyx_int_126472276, __pyx_int_241444487, __pyx_int_168508340); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_tuple__21 = PyTuple_Pack(3, __pyx_int_88468051, __pyx_int_261002861, __pyx_int_57427254); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_int_158473136, __pyx_int_211607581, __pyx_int_264265876); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_tuple__23 = PyTuple_Pack(3, __pyx_int_89314571, __pyx_int_42909475, __pyx_int_253263515); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_int_217770541, __pyx_int_89160345, __pyx_int_114850062); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_int_60997603, __pyx_int_19056541, __pyx_int_257251258); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_int_19657503, __pyx_int_201276235, __pyx_int_60181930); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_tuple__27 = PyTuple_Pack(3, __pyx_int_237118084, __pyx_int_209144725, __pyx_int_54761546); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_tuple__28 = PyTuple_Pack(3, __pyx_int_64598595, __pyx_int_58806657, __pyx_int_222754444); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_tuple__29 = PyTuple_Pack(3, __pyx_int_161766002, __pyx_int_251934659, __pyx_int_209608191); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_tuple__30 = PyTuple_Pack(3, __pyx_int_231565212, __pyx_int_18525531, __pyx_int_5413075); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_int_256597255, __pyx_int_134496108, __pyx_int_211749586); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "flitter/language/tree.pyx":50
 * 
 * cdef class Expression:
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):             # <<<<<<<<<<<<<<
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:
 */
  __pyx_tuple__33 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_initial_lnames, __pyx_n_s_initial_errors, __pyx_n_s_log_errors, __pyx_n_s_program, __pyx_n_s_error); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_tree_pyx, __pyx_n_s_compile, 50, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_tuple__35 = PyTuple_Pack(3, __pyx_empty_tuple, Py_None, Py_True); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "flitter/language/tree.pyx":61
 *         return program
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):             # <<<<<<<<<<<<<<
 *         cdef dict context_vars = {}
 *         cdef str key
 */
  __pyx_tuple__36 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_static, __pyx_n_s_dynamic, __pyx_n_s_parent, __pyx_n_s_path, __pyx_n_s_return_context, __pyx_n_s_context_vars, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_context, __pyx_n_s_expr, __pyx_n_s_exc, __pyx_n_s_error); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(7, 0, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_language_tree_pyx, __pyx_n_s_simplify, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_tuple__38 = PyTuple_Pack(6, Py_None, Py_None, Py_None, Py_None, Py_None, Py_False); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__39 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)
 */
  __pyx_tuple__41 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)
 */
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)
 */
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)
 */
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)
 */
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)
 */
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)
 */
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)
 */
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)
 */
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)
 */
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)
 */
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)
 */
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)
 */
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)
 */
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)
 */
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)
 */
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)
 */
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)
 */
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)
 */
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)
 */
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_And__set_state(self, __pyx_state)
 */
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)
 */
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)
 */
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)
 */
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)
 */
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)
 */
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)
 */
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)
 */
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)
 */
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)
 */
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_For__set_state(self, __pyx_state)
 */
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)
 */
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)
 */
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Expression(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__137 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Expression, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Top, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Export, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Import, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Sequence, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Literal, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Name, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__145 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Lookup, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__145)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Range, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UnaryOperation, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Negative, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Positive, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Not, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Ceil, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Floor, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Fract, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BinaryOperation, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MathsBinaryOperat, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Add, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__157 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Subtract, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__157)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Multiply, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__159 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Divide, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__159)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FloorDivide, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__161 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Modulo, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__161)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Power, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__163 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Comparison, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__163)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_EqualTo, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_NotEqualTo, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LessThan, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__167 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GreaterThan, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__167)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LessThanOrEqualTo, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__169 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GreaterThanOrEqua, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__169)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_And, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__171 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Or, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__171)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Xor, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__173 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Slice, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__173)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Call, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__175 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_NodeModifier, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__175)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Tag, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__177 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Attributes, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__177)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Append, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__179 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Binding, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__179)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PolyBinding, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__181 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Let, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__181)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_For, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__183 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IfCondition, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__183)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IfElse, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_codeobj__185 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Function, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__185)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_get.method_name = &__pyx_n_s_get;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyList_Type_pop.method_name = &__pyx_n_s_pop;
  __pyx_umethod_PySet_Type_update.type = (PyObject*)&PySet_Type;
  __pyx_umethod_PySet_Type_update.method_name = &__pyx_n_s_update;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 2, __pyx_L1_error);
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_5413075 = PyInt_FromLong(5413075L); if (unlikely(!__pyx_int_5413075)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_16586848 = PyInt_FromLong(16586848L); if (unlikely(!__pyx_int_16586848)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_18525531 = PyInt_FromLong(18525531L); if (unlikely(!__pyx_int_18525531)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_19056541 = PyInt_FromLong(19056541L); if (unlikely(!__pyx_int_19056541)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_19136085 = PyInt_FromLong(19136085L); if (unlikely(!__pyx_int_19136085)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_19657503 = PyInt_FromLong(19657503L); if (unlikely(!__pyx_int_19657503)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_20102196 = PyInt_FromLong(20102196L); if (unlikely(!__pyx_int_20102196)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_21458427 = PyInt_FromLong(21458427L); if (unlikely(!__pyx_int_21458427)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_25936640 = PyInt_FromLong(25936640L); if (unlikely(!__pyx_int_25936640)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_29776540 = PyInt_FromLong(29776540L); if (unlikely(!__pyx_int_29776540)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_33963030 = PyInt_FromLong(33963030L); if (unlikely(!__pyx_int_33963030)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_39246950 = PyInt_FromLong(39246950L); if (unlikely(!__pyx_int_39246950)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_42909475 = PyInt_FromLong(42909475L); if (unlikely(!__pyx_int_42909475)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_46599698 = PyInt_FromLong(46599698L); if (unlikely(!__pyx_int_46599698)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_48890310 = PyInt_FromLong(48890310L); if (unlikely(!__pyx_int_48890310)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_54761546 = PyInt_FromLong(54761546L); if (unlikely(!__pyx_int_54761546)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_57427254 = PyInt_FromLong(57427254L); if (unlikely(!__pyx_int_57427254)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_58806657 = PyInt_FromLong(58806657L); if (unlikely(!__pyx_int_58806657)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_60181930 = PyInt_FromLong(60181930L); if (unlikely(!__pyx_int_60181930)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_60997603 = PyInt_FromLong(60997603L); if (unlikely(!__pyx_int_60997603)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_61976268 = PyInt_FromLong(61976268L); if (unlikely(!__pyx_int_61976268)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_63365304 = PyInt_FromLong(63365304L); if (unlikely(!__pyx_int_63365304)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_64598595 = PyInt_FromLong(64598595L); if (unlikely(!__pyx_int_64598595)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_65788850 = PyInt_FromLong(65788850L); if (unlikely(!__pyx_int_65788850)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_81599198 = PyInt_FromLong(81599198L); if (unlikely(!__pyx_int_81599198)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_87851177 = PyInt_FromLong(87851177L); if (unlikely(!__pyx_int_87851177)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_88468051 = PyInt_FromLong(88468051L); if (unlikely(!__pyx_int_88468051)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_89160345 = PyInt_FromLong(89160345L); if (unlikely(!__pyx_int_89160345)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_89314571 = PyInt_FromLong(89314571L); if (unlikely(!__pyx_int_89314571)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_101425629 = PyInt_FromLong(101425629L); if (unlikely(!__pyx_int_101425629)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_103159781 = PyInt_FromLong(103159781L); if (unlikely(!__pyx_int_103159781)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_114850062 = PyInt_FromLong(114850062L); if (unlikely(!__pyx_int_114850062)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_126472276 = PyInt_FromLong(126472276L); if (unlikely(!__pyx_int_126472276)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_132590791 = PyInt_FromLong(132590791L); if (unlikely(!__pyx_int_132590791)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_134496108 = PyInt_FromLong(134496108L); if (unlikely(!__pyx_int_134496108)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_147834117 = PyInt_FromLong(147834117L); if (unlikely(!__pyx_int_147834117)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_147971132 = PyInt_FromLong(147971132L); if (unlikely(!__pyx_int_147971132)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_149349987 = PyInt_FromLong(149349987L); if (unlikely(!__pyx_int_149349987)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_158473136 = PyInt_FromLong(158473136L); if (unlikely(!__pyx_int_158473136)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_161766002 = PyInt_FromLong(161766002L); if (unlikely(!__pyx_int_161766002)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_168508340 = PyInt_FromLong(168508340L); if (unlikely(!__pyx_int_168508340)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_174256674 = PyInt_FromLong(174256674L); if (unlikely(!__pyx_int_174256674)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_193537176 = PyInt_FromLong(193537176L); if (unlikely(!__pyx_int_193537176)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_201276235 = PyInt_FromLong(201276235L); if (unlikely(!__pyx_int_201276235)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_209144725 = PyInt_FromLong(209144725L); if (unlikely(!__pyx_int_209144725)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_209608191 = PyInt_FromLong(209608191L); if (unlikely(!__pyx_int_209608191)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_211607581 = PyInt_FromLong(211607581L); if (unlikely(!__pyx_int_211607581)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_211749586 = PyInt_FromLong(211749586L); if (unlikely(!__pyx_int_211749586)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_215229444 = PyInt_FromLong(215229444L); if (unlikely(!__pyx_int_215229444)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_217770541 = PyInt_FromLong(217770541L); if (unlikely(!__pyx_int_217770541)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_221199149 = PyInt_FromLong(221199149L); if (unlikely(!__pyx_int_221199149)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_222754444 = PyInt_FromLong(222754444L); if (unlikely(!__pyx_int_222754444)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_228825662 = PyInt_FromLong(228825662L); if (unlikely(!__pyx_int_228825662)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_231565212 = PyInt_FromLong(231565212L); if (unlikely(!__pyx_int_231565212)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_237118084 = PyInt_FromLong(237118084L); if (unlikely(!__pyx_int_237118084)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_238750788 = PyInt_FromLong(238750788L); if (unlikely(!__pyx_int_238750788)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_241444487 = PyInt_FromLong(241444487L); if (unlikely(!__pyx_int_241444487)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_251934659 = PyInt_FromLong(251934659L); if (unlikely(!__pyx_int_251934659)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_253263515 = PyInt_FromLong(253263515L); if (unlikely(!__pyx_int_253263515)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_254981756 = PyInt_FromLong(254981756L); if (unlikely(!__pyx_int_254981756)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_256597255 = PyInt_FromLong(256597255L); if (unlikely(!__pyx_int_256597255)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_256930855 = PyInt_FromLong(256930855L); if (unlikely(!__pyx_int_256930855)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_257251258 = PyInt_FromLong(257251258L); if (unlikely(!__pyx_int_257251258)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_257836063 = PyInt_FromLong(257836063L); if (unlikely(!__pyx_int_257836063)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_261002861 = PyInt_FromLong(261002861L); if (unlikely(!__pyx_int_261002861)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_262293510 = PyInt_FromLong(262293510L); if (unlikely(!__pyx_int_262293510)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_264124616 = PyInt_FromLong(264124616L); if (unlikely(!__pyx_int_264124616)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_264265876 = PyInt_FromLong(264265876L); if (unlikely(!__pyx_int_264265876)) __PYX_ERR(0, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_7flitter_8language_4tree_NoOp = ((struct __pyx_obj_7flitter_8language_4tree_Literal *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_7flitter_8language_4tree_Expression = &__pyx_vtable_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Expression._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_10Expression__compile;
  __pyx_vtable_7flitter_8language_4tree_Expression._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_10Expression__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_4tree_Expression = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Expression_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Expression)) __PYX_ERR(0, 49, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Expression_spec, __pyx_ptype_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Expression = &__pyx_type_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Expression->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Expression->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Expression->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Expression->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Expression, __pyx_vtabptr_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Expression, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Expression) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Top = &__pyx_vtable_7flitter_8language_4tree_Top;
  __pyx_vtable_7flitter_8language_4tree_Top.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Top.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_3Top__compile;
  __pyx_vtable_7flitter_8language_4tree_Top.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3Top__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Top = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Top_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Top)) __PYX_ERR(0, 92, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Top_spec, __pyx_ptype_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Top = &__pyx_type_7flitter_8language_4tree_Top;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Top->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Top->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Top->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Top->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Top->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Top, __pyx_vtabptr_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Top_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Top) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Export = &__pyx_vtable_7flitter_8language_4tree_Export;
  __pyx_vtable_7flitter_8language_4tree_Export.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Export.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_6Export__compile;
  __pyx_vtable_7flitter_8language_4tree_Export.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_6Export__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Export = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Export_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Export)) __PYX_ERR(0, 126, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Export_spec, __pyx_ptype_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Export = &__pyx_type_7flitter_8language_4tree_Export;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Export->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Export->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Export->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Export->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Export->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Export, __pyx_vtabptr_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Export_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Export) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Import = &__pyx_vtable_7flitter_8language_4tree_Import;
  __pyx_vtable_7flitter_8language_4tree_Import.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Import.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_6Import__compile;
  __pyx_vtable_7flitter_8language_4tree_Import.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_6Import__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Import = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Import_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Import)) __PYX_ERR(0, 166, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Import_spec, __pyx_ptype_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Import = &__pyx_type_7flitter_8language_4tree_Import;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Import->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Import->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Import->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Import->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Import->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Import, __pyx_vtabptr_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Import_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Import) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Sequence = &__pyx_vtable_7flitter_8language_4tree_Sequence;
  __pyx_vtable_7flitter_8language_4tree_Sequence.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Sequence.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_8Sequence__compile;
  __pyx_vtable_7flitter_8language_4tree_Sequence.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_8Sequence__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Sequence = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Sequence_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Sequence)) __PYX_ERR(0, 234, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Sequence_spec, __pyx_ptype_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Sequence = &__pyx_type_7flitter_8language_4tree_Sequence;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Sequence->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Sequence->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Sequence->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Sequence->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Sequence->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Sequence, __pyx_vtabptr_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Sequence_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Sequence) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Literal = &__pyx_vtable_7flitter_8language_4tree_Literal;
  __pyx_vtable_7flitter_8language_4tree_Literal.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Literal.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_7Literal__compile;
  __pyx_vtable_7flitter_8language_4tree_Literal.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_7Literal__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Literal = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Literal_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Literal)) __PYX_ERR(0, 276, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Literal_spec, __pyx_ptype_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Literal = &__pyx_type_7flitter_8language_4tree_Literal;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Literal->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Literal->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Literal->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Literal->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Literal->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Literal, __pyx_vtabptr_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Literal_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Literal) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Name = &__pyx_vtable_7flitter_8language_4tree_Name;
  __pyx_vtable_7flitter_8language_4tree_Name.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Name.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_4Name__compile;
  __pyx_vtable_7flitter_8language_4tree_Name.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_4Name__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Name = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Name_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Name)) __PYX_ERR(0, 292, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Name_spec, __pyx_ptype_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Name = &__pyx_type_7flitter_8language_4tree_Name;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Name->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Name->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Name->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Name->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Name->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Name, __pyx_vtabptr_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Name_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Name) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Lookup = &__pyx_vtable_7flitter_8language_4tree_Lookup;
  __pyx_vtable_7flitter_8language_4tree_Lookup.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Lookup.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_6Lookup__compile;
  __pyx_vtable_7flitter_8language_4tree_Lookup.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_6Lookup__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Lookup = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Lookup_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Lookup)) __PYX_ERR(0, 338, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Lookup_spec, __pyx_ptype_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Lookup = &__pyx_type_7flitter_8language_4tree_Lookup;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Lookup->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Lookup->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Lookup->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Lookup->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Lookup->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Lookup, __pyx_vtabptr_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Lookup_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Lookup) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Range = &__pyx_vtable_7flitter_8language_4tree_Range;
  __pyx_vtable_7flitter_8language_4tree_Range.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Range.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_5Range__compile;
  __pyx_vtable_7flitter_8language_4tree_Range.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_5Range__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Range = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Range_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Range)) __PYX_ERR(0, 366, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Range_spec, __pyx_ptype_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Range = &__pyx_type_7flitter_8language_4tree_Range;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Range->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Range->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Range->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Range->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Range->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Range, __pyx_vtabptr_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Range_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Range) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_UnaryOperation = &__pyx_vtable_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_UnaryOperation.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_UnaryOperation.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_14UnaryOperation__compile;
  __pyx_vtable_7flitter_8language_4tree_UnaryOperation._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_14UnaryOperation__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_UnaryOperation = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_UnaryOperation_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_UnaryOperation)) __PYX_ERR(0, 399, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_UnaryOperation_spec, __pyx_ptype_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_UnaryOperation = &__pyx_type_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_UnaryOperation->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_UnaryOperation->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_UnaryOperation->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_UnaryOperation->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_UnaryOperation->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_UnaryOperation, __pyx_vtabptr_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UnaryOperation, (PyObject *) __pyx_ptype_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_UnaryOperation) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Negative = &__pyx_vtable_7flitter_8language_4tree_Negative;
  __pyx_vtable_7flitter_8language_4tree_Negative.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Negative.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_8Negative__simplify;
  __pyx_vtable_7flitter_8language_4tree_Negative.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_8Negative__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Negative = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Negative_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Negative)) __PYX_ERR(0, 416, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Negative_spec, __pyx_ptype_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Negative = &__pyx_type_7flitter_8language_4tree_Negative;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Negative->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Negative->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Negative->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Negative->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Negative->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Negative, __pyx_vtabptr_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Negative, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Negative) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Positive = &__pyx_vtable_7flitter_8language_4tree_Positive;
  __pyx_vtable_7flitter_8language_4tree_Positive.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Positive.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_8Positive__simplify;
  __pyx_vtable_7flitter_8language_4tree_Positive.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_8Positive__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Positive = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Positive_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Positive)) __PYX_ERR(0, 451, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Positive_spec, __pyx_ptype_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Positive = &__pyx_type_7flitter_8language_4tree_Positive;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Positive->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Positive->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Positive->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Positive->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Positive->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Positive, __pyx_vtabptr_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Positive, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Positive) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Not = &__pyx_vtable_7flitter_8language_4tree_Not;
  __pyx_vtable_7flitter_8language_4tree_Not.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Not.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3Not__simplify;
  __pyx_vtable_7flitter_8language_4tree_Not.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_3Not__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Not = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Not_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Not)) __PYX_ERR(0, 466, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Not_spec, __pyx_ptype_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Not = &__pyx_type_7flitter_8language_4tree_Not;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Not->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Not->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Not->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Not->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Not->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Not, __pyx_vtabptr_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Not, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Not) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Ceil = &__pyx_vtable_7flitter_8language_4tree_Ceil;
  __pyx_vtable_7flitter_8language_4tree_Ceil.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Ceil.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_4Ceil__simplify;
  __pyx_vtable_7flitter_8language_4tree_Ceil.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_4Ceil__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Ceil = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Ceil_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Ceil)) __PYX_ERR(0, 479, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Ceil_spec, __pyx_ptype_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Ceil = &__pyx_type_7flitter_8language_4tree_Ceil;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Ceil->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Ceil->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Ceil->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Ceil->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Ceil->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Ceil, __pyx_vtabptr_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Ceil, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Ceil) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Floor = &__pyx_vtable_7flitter_8language_4tree_Floor;
  __pyx_vtable_7flitter_8language_4tree_Floor.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Floor.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_5Floor__simplify;
  __pyx_vtable_7flitter_8language_4tree_Floor.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_5Floor__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Floor = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Floor_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Floor)) __PYX_ERR(0, 492, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Floor_spec, __pyx_ptype_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Floor = &__pyx_type_7flitter_8language_4tree_Floor;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Floor->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Floor->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Floor->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Floor->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Floor->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Floor, __pyx_vtabptr_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Floor, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Floor) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Fract = &__pyx_vtable_7flitter_8language_4tree_Fract;
  __pyx_vtable_7flitter_8language_4tree_Fract.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_UnaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Fract.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_5Fract__simplify;
  __pyx_vtable_7flitter_8language_4tree_Fract.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_UnaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_5Fract__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Fract = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Fract_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Fract)) __PYX_ERR(0, 505, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Fract_spec, __pyx_ptype_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Fract = &__pyx_type_7flitter_8language_4tree_Fract;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Fract->tp_base = __pyx_ptype_7flitter_8language_4tree_UnaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Fract->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Fract->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Fract->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Fract->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Fract, __pyx_vtabptr_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Fract, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Fract) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_BinaryOperation = &__pyx_vtable_7flitter_8language_4tree_BinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_15BinaryOperation__compile;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_15BinaryOperation__simplify;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_15BinaryOperation__compile_op;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_15BinaryOperation_op;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.constant_left = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_left;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_15BinaryOperation_constant_right;
  __pyx_vtable_7flitter_8language_4tree_BinaryOperation.additional_rules = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_15BinaryOperation_additional_rules;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_BinaryOperation = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_BinaryOperation_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_BinaryOperation)) __PYX_ERR(0, 518, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_BinaryOperation_spec, __pyx_ptype_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_BinaryOperation = &__pyx_type_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_BinaryOperation->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_BinaryOperation->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_BinaryOperation->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_BinaryOperation->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_BinaryOperation->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_BinaryOperation, __pyx_vtabptr_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BinaryOperation, (PyObject *) __pyx_ptype_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_BinaryOperation) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation = &__pyx_vtable_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_MathsBinaryOperation.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_MathsBinaryOperation.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_20MathsBinaryOperation__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_MathsBinaryOperation_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation)) __PYX_ERR(0, 570, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_MathsBinaryOperation_spec, __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation = &__pyx_type_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation->tp_base = __pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation, __pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MathsBinaryOperation, (PyObject *) __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Add = &__pyx_vtable_7flitter_8language_4tree_Add;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_3Add__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_3Add_op;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base.__pyx_base.constant_left = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_3Add_constant_left;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_3Add_constant_right;
  __pyx_vtable_7flitter_8language_4tree_Add.__pyx_base.__pyx_base.additional_rules = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_3Add_additional_rules;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Add = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Add_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Add)) __PYX_ERR(0, 581, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Add_spec, __pyx_ptype_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Add = &__pyx_type_7flitter_8language_4tree_Add;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Add->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Add->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Add->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Add->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Add->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Add, __pyx_vtabptr_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Add, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Add) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Subtract = &__pyx_vtable_7flitter_8language_4tree_Subtract;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_8Subtract__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_8Subtract_op;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base.__pyx_base.constant_left = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_8Subtract_constant_left;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_8Subtract_constant_right;
  __pyx_vtable_7flitter_8language_4tree_Subtract.__pyx_base.__pyx_base.additional_rules = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_8Subtract_additional_rules;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Subtract = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Subtract_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Subtract)) __PYX_ERR(0, 608, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Subtract_spec, __pyx_ptype_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Subtract = &__pyx_type_7flitter_8language_4tree_Subtract;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Subtract->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Subtract->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Subtract->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Subtract->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Subtract->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Subtract, __pyx_vtabptr_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Subtract, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Subtract) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Multiply = &__pyx_vtable_7flitter_8language_4tree_Multiply;
  __pyx_vtable_7flitter_8language_4tree_Multiply.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Multiply.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_8Multiply__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Multiply.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_8Multiply_op;
  __pyx_vtable_7flitter_8language_4tree_Multiply.__pyx_base.__pyx_base.constant_left = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_8language_4tree_Expression *))__pyx_f_7flitter_8language_4tree_8Multiply_constant_left;
  __pyx_vtable_7flitter_8language_4tree_Multiply.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_8Multiply_constant_right;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Multiply = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Multiply_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Multiply)) __PYX_ERR(0, 629, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Multiply_spec, __pyx_ptype_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Multiply = &__pyx_type_7flitter_8language_4tree_Multiply;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Multiply->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Multiply->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Multiply->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Multiply->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Multiply->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Multiply, __pyx_vtabptr_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Multiply, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Multiply) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Divide = &__pyx_vtable_7flitter_8language_4tree_Divide;
  __pyx_vtable_7flitter_8language_4tree_Divide.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Divide.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_6Divide__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Divide.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_6Divide_op;
  __pyx_vtable_7flitter_8language_4tree_Divide.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_6Divide_constant_right;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Divide = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Divide_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Divide)) __PYX_ERR(0, 663, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Divide_spec, __pyx_ptype_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Divide = &__pyx_type_7flitter_8language_4tree_Divide;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Divide->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Divide->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Divide->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Divide->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Divide->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Divide, __pyx_vtabptr_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Divide, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Divide) < 0) __PYX_ERR(0, 663, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_FloorDivide = &__pyx_vtable_7flitter_8language_4tree_FloorDivide;
  __pyx_vtable_7flitter_8language_4tree_FloorDivide.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_FloorDivide.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_11FloorDivide__compile_op;
  __pyx_vtable_7flitter_8language_4tree_FloorDivide.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_11FloorDivide_op;
  __pyx_vtable_7flitter_8language_4tree_FloorDivide.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_11FloorDivide_constant_right;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_FloorDivide = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_FloorDivide_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_FloorDivide)) __PYX_ERR(0, 676, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_FloorDivide_spec, __pyx_ptype_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_FloorDivide = &__pyx_type_7flitter_8language_4tree_FloorDivide;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_FloorDivide->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_FloorDivide->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_FloorDivide->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_FloorDivide->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_FloorDivide->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_FloorDivide, __pyx_vtabptr_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FloorDivide, (PyObject *) __pyx_ptype_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_FloorDivide) < 0) __PYX_ERR(0, 676, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Modulo = &__pyx_vtable_7flitter_8language_4tree_Modulo;
  __pyx_vtable_7flitter_8language_4tree_Modulo.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Modulo.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_6Modulo__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Modulo.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_6Modulo_op;
  __pyx_vtable_7flitter_8language_4tree_Modulo.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_6Modulo_constant_right;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Modulo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Modulo_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Modulo)) __PYX_ERR(0, 688, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Modulo_spec, __pyx_ptype_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Modulo = &__pyx_type_7flitter_8language_4tree_Modulo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Modulo->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Modulo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Modulo->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Modulo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Modulo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Modulo, __pyx_vtabptr_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Modulo, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Modulo) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Power = &__pyx_vtable_7flitter_8language_4tree_Power;
  __pyx_vtable_7flitter_8language_4tree_Power.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_MathsBinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Power.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_5Power__compile_op;
  __pyx_vtable_7flitter_8language_4tree_Power.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_5Power_op;
  __pyx_vtable_7flitter_8language_4tree_Power.__pyx_base.__pyx_base.constant_right = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_5Power_constant_right;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Power = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Power_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Power)) __PYX_ERR(0, 700, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Power_spec, __pyx_ptype_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Power = &__pyx_type_7flitter_8language_4tree_Power;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Power->tp_base = __pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Power->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Power->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Power->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Power->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Power, __pyx_vtabptr_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Power, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Power) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Comparison = &__pyx_vtable_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_Comparison.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Comparison = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Comparison_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Comparison)) __PYX_ERR(0, 718, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Comparison_spec, __pyx_ptype_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Comparison = &__pyx_type_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Comparison->tp_base = __pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Comparison->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Comparison->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Comparison->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Comparison->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Comparison, __pyx_vtabptr_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Comparison, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Comparison) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_EqualTo = &__pyx_vtable_7flitter_8language_4tree_EqualTo;
  __pyx_vtable_7flitter_8language_4tree_EqualTo.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_EqualTo.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_7EqualTo__compile_op;
  __pyx_vtable_7flitter_8language_4tree_EqualTo.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_7EqualTo_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_EqualTo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_EqualTo_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_EqualTo)) __PYX_ERR(0, 722, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_EqualTo_spec, __pyx_ptype_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_EqualTo = &__pyx_type_7flitter_8language_4tree_EqualTo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_EqualTo->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_EqualTo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_EqualTo->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_EqualTo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_EqualTo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_EqualTo, __pyx_vtabptr_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EqualTo, (PyObject *) __pyx_ptype_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_EqualTo) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_NotEqualTo = &__pyx_vtable_7flitter_8language_4tree_NotEqualTo;
  __pyx_vtable_7flitter_8language_4tree_NotEqualTo.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_NotEqualTo.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_10NotEqualTo__compile_op;
  __pyx_vtable_7flitter_8language_4tree_NotEqualTo.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_10NotEqualTo_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_NotEqualTo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_NotEqualTo_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_NotEqualTo)) __PYX_ERR(0, 730, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_NotEqualTo_spec, __pyx_ptype_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_NotEqualTo = &__pyx_type_7flitter_8language_4tree_NotEqualTo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_NotEqualTo->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_NotEqualTo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_NotEqualTo->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_NotEqualTo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_NotEqualTo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_NotEqualTo, __pyx_vtabptr_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NotEqualTo, (PyObject *) __pyx_ptype_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_NotEqualTo) < 0) __PYX_ERR(0, 730, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_LessThan = &__pyx_vtable_7flitter_8language_4tree_LessThan;
  __pyx_vtable_7flitter_8language_4tree_LessThan.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_LessThan.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_8LessThan__compile_op;
  __pyx_vtable_7flitter_8language_4tree_LessThan.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_8LessThan_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_LessThan = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_LessThan_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_LessThan)) __PYX_ERR(0, 738, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_LessThan_spec, __pyx_ptype_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_LessThan = &__pyx_type_7flitter_8language_4tree_LessThan;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_LessThan->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_LessThan->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_LessThan->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_LessThan->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_LessThan->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_LessThan, __pyx_vtabptr_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LessThan, (PyObject *) __pyx_ptype_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_LessThan) < 0) __PYX_ERR(0, 738, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_GreaterThan = &__pyx_vtable_7flitter_8language_4tree_GreaterThan;
  __pyx_vtable_7flitter_8language_4tree_GreaterThan.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_GreaterThan.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_11GreaterThan__compile_op;
  __pyx_vtable_7flitter_8language_4tree_GreaterThan.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_11GreaterThan_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_GreaterThan = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_GreaterThan_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_GreaterThan)) __PYX_ERR(0, 746, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_GreaterThan_spec, __pyx_ptype_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_GreaterThan = &__pyx_type_7flitter_8language_4tree_GreaterThan;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_GreaterThan->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_GreaterThan->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_GreaterThan->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_GreaterThan->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_GreaterThan->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_GreaterThan, __pyx_vtabptr_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GreaterThan, (PyObject *) __pyx_ptype_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_GreaterThan) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_LessThanOrEqualTo = &__pyx_vtable_7flitter_8language_4tree_LessThanOrEqualTo;
  __pyx_vtable_7flitter_8language_4tree_LessThanOrEqualTo.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_LessThanOrEqualTo.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo__compile_op;
  __pyx_vtable_7flitter_8language_4tree_LessThanOrEqualTo.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_17LessThanOrEqualTo_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo)) __PYX_ERR(0, 754, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo_spec, __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo = &__pyx_type_7flitter_8language_4tree_LessThanOrEqualTo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo, __pyx_vtabptr_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LessThanOrEqualTo, (PyObject *) __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_GreaterThanOrEqualTo = &__pyx_vtable_7flitter_8language_4tree_GreaterThanOrEqualTo;
  __pyx_vtable_7flitter_8language_4tree_GreaterThanOrEqualTo.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Comparison;
  __pyx_vtable_7flitter_8language_4tree_GreaterThanOrEqualTo.__pyx_base.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo__compile_op;
  __pyx_vtable_7flitter_8language_4tree_GreaterThanOrEqualTo.__pyx_base.__pyx_base.op = (struct __pyx_obj_7flitter_5model_Vector *(*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_8language_4tree_20GreaterThanOrEqualTo_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo)) __PYX_ERR(0, 762, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo_spec, __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo = &__pyx_type_7flitter_8language_4tree_GreaterThanOrEqualTo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo->tp_base = __pyx_ptype_7flitter_8language_4tree_Comparison;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo, __pyx_vtabptr_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GreaterThanOrEqualTo, (PyObject *) __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_And = &__pyx_vtable_7flitter_8language_4tree_And;
  __pyx_vtable_7flitter_8language_4tree_And.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_And.__pyx_base.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_3And__compile;
  __pyx_vtable_7flitter_8language_4tree_And.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3And__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_And = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_And_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_And)) __PYX_ERR(0, 770, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_And_spec, __pyx_ptype_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_And = &__pyx_type_7flitter_8language_4tree_And;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_And->tp_base = __pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_And->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_And->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_And->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_And->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_And, __pyx_vtabptr_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_And, (PyObject *) __pyx_ptype_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_And) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Or = &__pyx_vtable_7flitter_8language_4tree_Or;
  __pyx_vtable_7flitter_8language_4tree_Or.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Or.__pyx_base.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_2Or__compile;
  __pyx_vtable_7flitter_8language_4tree_Or.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_2Or__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Or = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Or_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Or)) __PYX_ERR(0, 792, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Or_spec, __pyx_ptype_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Or = &__pyx_type_7flitter_8language_4tree_Or;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Or->tp_base = __pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Or->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Or->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Or->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Or->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Or, __pyx_vtabptr_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Or, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Or) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Xor = &__pyx_vtable_7flitter_8language_4tree_Xor;
  __pyx_vtable_7flitter_8language_4tree_Xor.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_BinaryOperation;
  __pyx_vtable_7flitter_8language_4tree_Xor.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3Xor__simplify;
  __pyx_vtable_7flitter_8language_4tree_Xor.__pyx_base._compile_op = (void (*)(struct __pyx_obj_7flitter_8language_4tree_BinaryOperation *, struct __pyx_obj_7flitter_8language_2vm_Program *))__pyx_f_7flitter_8language_4tree_3Xor__compile_op;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Xor = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Xor_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Xor)) __PYX_ERR(0, 814, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Xor_spec, __pyx_ptype_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Xor = &__pyx_type_7flitter_8language_4tree_Xor;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Xor->tp_base = __pyx_ptype_7flitter_8language_4tree_BinaryOperation;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Xor->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Xor->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Xor->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Xor->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Xor, __pyx_vtabptr_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Xor, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Xor) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Slice = &__pyx_vtable_7flitter_8language_4tree_Slice;
  __pyx_vtable_7flitter_8language_4tree_Slice.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Slice.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_5Slice__compile;
  __pyx_vtable_7flitter_8language_4tree_Slice.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_5Slice__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Slice = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Slice_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Slice)) __PYX_ERR(0, 834, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Slice_spec, __pyx_ptype_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Slice = &__pyx_type_7flitter_8language_4tree_Slice;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Slice->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Slice->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Slice->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Slice->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Slice->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Slice, __pyx_vtabptr_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Slice_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Slice) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Call = &__pyx_vtable_7flitter_8language_4tree_Call;
  __pyx_vtable_7flitter_8language_4tree_Call.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Call.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_4Call__compile;
  __pyx_vtable_7flitter_8language_4tree_Call.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_4Call__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Call = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Call_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Call)) __PYX_ERR(0, 867, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Call_spec, __pyx_ptype_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Call = &__pyx_type_7flitter_8language_4tree_Call;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Call->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Call->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Call->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Call->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Call->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Call, __pyx_vtabptr_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Call_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Call) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_NodeModifier = &__pyx_vtable_7flitter_8language_4tree_NodeModifier;
  __pyx_vtable_7flitter_8language_4tree_NodeModifier.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_NodeModifier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_NodeModifier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_NodeModifier)) __PYX_ERR(0, 975, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_NodeModifier_spec, __pyx_ptype_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_NodeModifier = &__pyx_type_7flitter_8language_4tree_NodeModifier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_NodeModifier->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_NodeModifier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_NodeModifier->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_NodeModifier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_NodeModifier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_NodeModifier, __pyx_vtabptr_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NodeModifier, (PyObject *) __pyx_ptype_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_NodeModifier) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Tag = &__pyx_vtable_7flitter_8language_4tree_Tag;
  __pyx_vtable_7flitter_8language_4tree_Tag.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_NodeModifier;
  __pyx_vtable_7flitter_8language_4tree_Tag.__pyx_base.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_3Tag__compile;
  __pyx_vtable_7flitter_8language_4tree_Tag.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3Tag__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Tag = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Tag_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Tag)) __PYX_ERR(0, 979, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Tag_spec, __pyx_ptype_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Tag = &__pyx_type_7flitter_8language_4tree_Tag;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Tag->tp_base = __pyx_ptype_7flitter_8language_4tree_NodeModifier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Tag->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Tag->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Tag->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Tag->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Tag, __pyx_vtabptr_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Tag_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Tag) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Attributes = &__pyx_vtable_7flitter_8language_4tree_Attributes;
  __pyx_vtable_7flitter_8language_4tree_Attributes.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_NodeModifier;
  __pyx_vtable_7flitter_8language_4tree_Attributes.__pyx_base.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_10Attributes__compile;
  __pyx_vtable_7flitter_8language_4tree_Attributes.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_10Attributes__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Attributes = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Attributes_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Attributes)) __PYX_ERR(0, 1018, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Attributes_spec, __pyx_ptype_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Attributes = &__pyx_type_7flitter_8language_4tree_Attributes;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Attributes->tp_base = __pyx_ptype_7flitter_8language_4tree_NodeModifier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Attributes->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Attributes->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Attributes->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Attributes->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Attributes, __pyx_vtabptr_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Attributes, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Attributes) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Append = &__pyx_vtable_7flitter_8language_4tree_Append;
  __pyx_vtable_7flitter_8language_4tree_Append.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_NodeModifier;
  __pyx_vtable_7flitter_8language_4tree_Append.__pyx_base.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_6Append__compile;
  __pyx_vtable_7flitter_8language_4tree_Append.__pyx_base.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_6Append__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Append = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Append_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Append)) __PYX_ERR(0, 1079, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Append_spec, __pyx_ptype_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Append = &__pyx_type_7flitter_8language_4tree_Append;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Append->tp_base = __pyx_ptype_7flitter_8language_4tree_NodeModifier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Append->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Append->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Append->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Append->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Append, __pyx_vtabptr_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Append, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Append) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_4tree_Binding = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Binding_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Binding)) __PYX_ERR(0, 1135, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Binding_spec, __pyx_ptype_7flitter_8language_4tree_Binding) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Binding = &__pyx_type_7flitter_8language_4tree_Binding;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Binding) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Binding->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Binding->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Binding->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Binding->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Binding_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Binding) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Binding) < 0) __PYX_ERR(0, 1135, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_4tree_PolyBinding = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_PolyBinding_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_4tree_PolyBinding)) __PYX_ERR(0, 1147, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_PolyBinding_spec, __pyx_ptype_7flitter_8language_4tree_PolyBinding) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_PolyBinding = &__pyx_type_7flitter_8language_4tree_PolyBinding;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_PolyBinding) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_PolyBinding->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_PolyBinding->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_PolyBinding->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_PolyBinding->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PolyBinding_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_PolyBinding) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_PolyBinding) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Let = &__pyx_vtable_7flitter_8language_4tree_Let;
  __pyx_vtable_7flitter_8language_4tree_Let.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Let.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_3Let__compile;
  __pyx_vtable_7flitter_8language_4tree_Let.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3Let__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Let = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Let_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Let)) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Let_spec, __pyx_ptype_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Let = &__pyx_type_7flitter_8language_4tree_Let;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Let->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Let->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Let->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Let->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Let->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Let, __pyx_vtabptr_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Let_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Let) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_For = &__pyx_vtable_7flitter_8language_4tree_For;
  __pyx_vtable_7flitter_8language_4tree_For.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_For.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_3For__compile;
  __pyx_vtable_7flitter_8language_4tree_For.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_3For__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_For = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_For_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_For)) __PYX_ERR(0, 1275, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_For_spec, __pyx_ptype_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_For = &__pyx_type_7flitter_8language_4tree_For;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_For->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_For->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_For->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_For->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_For->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_For, __pyx_vtabptr_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_For_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_For) < 0) __PYX_ERR(0, 1275, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_4tree_IfCondition = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_IfCondition_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_4tree_IfCondition)) __PYX_ERR(0, 1335, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_IfCondition_spec, __pyx_ptype_7flitter_8language_4tree_IfCondition) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_IfCondition = &__pyx_type_7flitter_8language_4tree_IfCondition;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_IfCondition) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_IfCondition->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_IfCondition->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_IfCondition->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_IfCondition->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IfCondition_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_IfCondition) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_IfCondition) < 0) __PYX_ERR(0, 1335, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_IfElse = &__pyx_vtable_7flitter_8language_4tree_IfElse;
  __pyx_vtable_7flitter_8language_4tree_IfElse.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_IfElse.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_6IfElse__compile;
  __pyx_vtable_7flitter_8language_4tree_IfElse.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_6IfElse__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_IfElse = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_IfElse_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_IfElse)) __PYX_ERR(0, 1347, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_IfElse_spec, __pyx_ptype_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_IfElse = &__pyx_type_7flitter_8language_4tree_IfElse;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_IfElse->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_IfElse->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_IfElse->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_IfElse->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_IfElse->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_IfElse, __pyx_vtabptr_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IfElse_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_IfElse) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_8language_4tree_Function = &__pyx_vtable_7flitter_8language_4tree_Function;
  __pyx_vtable_7flitter_8language_4tree_Function.__pyx_base = *__pyx_vtabptr_7flitter_8language_4tree_Expression;
  __pyx_vtable_7flitter_8language_4tree_Function.__pyx_base._compile = (void (*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_8language_2vm_Program *, PyObject *))__pyx_f_7flitter_8language_4tree_8Function__compile;
  __pyx_vtable_7flitter_8language_4tree_Function.__pyx_base._simplify = (struct __pyx_obj_7flitter_8language_4tree_Expression *(*)(struct __pyx_obj_7flitter_8language_4tree_Expression *, struct __pyx_obj_7flitter_5model_Context *))__pyx_f_7flitter_8language_4tree_8Function__simplify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_4tree_Function = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree_Function_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_8language_4tree_Function)) __PYX_ERR(0, 1403, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree_Function_spec, __pyx_ptype_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree_Function = &__pyx_type_7flitter_8language_4tree_Function;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_8language_4tree_Function->tp_base = __pyx_ptype_7flitter_8language_4tree_Expression;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree_Function->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree_Function->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree_Function->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree_Function->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_8language_4tree_Function, __pyx_vtabptr_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Function_2, (PyObject *) __pyx_ptype_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_8language_4tree_Function) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr)) __PYX_ERR(0, 221, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr_spec, __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr = &__pyx_type_7flitter_8language_4tree___pyx_scope_struct__genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr->tp_dictoffset && __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_8language_4tree___pyx_scope_struct__genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_5model_Vector = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Vector", sizeof(struct __pyx_obj_7flitter_5model_Vector), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Vector),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Vector) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Vector = (struct __pyx_vtabstruct_7flitter_5model_Vector*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Vector); if (unlikely(!__pyx_vtabptr_7flitter_5model_Vector)) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix33 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix33", sizeof(struct __pyx_obj_7flitter_5model_Matrix33), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix33),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix33) __PYX_ERR(3, 107, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix33 = (struct __pyx_vtabstruct_7flitter_5model_Matrix33*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix33); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix33)) __PYX_ERR(3, 107, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix44 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix44", sizeof(struct __pyx_obj_7flitter_5model_Matrix44), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix44),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix44) __PYX_ERR(3, 128, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix44 = (struct __pyx_vtabstruct_7flitter_5model_Matrix44*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix44); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix44)) __PYX_ERR(3, 128, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Quaternion = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Quaternion", sizeof(struct __pyx_obj_7flitter_5model_Quaternion), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Quaternion),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Quaternion) __PYX_ERR(3, 178, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Quaternion = (struct __pyx_vtabstruct_7flitter_5model_Quaternion*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Quaternion); if (unlikely(!__pyx_vtabptr_7flitter_5model_Quaternion)) __PYX_ERR(3, 178, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Node = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Node", sizeof(struct __pyx_obj_7flitter_5model_Node), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Node),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Node) __PYX_ERR(3, 198, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Node = (struct __pyx_vtabstruct_7flitter_5model_Node*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Node); if (unlikely(!__pyx_vtabptr_7flitter_5model_Node)) __PYX_ERR(3, 198, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_StateDict = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "StateDict", sizeof(struct __pyx_obj_7flitter_5model_StateDict), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_StateDict),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_StateDict) __PYX_ERR(3, 222, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_StateDict = (struct __pyx_vtabstruct_7flitter_5model_StateDict*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_StateDict); if (unlikely(!__pyx_vtabptr_7flitter_5model_StateDict)) __PYX_ERR(3, 222, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Context = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Context", sizeof(struct __pyx_obj_7flitter_5model_Context), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Context),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Context) __PYX_ERR(3, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flitter.language.vm"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_2vm_VectorStack = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "VectorStack", sizeof(struct __pyx_obj_7flitter_8language_2vm_VectorStack), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_VectorStack),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_VectorStack) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_vtabptr_7flitter_8language_2vm_VectorStack = (struct __pyx_vtabstruct_7flitter_8language_2vm_VectorStack*)__Pyx_GetVtable(__pyx_ptype_7flitter_8language_2vm_VectorStack); if (unlikely(!__pyx_vtabptr_7flitter_8language_2vm_VectorStack)) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_Instruction = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "Instruction", sizeof(struct __pyx_obj_7flitter_8language_2vm_Instruction), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_Instruction),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_Instruction) __PYX_ERR(7, 90, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionVector = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionVector", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionVector), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionVector),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionVector) __PYX_ERR(7, 94, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionNode = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionNode", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionNode), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionNode),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionNode) __PYX_ERR(7, 98, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionStr = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionStr", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionStr), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionStr),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionStr) __PYX_ERR(7, 102, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionTuple = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionTuple", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionTuple),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionTuple) __PYX_ERR(7, 106, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionInt = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionInt", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionInt), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionInt),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionInt) __PYX_ERR(7, 110, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionIntTuple = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionIntTuple", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionIntTuple),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionIntTuple) __PYX_ERR(7, 114, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionLabel = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionLabel", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionLabel),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionLabel) __PYX_ERR(7, 119, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionJump = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionJump", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionJump), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionJump),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionJump) __PYX_ERR(7, 123, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionFunc = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionFunc", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionFunc),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionFunc) __PYX_ERR(7, 128, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_InstructionObjectInt = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "InstructionObjectInt", sizeof(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_InstructionObjectInt),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_InstructionObjectInt) __PYX_ERR(7, 134, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_2vm_Program = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.vm", "Program", sizeof(struct __pyx_obj_7flitter_8language_2vm_Program), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_2vm_Program),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_2vm_Program) __PYX_ERR(7, 139, __pyx_L1_error)
  __pyx_vtabptr_7flitter_8language_2vm_Program = (struct __pyx_vtabstruct_7flitter_8language_2vm_Program*)__Pyx_GetVtable(__pyx_ptype_7flitter_8language_2vm_Program); if (unlikely(!__pyx_vtabptr_7flitter_8language_2vm_Program)) __PYX_ERR(7, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "HASH_START", (void **)&__pyx_vp_7flitter_5model_HASH_START, "uint64_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "null_", (void **)&__pyx_vp_7flitter_5model_null_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "true_", (void **)&__pyx_vp_7flitter_5model_true_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "false_", (void **)&__pyx_vp_7flitter_5model_false_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "minusone_", (void **)&__pyx_vp_7flitter_5model_minusone_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "inf_", (void **)&__pyx_vp_7flitter_5model_inf_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "nan_", (void **)&__pyx_vp_7flitter_5model_nan_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flitter.language.vm"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "static_builtins", (void **)&__pyx_vp_7flitter_8language_2vm_static_builtins, "PyObject *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "dynamic_builtins", (void **)&__pyx_vp_7flitter_8language_2vm_dynamic_builtins, "PyObject *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "builtins", (void **)&__pyx_vp_7flitter_8language_2vm_builtins, "PyObject *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_tree(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_tree},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "tree",
      __pyx_k_Language_abstract_syntax_tree_T, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC inittree(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC inittree(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_tree(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_tree(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_tree(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'tree' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("tree", __pyx_methods, __pyx_k_Language_abstract_syntax_tree_T, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 2, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "tree" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 2, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 2, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_tree(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 2, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flitter__language__tree) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 2, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flitter.language.tree")) {
      if (unlikely((PyDict_SetItemString(modules, "flitter.language.tree", __pyx_m) < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif

  /* "flitter/language/tree.pyx":10
 * """
 * 
 * from loguru import logger             # <<<<<<<<<<<<<<
 * 
 * from libc.stdint cimport int64_t
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(0, 10, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_loguru, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":14
 * from libc.stdint cimport int64_t
 * 
 * from .. import name_patch             # <<<<<<<<<<<<<<
 * from ..cache import SharedCache
 * from ..model cimport Vector, Node, Context, StateDict, null_, true_, false_, minusone_
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_name_patch);
  __Pyx_GIVEREF(__pyx_n_s_name_patch);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_name_patch)) __PYX_ERR(0, 14, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s__32, __pyx_t_3, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_name_patch, __pyx_t_3) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/language/tree.pyx":15
 * 
 * from .. import name_patch
 * from ..cache import SharedCache             # <<<<<<<<<<<<<<
 * from ..model cimport Vector, Node, Context, StateDict, null_, true_, false_, minusone_
 * from .vm cimport Program, Instruction, InstructionInt, InstructionVector, OpCode, static_builtins, dynamic_builtins
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_SharedCache);
  __Pyx_GIVEREF(__pyx_n_s_SharedCache);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_SharedCache)) __PYX_ERR(0, 15, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_cache, __pyx_t_2, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_SharedCache); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SharedCache, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/language/tree.pyx":20
 * 
 * 
 * logger = name_patch(logger, __name__)             # <<<<<<<<<<<<<<
 * 
 * cdef Literal NoOp = Literal(null_)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_4) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":22
 * logger = name_patch(logger, __name__)
 * 
 * cdef Literal NoOp = Literal(null_)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal), ((PyObject *)__pyx_v_7flitter_5model_null_)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_8language_4tree_NoOp);
  __Pyx_DECREF_SET(__pyx_v_7flitter_8language_4tree_NoOp, ((struct __pyx_obj_7flitter_8language_4tree_Literal *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":50
 * 
 * cdef class Expression:
 *     def compile(self, tuple initial_lnames=(), set initial_errors=None, bint log_errors=True):             # <<<<<<<<<<<<<<
 *         cdef Program program = Program.__new__(Program, initial_lnames)
 *         if initial_errors:
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Expression_1compile, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Expression_compile, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__35);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression, __pyx_n_s_compile, __pyx_t_4) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Expression);

  /* "flitter/language/tree.pyx":61
 *         return program
 * 
 *     def simplify(self, StateDict state=None, dict static=None, dynamic=None, Context parent=None, path=None, bint return_context=False):             # <<<<<<<<<<<<<<
 *         cdef dict context_vars = {}
 *         cdef str key
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Expression_3simplify, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Expression_simplify, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__38);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression, __pyx_n_s_simplify, __pyx_t_4) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Expression);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Expression_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Expression___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Expression);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Expression, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Expression__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Expression_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Expression___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Expression, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Expression);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Top_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Top___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Top, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Top);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Top, (type(self), 0xb892498, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Top__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Top_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Top___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Top, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Top);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Export_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Export___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Export, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Export);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Export, (type(self), 0x123fe55, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Export__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Export_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Export___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Export, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Export);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Import_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Import___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Import, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Import);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Import, (type(self), 0x3ebdbb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Import__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Import_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Import___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Import, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Import);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Sequence_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Sequence___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Sequence);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Sequence, (type(self), 0x18bc300, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Sequence__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Sequence_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Sequence___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Sequence, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Sequence);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7Literal_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Literal___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Literal);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Literal, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Literal__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7Literal_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Literal___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Literal, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Literal);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Name_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Name___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Name, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Name);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Name, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Name__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Name_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Name___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Name, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Name);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Lookup_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Lookup___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Lookup, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Lookup);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Lookup, (type(self), 0x2c70e12, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Lookup__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Lookup_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Lookup___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Lookup, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Lookup);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Range_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Range___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Range, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Range);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Range, (type(self), 0xfbe38c8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Range__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Range_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Range___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Range, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Range);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_14UnaryOperation_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnaryOperation___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_UnaryOperation);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnaryOperation, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnaryOperation__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_14UnaryOperation_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnaryOperation___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_UnaryOperation, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_UnaryOperation);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Negative_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Negative___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Negative);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Negative, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Negative__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Negative_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Negative___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Negative, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Negative);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Positive_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Positive___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Positive);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Positive, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Positive__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Positive_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Positive___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Positive, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Positive);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Not_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Not___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Not, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Not);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Not, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Not__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Not_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Not___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Not, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Not);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Ceil_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Ceil___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Ceil, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Ceil);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Ceil, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Ceil__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Ceil_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Ceil___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Ceil, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Ceil);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Floor_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Floor___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Floor);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Floor, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Floor__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Floor_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Floor___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Floor, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Floor);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Fract_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Fract___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Fract);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Fract, (type(self), 0x60ba1dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Fract__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Fract_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Fract___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Fract, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Fract);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_15BinaryOperation_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BinaryOperation___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_BinaryOperation);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryOperation__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_15BinaryOperation_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BinaryOperation___setstate_cytho, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_BinaryOperation, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_BinaryOperation);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_20MathsBinaryOperation_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MathsBinaryOperation___reduce_cy, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MathsBinaryOperation, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MathsBinaryOperation__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_20MathsBinaryOperation_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MathsBinaryOperation___setstate, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_MathsBinaryOperation);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Add_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Add___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Add);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Add, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Add__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Add_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Add___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Add, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Add);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Subtract_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Subtract___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Subtract, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Subtract);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Subtract, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Subtract__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Subtract_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Subtract___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Subtract, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Subtract);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Multiply_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Multiply___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Multiply);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Multiply, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Multiply__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Multiply_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Multiply___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Multiply, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Multiply);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Divide_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Divide___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Divide, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Divide);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Divide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Divide__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Divide_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Divide___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Divide, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Divide);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11FloorDivide_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloorDivide___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_FloorDivide, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_FloorDivide);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloorDivide, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloorDivide__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11FloorDivide_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloorDivide___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_FloorDivide, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_FloorDivide);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Modulo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Modulo___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Modulo, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Modulo);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Modulo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Modulo__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Modulo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Modulo___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Modulo, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Modulo);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Power_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Power___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Power, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Power);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Power, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Power__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Power_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Power___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Power, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Power);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Comparison_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Comparison___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Comparison);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Comparison, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Comparison__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Comparison_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Comparison___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Comparison, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Comparison);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7EqualTo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EqualTo___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_EqualTo, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_EqualTo);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EqualTo__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7EqualTo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EqualTo___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_EqualTo, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_EqualTo);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10NotEqualTo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NotEqualTo___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_NotEqualTo, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_NotEqualTo);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NotEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NotEqualTo__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10NotEqualTo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NotEqualTo___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_NotEqualTo, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_NotEqualTo);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8LessThan_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LessThan___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThan, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_LessThan);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThan__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8LessThan_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LessThan___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThan, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_LessThan);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11GreaterThan_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GreaterThan___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThan, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_GreaterThan);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThan, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThan__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11GreaterThan_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GreaterThan___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThan, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_GreaterThan);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_17LessThanOrEqualTo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LessThanOrEqualTo___reduce_cytho, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LessThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LessThanOrEqualTo__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_17LessThanOrEqualTo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LessThanOrEqualTo___setstate_cyt, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_LessThanOrEqualTo);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_20GreaterThanOrEqualTo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GreaterThanOrEqualTo___reduce_cy, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GreaterThanOrEqualTo, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GreaterThanOrEqualTo__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_20GreaterThanOrEqualTo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GreaterThanOrEqualTo___setstate, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_GreaterThanOrEqualTo);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3And_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_And___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_And, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_And);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_And, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_And__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3And_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_And___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_And, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_And);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_2Or_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Or___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Or, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Or);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Or, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Or__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_2Or_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Or___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Or, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Or);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Xor_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Xor___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Xor, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Xor);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Xor, (type(self), 0x132bc34, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Xor__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Xor_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Xor___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Xor, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Xor);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Slice_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Slice___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Slice, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Slice);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Slice, (type(self), 0x3b1aecc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Slice__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5Slice_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Slice___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Slice, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Slice);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Call_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Call___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Call, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Call);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Call, (type(self), 0x789d054, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Call__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_4Call_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Call___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Call, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Call);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_12NodeModifier_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NodeModifier___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_NodeModifier);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NodeModifier, (type(self), 0x545ea53, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NodeModifier__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_12NodeModifier_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NodeModifier___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_NodeModifier, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_NodeModifier);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Tag_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Tag___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Tag, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Tag);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Tag, (type(self), 0x9721bb0, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Tag__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Tag_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Tag___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Tag, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Tag);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Attributes_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Attributes___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__119)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Attributes, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Attributes);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Attributes, (type(self), 0x552d50b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Attributes__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_10Attributes_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Attributes___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Attributes, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Attributes);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Append_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Append___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Append);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Append, (type(self), 0xcfaea2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Append__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6Append_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Append___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Append, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Append);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7Binding_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Binding___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Binding);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Binding, (type(self), 0x3a2bfe3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Binding__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7Binding_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Binding___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Binding, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Binding);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11PolyBinding_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PolyBinding___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_PolyBinding);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PolyBinding, (type(self), 0x12bf31f, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PolyBinding__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11PolyBinding_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PolyBinding___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_PolyBinding, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_PolyBinding);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Let_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Let___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Let);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Let, (type(self), 0xe222284, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Let__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3Let_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Let___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Let, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Let);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3For_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_For___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__129)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_For, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_For);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_For, (type(self), 0x3d9b243, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_For__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3For_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_For___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_For, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_For);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11IfCondition_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IfCondition___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__131)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfCondition, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_IfCondition);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfCondition, (type(self), 0x9a45a72, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfCondition__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11IfCondition_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IfCondition___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfCondition, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_IfCondition);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6IfElse_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IfElse___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__133)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfElse, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_IfElse);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IfElse, (type(self), 0xdcd679c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IfElse__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_6IfElse_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IfElse___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_IfElse, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_IfElse);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Function_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Function___reduce_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Function, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Function);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Function, (type(self), 0xf4b5d07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Function__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_8Function_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Function___setstate_cython, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_8language_4tree_Function, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_8language_4tree_Function);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Expression(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_1__pyx_unpickle_Expression, 0, __pyx_n_s_pyx_unpickle_Expression, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Expression, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Expression__set_state(<Expression> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Expression__set_state(Expression __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_3__pyx_unpickle_Top, 0, __pyx_n_s_pyx_unpickle_Top, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Top, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Export(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_5__pyx_unpickle_Export, 0, __pyx_n_s_pyx_unpickle_Export, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Export, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Export__set_state(<Export> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Export__set_state(Export __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.static_exports = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_7__pyx_unpickle_Import, 0, __pyx_n_s_pyx_unpickle_Import, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Import, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Sequence(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_9__pyx_unpickle_Sequence, 0, __pyx_n_s_pyx_unpickle_Sequence, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Sequence, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Sequence__set_state(<Sequence> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Sequence__set_state(Sequence __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expressions = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_11__pyx_unpickle_Literal, 0, __pyx_n_s_pyx_unpickle_Literal, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__143)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Literal, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Name(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_13__pyx_unpickle_Name, 0, __pyx_n_s_pyx_unpickle_Name, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Name, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Name__set_state(<Name> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Name__set_state(Name __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_15__pyx_unpickle_Lookup, 0, __pyx_n_s_pyx_unpickle_Lookup, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__145)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Lookup, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Range(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_17__pyx_unpickle_Range, 0, __pyx_n_s_pyx_unpickle_Range, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Range, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Range__set_state(<Range> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Range__set_state(Range __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.start = __pyx_state[0]; __pyx_result.step = __pyx_state[1]; __pyx_result.stop = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_19__pyx_unpickle_UnaryOperation, 0, __pyx_n_s_pyx_unpickle_UnaryOperation, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UnaryOperation, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Negative(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_21__pyx_unpickle_Negative, 0, __pyx_n_s_pyx_unpickle_Negative, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Negative, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Negative__set_state(<Negative> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Negative__set_state(Negative __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_23__pyx_unpickle_Positive, 0, __pyx_n_s_pyx_unpickle_Positive, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__149)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Positive, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Not(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_25__pyx_unpickle_Not, 0, __pyx_n_s_pyx_unpickle_Not, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Not, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Not__set_state(<Not> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Not__set_state(Not __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_27__pyx_unpickle_Ceil, 0, __pyx_n_s_pyx_unpickle_Ceil, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Ceil, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Floor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_29__pyx_unpickle_Floor, 0, __pyx_n_s_pyx_unpickle_Floor, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Floor, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Floor__set_state(<Floor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Floor__set_state(Floor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_31__pyx_unpickle_Fract, 0, __pyx_n_s_pyx_unpickle_Fract, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Fract, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BinaryOperation(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_33__pyx_unpickle_BinaryOperation, 0, __pyx_n_s_pyx_unpickle_BinaryOperation, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BinaryOperation, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_BinaryOperation__set_state(<BinaryOperation> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryOperation__set_state(BinaryOperation __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_35__pyx_unpickle_MathsBinaryOperation, 0, __pyx_n_s_pyx_unpickle_MathsBinaryOperat, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MathsBinaryOperat, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Add(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_37__pyx_unpickle_Add, 0, __pyx_n_s_pyx_unpickle_Add, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Add, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Add__set_state(<Add> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Add__set_state(Add __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_39__pyx_unpickle_Subtract, 0, __pyx_n_s_pyx_unpickle_Subtract, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__157)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Subtract, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Multiply(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_41__pyx_unpickle_Multiply, 0, __pyx_n_s_pyx_unpickle_Multiply, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Multiply, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Multiply__set_state(<Multiply> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Multiply__set_state(Multiply __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_43__pyx_unpickle_Divide, 0, __pyx_n_s_pyx_unpickle_Divide, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__159)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Divide, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FloorDivide(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_45__pyx_unpickle_FloorDivide, 0, __pyx_n_s_pyx_unpickle_FloorDivide, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FloorDivide, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_FloorDivide__set_state(<FloorDivide> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FloorDivide__set_state(FloorDivide __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_47__pyx_unpickle_Modulo, 0, __pyx_n_s_pyx_unpickle_Modulo, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__161)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Modulo, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Power(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_49__pyx_unpickle_Power, 0, __pyx_n_s_pyx_unpickle_Power, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Power, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Power__set_state(<Power> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Power__set_state(Power __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_51__pyx_unpickle_Comparison, 0, __pyx_n_s_pyx_unpickle_Comparison, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__163)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Comparison, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_53__pyx_unpickle_EqualTo, 0, __pyx_n_s_pyx_unpickle_EqualTo, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_EqualTo, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_EqualTo__set_state(<EqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EqualTo__set_state(EqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_55__pyx_unpickle_NotEqualTo, 0, __pyx_n_s_pyx_unpickle_NotEqualTo, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_NotEqualTo, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LessThan(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_57__pyx_unpickle_LessThan, 0, __pyx_n_s_pyx_unpickle_LessThan, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LessThan, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_LessThan__set_state(<LessThan> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThan__set_state(LessThan __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_59__pyx_unpickle_GreaterThan, 0, __pyx_n_s_pyx_unpickle_GreaterThan, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__167)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GreaterThan, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LessThanOrEqualTo(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_61__pyx_unpickle_LessThanOrEqualTo, 0, __pyx_n_s_pyx_unpickle_LessThanOrEqualTo, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LessThanOrEqualTo, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_LessThanOrEqualTo__set_state(<LessThanOrEqualTo> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LessThanOrEqualTo__set_state(LessThanOrEqualTo __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_63__pyx_unpickle_GreaterThanOrEqualTo, 0, __pyx_n_s_pyx_unpickle_GreaterThanOrEqua, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__169)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GreaterThanOrEqua, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_And(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_65__pyx_unpickle_And, 0, __pyx_n_s_pyx_unpickle_And, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_And, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_And__set_state(<And> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_And__set_state(And __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_67__pyx_unpickle_Or, 0, __pyx_n_s_pyx_unpickle_Or, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__171)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Or, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Xor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_69__pyx_unpickle_Xor, 0, __pyx_n_s_pyx_unpickle_Xor, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Xor, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Xor__set_state(<Xor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Xor__set_state(Xor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.left = __pyx_state[0]; __pyx_result.right = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_71__pyx_unpickle_Slice, 0, __pyx_n_s_pyx_unpickle_Slice, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__173)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Slice, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Call(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_73__pyx_unpickle_Call, 0, __pyx_n_s_pyx_unpickle_Call, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Call, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Call__set_state(<Call> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Call__set_state(Call __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.function = __pyx_state[1]; __pyx_result.keyword_args = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_75__pyx_unpickle_NodeModifier, 0, __pyx_n_s_pyx_unpickle_NodeModifier, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__175)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_NodeModifier, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Tag(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_77__pyx_unpickle_Tag, 0, __pyx_n_s_pyx_unpickle_Tag, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Tag, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Tag__set_state(<Tag> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Tag__set_state(Tag __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.node = __pyx_state[0]; __pyx_result.tag = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_79__pyx_unpickle_Attributes, 0, __pyx_n_s_pyx_unpickle_Attributes, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__177)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Attributes, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Append(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_81__pyx_unpickle_Append, 0, __pyx_n_s_pyx_unpickle_Append, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Append, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Append__set_state(<Append> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Append__set_state(Append __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.node = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_83__pyx_unpickle_Binding, 0, __pyx_n_s_pyx_unpickle_Binding, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__179)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Binding, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PolyBinding(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_85__pyx_unpickle_PolyBinding, 0, __pyx_n_s_pyx_unpickle_PolyBinding, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PolyBinding, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PolyBinding__set_state(<PolyBinding> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PolyBinding__set_state(PolyBinding __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.expr = __pyx_state[0]; __pyx_result.names = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_87__pyx_unpickle_Let, 0, __pyx_n_s_pyx_unpickle_Let, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__181)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Let, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_For(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_89__pyx_unpickle_For, 0, __pyx_n_s_pyx_unpickle_For, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_For, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_For__set_state(<For> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_For__set_state(For __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.body = __pyx_state[0]; __pyx_result.names = __pyx_state[1]; __pyx_result.source = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_91__pyx_unpickle_IfCondition, 0, __pyx_n_s_pyx_unpickle_IfCondition, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__183)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IfCondition, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IfElse(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_93__pyx_unpickle_IfElse, 0, __pyx_n_s_pyx_unpickle_IfElse, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IfElse, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_IfElse__set_state(<IfElse> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IfElse__set_state(IfElse __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.else_ = __pyx_state[0]; __pyx_result.tests = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_8language_4tree_95__pyx_unpickle_Function, 0, __pyx_n_s_pyx_unpickle_Function, NULL, __pyx_n_s_flitter_language_tree, __pyx_d, ((PyObject *)__pyx_codeobj__185)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Function, __pyx_t_4) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/language/tree.pyx":2
 * 
 * """             # <<<<<<<<<<<<<<
 * Language abstract syntax tree
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init flitter.language.tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flitter.language.tree");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (__Pyx_IS_TYPE(L, &PySet_Type)) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (length <= 0) {
        return PyList_New(0);
    }
    return __Pyx_PyList_FromArray(((PyListObject*)src)->ob_item + start, length);
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    return __Pyx_PyTuple_FromArray(((PyTupleObject*)src)->ob_item + start, length);
}
#endif

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_discard */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    return found;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__8);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__186);
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_11
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_11
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
